# $Id$
########################################################################################################################
#
#     73_km200.pm
#     Creates the possibility to access the Buderus central heating system via
#     Buderus KM200, KM100 or KM50 communication module. It uses HttpUtils_NonblockingGet
#     from Rudolf Koenig to avoid a full blockage of the fhem main system during the
#     polling procedure.
#
#     Author                     : Matthias Deeke 
#     Contributions              : Olaf Droegehorn, Andreas Hahn, Rudolf Koenig, Markus Bloch,
#     Contributions (continued)  : Stefan M., Furban, KaiKr, grossi33, Morkin, DLindner
#     e-mail                     : matthias.deeke(AT)deeke(PUNKT)eu
#     Fhem Forum                 : http://forum.fhem.de/index.php/topic,25540.0.html
#     Fhem Wiki                  : http://www.fhemwiki.de/wiki/Buderus_Web_Gateway
#
#     This file is part of fhem.
#
#     Fhem is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 2 of the License, or
#     (at your option) any later version.
#
#     Fhem is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with fhem.  If not, see <http://www.gnu.org/licenses/>.
#
#     fhem.cfg: define <devicename> km200 <IPv4-address> <GatewayPassword> <PrivatePassword>
#
#     Example 1 - Bare Passwords:
#     define myKm200 km200 192.168.178.200 GatewayGeheim        PrivateGeheim
#
#     Example 2 - base64 encoded passwords: Both passwords may be pre-encode with base64
#     define myKm200 km200 192.168.178.200 R2F0ZXdheUdlaGVpbQ== UHJpdmF0ZUdlaGVpbQ==
#
########################################################################################################################

########################################################################################################################
# List of open Problems:
#
#
########################################################################################################################

package main;

use strict;
use warnings;
use Blocking;
use FHEM::Meta;
use Time::HiRes qw(gettimeofday sleep usleep);
use Digest::MD5 qw(md5 md5_hex md5_base64);
use base qw( Exporter );
use List::MoreUtils qw(first_index);
use MIME::Base64;
use LWP::UserAgent;
use JSON;
use Crypt::Rijndael;
use HttpUtils;
use Encode;
use Text::Wrap;
use constant false => 0;
use constant true  => 1;

sub km200_Define($$);
sub km200_Undefine($$);
sub km200_GetErrorMessage($$$$);

###START###### Initialize module ##############################################################################START####
sub km200_Initialize($)
{
    my ($hash)  = @_;

    $hash->{STATE}           = "Init";
    $hash->{DefFn}           = "km200_Define";
    $hash->{UndefFn}         = "km200_Undefine";
    $hash->{SetFn}           = "km200_Set";
    $hash->{GetFn}           = "km200_Get";
    $hash->{AttrFn}          = "km200_Attr";
	$hash->{DbLog_splitFn}   = "km200_DbLog_splitFn";

    $hash->{AttrList}        = "do_not_notify:1,0 " .
							   "header " .
							   "disable:1,0 " .
						       "loglevel:0,1,2,3,4,5,6 " .
						       "IntervalDynVal " .
						       "PollingTimeout " .
							   "DoNotPoll " .
							   "ReadBackDelay " .
							   "ReadingTextLen " .
						       $readingFnAttributes;
	return FHEM::Meta::InitMod( __FILE__, $hash );
}
####END####### Initialize module ###############################################################################END#####


###START######  Activate module after module has been used via fhem command "define" ##########################START####
sub km200_Define($$)
{
	my ($hash, $def)            = @_;
	my @a						= split("[ \t][ \t]*", $def);
	my $name					= $a[0];
								 #$a[1] just contains the "km200" module name and we already know that! :-)
	my $url						= $a[2];
	my $km200_gateway_password	= $a[3];
	my $km200_private_password	= $a[4];

	$hash->{NAME}				= $name;
	$hash->{STATE}              = "define";

	Log3 $name, 4, $name. " : km200 - Starting to define module";

	### Stop the current timer if one exists errornous 
	RemoveInternalTimer($hash);
	Log3 $name, 4, $name. " : km200 - InternalTimer has been removed.";

	
	###START###### Define known services of gateway ###########################################################START####
	my @KM200_AllServices = (
	"/",
	"/dhwCircuits",
	"/gateway",
	"/heatingCircuits",
	"/heatSources",
	"/notifications",
	"/recordings",
	"/solarCircuits",
	"/system",
	"/dhwCircuits",
	);
	####END####### Define known services of gateway ############################################################END#####
 
	
    ###START### Check whether all variables are available #####################################################START####
	if (int(@a) == 5) 
	{
		###START### Check whether IPv4 address is valid
		if ($url =~ m/^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$/)
		{
			Log3 $name, 4, $name. " : km200 - IPv4-address is valid                  : " . $url;
		}
		else
		{
			return $name .": Error - IPv4 address is not valid \n Please use \"define <devicename> km200 <IPv4-address> <interval/[s]> <GatewayPassword> <PrivatePassword>\" instead";
		}
		####END#### Check whether IPv4 address is valid	

		
		###START### Check whether gateway password is base64 encoded or bare, has the right length and delete "-" if required
		my $PasswordEncrypted = false;
		my $EvalPassWord = $km200_gateway_password;
		$EvalPassWord =~ tr/-//d;

		if ( length($EvalPassWord) == 16)
		{
			$km200_gateway_password = $EvalPassWord;
			Log3 $name,4, $name. " : km200 - Provided GatewayPassword provided as bareword has the correct length at least.";		
		}
		else # Check whether the password is eventually base64 encoded 
		{
		    # Remove additional encoding with base64
			my $decryptData = decode_base64($km200_gateway_password);
			$decryptData =~ tr/-//d;
			$decryptData =~ s/\r|\n//g;
			if ( length($decryptData) == 16)
			{
				$km200_gateway_password = $decryptData;
				$PasswordEncrypted = true;
				Log3 $name, 4, $name. " : km200 - Provided GatewayPassword encoded with base64 has the correct length at least.";
			}
			else
			{
				return $name .": Error - GatewayPassword does not have the correct length.\n". 
							  "          Please enter gateway password in the format of \"aaaabbbbccccdddd\" or \"aaaa-bbbb-cccc-dddd\"\n". 
							  "          You may encode your password with base64 first, in order to prevent bare passwords in fhem.cfg.\n".
							  "          If you choose to encrypt your gateway password with base64, you also must encrypt your private password the same way\n"; 

				Log3 $name, 3, $name. " : km200 - Provided Gateway Password does not follow the specifications";
			}
		}
		####END#### Check whether gateway password has the right length and delete "-" if required

		###START### Check whether private password is available and decode it with base64 if encoding is used
		if ($PasswordEncrypted == true)
		{
			my $decryptData = decode_base64($km200_private_password);
			$decryptData =~ s/\r|\n//g;
			if (length($decryptData) > 0)
			{
				$km200_private_password = $decryptData;
				Log3 $name, 4, $name. " : km200 - Provided PrivatePassword exists at least";
			}
			else
			{
				return $name .": Error - PrivatePassword does not have the minimum length.\n".
							  "          You may encode your password with base64 first, in order to prevent bare passwords in fhem.cfg.\n".
							  "          If you choose to encrypt your private password with base64, you also must encrypt your gateway password the same way\n"; 
			}
		}
		else # If private password is provided as bare word
		{
			if (length($km200_private_password) > 0)
			{
					Log3 $name, 4, $name. " : km200 - Provided PrivatePassword exists at least";		
			}
			else
			{
				return $name .": Error - PrivatePassword has not been provided.\n".
							  "          You may encode your password with base64 first, in order to prevent bare passwords in fhem.cfg.\n".
							  "          If you choose to encrypt your private password with base64, you also must encrypt your gateway password the same way\n"; 
				Log3 $name, 3, $name. " : km200 - Provided Private Password does not follow the specifications";
		  }
		}
		####END#### Check whether private password is available and decode it with base64 if encoding is used
	}
	else
	{
	    return $name .": km200 - Error - Not enough parameter provided." . "\n" . "Gateway IPv4 address, Gateway and Private Passwords must be provided" ."\n". "Please use \"define <devicename> km200 <IPv4-address> <GatewayPassword> <PrivatePassword>\" instead";
	}
	####END#### Check whether all variables are available ######################################################END#####

	
	###START###### Create the secret SALT of the MD5-Hash for AES-encoding ####################################START####
	my $Buderus_MD5Salt = pack(
	'C*',
	0x86, 0x78, 0x45, 0xe9, 0x7c, 0x4e, 0x29, 0xdc,
	0xe5, 0x22, 0xb9, 0xa7, 0xd3, 0xa3, 0xe0, 0x7b,
	0x15, 0x2b, 0xff, 0xad, 0xdd, 0xbe, 0xd7, 0xf5,
	0xff, 0xd8, 0x42, 0xe9, 0x89, 0x5a, 0xd1, 0xe4
	);
	####END####### Create the secret SALT of the MD5-Hash for AES-encoding #####################################END#####
	

	###START###### Create keys with MD5 #######################################################################START####
	# Copy Salt
	my $km200_crypt_md5_salt = $Buderus_MD5Salt;

	# First half of the key: MD5 of (km200GatePassword . Salt)
	my $key_1 = md5($km200_gateway_password . $km200_crypt_md5_salt);  
	  
	# Second half of the key: - Initial: MD5 of ( Salt)
	my $key_2_initial = md5($km200_crypt_md5_salt);

	# Second half of the key: -  private: MD5 of ( Salt . km200PrivatePassword)
	my $key_2_private = md5($km200_crypt_md5_salt . $km200_private_password);

	# Create keys
	my $km200_crypt_key_initial = ($key_1 . $key_2_initial);
	my $km200_crypt_key_private = ($key_1 . $key_2_private);
	####END####### Create keys with MD5 #########################################################################END#####
  
  
	###START###### Writing values to global hash ###############################################################START####
	  $hash->{NAME}                             = $name;
	  $hash->{URL}                              = $url;
	  $hash->{INTERVALDYNVAL}                   = 300;
	  $hash->{DELAYDYNVAL}                      = 60;
	  $hash->{POLLINGTIMEOUT}                   = 5;
	  $hash->{READBACKDELAY}					= 100;
	  $hash->{READTEXTLEN}                      = 0;
	  $hash->{temp}{ServiceCounterInit}         = 0;
	  $hash->{temp}{ServiceCounterDyn}          = 0;
	  $hash->{temp}{ServiceDbLogSplitHash}      = ();
	  $hash->{status}{FlagInitRequest}          = false;
	  $hash->{status}{FlagGetRequest}           = false;
	  $hash->{status}{FlagSetRequest}           = false;
	  $hash->{status}{FlagDynRequest}           = false;
	  $hash->{status}{FlagStatRequest}          = false;
	  $hash->{Secret}{CRYPTKEYPRIVATE}          = $km200_crypt_key_private;
	  $hash->{Secret}{CRYPTKEYINITIAL}          = $km200_crypt_key_initial;
	@{$hash->{Secret}{KM200ALLSERVICES}}        = sort @KM200_AllServices;
	@{$hash->{Secret}{KM200ALLSERVICESBACKUP}}  = sort @KM200_AllServices;
	@{$hash->{Secret}{KM200RESPONDINGSERVICES}} = ();
	@{$hash->{Secret}{KM200WRITEABLESERVICES}}  = ();
	@{$hash->{Secret}{KM200DONOTPOLL}}  		= ();
	####END####### Writing values to global hash ################################################################END#####

	###START###### Reset fullResponse error message ############################################################START####
	readingsSingleUpdate( $hash, "fullResponse", "OK", 1);
	####END####### Reset fullResponse error message #############################################################END#####	
	
	###START###### For Debugging purpose only ##################################################################START####  
	Log3 $name, 4, $name. " : km200 - Define H                               : " .$hash;
	Log3 $name, 4, $name. " : km200 - Define D                               : " .$def;
	Log3 $name, 4, $name. " : km200 - Define A                               : " .@a;
	Log3 $name, 4, $name. " : km200 - Define Name                            : " .$name;
	Log3 $name, 4, $name. " : km200 - Define Adr                             : " .$url;
	####END####### For Debugging purpose only ###################################################################END#####

	
	###START###### Check whether communication to the physical unit is possible ################################START####
	my $Km200Info ="";
	$hash->{temp}{service} = "/gateway/DateTime";
	$Km200Info = km200_GetSingleService($hash);
	
	if ($Km200Info eq "ERROR") 
	{
		$Km200Info = $hash->{temp}{TransferValue};
		$hash->{temp}{TransferValue} = "";

		## Communication with Gateway WRONG !! ##
		$hash->{STATE}="Error - No Communication";
		return ($name .": km200 - ERROR - The communication between fhem and the Buderus KM200 failed! \n". 
		               "                  Please check physical connection, IP-address and passwords! \n");
	} 
	elsif ($Km200Info eq "SERVICE NOT AVAILABLE") ## Communication OK but service not available ##
	{
		Log3 $name, 4, $name. " : km200 -  /gateway/DateTime                     : NOT AVAILABLE";
	} 
	else ## Communication OK and service is available ##
	{
		Log3 $name, 4, $name. " : km200 - /gateway/DateTime                      : AVAILABLE";				
	}
	####END####### Check whether communication to the physical unit is possible ################################END#####

	###START###### Initiate the timer for first time polling of  values from KM200 but wait 10s ###############START####
	InternalTimer(gettimeofday()+10, "km200_GetInitService", $hash, 1);
	Log3 $name, 4, $name. " : km200 - Internal timer for Initialisation of services started for the first time.";
	####END####### Initiate the timer for first time polling of  values from KM200 but wait 60s ################END#####
	
	return undef;
}
####END####### Activate module after module has been used via fhem command "define" ############################END#####


###START###### To bind unit of value to DbLog entries #########################################################START####
sub km200_DbLog_splitFn($$)
{
	my ($event, $name)	= @_;
	my ($reading, $value, $unit);
    my $hash 			= $defs{$name};
	my @argument		= split("[ \t][ \t]*", $event);
	
	### Delete ":" and everything behind in readings name
	$argument[0] =~ s/:.*//;
 
	### Log entries for debugging
	Log3 $name, 5, $name. " : km200_DbLog_splitFn - Content of event             : " . $event;
	Log3 $name, 5, $name. " : km200_DbLog_splitFn - Content of argument[0]       : " . $argument[0];
#	Log3 $name, 5, $name. " : km200_DbLog_splitFn - Content of argument[1]       : " . $argument[1];

	### If the service to be changed is identical to the one where the unit received from
	if ($argument[0] = $hash->{temp}{ServiceDbLogSplitHash}{id})
	{
		### Get values being changed from hash
		$reading = $argument[0];
		$value   = $argument[1];
		$unit    = $hash->{temp}{ServiceDbLogSplitHash}{unitOfMeasure};
	}

	### Delete temporary json-hash for DbLog-Split
	$hash->{temp}{ServiceDbLogSplitHash} = ();
	### Delete temporary json-hash for DbLog-Split

   return ($reading, $value, $unit);
}
####END####### To bind unit of value to DbLog entries ##########################################################END#####


###START###### Deactivate module module after "undefine" command by fhem ######################################START####
sub km200_Undefine($$)
{
	my ($hash, $def)  = @_;
	my $name = $hash->{NAME};	
	my $url  = $hash->{URL};

  	### Stop the internal timer for this module
	RemoveInternalTimer($hash);

	Log3 $name, 3, $name. " - km200 has been undefined. The KM unit at $url will no longer polled.";

	return undef;
}
####END####### Deactivate module module after "undefine" command by fhem #######################################END#####


###START###### Handle attributes after changes via fhem GUI ###################################################START####
sub km200_Attr(@)
{
	my @a               = @_;
	my $name            = $a[1];
	my $hash            = $defs{$name};
	my $IntervalDynVal  = $hash->{INTERVALDYNVAL};
	my $DelayDynVal     = $hash->{DELAYDYNVAL};
	my $ReadBackDelay   = $hash->{READBACKDELAY};
	my $ReadingTextLen  = $hash->{READTEXTLEN};
	
	
	### Check whether disable attribute has been provided
	if ($a[2] eq "disable")
	{
		###START### Check whether device shall be disabled
		if ($a[3] == 1)
		{
			### Set new status
			$hash->{STATE} = "Disabled";
			
			### Stop the current timer
			RemoveInternalTimer($hash);
			Log3 $name, 4, $name. " : km200 - InternalTimer has been removed.";

			### Delete all Readings
			fhem( "deletereading $name .*" );
			
			### Recover list of root services
			@{$hash->{Secret}{KM200ALLSERVICES}}= @{$hash->{Secret}{KM200ALLSERVICESBACKUP}};
			
			Log3 $name, 3, $name. " : km200 - Device disabled as per attribute.";
		}
		else
		{
			### Initiate the timer for first time polling of  values from KM200 but wait 10s
			InternalTimer(gettimeofday()+10, "km200_GetInitService", $hash, 1);
			Log3 $name, 4, $name. " : km200 - Internal timer for Initialisation of services re-started.";

			### Set new status
			$hash->{STATE} = "Initiating Sounding...";
			
			Log3 $name, 4, $name. " : km200 - Device enabled as per attribute.";
		}
		####END#### Check whether device shall be disabled
	}
	### Check whether dynamic interval attribute has been provided
	elsif ($a[2] eq "IntervalDynVal")
	{

		$IntervalDynVal = $a[3];
		###START### Check whether polling interval is not too short
		if ($IntervalDynVal > 19)
		{
			$hash->{INTERVALDYNVAL} = $IntervalDynVal;
			Log3 $name, 4, $name. " : km200 - IntervalDynVal set to attribute value:" . $IntervalDynVal ." s";
		}
		else
		{
			return $name .": Error - Gateway interval for IntervalDynVal too small - server response time longer than defined interval, please use something >=20, default is 90";
		}
		####END#### Check whether polling interval is not too short
	}
	### Check whether polling timeout attribute has been provided
	elsif($a[2] eq "PollingTimeout") 
	{
		###START### Check whether timeout is not too short
		if ($a[3] >= 5)
		{
			$hash->{POLLINGTIMEOUT} = $a[3];
			Log3 $name, 4, $name. " : km200 - Polling timeout set to attribute value:" . $a[3] ." s";
		}
		else
		{
			Log3 $name, 4, $name. " : km200 - Error - Gateway polling timeout attribute too small: " . $a[3] ." s";
			return $name .": Error - Gateway polling timeout attribute is too small - server response time is 5s minimum, default is 5";
		}
		####END#### Check whether timeout is not too short
	}
	### Check whether DoNotPoll attribute have been provided
	elsif($a[2] eq "DoNotPoll") 
	{
		my @KM200_DONOTPOLL   = ();
		my @temp              = @a;

		### Stop the current timer
		RemoveInternalTimer($hash);
		Log3 $name, 4, $name. " : km200 - InternalTimer has been removed.";
		
		### Delete the first 3 items of the array
		splice @temp, 0, 3;

		### Insert empty field as minimum entry 
		push @temp, "";
		
		### Transform string entries seperated by blank into array
		@KM200_DONOTPOLL = split(/\s+/, $temp[0]);

		### Remove trailing slash of each item if available
		
		### For each item found in this empty parent directory
		foreach my $item (@KM200_DONOTPOLL)
		{
			### Delete trailing slash
			$item =~ s/\/$//;
		}
		
		### Save list of services not to be polled into hash
		@{$hash->{Secret}{KM200DONOTPOLL}} = @KM200_DONOTPOLL;

		### Get original list of root services back
		@{$hash->{Secret}{KM200ALLSERVICES}} = @{$hash->{Secret}{KM200ALLSERVICESBACKUP}};
		
		### For every blacklisted service
		foreach my $SearchWord(@KM200_DONOTPOLL)
		{
			### Filter all blocked root services out of services to be polled 
			my $FoundPosition = first_index{ $_ eq $SearchWord }@{$hash->{Secret}{KM200ALLSERVICES}};
			if ($FoundPosition >= 0)
			{
				splice(@{$hash->{Secret}{KM200ALLSERVICES}}, $FoundPosition, 1);
			}
		}
		### Message for debugging purposes
		Log3 $name, 5, $name. "km200 module is only polling the following services! \n @{$hash->{Secret}{KM200ALLSERVICES}}";
		Log3 $name, 5, $name. "km200 module is NOT  polling the following services! \n @{$hash->{Secret}{KM200DONOTPOLL}}";
		Log3 $name, 4, $name. " : km200 - The following services will not be polled: ". $a[3];
		
		### Interrupting all currently running Polling
		@{$hash->{Secret}{KM200DYNSERVICES}} = "";
		$hash->{temp}{ServiceCounterDyn} = 0;

		### Delete all Readings
		fhem( "deletereading $name .*" );

		### Re-start the sounding of  values from KM200 but wait the period of $hash->{POLLINGTIMEOUT} + 10s
		InternalTimer(gettimeofday()+$hash->{POLLINGTIMEOUT}+10, "km200_GetInitService", $hash, 1);
		Log3 $name, 4, $name. " : km200 - Sounding of services re-started after change of DoNotPoll attribute";
	}
	### Check whether time-out for Read-Back has been provided
	if($a[2] eq "ReadBackDelay") 
	{
		$ReadBackDelay = $a[3];
		###START### Check whether ReadBackDelay is valid
		if ($ReadBackDelay >= 0)
		{
			$hash->{READBACKDELAY} = $ReadBackDelay;
			Log3 $name, 4, $name. " : km200 - ReadBackDelay set to attribute value:" . $ReadBackDelay ." s";
		}
		else
		{
			return $name .": Error - Read-Back delay time must be positive. Default is 0us";
		}
		####END#### Check whether ReadBackDelay is valid
	}
	### Check whether length for Readins has been provided
	if($a[2] eq "ReadingTextLen") 
	{
		$ReadingTextLen = $a[3];
		###START### Check whether ReadBackDelay is valid
		if ($ReadingTextLen >= 0)
		{
			$hash->{READTEXTLEN} = $ReadingTextLen;
			Log3 $name, 4, $name. " : km200 - ReadingTextLen set to attribute value:" . $ReadingTextLen ." s";
		}
		else
		{
			return $name .": Error - Text Length must be positive. Default is 0";
		}
		####END#### Check whether ReadBackDelay is valid
	}
	### If no attributes of the above known ones have been selected
	else
	{
		# Do nothing
	}
	return undef;
}
####END####### Handle attributes after changes via fhem GUI ####################################################END#####


###START###### Obtain value after "get" command by fhem #######################################################START####
sub km200_Get($@)
{
	my ( $hash, @a ) = @_;
	
	### If not enough arguments have been provided
	if ( @a < 2 )
	{
		return "\"get km200\" needs at least one argument";
	}
		
	my $name     = shift @a;
	my $service  = shift @a;
	my $option   = shift @a;
	my %km200_gets;
	my $ReturnValue;
	my $ReturnMessage;
	
	### Get the list of possible services and create a hash out of it
	my @GetServices = @{$hash->{Secret}{KM200ALLSERVICES}};

	foreach my $item(@GetServices) 	
	{
		$km200_gets{$item} = ("1");
	}
	
	### Remove trailing slash if available
	$service = $1 if($service=~/(.*)\/$/);

	### For programming purpose only
	if ($service eq "TestErrorMsg") {
		my $classdescription;
		my $classworkaround;
		my $reason;
		my $checklist;
		my $correction;
		
		($classdescription, $classworkaround, $reason, $checklist, $correction) = km200_GetErrorMessage("A11", "3084", "4", $hash);
		$ReturnMessage = 
		"<style>table, th, td {border: 1px solid black; border-collapse: collapse;}</style>" .
		"<table n>" .
		"<tr><td>The error-class description is : </td><td>" . $classdescription . "</td></tr>" .
		"<tr><td>The error-class workaround is :  </td><td>" . $classworkaround .  "</td></tr>" .
		"<tr><td>The error-reason is :            </td><td>" . $reason .           "</td></tr>" .
		"<tr><td>The checklist is :               </td><td>" . $checklist .        "</td></tr>" .
		"<tr><td>The correction is :              </td><td>" . $correction .       "</td></tr>" .
		"</table>";
		return($ReturnMessage);
	}

	### If service chosen in GUI does not exist
	if(!$km200_gets{$service}) 
	{
		my @cList = keys %km200_gets;
		push(@cList, "TestErrorMsg");
		return "Unknown argument $service, choose one of " . join(" ", @cList);
	}

	### Check whether the initialisation process has been finished
	if ($hash->{temp}{ServiceCounterInit} == false)
	{
		### Save chosen service into hash
		$hash->{temp}{service} = $service;

		### Read service-hash
		$ReturnValue = km200_GetSingleService($hash);

		### If the "get" - option has been set to "Json" for the return of the raw Json-string
		if ($option =~ m/json/i)
		{
			$ReturnMessage = $hash->{temp}{JsonRaw};
		}
		### If no option has been chosen, just return the result of the value.
		else
		{
			### If type is a floatvalue then format decimals
			if ($ReturnValue->{type} eq "floatValue")
			{
				$ReturnMessage = sprintf("%.1f", $ReturnValue->{value});
			}
			### If type is something else just pass throught
			else
			{
				$ReturnMessage = $ReturnValue->{value};
			}
		}
	}
	### If the initialisation process has NOT been finished
	else
	{
		$ReturnMessage = "The initialisation process is still ongoing. Please wait for the STATE changing to \"Standby\"";
	}
		
	
	### Delete temporary values
	$hash->{temp}{service} = "";
	$hash->{temp}{JsonRaw} = "";
	

	### Return value
	return($ReturnMessage);
}
####END####### Obtain value after "get" command by fhem ########################################################END#####


###START###### Manipulate service after "set" command by fhem #################################################START####
sub km200_Set($@)
{
	my ( $hash, @a ) = @_;
	
	### If not enough arguments have been provided
	if ( @a < 2 )
	{
		return "\"set km200\" needs at least one argument";
	}
		
	my $name = shift @a;
	my $service  = shift @a;
	my $value = join(" ", @a);
	my %km200_sets;
	my $ReturnMessage;

	### Get the list of possible services and create a hash out of it
	my @WriteableServices = @{$hash->{Secret}{KM200WRITEABLESERVICES}};

	

	foreach my $item(@WriteableServices) 	
	{
		$km200_sets{$item} = ("1");
	}
	
	### If service chosen in GUI does not exist
	if(!$km200_sets{$service}) 
	{
		my @cList = keys %km200_sets;
		return "Unknown argument $service, choose one of " . join(" ", @cList);
	}

	### Check whether the initialisation process has been finished
	if ($hash->{temp}{ServiceCounterInit} == false)
	{
		### Save chosen service into hash
		$hash->{temp}{service}  = $service;
		$hash->{temp}{postdata} = $value;	
		
		### Call set sub
		$ReturnMessage = km200_PostSingleService($hash);
	}
	### If the initialisation process has NOT been finished
	else
	{
		$ReturnMessage = "The initialisation process is still ongoing. Please wait for the STATE changing to \"Standby\"";
	}
	
	### Delete temporary hash values
	$hash->{temp}{postdata} = "";
	$hash->{temp}{service}  = "";
	
	return($ReturnMessage);
}
####END####### Manipulate service after "Set" command by fhem ##################################################END#####


###START####### Repeats "string" for "count" times ############################################################START####
sub str_repeat($$)
{
    my $string = $_[0];
    my $count  = $_[1];
    return(${string}x${count});
}
####END######## Repeats "string" for "count" times #############################################################END#####


###START###### Subroutine Encrypt Data ########################################################################START####
sub km200_Encrypt($)
{
	my ($hash, $def)            = @_;

	my $km200_crypt_key_private = $hash->{Secret}{CRYPTKEYPRIVATE};
	my $name                    = $hash->{NAME};
	my $encryptData             = $hash->{temp}{jsoncontent};

    # Create Rijndael encryption object
    my $cipher = Crypt::Rijndael->new($km200_crypt_key_private, Crypt::Rijndael::MODE_ECB() );

    # Get blocksize and add PKCS #7 padding
    my $blocksize =  $cipher->blocksize();
    my $encrypt_padchar = $blocksize - ( length( $encryptData ) % $blocksize );
    $encryptData .= str_repeat( chr( $encrypt_padchar ), $encrypt_padchar );

	# Do the encryption
    my $ciphertext = $cipher->encrypt( $encryptData );

	# Do additional encoding with base64
    $ciphertext = encode_base64($ciphertext);
   
    # Return the encoded text
    return($ciphertext);
}
####END####### Subroutine Encrypt Data #########################################################################END#####


###START###### Subroutine Decrypt Data ########################################################################START####
sub km200_Decrypt($)
{
	my ($hash, $def)            = @_;

	my $km200_crypt_key_private = $hash->{Secret}{CRYPTKEYPRIVATE};
	my $name                    = $hash->{NAME};
	my $decryptData             = $hash->{temp}{decodedcontent};

	### Log entries for debugging purposes
    #Log3 $name, 5, $name. " : km200 - decryptData2 - decryptData                : " .$decryptData;
    
	# Remove additional encoding with base64
	$decryptData = decode_base64($decryptData);
	
	### Log entries for debugging purposes
	#Log3 $name, 5, $name. " : km200 - decryptData2 - base64decode               : " .$decryptData;

	# Check whether the length of the decryptData is NOT multiplies of 16
	if ((length($decryptData)&0xF) != 0)
	{
		# Return nothing which will end this subroutine
		return "";
	}
	
    # Create Rijndael decryption object and do the decryption
    my $cipher = Crypt::Rijndael->new($km200_crypt_key_private, Crypt::Rijndael::MODE_ECB() );
    my $deciphertext = $cipher->decrypt( $decryptData );

    # Remove zero padding
    $deciphertext =~ s/\x00+$//;

    # Remove PKCS #7 padding   
    my $decipher_len = length($deciphertext);
    my $decipher_padchar = ord(substr($deciphertext,($decipher_len - 1),1));
   
    my $i = 0;
    
    for ( $i = 0; $i < $decipher_padchar ; $i++ )
    {
        if ( $decipher_padchar != ord( substr($deciphertext,($decipher_len - $i - 1),1)))
        {
            last;
        }
    }
    
    # Return decrypted text
    if ( $i != $decipher_padchar )
    {
		### Log entries for debugging purposes
		Log3 $name, 5, $name. " : km200 - decryptData1 - decipher_len                : " .$decipher_len;
		$deciphertext =~ s/\x00+$//;
		Log3 $name, 5, $name. " : km200 - decryptData1 - deciphertext                : " .$deciphertext;
		### Log entries for debugging purposes
        return $deciphertext;
    }
    
	else
    {
		$deciphertext = substr($deciphertext,0,$decipher_len - $decipher_padchar);
		### Log entries for debugging purposes
		Log3 $name, 5, $name. " : km200 - decryptData2 - decipher_len                : " .$decipher_len;
		$deciphertext =~ s/\x00+$//;
		Log3 $name, 5, $name. " : km200 - decryptData2 - deciphertext                : " .$deciphertext;
		### Log entries for debugging purposes
        return $deciphertext;
    }
}
####END####### Subroutine Decrypt Data #########################################################################END#####

###START###### Subroutine set individual data value ###########################################################START####
sub km200_PostSingleService($)
{
	my ($hash, $def)                = @_;
	my $Service                     = $hash->{temp}{service};
	my $km200_gateway_host          = $hash->{URL} ;
	my $name                        = $hash->{NAME} ;
	my $PollingTimeout              = $hash->{POLLINGTIMEOUT};
	my $err;
	my $data;
	my $jsonSend;
	my $jsonRead;
	my $JsonContent;
	
	### Log file entry for debugging
	Log3 $name, 5, $name. ("km200_Set - Writing value: " . $hash->{temp}{postdata} . " to the service                     : ". $Service . "\n");
	
	### Read the current json string 
	$jsonRead = km200_GetSingleService($hash);	
	
	#### If the get-command returns an error due to an unknown Service requested
	if ($jsonRead -> {type} eq "ERROR")
	{
		### Rescue original Service request
		my $WriteService = $Service;

		### Try to replace the Post-String with nothing
		$Service  =~ s/\/1-Mo//i;
		$Service  =~ s/\/2-Tu//i;
		$Service  =~ s/\/3-We//i;
		$Service  =~ s/\/4-Th//i;
		$Service  =~ s/\/5-Fr//i;
		$Service  =~ s/\/6-Sa//i;
		$Service  =~ s/\/7-Su//i;

		### Save corrected string in hash
		$hash->{temp}{service} = $Service;

		### Log file entry for debugging
		Log3 $name, 5, $name. "km200_Set - Trying to re-read Service - Assuming its a switchProgram list";

		### Try again to read the current json string again with the corrected service
		$jsonRead = km200_GetSingleService($hash);	

		### Check whether the type is an switchProgram. 
		### If true, the requested service was a particular week of the switchProgram
		if ($jsonRead -> {type} eq "switchProgram")
		{
			### Log file entry for debugging
			Log3 $name, 5, $name. "km200_Set - It is a switchProgram list!";

			### For each weekday, get current readings, delete all unnecessary blanks and transform to array
			my $TempReadingVal;
			
			   $TempReadingVal		= ReadingsVal($name,($Service . "/1-Mo"),"");
			   $TempReadingVal		=~ s/\s+/ /g;
			   $TempReadingVal		=~ s/\s+$//g;
			my @TempReadingMo 	    = split(/\s+/, $TempReadingVal,0);
			   
			   $TempReadingVal		= ReadingsVal($name,($Service . "/2-Tu"),"");
			   $TempReadingVal		=~ s/\s+/ /g;
			   $TempReadingVal		=~ s/\s+$//g;
			my @TempReadingTu 		= split(/\s+/, $TempReadingVal,0);
			   
			   $TempReadingVal		= ReadingsVal($name,($Service . "/3-We"),"");
			   $TempReadingVal		=~ s/\s+/ /g;
			   $TempReadingVal		=~ s/\s+$//g;
			my @TempReadingWe 		= split(/\s+/, $TempReadingVal,0);

			   $TempReadingVal		= ReadingsVal($name,($Service . "/4-Th"),"");
			   $TempReadingVal		=~ s/\s+/ /g;
			   $TempReadingVal		=~ s/\s+$//g;
			my @TempReadingTh 		= split(/\s+/, $TempReadingVal,0);
			
			   $TempReadingVal		= ReadingsVal($name,($Service . "/5-Fr"),"");
			   $TempReadingVal		=~ s/\s+/ /g;
			   $TempReadingVal		=~ s/\s+$//g;
			my @TempReadingFr 		= split(/\s+/, $TempReadingVal,0);
			
			   $TempReadingVal		= ReadingsVal($name,($Service . "/6-Sa"),"");
			   $TempReadingVal		=~ s/\s+/ /g;
			   $TempReadingVal		=~ s/\s+$//g;
			my @TempReadingSa 		= split(/\s+/, $TempReadingVal,0);
			
			   $TempReadingVal		= ReadingsVal($name,($Service . "/7-Su"),"");
			   $TempReadingVal		=~ s/\s+/ /g;
			   $TempReadingVal		=~ s/\s+$//g;
			my @TempReadingSu 		= split(/\s+/, $TempReadingVal,0);

			
			### For value to be written, delete all unnecessary blanks and transform to array and get length of array
			my $ReturnString		= $hash->{temp}{postdata};
			   $ReturnString 		=~ s/\s+/ /g;
			   $ReturnString 		=~ s/\s+$//g;
			my @TempReading			= split(/\s+/, $ReturnString);
			my $TempReadingLength	= @TempReading;

			
			### Obtain the allowed terminology for setpoints
			$hash->{temp}{service}	= $jsonRead -> {setpointProperty}{id};
			my $TempSetpointsJson	= km200_GetSingleService($hash);

			
			my @TempSetpointNames	=();
			### For each item found in this empty parent directory
			foreach my $item (@{ $TempSetpointsJson->{references} })
			{
				my $TempSetPoint = substr($item->{id}, (rindex($item->{id}, "/") - length($item->{id}) +1));
				
				### Add service, which is one of the allowed terminologies at the same time, to the list of all known services
				push (@TempSetpointNames, $TempSetPoint);
			}
			
			### Restore the original service
			$hash->{temp}{service}	= $Service;
			
			### If number of switchpoints exceeds maximum allowed
			if (($TempReadingLength / 2) > $jsonRead -> {maxNbOfSwitchPointsPerDay})
			{
				return ("ERROR - Too much Switchpoints for weeklist inserted. \n Do not add more than " . $jsonRead -> {maxNbOfSwitchPointsPerDay} . " SwitchPoints per day!\n");	
			}
			
			### If content of array is not even
			if (($TempReadingLength % 2)  != 0)
			{
				return "ERROR - At least one Switchtime or Switchpoint is missing. \n Make sure you always have couples of Switchtime and Switchpoint!\n";	
			}
			
			### Check whether description of setpoints is the same as referenced and the data is in the right order
			for (my $i=0;$i<$TempReadingLength;$i+=2)
			{ 
				### If the even element behind the uneven index [1, 3, 5, ...] is not one of the pre-defined setpoints
				if (! grep /($TempReading[$i+1])/,@TempSetpointNames)
				{
					return "ERROR - At least for one Switchpoint the wrong terminology has been used. Only use one of the following items: " . join(' , ',@TempSetpointNames) ."\n";;
				}

				### If the uneven element behind the even index [0, 2, 4, ...]is not a number, hand back an error message
				if ($TempReading[$i] !~ /^[0-9.-]+$/)
				{
					return "ERROR - At least for one Switchtime a number is expected at that position. \n Ensure the correct syntax of time and switchpoint. (E.g. 0600 eco)\n";
				}
			
				### Convert timepoint into raster of defined switchPointTimeRaster
				my $TempHours    = substr($TempReading[$i], 0, length($TempReading[$i])-2);
				if ($TempHours > 23)
				{
					$TempHours = 23;
				}
				my $TempMinutes	 = substr($TempReading[$i], -2);
				if ($TempMinutes > 59)
				{
					$TempMinutes = 59;
				}
				$TempMinutes 	 = $TempMinutes / ($jsonRead -> {switchPointTimeRaster});
				$TempMinutes 	 =~ s/^(.*?)\..*$/$1/;
				$TempMinutes 	 = $TempMinutes * ($jsonRead -> {switchPointTimeRaster});
				$TempMinutes 	 = sprintf ('%02d', $TempMinutes);
				$TempReading[$i] = ($TempHours . $TempMinutes);

			}
			
			$hash->{temp}{postdata} = join(" ", @TempReading);
			
			### For the requested day to be changed, save new value
			if    ($WriteService =~ m/1-Mo/i)
			{
				@TempReadingMo   = @TempReading;
			}
			elsif ($WriteService =~ m/2-Tu/i)
			{
				@TempReadingTu   = @TempReading;
			}
			elsif ($WriteService =~ m/3-We/i)
			{
				@TempReadingWe   = @TempReading;
			}	
			elsif ($WriteService =~ m/4-Th/i)
			{
				@TempReadingTh   = @TempReading;
			}	
			elsif ($WriteService =~ m/5-Fr/i)
			{
				@TempReadingFr   = @TempReading;
			}	
			elsif ($WriteService =~ m/6-Sa/i)
			{
				@TempReadingSa   = @TempReading;
			}	
			elsif ($WriteService =~ m/7-Su/i)
			{
				@TempReadingSu   = @TempReading;
			}	

			
			### For every weekday create setpoint hash and push it to array of hashes of switchpoints to be send
			my @SwitchPointsSend =();
			
			for (my $i=0;$i<$#TempReadingMo;$i+=2)
			{
				my $TempHashSend;
				$TempHashSend->{"dayOfWeek"} = "Mo";
				my $TempHours                = substr($TempReadingMo[$i], 0, length($TempReadingMo[$i])-2);
				my $TempMinutes	             = substr($TempReadingMo[$i], -2);
				$TempHashSend->{"time"}      = ($TempHours * 60 ) + $TempMinutes;
				$TempHashSend->{"setpoint"}  = $TempReadingMo[$i+1];
				push @SwitchPointsSend, $TempHashSend;
			}

			for (my $i=0;$i<$#TempReadingTu;$i+=2)
			{
				my $TempHashSend;
				$TempHashSend->{"dayOfWeek"} = "Tu";
				my $TempHours                = substr($TempReadingTu[$i], 0, length($TempReadingTu[$i])-2);
				my $TempMinutes	             = substr($TempReadingTu[$i], -2);
				$TempHashSend->{"time"}      = ($TempHours * 60 ) + $TempMinutes;				
				$TempHashSend->{"setpoint"}  = $TempReadingTu[$i+1];
				push @SwitchPointsSend, $TempHashSend;
			}			
			
			for (my $i=0;$i<$#TempReadingWe;$i+=2)
			{
				my $TempHashSend;
				$TempHashSend->{"dayOfWeek"} = "We";
				my $TempHours                = substr($TempReadingWe[$i], 0, length($TempReadingWe[$i])-2);
				my $TempMinutes	             = substr($TempReadingWe[$i], -2);
				$TempHashSend->{"time"}      = ($TempHours * 60 ) + $TempMinutes;
				$TempHashSend->{"setpoint"}  = $TempReadingWe[$i+1];
				push @SwitchPointsSend, $TempHashSend;
			}
	
			for (my $i=0;$i<$#TempReadingTh;$i+=2)
			{
				my $TempHashSend;
				$TempHashSend->{"dayOfWeek"} = "Th";
				my $TempHours                = substr($TempReadingTh[$i], 0, length($TempReadingTh[$i])-2);
				my $TempMinutes	             = substr($TempReadingTh[$i], -2);
				$TempHashSend->{"time"}      = ($TempHours * 60 ) + $TempMinutes;
				$TempHashSend->{"setpoint"}  = $TempReadingTh[$i+1];
				push @SwitchPointsSend, $TempHashSend;
			}
			
			for (my $i=0;$i<$#TempReadingFr;$i+=2)
			{
				my $TempHashSend;
				$TempHashSend->{"dayOfWeek"} = "Fr";
				my $TempHours                = substr($TempReadingFr[$i], 0, length($TempReadingFr[$i])-2);
				my $TempMinutes	             = substr($TempReadingFr[$i], -2);
				$TempHashSend->{"time"}      = ($TempHours * 60 ) + $TempMinutes;
				$TempHashSend->{"setpoint"}  = $TempReadingFr[$i+1];
				push @SwitchPointsSend, $TempHashSend;
			}
			
			for (my $i=0;$i<$#TempReadingSa;$i+=2)
			{
				my $TempHashSend;
				$TempHashSend->{"dayOfWeek"} = "Sa";
				my $TempHours                = substr($TempReadingSa[$i], 0, length($TempReadingSa[$i])-2);
				my $TempMinutes	             = substr($TempReadingSa[$i], -2);
				$TempHashSend->{"time"}      = ($TempHours * 60 ) + $TempMinutes;
				$TempHashSend->{"setpoint"}  = $TempReadingSa[$i+1];
				push @SwitchPointsSend, $TempHashSend;
			}
			
			for (my $i=0;$i<$#TempReadingSu;$i+=2)
			{
				my $TempHashSend;
				$TempHashSend->{"dayOfWeek"} = "Su";
				my $TempHours                = substr($TempReadingSu[$i], 0, length($TempReadingSu[$i])-2);
				my $TempMinutes	 = substr($TempReadingSu[$i], -2);
				$TempHashSend->{"time"}      = ($TempHours * 60 ) + $TempMinutes;
				$TempHashSend->{"setpoint"}  = $TempReadingSu[$i+1];
				push @SwitchPointsSend, $TempHashSend;
			}

			### Save array of hashes of switchpoints into json hash to be send
			@{$jsonSend->{switchPoints}} = @SwitchPointsSend;
			
			### Create full URL of the current Service to be written
			my $url ="http://" . $km200_gateway_host . $Service;

			### Encode as json
			$JsonContent = encode_json($jsonSend);

			### Delete the name of hash, "{" and "}" out of json String. No idea why but result of Try-and-Error method
			$JsonContent =~ s/{"switchPoints"://;
			$JsonContent =~ s/]}/]/g;

			### Log file entry for debugging
			Log3 $name, 5, $name. "km200_Set - JsonContent                              : " . $JsonContent;

			### Encrypt 
			$hash->{temp}{jsoncontent} = $JsonContent;
			$data = km200_Encrypt($hash);

			### Log file entry for debugging
			Log3 $name, 5, $name. "km200_Set - Trying to push switchPoint list to KM-Unit";
			
			### Create parameter set for HttpUtils_BlockingGet
			my $param = {
							url        => $url,
							timeout    => $PollingTimeout * 5,
							data       => $data,
							method     => "POST",
							header     => "agent: TeleHeater/2.2.3\r\nUser-Agent: TeleHeater/2.2.3\r\nAccept: application/json",
						};
						
			### Block other scheduled and unscheduled routines
			$hash->{status}{FlagSetRequest} = true;

			### Write value with HttpUtils_BlockingGet
			($err, $data) = HttpUtils_BlockingGet($param);

			### Reset flag
			$hash->{status}{FlagSetRequest} = false;
			
			### If error message has been returned
			if($err ne "") 
			{
				Log3 $name, 2, $name . " - ERROR: $err";
				return $err;	
			}

			Log3 $name, 5, $name. ("Waiting for processing time (READBACKDELAY / [ms])     : " . $hash->{READBACKDELAY} . " \n");

			### Make a pause before ReadBack
			usleep ($hash->{READBACKDELAY}*1000);

		### Read service-hash and format it so it is compareable to the sent content
		my $ReReadContent = km200_GetSingleService($hash);
		   $ReReadContent = $ReReadContent->{switchPoints};
		   $ReReadContent = encode_json($ReReadContent);
		   $ReReadContent =~ s/{"switchPoints"://;
		   $ReReadContent =~ s/]}/]/g;
		   
		### Transform back into array of hashes
		eval 
			{
				$ReReadContent = decode_json(encode_utf8($ReReadContent));
				$JsonContent = decode_json(encode_utf8($JsonContent));
				1;
			}
		or do 
			{
			};
		
		### Set Counter for found items in SwitchPrograms
		my $FoundJsonItem = 0;
		
				
		### For every item of the array of SwitchPrograms to be send
		foreach my $ReReadItem (@{$ReReadContent})
		{
			### Set Counter for found items of ReRead values
			my $FoundReReadItem = 0;
			
			### For every item of the array of SwitchPrograms after Re-Reading 
			foreach my $JsonItem (@{$JsonContent})
			{
				### If the current Switchprogram - hash does not have the same amount of keys
				if (%$ReReadItem ne %$JsonItem) 
				{
					### Do nothing
					#print "they don't have the same number of keys\n";
				} 
				### If the current Switchprogram - hash do have the same amount of keys
				else 
				{
					### Compare key names and values
					my %cmp = map { $_ => 1 } keys %$ReReadItem;
					for my $key (keys %$JsonItem) 
					{
						last unless exists $cmp{$key};
						last unless $$ReReadItem{$key} eq $$JsonItem{$key};
						delete $cmp{$key};
					}
					if (%cmp) 
					{
						### Do nothing
						#print "they don't have the same keys or values\n";
					} 
					else 
					{
						### Inkrement Counter
						$FoundReReadItem = 1;
						#print "they have the same keys and values\n";
					}
				}
			}

			### If item has been found
			if ($FoundReReadItem == 1)
			{ 
				### Inkrement Counter for found identical SwitchPoints
				$FoundJsonItem++;
			}
		}
		
		my $ReturnValue;
		
		if 	($FoundJsonItem == @{$ReReadContent})
		{
			$ReturnValue = "The service " . $Service . " has been changed succesfully!";
			Log3 $name, 5, $name. "Writing $Service succesfully \n";
		}	
		else
		{
			$ReturnValue = "ERROR - The service " . $Service . " could not changed! \n";
		}

			### Return the status message
			return $ReturnValue;
		}
	}	
	### Check whether the type is an switchProgram. 
	### If true, the requested service is referring to the entire week but not a particular week.
	if ($jsonRead -> {type} eq "switchProgram")
	{
		### Create full URL of the current Service to be written
		my $url ="http://" . $km200_gateway_host . $Service;

		### Get the string to be send
		$JsonContent = $hash->{temp}{postdata};			

		### Encrypt 
		$hash->{temp}{jsoncontent} = $JsonContent;
		$data = km200_Encrypt($hash);
		
		### Create parameter set for HttpUtils_BlockingGet
		my $param = {
						url        => $url,
						timeout    => $PollingTimeout * 5,
						data       => $data,
						method     => "POST",
						header     => "agent: TeleHeater/2.2.3\r\nUser-Agent: TeleHeater/2.2.3\r\nAccept: application/json",
					};
					
		### Block other scheduled and unscheduled routines
		$hash->{status}{FlagSetRequest} = true;

		### Write value with HttpUtils_BlockingGet
		($err, $data) = HttpUtils_BlockingGet($param);

		### Reset flag
		$hash->{status}{FlagSetRequest} = false;
		
		### If error message has been returned
		if($err ne "") 
		{
			Log3 $name, 2, $name . " - ERROR: $err";
			return $err;	
		}

		Log3 $name, 5, $name. ("Waiting for processing time (READBACKDELAY / [ms])     : " . $hash->{READBACKDELAY} . " \n");
		### Make a pause before ReadBack
		usleep ($hash->{READBACKDELAY}*1000);

		### Read service-hash and format it so it is compareable to the sent content
		my $ReReadContent = km200_GetSingleService($hash);
		   $ReReadContent = $ReReadContent->{switchPoints};
		   $ReReadContent = encode_json($ReReadContent);
		   $ReReadContent =~ s/{"switchPoints"://;
		   $ReReadContent =~ s/]}/]/g;
		   
		### Transform back into array of hashes
		eval 
			{
				$ReReadContent = decode_json(encode_utf8($ReReadContent));
				$JsonContent = decode_json(encode_utf8($JsonContent));
				1;
			}
		or do 
			{
			};
		
		### Set Counter for found items in SwitchPrograms
		my $FoundJsonItem = 0;
		
				
		### For every item of the array of SwitchPrograms to be send
		foreach my $ReReadItem (@{$ReReadContent})
		{
			### Set Counter for found items of ReRead values
			my $FoundReReadItem = 0;
			
			### For every item of the array of SwitchPrograms after Re-Reading 
			foreach my $JsonItem (@{$JsonContent})
			{

				### If the current Switchprogram - hash does not have the same amount of keys
				if (%$ReReadItem != %$JsonItem) 
				{
					### Do nothing
					#print "they don't have the same number of keys\n";
				} 
				### If the current Switchprogram - hash do have the same amount of keys
				else 
				{
					### Compare key names and values
					my %cmp = map { $_ => 1 } keys %$ReReadItem;
					for my $key (keys %$JsonItem) 
					{
						last unless exists $cmp{$key};
						last unless $$ReReadItem{$key} eq $$JsonItem{$key};
						delete $cmp{$key};
					}
					if (%cmp) 
					{
						### Do nothing
						#print "they don't have the same keys or values\n";
					} 
					else 
					{
						### Inkrement Counter
						$FoundReReadItem = 1;
						#print "they have the same keys and values\n";
					}
				}
			}

			### If item has been found
			if ($FoundReReadItem == 1)
			{ 
				### Inkrement Counter for found identical SwitchPoints
				$FoundJsonItem++;
			}
		}
		
		my $ReturnValue;
		
		if 	($FoundJsonItem == @{$ReReadContent})
		{
			$ReturnValue = "The service " . $Service . " has been changed succesfully!";
			Log3 $name, 5, $name.  "The service $Service has been changed succesfully!";
			}	
		else
		{
			$ReturnValue = "ERROR - The service " . $Service . " could not changed! \n";
			Log3 $name, 5, $name. "Writing $Service was NOT succesfully";
		}

		### Return the status message
		return $ReturnValue;
	}	
	## Check whether the type is a single value containing a string
	elsif($jsonRead->{type} eq "stringValue") 
	{	
		### Save chosen value into hash to be send
		$jsonSend->{value} = $hash->{temp}{postdata};

		### Log file entry for debugging
		Log3 $name, 5, $name. "km200_Set - String value";		

		### Create full URL of the current Service to be written
		my $url ="http://" . $km200_gateway_host . $Service;

		### Encode as json
		$JsonContent = encode_json($jsonSend);
		
		### Encrypt 
		$hash->{temp}{jsoncontent} = $JsonContent;
		$data = km200_Encrypt($hash);

		
		### Create parameter set for HttpUtils_BlockingGet
		my $param = {
						url        => $url,
						timeout    => $PollingTimeout,
						data       => $data,
						method     => "POST",
						header     => "agent: TeleHeater/2.2.3\r\nUser-Agent: TeleHeater/2.2.3\r\nAccept: application/json",
					};
					
		### Block other scheduled and unscheduled routines
		$hash->{status}{FlagSetRequest} = true;

		### Write value with HttpUtils_BlockingGet
		($err, $data) = HttpUtils_BlockingGet($param);

		### Reset flag
		$hash->{status}{FlagSetRequest} = false;
		
		### If error message has been returned
		if($err ne "") 
		{
			Log3 $name, 2, $name . " - ERROR: $err";
			return $err;	
		}

		### Make a pause before ReadBack
		usleep ($hash->{READBACKDELAY}*1000);
		
		### Read service-hash
		my $ReadValue = km200_GetSingleService($hash);
		
		### Return value
		my $ReturnValue = "";
		if ($ReadValue->{value} eq $hash->{temp}{postdata})
		{	
			$ReturnValue = "The service " . $Service . " has been changed to: " . $ReadValue->{value};
			Log3 $name, 5, $name. "km200_Set - Writing " . $Service . " succesfully with value: " . $hash->{temp}{postdata};
		}
		else
		{
			$ReturnValue = "ERROR - The service " . $Service . " could not changed.";
			Log3 $name, 5, $name. "km200_Set - Writing " . $Service . " was NOT successful";
		}

		### Return the status message
		return $ReturnValue;
	}	
	## Check whether the type is a single value containing a float value
	elsif($jsonRead -> {type} eq "floatValue")
	{
		### Check whether value to be sent is numeric
		if ($hash->{temp}{postdata} =~ /^[0-9.-]+$/) 
		{
			### Save chosen value into hash to be send
			$jsonSend->{value} = ($hash->{temp}{postdata}) * 1;	

			### Log file entry for debugging
			Log3 $name, 5, $name. "km200_Set - Numeric value";
		
			### Create full URL of the current Service to be written
			my $url ="http://" . $km200_gateway_host . $Service;

			### Encode as json
			$JsonContent = encode_json($jsonSend);
			
			### Encrypt 
			$hash->{temp}{jsoncontent} = $JsonContent;
			$data = km200_Encrypt($hash);

			
			### Create parameter set for HttpUtils_BlockingGet
			my $param = {
							url        => $url,
							timeout    => $PollingTimeout,
							data       => $data,
							method     => "POST",
							header     => "agent: TeleHeater/2.2.3\r\nUser-Agent: TeleHeater/2.2.3\r\nAccept: application/json",
						};
						
			### Block other scheduled and unscheduled routines
			$hash->{status}{FlagSetRequest} = true;

			### Write value with HttpUtils_BlockingGet
			($err, $data) = HttpUtils_BlockingGet($param);

			### Reset flag
			$hash->{status}{FlagSetRequest} = false;
			
			### If error messsage has been returned
			if($err ne "") 
			{
				Log3 $name, 2, $name . " - ERROR: $err";
				return $err;	
			}

			### Make a pause before ReadBack
			usleep ($hash->{READBACKDELAY}*1000);	
			
			### Read service-hash
			my $ReadValue = km200_GetSingleService($hash);
			

			### Return value
			my $ReturnValue = "";
			if ($ReadValue->{value} eq $hash->{temp}{postdata})
			{	
				$ReturnValue = "The service " . $Service . " has been changed to: " . $ReadValue->{value} . "\n";
				Log3 $name, 5, $name. "km200_Set - Writing " . $Service . " succesfully with value: " . $hash->{temp}{postdata};
			}
			elsif ($jsonRead -> {value} == $ReadValue->{value})
			{
				$ReturnValue = "ERROR - The service " . $Service . " could not changed to: " . $hash->{temp}{postdata} . "\n The value is: " . $ReadValue->{value} . "\n";
				Log3 $name, 5, $name. "km200_Set - Writing " . $Service . " was NOT successful";
			}
			else
			{
				$ReturnValue = "The service " . $Service . " has been rounded to: " . $ReadValue->{value} . "\n";
				Log3 $name, 5, $name. "km200_Set - Writing " . $Service . " was rounded and changed successful";
			}

			### Return the status message
			return $ReturnValue;
		}
		### If the value to be sent is NOT numeric
		else
		{
			### Log file entry for debugging
			Log3 $name, 5, $name. "km200_Set - ERROR - Float value expected!";
			return ("km200_Set - ERROR - Float value expected!\n");
		}
	}
	## If the type is unknown
	else
	{
		### Log entries for debugging purposes
		Log3 $name, 4, $name. " : km200_SetSingleService - type unknown for : " .$Service;
	}
}
####END####### Subroutine set individual data value ############################################################END##### 


###START###### Subroutine get individual data value ###########################################################START####
sub km200_GetSingleService($)
{
	my ($hash, $def)            	= @_;
	my $Service                 	= $hash->{temp}{service};
	my $km200_gateway_host      	= $hash->{URL};
	my $name                    	= $hash->{NAME};
	my $PollingTimeout              = $hash->{POLLINGTIMEOUT};
	my $json -> {type}				= "";
	   $json -> {value}				= "";
	my $err;
	my $data;

	### Log entries for debugging purposes
    Log3 $name, 5, $name. " : km200 - GetSingleService - service            : " .$Service;
	
	### Create full URL of the current Service to be read
	my $url ="http://" . $km200_gateway_host . $Service;

	### Log entries for debugging purposes
    Log3 $name, 5, $name. " : km200 - GetSingleService - url                : " .$url;
	
	### Create parameter set for HttpUtils_BlockingGet
	my $param = {
					url        => $url,
					timeout    => $PollingTimeout,
					method     => "GET",
					header     => "agent: TeleHeater/2.2.3\r\nUser-Agent: TeleHeater/2.2.3\r\nAccept: application/json",
				};

	### Block other scheduled and unscheduled routines
	$hash->{status}{FlagGetRequest} = true;

	### Retrieve data from km200
	($err, $data) = HttpUtils_BlockingGet($param);

	### Log entries for debugging purposes
    Log3 $name, 5, $name. " : km200 - GetSingleService - err                : " .$err;
	Log3 $name, 5, $name. " : km200 - GetSingleService - data               : " .$data;
	
	### Block other scheduled and unscheduled routines
	$hash->{status}{FlagGetRequest} = false;
	
	### If error message has been reported
	if($err ne "") 
	{
		Log3 $name, 2, $name . " : ERROR: Service: ".$Service. ": No proper Communication with Gateway: " .$err;
		my $ReturnMessage ="ERROR";
		$json -> {type}   = $ReturnMessage;
		$json -> {value}  = $ReturnMessage;
		return $json;
	}
	### If NO error message has been reported
	else
	{
		$hash->{temp}{decodedcontent} = $data;
		my $decodedContent = km200_Decrypt($hash);

		if ($decodedContent ne "")
		{
			eval 
			{
				$json = decode_json(encode_utf8($decodedContent));
				1;
			}
			or do 
			{
			Log3 $name, 5, $name. " : km200_GetSingleService - Data cannot be parsed by JSON on km200 for http://" . $param->{url};
			};

			### Check whether the type is a single value containing a string or float value
			if(($json -> {type} eq "stringValue") || ($json -> {type} eq "floatValue"))
			{
				my $JsonId         = $json->{id};
				my $JsonType       = $json->{type};
				my $JsonValue      = $json->{value};

				### Save json-hash for DbLog-Split
				$hash->{temp}{ServiceDbLogSplitHash} = $json;
				$hash->{temp}{JsonRaw} = $decodedContent;
				
				
				### Write reading for fhem
				readingsSingleUpdate( $hash, $JsonId, $JsonValue, 1);

				return $json
			}	
			
			### Check whether the type is an switchProgram
			elsif ($json -> {type} eq "switchProgram")
			{
				my $JsonId         = $json->{id};
				my $JsonType       = $json->{type};

				### Log entries for debugging purposes
				Log3 $name, 4, $name. " : km200_GetSingleService  - value found for          : " .$Service;
				Log3 $name, 5, $name. " : km200_GetSingleService  - id                       : " .$JsonId;
				Log3 $name, 5, $name. " : km200_GetSingleService  - type                     : " .$JsonType;
				
				### Set up variables
				my $TempReturnVal = "";
				my $TempReadingMo = "";
				my $TempReadingTu = "";
				my $TempReadingWe = "";
				my $TempReadingTh = "";
				my $TempReadingFr = "";
				my $TempReadingSa = "";
				my $TempReadingSu = "";
				
				foreach my $item (@{ $json->{switchPoints} })
				{
					### Create string for time and switchpoint in fixed format and write part of Reading String
					
					my $time = $item->{time};
					my $temptime     = $time / 60;
					my $temptimeHH   = int($temptime);
					my $temptimeMM   = ($time - ($temptimeHH * 60));
					
					$temptimeHH = sprintf ('%02d', $temptimeHH);
					$temptimeMM = sprintf ('%02d', $temptimeMM);
					$temptime = $temptimeHH . $temptimeMM;
					
					my $tempsetpoint =  $item->{setpoint};
					$tempsetpoint    =~ s/^(.+)$/sprintf("%s%s", $1, ' 'x(8-length($1)))/e;
					my $TempReading  = $temptime . " " . $tempsetpoint;

					### Create ValueString for this day
					if ($item->{dayOfWeek} eq "Mo")
					{
						### If it is the first entry for this day
						if ($TempReadingMo eq "")
						{
							### Write the first entry
							$TempReadingMo = $TempReading;
						}
						### If it is NOT the first entry for this day
						else
						{
							### Add the next entry
							$TempReadingMo = $TempReadingMo . " " . $TempReading;
						}
					}
					elsif ($item->{dayOfWeek} eq "Tu")
					{
						### If it is the first entry for this day
						if ($TempReadingTu eq "")
						{
							### Write the first entry
							$TempReadingTu = $TempReading;
						}
						### If it is NOT the first entry for this day
						else
						{
							### Add the next entry
							$TempReadingTu = $TempReadingTu . " " . $TempReading;
						}
					}
					elsif ($item->{dayOfWeek} eq "We")
					{
						### If it is the first entry for this day
						if ($TempReadingWe eq "")
						{
							### Write the first entry
							$TempReadingWe = $TempReading;
						}
						### If it is NOT the first entry for this day
						else
						{
							### Add the next entry
							$TempReadingWe = $TempReadingWe . " " . $TempReading;
						}
					}
					elsif ($item->{dayOfWeek} eq "Th")
					{
						### If it is the first entry for this day
						if ($TempReadingTh eq "")
						{
							### Write the first entry
							$TempReadingTh = $TempReading;
						}
						### If it is NOT the first entry for this day
						else
						{
							### Add the next entry
							$TempReadingTh = $TempReadingTh . " " . $TempReading;
						}
					}
					elsif ($item->{dayOfWeek} eq "Fr")
					{
						### If it is the first entry for this day
						if ($TempReadingFr eq "")
						{
							### Write the first entry
							$TempReadingFr = $TempReading;
						}
						### If it is NOT the first entry for this day
						else
						{
							### Add the next entry
							$TempReadingFr = $TempReadingFr . " " . $TempReading;
						}
					}
					elsif ($item->{dayOfWeek} eq "Sa")
					{
						### If it is the first entry for this day
						if ($TempReadingSa eq "")
						{
							### Write the first entry
							$TempReadingSa = $TempReading;
						}
						### If it is NOT the first entry for this day
						else
						{
							### Add the next entry
							$TempReadingSa = $TempReadingSa . " " . $TempReading;
						}
					}
					elsif ($item->{dayOfWeek} eq "Su")
					{
						### If it is the first entry for this day
						if ($TempReadingSu eq "")
						{
							### Write the first entry
							$TempReadingSu = $TempReading;
						}
						### If it is NOT the first entry for this day
						else
						{
							### Add the next entry
							$TempReadingSu = $TempReadingSu . " " . $TempReading;
						}
					}
					else
					{
						Log3 $name, 5, $name. "dayOfWeek of unknow day: " . $item->{dayOfWeek};
					}
				}

				### Create new Service and write reading for fhem
				$TempReturnVal =                  "1-Mo: " . $TempReadingMo . "\n";
				$TempReturnVal = $TempReturnVal . "2-Tu: " . $TempReadingTu . "\n";
				$TempReturnVal = $TempReturnVal . "3-We: " . $TempReadingWe . "\n";
				$TempReturnVal = $TempReturnVal . "4-Th: " . $TempReadingTh . "\n";
				$TempReturnVal = $TempReturnVal . "5-Fr: " . $TempReadingFr . "\n";
				$TempReturnVal = $TempReturnVal . "6-Sa: " . $TempReadingSa . "\n";
				$TempReturnVal = $TempReturnVal . "7-Su: " . $TempReadingSu . "\n";

				### Save weeklist in "value"
				$json->{value} = $TempReturnVal;
				
				### Save raw Json string
				$hash->{temp}{JsonRaw} = $decodedContent;
	
				
				my $TempJsonId;
				
				### Create new Service and write reading for fhem
				$TempJsonId = $JsonId . "/" . "1-Mo";
				readingsSingleUpdate( $hash, $TempJsonId, $TempReadingMo, 1);

				### Create new Service and write reading for fhem
				$TempJsonId = $JsonId . "/" . "2-Tu";
				readingsSingleUpdate( $hash, $TempJsonId, $TempReadingTu, 1);

				### Create new Service and write reading for fhem
				$TempJsonId = $JsonId . "/" . "3-We";
				readingsSingleUpdate( $hash, $TempJsonId, $TempReadingWe, 1);

				### Create new Service and write reading for fhem
				$TempJsonId = $JsonId . "/" . "4-Th";
				readingsSingleUpdate( $hash, $TempJsonId, $TempReadingTh, 1);

				### Create new Service and write reading for fhem
				$TempJsonId = $JsonId . "/" . "5-Fr";
				readingsSingleUpdate( $hash, $TempJsonId, $TempReadingFr, 1);

				### Create new Service and write reading for fhem
				$TempJsonId = $JsonId . "/" . "6-Sa";
				readingsSingleUpdate( $hash, $TempJsonId, $TempReadingSa, 1);

				### Create new Service and write reading for fhem
				$TempJsonId = $JsonId . "/" . "7-Su";
				readingsSingleUpdate( $hash, $TempJsonId, $TempReadingSu, 1);

				return $json
			}
			### Check whether the type is an errorlist
			elsif ($json -> {type} eq "errorList")
			{
				my $TempErrorList    = "";
				
				### Sort list by timestamps descending
				my @TempSortedErrorList =  sort { $b->{t} cmp $a->{t} } @{ $json->{values} };
#				my @TempSortedErrorList =  @{ $json->{values} } ;
			
				### For every notification do
				foreach my $item (@TempSortedErrorList)
				{					
					### Create message string with fixed blocksize
					my $TempTime      = $item->{t};
					if ($TempTime) {$TempTime      =~ s/^(.+)$/sprintf("%s%s", $1, ' 'x(20-length($1)))/e;}
					my $TempErrorCode = $item->{dcd};
					   $TempErrorCode =~ s/^(.+)$/sprintf("%s%s", $1, ' 'x(3 -length($1)))/e;
					my $TempAddCode   = $item->{ccd};    
					   $TempAddCode   =~ s/^(.+)$/sprintf("%s%s", $1, ' 'x(4 -length($1)))/e;
					my $TempClassCode = $item->{cat};    
					   $TempClassCode =~ s/^(.+)$/sprintf("%s%s", $1, ' 'x(2- length($1)))/e;
					my $TempErrorMessage = "Time: " . $TempTime . "-ErrorCode: " . $TempErrorCode . " -AddCode: " . $TempAddCode . " -Category: " . $TempClassCode;

					### Create List
					$TempErrorList = $TempErrorList . $TempErrorMessage . "\n";
				}
				### Save raw Json string
				$hash->{temp}{JsonRaw} = $decodedContent;
				
				### Save errorList
				$json->{value} = $TempErrorList;
				
				return $json;
			}
			### Check whether the type is an refEnum which is indicating an empty parent directory
			elsif ($json -> {type} eq "refEnum")
			{
				### Initialise Return Message
				my $ReturnMessage = "";

				### For each item found in this empty parent directory
				foreach my $item (@{ $json->{references} })
				{
					### If it is the first item in the list
					if ($ReturnMessage eq "")
					{				
						$ReturnMessage = $item->{id};
					}
					### If it is not the first item in the list
					else
					{
						$ReturnMessage = $ReturnMessage . "\n" . $item->{id};
					}
				}
				### Return list of available directories
				$json->{value} = $ReturnMessage;
				
				### Save raw Json string
				$hash->{temp}{JsonRaw} = $decodedContent;
				
				return $json;
			}		
			### Check whether the type is a systeminfo
			elsif ($json -> {type} eq "systeminfo")
			{
				my $JsonId         = $json->{id};
				my $JsonType       = $json->{type};
				my @JsonValues     = $json->{values};

				### Log entries for debugging purposes
				Log3 $name, 4, $name. " : km200_GetSingleService  - value found for          : " .$Service;
				Log3 $name, 5, $name. " : km200_GetSingleService  - id                       : " .$JsonId;
				Log3 $name, 5, $name. " : km200_GetSingleService  - type                     : " .$JsonType;

				### Initialise Return Message
				my $ReturnMessage = "";
								
				### Initialise ArrayCounter 
				my $ArrayCounter = 0;
				
				foreach my $ArrayItem (@{ $json->{values} })
				{
					### Incrementation of ArrayCounter
					$ArrayCounter++;

					### Get array from scalar
					my %ArrayHash = %{$ArrayItem};
					
					while( my( $SystemInfoHashKey, $SystemInfoHashValue ) = each %ArrayHash )
					{
						### Create new Service and write reading for fhem
						my $TempJsonId = $JsonId . "/" . sprintf ('%02d', $ArrayCounter) . "/" . $SystemInfoHashKey;
						readingsSingleUpdate( $hash, $TempJsonId, $SystemInfoHashValue, 1);

						### If it is the first item in the list
						if ($ReturnMessage eq "")
						{				
							$ReturnMessage = $TempJsonId . " = " . $SystemInfoHashValue;
						}
						### If it is not the first item in the list
						else
						{
							$ReturnMessage = $ReturnMessage . "\n" . $TempJsonId . " = " . $SystemInfoHashValue;
						}
					}
				}

				### Return list of available directories
				$json->{value} = $ReturnMessage;
				
				### Save raw Json string
				$hash->{temp}{JsonRaw} = $decodedContent;
				
				return $json;
			}
			### If the type is unknown
			else
			{
				### Log entries for debugging purposes
				Log3 $name, 4, $name. " : km200_GetSingleService - type unknown for : " .$Service;
				### Log entries for debugging purposes
			}
		}
		else 
		{
			Log3 $name, 4, $name. " : km200_GetSingleService: ". $Service . " NOT available";
		
			my $ReturnMessage = "ERROR";
			$json -> {type}   = $ReturnMessage;
			$json -> {value}  = $ReturnMessage;
			return $json;
		}
	}
}
####END####### Subroutine get individual data value ############################################################END#####


###START###### Subroutine initial contact of services via HttpUtils ###########################################START####
sub km200_GetInitService($)
{
	my ($hash, $def)                 = @_;
	my $km200_gateway_host           =   $hash->{URL} ;
	my $name                         =   $hash->{NAME} ;
	$hash->{status}{FlagInitRequest} = true;
	my @KM200_InitServices           = @{$hash->{Secret}{KM200ALLSERVICES}};
	my $ServiceCounterInit           = $hash->{temp}{ServiceCounterInit};
	my $PollingTimeout               = $hash->{POLLINGTIMEOUT};
	my $Service                      = $KM200_InitServices[$ServiceCounterInit];

	### Stop the current timer
	RemoveInternalTimer($hash);

	### If this this loop is accessed for the first time, stop the timer and set status
	if ($ServiceCounterInit == 0)
	{
		### Log file entry for debugging
		Log3 $name, 5, $name. "Sounding and importing of services started";
		
		### Set status of km200 fhem module
		$hash->{STATE} = "Sounding...";
	}


	### Get the values
	my $url ="http://" . $km200_gateway_host . $Service;

	my $param = {
					url        => $url,
					timeout    => $PollingTimeout,
					hash       => $hash,
					method     => "GET",
					header     => "agent: TeleHeater/2.2.3\r\nUser-Agent: TeleHeater/2.2.3\r\nAccept: application/json",
					callback   =>  \&km200_ParseHttpResponseInit
				};

	### Set flag for initialisation
	$hash->{status}{FlagInitRequest} = true;
	
	### Get the value
	HttpUtils_NonblockingGet($param);
}
####END####### Subroutine initial contact of services via HttpUtils ############################################END#####


###START###### Subroutine to download complete initial data set from gateway ##################################START####
# For all known, but not excluded services by attribute "DoNotPoll", try reading the respective values from gateway
sub km200_ParseHttpResponseInit($)
{
    my ($param, $err, $data)     = @_;
    my $hash                     =   $param->{hash};
    my $name                     =   $hash ->{NAME};
	my $ServiceCounterInit       =   $hash ->{temp}{ServiceCounterInit};
	my @KM200_RespondingServices = @{$hash ->{Secret}{KM200RESPONDINGSERVICES}};
	my @KM200_WriteableServices  = @{$hash ->{Secret}{KM200WRITEABLESERVICES}};
	my @KM200_InitServices       = @{$hash ->{Secret}{KM200ALLSERVICES}};	
	my $NumberInitServices       = "";
	my $Service                  = $KM200_InitServices[$ServiceCounterInit];
	my $type;
    my $json ->{type} = "";
	
	if($err ne "") 
	{
		### Create Log entry
		Log3 $name, 2, $name . " : km200_ParseHttpResponseInit - ERROR                : ".$Service. ": No proper Communication with Gateway: " .$err;
		
		### Set status of km200 fhem module
		$hash->{STATE} = "ERROR - Initial Connection failed... Try to re-connect in 10s";
		
		### Start the timer for polling again but wait 10s
		InternalTimer(gettimeofday()+10, "km200_GetInitService", $hash, 1);
		
		### Create Log entry
		Log3 $name, 2, $name . " : km200_ParseHttpResponseInit - ERROR                : Timer restarted to try again in 10s";
		Log3 $name, 5, $name . "______________________________________________________________________________________________________________________";
		return "ERROR";	
	}

	$hash->{temp}{decodedcontent} = $data;
	my $decodedContent = km200_Decrypt($hash);
	
	### Check whether the decoded content is not empty and therefore available
	if ($decodedContent ne "")
	{	
		eval 
		{
			$json = decode_json(encode_utf8($decodedContent));
			1;
		}
		or do 
		{
			Log3 $name, 4, $name. " : km200_ParseHttpResponseInit - CANNOT be parsed    : ". $Service;
		};

		### Check whether the type is a single value containing a string or float value
		if(($json -> {type} eq "stringValue") || ($json -> {type} eq "floatValue"))
		{
			my $JsonId         = $json->{id};
			my $JsonType       = $json->{type};
			my $JsonValue      = $json->{value};

			### Log entries for debugging purposes
			Log3 $name, 5, $name. " : km200_ParseHttpResponseInit - value found for      : " .$Service;
			Log3 $name, 5, $name. " : km200_ParseHttpResponseInit - id                   : " .$JsonId;
			Log3 $name, 5, $name. " : km200_ParseHttpResponseInit - type                 : " .$JsonType;
			Log3 $name, 5, $name. " : km200_ParseHttpResponseInit - value                : " .$JsonValue;

			### Add service to the list of responding services
			push (@KM200_RespondingServices, $Service);
			
			### Delete double entries in the list of responding services and sort in alphabetical order
			my %FilteredKM200RespondingServices;
			$FilteredKM200RespondingServices{$_}=0 for @KM200_RespondingServices;
			@KM200_RespondingServices = (keys %FilteredKM200RespondingServices);
			@KM200_RespondingServices = sort @KM200_RespondingServices;
			
			### Save json-hash for DbLog-Split
			$hash->{temp}{ServiceDbLogSplitHash} = $json;

			### Write reading for fhem
			readingsSingleUpdate( $hash, $JsonId, $JsonValue, 1);

			### Log file entry for debugging
			my $LogMessage = " : The following Service can be read";

			
			### Check whether service is writeable and write name of service in array
			if ($json->{writeable} == 1)
			{
				$LogMessage = $LogMessage . " and is writeable";
				push (@KM200_WriteableServices, $Service);
			}
			else
			{
				# Do nothing
				$LogMessage = $LogMessage . "                 ";
			}

			### Log file entry for debugging
			$LogMessage = $LogMessage . " : " . $JsonId;
			Log3 $name, 4, $name. $LogMessage;
		}	
		### Check whether the type is an switchProgram
		elsif ($json -> {type} eq "switchProgram")
		{
			my $JsonId         = $json->{id};
			my $JsonType       = $json->{type};
			my @JsonValues     = $json->{switchPoints};

			### Log entries for debugging purposes
			Log3 $name, 5, $name. " : km200_ParseHttpResponseInit - value found for      : " .$Service;
			Log3 $name, 5, $name. " : km200_ParseHttpResponseInit - id                   : " .$JsonId;
			Log3 $name, 5, $name. " : km200_ParseHttpResponseInit - type                 : " .$JsonType;
			Log3 $name, 5, $name. " : km200_ParseHttpResponseInit - value                : " .@JsonValues;

			### Add service to the list of responding services
			push (@KM200_RespondingServices, $Service);

			### Delete double entries in the list of responding services and sort in alphabetical order
			my %FilteredKM200RespondingServices;
			$FilteredKM200RespondingServices{$_}=0 for @KM200_RespondingServices;
			@KM200_RespondingServices = (keys %FilteredKM200RespondingServices);
			@KM200_RespondingServices = sort @KM200_RespondingServices;
			
			### Log file entry for debugging
			my $LogMessage = " : The following Service can be read";
			
			### Check whether service is writeable and write name of service in array
			if ($json->{writeable} == 1)
			{
				$LogMessage = $LogMessage . " and is writeable";
				push (@KM200_WriteableServices, $Service);
			}
			else
			{
				# Do nothing
				$LogMessage = $LogMessage . "                 ";
			}
			$LogMessage = $LogMessage . " : " .$JsonId;
			Log3 $name, 4, $name . $LogMessage;
			
			### Set up variables
			my $TempJsonId    = "";
			my $TempReadingMo = "";
			my $TempReadingTu = "";
			my $TempReadingWe = "";
			my $TempReadingTh = "";
			my $TempReadingFr = "";
			my $TempReadingSa = "";
			my $TempReadingSu = "";
			
			foreach my $item (@{ $json->{switchPoints} })
			{
				### Create string for time and switchpoint in fixed format and write part of Reading String
				my $time         = $item->{time};
				my $temptime     = $time / 60;
				my $temptimeHH   = int($temptime);
				my $temptimeMM   = ($time - ($temptimeHH * 60));

				$temptimeHH = sprintf ('%02d', $temptimeHH);
				$temptimeMM = sprintf ('%02d', $temptimeMM);
				$temptime = $temptimeHH . $temptimeMM;
				
				my $tempsetpoint =  $item->{setpoint};
				$tempsetpoint    =~ s/^(.+)$/sprintf("%s%s", $1, ' 'x(8-length($1)))/e;
				my $TempReading  = $temptime . " " . $tempsetpoint;

				### Create ValueString for this day
				if ($item->{dayOfWeek} eq "Mo")
				{
					### If it is the first entry for this day
					if ($TempReadingMo eq "")
					{
						### Write the first entry
						$TempReadingMo = $TempReading;
					}
					### If it is NOT the first entry for this day
					else
					{
						### Add the next entry
						$TempReadingMo = $TempReadingMo . " " . $TempReading;
					}
				}
				elsif ($item->{dayOfWeek} eq "Tu")
				{
					### If it is the first entry for this day
					if ($TempReadingTu eq "")
					{
						### Write the first entry
						$TempReadingTu = $TempReading;
					}
					### If it is NOT the first entry for this day
					else
					{
						### Add the next entry
						$TempReadingTu = $TempReadingTu . " " . $TempReading;
					}
				}
				elsif ($item->{dayOfWeek} eq "We")
				{
					### If it is the first entry for this day
					if ($TempReadingWe eq "")
					{
						### Write the first entry
						$TempReadingWe = $TempReading;
					}
					### If it is NOT the first entry for this day
					else
					{
						### Add the next entry
						$TempReadingWe = $TempReadingWe . " " . $TempReading;
					}
				}
				elsif ($item->{dayOfWeek} eq "Th")
				{
					### If it is the first entry for this day
					if ($TempReadingTh eq "")
					{
						### Write the first entry
						$TempReadingTh = $TempReading;
					}
					### If it is NOT the first entry for this day
					else
					{
						### Add the next entry
						$TempReadingTh = $TempReadingTh . " " . $TempReading;
					}
				}
				elsif ($item->{dayOfWeek} eq "Fr")
				{
					### If it is the first entry for this day
					if ($TempReadingFr eq "")
					{
						### Write the first entry
						$TempReadingFr = $TempReading;
					}
					### If it is NOT the first entry for this day
					else
					{
						### Add the next entry
						$TempReadingFr = $TempReadingFr . " " . $TempReading;
					}
				}
				elsif ($item->{dayOfWeek} eq "Sa")
				{
					### If it is the first entry for this day
					if ($TempReadingSa eq "")
					{
						### Write the first entry
						$TempReadingSa = $TempReading;
					}
					### If it is NOT the first entry for this day
					else
					{
						### Add the next entry
						$TempReadingSa = $TempReadingSa . " " . $TempReading;
					}
				}
				elsif ($item->{dayOfWeek} eq "Su")
				{
					### If it is the first entry for this day
					if ($TempReadingSu eq "")
					{
						### Write the first entry
						$TempReadingSu = $TempReading;
					}
					### If it is NOT the first entry for this day
					else
					{
						### Add the next entry
						$TempReadingSu = $TempReadingSu . " " . $TempReading;
					}
				}
				else
				{
					Log3 $name, 5, $name. "dayOfWeek of unknow day: " . $item->{dayOfWeek};
				}
			}

			### Create new Service and write reading for fhem
			$TempJsonId = $JsonId . "/" . "1-Mo";
			readingsSingleUpdate( $hash, $TempJsonId, $TempReadingMo, 1);
			### Log file entry for debugging
			Log3 $name, 5, $name. " : The following Service can be read and is writeable : " . $TempJsonId;
			### Add service to the list of writeable services
			push (@KM200_WriteableServices, $TempJsonId);

			
			### Create new Service and write reading for fhem
			$TempJsonId = $JsonId . "/" . "2-Tu";
			readingsSingleUpdate( $hash, $TempJsonId, $TempReadingTu, 1);
			### Log file entry for debugging
			Log3 $name, 5, $name. " : The following Service can be read and is writeable : " . $TempJsonId;
			### Add service to the list of writeable services
			push (@KM200_WriteableServices, $TempJsonId);
			
			
			### Create new Service and write reading for fhem
			$TempJsonId = $JsonId . "/" . "3-We";
			readingsSingleUpdate( $hash, $TempJsonId, $TempReadingWe, 1);
			### Log file entry for debugging
			Log3 $name, 5, $name. " : The following Service can be read and is writeable : " . $TempJsonId;
			### Add service to the list of writeable services
			push (@KM200_WriteableServices, $TempJsonId);

			
			### Create new Service and write reading for fhem
			$TempJsonId = $JsonId . "/" . "4-Th";
			readingsSingleUpdate( $hash, $TempJsonId, $TempReadingTh, 1);
			### Log file entry for debugging
			Log3 $name, 5, $name. " : The following Service can be read and is writeable : " . $TempJsonId;
			### Add service to the list of writeable services
			push (@KM200_WriteableServices, $TempJsonId);

			
			### Create new Service and write reading for fhem
			$TempJsonId = $JsonId . "/" . "5-Fr";
			readingsSingleUpdate( $hash, $TempJsonId, $TempReadingFr, 1);
			### Log file entry for debugging
			Log3 $name, 5, $name. " : The following Service can be read and is writeable : " . $TempJsonId;
			### Add service to the list of writeable services
			push (@KM200_WriteableServices, $TempJsonId);

			
			### Create new Service and write reading for fhem
			$TempJsonId = $JsonId . "/" . "6-Sa";
			readingsSingleUpdate( $hash, $TempJsonId, $TempReadingSa, 1);
			### Log file entry for debugging
			Log3 $name, 5, $name. " : The following Service can be read and is writeable : " . $TempJsonId;
			### Add service to the list of writeable services
			push (@KM200_WriteableServices, $TempJsonId);

			
			### Create new Service and write reading for fhem
			$TempJsonId = $JsonId . "/" . "7-Su";
			readingsSingleUpdate( $hash, $TempJsonId, $TempReadingSu, 1);
			### Log file entry for debugging
			Log3 $name, 5, $name. " : The following Service can be read and is writeable : " . $TempJsonId;
			### Add service to the list of writeable services
			push (@KM200_WriteableServices, $TempJsonId);

		}
		### Check whether the type is an errorlist
		elsif ($json -> {type} eq "errorList")
		{
			my $JsonId         = $json->{id};
			my $JsonType       = $json->{type};

			### Log entries for debugging purposes
			Log3 $name, 5, $name. " : km200_ParseHttpResponseInit - value found for      : " .$Service;
			Log3 $name, 5, $name. " : km200_ParseHttpResponseInit - id                   : " .$JsonId;
			Log3 $name, 5, $name. " : km200_ParseHttpResponseInit - type                 : " .$JsonType;

			### Add service to the list of responding services
			push (@KM200_RespondingServices, $Service);

			### Delete double entries in the list of responding services and sort in alphabetical order
			my %FilteredKM200RespondingServices;
			$FilteredKM200RespondingServices{$_}=0 for @KM200_RespondingServices;
			@KM200_RespondingServices = (keys %FilteredKM200RespondingServices);
			@KM200_RespondingServices = sort @KM200_RespondingServices;
			
			### Log file entry for debugging
			my $LogMessage = " : The following Service can be read";
			
			### Check whether service is writeable and write name of service in array
			if ($json->{writeable} == 1)
			{
				$LogMessage = $LogMessage . " and is writeable     ";
				push (@KM200_WriteableServices, $Service);
			}
			else
			{
				# Do nothing
				$LogMessage = $LogMessage . "                 ";
			}
			### Log file entry for debugging
			$LogMessage = $LogMessage . " : " . $JsonId;
			Log3 $name, 4, $name . $LogMessage;
			
			### Sort list by timestamps descending
			my $TempServiceIndex = 0;
			my @TempSortedErrorList =  sort { $b->{t} cmp $a->{t} } @{ $json->{values} };
           #my @TempSortedErrorList =  @{ $json->{values} };
			
			foreach my $item (@TempSortedErrorList)
			{
				### Increment Service-Index
				$TempServiceIndex++;
				
				### Create message string with fixed blocksize
				my $TempTime      = $item->{t};
				if ($TempTime) {$TempTime =~ s/^(.+)$/sprintf("%s%s", $1, ' 'x(20-length($1)))/e;} 
				my $TempErrorCode = $item->{dcd};
				   $TempErrorCode =~ s/^(.+)$/sprintf("%s%s", $1, ' 'x(3 -length($1)))/e;
				my $TempAddCode   = $item->{ccd};    
				   $TempAddCode   =~ s/^(.+)$/sprintf("%s%s", $1, ' 'x(4 -length($1)))/e;
				my $TempClassCode = $item->{cat};    
				   $TempClassCode =~ s/^(.+)$/sprintf("%s%s", $1, ' 'x(2- length($1)))/e;
				my $TempErrorMessage = "Time: " . $TempTime . "-ErrorCode: " . $TempErrorCode . " -AddCode: " . $TempAddCode . " -Category: " . $TempClassCode;
				
				### Create Service with Increment
				my $TempServiceString = $Service . "/Error-" . (sprintf("%02d", $TempServiceIndex));
				
				### Write Reading
				readingsSingleUpdate( $hash, $TempServiceString, $TempErrorMessage, 1);
				
				### Log file entry for debugging
				Log3 $name, 5, $name. " : The following Service can be read                       : " .$TempServiceString;
			}
		}
		### Check whether the type is an refEnum which is indicating an empty parent directory
		elsif ($json -> {type} eq "refEnum")
		{
			my $JsonId         = $json->{id};
			my $JsonType       = $json->{type};
			my @JsonReferences = $json->{references};
			
			### Log file entry for debugging
			Log3 $name, 5, $name. " : The following Service is an empty parent directory : " . $JsonId;
			
			### For each item found in this empty parent directory
			foreach my $item (@{ $json->{references} })
			{
				my $SearchWord = $item->{id};
				
				### If the Service found is listed as blocked service
				if ((grep {$_ eq $SearchWord} @{$hash->{Secret}{KM200DONOTPOLL}}) == 1)
				{
					### Do nothing
					
					### Log file entry for debugging
					Log3 $name, 5, $name. "The following Service has been found but is blacklisted: " . $item->{id};
				}
				### If the Service found is NOT listed as blocked service
				else
				{
					### Add service to the list of all known services
					push (@{$hash ->{Secret}{KM200ALLSERVICES}}, $item->{id});
				}
			}
			### Sort the list of all services alphabetically
			@{$hash ->{Secret}{KM200ALLSERVICES}} = sort @{$hash ->{Secret}{KM200ALLSERVICES}};
		}
		### Check whether the type is a systeminfo
		elsif ($json -> {type} eq "systeminfo")
		{
			my $JsonId         = $json->{id};
			my $JsonType       = $json->{type};
			my @JsonValues     = $json->{values};

			### Log entries for debugging purposes
			Log3 $name, 5, $name. " : km200_ParseHttpResponseInit - value found for      : " .$Service;
			Log3 $name, 5, $name. " : km200_ParseHttpResponseInit - id                   : " .$JsonId;
			Log3 $name, 5, $name. " : km200_ParseHttpResponseInit - type                 : " .$JsonType;
			Log3 $name, 5, $name. " : km200_ParseHttpResponseInit - value                : " .@JsonValues;

			### Add service to the list of responding services
			push (@KM200_RespondingServices, $Service);

			### Delete double entries in the list of responding services and sort in alphabetical order
			my %FilteredKM200RespondingServices;
			$FilteredKM200RespondingServices{$_}=0 for @KM200_RespondingServices;
			@KM200_RespondingServices = (keys %FilteredKM200RespondingServices);
			@KM200_RespondingServices = sort @KM200_RespondingServices;
			
			### Log file entry for debugging
			Log3 $name, 4, $name . " : The following Service can be read                  : " .$JsonId;

			### Initialise ArrayCounter 
			my $ArrayCounter = 0;
			
			foreach my $ArrayItem (@{ $json->{values} })
			{
				### Incrementation of ArrayCounter
				$ArrayCounter++;

				### Log file entry for debugging
				Log3 $name, 5, $name . " : The ArrayItem is                                   : " .       $ArrayItem  ;
				Log3 $name, 5, $name . " : The keys ArrayItem is                              : " . (keys %{$ArrayItem}) ;
				
				### Get array from scalar
				my %ArrayHash = %{$ArrayItem};
				
				while( my( $SystemInfoHashKey, $SystemInfoHashValue ) = each %ArrayHash )
				{
					### Log file entry for debugging
					Log3 $name, 5, $name . " : The ArrayHashKey is                                : " . $SystemInfoHashKey;
					Log3 $name, 5, $name . " : The ArrayHashValue is                              : " . $SystemInfoHashValue;

					### Create new Service and write reading for fhem
					my $TempJsonId = $JsonId . "/" . sprintf ('%02d', $ArrayCounter) . "/" . $SystemInfoHashKey;
					readingsSingleUpdate( $hash, $TempJsonId, $SystemInfoHashValue, 1);
					### Log file entry for debugging
					Log3 $name, 5, $name . " : The following Service can be read                  : " . $TempJsonId;

				}
			}
		}
		### Check whether the type is unknown
		else
		{
			### Log entries for debugging purposes
			Log3 $name, 4, $name. " : km200_ParseHttpResponseInit - type unknown for     : " .$Service;
		}
	}
	### Check whether the decoded content is empty and therefore NOT available
	else 
	{
		### Log entries for debugging purposes
		Log3 $name, 4, $name. " : km200_ParseHttpResponseInit -  NOT available       : ". $Service;
	}

	### Log entries for debugging purposes
	#Log3 $name, 5, $name. " : km200_ParseHttpResponseInit    : response          : " .$data;
	Log3 $name, 5, $name . "______________________________________________________________________________________________________________________";
	### Get the size of the array
	@KM200_InitServices       = @{$hash ->{Secret}{KM200ALLSERVICES}};
	$NumberInitServices       = @KM200_InitServices;
	
	### If the list of KM200ALLSERVICES has not been finished yet
	if ($ServiceCounterInit < ($NumberInitServices-1))
	{
		++$ServiceCounterInit;
		$hash->{temp}{ServiceCounterInit}           = $ServiceCounterInit;
		@{$hash->{Secret}{KM200RESPONDINGSERVICES}} = @KM200_RespondingServices;
		@{$hash->{Secret}{KM200WRITEABLESERVICES}}  = @KM200_WriteableServices;
		km200_GetInitService($hash);
	}
	### If the list of KM200ALLSERVICES is finished
	else
	{
		my @KM200_DynServices = @KM200_RespondingServices;
		
		### Save arrays of services in hash
		@{$hash->{Secret}{KM200RESPONDINGSERVICES}} = @KM200_RespondingServices;
		@{$hash->{Secret}{KM200WRITEABLESERVICES}}  = @KM200_WriteableServices;
		@{$hash->{Secret}{KM200DYNSERVICES}}        = @KM200_DynServices;

		
		### Reset flag for initialisation
		$hash->{status}{FlagInitRequest}            = false;
		
		
		###START###### Initiate the timer for continuous polling of dynamical values from KM200 ###################START####
		InternalTimer(gettimeofday()+($hash->{INTERVALDYNVAL}), "km200_GetDynService", $hash, 1);
		Log3 $name, 4, $name. " : km200 - Define: InternalTimer for dynamic values started with interval of: ".($hash->{INTERVALDYNVAL});
		####END####### Initiate the timer for continuous polling of dynamical values from KM200 ####################END#####

		### Reset fullResponse error message
		readingsSingleUpdate( $hash, "fullResponse", "OK", 1);
		
		### Log file entry for debugging
		Log3 $name, 5, $name. "Sounding and importing of services is completed";

		### Set status of km200 fhem module
		$hash->{STATE} = "Standby";

		### Disable flag
		$hash->{temp}{ServiceCounterInit} = false;
	}
	### If the Initialisation process has been interuppted with an error message
	if (ReadingsVal($name,"fullResponse",0) eq "ERROR")
	{
		### Reset fullResponse error message
		readingsSingleUpdate( $hash, "fullResponse", "Restarted after ERROR", 1);
		
		### Reset timer for init procedure and start over again until it works
		InternalTimer(gettimeofday()+5, "km200_GetInitService", $hash, 1);
		Log3 $name, 5, $name. " : km200 - Internal timer for Initialisation of services restarted after fullResponse - error.";
	}
	
	### Clear up temporary variables
	$hash->{temp}{decodedcontent} = "";	
	
	return;
}
####END####### Subroutine to download complete initial data set from gateway ###################################END#####


###START###### Subroutine obtaining dynamic services via HttpUtils ############################################START####
sub km200_GetDynService($)
{
	my ($hash, $def)                 = @_;
	my $km200_gateway_host           =   $hash->{URL};
	my $name                         =   $hash->{NAME};
	$hash->{STATE}                   = "Polling";
	my @KM200_DynServices            = @{$hash->{Secret}{KM200DYNSERVICES}};
	my $ServiceCounterDyn            =   $hash->{temp}{ServiceCounterDyn};
	my $PollingTimeout               =   $hash->{POLLINGTIMEOUT};
	
	### Stop the current timer
	RemoveInternalTimer($hash);

	### If at least one service to be polled is available
	if (@KM200_DynServices != 0)
	{
		my $Service                  =   $KM200_DynServices[$ServiceCounterDyn];
		
		### Log file entry for debugging
		if ($ServiceCounterDyn == 0)
		{
			Log3 $name, 5, $name. "Starting download of dynamic services";
		}
		
		### Log file entry for debugging
		Log3 $name, 5, $name . " - km200_GetDynService - Polling                      : " . $Service;
		
		my $url = "http://" . $km200_gateway_host . $Service;
		my $param = {
						url        => $url,
						timeout    => $PollingTimeout,
						hash       => $hash,
						method     => "GET",
						header     => "agent: TeleHeater/2.2.3\r\nUser-Agent: TeleHeater/2.2.3\r\nAccept: application/json",
						callback   =>  \&km200_ParseHttpResponseDyn
					};

		### Set Status Flag in order state running dynamic request
		$hash->{status}{FlagDynRequest}           = true;

		### Get data
		HttpUtils_NonblockingGet($param);
	}
	### If no service to be polled is available
	else
	{
		Log3 $name, 5, $name . " : No dynamic values available to be read. Skipping download.";
	}
}
####END####### Subroutine get dynamic data value ###############################################################END#####

###START###### Subroutine to download complete dynamic data set from gateway ##################################START####
# For all responding dynamic services read the respective values from gateway
sub km200_ParseHttpResponseDyn($)
{
    my ($param, $err, $data)    = @_;
    my $hash                    =   $param->{hash};
    my $name                    =   $hash ->{NAME};
	my $ServiceCounterDyn       =   $hash ->{temp}{ServiceCounterDyn};
	my @KM200_DynServices       = @{$hash ->{Secret}{KM200DYNSERVICES}};	
	my $NumberDynServices       = @KM200_DynServices;
	my $Service                 = $KM200_DynServices[$ServiceCounterDyn];
	my $type;
    my $json ->{type} = "";
	
	Log3 $name, 5, $name. " : Parsing response of dynamic service received for   : " . $Service;

	### Reset Status Flag
	$hash->{status}{FlagDynRequest}           = false;
	
	if($err ne "")
	{
		Log3 $name, 2, $name . " : ERROR: Service: ".$Service. ": No proper Communication with Gateway: " .$err;
        readingsSingleUpdate($hash, "fullResponse", "ERROR", 1);
	}

	$hash->{temp}{decodedcontent} = $data;
	my $decodedContent = km200_Decrypt($hash);
	
	if ($decodedContent ne "")
	{
		eval 
		{
			$json = decode_json(encode_utf8($decodedContent));
			1;
		}
		or do 
		{
			Log3 $name, 5, $name. " - km200_parseHttpResponseDyn : Data cannot be parsed by JSON on km200 for http://" . $param->{url};
		};
		
		### Check whether the type is a single value containing a string or float value
		if(($json -> {type} eq "stringValue") || ($json -> {type} eq "floatValue"))
		{
			my $JsonId         = $json->{id};
			my $JsonType       = $json->{type};
			my $JsonValue      = $json->{value};
			
			### Log entries for debugging purposes
			Log3 $name, 4, $name. " : km200_parseHttpResponseDyn  - value found for      : " .$Service;
			Log3 $name, 5, $name. " : km200_parseHttpResponseDyn  - id                   : " .$JsonId;
			Log3 $name, 5, $name. " : km200_parseHttpResponseDyn  - type                 : " .$JsonType;
			Log3 $name, 5, $name. " : km200_parseHttpResponseDyn  - value                : " .$JsonValue;
			### Log entries for debugging purposes

			### Save json-hash for DbLog-Split
			$hash->{temp}{ServiceDbLogSplitHash} = $json;
			### Save json-hash for DbLog-Split

			### Write reading
			readingsSingleUpdate( $hash, $JsonId, $JsonValue, 1);
			### Write reading
		}			
		### Check whether the type is an switchProgram
		elsif ($json -> {type} eq "switchProgram")
		{
			my $JsonId         = $json->{id};
			my $JsonType       = $json->{type};

			### Log entries for debugging purposes
			Log3 $name, 4, $name. " : km200_parseHttpResponseDyn  - value found for      : " .$Service;
			Log3 $name, 5, $name. " : km200_parseHttpResponseDyn  - id                   : " .$JsonId;
			Log3 $name, 5, $name. " : km200_parseHttpResponseDyn  - type                 : " .$JsonType;
			
			### Set up variables
			my $TempJsonId    = "";
			my $TempReadingMo = "";
			my $TempReadingTu = "";
			my $TempReadingWe = "";
			my $TempReadingTh = "";
			my $TempReadingFr = "";
			my $TempReadingSa = "";
			my $TempReadingSu = "";
			
			foreach my $item (@{ $json->{switchPoints} })
			{
				### Create string for time and switchpoint in fixed format and write part of Reading String
				my $time         = $item->{time};
				my $temptime     = $time / 60;
				my $temptimeHH   = int($temptime);
				my $temptimeMM   = ($time - ($temptimeHH * 60));

				$temptimeHH = sprintf ('%02d', $temptimeHH);
				$temptimeMM = sprintf ('%02d', $temptimeMM);
				$temptime = $temptimeHH . $temptimeMM;
				
				my $tempsetpoint =  $item->{setpoint};
				$tempsetpoint    =~ s/^(.+)$/sprintf("%s%s", $1, ' 'x(8-length($1)))/e;
				my $TempReading  = $temptime . " " . $tempsetpoint;

				### Create ValueString for this day
				if ($item->{dayOfWeek} eq "Mo")
				{
					### If it is the first entry for this day
					if ($TempReadingMo eq "")
					{
						### Write the first entry
						$TempReadingMo = $TempReading;
					}
					### If it is NOT the first entry for this day
					else
					{
						### Add the next entry
						$TempReadingMo = $TempReadingMo . " " . $TempReading;
					}
				}
				elsif ($item->{dayOfWeek} eq "Tu")
				{
					### If it is the first entry for this day
					if ($TempReadingTu eq "")
					{
						### Write the first entry
						$TempReadingTu = $TempReading;
					}
					### If it is NOT the first entry for this day
					else
					{
						### Add the next entry
						$TempReadingTu = $TempReadingTu . " " . $TempReading;
					}
				}
				elsif ($item->{dayOfWeek} eq "We")
				{
					### If it is the first entry for this day
					if ($TempReadingWe eq "")
					{
						### Write the first entry
						$TempReadingWe = $TempReading;
					}
					### If it is NOT the first entry for this day
					else
					{
						### Add the next entry
						$TempReadingWe = $TempReadingWe . " " . $TempReading;
					}
				}
				elsif ($item->{dayOfWeek} eq "Th")
				{
					### If it is the first entry for this day
					if ($TempReadingTh eq "")
					{
						### Write the first entry
						$TempReadingTh = $TempReading;
					}
					### If it is NOT the first entry for this day
					else
					{
						### Add the next entry
						$TempReadingTh = $TempReadingTh . " " . $TempReading;
					}
				}
				elsif ($item->{dayOfWeek} eq "Fr")
				{
					### If it is the first entry for this day
					if ($TempReadingFr eq "")
					{
						### Write the first entry
						$TempReadingFr = $TempReading;
					}
					### If it is NOT the first entry for this day
					else
					{
						### Add the next entry
						$TempReadingFr = $TempReadingFr . " " . $TempReading;
					}
				}
				elsif ($item->{dayOfWeek} eq "Sa")
				{
					### If it is the first entry for this day
					if ($TempReadingSa eq "")
					{
						### Write the first entry
						$TempReadingSa = $TempReading;
					}
					### If it is NOT the first entry for this day
					else
					{
						### Add the next entry
						$TempReadingSa = $TempReadingSa . " " . $TempReading;
					}
				}
				elsif ($item->{dayOfWeek} eq "Su")
				{
					### If it is the first entry for this day
					if ($TempReadingSu eq "")
					{
						### Write the first entry
						$TempReadingSu = $TempReading;
					}
					### If it is NOT the first entry for this day
					else
					{
						### Add the next entry
						$TempReadingSu = $TempReadingSu . " " . $TempReading;
					}
				}
				else
				{
					Log3 $name, 5, $name. "dayOfWeek of unknow day: " . $item->{dayOfWeek};
				}
			}

			### Create new Service and write reading for fhem
			$TempJsonId = $JsonId . "/" . "1-Mo";
			readingsSingleUpdate( $hash, $TempJsonId, $TempReadingMo, 1);
			
			### Create new Service and write reading for fhem
			$TempJsonId = $JsonId . "/" . "2-Tu";
			readingsSingleUpdate( $hash, $TempJsonId, $TempReadingTu, 1);
			
			### Create new Service and write reading for fhem
			$TempJsonId = $JsonId . "/" . "3-We";
			readingsSingleUpdate( $hash, $TempJsonId, $TempReadingWe, 1);
			
			### Create new Service and write reading for fhem
			$TempJsonId = $JsonId . "/" . "4-Th";
			readingsSingleUpdate( $hash, $TempJsonId, $TempReadingTh, 1);
			
			### Create new Service and write reading for fhem
			$TempJsonId = $JsonId . "/" . "5-Fr";
			readingsSingleUpdate( $hash, $TempJsonId, $TempReadingFr, 1);
			
			### Create new Service and write reading for fhem
			$TempJsonId = $JsonId . "/" . "6-Sa";
			readingsSingleUpdate( $hash, $TempJsonId, $TempReadingSa, 1);
			
			### Create new Service and write reading for fhem
			$TempJsonId = $JsonId . "/" . "7-Su";
			readingsSingleUpdate( $hash, $TempJsonId, $TempReadingSu, 1);
		}
		### Check whether the type is an errorlist
		elsif ($json -> {type} eq "errorList")
		{
			my $JsonId           = $json->{id};
			my $JsonType         = $json->{type};
			my $TempServiceIndex = 0;

			### Sort list by timestamps descending
			my @TempSortedErrorList =  sort { $b->{t} cmp $a->{t} } @{ $json->{values} };
#			my @TempSortedErrorList =  @{ $json->{values} } ;
			
			### For every notification do
			foreach my $item (@TempSortedErrorList)
			{
				### Increment Service-Index
				$TempServiceIndex++;
				
				### Create message string with fixed blocksize
				my $TempTime      = $item->{t};
				if ($TempTime) 
				{
					$TempTime     =~ s/^(.+)$/sprintf("%s%s", $1, ' 'x(20-length($1)))/e;
				}
				else
				{
					$TempTime     = "unknown";
				}
				
				my $TempErrorCode = $item->{dcd};
				   $TempErrorCode =~ s/^(.+)$/sprintf("%s%s", $1, ' 'x(3 -length($1)))/e;
				my $TempAddCode   = $item->{ccd};    
				   $TempAddCode   =~ s/^(.+)$/sprintf("%s%s", $1, ' 'x(4 -length($1)))/e;
				my $TempClassCode = $item->{cat};    
				   $TempClassCode =~ s/^(.+)$/sprintf("%s%s", $1, ' 'x(2- length($1)))/e;
				my $TempErrorMessage = "Time: " . $TempTime . "-ErrorCode: " . $TempErrorCode . " -AddCode: " . $TempAddCode . " -Category: " . $TempClassCode;

				### Create Service with Increment and leading 0
				my $TempServiceString = $Service . "/Error-" . (sprintf("%02d", $TempServiceIndex));

				### Open readings bulk update
				readingsBeginUpdate($hash);
				
				### Write Reading
				readingsBulkUpdate( $hash, $TempServiceString, $TempErrorMessage, 1);

				### Obtain detailed Information about error codes
				my $classdescription;
				my $classworkaround;
				my $reason;
				my $checklist;
				my $correction;
				($classdescription, $classworkaround, $reason, $checklist, $correction) = km200_GetErrorMessage($TempErrorCode, $TempAddCode, $TempClassCode, $hash);

				### Write Reading
				readingsBulkUpdate( $hash, $TempServiceString . "_Class-Description", $classdescription, 1);
				readingsBulkUpdate( $hash, $TempServiceString . "_Class-Workaround",  $classworkaround,  1);
				readingsBulkUpdate( $hash, $TempServiceString . "_Error-Reason",      $reason,           1);
				readingsBulkUpdate( $hash, $TempServiceString . "_Error-Checklist",   $checklist,        1);
				readingsBulkUpdate( $hash, $TempServiceString . "_Error-Correction",  $correction,       1);

				### Close readings bulk update
				readingsEndUpdate($hash, 1);
			}
		}
		### Check whether the type is a systeminfo
		elsif ($json -> {type} eq "systeminfo")
		{
			my $JsonId         = $json->{id};
			my $JsonType       = $json->{type};
			my @JsonValues     = $json->{values};

			### Log entries for debugging purposes
			Log3 $name, 5, $name. " : km200_ParseHttpResponseInit - value found for      : " .$Service;
			Log3 $name, 5, $name. " : km200_ParseHttpResponseInit - id                   : " .$JsonId;
			Log3 $name, 5, $name. " : km200_ParseHttpResponseInit - type                 : " .$JsonType;

			### Initialise ArrayCounter 
			my $ArrayCounter = 0;
			
			foreach my $ArrayItem (@{ $json->{values} })
			{
				### Incrementation of ArrayCounter
				$ArrayCounter++;


				### Get array from scalar
				my %ArrayHash = %{$ArrayItem};
				
				while( my( $SystemInfoHashKey, $SystemInfoHashValue ) = each %ArrayHash )
				{
					### Create new Service and write reading for fhem
					my $TempJsonId = $JsonId . "/" . sprintf ('%02d', $ArrayCounter) . "/" . $SystemInfoHashKey;
					readingsSingleUpdate( $hash, $TempJsonId, $SystemInfoHashValue, 1);
				}
			}
		}
		### Check whether the type is unknown
		else
		{
			### Log entries for debugging purposes
			Log3 $name, 4, $name. " : km200_parseHttpResponseDyn  - type unknown for : " .$Service;
		}
	}
	else 
	{
		Log3 $name, 5, $name. " : km200_parseHttpResponseDyn  - Data not available on km200 for http://" . $param->{url};
	}

	Log3 $name, 5, $name . "______________________________________________________________________________________________________________________";
	### Clear up temporary variables
	$hash->{temp}{decodedcontent} = "";	
	$hash->{temp}{service}        = "";
	### Clear up temporary variables

	### If list is not complete yet
	if ($ServiceCounterDyn < ($NumberDynServices-1))
	{
		++$ServiceCounterDyn;
		$hash->{temp}{ServiceCounterDyn} = $ServiceCounterDyn;
		km200_GetDynService($hash);
	}
	### If list is complete
	else
	{
		$hash->{STATE}                   = "Standby";
		$hash->{temp}{ServiceCounterDyn} = 0;
		
		###START###### Re-Start the timer #####################################START####
		InternalTimer(gettimeofday()+$hash->{INTERVALDYNVAL}, "km200_GetDynService", $hash, 1);
		####END####### Re-Start the timer ######################################END#####
		
		### Update fullResponse Reading
		readingsSingleUpdate( $hash, "fullResponse", "OK", 1);
		
		$hash->{status}{FlagDynRequest}  = false;
		
		Log3 $name, 5, $name . "______________________________________________________________________________________________________________________";
	}
	return undef;
}
####END####### Subroutine to download complete dynamic data set from gateway ###################################END#####


###START###### Load database for errorcodes in hash ###########################################################START####
sub km200_GetErrorMessage($$$$)
{
	my ($ErrorCode, $SubCode, $Class, $hash) = @_;
	my $name                         =   $hash ->{NAME};
	my %ErrorCodeList;
	my %ClassCodeList;

	### Define database for error codes as hash
	$ErrorCodeList{"0001"} = {errorcode => "--",                                     subcode => "200",                                                class => "8",           reason => "W&auml;rmeerzeuger im Heizbetrieb",                                                                                                                                                                                                                                                                                                                checklist => "W&auml;rmeerzeuger ist im Heizbetrieb",                                                                                                                                                                                                                                                                                                                                                                                                                                                               correction => "Betriebsmeldung, keine St&ouml;rung"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"0002"} = {errorcode => "--",                                     subcode => "201",                                                class => "8",           reason => "W&auml;rmeerzeuger im WW-Betrieb",                                                                                                                                                                                                                                                                                                                 checklist => "W&auml;rmeerzeuger ist im Warmwasserbetrieb",                                                                                                                                                                                                                                                                                                                                                                                                                                                         correction => "Betriebsmeldung, keine St&ouml;rung"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"0003"} = {errorcode => "--",                                     subcode => "202",                                                class => "8",           reason => "Ger&auml;t im Schaltoptimierungsprogramm",                                                                                                                                                                                                                                                                                                         checklist => "Einstellungen am  am Basiscontroller/ Steuerger&auml;t &uuml;berpr&uuml;fen.",                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Einstellungen im Basiscontroller/Steuerger&auml;t auf die Anlagenbedingungen anpassen"                                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"0004"} = {errorcode => "--",                                     subcode => "203",                                                class => "8",           reason => "Ger&auml;t in Betriebsbereitschaft, kein W&auml;rmebedarf vorhanden",                                                                                                                                                                                                                                                                              checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Betriebsmeldung, keine St&ouml;rung"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"0005"} = {errorcode => "--",                                     subcode => "204",                                                class => "8",           reason => "Aktuelle Heizwassertemperatur des W&auml;rmeerzeugers h&ouml;her als Sollwert",                                                                                                                                                                                                                                                                    checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Betriebsmeldung, keine St&ouml;rung"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"0006"} = {errorcode => "--",                                     subcode => "208",                                                class => "8",           reason => "W&auml;rmeanforderung wegen Abgastest",                                                                                                                                                                                                                                                                                                            checklist => "Schornsteinfegerfunktion ist aktiviert",                                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0007"} = {errorcode => "--",                                     subcode => "265",                                                class => "8",           reason => "W&auml;rmebedarf geringer als gelieferte Energie",                                                                                                                                                                                                                                                                                                 checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0008"} = {errorcode => "--",                                     subcode => "268",                                                class => "8",           reason => "Relaistest wurde aktiviert",                                                                                                                                                                                                                                                                                                                       checklist => "Im Servicemen&uuml; wurde der Relaistest aktiviert. Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                             correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0009"} = {errorcode => "--",                                     subcode => "283",                                                class => "8",           reason => "Brenner wird gestartet",                                                                                                                                                                                                                                                                                                                           checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0010"} = {errorcode => "--",                                     subcode => "284",                                                class => "8",           reason => "Gasarmatur/&Ouml;lMagnetventile wird ge&ouml;ffnet; Erste Sicherheitszeit",                                                                                                                                                                                                                                                                        checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Betriebsmeldung, keine St&ouml;rung"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"0011"} = {errorcode => "--",                                     subcode => "305",                                                class => "8",           reason => "Kessel kann vor&uuml;bergehend nach Warmwasservorrang nicht starten",                                                                                                                                                                                                                                                                              checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Betriebsmeldung, keine St&ouml;rung"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"0012"} = {errorcode => "--",                                     subcode => "309",                                                class => "8",           reason => "Heizund W&auml;rmebetrieb parallel",                                                                                                                                                                                                                                                                                                               checklist => "Betriebscode, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0013"} = {errorcode => "--",                                     subcode => "357",                                                class => "8",           reason => "Entl&uuml;ftungsprogramm",                                                                                                                                                                                                                                                                                                                         checklist => "Entl&uuml;ftungsprogramm: Das Entl&uuml;ftungsprogramm des W&auml;rmetauchers l&auml;uft. Hierbei taktet die Pumpe und das 3Wegeventil",                                                                                                                                                                                                                                                                                                                                                              correction => "Betriebsmeldung, keine St&ouml;rung"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"0014"} = {errorcode => "--",                                     subcode => "358",                                                class => "8",           reason => "Blockierschutz aktiv",                                                                                                                                                                                                                                                                                                                             checklist => "Blockierschutz f&uuml;r die Heizungspumpe und 3-Wegeventil aktiv",                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "Betriebsmeldung, keine St&ouml;rung"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"0015"} = {errorcode => "--",                                     subcode => "370",                                                class => "4",           reason => "Zu geringer K&auml;ltemitteldruck",                                                                                                                                                                                                                                                                                                                checklist => "Verdampfer vereist&bull;K&auml;ltemittelkreis undicht",                                                                                                                                                                                                                                                                                                                                                                                                                                               correction => "Eisbildung am Verdampfer entfernen&bull;W&auml;rmepumpeneinheit austauschen"                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0016"} = {errorcode => "--",                                     subcode => "471",                                                class => "8",           reason => "Vorlaufphase Pumpe in der Inneneinheit",                                                                                                                                                                                                                                                                                                           checklist => "Betriebscode, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0017"} = {errorcode => "--",                                     subcode => "472",                                                class => "8",           reason => "Vorheizphase der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                 checklist => "Betriebscode, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0018"} = {errorcode => "--",                                     subcode => "473",                                                class => "8",           reason => "W&auml;rmepumpe in Betrieb",                                                                                                                                                                                                                                                                                                                       checklist => "Betriebscode, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0019"} = {errorcode => "--",                                     subcode => "474",                                                class => "8",           reason => "Pumpe Hybrid-Manager: Nachlaufphase",                                                                                                                                                                                                                                                                                                              checklist => "Betriebscode, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0020"} = {errorcode => "--",                                     subcode => "475",                                                class => "8",           reason => "W&auml;rmepumpe im Abtaubetrieb",                                                                                                                                                                                                                                                                                                                  checklist => "Betriebscode, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0021"} = {errorcode => "--",                                     subcode => "476",                                                class => "8",           reason => "Interner Status",                                                                                                                                                                                                                                                                                                                                  checklist => "Betriebscode, keine zSt&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0022"} = {errorcode => "--",                                     subcode => "477",                                                class => "8",           reason => "Nur EMS-W&auml;rmeerzeuger in Betrieb",                                                                                                                                                                                                                                                                                                            checklist => "Betriebscode, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0023"} = {errorcode => "--",                                     subcode => "478",                                                class => "8",           reason => "W&auml;rmepumpe blockiert: mehr als 4 Starts pro Stunde",                                                                                                                                                                                                                                                                                          checklist => "Betriebscode, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0024"} = {errorcode => "--",                                     subcode => "480",                                                class => "8",           reason => "Differenztemperatur au&szlig;erhalb des zul&auml;ssigen Bereichs",                                                                                                                                                                                                                                                                                 checklist => "Indikator f&uuml;r geringen Durchfluss im System. Filter auf Verstopfung pr&uuml;fen.",                                                                                                                                                                                                                                                                                                                                                                                                               correction => "Filter ggf. reinigen"                                                                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"0025"} = {errorcode => "--",                                     subcode => "481",                                                class => "8",           reason => "W&auml;rmepumpe l&auml;uft bei blockiertem Heizkessel",                                                                                                                                                                                                                                                                                            checklist => "Betriebscode, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0026"} = {errorcode => "--",                                     subcode => "482",                                                class => "8",           reason => "W&auml;rmepumpe arbeitet im Servicemodus",                                                                                                                                                                                                                                                                                                         checklist => "Betriebscode, keine St&ouml;rung Die W&auml;rmepumpe arbeitet im Servicemodus (100 % Leistung)",                                                                                                                                                                                                                                                                                                                                                                                                      correction => "Nach erfolgten Servicearbeiten auf Normalbetrieb umschalten. Der Servicemodus endet automatisch nach 15 Minuten."                                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"0027"} = {errorcode => "--",                                     subcode => "2908",                                               class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik / Basiscontroller",                                                                                                                                                                                                                                                                                       checklist => "Neustart Ger&auml;t, um zu sehen, ob das Problem erneut auftritt.",                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Bleibt die St&ouml;rung nach Reset erhalten, ist der Feuerungsautomat oder Fremdbrennermodul defekt und muss ausgetauscht werden"                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"0028"} = {errorcode => "--",                                     subcode => "2909",                                               class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik / Basiscontroller",                                                                                                                                                                                                                                                                                       checklist => "Restart-Appliance, um zu sehen, wenn das Problem erneut auftritt.",                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Bleibt die St&ouml;rung nach Reset erhalten, ist der Feuerungsautomat oder Fremdbrennermodul defekt und muss ausgetauscht werden"                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"0029"} = {errorcode => "--",                                     subcode => "2910",                                               class => "8",           reason => "Fehler im Abgassystem",                                                                                                                                                                                                                                                                                                                            checklist => "&Uuml;berpr&uuml;fen, ob ein Abgassystem angeschlossen ist&bull;Abgassystem auf freien Querschnitt &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                             correction => "Abgassystem montieren&bull;Ablagerungen im Abgasssystem entfernen"                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0030"} = {errorcode => "--",                                     subcode => "2911",                                               class => "8",           reason => "Kalibrierung fehlgeschlagen",                                                                                                                                                                                                                                                                                                                      checklist => "Verkabelung pr&uuml;fen&bull;Ionistationselektrode pr&uuml;fen&bull;Gas-Ventil pr&uuml;fen&bull;Feuerungsautomat pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                        correction => "Fehlerhafte Komponente ersetzen&bull;-"                                                                                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"0031"} = {errorcode => "--",                                     subcode => "2912",                                               class => "8",           reason => "Kein Flammensignal w&auml;hrend der Kalibrierung",                                                                                                                                                                                                                                                                                                 checklist => "Verkabelung pr&uuml;fen&bull;Ionistationselektrode pr&uuml;fen&bull;Gas-Ventil pr&uuml;fen&bull;Feuerungsautomat pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                        correction => "Fehlerhafte Komponente ersetzen&bull;-"                                                                                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"0032"} = {errorcode => "--",                                     subcode => "2913",                                               class => "8",           reason => "Flammensignal zu niedrig in der Kalibrierung",                                                                                                                                                                                                                                                                                                     checklist => "Verkabelung pr&uuml;fen&bull;Ionistationselektrode pr&uuml;fen&bull;Gas-Ventil pr&uuml;fen&bull;Feuerungsautomat pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                        correction => "Ionistationselektrode ersetzen&bull;-"                                                                                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"0033"} = {errorcode => "--",                                     subcode => "2914",                                               class => "4",           reason => "Systemst&ouml;rung Ger&auml;teelektronik",                                                                                                                                                                                                                                                                                                         checklist => "Reset am W&auml;rmeerzeuger durchf&uuml;hren.",                                                                                                                                                                                                                                                                                                                                                                                                                                                       correction => "Bleibt die St&ouml;rung nach Reset erhalten, ist die Ger&auml;testeuerung oder das Brennermodul defekt und muss ausgetauscht werden"                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"0034"} = {errorcode => "--",                                     subcode => "2915",                                               class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik",                                                                                                                                                                                                                                                                                                         checklist => "Verkabelung pr&uuml;fen&bull;Feuerungsautomat pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                           correction => "1. Reset am Ger&auml;t ausf&uuml;hren; 2. W&auml;rmeanforderung abwarten/erzeugen; 3. Ende der W&auml;rmeanforderung abwarten/ W&auml;rmeanforderung beenden.; Wenn der Fehler erneut erscheint, muss der Feuerungsautomat getauscht werden&bull;-"                                                                                                                                              };
	$ErrorCodeList{"0035"} = {errorcode => "--",                                     subcode => "2916",                                               class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik",                                                                                                                                                                                                                                                                                                         checklist => "Verkabelung pr&uuml;fen&bull;Feuerungsautomat pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                           correction => "1. Reset am Ger&auml;t ausf&uuml;hren; 2. W&auml;rmeanforderung abwarten/erzeugen; 3. Ende der W&auml;rmeanforderung abwarten/ W&auml;rmeanforderung beenden.; Wenn der Fehler erneut erscheint, muss der Feuerungsautomat getauscht werden&bull;-"                                                                                                                                              };
	$ErrorCodeList{"0036"} = {errorcode => "--",                                     subcode => "2917",                                               class => "8",           reason => "Kein Flammensignal w&auml;hrend der &Uuml;berpr&uuml;fung der Verbrennungsregelung",                                                                                                                                                                                                                                                               checklist => "Gas-Ventil pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "1. Ger&auml;t auslaufen lassen; 2. W&auml;rmeanforderung erzeugen/abwarten; 3. 5 Minuten abwarten. Wenn der Fehler erneut auftritt, dann Ger&auml;t resetten. Bei erneutem Auftreten des Fehlers Feuerungsautomat tauschen"                                                                                                                                                                      };
	$ErrorCodeList{"0037"} = {errorcode => "--",                                     subcode => "2918",                                               class => "8",           reason => "St&ouml;rung in der Verbrennungsregelung",                                                                                                                                                                                                                                                                                                         checklist => "Siphon pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Siphon und Kondensatabflu&szlig; reinigen"                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0038"} = {errorcode => "--",                                     subcode => "2919",                                               class => "8",           reason => "St&ouml;rung in der Verbrennungsregelung",                                                                                                                                                                                                                                                                                                         checklist => "Verkabelung pr&uuml;fen&bull;Ionistationselektrode pr&uuml;fen&bull;Gas-Ventil pr&uuml;fen&bull;Feuerungsautomat pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                        correction => "Fehlerhafte Komponente tauschen&bull;-"                                                                                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"0039"} = {errorcode => "--",                                     subcode => "2920",                                               class => "8",           reason => "St&ouml;rung Flammen&uuml;berwachung",                                                                                                                                                                                                                                                                                                             checklist => "Verkabelung pr&uuml;fen&bull;Brennerd&uuml;se pr&uuml;fen&bull;Feuerungsautomat pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                         correction => "Feuerungsautomat austauschen&bull;-"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"0040"} = {errorcode => "--",                                     subcode => "2921",                                               class => "4",           reason => "Ger&auml;teelektronik im Testmode",                                                                                                                                                                                                                                                                                                                checklist => "Warten bis Test beendet",                                                                                                                                                                                                                                                                                                                                                                                                                                                                             correction => "Keine Ma&szlig;nahme erforderlich (W&auml;rmeerzeuger startet automatisch)"                                                                                                                                                                                                                                                                                                                      };
	$ErrorCodeList{"0041"} = {errorcode => "--",                                     subcode => "2922",                                               class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik",                                                                                                                                                                                                                                                                                                         checklist => "Verkabelung pr&uuml;fen&bull;Feuerungsautomat pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                           correction => "Feuerungsautomat austauschen&bull;-"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"0042"} = {errorcode => "--",                                     subcode => "2923",                                               class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik",                                                                                                                                                                                                                                                                                                         checklist => "Reset am W&auml;rmeerzeuger durchf&uuml;hren.",                                                                                                                                                                                                                                                                                                                                                                                                                                                       correction => "Ger&auml;teelektronik austauschen"                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0043"} = {errorcode => "--",                                     subcode => "2924",                                               class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik",                                                                                                                                                                                                                                                                                                         checklist => "Reset am W&auml;rmeerzeuger durchf&uuml;hren.",                                                                                                                                                                                                                                                                                                                                                                                                                                                       correction => "Ger&auml;teelektronik austauschen"                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0044"} = {errorcode => "--",                                     subcode => "2925",                                               class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik",                                                                                                                                                                                                                                                                                                         checklist => "Verkabelung pr&uuml;fen&bull;Ionistationselektrode pr&uuml;fen&bull;Feuerungsautomat pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                    correction => "Ger&auml;teelektronik austauschen&bull;-"                                                                                                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"0045"} = {errorcode => "--",                                     subcode => "2926",                                               class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik",                                                                                                                                                                                                                                                                                                         checklist => "Verkabelung pr&uuml;fen&bull;Ionistationselektrode pr&uuml;fen&bull;Feuerungsautomat pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                    correction => "Ger&auml;teelektronik austauschen&bull;-"                                                                                                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"0046"} = {errorcode => "--",                                     subcode => "2626",                                               class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik",                                                                                                                                                                                                                                                                                                         checklist => "Verkabelung pr&uuml;fen&bull;Ionistationselektrode pr&uuml;fen&bull;Feuerungsautomat pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                    correction => "Ger&auml;teelektronik austauschen&bull;-"                                                                                                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"0047"} = {errorcode => "--",                                     subcode => "2927",                                               class => "4",           reason => "Flamme w&auml;hrend Brennerbetrieb ausgefallen",                                                                                                                                                                                                                                                                                                   checklist => "Hauptabsperreinrichtung &uuml;berpr&uuml;fen&bull;Ger&auml;teabsperrhahn &uuml;berpr&uuml;fen&bull;Gasanschlu&szlig;flie&szlig;druck bei Nennw&auml;rmebelastung messen&bull;Ionisationsstrom messen&bull;Ionisationselektrode &uuml;berpr&uuml;fen&bull;Schutzleiteranschlu&szlig; im Steuerger&auml;t &uuml;berpr&uuml;fen&bull;Z&uuml;ndkabel auf Besch&auml;digungen &uuml;berpr&uuml;fen&bull;Anschlusskabel f&uuml;r die Ionisationselektrode auf Besch&auml;digungen &uuml;berpr&uuml;fen&bull;Ohmschen Widerstand der Sicherheitsventile an der Gasarmatur messen&bull;Brennereinstellwert bei Nennw&auml;rmebelastung bzw. eingebaute Brennerd&uuml;sen &uuml;berpr&uuml;fen&bull;Brennereinstellwert bei kleinster Leistung &uuml;berpr&uuml;fen&bull;Abgasanlage &uuml;berpr&uuml;fen&bull;Verbrennungsluftversorgung &uuml;berpr&uuml;fen&bull;W&auml;rmeblock auf abgasseitige Ablagerungen &uuml;berpr&uuml;fen&bull;-",correction => "Hauptabsperreinrichtung &ouml;ffnen&bull;Ger&auml;teabsperrhahn &ouml;ffnen&bull;Ger&auml;t stillegen und Gasleitung &uuml;berpr&uuml;fen&bull;Signalauswertung auf Leiterplatte defekt&bull;Ionisationselektrode austauschen&bull;Schutzleiteranschluss (PE) im Schaltkasten herstellen&bull;Z&uuml;ndkabel austauschen&bull;Anschlusskabel zur Ionisationselektrode austauschen&bull;Gasarmatur austauschen&bull;Brenner korrekt einstellen bzw. Brennerd&uuml;sen austauschen&bull;Brenner bei minimaler Nennbelastung einstellen&bull;Abgasanlage umbauen&bull;Verbrennungsluftverbund zu klein bzw. zu geringe Gr&ouml;&szlig;e der L&uuml;ftungs&ouml;ffnung&bull;W&auml;rmeblock abgasseitig reinigen&bull;Steuerger&auml;t/Feuerungsautomat austauschen"};
	$ErrorCodeList{"0048"} = {errorcode => "--",                                     subcode => "2928",                                               class => "8",           reason => "Interner Fehler",                                                                                                                                                                                                                                                                                                                                  checklist => "Ger&auml;t neustarten, pr&uuml;fen ob Fehler wieder erscheint",                                                                                                                                                                                                                                                                                                                                                                                                                                       correction => "Steuerger&auml;t / Feuerungsautomat austauschen, falls das Problem erneut auftritt"                                                                                                                                                                                                                                                                                                              };
	$ErrorCodeList{"0049"} = {errorcode => "--",                                     subcode => "2929",                                               class => "8",           reason => "Siphon F&uuml;llprogramm",                                                                                                                                                                                                                                                                                                                         checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0050"} = {errorcode => "0",                                      subcode => "271; 471",                                           class => "BC",          reason => "Vorlaufphase Pumpe in der Inneneinheit",                                                                                                                                                                                                                                                                                                           checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0051"} = {errorcode => "0",                                      subcode => "370",                                                class => "8",           reason => "Zu geringer K&auml;ltemitteldruck",                                                                                                                                                                                                                                                                                                                checklist => "Verdampfer auf Vereisung &uuml;berpr&uuml;fen&bull;-",                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Eisbildung am Verdampfer entfernen&bull;W&auml;rmepumpeneinheit austauschen"                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0052"} = {errorcode => "0",                                      subcode => "2516",                                               class => "BC",          reason => "Interner Status",                                                                                                                                                                                                                                                                                                                                  checklist => "Interner Status",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                     correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0053"} = {errorcode => "1",                                      subcode => "272; 472",                                           class => "BC",          reason => "Vorheizphase der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                 checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0054"} = {errorcode => "1",                                      subcode => "470",                                                class => "10",          reason => "Keine Kommunikation mit dem Systemregler",                                                                                                                                                                                                                                                                                                         checklist => "Kontakte, Kabel und Verbindungen zwischen Hybridmodul und Systemregler pr&uuml;fen&bull;Systemregler pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                    correction => "Besch&auml;digte Leitungen und Kontakte austauschen&bull;Ggf. ersetzen"                                                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"0055"} = {errorcode => "2",                                      subcode => "273; 473",                                           class => "BC",          reason => "W&auml;rmepumpe in Betrieb",                                                                                                                                                                                                                                                                                                                       checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0056"} = {errorcode => "3",                                      subcode => "274; 474",                                           class => "BC",          reason => "Pumpe Hybrid-Manager: Nachlaufphase",                                                                                                                                                                                                                                                                                                              checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0057"} = {errorcode => "4",                                      subcode => "275; 475",                                           class => "BC",          reason => "W&auml;rmepumpe im Abtaubetrieb",                                                                                                                                                                                                                                                                                                                  checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0058"} = {errorcode => "5",                                      subcode => "276; 476",                                           class => "BC",          reason => "W&auml;rmepumpe im St&ouml;rungsmodus",                                                                                                                                                                                                                                                                                                            checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0059"} = {errorcode => "6",                                      subcode => "277; 477",                                           class => "BC",          reason => "Nur EMS-W&auml;rmeerzeuger in Betrieb",                                                                                                                                                                                                                                                                                                            checklist => "Pr&uuml;fen, ob ein Durchfluss gegeben ist. Ursache: Filter in der Logatherm  WHM (Inneneinheit Logatherm WPL); verschmutzt oder Absperrventile geschlossen&bull;Funktion und Anschluss des Str&ouml;mungsschalters &uuml;berpr&uuml;fen&bull;Au&szlig;eneinheit mit dem Servicetool pr&uuml;fen",                                                                                                                                                                                                    correction => "Filter reinigen und Absperrventile &ouml;ffnen&bull;Str&ouml;mungsschalter anschlie&szlig;en oder ggf. Str&ouml;mungschalter austauschen&bull;St&ouml;rung der Au&szlig;eneinheit beheben"                                                                                                                                                                                                       };
	$ErrorCodeList{"0060"} = {errorcode => "7",                                      subcode => "278; 478",                                           class => "BC",          reason => "W&auml;rmepumpe blockiert: mehr als 4 Starts pro Stunde",                                                                                                                                                                                                                                                                                          checklist => "W&auml;rmepumpe blockiert w&auml;hrend der Startversuche; dieser Zustand tritt auf, wenn das System mehr als 4-mal in einem 60-min&uuml;tigen Zeitraum startet; die Z&auml;hlung wird nach 60 Minuten zur&uuml;ckgesetzt; dieser Zustand tritt auch auf, um sicherzustellen, dass die Luft-Wasser-W&auml;rmepumpe f&uuml;r mindestens 20 Minuten ausgeschaltet ist",                                                                                                                                  correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0061"} = {errorcode => "8",                                      subcode => "479",                                                class => "BC",          reason => "St&ouml;rung Str&ouml;mungsschalter W&auml;rmepumpe w&auml;hrend Selbsttest",                                                                                                                                                                                                                                                                      checklist => "Str&ouml;mungsschalter pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Ggf. Str&ouml;mungsschalter austauschen"                                                                                                                                                                                                                                                                                                                                                         };
	$ErrorCodeList{"0062"} = {errorcode => "9",                                      subcode => "280; 480",                                           class => "BC",          reason => "Differenztemperatur au&szlig;erhalb des zul&auml;ssigen Bereichs",                                                                                                                                                                                                                                                                                 checklist => "Indikator f&uuml;r geringen Durchfluss im System; Filter auf Verstopfung pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Ggf. reinigen"                                                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"0063"} = {errorcode => "0A",                                     subcode => "202",                                                class => "BC",          reason => "Ger&auml;t im Schaltoptimierungsprogramm",                                                                                                                                                                                                                                                                                                         checklist => "Leistungseinstellung am Basiscontroller pr&uuml;fen&bull;Regelungseinstellungen im Regelger&auml;t Bedieneinheit pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                        correction => "Kesselleistung auf den erforderlichen W&auml;rmebedarf des Geb&auml;udes abstimmen&bull;Regelungseinstellungen an die Anlagenbedingungen anpassen"                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0064"} = {errorcode => "0A",                                     subcode => "305",                                                class => "BC",          reason => "Kessel kann vor&uuml;bergehend nach Warmwasser-Vorrang nicht starten",                                                                                                                                                                                                                                                                             checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0065"} = {errorcode => "0A",                                     subcode => "333",                                                class => "10",          reason => "Kessel vor&uuml;bergehend wegen zu geringen Wasserdrucks abgeschaltet",                                                                                                                                                                                                                                                                            checklist => "Kessel hat wegen kurzzeitig zu geringen Wasserdrucks abgeschaltet; Kessel l&auml;uft selbstst&auml;ndig nach kurzer Zeit wieder an",                                                                                                                                                                                                                                                                                                                                                                  correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0066"} = {errorcode => "0A",                                     subcode => "2505",                                               class => "BC",          reason => "W&auml;rmeanforderung blockiert wegen Antipendel",                                                                                                                                                                                                                                                                                                 checklist => "W&auml;rmeanforderung blockiert durch Antipendelzeit",                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0067"} = {errorcode => "0C",                                     subcode => "283",                                                class => "BC",          reason => "Brenner wird gestartet",                                                                                                                                                                                                                                                                                                                           checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0068"} = {errorcode => "0C",                                     subcode => "2517",                                               class => "BC",          reason => "Vorbel&uuml;ftung",                                                                                                                                                                                                                                                                                                                                checklist => "Vorbel&uuml;ftung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0069"} = {errorcode => "0C",                                     subcode => "2518",                                               class => "BC",          reason => "Warten Mischraumtemperatur",                                                                                                                                                                                                                                                                                                                       checklist => "Warten, dass Mischraumtemperatur erreicht wird",                                                                                                                                                                                                                                                                                                                                                                                                                                                      correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0070"} = {errorcode => "0C",                                     subcode => "2519",                                               class => "BC",          reason => "Flamme bilden",                                                                                                                                                                                                                                                                                                                                    checklist => "Flamme bilden",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0071"} = {errorcode => "0d",                                     subcode => "2524",                                               class => "BC",          reason => "Nachfackelkontrolle aus Startphase",                                                                                                                                                                                                                                                                                                               checklist => "Nachfackelkontrolle aus Startphase",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0072"} = {errorcode => "0d",                                     subcode => "2525",                                               class => "BC",          reason => "Nachfackelkontrolle aus Station&auml;rbetrieb",                                                                                                                                                                                                                                                                                                    checklist => "Nachfackelkontrolle aus Station&auml;rbetrieb",                                                                                                                                                                                                                                                                                                                                                                                                                                                       correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0073"} = {errorcode => "0d",                                     subcode => "2526",                                               class => "BC",          reason => "Nachbel&uuml;ftung aus Startphase",                                                                                                                                                                                                                                                                                                                checklist => "Nachbel&uuml;ftung aus Startphase",                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0074"} = {errorcode => "0d",                                     subcode => "2527",                                               class => "BC",          reason => "Nachbel&uuml;ftung aus Station&auml;rbetrieb",                                                                                                                                                                                                                                                                                                     checklist => "Nachbel&uuml;ftung aus Station&auml;rbetrieb",                                                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0075"} = {errorcode => "0d",                                     subcode => "2528",                                               class => "BC",          reason => "Gebl&auml;se aus",                                                                                                                                                                                                                                                                                                                                 checklist => "Gebl&auml;se aus",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0076"} = {errorcode => "0d",                                     subcode => "2529",                                               class => "BC",          reason => "Sicherheitsrelais aus",                                                                                                                                                                                                                                                                                                                            checklist => "Sicherheitsrelais aus",                                                                                                                                                                                                                                                                                                                                                                                                                                                                               correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0077"} = {errorcode => "0E",                                     subcode => "265",                                                class => "BC",          reason => "W&auml;rmeerzeuger in Betriebsbereitschaft und W&auml;rmebedarf vorhanden, es wird jedoch zu viel Energie geliefert",                                                                                                                                                                                                                              checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0078"} = {errorcode => "0E",                                     subcode => "2512",                                               class => "BC",          reason => "W&auml;rmeanforderung blockiert aufgrund einer Leistungsbegrenzung",                                                                                                                                                                                                                                                                               checklist => "W&auml;rmeanforderung blockiert aufgrund einer Leistungsbegrenzung",                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0079"} = {errorcode => "0F",                                     subcode => "-",                                                  class => "",            reason => "Gasarmatur/&Ouml;lMagnetventil wird ge&ouml;ffnet",                                                                                                                                                                                                                                                                                                checklist => "Vorlauftemperatur mit Basiscontroller kontrollieren. R&uuml;cklauftemperaturf&uuml;hler mit Bedieneinheit oder Service Key kontrollieren&bull;Widerstandswert des Kesself&uuml;hlers (Sicherheitstemperaturbegrenzer) messen und mit der Kennlinie vergleichen&bull;Pr&uuml;fen, ob ein Gussglied durch Schmutz verstopft ist",                                                                                                                                                                       correction => "Einstellung der Kesselkreispumpe anpassen&bull;Oberf&auml;chentemperatur des mit dem Sicherheitstemperaturf&uuml;hler best&uuml;ckten Gussglieds mit Temperaturmessger&auml;t pr&uuml;fen&bull;Gussglied reinigen"                                                                                                                                                                               };
	$ErrorCodeList{"0080"} = {errorcode => "0F",                                     subcode => "2513",                                               class => "BC",          reason => "W&auml;rmeanforderung blockiert aufgrund von Temperaturdifferenzen",                                                                                                                                                                                                                                                                               checklist => "W&auml;rmeanforderung blockiert aufgrund; zu hoher Temperaturdifferenzen zwischen Vorlauf und R&uuml;cklauf",                                                                                                                                                                                                                                                                                                                                                                                         correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0081"} = {errorcode => "0H",                                     subcode => "203",                                                class => "BC",          reason => "Ger&auml;t in Betriebsbereitschaft, kein W&auml;rmebedarf vorhanden",                                                                                                                                                                                                                                                                              checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0082"} = {errorcode => "0H",                                     subcode => "2500",                                               class => "BC",          reason => "Keine W&auml;rmeanforderung",                                                                                                                                                                                                                                                                                                                      checklist => "Keine W&auml;rmeanforderung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                         correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0083"} = {errorcode => "0H",                                     subcode => "2530",                                               class => "BC",          reason => "Interner Status",                                                                                                                                                                                                                                                                                                                                  checklist => "Interner Status",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                     correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0084"} = {errorcode => "0L",                                     subcode => "284",                                                class => "BC",          reason => "Gasarmatur/&Ouml;l-Magnetventile werden ge&ouml;ffnet Erste Sicherheitszeit",                                                                                                                                                                                                                                                                      checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0085"} = {errorcode => "0L",                                     subcode => "2520",                                               class => "BC",          reason => "Flamme stabilisieren",                                                                                                                                                                                                                                                                                                                             checklist => "Flamme stabilisieren",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0086"} = {errorcode => "0L",                                     subcode => "2521",                                               class => "BC",          reason => "Stabilisieren W&auml;rmetauscher",                                                                                                                                                                                                                                                                                                                 checklist => "Stabilisieren W&auml;rmetauscher",                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0087"} = {errorcode => "0L",                                     subcode => "2522",                                               class => "BC",          reason => "Warten Aufheizung W&auml;rmetauscher",                                                                                                                                                                                                                                                                                                             checklist => "Warten Aufheizung W&auml;rmetauscher",                                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0088"} = {errorcode => "0L",                                     subcode => "2523",                                               class => "BC",          reason => "Umschaltphase (von Start auf station&auml;r)",                                                                                                                                                                                                                                                                                                     checklist => "Umschaltphase (von Start auf station&auml;r)",                                                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0089"} = {errorcode => "0P",                                     subcode => "205",                                                class => "BC",          reason => "W&auml;rmeerzeuger wartet auf Luftstr&ouml;mung",                                                                                                                                                                                                                                                                                                  checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0090"} = {errorcode => "0U",                                     subcode => "270",                                                class => "BC",          reason => "W&auml;rmeerzeuger wird hochgefahren",                                                                                                                                                                                                                                                                                                             checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0091"} = {errorcode => "0Y",                                     subcode => "204",                                                class => "BC",          reason => "Aktuelle Heizwassertemperatur des W&auml;rmeerzeugers h&ouml;her als der Sollwert",                                                                                                                                                                                                                                                                checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0092"} = {errorcode => "0Y",                                     subcode => "276",                                                class => "10",          reason => "Temperatur am Vorlauftemperaturf&uuml;hler zu hoch",                                                                                                                                                                                                                                                                                               checklist => "Sind alle Absperrh&auml;hne ge&ouml;ffnet?&bull;Betr&auml;gt der Wasserdruck in der Anlage mindestens 1 bar?&bull;Ist mindestens ein Thermostatventil ge&ouml;ffnet?&bull;Umw&auml;lzpumpe pr&uuml;fen&bull;Vorlauftemperaturf&uuml;hler pr&uuml;fen&bull;Pumpenleistung oder Pumpenkennfeld pr&uuml;fen",                                                                                                                                                                                            correction => "Absperrh&auml;hne &ouml;ffnen&bull;Bei zu geringem Wasserdruck,; Wasser nachf&uuml;llen und Anlage entl&uuml;ften&bull;Ein Thermostatventil &ouml;ffnen&bull;Ggf. Umw&auml;lzpumpe austauschen&bull;Ggf. Vorlauftemperaturf&uuml;hler austauschen&bull;Pumpenleistung oder Pumpenkennfeld korrekt einstellen und auf maximale Leistung anpassen"                                                 };
	$ErrorCodeList{"0093"} = {errorcode => "0Y",                                     subcode => "277",                                                class => "10",          reason => "Temperatur am Sicherheitstemperaturf&uuml;hler zu hoch",                                                                                                                                                                                                                                                                                           checklist => "Sind alle Absperrh&auml;hne ge&ouml;ffnet?&bull;Betr&auml;gt der Wasserdruck in der Anlage mindestens 1 bar?&bull;Ist mindestens ein Thermostatventil ge&ouml;ffnet?&bull;Umw&auml;lzpumpe pr&uuml;fen&bull;Sicherheitstemperaturf&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                            correction => "Absperrh&auml;hne &ouml;ffnen&bull;Bei zu geringem Wasserdruck, Wasser nachf&uuml;llen und Anlage entl&uuml;ften&bull;Ein Thermostatventil &ouml;ffnen&bull;Ggf. Umw&auml;lzpumpe austauschen&bull;Ggf. Sicherheitstemperaturf&uuml;hler austauschen"                                                                                                                                            };
	$ErrorCodeList{"0094"} = {errorcode => "0Y",                                     subcode => "285",                                                class => "10",          reason => "Temperatur am R&uuml;cklauftemperaturf&uuml;hler zu hoch",                                                                                                                                                                                                                                                                                         checklist => "Sind alle Absperrh&auml;hne ge&ouml;ffnet?&bull;Betr&auml;gt der Wasserdruck in der Anlage mindestens 1 bar?&bull;Ist mindestens ein Thermostatventil ge&ouml;ffnet?&bull;Umw&auml;lzpumpe pr&uuml;fen&bull;R&uuml;cklauftemperaturf&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                          correction => "Absperrh&auml;hne &ouml;ffnen&bull;Bei zu geringem Wasserdruck Wasser nachf&uuml;llen und Anlage entl&uuml;ften&bull;Ein Thermostatventil &ouml;ffnen&bull;Ggf. Umw&auml;lzpumpe austauschen&bull;Ggf. R&uuml;cklauftemperaturf&uuml;hler austauschen"                                                                                                                                           };
	$ErrorCodeList{"0095"} = {errorcode => "0Y",                                     subcode => "359",                                                class => "10",          reason => "Temperatur am Warmwasser-Temperaturf&uuml;hler zu hoch",                                                                                                                                                                                                                                                                                           checklist => "Korrekte Einbaulage des Temperaturf&uuml;hlers pr&uuml;fen&bull;Temperaturf&uuml;hler und Anschlusskabel auf Unterbrechung oder Kurzschluss pr&uuml;fen&bull;Kesselidentifikationsmodul richtig aufstecken",                                                                                                                                                                                                                                                                                          correction => "Temperaturf&uuml;hler korrekt einbauen&bull;Ggf. austauschen&bull;Ggf. austauschen"                                                                                                                                                                                                                                                                                                              };
	$ErrorCodeList{"0096"} = {errorcode => "0Y",                                     subcode => "2511",                                               class => "BC",          reason => "W&auml;rmeanforderung blockiert, weil GPA nicht kalibriert",                                                                                                                                                                                                                                                                                       checklist => "W&auml;rmeanforderung blockiert, weil Luftklappenstellmotor (GPA) nicht kalibriert",                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0097"} = {errorcode => "0Y",                                     subcode => "2515",                                               class => "BC",          reason => "W&auml;rmeanforderung blockiert, weil Kessel warm genug",                                                                                                                                                                                                                                                                                          checklist => "W&auml;rmeanforderung blockiert, weil Kessel warm genug",                                                                                                                                                                                                                                                                                                                                                                                                                                             correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0098"} = {errorcode => "0Y",                                     subcode => "2531",                                               class => "BC",          reason => "W&auml;rmeanforderung blockiert weil Mischraum zu kalt ist",                                                                                                                                                                                                                                                                                       checklist => "W&auml;rmeanforderung blockiert weil Mischraum zu kalt ist.",                                                                                                                                                                                                                                                                                                                                                                                                                                         correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0099"} = {errorcode => "10",                                     subcode => "481",                                                class => "BC",          reason => "W&auml;rmepumpe l&auml;uft; bei blockiertem Heizkessel",                                                                                                                                                                                                                                                                                           checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0100"} = {errorcode => "11",                                     subcode => "482",                                                class => "BC",          reason => "W&auml;rmepumpe arbeitet im Servicemodus",                                                                                                                                                                                                                                                                                                         checklist => "Die W&auml;rmepumpe arbeitet im Servicemodus (100 % Leistung)",                                                                                                                                                                                                                                                                                                                                                                                                                                       correction => "Nach erfolgten Servicearbeiten auf Normalbetrieb umschalten; der Servicemodus endet automatisch nach 15 min"                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0101"} = {errorcode => "1C",                                     subcode => "210",                                                class => "4",           reason => "Abgasthermostat hat angesprochen",                                                                                                                                                                                                                                                                                                                 checklist => "Abgastemperaturf&uuml;hler unter 105°C abk&uuml;hlen lassen und anschlie&szlig;end auf Durchgang pr&uuml;fen.&bull;W&auml;rmetauscher und Brenner auf Verschmutzung pr&uuml;fen",                                                                                                                                                                                                                                                                                                                    correction => "Ist kein Durchgang vorhanden, Abgastemperaturf&uuml;hler austauschen&bull;W&auml;rmetauscher und Brenner entsprechend Montageund Wartungsanleitung reinigen"                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0102"} = {errorcode => "1C",                                     subcode => "526",                                                class => "4",           reason => "F&uuml;hlerdifferenz Abgastemperaturf&uuml;hler zu gro&szlig;",                                                                                                                                                                                                                                                                                    checklist => "F&uuml;hlerleitung pr&uuml;fen&bull;Steckverbindung pr&uuml;fen&bull;-&bull;F&uuml;hlerwerte nach Tabelle pr&uuml;fen&bull;Spannungswerte am F&uuml;hler laut Tabelle pr&uuml;fen",                                                                                                                                                                                                                                                                                                                   correction => "Bei Besch&auml;digung austauschen&bull;Bei Verschmutzung reinigen oder ggf. austauschen&bull;Bei Besch&auml;digung austauschen&bull;Bei losem Stecker diesen wieder aufstecken&bull;Bei Abweichungen den F&uuml;hler austauschen&bull;Bei Abweichungen Ger&auml;teelektronik SAFe austauschen"                                                                                                   };
	$ErrorCodeList{"0103"} = {errorcode => "1F",                                     subcode => "525",                                                class => "4",           reason => "Abgastemperatur zu hoch",                                                                                                                                                                                                                                                                                                                          checklist => "Im Men&uuml; „Service“ der Bedieneinheit den Wert der Abgastemperatur pr&uuml;fen und mit der wirklichen Abgastemperatur vergleichen&bull;Wenn Abweichungen bestehen, dann muss der Abgastemperaturf&uuml;hler mit einem Ohmmeter gemessen und nach Tabelle gepr&uuml;ft werden&bull;Kessel auf Verschmutzung pr&uuml;fen&bull;Die Position des Abgastemperaturf&uuml;hlers pr&uuml;fen&bull;Pr&uuml;fen, ob die Heizgaslenkplatten vorhanden oder eventuell besch&auml;digt sind",               correction => "-&bull;Bei Abweichungen der Messwerte den F&uuml;hler austauschen&bull;Bei verschmutztem Kessel diesen reinigen&bull;Wenn der Abgastemperaturf&uuml;hler falsch positioniert ist, diesen an der richtigen Stelle montieren&bull;Heizgaslenkplatten ersetzen bzw. vervollst&auml;ndigen"                                                                                                          };
	$ErrorCodeList{"0104"} = {errorcode => "1H",                                     subcode => "530",                                                class => "10",          reason => "Abgastemperatur zu hoch",                                                                                                                                                                                                                                                                                                                          checklist => "F&uuml;hlerleitung pr&uuml;fen&bull;Steckverbindung pr&uuml;fen&bull;-&bull;F&uuml;hlerwerte nach Tabelle pr&uuml;fen&bull;Spannungswerte am F&uuml;hler laut Tabelle pr&uuml;fen",                                                                                                                                                                                                                                                                                                                   correction => "Bei Besch&auml;digung austauschen&bull;Bei Verschmutzung reinigen oder ggf. austauschen&bull;Bei Besch&auml;digung austauschen&bull;Bei losem Stecker diesen wieder aufstecken&bull;Bei Abweichungen den F&uuml;hler austauschen&bull;Bei Abweichungen Ger&auml;teelektronik SAFe austauschen"                                                                                                   };
	$ErrorCodeList{"0105"} = {errorcode => "1H",                                     subcode => "562",                                                class => "10",          reason => "Abgasaustrittssicherung zu hohe Temperatur",                                                                                                                                                                                                                                                                                                       checklist => "Zugbedarf im Abgasrohr pr&uuml;fen&bull;Schornsteindimensionierung pr&uuml;fen&bull;Abgas&uuml;berwachungseinrichtung defekt",                                                                                                                                                                                                                                                                                                                                                                        correction => "Abgasweg frei machen&bull;-&bull;Abgas&uuml;berwachungseinrichtung pr&uuml;fen bzw. austauschen"                                                                                                                                                                                                                                                                                                 };
	$ErrorCodeList{"0106"} = {errorcode => "1H",                                     subcode => "563",                                                class => "4",           reason => "Zu h&auml;ufiger Abgasaustritt",                                                                                                                                                                                                                                                                                                                   checklist => "Zugbedarf im Abgasrohr pr&uuml;fen&bull;Schornsteindimensionierung pr&uuml;fen&bull;Abgas&uuml;berwachungseinrichtung defekt",                                                                                                                                                                                                                                                                                                                                                                        correction => "Abgasweg frei machen&bull;-&bull;Abgas&uuml;berwachungseinrichtung pr&uuml;fen bzw. austauschen"                                                                                                                                                                                                                                                                                                 };
	$ErrorCodeList{"0107"} = {errorcode => "1L",                                     subcode => "211",                                                class => "10",          reason => "Installation Ger&auml;teelektronik UBA fehlerhaft",                                                                                                                                                                                                                                                                                                checklist => "Kontakt zwischen Ger&auml;teelektronik UBA und Montagefu&szlig; pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                         correction => "Ger&auml;teelektronik UBA ordnungsgem&auml;&szlig; aufsetzen und nachziehen, Abdeckung hinter Ger&auml;teelektronik UBA abnehmen und Stecker richtig auf Ger&auml;teelektronik-R&uuml;ckseite aufstecken"                                                                                                                                                                                        };
	$ErrorCodeList{"0108"} = {errorcode => "1L",                                     subcode => "527",                                                class => "4",           reason => "Kurzschluss Abgastemperaturf&uuml;hler",                                                                                                                                                                                                                                                                                                           checklist => "F&uuml;hlerleitung pr&uuml;fen&bull;Steckverbindung pr&uuml;fen&bull;-&bull;F&uuml;hlerwerte nach Tabelle pr&uuml;fen&bull;Spannungswerte am F&uuml;hler laut Tabelle pr&uuml;fen",                                                                                                                                                                                                                                                                                                                   correction => "Bei Besch&auml;digung austauschen&bull;Bei Verschmutzung reinigen oder ggf. austauschen&bull;Bei Besch&auml;digung austauschen&bull;Bei losem Stecker diesen wieder aufstecken&bull;Bei Abweichungen den F&uuml;hler austauschen&bull;Bei Abweichungen Ger&auml;teelektronik UBA austauschen"                                                                                                    };
	$ErrorCodeList{"0109"} = {errorcode => "1L",                                     subcode => "529",                                                class => "4",           reason => "Kurzschluss Abgastemperaturf&uuml;hler",                                                                                                                                                                                                                                                                                                           checklist => "F&uuml;hlerleitung pr&uuml;fen&bull;Steckverbindung pr&uuml;fen&bull;-&bull;F&uuml;hlerwerte nach Tabelle pr&uuml;fen&bull;Spannungswerte am F&uuml;hler laut Tabelle pr&uuml;fen",                                                                                                                                                                                                                                                                                                                   correction => "Bei Besch&auml;digung austauschen&bull;Bei Verschmutzung reinigen oder ggf. austauschen&bull;Bei Besch&auml;digung austauschen&bull;Bei losem Stecker diesen wieder aufstecken&bull;Bei Abweichungen den F&uuml;hler austauschen&bull;Bei Abweichungen Ger&auml;teelektronik SAFe austauschen"                                                                                                   };
	$ErrorCodeList{"0110"} = {errorcode => "1P",                                     subcode => "528",                                                class => "4",           reason => "Unterbrechung Abgastemperaturf&uuml;hler",                                                                                                                                                                                                                                                                                                         checklist => "F&uuml;hlerleitung pr&uuml;fen&bull;Steckverbindung pr&uuml;fen&bull;-&bull;F&uuml;hlerwerte nach Tabelle pr&uuml;fen&bull;Spannungswerte am F&uuml;hler laut Tabelle pr&uuml;fen",                                                                                                                                                                                                                                                                                                                   correction => "Bei Besch&auml;digung austauschen&bull;Bei Verschmutzung reinigen oder ggf. austauschen&bull;Bei Besch&auml;digung austauschen&bull;Bei losem Stecker diesen wieder aufstecken&bull;Bei Abweichungen den F&uuml;hler austauschen&bull;Bei Abweichungen Ger&auml;teelektronik SAFe austauschen"                                                                                                   };
	$ErrorCodeList{"0111"} = {errorcode => "2A",                                     subcode => "531",                                                class => "4",           reason => "Wassermangel im W&auml;rmeerzeuger",                                                                                                                                                                                                                                                                                                               checklist => "Pr&uuml;fen, ob die Absperrh&auml;hne ge&ouml;ffnet sind&bull;-&bull;Ist Luft im System?&bull;Betr&auml;gt der Wasserdruck in der Anlage mindestens 1 bar?",                                                                                                                                                                                                                                                                                                                                          correction => "Anlage entl&uuml;ften&bull;Absperrh&auml;hne &ouml;ffnen&bull;Anlage entl&uuml;ften&bull;Bei zu geringem Wasserdruck, Wasser nachf&uuml;llen und Anlage entl&uuml;ften"                                                                                                                                                                                                                          };
	$ErrorCodeList{"0112"} = {errorcode => "2E",                                     subcode => "207",                                                class => "4",           reason => "Betriebsdruck zu niedrig",                                                                                                                                                                                                                                                                                                                         checklist => "Ist Luft im System?&bull;Betr&auml;gt der Wasserdruck in der Anlage mindestens 1 bar?&bull;Drucksensor &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                         correction => "Anlage entl&uuml;ften&bull;Bei zu geringem Wasserdruck, Wasser nachf&uuml;llen und Anlage entl&uuml;ften&bull;Drucksensor austauschen"                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"0113"} = {errorcode => "2E",                                     subcode => "357",                                                class => "BC",          reason => "Entl&uuml;ftungsprogramm",                                                                                                                                                                                                                                                                                                                         checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0114"} = {errorcode => "2E",                                     subcode => "358",                                                class => "BC",          reason => "Entl&uuml;ftungsprogramm",                                                                                                                                                                                                                                                                                                                         checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0115"} = {errorcode => "2F",                                     subcode => "260",                                                class => "12",          reason => "Kein Temperaturanstieg nach Brennerstart",                                                                                                                                                                                                                                                                                                         checklist => "Sind alle Absperrh&auml;hne ge&ouml;ffnet?&bull;Betr&auml;gt der Wasserdruck in der Anlage mindestens 1 bar?&bull;Ist mindestens ein Thermostatventil ge&ouml;ffnet?&bull;Umw&auml;lzpumpe pr&uuml;fen&bull;Sicherheitsund Vorlauftemperaturf&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                 correction => "Absperrh&auml;hne &ouml;ffnen&bull;Bei zu geringem Wasserdruck Wasser nachf&uuml;llen und Anlage entl&uuml;ften&bull;Ein Thermostatventil &ouml;ffnen&bull;Ggf. Umw&auml;lzpumpe austauschen&bull;Ggf. Sicherheitsoder Vorlauftemperaturf&uuml;hler austauschen"                                                                                                                                 };
	$ErrorCodeList{"0116"} = {errorcode => "2F",                                     subcode => "271",                                                class => "4",           reason => "Temperaturdifferenz Kesselvorlauf-/Sicherheitstemperaturf&uuml;hler; zu gro&szlig;",                                                                                                                                                                                                                                                               checklist => "Sind alle Absperrh&auml;hne ge&ouml;ffnet?&bull;Ist Luft im System?&bull;Ist mindestens ein Thermostatventil ge&ouml;ffnet?&bull;Umw&auml;lzpumpe pr&uuml;fen&bull;Sicherheitsund Vorlauftemperaturf&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                                                          correction => "Absperrh&auml;hne &ouml;ffnen&bull;Anlage entl&uuml;ften&bull;Ein Thermostatventil &ouml;ffnen&bull;Ggf. Umw&auml;lzpumpe austauschen&bull;Ggf. Sicherheitsoder Vorlauftemperaturf&uuml;hler austauschen"                                                                                                                                                                                        };
	$ErrorCodeList{"0117"} = {errorcode => "2F",                                     subcode => "338",                                                class => "4",           reason => "Zu viele erfolglose Brennerstartversuche",                                                                                                                                                                                                                                                                                                         checklist => "Sind alle Absperrh&auml;hne ge&ouml;ffnet?&bull;Betr&auml;gt der Wasserdruck in der Anlage mindestens 1 bar?&bull;Ist mindestens ein Thermostatventil ge&ouml;ffnet?&bull;Umw&auml;lzpumpe pr&uuml;fen&bull;Sicherheitsund Vorlauftemperaturf&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                 correction => "Absperrh&auml;hne &ouml;ffnen&bull;Bei zu geringem Wasserdruck Wasser nachf&uuml;llen und Anlage entl&uuml;ften&bull;Ein Thermostatventil &ouml;ffnen&bull;Ggf. Umw&auml;lzpumpe austauschen&bull;Ggf. Sicherheitsoder Vorlauftemperaturf&uuml;hler austauschen"                                                                                                                                 };
	$ErrorCodeList{"0118"} = {errorcode => "2H",                                     subcode => "357",                                                class => "BC",          reason => "Blockierschutz aktiv",                                                                                                                                                                                                                                                                                                                             checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0119"} = {errorcode => "2H",                                     subcode => "358",                                                class => "BC",          reason => "Blockierschutz aktiv",                                                                                                                                                                                                                                                                                                                             checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0120"} = {errorcode => "2L",                                     subcode => "266",                                                class => "4",           reason => "Pumpendruckerh&ouml;hung zu niedrig",                                                                                                                                                                                                                                                                                                              checklist => "Ist Luft im System?&bull;Absperrh&auml;hne nicht vollst&auml;ndig ge&ouml;ffnet&bull;Durchfluss zu gest&ouml;rt, nicht ausreichend&bull;Umw&auml;lzpumpe auf mechanisches Klemmen pr&uuml;fen&bull;Ansteuerung der Umw&auml;lzpumpe pr&uuml;fen&bull;Laufrad der Umw&auml;lzpumpe auf Verschmutzungen pr&uuml;fen&bull;Druckf&uuml;hler auf Verschmutzung pr&uuml;fen&bull;Pr&uuml;fen, ob bei Einsatz einer hydraulischen Weiche Vorund R&uuml;cklauf an den Heizkreisgruppen richtig angeschlossen sind&bull;Pr&uuml;fen, ob das Ausdehnungsgef&auml;&szlig; am Heizkesselvorlauf angeschlossen ist",correction => "Anlage entl&uuml;ften&bull;Absperrh&auml;hne ganz &ouml;ffnen&bull;Ggf. Filter reinigen bzw. ersetzen&bull;Versuchen, die Umw&auml;lzpumpe zu l&ouml;sen&bull;Erfolgt keine Spannungsversorgung der Umw&auml;lzpumpe, Versorgungskabel durchmessen und bei Besch&auml;digung austauschen&bull;Laufrad der Umw&auml;lzpumpe reinigen&bull;Druckf&uuml;hler reinigen ggf. ersetzen&bull;Ggf. Anlagenhydraulik ab&auml;ndern&bull;Hydraulik ab&auml;ndern, Ausdehnungsgef&auml;&szlig; muss in den Anlagenr&uuml;cklauf"};
	$ErrorCodeList{"0121"} = {errorcode => "2L",                                     subcode => "329",                                                class => "10",          reason => "Pumpendruckerh&ouml;hung zu niedrig",                                                                                                                                                                                                                                                                                                              checklist => "Umw&auml;lzpumpe auf mechanisches Klemmen pr&uuml;fen&bull;Ansteuerung der Umw&auml;lzpumpe pr&uuml;fen&bull;Laufrad der Umw&auml;lzpumpe auf Verschmutzung pr&uuml;fen&bull;Pr&uuml;fen, ob das Ausdehnungsgef&auml;&szlig; am Heizkesselvorlauf angeschlossen ist&bull;Pr&uuml;fen, ob bei Einsatz einer hydraulischen Weiche Vorund R&uuml;cklauf an den Heizkreisgruppen richtig angeschlossen sind&bull;Druckf&uuml;hler auf Verschmutzung pr&uuml;fen",                                         correction => "Versuchen, die Umw&auml;lzpumpe zu l&ouml;sen oder die Umw&auml;lzpumpe auszutauschen&bull;Erfolgt keine Spannungsversorgung der Umw&auml;lzpumpe, Versorgungskabel durchmessen und bei Besch&auml;digung austauschen&bull;Laufrad der Umw&auml;lzpumpe reinigen&bull;Hydraulik &auml;ndern, Ausdehnungsgef&auml;&szlig; muss in den Anlagenr&uuml;cklauf&bull;Ggf. Anlagenhydraulik &auml;ndern&bull;Druckf&uuml;hler reinigen, ggf. ersetzen"};
	$ErrorCodeList{"0122"} = {errorcode => "2P",                                     subcode => "212",                                                class => "12",          reason => "Temperaturanstieg Sicherheitsoder Kesselvorlauftemperaturf&uuml;hler zu schnell",                                                                                                                                                                                                                                                                  checklist => "Sind alle Absperrh&auml;hne ge&ouml;ffnet?&bull;Betr&auml;gt der Wasserdruck in der Anlage mindestens 1 bar?&bull;Ist mindestens ein Thermostatventil ge&ouml;ffnet?&bull;Umw&auml;lzpumpe pr&uuml;fen&bull;Sicherheitsund Vorlauftemperaturf&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                 correction => "Absperrh&auml;hne &ouml;ffnen&bull;Bei zu geringem Wasserdruck Wasser nachf&uuml;llen und Anlage entl&uuml;ften&bull;Ein Thermostatventil &ouml;ffnen&bull;Ggf. Umw&auml;lzpumpe austauschen&bull;Ggf. Sicherheitsoder Vorlauftemperaturf&uuml;hler austauschen"                                                                                                                                 };
	$ErrorCodeList{"0123"} = {errorcode => "2P",                                     subcode => "341",                                                class => "10",          reason => "Temperaturanstieg W&auml;rmeerzeugertemperatur zu schnell",                                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Kessel durchstr&ouml;mt wird&bull;Stecker pr&uuml;fen&bull;Systemwasserdruck sowie Pumpe pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                correction => "W&auml;rmeabnahme sicherstellen&bull;Ggf. richtig aufstecken&bull;Pumpenleistung oder Pumpenkennfeld korrekt einstellen und an maximale Leistung anpassen; W&auml;rmeabnahme sicherstellen"                                                                                                                                                                                                      };
	$ErrorCodeList{"0124"} = {errorcode => "2P",                                     subcode => "342",                                                class => "12",          reason => "Temperaturanstieg Warmwasserbetrieb zu schnell",                                                                                                                                                                                                                                                                                                   checklist => "Betr&auml;gt der Wasserdruck in der Anlage mindestens 1 bar?&bull;Pumpe pr&uuml;fen&bull;Vorlauftemperaturf&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                   correction => "Bei zu geringem Wasserdruck Wasser nachf&uuml;llen und Anlage entl&uuml;ften&bull;Ggf. Pumpe austauschen&bull;Ggf. Vorlauftemperaturf&uuml;hler austauschen"                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0125"} = {errorcode => "2P",                                     subcode => "564",                                                class => "10",          reason => "Temperaturanstieg W&auml;rmeerzeugertemperatur zu schnell",                                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Kessel durchstr&ouml;mt wird&bull;Pumpe pr&uuml;fen&bull;Ablagerungen im Kesselw&auml;rmetauscher (Kalk, Schmutz aus der Heizungsanlage)",                                                                                                                                                                                                                                                                                                                                            correction => "Bei zu geringem Wasserdruck Wasser nachf&uuml;llen und Anlage entl&uuml;ften&bull;W&auml;rmeabnahme sicherstellen&bull;W&auml;rmetauscher mit einem f&uuml;r Aluminium geeigneten und freigegebenen Mittel reinigen"                                                                                                                                                                             };
	$ErrorCodeList{"0126"} = {errorcode => "2U",                                     subcode => "213",                                                class => "12",          reason => "Differenz zwischen Vorlaufund R&uuml;cklauftemperatur zu gro&szlig;; (> 50 K)",                                                                                                                                                                                                                                                                    checklist => "Sind alle Absperrh&auml;hne ge&ouml;ffnet?; Betr&auml;gt der Wasserdruck in der Anlage mindestens 1 bar?; Ist mindestens ein Thermostatventil ge&ouml;ffnet? Umw&auml;lzpumpe pr&uuml;fen; Vorlaufund R&uuml;cklauftemperaturf&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                                correction => "Absperrh&auml;hne &ouml;ffnen; Bei zu geringem Wasserdruck Wasser nachf&uuml;llen und Anlage entl&uuml;ften; Ein Thermostatventil &ouml;ffnen; Ggf. Umw&auml;lzpumpe austauschen; Ggf. Vorlaufod. R&uuml;cklauftemperaturf&uuml;hler austauschen"                                                                                                                                                };
	$ErrorCodeList{"0127"} = {errorcode => "2U",                                     subcode => "533",                                                class => "4",           reason => "Regelung des W&auml;rmeerzeugers hat wasserseitig falsche Durchstr&ouml;mung erkannt",                                                                                                                                                                                                                                                             checklist => "Pr&uuml;fen, ob Kesselvorund -r&uuml;cklauf vertauscht sind&bull;Umw&auml;lzpumpen auf korrekte Flie&szlig;richtung pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                     correction => "Kesselvorund r&uuml;cklauf korrekt anschlie&szlig;en&bull;Korrekte Flie&szlig;richtung der Pumpen sicherstellen"                                                                                                                                                                                                                                                                                 };
	$ErrorCodeList{"0128"} = {errorcode => "2U",                                     subcode => "565",                                                class => "10",          reason => "Differenz zwischen Vorlaufund R&uuml;cklauftemperatur zu gro&szlig;",                                                                                                                                                                                                                                                                              checklist => "Kann bei normaler Anlagenkonfiguration nicht auftreten",                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Anlagenhydraulik pr&uuml;fen"                                                                                                                                                                                                                                                                                                                                                                    };
	$ErrorCodeList{"0129"} = {errorcode => "2U",                                     subcode => "575",                                                class => "4",           reason => "Kesselvorlauftemperatur hat maximal zul&auml;ssigen Wert &uuml;berschritten",                                                                                                                                                                                                                                                                      checklist => "Wasserseitige Durchstr&ouml;mung pr&uuml;fen&bull;-",                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "Ausreichende Durchstr&ouml;mung sicherstellen&bull;Kesself&uuml;hler/Sicherheitstemperaturbegrenzer-F&uuml;hler austauschen&bull;Z&uuml;nd-/&Uuml;berwachungselektrode austauschen"                                                                                                                                                                                                              };
	$ErrorCodeList{"0130"} = {errorcode => "2U",                                     subcode => "2050",                                               class => "10",          reason => "Falschdurchstr&ouml;mung des Kessels",                                                                                                                                                                                                                                                                                                             checklist => "Fehlerhafter hydraulischer Anschluss des Kessels&bull;Fehlerhafter Einbau der Heizkreispumpe(n) (Flie&szlig;richtung)&bull;Fehlerhafte Montage des HeizkreisSchnellmontage-Sets&bull;R&uuml;cklauftemperaturf&uuml;hler besch&auml;digt",                                                                                                                                                                                                                                                             correction => "Hydraulischer Anschluss (Vorlauf/R&uuml;cklauf) am Kessel pr&uuml;fen, ggf. korrigieren&bull;Einbaulage der Heizkreispumpe(n) pr&uuml;fen, ggf. korrigieren&bull;Montage des Heizkreis-Schnellmontage-Sets pr&uuml;fen, ggf. korrigieren&bull;R&uuml;cklauftemperaturf&uuml;hler pr&uuml;fen, ggf. austauschen"                                                                                  };
	$ErrorCodeList{"0131"} = {errorcode => "2Y",                                     subcode => "281",                                                class => "12",          reason => "Pumpendruckerh&ouml;hung zu niedrig",                                                                                                                                                                                                                                                                                                              checklist => "Luft in der Umw&auml;lzpumpe&bull;Tachokabel auf Kabelbruch pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                             correction => "Pumpe entfl&uuml;ften&bull;Ggf. Tachokabel der Umw&auml;lzpumpe austauschen"                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0132"} = {errorcode => "2Y",                                     subcode => "282",                                                class => "10",          reason => "Keine Drehzahlr&uuml;ckmeldung Kesselpumpe",                                                                                                                                                                                                                                                                                                       checklist => "Pr&uuml;fen, ob die Steckverbindung des  Tachokabels und Versorgungskabel der Umw&auml;lzpumpe richtig montiert sind&bull;Tachokabel auf Kabelbruch pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                     correction => "Steckverbindung des Tachokabels der Umw&auml;lzpumpe wiederherstellen&bull;Ggf. Tachokabel der Umw&auml;lzpumpe austauschen"                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0133"} = {errorcode => "2Y",                                     subcode => "307",                                                class => "12",          reason => "Kesselpumpe blockiert",                                                                                                                                                                                                                                                                                                                            checklist => "Kesselinterne Pumpe defekt",                                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Kesselinterne Pumpe austauschen"                                                                                                                                                                                                                                                                                                                                                                 };
	$ErrorCodeList{"0134"} = {errorcode => "2Y",                                     subcode => "308",                                                class => "12",          reason => "Kesselpumpe dreht ohne Widerstand",                                                                                                                                                                                                                                                                                                                checklist => "Kesselinterne Pumpe defekt",                                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Kesselinterne Pumpe austauschen"                                                                                                                                                                                                                                                                                                                                                                 };
	$ErrorCodeList{"0135"} = {errorcode => "3A",                                     subcode => "264",                                                class => "10",          reason => "Lufttransport w&auml;hrend Betriebsphase ausgefallen",                                                                                                                                                                                                                                                                                             checklist => "Beide Steckverbindungen am Gebl&auml;se pr&uuml;fen&bull;Ansteuerung 230 V AC des Gebl&auml;ses pr&uuml;fen, Versorgungskabel des Gebl&auml;ses pr&uuml;fen&bull;Tachokabel des Gebl&auml;ses pr&uuml;fen&bull;-",                                                                                                                                                                                                                                                                                    correction => "Steckverbindungen am Gebl&auml;se ordnungsgem&auml;&szlig; wiederherstellen&bull;Ggf. Versorgungskabel des Gebl&auml;ses austauschen&bull;Ggf. Tachokabel des Gebl&auml;ses austauschen&bull;Ggf. Gebl&auml;se austauschen"                                                                                                                                                                      };
	$ErrorCodeList{"0136"} = {errorcode => "3C",                                     subcode => "217",                                                class => "4",           reason => "Kein Lufttransport nach mehreren Minuten",                                                                                                                                                                                                                                                                                                         checklist => "Feuerungsautomat, Ger&auml;teelektronik oder Gebl&auml;se defekt&bull;Gebl&auml;sekabel mit Stecker und Gebl&auml;se pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                    correction => "Feuerungsautomat oder Gebl&auml;se austauschen&bull;Ggf. austauschen"                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"0137"} = {errorcode => "3C",                                     subcode => "537",                                                class => "4",           reason => "Keine Drehzahlr&uuml;ckmeldung vom Gebl&auml;se",                                                                                                                                                                                                                                                                                                  checklist => "Verbindungsleitung zwischen Gebl&auml;se und Feuerungsautomat SAFe pr&uuml;fen&bull;Steckverbindung an Feuerungsautomat SAFe und Gebl&auml;se pr&uuml;fen&bull;-",                                                                                                                                                                                                                                                                                                                                    correction => "Wenn besch&auml;digt, austauschen&bull;Wenn besch&auml;digt, austauschen oder Stecker wieder aufstecken&bull;Wenn Spannung anliegt und das Gebl&auml;se nicht l&auml;uft, ist das Gebl&auml;se defekt und muss ausgetauscht werden&bull;Wenn keine Spannung anliegt, ist die Ger&auml;teelektronik SAFe defekt und muss ausgetauscht werden"                                                     };
	$ErrorCodeList{"0138"} = {errorcode => "3C",                                     subcode => "538",                                                class => "4",           reason => "Gebl&auml;se zu langsam",                                                                                                                                                                                                                                                                                                                          checklist => "Gebl&auml;se auf Verschmutzung oder Defekt pr&uuml;fen&bull;-",                                                                                                                                                                                                                                                                                                                                                                                                                                       correction => "Gebl&auml;se reinigen&bull;Ggf. Gebl&auml;se austauschen"                                                                                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"0139"} = {errorcode => "3C",                                     subcode => "539",                                                class => "4",           reason => "Gebl&auml;sedrehzahl au&szlig;erhalb des zul&auml;ssigen Bereichs",                                                                                                                                                                                                                                                                                checklist => "&Uuml;berp&uuml;fen des PWM-Verbindungskabels&bull;Brennergebl&auml;se auf Funktion pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                     correction => "Ggf. PWM-Verbindungskabel austauschen&bull;Ggf. Brennergebl&auml;se austauschen"                                                                                                                                                                                                                                                                                                                 };
	$ErrorCodeList{"0140"} = {errorcode => "3C",                                     subcode => "540",                                                class => "4",           reason => "Gebl&auml;se zu schnell",                                                                                                                                                                                                                                                                                                                          checklist => "Pr&uuml;fen, ob PWM-Signal &uuml;bertragen wird",                                                                                                                                                                                                                                                                                                                                                                                                                                                     correction => "PWM Verbindungskabel pr&uuml;fen, ggf. austauschen"                                                                                                                                                                                                                                                                                                                                              };
	$ErrorCodeList{"0141"} = {errorcode => "3C",                                     subcode => "2036",                                               class => "4",           reason => "Gebl&auml;sedrehzahl entspricht nicht Sollwert",                                                                                                                                                                                                                                                                                                   checklist => "Verbindungsleitung (PWM-Signal) zwischen Feuerungsautomat und Brennergebl&auml;se besch&auml;digt oder nicht aufgesteckt&bull;Spannungsversorgung des Brennergebl&auml;ses fehlerhaft&bull;Brennergebl&auml;se besch&auml;digt&bull;Feuerungsautomat besch&auml;digt",                                                                                                                                                                                                                                correction => "Verbindungsleitung (PWM-Signal) zwischen Feuerungsautomat und Brennergebl&auml;se pr&uuml;fen, aufstecken, ggf. austauschen&bull;Spannungsversorgung des Brennergebl&auml;ses pr&uuml;fen, ggf. austauschen&bull;L&uuml;fterrad des Brennergebl&auml;ses pr&uuml;fen, reinigen oder Brennergebl&auml;se ggf. austauschen&bull;Feuerungsautomat pr&uuml;fen, ggf. austauschen"                    };
	$ErrorCodeList{"0142"} = {errorcode => "3C",                                     subcode => "2037",                                               class => "4",           reason => "Startdrehzahl am Gebl&auml;se nicht erreicht",                                                                                                                                                                                                                                                                                                     checklist => "Verbindungsleitung (PWM-Signal) zwischen Feuerungsautomat und Brennergebl&auml;se besch&auml;digt&bull;Brennergebl&auml;se verschmutzt&bull;Brennergebl&auml;se besch&auml;digt&bull;Feuerungsautomat besch&auml;digt",                                                                                                                                                                                                                                                                               correction => "Verbindungsleitung (PWM-Signal) zwischen Feuerungsautomat und Brennergebl&auml;se pr&uuml;fen, ggf. austauschen&bull;Brennergebl&auml;se auf Verschmutzung pr&uuml;fen, ggf. reinigen&bull;Brennergebl&auml;se pr&uuml;fen, ggf. austauschen&bull;Feuerungsautomat pr&uuml;fen, ggf. austauschen"                                                                                                };
	$ErrorCodeList{"0143"} = {errorcode => "3C",                                     subcode => "2046",                                               class => "4",           reason => "Mindestdrehzahl Gebl&auml;se unterschritten",                                                                                                                                                                                                                                                                                                      checklist => "Verbindungsleitung (PWM-Signal) zwischen Feuerungsautomat und Brennergebl&auml;se besch&auml;digt oder nicht aufgesteckt&bull;Spannungsversorgung des Brennergebl&auml;ses fehlerhaft&bull;Brennergebl&auml;se besch&auml;digt&bull;Feuerungsautomat besch&auml;digt",                                                                                                                                                                                                                                correction => "Verbindungsleitung (PWM-Signal) zwischen Feuerungsautomat und Brennergebl&auml;se pr&uuml;fen, aufstecken, ggf. austauschen&bull;Spannungsversorgung des Brennergebl&auml;ses pr&uuml;fen, ggf. herstellen&bull;Brennergebl&auml;se pr&uuml;fen, ggf. austauschen&bull;Feuerungsautomat pr&uuml;fen, ggf. austauschen"                                                                           };
	$ErrorCodeList{"0144"} = {errorcode => "3C",                                     subcode => "2114",                                               class => "4",           reason => "Schwerg&auml;ngiges Gebl&auml;se Das Ansteuer-Signal (PWM) des Gebl&auml;ses passt nicht zur Drehzahl",                                                                                                                                                                                                                                            checklist => "Schwerg&auml;ngiges Gebl&auml;se&bull;-",                                                                                                                                                                                                                                                                                                                                                                                                                                                             correction => "Gebl&auml;se pr&uuml;fen, ggf. reinigen&bull;Gebl&auml;se austauschen"                                                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"0145"} = {errorcode => "3F",                                     subcode => "273",                                                class => "10",          reason => "Betriebsunterbrechung: Brenner und Gebl&auml;se",                                                                                                                                                                                                                                                                                                  checklist => "W&auml;rmeanforderung vollst&auml;ndig wegschalten und nach einer Minute pr&uuml;fen, ob das Gebl&auml;se in Betrieb bleibt",                                                                                                                                                                                                                                                                                                                                                                         correction => "Ggf. Ger&auml;teelektronik UBA austauschen"                                                                                                                                                                                                                                                                                                                                                      };
	$ErrorCodeList{"0146"} = {errorcode => "3H",                                     subcode => "535",                                                class => "10",          reason => "Lufttemperatur zu hoch",                                                                                                                                                                                                                                                                                                                           checklist => "Pr&uuml;fen, ob der Luftf&uuml;hler richtig positioniert ist&bull;Widerstandswerte des F&uuml;hlers nach Tabelle pr&uuml;fen&bull;F&uuml;hlerleitung auf Besch&auml;digung pr&uuml;fen&bull;Kessel auf Verschmutzung pr&uuml;fen",                                                                                                                                                                                                                                                                    correction => "F&uuml;hler richtig positionieren&bull;Bei Abweichungen F&uuml;hler austauschen&bull;Bei Besch&auml;digungen F&uuml;hler austauschen&bull;Wenn erforderlich, Kessel reinigen"                                                                                                                                                                                                                    };
	$ErrorCodeList{"0147"} = {errorcode => "3L",                                     subcode => "214",                                                class => "4",           reason => "Gebl&auml;se wird w&auml;hrend Sicherheitszeit abgeschaltet",                                                                                                                                                                                                                                                                                      checklist => "Beide Steckverbindungen am Gebl&auml;se pr&uuml;fen&bull;Ansteuerung 230 V AC des Gebl&auml;ses pr&uuml;fen, Versorgungskabel des Gebl&auml;ses pr&uuml;fen&bull;Tachokabel des Gebl&auml;ses auf Kabelbruch pr&uuml;fen&bull;Gebl&auml;sekabel mit Stecker und Gebl&auml;se pr&uuml;fen",                                                                                                                                                                                                            correction => "Steckverbindungen am Gebl&auml;se ordnungsgem&auml;&szlig; wiederherstellen&bull;Ggf. Versorgungskabel des Gebl&auml;ses austauschen&bull;Ggf. Tachokabel des Gebl&auml;ses austauschen&bull;Ggf. Gebl&auml;se austauschen"                                                                                                                                                                      };
	$ErrorCodeList{"0148"} = {errorcode => "3P",                                     subcode => "216",                                                class => "4",           reason => "Gebl&auml;se zu langsam",                                                                                                                                                                                                                                                                                                                          checklist => "Pr&uuml;fen, ob das Gebl&auml;se verschmutzt oder feucht ist&bull;Pr&uuml;fen, ob die Netzspannung des Ger&auml;ts zwischen 195 V AC und 253 V AC liegt",                                                                                                                                                                                                                                                                                                                                             correction => "Gebl&auml;se reinigen oder austauschen&bull;Elektroanlage pr&uuml;fen"                                                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"0149"} = {errorcode => "3P",                                     subcode => "2035",                                               class => "4",           reason => "Luftklappenstellung entspricht nicht dem Sollwert",                                                                                                                                                                                                                                                                                                checklist => "Luftklappendichtung besch&auml;digt&bull;Luftkanal verschmutzt&bull;Luftklappe besch&auml;digt&bull;Stellmotor besch&auml;digt",                                                                                                                                                                                                                                                                                                                                                                      correction => "Luftklappendichtung pr&uuml;fen, ggf. austauschen&bull;Luftkanal auf Verschmutzung pr&uuml;fen, ggf. reinigen&bull;Luftklappe pr&uuml;fen, ggf. austauschen&bull;Stellmotor pr&uuml;fen, ggf. austauschen"                                                                                                                                                                                       };
	$ErrorCodeList{"0150"} = {errorcode => "3P",                                     subcode => "2042",                                               class => "4",           reason => "Heizpatronentemperatur entspricht nicht Vorgabe Heizpatronentemperatur zu hoch",                                                                                                                                                                                                                                                                   checklist => "Heizpatrone wird nicht korrekt durchstr&ouml;mt&bull;Luftklappe undicht&bull;Luftklappe ist in fehlerhafter Position&bull;Startluftkanal ist verschmutzt&bull;Stellmotor ist fehlerhaft montiert&bull;Stellmotor ist besch&auml;digt&bull;Gebl&auml;se ist verschmutzt&bull;Gebl&auml;se ist besch&auml;digt",                                                                                                                                                                                        correction => "Luftklappendichtung pr&uuml;fen, ggf. austauschen&bull;-&bull;Startluftkanal auf Verschmutzung pr&uuml;fen, ggf. reinigen&bull;Stellmotor pr&uuml;fen, ggf. austauschen&bull;-&bull;Gebl&auml;se pr&uuml;fen, ggf. L&uuml;fterrad reinigen oder Gebl&auml;se austauschen&bull;-"                                                                                                                 };
	$ErrorCodeList{"0151"} = {errorcode => "3P",                                     subcode => "2083",                                               class => "4",           reason => "Positionskalibrierung Luftklappe fehlgeschlagen",                                                                                                                                                                                                                                                                                                  checklist => "Luftklappendichtung fehlerhaft&bull;Luftklappe besch&auml;digt&bull;Startluftkanal verschmutzt&bull;Stellmotor fehlerhaft montiert&bull;Stellmotor besch&auml;digt",                                                                                                                                                                                                                                                                                                                                  correction => "Luftklappendichtung pr&uuml;fen, ggf. austauschen&bull;Luftklappe auf Besch&auml;digung pr&uuml;fen, ggf. austauschen&bull;Startluftkanal auf Verschmutzung pr&uuml;fen, ggf. reinigen&bull;Stellmotor und Luftklappe auf fehlerhafte Montage pr&uuml;fen, ggf. korrigieren&bull;Stellmotor pr&uuml;fen, ggf. austauschen"                                                                       };
	$ErrorCodeList{"0152"} = {errorcode => "3P",                                     subcode => "2091",                                               class => "4",           reason => "Stellklappe schlie&szlig;t schwerg&auml;ngig; Der Strom des Luftklappenstellmotors (GPA) ist im oberen Anschlag zu hoch",                                                                                                                                                                                                                          checklist => "Luftklappendichtung fehlerhaft&bull;Luftklappe besch&auml;digt&bull;Startluftkanal verschmutzt&bull;Stellmotor fehlerhaft montiert&bull;Stellmotor besch&auml;digt",                                                                                                                                                                                                                                                                                                                                  correction => "Luftklappendichtung pr&uuml;fen, ggf. austauschen&bull;Luftklappe auf Besch&auml;digung pr&uuml;fen, ggf. austauschen&bull;Startluftkanal auf Verschmutzung pr&uuml;fen, ggf. reinigen&bull;Stellmotor und Luftklappe auf fehlerhafte Montage pr&uuml;fen, ggf. korrigieren&bull;Stellmotor pr&uuml;fen, ggf. austauschen"                                                                       };
	$ErrorCodeList{"0153"} = {errorcode => "3P",                                     subcode => "2112",                                               class => "4",           reason => "Heizpatrone k&uuml;hlt nach Abschaltung nicht ab",                                                                                                                                                                                                                                                                                                 checklist => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Entriegeln"                                                                                                                                                                                                                                                                                                                                                                                      };
	$ErrorCodeList{"0154"} = {errorcode => "3U",                                     subcode => "536",                                                class => "4",           reason => "Lufttemperaturf&uuml;hler/Abgastemperaturf&uuml;hler falsch positioniert",                                                                                                                                                                                                                                                                         checklist => "Positionierung des Abgastemperaturf&uuml;hlers pr&uuml;fen&bull;Positionierung des Lufttemperaturf&uuml;hlers pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                           correction => "Bei falscher Positionierung die F&uuml;hlerposition korrigieren&bull;Bei falscher Positionierung die F&uuml;hlerposition korrigieren"                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"0155"} = {errorcode => "3Y",                                     subcode => "215",                                                class => "4",           reason => "Gebl&auml;se zu schnell",                                                                                                                                                                                                                                                                                                                          checklist => "Steckverbindung des Tachokabels am Gebl&auml;se pr&uuml;fen&bull;Tachokabel des Gebl&auml;ses auf Kabelbruch pr&uuml;fen&bull;Pr&uuml;fen, ob die Netzspannung des Ger&auml;ts zwischen 195 V AC und 253 V AC liegt&bull;Pr&uuml;fen, ob eine Verstopfung im Brenner, W&auml;rmetauscher oder Abgassystem vorliegt&bull;Pr&uuml;fen, ob das Gebl&auml;serad lose auf der Motorwelle sitzt",                                                                                                           correction => "Steckverbindung ordnungsgem&auml;&szlig; wiederherstellen&bull;Ggf. Tachokabel des Gebl&auml;ses austauschen&bull;Elektroanlage pr&uuml;fen&bull;Verstopfung/Verunreinigung beheben&bull;Ggf. Gebl&auml;se austauschen"                                                                                                                                                                          };
	$ErrorCodeList{"0156"} = {errorcode => "4A",                                     subcode => "218; 332",                                           class => "4",           reason => "Temperatur am Kesselvorlauftemperaturf&uuml;hler zu hoch",                                                                                                                                                                                                                                                                                         checklist => "Sind alle Absperrh&auml;hne ge&ouml;ffnet?&bull;Betr&auml;gt der Wasserdruck in der Anlage mindestens 1 bar?&bull;Ist mindestens ein Thermostatventil ge&ouml;ffnet?&bull;Umw&auml;lzpumpe pr&uuml;fen&bull;Vorlauftemperaturf&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                                correction => "Absperrh&auml;hne &ouml;ffnen&bull;Bei zu geringem Wasserdruck Wasser nachf&uuml;llen und Anlage entl&uuml;ften&bull;Ein Thermostatventil &ouml;ffnen&bull;Ggf. Umw&auml;lzpumpe austauschen&bull;Ggf. Vorlauftemperaturf&uuml;hler austauschen"                                                                                                                                                 };
	$ErrorCodeList{"0157"} = {errorcode => "4A",                                     subcode => "505",                                                class => "4",           reason => "Kein Temperaturanstieg am Sicherheitstemperaturbegrenzer",                                                                                                                                                                                                                                                                                         checklist => "Pr&uuml;fen, ob der Sicherheitstemperaturbegrenzer richtig in der Tauchh&uuml;lse steckt",                                                                                                                                                                                                                                                                                                                                                                                                            correction => "Sicherheitstemperaturbegrenzer richtig positionieren"                                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"0158"} = {errorcode => "4A",                                     subcode => "506",                                                class => "4",           reason => "Temperaturanstieg am Sicherheitstemperaturbegrenzer zu hoch",                                                                                                                                                                                                                                                                                      checklist => "Hydraulik der Anlage bzw. Position des F&uuml;hlers pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                     correction => "F&uuml;r eine ausreichende Durchstr&ouml;mung; des Kessels sorgen bzw. F&uuml;hler richtig positionieren"                                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"0159"} = {errorcode => "4A",                                     subcode => "507",                                                class => "10",          reason => "Sicherheitstemperaturbegrenzer hat w&auml;hrend Sicherheitstemperaturbegrenzer-Test ausgel&ouml;st",                                                                                                                                                                                                                                               checklist => "Servicetaste an der Ger&auml;teelektronik BC10 h&auml;ngt oder wurde zu lange gedr&uuml;ckt",                                                                                                                                                                                                                                                                                                                                                                                                         correction => "Bei h&auml;ngender Taste die Taste l&ouml;sen oder Ger&auml;teelektronik BC10 austauschen"                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0160"} = {errorcode => "4A",                                     subcode => "520",                                                class => "4",           reason => "Vorlauftemperatur hat maximal zul&auml;ssigen Wert &uuml;berschritten (100°C)",                                                                                                                                                                                                                                                                   checklist => "Da &uuml;ber den Kesself&uuml;hler der Temperatur-anstieg im Kessel &uuml;berwacht wird und dadurch der Brenner rechtzeitig ausge-; schaltet wird, kann diese St&ouml;rungsmeldung unter normalen Umst&auml;nden nicht erscheinen Anlagenhydraulik pr&uuml;fen, ggf. &auml;ndern",                                                                                                                                                                                                                    correction => "Die St&ouml;rung kann nur bei ung&uuml;nstiger Hydraulik in Zweikessel-Anlagen auftreten, wenn Kessel sich z. B. &uuml;ber den R&uuml;cklauf oder Vorlauf gegenseitig beeinflussen"                                                                                                                                                                                                              };
	$ErrorCodeList{"0161"} = {errorcode => "4A",                                     subcode => "575",                                                class => "4",           reason => "Kesselvorlauftemperatur hat maximal zul&auml;ssigen Wert &uuml;berschritten",                                                                                                                                                                                                                                                                      checklist => "Wasserseitige Durchstr&ouml;mung pr&uuml;fen Pr&uuml;fung ob Fremdw&auml;rmequelle vorhanden (FestbrennstoffKessel)&bull;Kesself&uuml;hler/SicherheitstemperaturbegrenzerF&uuml;hler pr&uuml;fen&bull;Z&uuml;nd-/&Uuml;berwachungselektrode pr&uuml;fen",                                                                                                                                                                                                                                             correction => "Ausreichende Durchstr&ouml;mung sicherstellen&bull;Ggf. Kesself&uuml;hler/Sicherheitstemperaturbegrenzer-F&uuml;hler austauschen&bull;Ggf. Z&uuml;nd-/&Uuml;berwachungselektrode austauschen"                                                                                                                                                                                                    };
	$ErrorCodeList{"0162"} = {errorcode => "4A",                                     subcode => "700",                                                class => "4",           reason => "Werksauslieferungszustand",                                                                                                                                                                                                                                                                                                                        checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Kessel durch Reset entriegeln"                                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"0163"} = {errorcode => "4A",                                     subcode => "2038",                                               class => "4",           reason => "Solltemperatur im Mischraum nicht erreicht",                                                                                                                                                                                                                                                                                                       checklist => "Dichtung Luftklappe verschlie&szlig;en&bull;Heizpatrone defekt",                                                                                                                                                                                                                                                                                                                                                                                                                                      correction => "Dichtung Luftklappe optisch pr&uuml;fen, ggf. austauschen&bull;Heizpatrone pr&uuml;fen, ggf. austauschen"                                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"0164"} = {errorcode => "4A",                                     subcode => "2043",                                               class => "4",           reason => "Mischraumtemperatur entspricht nicht Vorgabe, Mischraumtemperatur zu niedrig oder zu hoch",                                                                                                                                                                                                                                                        checklist => "-&bull;Mischraum-Temperaturf&uuml;hler besch&auml;digt&bull;Dichtung Luftklappe undicht&bull;Verbrennungsluft-W&auml;rmetauscher verstopft&bull;Zu hoher Restsauerstoffanteil (Lambda) im Abgas&bull;-",                                                                                                                                                                                                                                                                                              correction => "Entriegeln&bull;Mischraum-Temperaturf&uuml;hler pr&uuml;fen, ggf. austauschen&bull;Luftklappendichtung pr&uuml;fen, ggf. austauschen&bull;Verbrennungsluft-W&auml;rmetauscher auf Verschmutzung pr&uuml;fen, ggf. reinigen&bull;Abgasmessung durchf&uuml;hren: CO2-Wert pr&uuml;fen&bull;&Ouml;lpumpendruck messen, ggf. &Ouml;lpumpe austauschen&bull;Einspritzventil pr&uuml;fen, ggf. austauschen&bull;Brennergebl&auml;se pr&uuml;fen, ggf. austauschen"};
	$ErrorCodeList{"0165"} = {errorcode => "4A",                                     subcode => "2090",                                               class => "4",           reason => "Temperaturanstieg der Heizpatrone zu gering",                                                                                                                                                                                                                                                                                                      checklist => "Heizpatronen-Temperaturf&uuml;hler nicht korrekt angeschlossen&bull;Heizpatrone besch&auml;digt&bull;Verbrennungsluftdurchstr&ouml;mung der Heizpatrone zu gro&szlig;&bull;Feuerungsautomat defekt",                                                                                                                                                                                                                                                                                                  correction => "Anschluss Heizpatronen-Temperaturf&uuml;hler pr&uuml;fen, ggf. korrigieren&bull;Spannungsversorgung Heizpatrone pr&uuml;fen, ggf. korrigieren&bull;Widerstand Heizpatronenwendel (< 100 Ohm) pr&uuml;fen, ggf. Heizpatrone austauschen&bull;Feuerungsautomat pr&uuml;fen, ggf. austauschen"                                                                                                      };
	$ErrorCodeList{"0166"} = {errorcode => "4A",                                     subcode => "2129",                                               class => "8",           reason => "W&auml;rmeanforderung blockiert weil Mischraum zu hei&szlig;",                                                                                                                                                                                                                                                                                     checklist => "Betriebsmeldung, keine St&ouml;rung; aktueller Betriebszustand des W&auml;rmeerzeugers",                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Keine Ma&szlig;nahme notwendig"                                                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0167"} = {errorcode => "4C",                                     subcode => "224",                                                class => "4",           reason => "Sicherheitsoder Abgastemperaturbegrenzer hat ausgel&ouml;st",                                                                                                                                                                                                                                                                                      checklist => "Sicherheitstemperaturbegrenzer auf Unterbrechung oder Kurzschluss pr&uuml;fen&bull;Abgastemperatur auf Unterbrechung oder Kurzschluss pr&uuml;fen&bull;Betr&auml;gt der Wasserdruck in der Anlage mindestens 1 bar?&bull;Pumpenanlauf pr&uuml;fen",                                                                                                                                                                                                                                                   correction => "Ggf. Sichheitstemperaturbegrenzer austauschen&bull;Ggf. Abgastemperaturbegrenzer austauschen&bull;Bei zu geringem Wasserdruck Wasser nachf&uuml;llen und Anlage entl&uuml;ften&bull;Ggf. Pumpe austauschen"                                                                                                                                                                                      };
	$ErrorCodeList{"0168"} = {errorcode => "4E",                                     subcode => "225",                                                class => "4",           reason => "Temperatur",                                                                                                                                                                                                                                                                                                                                       checklist => "Sind alle Absperrh&auml;hne ge&ouml;ffnet?&bull;Betr&auml;gt der Wasserdruck in der Anlage mindestens 1 bar?&bull;Ist mindestens ein Thermostatventil ge&ouml;ffnet?&bull;Umw&auml;lzpumpe pr&uuml;fen&bull;Sicherheitsund Vorlauftemperaturf&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                 correction => "Absperrh&auml;hne &ouml;ffnen&bull;Bei zu geringem Wasserdruck Wasser nachf&uuml;llen und Anlage entl&uuml;ften&bull;Ein Thermostatventil &ouml;ffnen&bull;Ggf. Umw&auml;lzpumpe austauschen&bull;Ggf. Doppeltemperaturf&uuml;hler austauschen"                                                                                                                                                  };
	$ErrorCodeList{"0169"} = {errorcode => "4E",                                     subcode => "278",                                                class => "10",          reason => "F&uuml;hlertest fehlgeschlagen",                                                                                                                                                                                                                                                                                                                   checklist => "Mittels Spannungsmessung pr&uuml;fen,; ob zwischen der Verkabelung des Vorlaufund des Sicherheitstemperaturf&uuml;hlers ein Kurzschluss vorliegt",                                                                                                                                                                                                                                                                                                                                                    correction => "Kabelbaum oder entsprechenden Teil davon ersetzen"                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0170"} = {errorcode => "4F",                                     subcode => "219",                                                class => "4",           reason => "Temperatur am Sicherheitstemperaturf&uuml;hler zu hoch",                                                                                                                                                                                                                                                                                           checklist => "Sind alle Absperrh&auml;hne ge&ouml;ffnet?&bull;Betr&auml;gt der Wasserdruck in der Anlage mindestens 1 bar?&bull;Ist mindestens ein Thermostatventil ge&ouml;ffnet?&bull;Umw&auml;lzpumpe pr&uuml;fen&bull;Sicherheitstemperaturf&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                            correction => "Absperrh&auml;hne &ouml;ffnen&bull;Bei zu geringem Wasserdruck,; Wasser nachf&uuml;llen und Anlage entl&uuml;ften&bull;Ein Thermostatventil &ouml;ffnen&bull;Ggf. Umw&auml;lzpumpe austauschen&bull;Ggf. Sicherheitstemperaturf&uuml;hler austauschen"                                                                                                                                           };
	$ErrorCodeList{"0171"} = {errorcode => "4L",                                     subcode => "220",                                                class => "4",           reason => "Temperatur am Sicherheitstemperaturf&uuml;hler zu hoch",                                                                                                                                                                                                                                                                                           checklist => "Sind alle Absperrh&auml;hne ge&ouml;ffnet?&bull;Betr&auml;gt der Wasserdruck in der Anlage mindestens 1 bar?&bull;Ist mindestens ein Thermostatventil ge&ouml;ffnet?&bull;Umw&auml;lzpumpe pr&uuml;fen&bull;Kabel im Kabelbaum zum Sicherheitsf&uuml;hler auf Kurzschluss pr&uuml;fen",                                                                                                                                                                                                               correction => "Absperrh&auml;hne &ouml;ffnen&bull;Bei zu geringem Wasserdruck,; Wasser nachf&uuml;llen und Anlage entl&uuml;ften&bull;Ein Thermostatventil &ouml;ffnen&bull;Ggf. Umw&auml;lzpumpe austauschen&bull;Bei Besch&auml;digung Kabelbaum austauschen"                                                                                                                                                 };
	$ErrorCodeList{"0172"} = {errorcode => "4P",                                     subcode => "221",                                                class => "4",           reason => "Keine Verbindung zum Sicherheitstemperaturf&uuml;hler",                                                                                                                                                                                                                                                                                            checklist => "Steckverbindung zum Sicherheitsf&uuml;hler auf losen Kontakt pr&uuml;fen&bull;Kabel im Kabelbaum zum Sicherheitsf&uuml;hler auf Besch&auml;digungen pr&uuml;fen&bull;Widerstandswert des Sicherheitsf&uuml;hlers pr&uuml;fen",                                                                                                                                                                                                                                                                        correction => "Steckverbindung ordnungsgem&auml;&szlig; anbringen&bull;Bei Besch&auml;digung Kabelbaum austauschen&bull;Ggf. Sicherheitsf&uuml;hler austauschen"                                                                                                                                                                                                                                                };
	$ErrorCodeList{"0173"} = {errorcode => "4U",                                     subcode => "222",                                                class => "4",           reason => "Kurzschluss Vorlauftemperaturf&uuml;hler",                                                                                                                                                                                                                                                                                                         checklist => "Kabel im Kabelbaum zum Vorlauftemperaturf&uuml;hler auf Kurzschluss pr&uuml;fen&bull;Vorlauftemperaturf&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                       correction => "Bei Besch&auml;digung Kabelbaum austauschen&bull;Ggf. Vorlauftemperaturf&uuml;hler austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0174"} = {errorcode => "4U",                                     subcode => "350",                                                class => "10",          reason => "Kurzschluss Vorlauftemperaturf&uuml;hler",                                                                                                                                                                                                                                                                                                         checklist => "Kabel im Kabelbaum zum Vorlauftemperaturf&uuml;hler auf Kurzschluss pr&uuml;fen&bull;-",                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Bei Besch&auml;digung Kabelbaum austauschen; Ggf. Vorlauftemperaturf&uuml;hler austauschen&bull;-"                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0175"} = {errorcode => "4U",                                     subcode => "521",                                                class => "4",           reason => "F&uuml;hlerdifferenz zwischen Kesselvorlauftemperaturf&uuml;hler 1 und 2 zu gro&szlig;",                                                                                                                                                                                                                                                           checklist => "Pr&uuml;fen, ob die „Reset“-Taste der Ger&auml;teelektronik SAFe leuchtet (am Brenner)&bull;Pr&uuml;fen, ob die R&uuml;ckschlagklappe an der Speicherladepumpe geschlossen ist&bull;Pr&uuml;fen, ob Heizkesselvorlauf und -r&uuml;cklauf richtig angeschlossen sind&bull;Steckverbindung am Kesself&uuml;hler und an Ger&auml;teelektronik SAFe oder Fremdbrennermodul auf Verschmutzung bzw. Besch&auml;digung pr&uuml;fen&bull;Widerstandswerte der Kesself&uuml;hler nach Tabelle pr&uuml;fen bzw. Stecker am F&uuml;hler optisch pr&uuml;fen&bull;Spannungswerte am Kesself&uuml;hler nach Tabelle pr&uuml;fen",correction => "„Reset“ an Ger&auml;teelektronik SAFe bet&auml;tigen (am Brenner)&bull;Sollte die Klappe nicht geschlossen sein, diese schlie&szlig;en&bull;Sollte eine Vertauschung vorliegen, Kessel richtig anschlie&szlig;en&bull;Ggf. Kontakte reinigen bzw. Verbindungsleitung austauschen&bull;Wenn Abweichungen der F&uuml;hlerwerte bestehen oder der Stecker defekt ist,; Kesself&uuml;hler austauschen&bull;Wenn Abweichungen bestehen, Ger&auml;teelektronik SAFe oder Fremdbrennermodul austauschen"};
	$ErrorCodeList{"0176"} = {errorcode => "4U",                                     subcode => "522",                                                class => "4",           reason => "Kurzschluss zwischen Kesselvorlauftemperaturf&uuml;hler 1 und 2",                                                                                                                                                                                                                                                                                  checklist => "F&uuml;hlerleitung pr&uuml;fen&bull;Steckverbindung pr&uuml;fen&bull;-&bull;F&uuml;hlerwerte nach Tabelle pr&uuml;fen&bull;Spannungswerte am F&uuml;hler laut Tabelle pr&uuml;fen",                                                                                                                                                                                                                                                                                                                   correction => "Bei Besch&auml;digung austauschen&bull;Bei Verschmutzung reinigen oder ggf. austauschen&bull;Bei Besch&auml;digung austauschen&bull;Bei losem Stecker diesen wieder aufstecken&bull;Bei Abweichungen den F&uuml;hler austauschen&bull;Bei Abweichungen Ger&auml;teelektronik SAFe oder Fremdbrennermodul austauschen"                                                                            };
	$ErrorCodeList{"0177"} = {errorcode => "4U",                                     subcode => "524",                                                class => "4",           reason => "Kurzschluss Kesselvorlauftemperaturf&uuml;hler",                                                                                                                                                                                                                                                                                                   checklist => "F&uuml;hlerleitung pr&uuml;fen&bull;Steckverbindung pr&uuml;fen&bull;-&bull;F&uuml;hlerwerte nach Tabelle pr&uuml;fen&bull;Spannungswerte am F&uuml;hler laut Tabelle pr&uuml;fen",                                                                                                                                                                                                                                                                                                                   correction => "Bei Besch&auml;digung austauschen&bull;Bei Verschmutzung reinigen oder ggf. austauschen&bull;Bei Besch&auml;digung austauschen&bull;Bei losem Stecker diesen wieder aufstecken&bull;Bei Abweichungen den F&uuml;hler austauschen&bull;Bei Abweichungen Ger&auml;teelektronik SAFe oder Fremdbrennermodul austauschen"                                                                            };
	$ErrorCodeList{"0178"} = {errorcode => "4U",                                     subcode => "532",                                                class => "10",          reason => "Netzspannung zu niedrig",                                                                                                                                                                                                                                                                                                                          checklist => "Verdrahtung oder Netzspannung pr&uuml;fen&bull;Fremdbrennermodul defekt&bull;EMV-Probleme",                                                                                                                                                                                                                                                                                                                                                                                                           correction => "Verdrahtung korrigieren oder f&uuml;r ausreichend Netzspannung sorgen&bull;Fremdbrennermodul austauschen&bull;EMV-Probleme beseitigen"                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"0179"} = {errorcode => "4U",                                     subcode => "2006",                                               class => "4",           reason => "Kurzschluss MischraumTemperaturf&uuml;hler",                                                                                                                                                                                                                                                                                                       checklist => "Zu geringe Temperatur vom MischraumTemperaturf&uuml;hler gemessen&bull;-",                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "Mischraum-Temperaturf&uuml;hler pr&uuml;fen, ggf. austauschen; F&uuml;hlerwerte nach Tabelle pr&uuml;fen, Spannungswerte am Kesself&uuml;hler nach Tabelle pr&uuml;fen&bull;Steckverbindung pr&uuml;fen, ggf. Kabelbaum oder Temperaturf&uuml;hler austauschen&bull;Feuerungsautomat austauschen"                                                                                                };
	$ErrorCodeList{"0180"} = {errorcode => "4U",                                     subcode => "2009",                                               class => "4",           reason => "Differenz zwischen Mischraum-Temperaturf&uuml;hler 1 und 2 zu gro&szlig;",                                                                                                                                                                                                                                                                         checklist => "Mischraum-Temperaturf&uuml;hler misst fehlerhafte Werte&bull;-",                                                                                                                                                                                                                                                                                                                                                                                                                                      correction => "F&uuml;hlerkabel pr&uuml;fen, ggf. austauschen&bull;Temperaturf&uuml;hler pr&uuml;fen, ggf. austauschen&bull;Feuerungsautomat austauschen&bull;Kundendienst verst&auml;ndigen"                                                                                                                                                                                                                   };
	$ErrorCodeList{"0181"} = {errorcode => "4U",                                     subcode => "2023",                                               class => "4",           reason => "Kurzschluss F&uuml;hler Heizpatrone",                                                                                                                                                                                                                                                                                                              checklist => "Heizpatronen-Temperaturf&uuml;hler liefert fehlerhafte Werte&bull;-",                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "F&uuml;hlerleitung pr&uuml;fen, ggf. austauschen&bull;Spannungsversorgung Kesselregelger&auml;t pr&uuml;fen, ggf. Anschluss korrigieren&bull;Polung der Anschlussleitung pr&uuml;fen, ggf. Anschluss korrigieren"                                                                                                                                                                                };
	$ErrorCodeList{"0182"} = {errorcode => "4U",                                     subcode => "2100",                                               class => "4",           reason => "Kurzschluss MischraumTemperaturf&uuml;hler",                                                                                                                                                                                                                                                                                                       checklist => "Mischraum-Temperaturf&uuml;hler defekt&bull;Steckverbindung Mischraum-Temperaturf&uuml;hler besch&auml;digt oder nicht aufgesteckt&bull;Feuerungsautomat defekt",                                                                                                                                                                                                                                                                                                                                     correction => "Temperaturf&uuml;hler pr&uuml;fen, ggf. austauschen&bull;Steckverbindung Mischraum-Temperaturf&uuml;hler pr&uuml;fen, ggf. aufstecken&bull;Feuerungsautomat austauschen"                                                                                                                                                                                                                         };
	$ErrorCodeList{"0183"} = {errorcode => "4Y",                                     subcode => "223",                                                class => "4",           reason => "Kontakt Vorlauftemperaturf&uuml;hler lose oder defekt",                                                                                                                                                                                                                                                                                            checklist => "Steckverbindung zum Vorlauftemperaturf&uuml;hler auf losen Kontakt pr&uuml;fen&bull;Kabel im Kabelbaum zum Vorlauftemperaturf&uuml;hler auf Besch&auml;digungen pr&uuml;fen&bull;Widerstandswert des Vorlauftemperaturf&uuml;hlers pr&uuml;fen",                                                                                                                                                                                                                                                      correction => "Steckverbindung ordnungsgem&auml;&szlig; anbringen&bull;Bei Besch&auml;digung Kabelbaum austauschen&bull;Ggf. Vorlauftemperaturf&uuml;hler austauschen"                                                                                                                                                                                                                                          };
	$ErrorCodeList{"0184"} = {errorcode => "4Y",                                     subcode => "351",                                                class => "10",          reason => "Unterbrechung Vorlauftemperaturf&uuml;hler",                                                                                                                                                                                                                                                                                                       checklist => "Kabel im Kabelbaum zum Vorlauftemperaturf&uuml;hler pr&uuml;fen&bull;Vorlauftemperaturf&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                       correction => "Bei Besch&auml;digung Kabelbaum austauschen&bull;Ggf. Vorlauftemperaturf&uuml;hler austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0185"} = {errorcode => "4Y",                                     subcode => "523",                                                class => "4",           reason => "Unterbrechung Kesselvorlauftemperaturf&uuml;hler",                                                                                                                                                                                                                                                                                                 checklist => "F&uuml;hlerleitung pr&uuml;fen&bull;Steckverbindung pr&uuml;fen&bull;-&bull;F&uuml;hlerwerte nach Tabelle pr&uuml;fen&bull;Spannungswerte am F&uuml;hler laut Tabelle pr&uuml;fen",                                                                                                                                                                                                                                                                                                                   correction => "Bei Besch&auml;digung austauschen&bull;Bei Verschmutzung reinigen oder ggf. austauschen&bull;Bei Besch&auml;digung austauschen&bull;Bei losem Stecker diesen wieder aufstecken&bull;Bei Abweichungen den F&uuml;hler austauschen&bull;Bei Abweichungen Ger&auml;teelektronik SAFe oder Fremdbrennermodul austauschen"                                                                            };
	$ErrorCodeList{"0186"} = {errorcode => "4Y",                                     subcode => "2005",                                               class => "4",           reason => "Unterbrechung MischraumTemperaturf&uuml;hler",                                                                                                                                                                                                                                                                                                     checklist => "Zu hohe Temperatur vom MischraumTemperaturf&uuml;hler gemessen&bull;Steckverbindung Mischraum-Temperaturf&uuml;hler fehlerhaft oder besch&auml;digt&bull;Feuerungsautomat defekt",                                                                                                                                                                                                                                                                                                                    correction => "Mischraum-Temperaturf&uuml;hler pr&uuml;fen, ggf. austauschen; F&uuml;hlerwerte nach Tabelle pr&uuml;fen, Spannungswerte am Kesself&uuml;hler nach Tabelle pr&uuml;fen&bull;Steckverbindung pr&uuml;fen, ggf. Kabelbaum oder Temperaturf&uuml;hler austauschen&bull;Feuerungsautomat austauschen"                                                                                                };
	$ErrorCodeList{"0187"} = {errorcode => "5A",                                     subcode => "275",                                                class => "20",          reason => "Ger&auml;teelektronik UBA im Testmode",                                                                                                                                                                                                                                                                                                            checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0188"} = {errorcode => "5A",                                     subcode => "507",                                                class => "BC",          reason => "Sicherheitstemperaturbegrenzer-Test erfolgreich durchgef&uuml;hrt",                                                                                                                                                                                                                                                                                checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0189"} = {errorcode => "5C",                                     subcode => "226",                                                class => "BC",          reason => "Kennzeichnung f&uuml;r Handterminal",                                                                                                                                                                                                                                                                                                              checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0190"} = {errorcode => "5E",                                     subcode => "586",                                                class => "4",           reason => "Ger&auml;teelektronik SAFe alter Softwarestand",                                                                                                                                                                                                                                                                                                   checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Ger&auml;teelektronik SAFe mit neuem Softwarestand (ab Version 4.X) einsetzen"                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"0191"} = {errorcode => "5H",                                     subcode => "268",                                                class => "BC",          reason => "Relaistest wurde aktiviert",                                                                                                                                                                                                                                                                                                                       checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0192"} = {errorcode => "5H",                                     subcode => "310",                                                class => "10",          reason => "Keine Kommunikation mit dem EMS W&auml;rmeerzeuger",                                                                                                                                                                                                                                                                                               checklist => "Es kann keine Kommunikation zwischen dem Hybridmodul (Inneneinheit) und dem EMSW&auml;rmeerzeuger hergestellt werden&bull;Kontakte, Kabel und Verbindungen zwischen Hybridmodul und EMS-W&auml;rmeerzeuger pr&uuml;fen",                                                                                                                                                                                                                                                                              correction => "EMS-Verbindung zwischen dem Hybridmodul (Inneneinheit) und dem EMS-W&auml;rmeerzeuger &uuml;berpr&uuml;fen&bull;Besch&auml;digte Leitungen und Kontakte austauschen, ggf. Ger&auml;teelektronik Hybridmodul (Inneneinheit) und EMS W&auml;rmeerzeuger MC10/ MC40/MC100/BC10/BC25 austauschen"                                                                                                    };
	$ErrorCodeList{"0193"} = {errorcode => "5H",                                     subcode => "323",                                                class => "n.a.",        reason => "EMS Bussignal kann im Ger&auml;t nicht verarbeitet werden",                                                                                                                                                                                                                                                                                        checklist => "&Uuml;berpr&uuml;fen, ob alle Anschlussstecker f&uuml;r den EMS BUS aufgesteckt sind&bull;Anschlussstecker auf korrekte Verdrahtung pr&uuml;fen&bull;Anschlusskabel pr&uuml;fen",                                                                                                                                                                                                                                                                                                                     correction => "Anschlussstecker mit der Beschriftung EMS im Steuerger&auml;t aufstecken&bull;Anschlusstecker mit der Beschrifung EMS austauschen&bull;Anschlusskabel f&uuml;r den 2 Draht Bus austauschen"                                                                                                                                                                                                      };
	$ErrorCodeList{"0194"} = {errorcode => "5H",                                     subcode => "470",                                                class => "10",          reason => "Keine Kommunikation mit dem Systemregler",                                                                                                                                                                                                                                                                                                         checklist => "Kontakte, Kabel und Verbindungen zwischen Hybridmodul und Systemregler pr&uuml;fen&bull;Systemregler pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                    correction => "Besch&auml;digte Leitungen und Kontakte austauschen&bull;Ggf. ersetzen"                                                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"0195"} = {errorcode => "5H",                                     subcode => "2113",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Entriegeln"                                                                                                                                                                                                                                                                                                                                                                                      };
	$ErrorCodeList{"0196"} = {errorcode => "5H",                                     subcode => "2504",                                               class => "BC",          reason => "W&auml;rmeanforderung wegen Relaistest",                                                                                                                                                                                                                                                                                                           checklist => "W&auml;rmeanforderung wegen Relaistest",                                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0197"} = {errorcode => "5L",                                     subcode => "542",                                                class => "10",          reason => "Kommunikation mit Ger&auml;teelektronik SAFe oder BRM10/Regelger&auml;t/Feuerungsautomat oder Fremdbrennermodul unvollst&auml;ndig",                                                                                                                                                                                                               checklist => "Leitungsverbindungen zwischen Feuerungsautomat oder Fremdbrennermodul und Regelger&auml;t pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                               correction => "Ger&auml;teelektronik SAFe oder Fremdbrennermodul austauschen"                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"0198"} = {errorcode => "5L",                                     subcode => "543",                                                class => "10",          reason => "Keine Kommunikation mit Ger&auml;teelektronik/ Fremdbrennermodul; (Auswirkung: schnelles Blinken; der LED am SAFe bzw. UM10 bzw. BRM10 (= Not-; betrieb))",                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob die Stecker der Kabel (Busleitung und Netzleitung) zwischen Ger&auml;teelektronik oder Fremdbrennermodul und Regelger&auml;t richtig aufgesteckt sind&bull;Im Regelger&auml;t an den Anschlussklemmen „Netz Ger&auml;teelektronik“ pr&uuml;fen ob 230 V AC anliegen&bull;Pr&uuml;fen, ob die Verbindungsleitungen (Busleitung und Netzleitung) zwischen Ger&auml;teelektronik und Regelger&auml;t evtl. besch&auml;digt sind&bull;Die Busverbindungsleitung zwischen Ger&auml;teelektronik trennen und pr&uuml;fen, ob der Kessel in Notbetrieb geht (l&auml;uft auf 60°C Kesseltemperatur)&bull;Pr&uuml;fen, ob am Ger&auml;teelektronik die gr&uuml;ne Signallampe leuchtet&bull;Wenn die Ger&auml;teelektronik dunkel bleibt, eine gewisse Zeit abwarten, da bei kalter Ger&auml;teelektronik das Ger&auml;t evtl. nicht anl&auml;uft&bull;Durch Austausch pr&uuml;fen, ob Regelger&auml;t ger&auml;teelektronik oder fremdbrennermodul defekt ist&bull;Durch Austausch pr&uuml;fen, ob der Ger&auml;teelektronik oder Fremdbrennermodul defekt ist",correction => "Stecker richtig aufstecken&bull;Wenn keine 230 V AC anliegen, dann ist das Regelger&auml;t defekt und muss ausgetauscht werden&bull;Verbindungsleitung austauschen&bull;Wenn die Lampen nicht leuchten, dann ist vermutlich die Ger&auml;teelektronik defekt und muss ausgetauscht werden&bull;Wenn der Kessel nicht anl&auml;uft, dann ist die Ger&auml;teelektronik defekt und muss ausgetauscht werden&bull;Max. 30 Minuten warten und pr&uuml;fen, ob dann die gr&uuml;ne Lampe an der Ger&auml;teelektronik; wieder leuchtet; Wenn dies nicht der Fall ist, dann muss die Ger&auml;teelektronik ausgetauscht werden&bull;Regelger&auml;t austauschen&bull;Max. 30 Minuten warten und pr&uuml;fen, ob; dann die gr&uuml;ne Lampe an der Ger&auml;teelektronik wieder leuchtet; Wenn dies nicht der Fall ist, dann muss die Ger&auml;teelektronik ausgetauscht werden"};
	$ErrorCodeList{"0199"} = {errorcode => "5L",                                     subcode => "2051",                                               class => "10",          reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "Sicherheits-Controller blockiert&bull;-",                                                                                                                                                                                                                                                                                                                                                                                                                                                             correction => "Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen&bull;Kundendienst verst&auml;ndigen"                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"0200"} = {errorcode => "5P",                                     subcode => "552",                                                class => "4",           reason => "„Reset“-Taste zu h&auml;ufig bet&auml;tigt; (Wird innerhalb kurzer Zeit die „Reset“-Taste am MC10 zu haufig betatigt, wird diese Fehler meldung erzeugt.; ACHTUNG: Dieser Fehler lasst sich nur uber die Taste am SAFe entriegeln!)",                                                                                                      checklist => "Pr&uuml;fen, ob die „Reset“-Taste am Regelger&auml;t evtl. klemmt&bull;-",                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Taste wieder l&ouml;sen&bull;Wenn die „Reset“-Taste in Ordnung ist, dann muss der Basiscontroller ausgetauscht werden"                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0201"} = {errorcode => "5U",                                     subcode => "582",                                                class => "10",          reason => "Keine Kommunikation mit Umschaltmodul",                                                                                                                                                                                                                                                                                                            checklist => "Verbindungsleitung zwischen Feuerungsautomat oder Fremdbrennermodul und Umschaltmodul pr&uuml;fen&bull;Sicherung am Umschaltmodul pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                       correction => "Bei defekten Kabeln diese austauschen, bei losem Stecker diesen richtig aufstecken&bull;Sicherung austauschen"                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"0202"} = {errorcode => "5U",                                     subcode => "588",                                                class => "4",           reason => "Mehr als ein Umschaltmodul im System",                                                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Alle Module bis auf ein Umschaltmodul entfernen"                                                                                                                                                                                                                                                                                                                                                 };
	$ErrorCodeList{"0203"} = {errorcode => "5Y",                                     subcode => "585",                                                class => "4",           reason => "Umschaltmodul nicht vorhanden",                                                                                                                                                                                                                                                                                                                    checklist => "Wenn das Umschaltmodul ausgebaut wird, muss es auch softwareseitig deinstalliert werden (siehe Montageund Bedienungsanleitung)",                                                                                                                                                                                                                                                                                                                                                                      correction => "Entsprechend der Montageund Bedienungsanleitung einen Reset am Raumcontroller durchf&uuml;hren, um das Umschaltmodul zu deinstallieren bzw. das Modul auszutauschen"                                                                                                                                                                                                                             };
	$ErrorCodeList{"0204"} = {errorcode => "6A",                                     subcode => "227",                                                class => "4",           reason => "Kein Flammensignal nach Z&uuml;ndung; (Keine Ionisationsmeldung nach; der Zundung. Nach dem 4. Versuch (GB172: 5. Versuch) wird die; St&ouml;rung 6A erzeugt. Zwischen erfolglosen Zundungsversuchen Anzeige blockierend (6A nicht blinkend). Nach 5 erfolglosen; Zundungsversuchen Anzeige; verriegelnd (6A blinkend),; „Reset“ zum Entriegeln)", checklist => "Gasanschlussdruck messen, Gasd&uuml;se pr&uuml;fen&bull;Ist die Gasleitung entl&uuml;ftet?&bull;Gl&uuml;hz&uuml;nder in Ordnung?&bull;Die Spannung zum Gl&uuml;hz&uuml;nder muss 120 V AC betragen&bull;&Ouml;ffnet die Gasarmatur bei 0L? Elektrischen Anschluss Gasarmatur pr&uuml;fen&bull;Zuluftund Abgasweg sowie Brennkammer auf Verschmutzung pr&uuml;fen&bull;Ionisationsstrom messen&bull;Ist das Kesselidentifikationsmodul richtig aufgesteckt?&bull;Bei raumluftabh&auml;ngiger Betriebsweise den Raumluftverbund oder die L&uuml;ftungs&ouml;ffnung pr&uuml;fen&bull;Abgasweg kontollieren und den Kondensatsiphon pr&uuml;fen&bull;W&auml;rmetauscher auf Verschmutzung pr&uuml;fen&bull;Ist ein Stromnetz (IT) vorhanden, in Deutschland praktisch nicht vorkommend (au&szlig;er Insell&ouml;sungen)?",correction => "Ist der ben&ouml;tigte Gasanschlussdruck nicht vorhanden, mit dem zust&auml;ndigen Gasversorger R&uuml;cksprache halten; Gas-Luft-Verh&auml;ltnis auf: 5 Pa einstellen F&uuml;r die Gasart korrekte Gasd&uuml;se einsetzen&bull;Gasleitung entl&uuml;ften&bull;Widerstand 50: 300 Ω, dann Gl&uuml;hz&uuml;nder in Ordnung, sonst auswechseln&bull;Wenn keine Spannung anliegt, Steckverbindung pr&uuml;fen oder defekten Trafo austauschen&bull;Spannung muss w&auml;hrend 0L zwischen 20 und 24 V AC liegen&bull;Verschmutzung beseitigen&bull;Ionisationsstrom muss in der Teillast > 1,4 μA betragen, sonst &Uuml;berwachungselektrode austauschen&bull;Kesselidentifikationsmodul richtig aufstecken oder ggf. austauschen&bull;-&bull;Ggf. Abgasweg reinigen bzw. instand setzen und den Kondensatsiphon reinigen&bull;Ggf. W&auml;rmetauscher reinigen&bull;Am Netzanschluss an der Leiterplatte muss dann ein Widerstand von 2 M Ω eingebaut werden"};
	$ErrorCodeList{"0205"} = {errorcode => "6A",                                     subcode => "504",                                                class => "10",          reason => "Brennerst&ouml;rung",                                                                                                                                                                                                                                                                                                                              checklist => "Der vorhandene Nicht-EMS-Brenner ist auf St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                               correction => "Brenner am Feuerungsautomaten entriegeln"                                                                                                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"0206"} = {errorcode => "6A",                                     subcode => "577",                                                class => "10",          reason => "Keine Flamme innerhalb Sicherheitszeit; (Innerhalb der Sicherheitszeit ist Ionisationsstrom <; 1.1 μA)",                                                                                                                                                                                                                                          checklist => "Pr&uuml;fen, ob der Gashahn ge&ouml;ffnet ist&bull;Gasanschlussdruck pr&uuml;fen&bull;Querschnitte der Gasleitung nicht ausreichend dimensioniert&bull;Gegendruck der Abgasanlage zu hoch durch ung&uuml;nstige Ausf&uuml;hrung (zu viele Umlenkungen, Querschnitte zu gering, zu lange waagerechte Strecken)&bull;Verbindungsleitung zwischen Ger&auml;teelektronik SAFe und &Uuml;berwachungselektrode auf fehlerhafte Kontaktierung, Unterbrechung und Besch&auml;digung pr&uuml;fen&bull;Elektrodenabst&auml;nde und Z&uuml;nd-/&Uuml;berwachungselektrode auf Besch&auml;digungen pr&uuml;fen&bull;Luft in der Gasleitung&bull;Startgasd&uuml;se verschmutzt&bull;Gasarmatur defekt (&uuml;ber Klickger&auml;usch der Armatur h&ouml;rbar)&bull;Masseanschluss der &Uuml;berwachungselektrode pr&uuml;fen&bull;&Uuml;berwachungselektrode verschmutzt&bull;Z&uuml;ndung und Flammen&uuml;berwachung pr&uuml;fen&bull;Z&uuml;ndtrafo defekt (kein oder verz&ouml;gerter Z&uuml;ndfunke, „harter Start“)&bull;Feuerungsautomat/Ger&auml;teelektronik SAFe defekt",correction => "Gashahn &ouml;ffnen&bull;Bei zu niedrigem Gasanschlussdruck den Gasversorger informieren&bull;Ausreichend dimensionierte Gasleitung installieren&bull;Abgasanlage korrekt dimensionieren&bull;Korrekte Kontaktierung herstellen, ggf. Kabel austauschen&bull;Brennstab und Elektrode ausrichten, defekte Elektroden austauschen&bull;Gasleitung entl&uuml;ften&bull;Startgasd&uuml;se reinigen&bull;Gasarmatur austauschen&bull;Masseverbindung der &Uuml;berwachungselektrode herstellen&bull;&Uuml;berwachungselektrode reinigen bzw. austauschen&bull;Austausch Z&uuml;ndung und Flammen&uuml;berwachung&bull;Z&uuml;ndtrafo austauschen&bull;Feuerungsautomat/Ger&auml;teelektronik SAFe austauschen"};
	$ErrorCodeList{"0207"} = {errorcode => "6A",                                     subcode => "578",                                                class => "4",           reason => "Wiederholt keine Flamme innerhalb Sicherheitszeit",                                                                                                                                                                                                                                                                                                checklist => "Hauptabsperreinrichtung &uuml;berpr&uuml;fen&bull;Ger&auml;teabsperrhan &uuml;berpr&uuml;fen&bull;Gasanschlu&szlig;flie&szlig;druck bei Nennw&auml;rmebelastung messen&bull;Anschlusstecker an der Ionisationselektrode &uuml;berpr&uuml;fen&bull;Ionisationselektrode &uuml;berpr&uuml;fen&bull;Z&uuml;ndkabel auf Besch&auml;digungen &uuml;berpr&uuml;fen&bull;Anschlusskabel f&uuml;r die Ionisationselektrode auf Besch&auml;digungen &uuml;berpr&uuml;fen&bull;Z&uuml;ndtransformator &uuml;berpr&uuml;fen&bull;Brennereinstellwert bei kleinster Leistung &uuml;berpr&uuml;fen&bull;Verbrennungsluftversorgung &uuml;berpr&uuml;fen",correction => "Hauptabsperreinrichtung &ouml;ffnen&bull;Ger&auml;teabsperrhahn &ouml;ffnen&bull;Ger&auml;t stilllegen und Gasleitung &uuml;berpr&uuml;fen&bull;Anschlussstecker an der Ionisationelektrode aufstecken&bull;Ionisationselektrode austauschen&bull;Z&uuml;ndkabel austauschen&bull;Anschlusskabel zur Ionisationselektrode austauschen&bull;Z&uuml;ndtransformator austauschen&bull;Brenner bei minimaler Nennbelastung einstellen&bull;Verbrennungsluftverbund zu klein bzw. zu geringe Gr&ouml;&szlig;e der L&uuml;ftungs&ouml;ffnung"};
	$ErrorCodeList{"0208"} = {errorcode => "6C",                                     subcode => "228",                                                class => "4",           reason => "Flammensignal trotz nicht vorhandener Flamme",                                                                                                                                                                                                                                                                                                     checklist => "&Uuml;berwachungselektrode auf Besch&auml;digung bzw. Verschmutzung pr&uuml;fen&bull;Leiterplatte auf Feuchtigkeit pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                      correction => "&Uuml;berwachungselektrode austauschen bzw. reinigen&bull;Leiterplatte trocknen"                                                                                                                                                                                                                                                                                                                 };
	$ErrorCodeList{"0209"} = {errorcode => "6C",                                     subcode => "306",                                                class => "4",           reason => "Flammensignal nach Schlie&szlig;en der Brennstoffversorgung",                                                                                                                                                                                                                                                                                      checklist => "Str&ouml;mt nach Abschaltung Gas-Luft-Gemisch in Brennkammer, obwohl Gasarmatur spannungslos ist?&bull;Elektroden und Anschlusskabel pr&uuml;fen&bull;Kondensatablauf pr&uuml;fen",                                                                                                                                                                                                                                                                                                                   correction => "Gasarmatur austauschen&bull;Ggf. austauschen&bull;Kondensatablauf reinigen"                                                                                                                                                                                                                                                                                                                      };
	$ErrorCodeList{"0210"} = {errorcode => "6C",                                     subcode => "508",                                                class => "4",           reason => "Flammensignal zu hoch",                                                                                                                                                                                                                                                                                                                            checklist => "„Reset“-Taste bet&auml;tigen und warten, ob die St&ouml;rung behoben ist",                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Bleibt die St&ouml;rung nach Reset erhalten, ist die Ger&auml;teelektronik SAFe defekt und muss ausgetauscht werden"                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"0211"} = {errorcode => "6C",                                     subcode => "509",                                                class => "4",           reason => "Eingang Flammenf&uuml;hler defekt; (Bei der &Uuml;berpr&uuml;fung der Eingangsbeschaltung des Flammenf&uuml;hlers wurde ein Fehler erkannt.)",                                                                                                                                                                                                     checklist => "&Uuml;ber die Bedieneinheit die Anlage in einen Ruhezustand, z. B. Absenkbetrieb, setzen, sodass der Brenner ausgeschaltet ist&bull;Im Ruhezustand in der Bedieneinheit im Men&uuml;; „Monitor“ den F&uuml;hlerstrom pr&uuml;fen&bull;-",                                                                                                                                                                                                                                                         correction => "-&bull;Falls der F&uuml;hlerstrom zu gro&szlig; ist, die Position des Flammenf&uuml;hlers pr&uuml;fen; Evtl. gelangt Fremdlicht an den Flammenf&uuml;hler&bull;Falls ein F&uuml;hlerstrom von ca. 0 μA angezeigt wird, ist die Ger&auml;teelektronik SAFe defekt und muss ausgetauscht werden"                                                                                                  };
	$ErrorCodeList{"0212"} = {errorcode => "6C",                                     subcode => "519",                                                class => "4",           reason => "Kein Flammenabriss/Nachbel&uuml;ftung; (W&auml;hrend der Nachbel&uuml;ftungsphase; ging das Flammensignal nicht aus.)",                                                                                                                                                                                                                            checklist => "Stecker Magnetventil 1. Stufe an Ger&auml;teelektronik SAFe entfernen und im Men&uuml; „Monitor“ der Bedieneinheit den Flammenstrom pr&uuml;fen&bull;Erneut den Flammenstrom im Men&uuml; „Monitor“ der Bedieneinheit pr&uuml;fen",                                                                                                                                                                                                                                                           correction => "Sollte ein Flammenstrom von mehr als ca. 0 μA angezeigt werden, das Magnetventil; der 1. Stufe austauschen&bull;Sollte weiterhin ein Flammenstrom von mehr als ca. 0 μA angezeigt werden, den Flammenf&uuml;hler austauschen"                                                                                                                                                                  };
	$ErrorCodeList{"0213"} = {errorcode => "6C",                                     subcode => "576",                                                class => "4",           reason => "Flammensignal w&auml;hrend Vorbel&uuml;ftung",                                                                                                                                                                                                                                                                                                     checklist => "Flammenf&uuml;hler auf Defekte pr&uuml;fen&bull;Gasarmatur auf Funktion pr&uuml;fen&bull;Wenn Gasarmatur in Ordnung ist",                                                                                                                                                                                                                                                                                                                                                                             correction => "Flammenf&uuml;hler austauschen&bull;Gasarmatur austauschen&bull;Ger&auml;teelektronik SAFe austauschen"                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"0214"} = {errorcode => "6C",                                     subcode => "2041",                                               class => "4",           reason => "Fremdlicht im Feuerraum w&auml;hrend Nachbel&uuml;ftung erkannt; Flamme erlischt nicht, nachdem Magnetventil &Ouml;lpumpe geschlossen",                                                                                                                                                                                                            checklist => "Magnetventil &Ouml;lpumpe defekt&bull;-",                                                                                                                                                                                                                                                                                                                                                                                                                                                             correction => "Entriegeln&bull;Magnetventileinsatz &Ouml;lpumpe austauschen"                                                                                                                                                                                                                                                                                                                                    };
	$ErrorCodeList{"0215"} = {errorcode => "6E",                                     subcode => "556",                                                class => "10",          reason => "Hauptflamme zu fr&uuml;h",                                                                                                                                                                                                                                                                                                                         checklist => "Diese Funktion ist zurzeit nicht aktiv.",                                                                                                                                                                                                                                                                                                                                                                                                                                                             correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0216"} = {errorcode => "6F",                                     subcode => "2510",                                               class => "BC",          reason => "W&auml;rmeanforderung blockiert, weil 24 Stunden vorbei",                                                                                                                                                                                                                                                                                          checklist => "Sicherheitsabschaltung nach 24 Stunden",                                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Keine Ma&szlig;nahme, W&auml;rmeerzeuger versucht Wiederanlauf"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0217"} = {errorcode => "6L",                                     subcode => "229",                                                class => "10",          reason => "Flamme w&auml;hrend Brennerbetrieb ausgefallen; (Flamme wahrend des Brenner-betriebs ausgefallen.; Der Brenner startet von Neuem, misslingt dieser Start auch, wird der Fehler zu einer blockierenden 6A Storung. Nach weiteren vier Startversuchen wird diese zu einer verriegelnden 6A St&ouml;rung.)",                                          checklist => "Gasanschlussdruck messen, Gas-Luft-Verh&auml;ltnis messen, Gasd&uuml;se pr&uuml;fen; Bei Fl&uuml;ssiggas: Gen&uuml;gend Gasvorrat im Tank oder Tank neu bef&uuml;llt? Kondensatsiphon pr&uuml;fen&bull;&Uuml;berwachungselektrode auf Besch&auml;digung bzw. Verschmutzung pr&uuml;fen&bull;Kondensatablauf pr&uuml;fen",                                                                                                                                                                             correction => "Ist der ben&ouml;tigte Gasanschlussdruck nicht vorhanden, mit dem zust&auml;ndigen Gasversorger R&uuml;cksprache halten; Gas-Luft-Verh&auml;ltnis auf -5 Pa einstellen  F&uuml;r die Gasart korrekte Gasd&uuml;se einsetzen&bull;&Uuml;berwachungselektrode austauschen bzw. reinigen&bull;Kondensatablauf reinigen"                                                                             };
	$ErrorCodeList{"0218"} = {errorcode => "6L",                                     subcode => "512",                                                class => "10",          reason => "Flammenabriss innerhalb Sicherheitszeit",                                                                                                                                                                                                                                                                                                          checklist => "Flammensignal ging innerhalb der Sicherheitszeit aus.",                                                                                                                                                                                                                                                                                                                                                                                                                                               correction => "Keine Ma&szlig;nahme, Ger&auml;teelektronik SAFe versucht Wiederanlauf"                                                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"0219"} = {errorcode => "6L",                                     subcode => "513",                                                class => "10",          reason => "Flammenabriss innerhalb Nachz&uuml;ndzeit",                                                                                                                                                                                                                                                                                                        checklist => "Das Flammensignal ging innerhalb der Nachzundzeit aus.",                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Keine Ma&szlig;nahme, Ger&auml;teelektronik SAFe versucht Wiederanlauf"                                                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"0220"} = {errorcode => "6L",                                     subcode => "514",                                                class => "10",          reason => "Flammenabriss innerhalb Stabilisierungszeit",                                                                                                                                                                                                                                                                                                      checklist => "Das Flammensignal ging innerhalb der Stabilisierungszeit aus.",                                                                                                                                                                                                                                                                                                                                                                                                                                       correction => "Keine Ma&szlig;nahme, Ger&auml;teelektronik SAFe versucht Wiederanlauf"                                                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"0221"} = {errorcode => "6L",                                     subcode => "515",                                                class => "10",          reason => "Flammenabriss in Betrieb; 1. + 2. Stufe",                                                                                                                                                                                                                                                                                                          checklist => "Das Flammensignal ging wahrend des Betriebs von der 1. + 2. Stufe aus.",                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Keine Ma&szlig;nahme, Ger&auml;teelektronik SAFe versucht Wiederanlauf"                                                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"0222"} = {errorcode => "6L",                                     subcode => "516",                                                class => "10",          reason => "Flammenabriss Umschaltung 1. Stufe",                                                                                                                                                                                                                                                                                                               checklist => "Das Flammensignal ging w&auml;hrend der Umschaltung auf die 1. Stufe aus.",                                                                                                                                                                                                                                                                                                                                                                                                                           correction => "Keine Ma&szlig;nahme, Ger&auml;teelektronik SAFe versucht Wiederanlauf"                                                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"0223"} = {errorcode => "6L",                                     subcode => "517",                                                class => "10",          reason => "Flammenabriss in Betrieb 1. Stufe",                                                                                                                                                                                                                                                                                                                checklist => "Das Flammensignal ging w&auml;hrend des Betriebs von der 1. Stufe aus.",                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Keine Ma&szlig;nahme, Ger&auml;teelektronik SAFe versucht Wiederanlauf"                                                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"0224"} = {errorcode => "6L",                                     subcode => "518",                                                class => "10",          reason => "Flammenabriss Umschaltung 1. + 2. Stufe",                                                                                                                                                                                                                                                                                                          checklist => "Das Flammensignal ging w&auml;hrend der Umschaltung von 1. Stufe auf 2. Stufe bzw. von 2. Stufe auf 1. Stufe aus.",                                                                                                                                                                                                                                                                                                                                                                                   correction => "Keine Ma&szlig;nahme, Ger&auml;teelektronik SAFe versucht Wiederanlauf"                                                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"0225"} = {errorcode => "6L",                                     subcode => "548",                                                class => "4",           reason => "Zu viele Brennerstartversuche; (W&auml;hrend einer W&auml;rmeanforderung sind 6 Flammenabrisse aufgetreten.)",                                                                                                                                                                                                                                     checklist => "In der Serviceebene der Bedieneinheit den Men&uuml;punkt „St&ouml;rungsspeicher“ und danach das Untermen&uuml; „blockierende St&ouml;rung“ aufrufen; Sollten ausschlie&szlig;lich St&ouml;rungsmeldungen 6U/511 vorliegen, folgendes Vorgehen w&auml;hlen:&bull;Pr&uuml;fen, ob eine korrekte Brennstoffversorgung sichergestellt ist; Kapitel „Brennstoffversorgung“ in der Serviceanleitung beachten!&bull;In der Serviceebene der Bedieneinheit den Men&uuml;punkt „Relaistest“ aufrufen, hier die Z&uuml;ndung einschalten und pr&uuml;fen, ob die Z&uuml;ndung korrekt arbeitet&bull;Sollte keine Z&uuml;ndung vorhanden sein, folgende Punkte pr&uuml;fen:&bull;Abstand Z&uuml;ndelektroden&bull;Zustand Z&uuml;ndelektroden&bull;Zustand der Z&uuml;ndleitungen zwischen Z&uuml;ndtrafo und Elektroden&bull;Sind die Stecker richtig auf die Z&uuml;ndelektroden aufgesteckt?&bull;Zustand der Verbindungsleitung zwischen Feuerungsautomat und Z&uuml;ndtrafo pr&uuml;fen&bull;Ist der Stecker des Z&uuml;ndtrafos richtig am Feuerungsautomaten aufgesteckt?&bull;In der Serviceebene der Bedieneinheit den Men&uuml;punkt „Monitor“ aufrufen und hier den Flammenstrom pr&uuml;fen; Sollte dieser schwanken oder st&auml;ndig kleiner als die Sollvorgaben aus der Serviceanleitung sein, folgendes Vorgehen w&auml;hlen:&bull;Flammen&uuml;berwachung auf Verschmutzung pr&uuml;fen&bull;Position der Flammen&uuml;berwachung pr&uuml;fen (bei &Ouml;lkesseln Winkelhalter pr&uuml;fen)&bull;Leitungsverbindung zwischen Feuerungsautomat und Flammenf&uuml;hler pr&uuml;fen&bull;Steckverbindung der Flammen&uuml;berwachung am Feuerungsautomaten pr&uuml;fen&bull;D&uuml;sen pr&uuml;fen&bull;&Ouml;labschlussventil bei &Ouml;lkesseln optisch pr&uuml;fen&bull;Mischsystem bei &Ouml;lkesseln pr&uuml;fen&bull;In der Serviceebene der Bedieneinheit den Men&uuml;punkt „Diagnose/Fehler/blockierende Fehler“ (RC35) bzw. „Diagnose/St&ouml;rungsanzeigen/ aktuelle St&ouml;rungen“ (RC300) pr&uuml;fen ob St&ouml;rmeldung 6L/516 vorliegt; sollte diese vorliegen, dann pr&uuml;fen, ob Magnetventile 1 und/oder 2 von der Zuordnung richtig am Feuerungsautomaten aufgesteckt sind&bull;Sollten andere oder auch keine blockierenden St&ouml;rungen vorliegen, folgendes Vorgehen w&auml;hlen:&bull;Bei raumluftunabh&auml;ngiger Betriebsweise den CO-Gehalt in der Zuluft messen, wenn CO gemessen wird, ist das Abgassystem undicht&bull;Pr&uuml;fen, ob eine korrekte Brennstoffversorgung sichergestellt ist; Kapitel „Brennstoffversorgung“ in der Serviceanleitung beachten!&bull;Magnetventil auf Funktion pr&uuml;fen, ob evtl. defekt",correction => "-&bull;Bei St&ouml;rung der Brennstoffversorgung diese beheben&bull;-&bull;Bei Abweichungen folgende Ma&szlig;nahmen durchf&uuml;hren:&bull;Korrekten Abstand herstellen&bull;Wenn Elektroden verbraucht oder besch&auml;digt sind, diese austauschen&bull;Wenn besch&auml;digt, austauschen&bull;Wenn die Kontakte nicht richtig befestigt sind, St&ouml;rung beheben&bull;Wenn besch&auml;digt, austauschen&bull;Ggf. Kontaktprobleme beheben (z. B. Stecker richtig aufstecken)&bull;-&bull;Bei Verschmutzung des Flammenf&uuml;hlers diesen reinigen&bull;Flammenf&uuml;hler richtig positionieren bzw. den Winkelhalter austauschen&bull;Bei defekten Kabeln diese austauschen&bull;Wenn der Stecker nicht richtig aufgesteckt ist, dann den Stecker wieder richtig aufstecken&bull;D&uuml;sen ggf. austauschen&bull;&Ouml;labschlussventil des &Ouml;lvorw&auml;rmers ggf. austauschen&bull;Mischsystem ggf. reinigen&bull;Bei Abweichungen korrigieren,; bei Vertauschungen diese beheben&bull;-&bull;Abgassystem abdichten, ggf. erneut montieren, Dichtheitspr&uuml;fung durchf&uuml;hren&bull;Bei St&ouml;rungen in der Brennstoffversorgung diese beheben&bull;Bei Defekt das Magnetventil austauschen"};
	$ErrorCodeList{"0226"} = {errorcode => "6L",                                     subcode => "553",                                                class => "4",           reason => "Zu viele Flammenabrisse (Wenn direkt hinter einander 15 Flammen abrisse registriert; werden, wird diese Storungs-; meldung erzeugt. ACHTUNG: Ein „Reset“ ist nur am SAFe moglich!)",                                                                                                                                                           checklist => "In der Serviceebene der Bedieneinheit den Men&uuml;punkt „Diagnose/Fehler“ (RC35) bzw. „Diagnose/ St&ouml;rungsanzeigen/aktuelle St&ouml;rungen“ (RC300) aufrufen „St&ouml;rungsspeicher“; Sollten ausschlie&szlig;lich St&ouml;rungsmeldungen 6U/511 vorliegen, folgendes Vorgehen w&auml;hlen:&bull;Pr&uuml;fen, ob eine korrekte Brennstoffversorgung sichergestellt ist; Kapitel „Brennstoffversorgung“ in der Serviceanleitung beachten!&bull;In der Serviceebene der Bedieneinheit den Men&uuml;punkt „Relaistest“ aufrufen, hier die Z&uuml;ndung einschalten und pr&uuml;fen, ob die Z&uuml;ndung korrekt arbeitet&bull;Sollte keine Z&uuml;ndung vorhanden sein, folgende Punkte pr&uuml;fen:&bull;Abstand Z&uuml;ndelektroden pr&uuml;fen&bull;Zustand Z&uuml;ndelektroden pr&uuml;fen&bull;Zustand der Z&uuml;ndleitungen zwischen Z&uuml;ndtrafo und Elektroden pr&uuml;fen&bull;Sind die Stecker richtig auf die Z&uuml;ndelektroden aufgesteckt?&bull;Zustand der Verbindungsleitung zwischen Feuerungsautomaten und Z&uuml;ndtrafo pr&uuml;fen&bull;Ist der Stecker des Z&uuml;ndtrafos richtig am Feuerungsautomat aufgesteckt?&bull;In der Serviceebene der Bedieneinheit den Men&uuml;punkt „Monitor“ aufrufen und hier den Flammenstrom pr&uuml;fen; sollte dieser schwanken oder st&auml;ndig kleiner als die Sollvorgaben aus der Serviceanleitung sein, folgendes Vorgehen w&auml;hlen:&bull;Flammen&uuml;berwachung auf Verschmutzung pr&uuml;fen&bull;Position der Flammen&uuml;berwachung pr&uuml;fen (Winkelhalter pr&uuml;fen)&bull;Leitungsverbindung zwischen Feuerungs-automat und Flammenf&uuml;hler pr&uuml;fen&bull;Steckverbindung der Flammen&uuml;berwachung am Feuerungsautomaten pr&uuml;fen&bull;D&uuml;sen pr&uuml;fen&bull;&Ouml;labschlussventil bei &Ouml;lkesseln optisch pr&uuml;fen&bull;Bei raumluftunabh&auml;ngiger Betriebsweise den CO-Gehalt in der Zuluft messen, wenn CO gemessen wird, ist das Abgassystem undicht&bull;Mischsystem bei &Ouml;lkesseln pr&uuml;fen&bull;Brennereinstellung pr&uuml;fen&bull;In der Serviceebene der Bedieneinheit den Men&uuml;punkt „Diagnose/Fehler/blockierende Fehler“ (RC35) bzw. „Diagnose/St&ouml;rungsanzeigen/ aktuelle St&ouml;rungen“ (RC300) auf-rufen; sollten St&ouml;rungsmeldungen 6L/516 vorliegen, pr&uuml;fen, ob Magnetventile 1 und/oder 2 von der Zuordnung richtig am Feuerungsautomaten aufgesteckt sind&bull;Sollten andere oder auch keine anderen blockierenden Fehler vorliegen, folgendes Vorgehen w&auml;hlen:&bull;Pr&uuml;fen, ob eine korrekte Brennstoffversorgung sichergestellt ist; entsprechendes Kapitel; „Brennstoffversorgung“ in der Serviceanleitung beachten!&bull;Magnetventil auf Funktion pr&uuml;fen, ob eventuell defekt",correction => "-&bull;Bei St&ouml;rung der Brennstoffversorgung diese beheben&bull;-&bull;Bei Abweichungen folgende Ma&szlig;nahmen durchf&uuml;hren:&bull;Korrekten Abstand herstellen&bull;Wenn Elektroden verbraucht oder besch&auml;digt sind, diese austauschen&bull;Wenn besch&auml;digt, austauschen&bull;Wenn die Kontakte nicht richtig befestigt sind, St&ouml;rung beheben&bull;Wenn besch&auml;digt, austauschen&bull;Ggf. Kontaktprobleme beheben (z. B. Stecker richtig aufstecken )&bull;-&bull;Bei Verschmutzung des Flammenf&uuml;hlers diesen reinigen&bull;Flammenf&uuml;hler richtig positionieren bzw. den Winkelhalter austauschen&bull;Bei defekten Kabeln diese austauschen&bull;Sollte der Stecker nicht richtig aufgesteckt sein, den Stecker wieder richtig aufstecken&bull;D&uuml;sen ggf. austauschen&bull;&Ouml;labschlussventil des &Ouml;lvorw&auml;rmers ggf. austauschen&bull;Abgassystem abdichten, ggf. erneut montieren, Dichtheitspr&uuml;fung durchf&uuml;hren&bull;Mischsystem ggf. reinigen&bull;Bei Abweichungen korrigieren&bull;Bei Vertauschungen diese beheben&bull;-&bull;Bei St&ouml;rung der &Ouml;lversorgung, diese beheben&bull;Bei Defekt austauschen"};
	$ErrorCodeList{"0227"} = {errorcode => "6L",                                     subcode => "555",                                                class => "10",          reason => "Flammenabriss innerhalb Stabilisierung Z&uuml;ndgas",                                                                                                                                                                                                                                                                                              checklist => "Kabelverbindung zur &Uuml;berwachungselektrode pr&uuml;fen&bull;&Uuml;berwachungselektrode auf Verschmutzung pr&uuml;fen&bull;Position der &Uuml;berwachungselektrode pr&uuml;fen",                                                                                                                                                                                                                                                                                                                   correction => "Defekte Kabelverbindung wiederherstellen&bull;Verschmutzung beheben bzw. Elektrode austauschen&bull;&Uuml;berwachungselektrode korrekt positionieren"                                                                                                                                                                                                                                            };
	$ErrorCodeList{"0228"} = {errorcode => "6L",                                     subcode => "557",                                                class => "10",          reason => "Flammenabriss bei Hauptgas ein",                                                                                                                                                                                                                                                                                                                   checklist => "Feuerungsautomat entriegeln Gasanschlussdruck pr&uuml;fen; Feuerungsautomat/Ger&auml;teelektronik SAFe pr&uuml;fen; D&uuml;sendruck laut technischer Dokumente des Kessels pr&uuml;fen",                                                                                                                                                                                                                                                                                                              correction => "Feuerungsautomat entriegeln; Bei Abweichungen ggf. mit Gasversorger in Verbindung setzen; Feuerungsautomat/Ger&auml;teelektronik SAFe auswechseln; D&uuml;sendruck ggf. richtig einstellen"                                                                                                                                                                                                      };
	$ErrorCodeList{"0229"} = {errorcode => "6L",                                     subcode => "561",                                                class => "4",           reason => "Spannungsunterbrechung w&auml;hrend Brennerstart (Wenn der Brennerautomat 5-mal hintereinander wahrend des ersten Brenner anlaufs direkt nach dem; „Power Up“ ausgeschaltet wurde, wird diese Meldung erzeugt. Tritt; die Meldung mehr als 5-mal nach -; einander auf, wird der Fehler ein; verriegelnder Fehler.)",                           checklist => "-&bull;Die 230-V-Spannungsversorgung zum Regelger&auml;t bzw. Feuerungsautomat pr&uuml;fen; Evtl. liegt ein Wackelkontakt vor und der Feuerungsautomat bekommt st&auml;ndig eine Spannungsunterbrechung&bull;Eventuell liegen St&ouml;rungen im Rahmen von EMV vor; daher das Umfeld der Anlage auf m&ouml;gliche St&ouml;rquellen pr&uuml;fen&bull;Sollten keine der genannten Probleme vorliegen, ist vermutlich der Feuerungsautomat defekt",                                                      correction => "Feuerungsautomat entriegeln&bull;Problem in der Spannungsversorgung beheben&bull;Ursachen f&uuml;r EMV-St&ouml;rungen beheben&bull;Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                          };
	$ErrorCodeList{"0230"} = {errorcode => "6L",                                     subcode => "587",                                                class => "10",          reason => "Flammenabriss Stabilisierung Teillast",                                                                                                                                                                                                                                                                                                            checklist => "Gasanschlussdruck messen&bull;Gasarmatur auf Funktion pr&uuml;fen&bull;Ionisationsstrom in der Serviceebene der Bedieneinheit pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                           correction => "F&uuml;r korrekten Gasanschlussdruck sorgen&bull;Bei Defekt Gasarmatur austauschen&bull;Bei Abweichungen Z&uuml;nd-/&Uuml;berwachungselektrode austauschen"                                                                                                                                                                                                                                      };
	$ErrorCodeList{"0231"} = {errorcode => "6P",                                     subcode => "269",                                                class => "4",           reason => "Flammen&uuml;berwachung",                                                                                                                                                                                                                                                                                                                          checklist => "BCM/Kesselidentifikationsmodul defekt",                                                                                                                                                                                                                                                                                                                                                                                                                                                               correction => "BCM bzw. Kesselidentifikationsmodul durch Buderus Service austauschen lassen"                                                                                                                                                                                                                                                                                                                    };
	$ErrorCodeList{"0232"} = {errorcode => "6P",                                     subcode => "363",                                                class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik / Basiscontroller",                                                                                                                                                                                                                                                                                       checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Steuerger&auml;t/ Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0233"} = {errorcode => "6U",                                     subcode => "511",                                                class => "10",          reason => "Keine Flamme innerhalb Sicherheitszeit",                                                                                                                                                                                                                                                                                                           checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Keine Ma&szlig;nahme, Ger&auml;teelektronik SAFe versucht Wiederanlauf&bull;Ggf. entriegeln/reseten&bull;Hinweis: siehe auch St&ouml;rung 6L/548 und 6L/553"                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0234"} = {errorcode => "6Y",                                     subcode => "510",                                                class => "4",           reason => "Fremdlicht Vorbel&uuml;ftung",                                                                                                                                                                                                                                                                                                                     checklist => "Position des Flammenf&uuml;hlers pr&uuml;fen, ob evtl. Fremdlicht einf&auml;llt&bull;Mit abgezogenem und mit der Hand abgedunkeltem Flammenf&uuml;hler Startversuch durchf&uuml;hren und pr&uuml;fen, ob St&ouml;rungsmeldung 6Y/510 angezeigt wird&bull;Das Magnetventil ist undicht, wodurch die Flamme nicht verlischt (im Sichtfenster am Brenner optisch pr&uuml;fen)&bull;Mit abgezogenem und mit der Hand abgedunkeltem Flammenf&uuml;hler Startversuch durchf&uuml;hren und pr&uuml;fen, ob St&ouml;rungsmeldung 6U/511 angezeigt wird; (Die Ger&auml;teelektronik SAFe versucht einen Wiederanlauf) Sollte dies der Fall sein, erkennt der Flammenf&uuml;hler Fremdlicht&bull;Wackelkontakt Flammenf&uuml;hler oder Stecker Ger&auml;teelektronik SAFe",correction => "Wenn erforderlich, F&uuml;hler richtig positionieren&bull;Der Flammenf&uuml;hler ist defekt, wenn immer noch die St&ouml;rungsmeldung 6Y/510 erscheint Der Flammenf&uuml;hler muss ausgetauscht werden&bull;Magnetventil austauschen&bull;Evtl. ist das Z&uuml;ndelement nicht richtig positioniert und sollte richtig eingebaut werden&bull;Flammenf&uuml;hler oder Ger&auml;teelektronik SAFe austauschen"};
	$ErrorCodeList{"0235"} = {errorcode => "6Y",                                     subcode => "2039",                                               class => "4",           reason => "Fremdlicht im Feuerraum w&auml;hrend Vorbel&uuml;ftung erkannt; Flamme wurde zu einem unzul&auml;ssigen Zeitpunkt erkannt",                                                                                                                                                                                                                        checklist => "Magnetventil undicht&bull;Einspritzventil undicht&bull;Ionisationsstromsensor besch&auml;digt&bull;Feuerungsautomat besch&auml;digt",                                                                                                                                                                                                                                                                                                                                                                 correction => "Magnetventil auf Dichtheit pr&uuml;fen, ggf. austauschen&bull;Einspritzventil auf Dichtheit pr&uuml;fen, ggf. austauschen&bull;&Uuml;berwachungselektrode pr&uuml;fen, ggf. austauschen&bull;Feuerungsautomat pr&uuml;fen, ggf. austauschen"                                                                                                                                                     };
	$ErrorCodeList{"0236"} = {errorcode => "7A",                                     subcode => "550",                                                class => "10",          reason => "Netzspannung zu niedrig",                                                                                                                                                                                                                                                                                                                          checklist => "Netzspannung pr&uuml;fen. Die Spannung darf; 187 V AC nicht unterschreiten (GB402 195 V AC)",                                                                                                                                                                                                                                                                                                                                                                                                         correction => "F&uuml;r eine korrekte Spannungsversorgung sorgen"                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0237"} = {errorcode => "7A",                                     subcode => "551",                                                class => "10",          reason => "Netzspannung kurzzeitig unterbrochen",                                                                                                                                                                                                                                                                                                             checklist => "Netzzuleitung auf evtl. Wackelkontakte pr&uuml;fen&bull;Verdrahtung und richtige Kontaktierung des Netzsteckers an Regelger&auml;t bzw. Ger&auml;teelektronik SAFe oder Fremdbrennermodul pr&uuml;fen",                                                                                                                                                                                                                                                                                               correction => "Ggf. Kontaktprobleme beheben&bull;Ggf. Kontaktprobleme beheben"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0238"} = {errorcode => "7C",                                     subcode => "231",                                                class => "4",           reason => "Netzspannungsunterbrechung nach St&ouml;rungsanzeige",                                                                                                                                                                                                                                                                                             checklist => "Die Netzspannung wurde nach einer St&ouml;rungsmeldung unterbrochen und wieder eingeschaltet.",                                                                                                                                                                                                                                                                                                                                                                                                       correction => "Taste „Reset“ bet&auml;tigen"                                                                                                                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"0239"} = {errorcode => "7H",                                     subcode => "328",                                                class => "4",           reason => "Netzspannung kurzzeitig unterbrochen; (Der UBA &uuml;berpr&uuml;ft die eingehende Netzspannung.; Wird ein Spannungsabfall registriert, geht der Kessel auf St&ouml;rung.)",                                                                                                                                                                        checklist => "Netzspannung pr&uuml;fen Lag ein Stromausfall vor?",                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Eventuell Trafo pr&uuml;fen, ggf. austauschen"                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"0240"} = {errorcode => "7H",                                     subcode => "356",                                                class => "4",           reason => "Versorgungsspannung f&uuml;r den W&auml;rmeerzeuger ist zu niedrig",                                                                                                                                                                                                                                                                               checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Versorgungsspannung von mindestens 196 VAC herstellen"                                                                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"0241"} = {errorcode => "7L",                                     subcode => "261",                                                class => "4",           reason => "Zeitfehler bei erster Sicherheitszeit",                                                                                                                                                                                                                                                                                                            checklist => "Kontakt zwischen Ger&auml;teelektronik UBA u. Montagefu&szlig; sowie Basiscontroller BC25 und der Grundplatte und allen sonstigen Steckverbindungen pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                     correction => "Kontaktproblem beseitigen, ggf. Ger&auml;teelektronik UBA bzw. Ger&auml;teelektronik BC25 austauschen"                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"0242"} = {errorcode => "7L",                                     subcode => "280",                                                class => "4",           reason => "Zeitfehler bei Wiederanlaufversuch",                                                                                                                                                                                                                                                                                                               checklist => "Kontakt zwischen Ger&auml;teelektronik UBA u. Montagefu&szlig; sowie Basiscontroller BC25 und der Grundplatte und allen sonstigen Steckverbindungen pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                     correction => "Kontaktproblem beseitigen, ggf. Ger&auml;teelektronik UBA bzw. Ger&auml;teelektronik BC25 austauschen"                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"0243"} = {errorcode => "7P",                                     subcode => "549",                                                class => "4",           reason => "Sicherheitskette offen (SI 17-18)",                                                                                                                                                                                                                                                                                                                checklist => "Den Wasserdruck der Anlage pr&uuml;fen. Der Wasserdruck darf 0,8 bar nicht unterschreiten&bull;Pr&uuml;fen, ob der Steckverbinder des Druckw&auml;chters richtig aufgesteckt ist&bull;Eingangsspannung Druckw&auml;chter pr&uuml;fen&bull;Ausgangsspannung Druckw&auml;chter pr&uuml;fen&bull;Pr&uuml;fen, ob der Gasdruckw&auml;chter geschaltet hat oder einen Verdrahtungsfehler vorliegt&bull;Angeschlossene Sicherheitsger&auml;te auf Ausl&ouml;sung pr&uuml;fen (Sicherheitskette SI17-18)&bull;Regelger&auml;t defekt",correction => "Wasser bis mindestens 1 bar Wasserdruck nachf&uuml;llen&bull;-&bull;Gasdruckw&auml;chter und Gasdruck pr&uuml;fen&bull;Angeschlossene Sicherheitsger&auml;te (z.B. AbgasSTB)  entriegeln, St&ouml;rungsursache beseitigen&bull;Regelger&auml;t austauschen"                                                                                                                                      };
	$ErrorCodeList{"0244"} = {errorcode => "7U",                                     subcode => "2052",                                               class => "10",          reason => "Maximale Einschaltdauer Z&uuml;ndtrafo &uuml;berschritten",                                                                                                                                                                                                                                                                                        checklist => "Zu viele Wiederanl&auml;ufe des Brenners haben die Einschaltdauer des Z&uuml;ndtrafos &uuml;berschritten&bull;Fehlerhafte &Ouml;lversorgung&bull;Fehlerhafte Brennerkomponenten&bull;Feuerungsautomat besch&auml;digt&bull;Hinweis: siehe auch Beschreibung St&ouml;rungscode 6L/548",                                                                                                                                                                                                                correction => "Fehler in der &Ouml;lversorgung pr&uuml;fen, ggf. beseitigen&bull;-&bull;Brennerkomponenten pr&uuml;fen, ggf. austauschen&bull;Feuerungsautomat pr&uuml;fen, ggf. austauschen&bull;-"                                                                                                                                                                                                            };
	$ErrorCodeList{"0245"} = {errorcode => "8L",                                     subcode => "534",                                                class => "10",          reason => "Kein Gasdruck oder zus&auml;tzlicher Abgasdruckbegrenzer abgeschaltet; (Obwohl das Magnetventil 1 ge&ouml;ffnet haben m&uuml;sste, steht kein Gasdruck an.; Brenner macht nach-einander 3 Startversuche, dann wartet er eine Stunde, um dann erneut 3 Startversuche durchzu-; f&uuml;hren.)",                                                      checklist => "Pr&uuml;fen, ob der Gashahn ge&ouml;ffnet ist&bull;Pr&uuml;fen, ob Gasdruck vorhanden ist&bull;Pr&uuml;fen, ob Abgasdruckbegrenzer geschaltet hat&bull;Elektrode pr&uuml;fen&bull;Z&uuml;ndtrafo pr&uuml;fen&bull;Gasfilter auf Verschmutzung pr&uuml;fen",                                                                                                                                                                                                                                           correction => "Ggf. Gasarmatur austauschen&bull;Gasdruck messen&bull;Abgasdruckbegrenzer entriegeln Abgasweg auf Verstopfung pr&uuml;fen&bull;Ggf. Elektrode austauschen&bull;Ggf. Z&uuml;ndtrafo austauschen&bull;Ggf. Gasfilter austauschen"                                                                                                                                                                  };
	$ErrorCodeList{"0246"} = {errorcode => "8L",                                     subcode => "579",                                                class => "4",           reason => "Kein Gasdruck; (Obwohl das Magnetventil 1 ge&ouml;ffnet haben m&uuml;sste, steht kein Gasdruck an. Der Brenner macht hinter einander 3 Startversuche, dann; wartet er eine Stunde, um erneut; 3 Startversuche durchzuf&uuml;hren.)",                                                                                                               checklist => "Pr&uuml;fen, ob der Gashahn ge&ouml;ffnet ist",                                                                                                                                                                                                                                                                                                                                                                                                                                                       correction => "Gasmagnetventil 1 austauschen"                                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"0247"} = {errorcode => "8U",                                     subcode => "364",                                                class => "L",           reason => "Magnetventil EV2 undicht",                                                                                                                                                                                                                                                                                                                         checklist => "Gasarmatur auf Verschmutzung pr&uuml;fen&bull;Gasfilter vorhanden?",                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Gasarmatur austauschen&bull;-"                                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"0248"} = {errorcode => "8U",                                     subcode => "365",                                                class => "L",           reason => "Magnetventil EV1 undicht",                                                                                                                                                                                                                                                                                                                         checklist => "Gasarmatur auf Verschmutzung pr&uuml;fen&bull;Gasfilter vorhanden?",                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Gasarmatur austauschen&bull;-"                                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"0249"} = {errorcode => "8U",                                     subcode => "581",                                                class => "4",           reason => "Magnetventil 2 undicht",                                                                                                                                                                                                                                                                                                                           checklist => "Gasarmatur auf Verschmutzung pr&uuml;fen Gasfilter vorhanden?",                                                                                                                                                                                                                                                                                                                                                                                                                                       correction => "Gasarmatur austauschen"                                                                                                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"0250"} = {errorcode => "8U",                                     subcode => "584",                                                class => "10",          reason => "Umschaltmodul keine R&uuml;ckmeldung",                                                                                                                                                                                                                                                                                                             checklist => "Externe Komponenten m&uuml;ssen &uuml;ber die Anschlussklemme 7 eine R&uuml;ckmeldung (230 V AC) geben; diese R&uuml;ckmeldung fehlt;; eventuell ist ein Draht lose, ein Kabel besch&auml;digt oder die externe Komponente defekt",                                                                                                                                                                                                                                                                   correction => "Defektes Kabel austauschen bzw. Steckverbindung richtig herstellen oder auch defekte externe Komponente austauschen"                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"0251"} = {errorcode => "8U",                                     subcode => "591",                                                class => "4",           reason => "Abgassperrklappe &ouml;ffnet nicht",                                                                                                                                                                                                                                                                                                               checklist => "Die Verdrahtung zwischen Universalschaltmodul und Abgassperrklappe pr&uuml;fen&bull;-&bull;Abgassperrklappenmotor defekt&bull;Universalschaltmodul defekt",                                                                                                                                                                                                                                                                                                                                           correction => "Bei Fehlverdrahtung den Verdrahtungsfehler korrigieren&bull;Bei Defekt Kabel ersetzen&bull;Abgassperrklappenmotor austauschen&bull;Modul austauschen"                                                                                                                                                                                                                                            };
	$ErrorCodeList{"0252"} = {errorcode => "8U",                                     subcode => "592",                                                class => "4",           reason => "Abgassperrklappe dauerhaft ge&ouml;ffnet",                                                                                                                                                                                                                                                                                                         checklist => "Die Verdrahtung zwischen Universalschaltmodul und Abgassperrklappe pr&uuml;fen&bull;Verbindungsleitung zwischen Universalschaltmodul und Abgassperrklappe pr&uuml;fen&bull;Abgassperrklappenmotor defekt&bull;Universalschaltmodul defekt",                                                                                                                                                                                                                                                           correction => "Bei Fehlverdrahtung den Verdrahtungsfehler korrigieren&bull;Bei Defekt Kabel ersetzen&bull;Abgassperrklappenmotor austauschen&bull;Modul austauschen"                                                                                                                                                                                                                                            };
	$ErrorCodeList{"0253"} = {errorcode => "8U",                                     subcode => "593",                                                class => "4",           reason => "Br&uuml;cke am Eingang K&uuml;chenl&uuml;fter (Dunstabzugshaube) fehlt",                                                                                                                                                                                                                                                                           checklist => "K&uuml;chenl&uuml;fter &uuml;berpr&uuml;fen; Verbindung K&uuml;chenl&uuml;fter zu Modul LM10 &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                   correction => "K&uuml;chenl&uuml;fter abschalten, Installation korrigieren, Br&uuml;cke einsetzen"                                                                                                                                                                                                                                                                                                              };
	$ErrorCodeList{"0254"} = {errorcode => "8Y",                                     subcode => "232",                                                class => "20",          reason => "W&auml;rmeerzeuger durch externen Schaltkontakt verriegelt (EV bzw. I3)",                                                                                                                                                                                                                                                                          checklist => "Pr&uuml;fen, ob ein externer Schaltkontakt, z. B. Fu&szlig;bodenthermostat oder Festbrennstoff-Kessel, ge&ouml;ffnet hat&bull;Kabelbaum zwischen Ger&auml;teelektronik UBA und der Klemmleiste pr&uuml;fen",                                                                                                                                                                                                                                                                                          correction => "Kein Handlungsbedarf, wenn der Zustand korrekt und gew&uuml;nscht ist&bull;Ggf. Kabelbaum oder entsprechenden Teil davon ersetzen"                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0255"} = {errorcode => "8Y",                                     subcode => "572",                                                class => "10",          reason => "Externe Sperrung durch Klemme EV",                                                                                                                                                                                                                                                                                                                 checklist => "Pr&uuml;fen, ob evtl. ein Kabel von den Anschlussklemmen EV defekt ist; Ansonsten liegt keine St&ouml;rung vor, da die Verriegelung eine gewollte Funktion ist",                                                                                                                                                                                                                                                                                                                                      correction => "Bei defektem Kabel oder losen Dr&auml;hten den Defekt beheben"                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"0256"} = {errorcode => "8Y",                                     subcode => "583",                                                class => "10",          reason => "Umschaltmodul externe Verriegelung",                                                                                                                                                                                                                                                                                                               checklist => "Pr&uuml;fen, ob evtl. ein Kabel von den Anschlussklemmen EV defekt ist; Ansonsten liegt keine St&ouml;rung vor, da die Verriegelung eine gewollte Funktion ist",                                                                                                                                                                                                                                                                                                                                      correction => "Bei defektem Kabel oder losen Dr&auml;hten den Defekt beheben"                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"0257"} = {errorcode => "8Y",                                     subcode => "589",                                                class => "10",          reason => "W&auml;rmeerzeuger durch externen Schaltkontakt verriegelt (BRM10)",                                                                                                                                                                                                                                                                               checklist => "Pr&uuml;fen, ob evtl. ein Kabel des BRM10 von den Anschlussklemmen 15/16 defekt ist; Ansonsten liegt keine St&ouml;rung vor, da die Verriegelung eine gewollte Funktion ist",                                                                                                                                                                                                                                                                                                                         correction => "Bei defektem Kabel oder losen Dr&auml;hten den Defekt beheben"                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"0258"} = {errorcode => "8Y",                                     subcode => "590",                                                class => "4",           reason => "Druckschalter w&auml;hrend Betrieb ge&ouml;ffnet",                                                                                                                                                                                                                                                                                                 checklist => "Unterbrechung Zuluftdruckschalter im Betrieb&bull;Unterbrechung Abgasdruckschalter im Betrieb&bull;Unterbrechung am Gaseingangsdruckschalter&bull;Unterbrechung an Ger&auml;teelektronik SAFe Druckschalter&bull;Unterbrechung am Gasausgangsdruckschalter",                                                                                                                                                                                                                                          correction => "Zuluftweg auf freien Durchgang pr&uuml;fen&bull;Abgasweg auf freien Durchgang pr&uuml;fen&bull;Einstellung des Druckschalters pr&uuml;fen (Erdgas Soll= 3 inch wc, FG Soll = 7 inch wc), Gas-Versorgungsdruck > 3,5 inch wc auch bei 100 %-Modulation sicherstellen&bull;Gasarmatur austauschen (Nulldruckregler defekt)&bull;Gasarmatur austauschen (Nulldruckregler defekt)"                   };
	$ErrorCodeList{"0259"} = {errorcode => "8Y",                                     subcode => "2514",                                               class => "BC",          reason => "W&auml;rmeanforderung blockiert wegen UM10",                                                                                                                                                                                                                                                                                                       checklist => "W&auml;rmeanforderung blockiert durch Umschaltmodul",                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0260"} = {errorcode => "92",                                     subcode => "84",                                                 class => "B",           reason => "R&uuml;cklauftemperaturf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                        checklist => "R&uuml;cklauftemperaturf&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                                      correction => "F&uuml;hler wieder richtig anschlie&szlig;en, ggf. austauschen"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0261"} = {errorcode => "92",                                     subcode => "85",                                                 class => "B",           reason => "R&uuml;cklauftemperaturf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                        checklist => "R&uuml;cklauftemperaturf&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                                      correction => "F&uuml;hler wieder richtig anschlie&szlig;en, ggf. austauschen"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0262"} = {errorcode => "93",                                     subcode => "86",                                                 class => "B",           reason => "Vorlauftemperaturf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                              checklist => "Vorlauftemperaturf&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "F&uuml;hler wieder richtig anschlie&szlig;en, ggf. austauschen"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0263"} = {errorcode => "93",                                     subcode => "87",                                                 class => "B",           reason => "Vorlauftemperaturf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                              checklist => "Vorlauftemperaturf&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "F&uuml;hler wieder richtig anschlie&szlig;en, ggf. austauschen"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0264"} = {errorcode => "94",                                     subcode => "88",                                                 class => "B",           reason => "H&ouml;here R&uuml;cklauftemperatur als Vorlauftemperatur",                                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Vorund R&uuml;cklauf vertauscht sind&bull;DIP-Schalter-Stellung auf der Leiterplatte auf dem W&auml;rmepumpenmodul pr&uuml;fen (Normalbetrieb)&bull;Die F&uuml;hler und die Position der F&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                                                    correction => "Vorund R&uuml;cklauf richtig anschlie&szlig;en&bull;DIP-Schalter auf Normalbetrieb stellen&bull;Die F&uuml;hler richtig positionieren"                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"0265"} = {errorcode => "95",                                     subcode => "90",                                                 class => "B",           reason => "Au&szlig;entemperaturf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                          checklist => "F&uuml;hler anhand der Spannungsund Widerstandswerte pr&uuml;fen&bull;F&uuml;hleranschluss pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                              correction => "Bei Defekt F&uuml;hler austauschen&bull;F&uuml;hler wieder richtig anschlie&szlig;en"                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"0266"} = {errorcode => "96",                                     subcode => "89",                                                 class => "10",          reason => "Vorlaufoder R&uuml;cklauftemperatur au&szlig;erhalb zul&auml;ssigen Bereichs",                                                                                                                                                                                                                                                                     checklist => "Vorlaufoder R&uuml;cklauftemperatur am HybridManager au&szlig;erhalb des zul&auml;ssigen Bereichs Der W&auml;rmeerzeuger erw&auml;rmt zun&auml;chst das Wasser, bevor die Luft-Wasser-W&auml;rmepumpe gestartet wird",                                                                                                                                                                                                                                                                                correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0267"} = {errorcode => "97",                                     subcode => "91",                                                 class => "4",           reason => "St&ouml;rung Au&szlig;eneinheit/ Wasserdurchfluss",                                                                                                                                                                                                                                                                                                checklist => "St&ouml;rung an der W&auml;rmepumpe/Wasserdurchfluss gest&ouml;rt&bull;LED-Signal am Hybridmodul pr&uuml;fen&bull;Durchfluss der Luft-Wasser-W&auml;rmepumpe pr&uuml;fen, Filter pr&uuml;fen&bull;Str&ouml;mungsschalter pr&uuml;fen&bull;System pr&uuml;fen",                                                                                                                                                                                                                                        correction => "-&bull;Ggf. Filter reinigen bzw. ersetzen&bull;Ggf. ersetzen&bull;-"                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"0268"} = {errorcode => "9A",                                     subcode => "235",                                                class => "8",           reason => "Versionskonflikt Ger&auml;teelektronik/Kesselidentifikationsmodul",                                                                                                                                                                                                                                                                                checklist => "Softwareversion im Kessel: IdentifikationsModul/ Kodierstecker ist nicht kompatibel zum Steuerger&auml;t/Feuerungsautomat",                                                                                                                                                                                                                                                                                                                                                                           correction => "Kesselidentifikationsmodul/Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                                                         };
	$ErrorCodeList{"0269"} = {errorcode => "9A",                                     subcode => "361",                                                class => "8",           reason => "Kesselidentifikationsmodul oder Ger&auml;teelektronik St&ouml;rung",                                                                                                                                                                                                                                                                               checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Kessel Idendifikations Modul/ Kodierstecker korrekt aufstecken bzw. austauschen"                                                                                                                                                                                                                                                                                                                 };
	$ErrorCodeList{"0270"} = {errorcode => "9H",                                     subcode => "237",                                                class => "4",           reason => "Systemst&ouml;rung",                                                                                                                                                                                                                                                                                                                               checklist => "Kesselidentifikationsmodul KIM oder Ger&auml;teelektronik UBA defekt oder Kurzschluss im Anschlusskabel der Gasarmatur",                                                                                                                                                                                                                                                                                                                                                                              correction => "Anschlusskabel der Gasarmatur austauschen, Ger&auml;teelektronik UBA austauschen oder Kesselidentifikationsmodul durch Buderus Service austauschen lassen"                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0271"} = {errorcode => "9H",                                     subcode => "267",                                                class => "4",           reason => "Systemst&ouml;rung",                                                                                                                                                                                                                                                                                                                               checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Ger&auml;teelektronik UBA austauschen"                                                                                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"0272"} = {errorcode => "9H",                                     subcode => "272",                                                class => "4",           reason => "Systemst&ouml;rung",                                                                                                                                                                                                                                                                                                                               checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Ger&auml;teelektronik UBA austauschen (Ger&auml;teelektronik mit aktuellerem Softwarestand einsetzen)"                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"0273"} = {errorcode => "9L",                                     subcode => "230",                                                class => "10",          reason => "St&ouml;rung Regelventil",                                                                                                                                                                                                                                                                                                                         checklist => "Spule des Regelventils unterbrochen, Kabel zum Regelventil unterbrochen, Regelventilansteuerung defekt",                                                                                                                                                                                                                                                                                                                                                                                              correction => "Bei defektem Kabel oder losen Dr&auml;hten den Defekt beheben"                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"0274"} = {errorcode => "9L",                                     subcode => "234",                                                class => "4",           reason => "Elektrische St&ouml;rung Gasarmatur",                                                                                                                                                                                                                                                                                                              checklist => "Anschlusskabel der Gasarmatur pr&uuml;fen&bull;Gasarmatur defekt",                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "Anschlusskabel austauschen und Reset nach dem Austausch&bull;Gasarmatur austauschen und Reset nach dem Austausch"                                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"0275"} = {errorcode => "9L",                                     subcode => "238",                                                class => "8",           reason => "Ger&auml;teelektronik ist defekt",                                                                                                                                                                                                                                                                                                                 checklist => "Relais f&uuml;r die Ansteuerung der Gasarmatur defekt",                                                                                                                                                                                                                                                                                                                                                                                                                                               correction => "Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                                                                                    };
	$ErrorCodeList{"0276"} = {errorcode => "9P",                                     subcode => "239",                                                class => "4",           reason => "Systemst&ouml;rung",                                                                                                                                                                                                                                                                                                                               checklist => "Ger&auml;teelektronik UBA defekt oder Kurzschluss im Anschlusskabel der Gasarmatur&bull;Kesselidentifikationsmodul nicht richtig aufgesteckt oder defekt",                                                                                                                                                                                                                                                                                                                                            correction => "Anschlusskabel der Gasarmatur austauschen, ggf. Ger&auml;teelektronik UBA austauschen&bull;Kesselidentifikationsmodul richtig aufstecken oder durch Buderus Service austauschen lassen"                                                                                                                                                                                                          };
	$ErrorCodeList{"0277"} = {errorcode => "9U",                                     subcode => "230",                                                class => "8",           reason => "Interne St&ouml;rung Ger&auml;teelektronik, Modulationsspule St&ouml;rung",                                                                                                                                                                                                                                                                        checklist => "Modulationsspule defekt, oder Dr&auml;hte an der Spule lose",                                                                                                                                                                                                                                                                                                                                                                                                                                         correction => "Ionisation pr&uuml;fen (au&szlig;erhalb Grenzwert), Steuerger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"0278"} = {errorcode => "9U",                                     subcode => "233",                                                class => "4",           reason => "St&ouml;rung Kesselidentifikationsmodul oder Ger&auml;teelektronik (KIM oder UBA defekt)",                                                                                                                                                                                                                                                         checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Ger&auml;teelektronik UBA austauschen oder Kesselidentifikationsmodul durch Buderus Service austauschen lassen; KesselidentifikationsmodulSteckverbindung wiederherstellen; Ggf. Kesselidentifikationsmodul, als letzte Ma&szlig;nahme Kabelbaum oder Ger&auml;teelektronik UBA austauschen"                                                                                                     };
	$ErrorCodeList{"0279"} = {errorcode => "9Y",                                     subcode => "500",                                                class => "8",           reason => "Interne St&ouml;rung Fremdbrennermodul, keine Spannung Sicherheitsrelais",                                                                                                                                                                                                                                                                         checklist => "„Reset“-Taste bet&auml;tigen und warten, ob die St&ouml;rung behoben ist",                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Bleibt die St&ouml;rung nach Reset erhalten, ist der Feuerungsautomat oder Fremdbrennermodul defekt und muss ausgetauscht werden"                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"0280"} = {errorcode => "9Y",                                     subcode => "501",                                                class => "8",           reason => "Interne St&ouml;rung Fremdbrennermodul, Sicherheitsrelais h&auml;ngt",                                                                                                                                                                                                                                                                             checklist => "„Reset“-Taste bet&auml;tigen und warten, ob die St&ouml;rung behoben ist",                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Bleibt die St&ouml;rung nach Reset erhalten, ist der Feuerungsautomat oder Fremdbrennermodul defekt und muss ausgetauscht werden"                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"0281"} = {errorcode => "9Y",                                     subcode => "502",                                                class => "4",           reason => "Interne St&ouml;rung Ger&auml;teelektronik SAFe/Fremdbrennermodul, keine Spannung Brennstoffrelais 1",                                                                                                                                                                                                                                             checklist => "„Reset“-Taste bet&auml;tigen und warten, ob die St&ouml;rung behoben ist",                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Bleibt die St&ouml;rung nach Reset erhalten, ist der Feuerungsautomat oder das Fremdbrennermodul defekt und muss ausgetauscht werden"                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"0282"} = {errorcode => "9Y",                                     subcode => "503",                                                class => "4",           reason => "Interne St&ouml;rung Ger&auml;teelektronik SAFe/Fremdbrennermodul, Brennstoffrelais 1 h&auml;ngt",                                                                                                                                                                                                                                                 checklist => "„Reset“-Taste bet&auml;tigen und warten, ob die St&ouml;rung behoben ist",                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Bleibt die St&ouml;rung nach Reset erhalten, ist der Feuerungsautomat oder das Fremdbrennermodul defekt und muss ausgetauscht werden"                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"0283"} = {errorcode => "9Y",                                     subcode => "2000",                                               class => "4",           reason => "Interne St&ouml;rung Feuerungsautomat, keine Spannung Z&uuml;ndtraforelais; hinter dem Z&uuml;ndtraforelais wird keine Spannung gemessen, obwohl das Relais eingeschaltet ist",                                                                                                                                                                    checklist => "St&ouml;rung im Feuerungsautomaten&bull;-",                                                                                                                                                                                                                                                                                                                                                                                                                                                           correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"0284"} = {errorcode => "9Y",                                     subcode => "2001",                                               class => "4",           reason => "Interne St&ouml;rung Feuerungsautomat, Z&uuml;ndtraforelais h&auml;ngt; hinter dem Z&uuml;ndtraforelais wird eine Spannung gemessen, obwohl das Relais abgeschaltet ist",                                                                                                                                                                          checklist => "St&ouml;rung im Feuerungsautomaten&bull;-",                                                                                                                                                                                                                                                                                                                                                                                                                                                           correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"0285"} = {errorcode => "9Y",                                     subcode => "2002",                                               class => "4",           reason => "Interne St&ouml;rung Feuerungsautomat, keine Spannung Heizpatronenrelais; hinter dem Heizpatronenrelais wird keine Spannung gemessen, obwohl das Relais eingeschaltet ist und hinter dem Sicherheitsrelais eine Spannung gemessen wird",                                                                                                           checklist => "St&ouml;rung im Feuerungsautomaten&bull;-",                                                                                                                                                                                                                                                                                                                                                                                                                                                           correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"0286"} = {errorcode => "9Y",                                     subcode => "2003",                                               class => "4",           reason => "Interne St&ouml;rung Feuerungsautomat, Heizpatronenrelais h&auml;ngt; hinter dem Heizpatronenrelais wird eine Spannung gemessen, obwohl das Relais abgeschaltet ist",                                                                                                                                                                              checklist => "St&ouml;rung im Feuerungsautomaten&bull;-",                                                                                                                                                                                                                                                                                                                                                                                                                                                           correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen-A"                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"0287"} = {errorcode => "A01",                                    subcode => "594",                                                class => "20",          reason => "Temperaturf&uuml;hler anstelle Kodierbr&uuml;cken angeschlossen",                                                                                                                                                                                                                                                                                  checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Verdrahtungsfehler im Bereich Kodierbr&uuml;cke beheben"                                                                                                                                                                                                                                                                                                                                         };
	$ErrorCodeList{"0288"} = {errorcode => "A01",                                    subcode => "800",                                                class => "14",          reason => "Au&szlig;entemperaturf&uuml;hler defekt; (Auswirkung: es wird die minimale Au&szlig;entemperatur angenommen)",                                                                                                                                                                                                                                     checklist => "Konfiguration pr&uuml;fen; Mit der gew&auml;hlten Einstellung ist ein Au&szlig;entemperaturf&uuml;hler erforderlich&bull;Verbindungsleitung zwischen Regelger&auml;t; und Au&szlig;entemperaturf&uuml;hler auf Durchgang pr&uuml;fen&bull;Elektrischen Anschluss der Verbindungsleitung im Au&szlig;entemperaturf&uuml;hler bzw. am Stecker im Regelger&auml;t pr&uuml;fen&bull;Au&szlig;entemperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Au&szlig;entemperaturf&uuml;hlers im Regelger&auml;t laut Tabelle pr&uuml;fen",correction => "Konfiguration &auml;ndern&bull;Wenn kein Durchgang vorhanden ist, die St&ouml;rung beheben&bull;Falls Schrauben oder ein Stecker lose sind, das Kontaktproblem beheben&bull;Wenn Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Regelger&auml;t austauschen"       };
	$ErrorCodeList{"0289"} = {errorcode => "A01",                                    subcode => "808",                                                class => "14",          reason => "Warmwasserbereitung: Warmwasser-Temperaturf&uuml;hler 1 defekt (Auswirkung: es wird kein Warmwasser geladen) Falls keine Warmwasserfunktion gew&uuml;nscht; ist, diese in der Bedieneinheit; deaktivieren",                                                                                                                                        checklist => "Verbindungsleitung zwischen Regelger&auml;t und Warmwasser-Temperaturf&uuml;hler pr&uuml;fen&bull;Elektrischen Anschluss der Verbindungsleitung im Regelger&auml;t pr&uuml;fen&bull;Warmwasser-Temperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Warmwasser-Temperaturf&uuml;hlers im Regelger&auml;t laut Tabelle pr&uuml;fen&bull;Beim GB142 sollte ein Reset durchgef&uuml;hrt werden",                                                                     correction => "Wenn ein Defekt vorliegt, F&uuml;hler austauschen&bull;Falls Schrauben oder ein Stecker lose sind, das Kontaktproblem beheben&bull;Wenn Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Regelger&auml;t austauschen&bull;Wenn nach dem Reset die St&ouml;rung noch besteht, dann ist die Ger&auml;teelektronik UBA; im W&auml;rmeerzeuger auszutauschen"};
	$ErrorCodeList{"0290"} = {errorcode => "A01",                                    subcode => "809",                                                class => "14",          reason => "Warmwasserbereitung: Warmwasser-Temperaturf&uuml;hler 2 defekt. Falls keine Warmwasserfunktion gew&uuml;nscht ist, diese in der Bedieneinheit; deaktivieren",                                                                                                                                                                                      checklist => "Verbindungsleitung zwischen Regelger&auml;t und Warmwasser-Temperaturf&uuml;hler pr&uuml;fen&bull;Elektrischen Anschluss der Verbindungsleitung im Regelger&auml;t pr&uuml;fen&bull;Warmwasser-Temperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Warmwasser-Temperaturf&uuml;hlers im Regelger&auml;t laut Tabelle pr&uuml;fen&bull;Beim GB142 sollte ein Reset durchgef&uuml;hrt werden",                                                                     correction => "Wenn ein Defekt vorliegt, F&uuml;hler austauschen&bull;Falls Schrauben oder ein Stecker lose sind, das Kontaktproblem beheben&bull;Wenn die Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Regelger&auml;t austauschen&bull;Wenn nach dem Reset die St&ouml;rung noch besteht, dann ist die Ger&auml;teelektronik UBA; im W&auml;rmeerzeuger auszutauschen"};
	$ErrorCodeList{"0291"} = {errorcode => "A01",                                    subcode => "810",                                                class => "14",          reason => "Warmwasser bleibt kalt. Warmwassertemperatur ist w&auml;hrend der Ladung; 2 Std. lang nicht angestiegen. Auswirkung: Warmwasserladung l&auml;uft weiter, Warmwasser-Vorrang wird deaktiviert.",                                                                                                                                                    checklist => "Pumpe oder 3-Wege-Umschaltventil defekt&bull;Pr&uuml;fen, ob evtl. st&auml;ndig Wasser aufgrund von Zapfungen oder einer Leckage aus dem Warmwasserspeicher entnommen wird&bull;Die Position des Warmwasser-Temperaturf&uuml;hlers pr&uuml;fen, evtl. ist dieser falsch angebracht&bull;Die Verbindungsrohre zwischen Kessel und Speicher kontrollieren und nach Montageanleitung pr&uuml;fen, ob diese richtig angeschlossen sind&bull;Pr&uuml;fen, ob die Heizschlange im Speicher vollst&auml;ndig entl&uuml;ftet ist&bull;Entsprechend den technischen Unterlagen pr&uuml;fen, ob die eingebaute Speicherladepumpe &uuml;ber die erforderliche Leistung verf&uuml;gt&bull;Wenn der Warmwasser-Vorrang abgew&auml;hlt wurde und Heizung und Warmwasser im Parallelbetrieb laufen, kann evtl. die Leistung des Kessels nicht ausreichen&bull;Zu gro&szlig;e Verluste Zirkulationsleitung&bull;Den Warmwasser-Temperaturf&uuml;hler laut Tabelle pr&uuml;fen",correction => "Funktionstest durchf&uuml;hren&bull;Ggf. st&auml;ndige Warmwasserentnahme unterbinden&bull;Warmwasser-Temperaturf&uuml;hler richtig positionieren&bull;Bei Fehlern in der Verrohrung diese beheben&bull;Ggf. entl&uuml;ften&bull;Wenn Abweichungen bestehen, die Pumpe austauschen&bull;Warmwasserbereitung auf „Vorrang“ einstellen&bull;Zirkulationsleitung pr&uuml;fen&bull;Bei Abweichungen zu den Tabellenwerten den F&uuml;hler austauschen"};
	$ErrorCodeList{"0292"} = {errorcode => "A01",                                    subcode => "811",                                                class => "16",          reason => "Warmwasserbereitung: thermische Desinfektion misslungen; (Nach 3 Std. wurde die eingestellte Temperatur nicht erreicht.; Auswirkung: nach St&ouml;rmeldung; wird die thermische Desinfektion abgebrochen.)",                                                                                                                                       checklist => "Pr&uuml;fen, ob evtl. st&auml;ndig Wasser aufgrund von Zapfungen oder einer Leckage aus dem Warmwasserspeicher entnommen wird&bull;Die Position des Warmwasser-Temperaturf&uuml;hlers pr&uuml;fen, evtl. ist dieser falsch angebracht&bull;Die Verbindungsrohre zwischen Kessel und Speicher kontrollieren und nach Montageanleitung pr&uuml;fen, ob diese richtig angeschlossen sind&bull;Pr&uuml;fen, ob die Heizschlange im Speicher vollst&auml;ndig entl&uuml;ftet ist&bull;Entsprechend den technischen Unterlagen pr&uuml;fen, ob die eingebaute Speicherladepumpe &uuml;ber die erforderliche Leistung verf&uuml;gt&bull;Wenn der Warmwasser-Vorrang abgew&auml;hlt wurde und Heizung und Warmwasser im Parallelbetrieb laufen, kann evtl. die Leistung des Kessels nicht ausreichen&bull;Zu gro&szlig;e Verluste Zirkulationsleitung&bull;Den Warmwasser-Temperaturf&uuml;hler laut Tabelle pr&uuml;fen",correction => "Ggf. st&auml;ndige Warmwasserentnahme unterbinden&bull;Warmwasser-Temperaturf&uuml;hler richtig positionieren&bull;Bei Fehlern in der Verrohrung diese beheben&bull;Ggf. entl&uuml;ften&bull;Wenn Abweichungen bestehen, die Pumpe austauschen&bull;Warmwasserbereitung auf „Vorrang“ einstellen&bull;Zirkulationsleitung pr&uuml;fen&bull;Bei Abweichungen zu den Tabellenwerten den F&uuml;hler austauschen"};
	$ErrorCodeList{"0293"} = {errorcode => "A01",                                    subcode => "815",                                                class => "14",          reason => "Temperaturf&uuml;hler hydraulische Weiche defekt (Pumpeneffizienzmodul)",                                                                                                                                                                                                                                                                          checklist => "F&uuml;hler FK des Pumpeneffizienzmoduls defekt oder nicht korrekt angeschlossen&bull;-",                                                                                                                                                                                                                                                                                                                                                                                                             correction => "F&uuml;hleranschluss pr&uuml;fen&bull;Weichentemperaturf&uuml;hler auf falsche Einbauposition oder auf Bruchstelle pr&uuml;fen"                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0294"} = {errorcode => "A01",                                    subcode => "816",                                                class => "A",           reason => "Keine Kommunikation mit Busmaster Ger&auml;teelektronik (UBA oder MC10 oder MC40 oder MC100)/Kaskadenmodul/Feuerungsautomat/HM10 (Auswirkung: es findet kein Heiz-betrieb statt)",                                                                                                                                                                 checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist; Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder einschalten; pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung&bull;-",                                                                                                                                                                                                                                   correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t ausund wieder einschalten&bull;Busleitung reparieren bzw. austauschen&bull;Defekten EMS-BUS-Teilnehmer austauschen (EMS-Modul bzw. Feuerungsautomat bzw. Ger&auml;teelektronik bzw. Regelger&auml;t bzw. HM10)"                                                                                                                                };
	$ErrorCodeList{"0295"} = {errorcode => "A01",                                    subcode => "817",                                                class => "14",          reason => "Lufttemperaturf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                                 checklist => "Lufttemperaturf&uuml;hler inkl. Steckverbindung der Ger&auml;teelektronik SAFe pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Steckverbindung richtig aufstecken oder ggf. Temperaturf&uuml;hler austauschen"                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0296"} = {errorcode => "A01",                                    subcode => "818",                                                class => "14",          reason => "W&auml;rmeerzeuger bleibt kalt",                                                                                                                                                                                                                                                                                                                   checklist => "Die Version der Ger&auml;teelektronik SAFe pr&uuml;fen&bull;Evtl. liegen Probleme im Bereich des Kesself&uuml;hlers bzw. Luftprobleme im Kessel vor&bull;Es wurde ein Weichenmodul installiert, obwohl dieses nicht installiert werden darf, und damit ist die Pumpenlogik au&szlig;er Betrieb&bull;Die Auslegung der Kesselleistung bezogen auf die Anlage pr&uuml;fen&bull;Unter Umst&auml;nden ist die Kesselleistung nicht f&uuml;r einen Parallelbetrieb von Warmwasser und Heizung ausgelegt und es ist trotzdem Parallelbetrieb von Heizung und Warmwasser parametriert",correction => "Ist die Feuerungsautomat-Version nicht mindestens 2.14, sollte die Ger&auml;teelektronik ausgetauscht werden&bull;Anlage entl&uuml;ften bzw. Anschl&uuml;sse und Anschlusskabel des Kesself&uuml;hlers pr&uuml;fen oder den F&uuml;hler austauschen&bull;Weichenmodul deaktivieren&bull;Ggf. gr&ouml;&szlig;eren Kessel ausw&auml;hlen&bull;Heizungsanlage auf „Warmwasser-Vorrang“ einstellen"};
	$ErrorCodeList{"0297"} = {errorcode => "A01",                                    subcode => "819",                                                class => "14",          reason => "&Ouml;lvorw&auml;rmer meldet Dauersignal",                                                                                                                                                                                                                                                                                                         checklist => "Verbindungsleitung zwischen Ger&auml;teelektronik SAFe und &Ouml;lvorw&auml;rmer optisch auf Besch&auml;digungen pr&uuml;fen&bull;Verbindungsleitung zwischen Ger&auml;teelektronik SAFe und &Ouml;lvorw&auml;rmer abziehen und auf Kurzschl&uuml;sse pr&uuml;fen&bull;&Ouml;lvorw&auml;rmer im kalten Zustand pr&uuml;fen",                                                                                                                                                                          correction => "Sollte das Kabel besch&auml;digt sein, dieses austauschen&bull;Sollten Kurzschl&uuml;sse vorhanden sein, Kabel austauschen&bull;Sollte eine Verbindung vorhanden sein, &Ouml;lvorw&auml;rmer austauschen"                                                                                                                                                                                        };
	$ErrorCodeList{"0298"} = {errorcode => "A01",                                    subcode => "820",                                                class => "14",          reason => "&Ouml;l-Betriebstemperatur wird nicht erreicht",                                                                                                                                                                                                                                                                                                   checklist => "Verbindungsleitung zwischen Ger&auml;teelektronik SAFe und &Ouml;lvorw&auml;rmer optisch auf Besch&auml;digungen pr&uuml;fen&bull;Pr&uuml;fen, ob an Ger&auml;teelektronik SAFe und am &Ouml;lvorw&auml;rmer die Stecker richtig aufgesteckt sind&bull;Verbindungsleitung zwischen Ger&auml;teelektronik SAFe und &Ouml;lvorw&auml;rmer auf; Durchgang pr&uuml;fen&bull;Im Men&uuml; „Service“ der Bedieneinheit den Men&uuml;punkt „Relaistest“ aufrufen und hier den &Ouml;lvorw&auml;rmer einschalten; Mit der Hand pr&uuml;fen, ob der &Ouml;lvorw&auml;rmer sich erw&auml;rmt&bull;-",correction => "Falls das Kabel besch&auml;digt ist, dieses austauschen&bull;Stecker richtig aufstecken&bull;Erw&auml;rmt sich der &Ouml;lvorw&auml;rmer und die St&ouml;rungsmeldung bleibt erhalten, dann ist der Schaltkontakt im &Ouml;lvorw&auml;rmer defekt; Der &Ouml;lvorw&auml;rmer muss ausgetauscht werden&bull;Erw&auml;rmt sich der &Ouml;lvorw&auml;rmer und die St&ouml;rungsmeldung bleibt erhalten, dann ist der Schaltkontakt im &Ouml;lvorw&auml;rmer defekt; Der &Ouml;lvorw&auml;rmer muss ausgetauscht werden&bull;Erw&auml;rmt sich der &Ouml;lvorw&auml;rmer nicht, ist das Heizelement im &Ouml;lvorw&auml;rmer defekt; Der &Ouml;lvorw&auml;rmer muss ausgetauscht werden"};
	$ErrorCodeList{"0299"} = {errorcode => "A01",                                    subcode => "828",                                                class => "14",          reason => "Wasserdruckf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                                    checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Wasserdruckf&uuml;hler austauschen"                                                                                                                                                                                                                                                                                                                                                              };
	$ErrorCodeList{"0300"} = {errorcode => "A01",                                    subcode => "845",                                                class => "14",          reason => "Hydraulische Konfiguration wird nicht unterst&uuml;tzt",                                                                                                                                                                                                                                                                                           checklist => "W&auml;rmeerzeuger unterst&uuml;tzt die vorgegebene hydraulische Konfiguration nicht (z. B. weil mehr Pumpenausg&auml;nge ben&ouml;tigt werden; als vorhanden)&bull;-",                                                                                                                                                                                                                                                                                                                               correction => "Warmwasser auf Modul konfigurieren oder deinstallieren&bull;Heizkreis 1 auf Modul konfigurieren oder deinstallieren&bull;Systempumpe auf „Keine“ stellen"                                                                                                                                                                                                                                    };
	$ErrorCodeList{"0301"} = {errorcode => "A01",                                    subcode => "1010",                                               class => "8",           reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS",                                                                                                                                                                                                                                                                                                 checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist. Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder einschalten. Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung.&bull;-",                                                                                                                                                                                                                                  correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t Ausund wieder Einschalten&bull;Busleitung reparieren bzw. austauschen.&bull;Defekten EMS-BUS-Teilnehmer austauschen"                                                                                                                                                                                                                           };
	$ErrorCodeList{"0302"} = {errorcode => "A01",                                    subcode => "1030",                                               class => "8",           reason => "Etagenstation: Interner Datenfehler, Regelungsmodul austauschen",                                                                                                                                                                                                                                                                                  checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Modul austauschen"                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0303"} = {errorcode => "A01",                                    subcode => "1033; 1034; 1035; 1036",                             class => "8",           reason => "Etagenstation: interner Fehler, Ger&auml;t tauschen",                                                                                                                                                                                                                                                                                              checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Modul austauschen"                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0304"} = {errorcode => "A01",                                    subcode => "1038",                                               class => "8",           reason => "Etagenstation: Zeit/Datum ung&uuml;ltiger Wert",                                                                                                                                                                                                                                                                                                   checklist => "Datum/Zeit noch nicht eingestellt&bull;Spannungsversorgung &uuml;ber l&auml;ngere Zeit ausgefallen",                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Datum/Zeit einstellen&bull;l&auml;ngere Spannungsausf&auml;lle vermeiden"                                                                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"0305"} = {errorcode => "A01",                                    subcode => "1042",                                               class => "8",           reason => "Etagenstation: Interner Fehler: Zugriff auf Uhrenbaustein blockiert",                                                                                                                                                                                                                                                                              checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Modul austauschen"                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0306"} = {errorcode => "A01",                                    subcode => "1705",                                               class => "8",           reason => "Etagenstation: Modul erh&auml;lt unzul&auml;ssige Werte vom R&uuml;cklauftemperaturf&uuml;hler (Prim&auml;rkreis)",                                                                                                                                                                                                                                checklist => "Anschlussstecker am R&uuml;cklauftemperaturf&uuml;hler nicht aufgesteckt&bull;Anschlusskabel zum R&uuml;cklauftemperaturf&uuml;hler besch&auml;digt&bull;R&uuml;cklauftemperaturf&uuml;hler (Prim&auml;rkreis) defekt",                                                                                                                                                                                                                                                                               correction => "Anschlussstecker am R&uuml;cklauftemperaturf&uuml;hler (Prim&auml;rkreis) aufstecken&bull;Anschlusskabel zum R&uuml;cklauftemperaturf&uuml;hler (Prim&auml;rkreis) austauschen&bull;R&uuml;cklauftemperaturf&uuml;hler (Prim&auml;rkreis) austauschen"                                                                                                                                           };
	$ErrorCodeList{"0307"} = {errorcode => "A01",                                    subcode => "1706",                                               class => "8",           reason => "Etagenstation: Modul erh&auml;lt unzul&auml;ssige Werte vom Vorlauftemperaturf&uuml;hler (Prim&auml;rkreis)",                                                                                                                                                                                                                                      checklist => "Anschlussstecker am Vorlauftemperaturf&uuml;hler (Prim&auml;rkreis) nicht aufgesteckt&bull;Anschlusskabel zum Vorlauftemperaturf&uuml;hler (Prim&auml;rkreis) besch&auml;digt&bull;Vorlauftemperaturf&uuml;hler (Prim&auml;rkreis) defekt",                                                                                                                                                                                                                                                           correction => "Anschlussstecker am Vorlauftemperaturf&uuml;hler (Prim&auml;rkreis) aufstecken&bull;Anschlusskabel zum Vorlauftemperaturf&uuml;hler (Prim&auml;rkreis) austauschen&bull;Vorlauftemperaturf&uuml;hler (Prim&auml;rkreis) austauschen"                                                                                                                                                             };
	$ErrorCodeList{"0308"} = {errorcode => "A01",                                    subcode => "1707",                                               class => "8",           reason => "Etagenstation: Ung&uuml;ltige Position des Koodierschalters f&uuml;r Warmwasser",                                                                                                                                                                                                                                                                  checklist => "Falsche Position f&uuml;r den Kodierschalter (Warmwasser) eingestellt&bull;Kodierschalter f&uuml;r Warmwasser defekt",                                                                                                                                                                                                                                                                                                                                                                                correction => "Position des Kodierschalters f&uuml;r Warmwasser auf die korrekte Kennzahl drehen.&bull;Modul austauschen"                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0309"} = {errorcode => "A01",                                    subcode => "1708",                                               class => "8",           reason => "Etagenstation: Ung&uuml;ltige Position des Kodierschalters f&uuml;r Heizung",                                                                                                                                                                                                                                                                      checklist => "Falsche Position f&uuml;r den Kodierschalter (Heizung) eingestellt&bull;Kodierschalter f&uuml;r Heizung defekt",                                                                                                                                                                                                                                                                                                                                                                                      correction => "Position des Kodierschalters f&uuml;r Heizung auf die korrekte Kennzahl drehen.&bull;Modul austauschen"                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"0310"} = {errorcode => "A01",                                    subcode => "1710",                                               class => "4",           reason => "Etagenstation: Vorlauftemperatur (Prim&auml;rkreis) zu niedrig",                                                                                                                                                                                                                                                                                   checklist => "Position der Kodierschalter am Modul &uuml;berpr&uuml;fen&bull;Vorlauftemperatur in der Systemauslegung zu niedrig&bull;Vorlauftemperaturf&uuml;hler (Prim&auml;r) ist nicht korrekt montiert",                                                                                                                                                                                                                                                                                                       correction => "Kodierschalter auf die notwendige Versorgungstemperatur anheben&bull;Einstellungen am W&auml;rmeerzeuger auf das System anpassen&bull;Vorlauftemperaturf&uuml;hler (Prim&auml;rkreis) an der korrekten Position montieren"                                                                                                                                                                       };
	$ErrorCodeList{"0311"} = {errorcode => "A01",                                    subcode => "1711",                                               class => "8",           reason => "Etagenstation: Vorlauftemperatur (Prim&auml;rkreis) zu hoch",                                                                                                                                                                                                                                                                                      checklist => "Systemauslegung bzw. Einstellungen; am W&auml;rmeerzeuger ist nicht auf die HDU abgestimmt&bull;Systemauslegung bei Fernw&auml;rmeanschluss nicht f&uuml;r HDU geeignet",                                                                                                                                                                                                                                                                                                                             correction => "System bzw. Einstellungen am W&auml;rmeerzeuger anpassen&bull;Fernw&auml;rmeanschlu&szlig; an HDU-Parameter anpassen."                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"0312"} = {errorcode => "A01",                                    subcode => "1712",                                               class => "8",           reason => "Etagenstation: Modul erh&auml;lt unzul&auml;ssige Werte f&uuml;r den Warmwassertemperaturf&uuml;hler",                                                                                                                                                                                                                                             checklist => "Anschlusskabel vom Warmwassertemperaturf&uuml;hler im Modul &uuml;berpr&uuml;fen&bull;Anschlusskabel zum Warmwassertemperaturf&uuml;hler &uuml;berpr&uuml;fen&bull;-",                                                                                                                                                                                                                                                                                                                                correction => "Anschlusskabel vom Warmwassertemperaturf&uuml;hler im Modul anschlie&szlig;en&bull;Warmwassertemperaturf&uuml;hler komplett austauschen&bull;Warmwassertemperaturf&uuml;hler komplett austauschen"                                                                                                                                                                                               };
	$ErrorCodeList{"0313"} = {errorcode => "A01",                                    subcode => "1713",                                               class => "8",           reason => "Etagenstation: Modul erh&auml;lt unzul&auml;ssige Werte f&uuml;r den Warmnwasservolumenstrom",                                                                                                                                                                                                                                                     checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Volumenstrommesser im Warmwasserkreis austauschen"                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0314"} = {errorcode => "A01",                                    subcode => "1716",                                               class => "8",           reason => "Etagenstation: Modul erh&auml;lt unzul&auml;ssige Werte f&uuml;r den Vorlauftemperaturf&uuml;hler (Sekund&auml;rkreis)",                                                                                                                                                                                                                           checklist => "Anschlussstecker am Vorlauftemperaturf&uuml;hler (Sekund&auml;rkreis) &uuml;berpr&uuml;fen&bull;Anschlusskabel zum Vorlauftemperaturf&uuml;hler (Sekund&auml;rkreis) &uuml;berpr&uuml;fen&bull;-",                                                                                                                                                                                                                                                                                                    correction => "Anschlussstecker am Vorlauftemperaturf&uuml;hler (Sekund&auml;rkreis) aufstecken&bull;Anschlusskabel zum Vorlauftemperaturf&uuml;hler (Sekund&auml;rkreis) austauschen&bull;Vorlauftemperaturf&uuml;hler (Sekund&auml;rkreis) austauschen"                                                                                                                                                       };
	$ErrorCodeList{"0315"} = {errorcode => "A01",                                    subcode => "1717",                                               class => "4",           reason => "Etagenstation: Modul erh&auml;lt unzul&auml;ssige Werte vom Au&szlig;entemperaturf&uuml;hler",                                                                                                                                                                                                                                                     checklist => "Konfiguration &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob ein Au&szlig;entemperaturf&uuml;hler montiert ist&bull;Anschluss des Au&szlig;entemperaturf&uuml;hlers am Modul &uuml;berpr&uuml;fen&bull;Anschlusskabel am Au&szlig;entemperaturf&uuml;hler &uuml;berpr&uuml;fen&bull;Au&szlig;entemperaturf&uuml;hler &uuml;berpr&uuml;fen&bull;Anschlusskabel zum Au&szlig;entemperatuf&uuml;hler &uuml;berpr&uuml;fen",                                                                          correction => "Konfiguration auf raumtemperaturgef&uuml;hrt&bull;Au&szlig;entemperaturf&uuml;hler montieren&bull;Anschlusskabel vom Au&szlig;entemperaturf&uuml;hler am Modul anschlie&szlig;en&bull;Anschlusskabel am Au&szlig;entemperaturf&uuml;hler anschlie&szlig;en&bull;Au&szlig;entemperaturf&uuml;hler austauschen&bull;Anschlusskabel zum Au&szlig;entemperaturf&uuml;hler instandsetzen oder austauschen"};
	$ErrorCodeList{"0316"} = {errorcode => "A01",                                    subcode => "1718",                                               class => "8",           reason => "Etagenstation: Modul erh&auml;lt unzul&auml;ssige Werte f&uuml;r den R&uuml;cklauftemperaturf&uuml;hler (Sekund&auml;rkreis)",                                                                                                                                                                                                                     checklist => "Anschlussstecker am R&uuml;cklauftemperaturf&uuml;hler (Sekund&auml;rkreis) &uuml;berpr&uuml;fen&bull;Anschlusskabel zum R&uuml;cklauftemperaturf&uuml;hler (Sekund&auml;rkreis) &uuml;berpr&uuml;fen&bull;-",                                                                                                                                                                                                                                                                                        correction => "Anschlussstecker am R&uuml;cklauftemperaturf&uuml;hler (Sekund&auml;rkreis) aufstecken&bull;Anschlusskabel zm R&uuml;cklauftemperaturf&uuml;hler (Sekund&auml;rkreis) austauschen&bull;R&uuml;cklauftemperaturf&uuml;hler (Sekund&auml;rkreis) austauschen"                                                                                                                                      };
	$ErrorCodeList{"0317"} = {errorcode => "A01",                                    subcode => "1719",                                               class => "8",           reason => "Etagenstation: Regelventil f&uuml;r Heizung in HDU defekt",                                                                                                                                                                                                                                                                                        checklist => "PWM-R&uuml;cksignal vom Regelventil (Heizung) liegt au&szlig;erhalb des zul&auml;ssigen Bereiches",                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelventil f&uuml;r den Heizbetrieb austauschen"                                                                                                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"0318"} = {errorcode => "A01",                                    subcode => "1720",                                               class => "8",           reason => "Etagenstation: Regelventil f&uuml;r Warmwasser in HDU defekt",                                                                                                                                                                                                                                                                                     checklist => "PWM-R&uuml;cksignal vom Regelventil (Warmwasser) liegt au&szlig;erhalb des zul&auml;ssigen Bereiches",                                                                                                                                                                                                                                                                                                                                                                                                correction => "Regelventil f&uuml;r den Warmwasserbetrieb austauschen"                                                                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"0319"} = {errorcode => "A01",                                    subcode => "1900",                                               class => "8",           reason => "AM200: Sensor defekt (Sys  Kabel zum Sensor pr&uuml;fen              ",                                                                                                                                                                                                                                                                            checklist => "Sensor ersetzen temvorlauftemperatur)             Elektrische Verbindung zum Sensor pr&uuml;fen            Sensorkabel ersetzen; Sensorkennlinie pr&uuml;fen       ",                                                                                                                                                                                                                                                                                                                                 correction => "Kaskadenmodul ersetzen Spannung am Sensoranschlu&szlig; pr&uuml;fen; Spannung am Kaskadenmodul abschalten/ anschalten"                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"0320"} = {errorcode => "A01",                                    subcode => "1901",                                               class => "8",           reason => "AM200: Sensor defekt (Systemr&uuml;cklauftemperatur)",                                                                                                                                                                                                                                                                                             checklist => "Kabel zum Sensor pr&uuml;fen&bull;Elektrische Verbindung zum Sensor pr&uuml;fen&bull;Sensorkennlinie pr&uuml;fen&bull;Spannung am Sensoranschlu&szlig; pr&uuml;fen&bull;Spannung am Kaskadenmodul abschalten/ anschalten",                                                                                                                                                                                                                                                                            correction => "Sensor ersetzen&bull;Sensorkabel ersetzen&bull;Kaskadenmodul ersetzen&bull;-"                                                                                                                                                                                                                                                                                                                    };
	$ErrorCodeList{"0321"} = {errorcode => "A01",                                    subcode => "1902",                                               class => "8",           reason => "AM200: Sensor defekt (Au&szlig;entemperatur)",                                                                                                                                                                                                                                                                                                     checklist => "Kabel zum Sensor pr&uuml;fen&bull;Elektrische Verbindung zum Sensor pr&uuml;fen&bull;Sensorkennlinie pr&uuml;fen&bull;Spannung am Sensoranschlu&szlig; pr&uuml;fen&bull;Spannung am Kaskadenmodul abschalten/ anschalten",                                                                                                                                                                                                                                                                            correction => "Sensor ersetzen&bull;Sensorkabel ersetzen&bull;Kaskadenmodul ersetzen&bull;-"                                                                                                                                                                                                                                                                                                                    };
	$ErrorCodeList{"0322"} = {errorcode => "A01",                                    subcode => "1903",                                               class => "8",           reason => "AM200: Schalterposition Drehcodierschalter ung&uuml;ltig",                                                                                                                                                                                                                                                                                         checklist => "Schalterstellung pr&uuml;fen&bull;-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "Schalterstellung korrigieren&bull;Netzspannung aus und wieder einschalten&bull;Kaskadenmodul ersetzen"                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"0323"} = {errorcode => "A01",                                    subcode => "1904",                                               class => "8",           reason => "Etagenstation: Sensor defekt (Systemr&uuml;cklauftemperatur)",                                                                                                                                                                                                                                                                                     checklist => "Konfiguration pr&uuml;fen. Mit der gew&auml;hlten Einstellung ist ein F&uuml;hler Versorgungsr&uuml;cklauf angebracht&bull;Verbindungsleitung zwischen Modul HDU und Versorgungsr&uuml;cklauf pr&uuml;fen&bull;F&uuml;hler Versorgungsr&uuml;cklauf laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des F&uuml;hler Versorgungsr&uuml;cklauf laut Tabelle; pr&uuml;fen",                                                                                                               correction => "Konfiguration &auml;ndern&bull;Verbindung ordnungsgem&auml;&szlig; herstellen&bull;Wenn die Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Modul austauschen"                                                                                                      };
	$ErrorCodeList{"0324"} = {errorcode => "A01",                                    subcode => "1905",                                               class => "8",           reason => "Etagenstation: Modul erh&auml;lt unzul&auml;ssige Werte f&uuml;r den Systemvorlauf-temperaturf&uuml;hler",                                                                                                                                                                                                                                         checklist => "Konfiguration pr&uuml;fen. Mit der gew&auml;hlten Einstellung ist ein F&uuml;hler Versorgungsvorlauf angebracht&bull;Verbindungsleitung zwischen Modul HDU und Versorgungsvorlauf pr&uuml;fen&bull;F&uuml;hler Versorgungsvorlauf laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des F&uuml;hler Versorgungsvorlauf laut Tabelle pr&uuml;fen",                                                                                                                                        correction => "Konfiguration &auml;ndern&bull;Verbindung ordnungsgem&auml;&szlig; herstellen&bull;Wenn die Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Modul austauschen"                                                                                                      };
	$ErrorCodeList{"0325"} = {errorcode => "A01",                                    subcode => "1906",                                               class => "8",           reason => "Etagenstation: Schalterposition ung&uuml;ltig",                                                                                                                                                                                                                                                                                                    checklist => "Modul verh&auml;lt sich wie in Stellung 'off', Funktionen sind deaktiviert&bull;-",                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Beide Poti-Drehknopfpositonen &uuml;berpr&uuml;fen&bull;Poti-Drehknopfpositon auf 0 und wieder auf neuen Wert setzen"                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"0326"} = {errorcode => "A01",                                    subcode => "1908",                                               class => "4",           reason => "Etagenstation: Systemvorlauftemperatur zu niedrig",                                                                                                                                                                                                                                                                                                checklist => "Konfiguration pr&uuml;fen. Mit der gew&auml;hlten Einstellung ben&ouml;tigt die HDU die an den; Poti-Drehkn&ouml;pfen angegebene Versorgungstemperatur&bull;Zuleitungstemperatur am Kessel bzw. Speicher pr&uuml;fen&bull;Auf Lufteinschl&uuml;sse pr&uuml;fen",                                                                                                                                                                                                                                      correction => "Versorgungstemperatur anheben&bull;Versorgungsleistung zur Verf&uuml;gung stellen, Volumenstrom anheben bzw. Anlage entl&uuml;ften&bull;-"                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0327"} = {errorcode => "A01",                                    subcode => "1909",                                               class => "8",           reason => "Etagenstation: System-vorlauftemperatur zu hoch",                                                                                                                                                                                                                                                                                                  checklist => "Wasserseitige Durchstr&ouml;mung pr&uuml;fen -; Pr&uuml;fung, ob Fremdw&auml;rmequelle vorhanden (Feststoffkessel)&bull;Kesself&uuml;hler/SicherheitstemperaturbegrenzerF&uuml;hler pr&uuml;fen.&bull;Z&uuml;nd-/&Uuml;berwachungselektrode pr&uuml;fen.",                                                                                                                                                                                                                                            correction => "Ausreichende Durchstr&ouml;mung sicherstellen&bull;Ggf. Kesself&uuml;hler/Sicherheitstemperaturbegrenzer-F&uuml;hler austauschen&bull;Ggf. Z&uuml;nd-/&Uuml;berwachungselektrode am Kessel austauschen"                                                                                                                                                                                          };
	$ErrorCodeList{"0328"} = {errorcode => "A01",                                    subcode => "1920",                                               class => "8",           reason => "AM200: Signal vom Vorlauftemperaturf&uuml;hler (System) liegt au&szlig;erhalb der Kennlinie",                                                                                                                                                                                                                                                      checklist => "F&uuml;hlerleitung pr&uuml;fen&bull;Steckverbindung pr&uuml;fen&bull;F&uuml;hlerwerte nach Tabelle pr&uuml;fen&bull;Spannungswerte am F&uuml;hler laut Tabelle pr&uuml;fen&bull;-",                                                                                                                                                                                                                                                                                                                   correction => "Bei Besch&auml;digung austauschen&bull;Bei Verschmutzung reinigen oder evtl. austauschen&bull;Bei Besch&auml;digung austauschen&bull;Bei losem Stecker diesen wieder aufstecken&bull;Bei Abweichungen den F&uuml;hler austauschen"                                                                                                                                                               };
	$ErrorCodeList{"0329"} = {errorcode => "A01",                                    subcode => "1921",                                               class => "8",           reason => "AM200: Signal vom R&uuml;cklauftemperaturf&uuml;hler (System) liegt au&szlig;erhalb der Kennlinie",                                                                                                                                                                                                                                                checklist => "F&uuml;hlerleitung pr&uuml;fen&bull;Steckverbindung pr&uuml;fen&bull;F&uuml;hlerwerte nach Tabelle pr&uuml;fen&bull;Spannungswerte am F&uuml;hler laut Tabelle pr&uuml;fen",                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0330"} = {errorcode => "A01",                                    subcode => "1922",                                               class => "8",           reason => "AM200: Signal vom Vorlauftemperaturf&uuml;hler am alternativen W&auml;rmeerzeuger liegt au&szlig;erhalb der Kennlinie",                                                                                                                                                                                                                            checklist => "F&uuml;hlerleitung pr&uuml;fen&bull;Steckverbindung pr&uuml;fen&bull;F&uuml;hlerwerte nach Tabelle pr&uuml;fen&bull;Spannungswerte am F&uuml;hler laut Tabelle pr&uuml;fen",                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0331"} = {errorcode => "A01",                                    subcode => "1923",                                               class => "8",           reason => "AM200: Signal vom R&uuml;cklauftemperaturf&uuml;hler am alternativen W&auml;rmeerzeuger liegt au&szlig;erhalb der Kennlinie",                                                                                                                                                                                                                      checklist => "F&uuml;hlerleitung pr&uuml;fen&bull;Steckverbindung pr&uuml;fen&bull;F&uuml;hlerwerte nach Tabelle pr&uuml;fen&bull;Spannungswerte am F&uuml;hler laut Tabelle pr&uuml;fen",                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0332"} = {errorcode => "A01",                                    subcode => "1924",                                               class => "8",           reason => "AM200: Signal vom Temperaturf&uuml;hler (Pufferspeicher oben) liegt au&szlig;erhalb der Kennlinie",                                                                                                                                                                                                                                                checklist => "F&uuml;hlerleitung pr&uuml;fen&bull;Steckverbindung pr&uuml;fen&bull;F&uuml;hlerwerte nach Tabelle pr&uuml;fen&bull;Spannungswerte am F&uuml;hler laut Tabelle pr&uuml;fen",                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0333"} = {errorcode => "A01",                                    subcode => "1925",                                               class => "8",           reason => "AM200: Signal vom Temperaturf&uuml;hler (Pufferspeicher Mitte) liegt au&szlig;erhalb der Kennlinie",                                                                                                                                                                                                                                               checklist => "F&uuml;hlerleitung pr&uuml;fen&bull;Steckverbindung pr&uuml;fen&bull;F&uuml;hlerwerte nach Tabelle pr&uuml;fen&bull;Spannungswerte am F&uuml;hler laut Tabelle pr&uuml;fen",                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0334"} = {errorcode => "A01",                                    subcode => "1926",                                               class => "8",           reason => "AM200: Signal vom Temperaturf&uuml;hler (Pufferspeicher unten) liegt au&szlig;erhalb der Kennlinie",                                                                                                                                                                                                                                               checklist => "F&uuml;hlerleitung pr&uuml;fen&bull;Steckverbindung pr&uuml;fen&bull;F&uuml;hlerwerte nach Tabelle pr&uuml;fen&bull;Spannungswerte am F&uuml;hler laut Tabelle pr&uuml;fen",                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0335"} = {errorcode => "A01",                                    subcode => "1927",                                               class => "8",           reason => "AM200: Signal vom Warmwasser Temperaturf&uuml;hler liegt au&szlig;erhalb der Kennlinie",                                                                                                                                                                                                                                                           checklist => "F&uuml;hlerleitung pr&uuml;fen&bull;Steckverbindung pr&uuml;fen&bull;F&uuml;hlerwerte nach Tabelle pr&uuml;fen&bull;Spannungswerte am F&uuml;hler laut Tabelle pr&uuml;fen",                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0336"} = {errorcode => "A01",                                    subcode => "1928",                                               class => "8",           reason => "AM200: Signal vom Au&szlig;entemperaturf&uuml;hler liegt au&szlig;erhalb der Kennlinie",                                                                                                                                                                                                                                                           checklist => "F&uuml;hlerleitung pr&uuml;fen&bull;Steckverbindung pr&uuml;fen&bull;F&uuml;hlerwerte nach Tabelle pr&uuml;fen&bull;Spannungswerte am F&uuml;hler laut Tabelle pr&uuml;fen",                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0337"} = {errorcode => "A01",                                    subcode => "1929",                                               class => "8",           reason => "AM200: Signal vom Abgastemperaturf&uuml;hler liegt au&szlig;erhalb der Kennlinie",                                                                                                                                                                                                                                                                 checklist => "F&uuml;hlerleitung pr&uuml;fen&bull;Steckverbindung pr&uuml;fen&bull;F&uuml;hlerwerte nach Tabelle pr&uuml;fen&bull;Spannungswerte am F&uuml;hler laut Tabelle pr&uuml;fen",                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0338"} = {errorcode => "A01",                                    subcode => "1930",                                               class => "8",           reason => "AM200: Systemkonfiguration nicht best&auml;tigt",                                                                                                                                                                                                                                                                                                  checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Best&auml;tigung Systemkonfiguration in der Bedieneinheit erforderlich"                                                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"0339"} = {errorcode => "A01",                                    subcode => "1931",                                               class => "8",           reason => "AM200: Schalterposition ung&uuml;ltig",                                                                                                                                                                                                                                                                                                            checklist => "Drehkodierschalter AM200 zwischen 2 g&uuml;ltigen Positionen&bull;Schalterstellung pr&uuml;fen&bull;Schalter defekt",                                                                                                                                                                                                                                                                                                                                                                                 correction => "Schalterstellung korrigieren&bull;Netzspannung aus und wieder einschalten&bull;Modul ersetzen"                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"0340"} = {errorcode => "A01",                                    subcode => "3818",                                               class => "4",           reason => "Keine Buskommunikation zwischen LM10/IUM10 und Ger&auml;teelektronik UBA-H3",                                                                                                                                                                                                                                                                      checklist => "Die Verdrahtung zwischen LM10/IUM10 und Ger&auml;teelektronik UBA-H3 pr&uuml;fen&bull;Verbindungsleitung zwischen LM10/IUM10 und Ger&auml;teelektronik UBA-H3 pr&uuml;fen&bull;LM10/IUM10 defekt",                                                                                                                                                                                                                                                                                                    correction => "Bei Fehlverdrahtung den Verdrahtungsfehler korrigieren&bull;Bei Defekt Datenleitung ersetzen&bull;LM10/IUM10 austauschen"                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"0341"} = {errorcode => "A01",                                    subcode => "4111",                                               class => "8",           reason => "Etagenstation: Trinkwasservolumensensor defekt",                                                                                                                                                                                                                                                                                                   checklist => "Konfiguration pr&uuml;fen. Mit der gew&auml;hlten Einstellung ist ein Trinkwasservolumenstromsensor erforderlich.&bull;Verbindungsleitung zwischen Regelger&auml;t und Trinkwasservolumenstromsensor auf Durchgang pr&uuml;fen&bull;Elektrischer Anschluss der Verbindungsleitung in Trinkwasservolumenstromsensor bzw. am Stecker im Regelger&auml;t pr&uuml;fen&bull;-",                                                                                                                            correction => "Konfiguration &auml;ndern&bull;Verbindung ordnungsgem&auml;&szlig; herstellen&bull;Wenn die Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Modul austauschen"                                                                                                      };
	$ErrorCodeList{"0342"} = {errorcode => "A01",                                    subcode => "5100; 5101; 5104",                                   class => "8",           reason => "Interner Fehler im Steuerger&auml;t (MCU)",                                                                                                                                                                                                                                                                                                        checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Versorgungsspannung kurz unterbrechen&bull;Steuerger&auml;t (MCU) austauschen"                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"0343"} = {errorcode => "A01",                                    subcode => "5105",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Interner Fehler im Steuerger&auml;t; (CU I/O)",                                                                                                                                                                                                                                                                               checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Versorgungsspannung kurz unterbrechen&bull;Steuerger&auml;t (CU I/O) austauschen"                                                                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"0344"} = {errorcode => "A01",                                    subcode => "5106",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Interner Fehler im Steuerger&auml;t  der Au&szlig;eneinheit",                                                                                                                                                                                                                                                                 checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Versorgungsspannung kurz unterbrechen&bull;Steuerger&auml;t in der Au&szlig;eneinheit austauschen"                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0345"} = {errorcode => "A01",                                    subcode => "5110",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Kein CAN Bussignal am Steuerger&auml;t (CU I/O) vorhanden",                                                                                                                                                                                                                                                                   checklist => "&Uuml;berpr&uuml;fen, ob das CAN  Buskabel  am Steuerger&auml;t (CU I/O) korrekt angeschlossen ist.&bull;&Uuml;berpr&uuml;fen, ob das CAN Buskabel zwischen den Steuerger&auml;ten MC und CU I/O besch&auml;digt ist.&bull;-",                                                                                                                                                                                                                                                                        correction => "CAN Buskabel im Steuerger&auml;t (CU I/O) korrekt anschlie&szlig;en&bull;CAN Buskabel zwischen den Steuerger&auml;ten MCU und CU I/O austauschen&bull;Steuerger&auml;t (CU I/O) austauschen"                                                                                                                                                                                                     };
	$ErrorCodeList{"0346"} = {errorcode => "A01",                                    subcode => "5111",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: St&ouml;rung in der CAN Buskommunikation zwischen Innenund Au&szlig;eneinheit",                                                                                                                                                                                                                                               checklist => "&Uuml;berpr&uuml;fen, ob das CAN -Buskabel am Steuerger&auml;t der Inneneinheit korrekt angeschlossen ist&bull;&Uuml;berpr&uuml;fen, ob das CAN Buskabel am Steuerger&auml;t der Au&szlig;eneinheit korrekt angeschlossen ist.&bull;Abstand zwischen dem CAN Buskabel und dem Anschlusskabel f&uuml;r die Spannungsversorgung der Au&szlig;eneinheit &uuml;berpr&uuml;fen.&bull;-",                                                                                                                   correction => "CAN Buskabel im Steuerger&auml;t der Inneneinheit korrekt anschlie&szlig;en&bull;CAN Buskabel im Steuerger&auml;t der Au&szlig;eneinheit korrekt anschlie&szlig;en&bull;Abstand von mindestens 100 mm zwischen CAN Buskabel und Anschlusskabel f&uuml;r die Spannungsversorgung sicherstellen&bull;CAN Buskabel zwischen Au&szlig;enund Inneneinheit instandsetzen oder austauschen"             };
	$ErrorCodeList{"0347"} = {errorcode => "A01",                                    subcode => "5112",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Kein CAN Bussignal am Steuerger&auml;t (GHP-SC) vorhanden",                                                                                                                                                                                                                                                                   checklist => "&Uuml;berpr&uuml;fen, ob das CAN  Buskabel  am Steuerger&auml;t (GHP-SC) korrekt angeschlossen ist.&bull;&Uuml;berpr&uuml;fen, ob das CAN  Buskabel  am Steuerger&auml;t (CU I/O) korrekt angeschlossen ist.&bull;&Uuml;berpr&uuml;fen, ob das CAN Buskabel zwischen den Steuerger&auml;ten GHP-SC und CU I/O besch&auml;digt ist.&bull;Spannungsversorgung am Steuerger&auml;t CU I/O pr&uuml;fen&bull;-",                                                                                           correction => "CAN Buskabel im Steuerger&auml;t (GHP-SC) korrekt anschlie&szlig;en&bull;CAN Buskabel im Steuerger&auml;t (CU I/O) korrekt anschlie&szlig;en&bull;CAN Buskabel zwischen den Steuerger&auml;ten GHP-SC und CU I/O austauschen&bull;Spannungsversorgung  am Steuerger&auml;t (CU I/O) korrigieren&bull;Steuerger&auml;t CU I/O tauschen&bull;Steuerger&auml;t GHP-SC tauschen"                     };
	$ErrorCodeList{"0348"} = {errorcode => "A01",                                    subcode => "5130",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: Steuerger&auml;t erh&auml;lt unzul&auml;ssige Werte vom Vorlauftemperaturf&uuml;hler",                                                                                                                                                                                                                                        checklist => "&Uuml;berpr&uuml;fen ob der Anschlussstecker am Vorlauftemperaturf&uuml;hler TTI-6 komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel vom Vorlauftemperaturf&uuml;hler TTI-6 am Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;Ohmschen Widerstand des Temperaturf&uuml;hlers TTI-6 messen&bull;&Uuml;berpr&uuml;fen, ob der Vorlauftemperaturf&uuml;hler TTI-6 korrekten Kontakt an der Vorlaufleitung hat.&bull;Position des Vorlauftemperaturf&uuml;hlers TTI-6 an der Vorlaufleitung &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob die Isolierung am Vorlauftemperaturf&uuml;hler TTI-6 vorhanden bzw. besch&auml;digt ist&bull;-",correction => "Anschlussstecker am Vorlauftemperaturf&uuml;hler TTI-6 komplett aufstecken.&bull;Anschlusskabel vom Vorlauftemperaturf&uuml;hler TTI-6 am Steuerger&auml;t (CU i/O) korrekt anschlie&szlig;en&bull;Vorlauftemperaturf&uuml;hler TTI-6 austauschen&bull;Kontakt zwischen Vorlauftemperaturf&uuml;hler TTI-6 und Vorlaufleitung herstellen&bull;Vorlauftemperaturf&uuml;hler TTI-6 an der korrekten Position der Vorlaufleitung montieren&bull;Isolierung am Vorlauftemperaturf&uuml;hler TTI-6 austauschen bzw. montieren&bull;Kabelbaum austauschen"};
	$ErrorCodeList{"0349"} = {errorcode => "A01",                                    subcode => "5131",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: Steuerger&auml;t erh&auml;lt unzul&auml;ssige Werte vom R&uuml;cklauftemperaturf&uuml;hle",                                                                                                                                                                                                                                   checklist => "&Uuml;berpr&uuml;fen ob der Anschlussstecker am  R&uuml;cklauftemperaturf&uuml;hler TTI-5 komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel vom R&uuml;cklauftemperaturf&uuml;hler TTI-5 am Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;Ohmschen Widerstand des R&uuml;cklauftemperaturf&uuml;hlers TTI-5 messen&bull;&Uuml;berpr&uuml;fen, ob der R&uuml;cklauftemperaturf&uuml;hler; TTI-5 korrekten Kontakt an der Vorlaufleitung hat.&bull;Position des R&uuml;cklauftemperaturf&uuml;hlers TTI-5 an der Vorlaufleitung &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob die Isolierung am R&uuml;cklauftemperaturf&uuml;hler TTI-5 vorhanden bzw. besch&auml;digt ist&bull;-",correction => "Anschlussstecker am R&uuml;cklauftemperaturf&uuml;hler TTI-5 komplett aufstecken.&bull;Anschlusskabel vom R&uuml;cklauftemperaturf&uuml;hler TTI-5 am Steuerger&auml;t (CU i/O) korrekt anschlie&szlig;en&bull;R&uuml;cklauftemperaturf&uuml;hler TTI-5 austauschen&bull;Kontakt zwischen R&uuml;cklauftemperaturf&uuml;hler TTI-5 und R&uuml;cklaufleitung herstellen&bull;R&uuml;cklauftemperaturf&uuml;hler TTI-5 an der korrekten Position der R&uuml;cklaufleitung montieren&bull;Isolierung am R&uuml;cklauftemperaturf&uuml;hler TTI-6 austauschen bzw. montieren&bull;Kabelbaum austauschen"};
	$ErrorCodeList{"0350"} = {errorcode => "A01",                                    subcode => "5132",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: Steuerger&auml;t erh&auml;lt unzul&auml;ssige Werte vom Temperaturf&uuml;hler im Solevorlauf",                                                                                                                                                                                                                                checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Temperaturf&uuml;hler TTI-7 komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel vom Temperaturf&uuml;hler TTI-7 im Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Temperaturf&uuml;hler TTI-7 besch&auml;digt ist&bull;Ohmschen Widerstand des Temperaturf&uuml;hlers TTI-7 &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-7 korrekten Kontakt zur Solevorlaufleitung hat&bull;&Uuml;berpr&uuml;fen, ob die Isolierung am Temperaturf&uuml;hler TTI-7 vorhanden bzw besch&auml;digt ist&bull;Solekreis auf Entnahmeleistung &uuml;berpr&uuml;fen&bull;-",correction => "Anschlussstecker am Temperaturf&uuml;hler TTI-7 komplett aufstecken&bull;Anschlusskabel vom Temperaturf&uuml;hler TTI-7 korrekt im Steuerger&auml;t (CU I/O) anschlie&szlig;en&bull;Kabelbaum autauschen&bull;Temperaturf&uuml;hler TTI-7 austauschen&bull;Temperaturf&uuml;hler TTI-7 an der korrekten Position der Solevorlaufleitung montieren&bull;Isolierung am Temperaturf&uuml;hler TTI-7 austauschen bzw. montieren&bull;Korrekten Volumenstrom im Solekreis einstellen&bull;Umw&auml;lzpumpe im Solekreis austauschen"};
	$ErrorCodeList{"0351"} = {errorcode => "A01",                                    subcode => "5133",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: Steuerger&auml;t erh&auml;lt unzul&auml;ssige Werte vom Soler&uuml;cklauftemperaturf&uuml;hler",                                                                                                                                                                                                                              checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Temperaturf&uuml;hler TTI-8 komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel vom Temperaturf&uuml;hler TTI-8 im Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Temperaturf&uuml;hler TTI-8 besch&auml;digt ist&bull;Ohmschen Widerstand des Temperaturf&uuml;hlers TTI-8 &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-8 korrekten Kontakt zur Soler&uuml;cklaufleitung hat&bull;&Uuml;berpr&uuml;fen, ob die Isolierung am Temperaturf&uuml;hler TTI-8 vorhanden bzw besch&auml;digt ist&bull;Solekreis auf Entnahmeleistung &uuml;berpr&uuml;fen&bull;-",correction => "Anschlussstecker am Temperaturf&uuml;hler TTI-8 komplett aufstecken&bull;Anschlusskabel vom Temperaturf&uuml;hler TTI-8 korrekt im Steuerger&auml;t (CU I/O) anschlie&szlig;en&bull;Kabelbaum autauschen&bull;Temperaturf&uuml;hler TTI-8 austauschen&bull;Temperaturf&uuml;hler TTI-8 an der korrekten Position der Soler&uuml;cklaufleitung montieren&bull;Isolierung am Temperaturf&uuml;hler TTI-8 austauschen bzw. montieren&bull;Korrekten Volumenstrom im Solekreis einstellen&bull;Umw&auml;lzpumpe im Solekreis austauschen"};
	$ErrorCodeList{"0352"} = {errorcode => "A01",                                    subcode => "5134",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: Steuerger&auml;t erh&auml;lt unzul&auml;ssige Werte vom Temperaturf&uuml;hler am Verfl&uuml;ssigerausgang",                                                                                                                                                                                                                   checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Temperaturf&uuml;hler TTI-3 komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel vom Temperaturf&uuml;hler TTI-3 im Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Temperaturf&uuml;hler TTI-3 besch&auml;digt ist&bull;Ohmschen Widerstand des Temperaturf&uuml;hlers TTI-3 &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-3 korrekten Kontakt zur K&auml;ltemittelleitung am Verfl&uuml;ssigerausgang hat&bull;-",correction => "Anschlussstecker am Temperaturf&uuml;hler TTI-3 komplett aufstecken&bull;Anschlusskabel vom Temperaturf&uuml;hler TTI-3 korrekt im Steuerger&auml;t (CU I/O) anschlie&szlig;en&bull;Kabelbaum autauschen&bull;Temperaturf&uuml;hler TTI-3 austauschen&bull;Temperaturf&uuml;hler TTI-3 an der korrekten Position der K&auml;ltemittelleitung am Verfl&uuml;ssigerausgang montieren&bull;Isolierung am Temperaturf&uuml;hler TTI-3 austauschen bzw. montieren"};
	$ErrorCodeList{"0353"} = {errorcode => "A01",                                    subcode => "5135",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: Steuerger&auml;t erh&auml;lt unzul&auml;ssige Werte vom Temperaturf&uuml;hler am Verfl&uuml;ssigerausgang (Heizwasser)",                                                                                                                                                                                                      checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Temperaturf&uuml;hler TTI-12 komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel vom Temperaturf&uuml;hler TTI-12 im Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Temperaturf&uuml;hler TTI-12 besch&auml;digt ist&bull;Ohmschen Widerstand des Temperaturf&uuml;hlers TTI-12 &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-12 korrekten Kontakt zur Heizwasserleitung am Verfl&uuml;ssigerausgang hat&bull;-",correction => "Anschlussstecker am Temperaturf&uuml;hler TTI-12 komplett aufstecken&bull;Anschlusskabel vom Temperaturf&uuml;hler TTI-12 korrekt im Steuerger&auml;t (CU I/O) anschlie&szlig;en&bull;Kabelbaum autauschen&bull;Temperaturf&uuml;hler TTI-12 austauschen&bull;Temperaturf&uuml;hler TTI-12 an der korrekten Position der Heizwasserleitung am Verfl&uuml;ssigerausgang montieren&bull;Isolierung am Temperaturf&uuml;hler TTI-12 austauschen bzw. montieren"};
	$ErrorCodeList{"0354"} = {errorcode => "A01",                                    subcode => "5136",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: Steuerger&auml;t erh&auml;lt unzul&auml;ssige Werte vom Temperaturf&uuml;hler am Verdampferausgang",                                                                                                                                                                                                                          checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Temperaturf&uuml;hler TTI-4 komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel vom Temperaturf&uuml;hler TTI-4 im Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Temperaturf&uuml;hler TTI-4 besch&auml;digt ist&bull;Ohmschen Widerstand des Temperaturf&uuml;hlers TTI-4 &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-12 korrekten Kontakt zur K&auml;ltemittelleitung am Verdampferausgang hat&bull;-",correction => "Anschlussstecker am Temperaturf&uuml;hler TTI-4 komplett aufstecken&bull;Anschlusskabel vom Temperaturf&uuml;hler TTI-4 korrekt im Steuerger&auml;t (CU I/O) anschlie&szlig;en&bull;Kabelbaum autauschen&bull;Temperaturf&uuml;hler TTI-4 austauschen&bull;Temperaturf&uuml;hler TTI-4 an der korrekten Position der K&auml;ltemittelleitung am Verdampferausgang montieren&bull;Isolierung am Temperaturf&uuml;hler TTI-4 austauschen bzw. montieren"};
	$ErrorCodeList{"0355"} = {errorcode => "A01",                                    subcode => "5137",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: Steuerger&auml;t erh&auml;lt unzul&auml;ssige Werte vom Temperaturf&uuml;hler am Austreiberfu&szlig;",                                                                                                                                                                                                                        checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Temperaturf&uuml;hler TTI-1 komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel vom Temperaturf&uuml;hler TTI-1 im Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Temperaturf&uuml;hler TTI-1 besch&auml;digt ist&bull;Ohmschen Widerstand des Temperaturf&uuml;hlers TTI-1 &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-1 korrekten Kontakt zum Generatorfu&szlig; hat&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-1 korrrekt isoliert ist&bull;Nennw&auml;rmebelastung am Brenner &uuml;berpr&uuml;fen&bull;Funktion der Umw&auml;lzpumpe im K&auml;ltemittelkreis (NH3) &uuml;berpr&uuml;fen",correction => "Anschlussstecker am Temperaturf&uuml;hler TTI-1 komplett aufstecken&bull;Anschlusskabel vom Temperaturf&uuml;hler TTI-1 korrekt im Steuerger&auml;t (CU I/O) anschlie&szlig;en&bull;Kabelbaum autauschen&bull;Temperaturf&uuml;hler TTI-1 austauschen&bull;Temperaturf&uuml;hler TTI-1 an der korrekten Position am Austreiberfu&szlig; montieren&bull;Isolierung am Temperaturf&uuml;hler T1I-4 austauschen bzw. montieren&bull;Brennereinstellwert f&uuml;r die Nennw&auml;rmebelastung korrigieren&bull;Werkskundendienst benachrichtigen"};
	$ErrorCodeList{"0356"} = {errorcode => "A01",                                    subcode => "5138",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: Steuerger&auml;t erh&auml;lt unzul&auml;ssige Werte vom Temperaturf&uuml;hler f&uuml;r Hei&szlig;gas",                                                                                                                                                                                                                        checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Temperaturf&uuml;hler TTI-2 komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel vom Temperaturf&uuml;hler TTI-2 im Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Temperaturf&uuml;hler TTI-2 besch&auml;digt ist&bull;Ohmschen Widerstand des Temperaturf&uuml;hlers TTI-2 &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-1 korrekten Kontakt zur K&auml;ltemittelleitung am Verfl&uuml;ssigerausgang hat&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-2 korrrekt isoliert ist&bull;Nennw&auml;rmebelastung am Brenner &uuml;berpr&uuml;fen&bull;Funktion der Umw&auml;lzpumpe im K&auml;ltemittelkreis (NH3) &uuml;berpr&uuml;fen",correction => "Anschlussstecker am Temperaturf&uuml;hler TTI-2 komplett aufstecken&bull;Anschlusskabel vom Temperaturf&uuml;hler TTI-2 korrekt im Steuerger&auml;t (CU I/O) anschlie&szlig;en&bull;Kabelbaum autauschen&bull;Temperaturf&uuml;hler TTI-2 austauschen&bull;Temperaturf&uuml;hler TTI-2 an der korrekten Position Der K&auml;ltemittelleitung am Verfl&uuml;ssigerausgang montieren&bull;Isolierung am Temperaturf&uuml;hler TT1-2 austauschen bzw. montieren&bull;Brennereinstellwert f&uuml;r die Nennw&auml;rmebelastung korrigieren&bull;Werkskundendienst benachrichtigen"};
	$ErrorCodeList{"0357"} = {errorcode => "A01",                                    subcode => "5139",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: Steuerger&auml;t erh&auml;lt unzul&auml;ssige Werte vom Niederdruckf&uuml;hler (K&auml;ltemittel)",                                                                                                                                                                                                                           checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Niederdruckf&uuml;hler PTI12 im K&auml;ltemittelkreis komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das  Anschlusskabel vom Niederdruckf&uuml;hler PTI12 am Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Niederdruckf&uuml;hler PTI12 im K&auml;ltemittelkreis besch&auml;digt ist&bull;&Uuml;berpr&uuml;fen, ob der Motor am 2 Wege Ventil korrekt montiert ist&bull;&Uuml;berpr&uuml;fen, ob der Motor am 2 Wege Ventil defekt ist&bull;Funktion des 2 Wege Ventiles vor dem Verfl&uuml;ssiger &uuml;berpr&uuml;fen&bull;-",correction => "Anschlussstecker am Niederdruckf&uuml;hler PTI12 im K&auml;ltemittelkreis komplett aufstecken&bull;Anschlusskabel vom Niederdruckf&uuml;hler PTI12 am Steuerger&auml;t (CU I/O) korrekt anschlie&szlig;en&bull;Kabelbaum austauschen&bull;Motor am 2 Wege Ventil korrekt auf das Ventilunterteil montieren&bull;Motor am 2 Wege Ventil austauschen&bull;2 Wege Ventil vor dem Verfl&uuml;ssiger austauschen&bull;Werkskundendienst benachrichtigen"};
	$ErrorCodeList{"0358"} = {errorcode => "A01",                                    subcode => "5140",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: Steuerger&auml;t erh&auml;lt unzul&auml;ssige Werte vom Hochdruckf&uuml;hler (K&auml;ltemittel)",                                                                                                                                                                                                                             checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Hochdruckf&uuml;hler PTI11 im K&auml;ltemittelkreis komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das  Anschlusskabel vom Hochdruckf&uuml;hler PTI11 am Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Motor am 2 -Wege Ventil  besch&auml;digt ist&bull;&Uuml;berpr&uuml;fen, ob der Motor am 2 Wege Ventil korrekt montiert ist&bull;&Uuml;berpr&uuml;fen, ob der Motor am 2 Wege Ventil defekt ist&bull;Funktion des 2 Wege Ventiles vor dem Verfl&uuml;ssiger &uuml;berpr&uuml;fen&bull;-",correction => "Anschlussstecker am Hochdruckf&uuml;hler PTI11 im K&auml;ltemittelkreis komplett aufstecken&bull;Anschlusskabel vom Hochdruckf&uuml;hler PTI11 am Steuerger&auml;t (CU I/O) korrekt anschlie&szlig;en&bull;Anschlusskabel zum Motor am 2 Wege Ventil austauschen&bull;Motor am 2 Wege Ventil korrekt auf das Ventilunterteil montieren&bull;Motor am 2 Wege Ventil austauschen&bull;2 Wege Ventil vor dem Verfl&uuml;ssiger austauschen&bull;Werkskundendienst benachrichtigen"};
	$ErrorCodeList{"0359"} = {errorcode => "A01",                                    subcode => "5141",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: Volumenstrom im Prim&auml;rkreis liegt au&szlig;erhalb des zul&auml;ssigen Bereichs",                                                                                                                                                                                                                                         checklist => "Stellung der Wartungsh&auml;hne zwischen W&auml;rmepumpe und Pufferspeicher &uuml;berpr&uuml;fen.&bull;&Uuml;berpr&uuml;fen, ob sich Luft in den Verbindungsleitungen zwischen W&auml;rmepumpe und Pufferspeicher befindet.&bull;Funktion der Umw&auml;lzpumpe &uuml;berpr&uuml;fen&bull;Volumenstrom &uuml;berpr&uuml;fen&bull;Volumenstrom &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Volumenstromsensor FTI18 aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel vom Volumenstromsensor FTI18 im Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;Volumenstrommesser FTI18 auf wasserseitige Verschmutzungen &uuml;berpr&uuml;fen&bull;-",correction => "Wartungsh&auml;hne zwischen W&auml;rmepumpe und Pufferspeicher &ouml;ffnen&bull;Verbindungsleitungen zwischen W&auml;rmepumpe und Pufferspeicher entl&uuml;ften&bull;Umw&auml;lzpumpe anwerfen ggf. austauschen&bull;Volumenstrom auf den zul&auml;ssigen Wert (7 100 l/ min) reduzieren&bull;Volumenstrom auf den zul&auml;ssigen Wert (7 100 l/ min) erh&ouml;hen&bull;Anschlussstecker am Volumenstromsensor FTI18 komplett aufstecken&bull;Anschlusskabel vom Volumenstromsensor FTI18 am Steuerger&auml;t (CU I/O) korrekt anschlie&szlig;en&bull;Ablagerungen/ Schmutz aus Volumenstromsensor FTI18 entfernen&bull;Volumenstromsensor FTI18 austauschen"};
	$ErrorCodeList{"0360"} = {errorcode => "A01",                                    subcode => "5142",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: Volumenstrom im Solekreis liegt au&szlig;erhalb des zul&auml;ssigen Bereichs",                                                                                                                                                                                                                                                checklist => "Stellung der Wartungsh&auml;hne im Solekreis &uuml;berpr&uuml;fen.&bull;Filter im Solekreis auf Verschmutzung &uuml;berpr&uuml;fen.&bull;Funktion der Umw&auml;lzpumpe  im Solekreis &uuml;berpr&uuml;fen&bull;Volumenstrom im Solekreis &uuml;berpr&uuml;fen&bull;Volumenstrom im Solekreis &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Volumenstromsensor FTI19 im Solekreis aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel vom Volumenstromsensor FTI19 im Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;Volumenstrommesser FTI19 im Solekreis auf wasserseitige Verschmutzungen &uuml;berpr&uuml;fen&bull;-",correction => "Wartungsh&auml;hne im Solekreis &ouml;ffnen&bull;Filter im Solekreis reinigen&bull;Umw&auml;lzpumpe anwerfen ggf. austauschen&bull;Volumenstrom auf den zul&auml;ssigen Wert im Solekreis  (12 100 l/min) reduzieren&bull;Volumenstrom auf den zul&auml;ssigen Wert im Solrkeis  (12 100 l/min) erh&ouml;hen&bull;Anschlussstecker am Volumenstromsensor FTI19 im Solekreis komplett aufstecken&bull;Anschlusskabel vom Volumenstromsensor FTI19 am Steuerger&auml;t (CU I/O) korrekt anschlie&szlig;en&bull;Ablagerungen/ Schmutz aus Volumenstromsensor FTI19 im Solekreis entfernen&bull;Volumenstromsensor FTI19 im Solekreisaustauschen"};
	$ErrorCodeList{"0361"} = {errorcode => "A01",                                    subcode => "5143",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: Unzul&auml;ssige Lufteintrittstemperatur an der Au&szlig;eneinheit",                                                                                                                                                                                                                                                          checklist => "&Uuml;berpr&uuml;fen, ob der Lufteintrittstemperaturf&uuml;hler an der Au&szlig;eneinheit verschmutzt ist&bull;&Uuml;berpr&uuml;fen, ob der Anschlusstecker am Lufteintrittstemperaturf&uuml;hler in der Au&szlig;eneinheit aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Lufteintrittstemperaturf&uuml;hler in der Au&szlig;eneinheit besch&auml;digt ist&bull;-",                                                                                                            correction => "Ablagerungen amLufteintrittstemperaturf&uuml;hler in der Au&szlig;eneinheit entfernen&bull;Anschlussstecker am Lufteintrittstemperaturf&uuml;hler in der Au&szlig;eneinheit aufstecken&bull;Anschlusskabel zum Lufteintrittstemperaturf&uuml;hler in der Au&szlig;eneinheit austauschen&bull;Lufteintrittstemperaturf&uuml;hler in der Au&szlig;eneinheit austauschen"                           };
	$ErrorCodeList{"0362"} = {errorcode => "A01",                                    subcode => "5144",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: Unzul&auml;ssige Luftaustrittstemperatur an der Au&szlig;eneinheit",                                                                                                                                                                                                                                                          checklist => "&Uuml;berpr&uuml;fen, ob der Luftaustrittstemperaturf&uuml;hler an der Au&szlig;eneinheit verschmutzt ist&bull;&Uuml;berpr&uuml;fen, ob der Anschlusstecker am Luftaustrittstemperaturf&uuml;hler in der Au&szlig;eneinheit aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Luftaustrittstemperaturf&uuml;hler in der Au&szlig;eneinheit besch&auml;digt ist&bull;-",                                                                                                            correction => "Ablagerungen amLuftaustrittstemperaturf&uuml;hler in der Au&szlig;eneinheit entfernen&bull;Anschlussstecker am Luftaustrittstemperaturf&uuml;hler in der Au&szlig;eneinheit aufstecken&bull;Anschlusskabel zum Luftaustrittstemperaturf&uuml;hler in der Au&szlig;eneinheit austauschen&bull;Luftaustrittstemperaturf&uuml;hler in der Au&szlig;eneinheit austauschen"                           };
	$ErrorCodeList{"0363"} = {errorcode => "A01",                                    subcode => "5150",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: Temperaturdifferenz im Solekreis liegt au&szlig;erhalb des zul&auml;ssigen Bereichs",                                                                                                                                                                                                                                         checklist => "Filter im Solekreis reinigen&bull;Volumenstrom im Solekreis &uuml;berpr&uuml;fen&bull;Volumenstrom im Solekreis &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-7 korrekten Kontakt zur Solevorlaufleitung hat&bull;&Uuml;berpr&uuml;fen, ob die Isolierung am Temperaturf&uuml;hler TTI-7 am Solevorlauf vorhanden bzw besch&auml;digt ist&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-8 korrekten Kontakt zur Soler&uuml;cklaufleitung hat.&bull;&Uuml;berpr&uuml;fen, ob ob die Isolierung am Temperaturf&uuml;hler TTI-8 am Soler&uuml;cklauf vorhanden  bzw. besch&auml;digt ist&bull;Entnahmeleistung der Erdsonde &uuml;berpr&uuml;fen&bull;-",correction => "Filter im Solekreis reinigen&bull;Volumenstrom im Solekreis auf den zul&auml;ssigen Wert reduzieren (12 100 l/min)&bull;Volumenstrom im Solekreis auf den zul&auml;ssigen Wert erh&ouml;hen (12 100 l/min)&bull;Temperaturf&uuml;hler TTI-7 an der korrekten Position der Solevorlaufleitung montieren&bull;Isolierung am Temperaturf&uuml;hler TTI-7 austauschen bzw. montieren&bull;Temperaturf&uuml;hler TTI-8 an der korrekten Position der Soler&uuml;cklaufleitung montieren&bull;Isolierung am Temperaturf&uuml;hler TTI-8 austauschen bzw. montieren&bull;Erdsonde &uuml;berpr&uuml;fen&bull;Werkskundendienst beauftragen"};
	$ErrorCodeList{"0364"} = {errorcode => "A01",                                    subcode => "5151",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: Temperaturdifferenz zwischen Sole und K&auml;ltemittelaustritt nach Verdampfer  ist au&szlig;erhalb des zul&auml;ssigen Bereiches",                                                                                                                                                                                           checklist => "&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-7 korrekten Kontakt zur Solevorlaufleitung hat&bull;&Uuml;berpr&uuml;fen, ob die Isolierung am Temperaturf&uuml;hler TTI-7 am Solevorlauf vorhanden bzw besch&auml;digt ist.&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-4 korrekten Kontakt zur K&auml;ltemittelleitung am Verdampferausgang hat.&bull;&Uuml;berpr&uuml;fen, ob die Isolierung am Temperaturf&uuml;hler TTI-4 vorhanden bzw. besch&auml;digt ist.&bull;Volumenstrom im Solekreis &uuml;berpr&uuml;fen",correction => "Temperaturf&uuml;hler TTI-7 an der korrekten Position der Solevorlaufleitung montieren&bull;Isolierung am Temperaturf&uuml;hler TTI-7 austauschen bzw. montieren&bull;Temperaturf&uuml;hler TTI-4 an der korrekten Position der K&auml;ltemittelleitung am Verdampferausgang montieren&bull;Isolierung am Temperaturf&uuml;hler TTI-4 austauschen bzw. montieren&bull;Volumenstrom im Solekreis auf die zul&auml;ssigen Werte (12 -100 l/min) erh&ouml;hen"};
	$ErrorCodeList{"0365"} = {errorcode => "A01",                                    subcode => "5152",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: Temperaturdifferenz zwischen Verl&uuml;ssigeraustritt K&auml;ltekreis und Heizungsr&uuml;cklauf ist au&szlig;erhalb des zul&auml;ssigen Bereiches",                                                                                                                                                                           checklist => "&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-3 korrekten Kontakt zur K&auml;ltemittelleitung am Verfl&uuml;ssigerausgang hat&bull;&Uuml;berpr&uuml;fen, ob die Isolierung am Temperaturf&uuml;hler TTI-3 vorhanden bzw. besch&auml;digt ist&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-5 korrekten Kontakt an der R&uuml;ckllaufleitung hat.&bull;Position des R&uuml;cklauftemperaturf&uuml;hlers TTI-5 an der Vorlaufleitung &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob die Isolierung am R&uuml;cklauftemperaturf&uuml;hler TTI-5 vorhanden bzw. besch&auml;digt ist",correction => "Temperaturf&uuml;hler TTI-3 an der korrekten Position der K&auml;ltemittelleitung am Verfl&uuml;ssigerausgang montieren&bull;Isolierung am Temperaturf&uuml;hler TTI-3 austauschen bzw. montieren&bull;Kontakt zwischen R&uuml;cklauftemperaturf&uuml;hler TTI-5 und R&uuml;cklaufleitung herstellen&bull;R&uuml;cklauftemperaturf&uuml;hler TTI-5 an der korrekten Position der R&uuml;cklaufleitung montieren&bull;Isolierung am R&uuml;cklauftemperaturf&uuml;hler TTI-5 austauschen bzw. montieren"};
	$ErrorCodeList{"0366"} = {errorcode => "A01",                                    subcode => "5153",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: Druckdifferenz im K&auml;ltemittelkreis ist au&szlig;erhalb des zul&auml;ssigen Bereichs",                                                                                                                                                                                                                                    checklist => "Funktion der L&ouml;sungspumpe im K&auml;ltemittelkreis (NH3) &uuml;berpr&uuml;fen&bull;Funktion der Regelventile im K&auml;ltemittelkreis (NH3) &uuml;berpr&uuml;fen&bull;-",                                                                                                                                                                                                                                                                                                                        correction => "Werkskundendienst benachrichtigen&bull;Werkskundendienst benachrichtigen&bull;Werkskundendienst benachrichtigen"                                                                                                                                                                                                                                                                                 };
	$ErrorCodeList{"0367"} = {errorcode => "A01",                                    subcode => "5161",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: Fehler im Frequenzumrichter f&uuml;r Ansteuerung L&ouml;sungspumpe",                                                                                                                                                                                                                                                          checklist => "Motor an der L&ouml;sungspumpe im K&auml;ltemittelkreis &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob alle Anschlusskabel im Frequenzumrichter aufgesteckt sind&bull;-",                                                                                                                                                                                                                                                                                                                         correction => "Motor an der L&ouml;sungspumpe im K&auml;ltemittelkreis austauschen&bull;Anschlusskabel im Frequenzumrichter aufstecken&bull;Frequenzumrichter austauschen"                                                                                                                                                                                                                                      };
	$ErrorCodeList{"0368"} = {errorcode => "A01",                                    subcode => "5162",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: Drehzahl vom Abluftventilator liegt au&szlig;erhalb des zul&auml;ssigen Bereichs",                                                                                                                                                                                                                                            checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker f&uuml;r die Drehzahlregelung am Abluftventilator komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel f&uuml;r die Drehzahlregelung am Abluftventilator besch&auml;digt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel f&uuml;r die Spannungsversorgung des Abluftventilators besch&auml;digt ist&bull;Funktion des Abluftventilators &uuml;berpr&uuml;fen&bull;-",                                                         correction => "Anschlussstecker f&uuml;r die Drehzahlregelung am Abluftventilator komplett aufstecken&bull;Anschlusskabel f&uuml;r die Drehzahlregelung des Abluftventilators austauschen&bull;Anschlusskabel f&uuml;r die Spannungsversorgung des Abluftventilators austauschen&bull;Abluftventilator austauschen&bull;Vorwiderstand f&uuml;r die Drehzahlbegrenzung des Abluftventilators austauschen"        };
	$ErrorCodeList{"0369"} = {errorcode => "A01",                                    subcode => "5163",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: Gebl&auml;sedrehzahl au&szlig;erhalb des zul&auml;ssigen Bereiches",                                                                                                                                                                                                                                                          checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker f&uuml;r die Drehzahlregelung am Gebl&auml;se  komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel f&uuml;r die Drehzahlregelung am Gebl&auml;se besch&auml;digt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel f&uuml;r die Spannungsversorgung des Gebl&auml;ses besch&auml;digt ist&bull;Funktion des Gebl&auml;sese &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Anschlussstecker f&uuml;r die Spannungsversorgung (230 VAC) am Gebl&auml;se komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob der Anschlussstecker f&uuml;r die Spannungsversorgung (230 VAC) des Gebl&auml;sese im Feuerungsautomat komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel f&uuml;r die Spannungsversorgung (230 VAC) zum Gebl&auml;se besch&auml;digt ist&bull;-",correction => "Anschlussstecker f&uuml;r die Drehzahlregelung am Gebl&auml;se komplett aufstecken&bull;Anschlusskabel f&uuml;r die Drehzahlregelung des Gebl&auml;ses austauschen&bull;Anschlusskabel f&uuml;r die Spannungsversorgung des Gebl&auml;ses austauschen&bull;Gebl&auml;se austauschen&bull;Anschlussstecker f&uuml;r die Spannungsversorgung (230 VAC) am Gebl&auml;se komplett aufstecken&bull;Anschlussstecker f&uuml;r die Spannungsversorgung (230 VAC) zum Gebl&auml;se im Feuerungsautomat komplett aufstecken&bull;Anschlusskabel f&uuml;r die Spannungsversorgung (230 VAC) zum Gebl&auml;se austauschen&bull;Feuerungsautomat austauschen"};
	$ErrorCodeList{"0370"} = {errorcode => "A01",                                    subcode => "5164",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: Fehler 2-Wege-Ventil am Verfl&uuml;ssiger (Heizungswasser)",                                                                                                                                                                                                                                                                  checklist => "&Uuml;berpr&uuml;fen Sie die Position des Motors am 2-Wege-Ventil&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel f&uuml;r das R&uuml;ckmeldesignal am Motor f&uuml;r das 2-Wege-Ventil besch&auml;digt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel f&uuml;r die Spannungsversorgung des Motors am 2-Wege-Ventil besch&auml;digt ist&bull;&Uuml;berpr&uuml;fen, ob der Motor vom 2-Wege-Ventil blockiert ist&bull;-",                                                                       correction => "Position des Motors am 2-Wege-Ventil korrigieren&bull;Anschlusskabel f&uuml;r das R&uuml;ckmeldesignal am Motor des 2-Wege-Ventiles austauschen&bull;Anschlusskabel f&uuml;r die Spannungsversorgung des Abluftventilators austauschen&bull;Motor am 2-Wege-Ventil austauschen&bull;2-Wege-Ventil austauschen"                                                                                   };
	$ErrorCodeList{"0371"} = {errorcode => "A01",                                    subcode => "5170",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: F&uuml;llstand des K&auml;ltemittels (NH3) im Austreiber zu niedrig",                                                                                                                                                                                                                                                         checklist => "Anschlusskabel zum F&uuml;llstandssensor &uuml;berpr&uuml;fen&bull;F&uuml;llstandssensor f&uuml;r das K&auml;ltemittel im Austreiber &uuml;berpr&uuml;fen&bull;Anlagenhydraulik &uuml;berpr&uuml;fen&bull;Fremdgas aus K&auml;ltemittelkreis (NH3) entfernen&bull;Funktion der Regelventile im K&auml;ltemittelkreis (NH3) &uuml;berpr&uuml;fen&bull;-",                                                                                                                                              correction => "Anschlusskabel zum F&uuml;llstandssensor austauschen&bull;Werkskundendienst benachrichtigen&bull;Anlagenhydraulik korrigieren&bull;Werkskundendienst benachrichtigen&bull;Werkskundendienst benachrichtigen&bull;Werkskundendienst benachrichtigen"                                                                                                                                              };
	$ErrorCodeList{"0372"} = {errorcode => "A01",                                    subcode => "5171",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: R&uuml;ckstau Abgaskondensat im Abgasw&auml;rmetauscher",                                                                                                                                                                                                                                                                     checklist => "&Uuml;berpr&uuml;fen, ob der Kondensatanschlu&szlig; am Abgasw&auml;rmetauscher durch Ablagerungen blockiert ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Sensor FLS besch&auml;digt ist&bull;Funktion des Sensors FLS &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob die Kondensatleitung mit Gef&auml;lle verlegt worden ist&bull;-",                                                                                                                                               correction => "R&uuml;ckst&auml;nde am Kondensatanschlu&szlig; im Abgasw&auml;rmetauscher entfernn&bull;Anschlusskabel zum Sensor FLS austauschen&bull;Sensor FLS am Abgasw&auml;rmetauscher austauschen&bull;Kondensatleitung mit Gef&auml;lle verlegen&bull;Externen Kondensatsiphon reinigen"                                                                                                                };
	$ErrorCodeList{"0373"} = {errorcode => "A01",                                    subcode => "5172",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: Aufheizzeit imK&auml;ltekreis nach Start zu lang",                                                                                                                                                                                                                                                                            checklist => "Ohmschen Widerstand des R&uuml;cklauftemperaturf&uuml;hlers TTI-5 messen&bull;&Uuml;berpr&uuml;fen, ob der R&uuml;cklauftemperaturf&uuml;hler; TTI-5 korrekten Kontakt an der Vorlaufleitung hat.&bull;Position des R&uuml;cklauftemperaturf&uuml;hlers TTI-5 an der Vorlaufleitung &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob die Isolierung am R&uuml;cklauftemperaturf&uuml;hler TTI-5 vorhanden bzw. besch&auml;digt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum R&uuml;cklauftemperaturf&uuml;hler TTI-5 besch&auml;digt ist&bull;Ohmschen Widerstand des Temperaturf&uuml;hlers TTI-6 messen&bull;&Uuml;berpr&uuml;fen, ob der Vorlauftemperaturf&uuml;hler TTI-6 korrekten Kontakt an der Vorlaufleitung hat&bull;Position des Vorlauftemperaturf&uuml;hlers TTI-6 an der Vorlaufleitung &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob die Isolierung am Vorlauftemperaturf&uuml;hler TTI-6 vorhanden bzw. besch&auml;digt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Temperaturf&uuml;hler TTI-6 besch&auml;digt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Temperaturf&uuml;hler TTI-12 besch&auml;digt ist&bull;Ohmschen Widerstand des Temperaturf&uuml;hlers TTI-12 &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-12 korrekten Kontakt zur Heizwasserleitung am Verfl&uuml;ssigerausgang hat&bull;&Uuml;berpr&uuml;fen, ob die Isolierung am Temperaturf&uuml;hler TTI-12 vorhanden bzw. besch&auml;digt ist&bull;Brennereinstellung bei minimaler Nennw&auml;rmebelastung &uuml;berpr&uuml;fen",correction => "R&uuml;cklauftemperaturf&uuml;hler TTI-5 austauschen&bull;Kontakt zwischen R&uuml;cklauftemperaturf&uuml;hler TTI-5 und R&uuml;cklaufleitung herstellen&bull;R&uuml;cklauftemperaturf&uuml;hler TTI-5 an der korrekten Position der R&uuml;cklaufleitung montieren&bull;Isolierung am R&uuml;cklauftemperaturf&uuml;hler TTI-5 austauschen bzw. montieren&bull;Kabelbaum austauschen&bull;Vorlauftemperaturf&uuml;hler TTI-6 austauschen&bull;Kontakt zwischen Vorlauftemperaturf&uuml;hler TTI-6 und Vorlaufleitung herstellen&bull;Vorlauftemperaturf&uuml;hler TTI-6 an der korrekten Position der Vorlaufleitung montieren&bull;Isolierung am Vorlauftemperaturf&uuml;hler TTI-6 austauschen bzw. montieren&bull;Kabelbaum austauschen&bull;Kabelbaum autauschen&bull;Temperaturf&uuml;hler TTI-12 austauschen&bull;Temperaturf&uuml;hler TTI-12 an der korrekten Position der Heizwasserleitung am Verfl&uuml;ssigerausgang montieren&bull;Isolierung am Temperaturf&uuml;hler TTI-12 austauschen bzw. montieren&bull;Brennereinstellung bei minimaler Nennw&auml;rmebelastung korrigieren"};
	$ErrorCodeList{"0374"} = {errorcode => "A01",                                    subcode => "5172",                                               class => "4",           reason => "",                                                                                                                                                                                                                                                                                                                                                 checklist => "Brennereinstellung  bei maximaler Nennw&auml;rmebelastung &uuml;berpr&uuml;fen&bull;Funktion der Regelventile im K&auml;ltemittelkreis &uuml;berp&uuml;fen&bull;-",                                                                                                                                                                                                                                                                                                                                   correction => "Brennereinstellung bei maximaler Nennw&auml;rmebelastung korrigieren&bull;Werkskundendienst benachrichtigen&bull;Werkskundendienst benachrichtigen&bull;Brennereinstellung bei maximaler Nennw&auml;rmebelastung korrigieren&bull;Werkskundendienst benachrichtigen"                                                                                                                             };
	$ErrorCodeList{"0375"} = {errorcode => "A01",                                    subcode => "5201",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: Warnung Aussentemperaturf&uuml;hler T1 Kabelbruch",                                                                                                                                                                                                                                                                           checklist => "Anschluss Au&szlig;entemperaturf&uuml;hler pr&uuml;fen Au&szlig;entemperatuf&uuml;hler liegt au&szlig;erhalb des Messbereichs (Ohm-Werte pr&uuml;fen, s.u. Werte in Installationsanleitung)",                                                                                                                                                                                                                                                                                                         correction => "Au&szlig;entemperaturf&uuml;hler montieren Au&szlig;entemperaturf&uuml;hler tauschen Installationsmodul tauschen"                                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"0376"} = {errorcode => "A01",                                    subcode => "5202",                                               class => "4",           reason => "Gas-W&auml;rmepumpe: Warnung Aussentemperaturf&uuml;hler T1 Kurzschluss",                                                                                                                                                                                                                                                                          checklist => "Anschlusskabel und Steckanschluss pr&uuml;fen (F&uuml;hlerwerte-Tabelle in Installationsanleitung)",                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Au&szlig;entemperaturf&uuml;hler tauschen Kabel erneuern Installationsmodul tauschen"                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"0377"} = {errorcode => "A01",                                    subcode => "5203",                                               class => "WPL AR",      reason => "Alarm Au&szlig;entemperaturf&uuml;hler T1 defekt",                                                                                                                                                                                                                                                                                                 checklist => "Siehe Pr&uuml;fvorgang/Ursache f&uuml;r Ursachencode 5201, 5202",                                                                                                                                                                                                                                                                                                                                                                                                                                     correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5201, 5202"                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"0378"} = {errorcode => "A01",                                    subcode => "5204",                                               class => "WPL AR",      reason => "Warnung Vorlauff&uuml;hler T0 Unterbrechung",                                                                                                                                                                                                                                                                                                      checklist => "Vorlauff&uuml;hler T0/Signalkabel Unterbrechung&bull;Schraubklemme an I/O-Board zu Vorlauff&uuml;hler T0 ej angezogen&bull;Vorlauff&uuml;hler T0/Signalkabel Unterbrechung&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                                      correction => "Ohmwert von F&uuml;hler T0/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen; das Innenteil ist dabei von I/O-Board-Klemme getrennt&bull;Schraubklemme &uuml;berpr&uuml;fen&bull;Vorlauff&uuml;hler T0 wechseln&bull;I/O-Board wechseln"                                                                                                                            };
	$ErrorCodeList{"0379"} = {errorcode => "A01",                                    subcode => "5205",                                               class => "WPL AR",      reason => "Warnung Vorlauff&uuml;hler T0 kurzgeschlossen",                                                                                                                                                                                                                                                                                                    checklist => "Vorlauff&uuml;hler T0/Signalkabel kurzgeschlossen&bull;Vorlauff&uuml;hler T0 kurzgeschlossen&bull;Signalkabel zu Vorlauff&uuml;hler T0 kurzgeschlossen&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                                                          correction => "Ohmwert von F&uuml;hler T0/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen; das Innenteil ist dabei von I/O-Board-Klemme getrennt&bull;Ohmwert im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen/Au&szlig;entemperatursensor T0 wechseln&bull;Signalkabel zu Vorlauff&uuml;hler T0 reparieren/wechseln&bull;I/O-Board wechseln"                  };
	$ErrorCodeList{"0380"} = {errorcode => "A01",                                    subcode => "5207",                                               class => "4",           reason => "Warnung Temperaturf&uuml;hler TC1 Kabelbruch",                                                                                                                                                                                                                                                                                                     checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Temperaturf&uuml;hler TC1 komplett aufgesteckt ist&bull;Ohmschen Widerstand des Temperaturf&uuml;hler TC1 messen&bull;-",                                                                                                                                                                                                                                                                                                                            correction => "Anschlussstecker am Temperaturf&uuml;hler TC1 komplett aufstecken.&bull;Temperaturf&uuml;hler TC 1 austauschen&bull;Anschlusskabel zum Temperatuf&uuml;hler TC1 austauschen"                                                                                                                                                                                                                     };
	$ErrorCodeList{"0381"} = {errorcode => "A01",                                    subcode => "5208",                                               class => "8",           reason => "Warnung Temperaturf&uuml;hler TC1 Kurzschluss",                                                                                                                                                                                                                                                                                                    checklist => "Ohmschen Widerstand des Temperaturf&uuml;hlers TC1 messen&bull;-",                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "Temperaturf&uuml;hler TC1 austauschen&bull;Anschlusskabel zum Temperatuf&uuml;hler TC1 austauschen"                                                                                                                                                                                                                                                                                              };
	$ErrorCodeList{"0382"} = {errorcode => "A01",                                    subcode => "5210",                                               class => "WPL AR",      reason => "Warnung Z2; W&auml;rmetr&auml;ger in F&uuml;hler TC1 Unterbrechung",                                                                                                                                                                                                                                                                               checklist => "W&auml;rmetr&auml;ger in F&uuml;hler TC1 W&auml;rmepumpe 2/Signalkabel Unterbrechung&bull;Schraubklemme an I/O-Board zu Vorlauff&uuml;hler TC1 W&auml;rmepumpe 2 nicht angezogen&bull;W&auml;rmetr&auml;ger in F&uuml;hler TC1 W&auml;rmepumpe 2/ Signalkabel Unterbrechung&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                     correction => "Ohmwert von F&uuml;hler TC1 W&auml;rmepumpe 2/ Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen. Das Innenteil ist dabei von I/O-Board-Klemme getrennt&bull;Schraubklemme &uuml;berpr&uuml;fen&bull;W&auml;rmetr&auml;ger in F&uuml;hler TC1 W&auml;rmepumpe 2 wechseln&bull;I/O-Board wechseln"                                                                   };
	$ErrorCodeList{"0383"} = {errorcode => "A01",                                    subcode => "5211",                                               class => "WPL AR",      reason => "Warnung Z2 W&auml;rmetr&auml;ger in F&uuml;hler TC1 kurzgschlossen",                                                                                                                                                                                                                                                                               checklist => "W&auml;rmetr&auml;ger in F&uuml;hler TC1 W&auml;rmepumpe 2/ Signalkabel kurzgeschlossen&bull;W&auml;rmetr&auml;ger in F&uuml;hler TC1 W&auml;rmepumpe 2 kurzgeschlossen&bull;Signalkabel zu Vorlauff&uuml;hler T0 kurzgeschlossen&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                               correction => "Ohmwert von F&uuml;hler TC1 W&auml;rmepumpe 2/ Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen. Das Innenteil ist dabei von I/O-Board-Klemme getrennt.&bull;Ohmwert im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen/W&auml;rmetr&auml;ger in F&uuml;hler TC1 W&auml;rmepumpe 2 wechseln&bull;Signalkabel zu W&auml;rmetr&auml;ger in F&uuml;hler TC1 reparieren/wechseln&bull;I/O-Board wechseln"};
	$ErrorCodeList{"0384"} = {errorcode => "A01",                                    subcode => "5213",                                               class => "WPL AR",      reason => "Warnung Z1; W&auml;rmetr&auml;ger in F&uuml;hler TC0 Unterbrechung",                                                                                                                                                                                                                                                                               checklist => "W&auml;rmetr&auml;ger in F&uuml;hler TC0/Signalkabel Unterbrechung&bull;Schraubklemme an I/O-Board zu Vorlauff&uuml;hler TC0 ej angezogen&bull;Vorlauff&uuml;hler TC0/Signalkabel Unterbrechung&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                 correction => "Ohmwert von F&uuml;hler TC0/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen; das Innenteil ist dabei von I/O-Board-Klemme getrennt&bull;Schraubklemme &uuml;berpr&uuml;fen&bull;Vorlauff&uuml;hler TC0 wechseln&bull;I/O-Board wechseln"                                                                                                                          };
	$ErrorCodeList{"0385"} = {errorcode => "A01",                                    subcode => "5214",                                               class => "WPL AR",      reason => "Warnung Z1; W&auml;rmetr&auml;ger in F&uuml;hler TC0 kurzgschlossen",                                                                                                                                                                                                                                                                              checklist => "W&auml;rmetr&auml;ger in F&uuml;hler TC0/Signalkabel kurzgeschlossen&bull;W&auml;rmetr&auml;ger in F&uuml;hler TC0 kurzgeschlossen&bull;Signalkabel zu W&auml;rmetr&auml;ger aus F&uuml;hler TC0 kurzgeschlossen&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                correction => "Ohmwert von F&uuml;hler TC0/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen; das Innenteil ist dabei von I/O-Board-Klemme getrennt&bull;Ohmwert im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen/W&auml;rmetr&auml;ger in F&uuml;hler TC0 wechseln&bull;Signalkabel zu W&auml;rmetr&auml;ger aus F&uuml;hler TC0 reparieren/wechseln&bull;I/O-Board wechseln"};
	$ErrorCodeList{"0386"} = {errorcode => "A01",                                    subcode => "5216",                                               class => "WPL AR",      reason => "Warnung Z2 W&auml;rmetr&auml;ger aus F&uuml;hler TC0 Unterbrechung",                                                                                                                                                                                                                                                                               checklist => "W&auml;rmetr&auml;ger aus F&uuml;hler TC0 W&auml;rmepumpe 2/ Signalkabel Unterbrechung&bull;Schraubklemme an I/O-Board zu Vorlauff&uuml;hler TC0 W&auml;rmepumpe 2 nicht angezogen&bull;Vorlauff&uuml;hler TC0/W&auml;rmepumpe 2 Signalkabel Unterbrechung&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                      correction => "Ohmwert von F&uuml;hler TC0 W&auml;rmepumpe 2/ Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen. Das Innenteil ist dabei von I/O-Board-Klemme getrennt.&bull;Schraubklemme &uuml;berpr&uuml;fen&bull;Vorlauff&uuml;hler TC1 W&auml;rmepumpe 2 wechseln&bull;I/O-Board wechseln"                                                                                    };
	$ErrorCodeList{"0387"} = {errorcode => "A01",                                    subcode => "5217",                                               class => "WPL AR",      reason => "Warnung Z2 W&auml;rmetr&auml;ger aus F&uuml;hler TC0 kurzgschlossen",                                                                                                                                                                                                                                                                              checklist => "W&auml;rmetr&auml;ger inaus F&uuml;hler TC0 W&auml;rmepumpe 2/Signalkabel kurzgeschlossen&bull;W&auml;rmetr&auml;ger in F&uuml;hler TC0 W&auml;rmepumpe 2 kurzgeschlossen&bull;Signalkabel zu W&auml;rmetr&auml;ger aus F&uuml;hler TC0 W&auml;rmepumpe 2 kurzgeschlossen&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                       correction => "Ohmwert von F&uuml;hler TC0 W&auml;rmepumpe 2/ Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen. Das Innenteil ist dabei  von I/O-Board-Klemme getrennt.&bull;Ohmwert im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen/W&auml;rmetr&auml;ger aus F&uuml;hler TC0 W&auml;rmepumpe 2 wechseln&bull;Signalkabel zu W&auml;rmetr&auml;ger aus F&uuml;hler TC0 W&auml;rmepumpe 2 reparieren/wechseln&bull;I/O-Board wechseln"};
	$ErrorCodeList{"0388"} = {errorcode => "A01",                                    subcode => "5220",                                               class => "8",           reason => "Warnung Z1 Sole Eintrittstemperaturf&uuml;hler TB0 Kurzschluss",                                                                                                                                                                                                                                                                                   checklist => "&Uuml;berpr&uuml;fen, ob das Anschlusskabel vom Temperaturf&uuml;hler TC3 im Steuerger&auml;t korrekt angeschlossen ist&bull;-",                                                                                                                                                                                                                                                                                                                                                                      correction => "Anschlusskabel vom Temperaturf&uuml;hler TC3 im Steuerger&auml;t korrekt anschlie&szlig;en&bull;Temperaturf&uuml;hler TC3 austauschen"                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"0389"} = {errorcode => "A01",                                    subcode => "5222",                                               class => "8",           reason => "Warnung Z2 Sole Eintrittstemperaturf&uuml;hler TB0 Kabelbruch",                                                                                                                                                                                                                                                                                    checklist => "F&uuml;hler TB0 (Sole-Eintrittstemperaturf&uuml;hler) pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Sole-Eintrittstemperaturf&uuml;hler tauschen Installationsmodul tauschen"                                                                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"0390"} = {errorcode => "A01",                                    subcode => "5234",                                               class => "WPL AR",      reason => "Warnung Schwimmbadtemperaturf&uuml;hler TP1 Unterbrechung",                                                                                                                                                                                                                                                                                        checklist => "Schwimmbadtemperaturf&uuml;hler TP1/ Signalkabel Unterbrechung&bull;Schraubklemme an I/O-Board zu Schwimmbadtemperaturf&uuml;hler TP1 nicht angezogen&bull;Schwimmbadtemperaturf&uuml;hler TP1/ Signalkabel Unterbrechung&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                       correction => "Ohmwert von F&uuml;hler TP1/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen; das Innenteil ist dabei von I/O-Board-Klemme getrennt&bull;Schraubklemme &uuml;berpr&uuml;fen&bull;Schwimmbadtemperaturf&uuml;hler TP1 wechseln&bull;I/O-Board wechseln"                                                                                                             };
	$ErrorCodeList{"0391"} = {errorcode => "A01",                                    subcode => "5235",                                               class => "WPL AR",      reason => "Warnung Schwimmbadtemperaturf&uuml;hler TP1 kurzgeschlossen",                                                                                                                                                                                                                                                                                      checklist => "Schwimmbadtemperaturf&uuml;hler TP1/ Signalkabel Unterbrechung&bull;Schwimmbadtemperatur TP1 kurzgeschlossen&bull;Signalkabel zu W&auml;rmetr&auml;ger aus Schwimmbadtemperaturf&uuml;hler TP1 kurzgeschlossen&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                  correction => "Ohmwert von Schwimmbadtemperaturf&uuml;hler TC1/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen; Das Innenteil ist dabei von I/O-BoardKlemme getrennt&bull;Ohmwert im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen/Schwimmbadtemperaturf&uuml;hler TP1 wechseln&bull;Signalkabel zu Schwimmbadtemperaturf&uuml;hler TP1 reparieren/wechseln&bull;I/O-Board wechseln"};
	$ErrorCodeList{"0392"} = {errorcode => "A01",                                    subcode => "5237",                                               class => "4",           reason => "Warnung Warmwasserf&uuml;hler TW1 Kabelbruch",                                                                                                                                                                                                                                                                                                     checklist => "&Uuml;berpr&uuml;fen, ob das Anschlusskabel vom Warmwassertemperaturf&uuml;hler TW1 im Steuerger&auml;t korrekt angeschlossen ist&bull;Ohmschen Widerstand des Warmwassertemperaturf&uuml;hlers TW1 messen&bull;-",                                                                                                                                                                                                                                                                                   correction => "Anschlusskabel vom Warmwassertemperaturf&uuml;hler TW1 im Steuerger&auml;t korrekt anschlie&szlig;en&bull;Warmwassertemperaturf&uuml;hler TW1 austauschen&bull;Anschlusskabel zum Warmwassertemperaturf&uuml;hler TW1 austauschen"                                                                                                                                                               };
	$ErrorCodeList{"0393"} = {errorcode => "A01",                                    subcode => "5238",                                               class => "4",           reason => "Warnung Warmwasserf&uuml;hler TW1 Kurzschluss",                                                                                                                                                                                                                                                                                                    checklist => "Ohmschen Widerstand des Warmwassertemperaturf&uuml;hlers TW1 messen&bull;-",                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Warmwassertemperaturf&uuml;hler TW1 austauschen&bull;Anschlusskabel zum Warmwassertemperaturf&uuml;hler TW1 austauschen"                                                                                                                                                                                                                                                                         };
	$ErrorCodeList{"0394"} = {errorcode => "A01",                                    subcode => "5246",                                               class => "WPL AR",      reason => "Alarm Z1 ElektroHeizkassette E2 ausgel&ouml;st &Uuml;berhitzungsschutz/ Druckw&auml;chter",                                                                                                                                                                                                                                                        checklist => "Niedriger Systemdruck, Luft im Heizsystem&bull;Schmutz im Systemfilter/Filterball SC1&bull;Schlechte Zirkulation im W&auml;rmetr&auml;gersystem/Heizsystem&bull;Sicherung F1 defekt&bull;Druckw&auml;chter defekt&bull;&Uuml;berhitzungsschutz defekt&bull;I/O-Board defekt, kein PWM-Signal zur Umw&auml;lzpumpe&bull;Sicherung F50 I/O-Board defekt&bull;I/O-Board defekt, keine 230 V an Klemme 01 (51, N)&bull;I/O-Board defekt&bull;Umw&auml;lzpumpe defekt&bull;Sch&uuml;tz zu elektr. Zuheizer bleibt in der geschlossenen Position stecken",correction => "Das Heizsystem laut Anweisungen in der Gebrauchsanweisung entl&uuml;ften, Wasser im Heizzystem&bull;Systemfilter/Filterball SC1 reinigen&bull;Einstellventile/Heizk&ouml;rperthermostate &uuml;berpr&uuml;fen&bull;Sicherung F1 wechseln&bull;Druckw&auml;chter wechseln&bull;Schalttemperatur (96°C) &uuml;berpr&uuml;fen &Uuml;berhitzungsschutz wechseln&bull;PWM-Signal an Klemme 36,37 am I/O-Board abschalten, Umw&auml;lzpumpe auf 100 % erh&ouml;hen&bull;Sicherung F50 wechseln&bull;&Uuml;berpr&uuml;fen, ob 230 V an Klemme 01 (51, N) am I/O-Board&bull;I/O-Board wechseln&bull;Umw&auml;lzpumpe wechseln&bull;&Uuml;berpr&uuml;fen/Sch&uuml;tz wechseln"};
	$ErrorCodeList{"0395"} = {errorcode => "A01",                                    subcode => "5247",                                               class => "WPL AR",      reason => "Alarm Z2 Elektro-Heiz-kassette E2 ausgel&ouml;st &Uuml;berhitzungsschutz/Druckw&auml;chter",                                                                                                                                                                                                                                                       checklist => "Siehe Alarm 5246 oben",                                                                                                                                                                                                                                                                                                                                                                                                                                                                               correction => "Siehe Ma&szlig;nahmen f&uuml;r ALARM 5246 oben"                                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0396"} = {errorcode => "A01",                                    subcode => "5265",                                               class => "WPL AR",      reason => "Warnung Heizpumpenkarte Z1 getrennt",                                                                                                                                                                                                                                                                                                              checklist => "Loser Kontakt in CANbus-Anschl&uuml;ssen an Installer-Board/ I/O-Board W&auml;rmepumpe&bull;Unterbrechung am CANbus-Kabel zwischen Tower und W&auml;rmepumpe&bull;Falscher Typ CANbus-Kabel&bull;CANbus-Kabel zusammen mit Stromversorgung der W&auml;rmepumpe verlegt&bull;Fehlerhafte Erdung des CANbus-Kabels",                                                                                                                                                                                    correction => "CANbus-Anschl&uuml;sse an Installer-Board/ I/O-Board W&auml;rmepumpe &uuml;berpr&uuml;fen&bull;CANbus-Kabel zwischen Tower und W&auml;rmepumpe &uuml;berpr&uuml;fen&bull;Zum richtigen Typ von Kabel wechseln&bull;CANbusund Stromkabel mit einem Abstand von mindestens 100 mm verlegen&bull;Kabelschutz zur/von der Erdung entfernen/abschlie&szlig;en"                                        };
	$ErrorCodeList{"0397"} = {errorcode => "A01",                                    subcode => "5267",                                               class => "WPL AR",      reason => "Warnung Heizpumpenkarte Z2 getrennt",                                                                                                                                                                                                                                                                                                              checklist => "Loser Kontakt in CANbus-Anschl&uuml;ssen an Installer-Board/ I/O-Board W&auml;rmepumpe 2&bull;Unterbrechung am CANbus-Kabel zwischen Tower und W&auml;rmepumpe 2&bull;Falscher Typ CANbus-Kabel&bull;CANbus-Kabel zusammen mit Stromversorgung der W&auml;rmepumpe 2 verlegt&bull;Fehlerhafte Erdung des CANbus-Kabels",                                                                                                                                                                              correction => "CANbus-Anschl&uuml;sse an Installer-Board/ I/O-Board W&auml;rmepumpe 2 &uuml;berpr&uuml;fen&bull;CANbus-Kabel zwischen Tower und W&auml;rmepumpe 2 wechseln&bull;Zum richtigen Typ von Kabel wechseln&bull;CANbusund Stromkabel mit einem Abstand von mindestens 100 mm verlegen&bull;Kabelschutz zur/von der Erdung entfernen/abschlie&szlig;en"                                                };
	$ErrorCodeList{"0398"} = {errorcode => "A01",                                    subcode => "5269",                                               class => "WPL AR",      reason => "Alarm Z1 Elektro-Heizkassette EE zu hohe Temperatur",                                                                                                                                                                                                                                                                                              checklist => "Schmutz im Systemfilter/Filterball SC1&bull;Schlechte Zirkulation im W&auml;rmetr&auml;gersystem/ Heizsystem&bull;F&uuml;hler TC1 defekt",                                                                                                                                                                                                                                                                                                                                                            correction => "Systemfilter/Filterball SC1 reinigen&bull;Einstellventile/Heizk&ouml;rperthermostate &uuml;berpr&uuml;fen&bull;F&uuml;hler TC1 im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen/F&uuml;hler TC1 wechseln"                                                                                                                                                                      };
	$ErrorCodeList{"0399"} = {errorcode => "A01",                                    subcode => "5270",                                               class => "WPL AR",      reason => "Alarm Z2 Elektro-Heizkassette EE zu hohe Temperatur",                                                                                                                                                                                                                                                                                              checklist => "Siehe Ursachen f&uuml;r Warnung 5269 oben",                                                                                                                                                                                                                                                                                                                                                                                                                                                           correction => "Siehe Ma&szlig;nahmen f&uuml;r ALARM 5269 oben"                                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0400"} = {errorcode => "A01",                                    subcode => "5271",                                               class => "WPL AR",      reason => "Alarm Heizsystem 1 hohe Vorlauftemperatur",                                                                                                                                                                                                                                                                                                        checklist => "F&uuml;hler T0 defekt&bull;3-Wege-Ventil VW1 wechselt nicht aus WW-Betriebsart&bull;&Uuml;berpr&uuml;fen, ob Klemme 53 (O3) an I/O-Board im WW-Betrieb 230 V abgibt&bull;&Uuml;berpr&uuml;fen, ob Klemme 53 (O3) an I/O-Board im WW-Betrieb 230 V abgibt&bull;&Uuml;berhitzungsschutz ausgel&ouml;st",                                                                                                                                                                                                correction => "F&uuml;hler TC0 im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen/ F&uuml;hler TC0 wechseln&bull;Funktion 3-Wege-Ventil VW1 &uuml;berpr&uuml;fen&bull;Wenn 230 V im WW-Betrieb, VW1 Motorteil/Kabel wechseln&bull;Wenn keine 230 V im WW-Betrieb, I/O-Board wechseln&bull;&Uuml;berhitzungsschutz r&uuml;cksetzen"                                                              };
	$ErrorCodeList{"0401"} = {errorcode => "A01",                                    subcode => "5272",                                               class => "WPL AR",      reason => "Alarm externe Zuheizung EM funktioniert nicht",                                                                                                                                                                                                                                                                                                    checklist => "Siehe Ursachen f&uuml;r Alarm 5246&bull;Sicherung in Stromzentrale ausgel&ouml;st",                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Siehe Ma&szlig;nahmen f&uuml;r Alarm 5246&bull;Sicherung in Stromzentrale wechseln/r&uuml;cksetzen"                                                                                                                                                                                                                                                                                              };
	$ErrorCodeList{"0402"} = {errorcode => "A01",                                    subcode => "5273",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z1 Phasenw&auml;chter",                                                                                                                                                                                                                                                                                                      checklist => "Sicherung im Tower ausgel&ouml;st&bull;Phase/Phasen fehlen an ankommender Klemme W&auml;rmepumpe&bull;Phase/Phasen fehlen an ankommender Klemme am EMI-Filter im Inverter&bull;Wenn alle Phasen am EMI-Filter unter Spannung stehen, ist der Inverter defekt",                                                                                                                                                                                                                                        correction => "Sicherung im Tower r&uuml;cksetzen&bull;&Uuml;berpr&uuml;fen, ob alle Phasen an ankommender Klemme der W&auml;rmepumpe vorhanden sind&bull;&Uuml;berpr&uuml;fen, ob alle Phasen an ankommender Klemme am EMI-Filter vorhanden sind&bull;Inverter wechseln"                                                                                                                                       };
	$ErrorCodeList{"0403"} = {errorcode => "A01",                                    subcode => "5274",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z2 Phasenw&auml;chter",                                                                                                                                                                                                                                                                                                      checklist => "Siehe Ursachen f&uuml;r Alarm 5273",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Siehe Ma&szlig;nahmen f&uuml;r Alarm 5273"                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0404"} = {errorcode => "A01",                                    subcode => "5298",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z1 zu hoher Druck an JR1",                                                                                                                                                                                                                                                                                                 checklist => "Schmutz im Systemfilter/Filterball SC1&bull;Schlechte Zirkulation im W&auml;rmetr&auml;gersystem/ Heizsystem&bull;Luft im W&auml;rmetr&auml;gersystem/Heizsystem&bull;Fehler an F&uuml;hler TC3, TC0, T0&bull;VW1 wechselt niciht von Warmwasser zu Heizung&bull;Installer-Board defekt&bull;Installer-Board defekt, kein PWM-Signal&bull;Installer-Board defekt, keine 230 V an Klemme 01 (51, N)&bull;Installer-Board defekt",                                                                      correction => "Systemfilter/Filterball SC1 reinigen&bull;Ausreichende Stromrate sicherstellen&bull;Das Heizsystem laut Anweisungen in der Gebrauchsanweisung entl&uuml;ften, Wasser im Heizzystem auff&uuml;llen&bull;F&uuml;hlerwert im Vergleich zur aktuellen F&uuml;hlertabelle und&bull;VW1 Position A = Warmwasser, B = Heizk&ouml;rper &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, dass Klemme 53 nur bei Warmwasserbedarf 230 V gibt&bull;PWM-Signal an Klemme 36,37 am InstallerBoard abschalten, Umw&auml;lzpumpe auf 100 % erh&ouml;hen, wenn nicht Umw&auml;lzpumpe wechseln&bull;&Uuml;berpr&uuml;fen, ob 230 V an Klemme 01 (51, N) am Installer-Board vorhanden sind&bull;Installer-Board wechseln"};
	$ErrorCodeList{"0405"} = {errorcode => "A01",                                    subcode => "5300",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z2 ausgel&ouml;ster Hochdruckpressostat",                                                                                                                                                                                                                                                                                  checklist => "Pressostat au&szlig;er Funktion&bull;Unterbrechung/schlechte Verbindung im Kabelbaum zwischen MR1 und Inverter&bull;Unterbrechung/schlechte Verbindung im Kabelbaum zwischen MR1 und Inverter",                                                                                                                                                                                                                                                                                                       correction => "Aufgebrochener Kreis <43bar = Pressostat wechseln&bull;Kabelbaum/Anschl&uuml;sse zwischen MR1 und Inverter &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob Hochdruckw&auml;chter angeschlossen ist"                                                                                                                                                                                           };
	$ErrorCodeList{"0406"} = {errorcode => "A01",                                    subcode => "5302",                                               class => "WPL AR",      reason => "Warnung Z1 hohe Temperatur am Kompressortreiber",                                                                                                                                                                                                                                                                                                  checklist => "Schlechte W&auml;rme&uuml;bertragung zur K&uuml;hlschlange",                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Verschraubung zur K&uuml;hlschlange &uuml;berpr&uuml;fen"                                                                                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"0407"} = {errorcode => "A01",                                    subcode => "5304",                                               class => "WPL AR",      reason => "Warnung Z2 hohe Temperatur am Kompressortreiber",                                                                                                                                                                                                                                                                                                  checklist => "Siehe Warnung 5302 oben",                                                                                                                                                                                                                                                                                                                                                                                                                                                                             correction => "Siehe Ma&szlig;nahmen f&uuml;r Warnung 5302 oben"                                                                                                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"0408"} = {errorcode => "A01",                                    subcode => "5310",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z1 Hei&szlig;gastemperatur zu hoch",                                                                                                                                                                                                                                                                                       checklist => "Sauggas-&Uuml;berhitzung zu hoch&bull;Sauggas-&Uuml;berhitzung &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "&Uuml;berpr&uuml;fen, dass Sauggas-&Uuml;berhitzung 10C nicht kontinuierlich &uuml;bersteigt&bull;Sauggas-&Uuml;berhitzung = TR5-JR0"                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"0409"} = {errorcode => "A01",                                    subcode => "5312",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z2 Hei&szlig;gastemperatur; zu hoch",                                                                                                                                                                                                                                                                                      checklist => "Sauggas-&Uuml;berhitzung zu hoch&bull;Sauggas-&Uuml;berhitzung &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "&Uuml;berpr&uuml;fen, dass Sauggas-&Uuml;berhitzung 10C nicht kontinuierlich &uuml;bersteigt&bull;Sauggas-&Uuml;berhitzung = TR5-JR0"                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"0410"} = {errorcode => "A01",                                    subcode => "5314",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z1 Heizgasf&uuml;hler TR6 Unterbrechung",                                                                                                                                                                                                                                                                                  checklist => "Sauggas-&Uuml;berhitzung &uuml;berpr&uuml;fen&bull;Heizgasf&uuml;hler TR6/Signalkabel Unterbrechung&bull;Schraubklemme an I/O-Board zu Heizgasf&uuml;hler TR6 nicht angezogen&bull;Heizgasf&uuml;hler TR6/Signalkabel Unterbrechung&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                             correction => "Sauggas-&Uuml;berhitzung = TR5-JR0&bull;Ohmwert von F&uuml;hler TR6/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle und der tats&auml;chlichen Temperatur, getrennt von der I/O-Board-Klemme, &uuml;berpr&uuml;fen&bull;Schraubklemme &uuml;berpr&uuml;fen&bull;Heizgasf&uuml;hler TR6 wechseln&bull;I/O-Board wechseln"                                                               };
	$ErrorCodeList{"0411"} = {errorcode => "A01",                                    subcode => "5315",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z1 Heizgasf&uuml;hler TR6 Kurzschluss",                                                                                                                                                                                                                                                                                    checklist => "Heizgasf&uuml;hler TR6/Signalkabel kurzgeschlossen&bull;Signalkabel zum Heizgasf&uuml;hler kurzgeschlossen&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                                                                                                      correction => "Ohmwert von F&uuml;hler TR6/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle und der tats&auml;chlichen Temperatur, getrennt von der I/O-Board-Klemme, &uuml;berpr&uuml;fen&bull;Signalkabel zum Heizgasf&uuml;hler TR6 reparieren/wechseln&bull;I/O-Board wechseln"                                                                                                                    };
	$ErrorCodeList{"0412"} = {errorcode => "A01",                                    subcode => "5317",                                               class => "WPL AR",      reason => "Warnung; W&auml;rmepumpe Z2 Heizgasf&uuml;hler TR6 Unterbrechung",                                                                                                                                                                                                                                                                                 checklist => "Heizgasf&uuml;hler TR6/Signalkabel Unterbrechung&bull;Schraubklemme an I/O-Board zu Heizgasf&uuml;hler TR6 nicht angezogen&bull;Heizgasf&uuml;hler TR6/Signalkabel Unterbrechung&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                                correction => "Ohmwert von F&uuml;hler TR6/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle und der tats&auml;chlichen Temperatur, getrennt von der I/O-Board-Klemme, &uuml;berpr&uuml;fen&bull;Schraubklemme &uuml;berpr&uuml;fen&bull;Heizgasf&uuml;hler TR6 wechseln&bull;I/O-Board wechseln"                                                                                                       };
	$ErrorCodeList{"0413"} = {errorcode => "A01",                                    subcode => "5318",                                               class => "WPL AR",      reason => "Warnung; W&auml;rmepumpe Z2 Heizgasf&uuml;hler TR6 Kurzschluss",                                                                                                                                                                                                                                                                                   checklist => "Heizgasf&uuml;hler TR6/Signalkabel kurzgeschlossen&bull;Heizgasf&uuml;hler TR6 kurzgeschlossen&bull;Signalkabel zum Heizgasf&uuml;hler kurzgeschlossen&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                                                          correction => "Ohmwert von F&uuml;hler TR6/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle und der tats&auml;chlichen Temperatur, getrennt von der I/OBoard-Klemme, &uuml;berpr&uuml;fen&bull;Ohmwert im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen/Heizgasf&uuml;hler TR6 wechseln&bull;Signalkabel zu W&auml;rmetr&auml;ger in F&uuml;hler TC1 reparieren/wechseln&bull;I/O-Board wechseln"};
	$ErrorCodeList{"0414"} = {errorcode => "A01",                                    subcode => "5320",                                               class => "8",           reason => "Warnung Temperaturf&uuml;hler TC3 Kabelbruch",                                                                                                                                                                                                                                                                                                     checklist => "&Uuml;berpr&uuml;fen, ob das Anschlusskabel vom Temperaturf&uuml;hler TC3 im Steuerger&auml;t korrekt angeschlossen ist&bull;-",                                                                                                                                                                                                                                                                                                                                                                      correction => "Anschlusskabel vom Temperaturf&uuml;hler TC3 im Steuerger&auml;t korrekt anschlie&szlig;en&bull;Temperaturf&uuml;hler TC3 austauschen"                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"0415"} = {errorcode => "A01",                                    subcode => "5321",                                               class => "4",           reason => "Warnung Temperaturf&uuml;hler TC3 Kurzschluss",                                                                                                                                                                                                                                                                                                    checklist => "Temperaturf&uuml;hler/Kabel TC3 pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                                         correction => "Temperaturf&uuml;hler TC3 austauschen"                                                                                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"0416"} = {errorcode => "A01",                                    subcode => "5323",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z2 Kondensatorf&uuml;hler TC3 Unterbrechung",                                                                                                                                                                                                                                                                              checklist => "Kondensatorf&uuml;hler TC3/Signalkabel kurzgeschlossen&bull;Kondensatorf&uuml;hler TC3 kurzgeschlossen&bull;Signalkabel zu Vorlauff&uuml;hler T0 kurzgeschlossen&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                                                correction => "Ohmwert von F&uuml;hler TC3/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle und der tats&auml;chlichen Temperatur, getrennt von der I/OBoard-Klemme, &uuml;berpr&uuml;fen&bull;Ohmwert im Vergleich zur aktuellen F&uuml;hlertabelle und tats&auml;chlichen Temperatur &uuml;ber-pr&uuml;fen/Kondensatorf&uuml;hler TC3 wechseln&bull;Signalkabel zu Kondensatorf&uuml;hler TC3 reparieren/wechseln&bull;I/O-Board wechseln"};
	$ErrorCodeList{"0417"} = {errorcode => "A01",                                    subcode => "5324",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z2 Kondensatorf&uuml;hler TC3 kurzgeschlossen",                                                                                                                                                                                                                                                                            checklist => "Kondensatorf&uuml;hler TC3/Signalkabel kurzgeschlossen&bull;Kondensatorf&uuml;hler TC3 kurzgeschlossen&bull;Signalkabel zu Kondensatorf&uuml;hler/ F&uuml;hler TC3 kurzgeschlossen&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                              correction => "Ohmwert von F&uuml;hler TC3/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle und der tats&auml;chlichen Temperatur, getrennt von der I/OBoard-Klemme, &uuml;berpr&uuml;fen&bull;Ohmwert im Vergleich zur aktuellen F&uuml;hler-tabelle und tats&auml;chlichen Temperatur &uuml;ber-pr&uuml;fen/Kondensatorf&uuml;hler TC3 wechseln&bull;Signalkabel/F&uuml;hler zu Kondensatorf&uuml;hler TC3 reparieren/wechseln&bull;I/O-Board wechseln"};
	$ErrorCodeList{"0418"} = {errorcode => "A01",                                    subcode => "5330",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z1 Kommunikations-; fehler zum Kompressortreiber",                                                                                                                                                                                                                                                                         checklist => "Keine Spannung zum Au&szlig;enteil&bull;St&ouml;rung am MODbus&bull;CANbus-Kabel zwischen IDU, ODU fehlerhaft verlegt",                                                                                                                                                                                                                                                                                                                                                                               correction => "230/400 V-Anschl&uuml;sse in Tower und ODU &uuml;berpr&uuml;fen&bull;MODbus-Kabel/Anschlussklemmen zwischen I/O-Board und Inverter &uuml;berpr&uuml;fen&bull;Kabelbaum und Verlegung (mindestens 100 mm zwischen Stromkabel und Canbus-Kabel) &uuml;berpr&uuml;fen"                                                                                                                              };
	$ErrorCodeList{"0419"} = {errorcode => "A01",                                    subcode => "5332",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z2 Kommunikations-; fehler zum Kompressortreiber",                                                                                                                                                                                                                                                                         checklist => "Keine Spannung zum Au&szlig;enteil&bull;St&ouml;rung am MODbus&bull;CANbus-Kabel zwischen IDU, ODU fehlerhaft verlegt",                                                                                                                                                                                                                                                                                                                                                                               correction => "230/400 V-Anschl&uuml;sse in Tower und ODU &uuml;berpr&uuml;fen&bull;MODbus-Kabel/Anschlussklemmen zwischen I/O-Board und Inverter &uuml;berpr&uuml;fen&bull;Kabelbaum und Verlegung (mindestens 100 mm zwischen Stromkabel und Canbus-Kabel) &uuml;berpr&uuml;fen"                                                                                                                              };
	$ErrorCodeList{"0420"} = {errorcode => "A01",                                    subcode => "5350",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z1 asynchroner Betrieb am Kompressor",                                                                                                                                                                                                                                                                                     checklist => "Spiel im Kabelbaum zwischen Kompressor und Inverter&bull;&Ouml;l/Fl&uuml;ssigkeit im Kompressor bei Startversuch&bull;Kompressorf&uuml;hler TR1 zeigt falschen Wert&bull;Wenn I/O-Board ausgetauscht wird, fehlerhafte Einstellung des Drehgebers am I/O-Board",                                                                                                                                                                                                                                      correction => "Kabelbaum/Anschl&uuml;sse zwischen Kompressor und Inverter &uuml;berpr&uuml;fen&bull;Kompressorf&uuml;hler TR1 zeigt falschen Wert&bull;Kompressorf&uuml;hler TR1 im Vergleich zur aktuellen F&uuml;hlertabelle und tats&auml;chlichen Temperatur &uuml;berpr&uuml;fen&bull;Einstellung im Vergleich zum aktuellen elektrischen Schaltplan &uuml;berpr&uuml;fen"                                 };
	$ErrorCodeList{"0421"} = {errorcode => "A01",                                    subcode => "5351",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z1 asynchroner Betrieb am Kompressor",                                                                                                                                                                                                                                                                                       checklist => "Siehe Pr&uuml;fvorgang/Ursachen f&uuml;r Ursachencode 5350",                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5350"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0422"} = {errorcode => "A01",                                    subcode => "5353",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z2 asynchroner Betrieb am Kompressor",                                                                                                                                                                                                                                                                                       checklist => "Siehe Pr&uuml;fvorgang/Ursachen f&uuml;r Ursachencode 5352",                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5352"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0423"} = {errorcode => "A01",                                    subcode => "5354",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z1 &Uuml;berstrom Kompressor",                                                                                                                                                                                                                                                                                             checklist => "Unterbrechung/Kurzschluss im Kabelbaum zwischen Inverter und Kompressor&bull;Stromfehler im Kompressor&bull;Inverter au&szlig;er Funktion",                                                                                                                                                                                                                                                                                                                                                           correction => "Kabelbaum zwischen Inverter und Kompressor &uuml;berpr&uuml;fen&bull;Widerstand zwischen Wicklungen und Erdung messen; bei Widerstand; < 10 kOhm, Kompressor wechseln&bull;Inverter wechseln"                                                                                                                                                                                                    };
	$ErrorCodeList{"0424"} = {errorcode => "A01",                                    subcode => "5356",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z2 &Uuml;berstrom Kompressor",                                                                                                                                                                                                                                                                                             checklist => "Unterbrechung/Kurzschluss im Kabelbaum zwischen Inverter und Kompressor&bull;Stromfehler im Kompressor&bull;Inverter au&szlig;er Funktion",                                                                                                                                                                                                                                                                                                                                                           correction => "Kabelbaum zwischen Inverter und Kompressor &uuml;berpr&uuml;fen&bull;Widerstand zwischen Wicklungen und Erdung messen; bei Widerstand < 10 kOhm, Kompressor wechseln&bull;Inverter wechseln"                                                                                                                                                                                                     };
	$ErrorCodeList{"0425"} = {errorcode => "A01",                                    subcode => "5360",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z2 PFC &Uuml;berstrom des Kompressortreibers",                                                                                                                                                                                                                                                                             checklist => "Kurze Tiefe bei der Strommessung/ Spannungsmessung&bull;Interner Fehler im Inverter",                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "Bei wiederholten Warnungen Kontakt zum Stromanbieter aufnehmen&bull;Inverter wechseln"                                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"0426"} = {errorcode => "A01",                                    subcode => "5362",                                               class => "WPL AR",      reason => "Info W&auml;rmepumpe Z1 &Uuml;berspannung",                                                                                                                                                                                                                                                                                                        checklist => "Zu hohe Spannung zur W&auml;rmepumpe&bull;Zu hohe eingehende Netzspannung",                                                                                                                                                                                                                                                                                                                                                                                                                           correction => "Hauptsicherungen in Bezug auf schlechten Kontakt &uuml;berpr&uuml;fen&bull;Bei wiederholten Warnungen Kontakt zum Stromanbieter aufnehmen"                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0427"} = {errorcode => "A01",                                    subcode => "5366",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z1 niedrige &Uuml;berhitzung",                                                                                                                                                                                                                                                                                             checklist => "Motor nicht ordnungsgem&auml;&szlig; montiert&bull;&Uuml;berpr&uuml;fen, dass Ventile sich in der richtigen Reihenfolge &ouml;ffnen/schlie&szlig;en&bull;Expansionsventilmotor defekt&bull;Expansionsventil VR1 &ouml;ffnet sich zu weit/bleibt in der offenen Stellung stecken",                                                                                                                                                                                                                     correction => "Ventilmotor &uuml;berpr&uuml;fen&bull;Test Au&szlig;eneinheit aktivieren&bull;Widerstand zwischen Kabel Orange, Rot, Gelb und Schwarz zu Grau messen; 46 kOhm = OK, bei Unterbrechung oder Kurzschluss, Ventilmotor wechseln&bull;Bei wiederholten Warnungen, Expansionsventil wechseln"                                                                                                         };
	$ErrorCodeList{"0428"} = {errorcode => "A01",                                    subcode => "5368",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z2 niedrige &Uuml;berhitzung",                                                                                                                                                                                                                                                                                             checklist => "Motor nicht ordnungsgem&auml;&szlig; montiert&bull;&Uuml;berpr&uuml;fen, dass Ventile sich in der richtigen Reihenfolge &ouml;ffnen/schlie&szlig;en&bull;Expansionsventilmotor defekt&bull;Expansionsventil VR1 &ouml;ffnet sich zu weit/ bleibt in der offenen Stellung stecken",                                                                                                                                                                                                                    correction => "Ventilmotor &uuml;berpr&uuml;fen&bull;Test Au&szlig;eneinheit aktivieren&bull;Widerstand zwischen Kabel Orange, Rot, Gelb und Schwarz zu Grau messen; 46 kOhm = OK, bei Unterbrechung oder Kurzschluss, Ventilmotor wechseln&bull;Bei wiederholten Warnungen, Expansionsventil wechseln"                                                                                                         };
	$ErrorCodeList{"0429"} = {errorcode => "A01",                                    subcode => "5374",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z1 Frostschutz des Kondensators",                                                                                                                                                                                                                                                                                          checklist => "Schlechte/ausgebliebene Zirkulation im W&auml;rmetr&auml;gersystem/Heizsystem&bull;Schmutz im Systemfilter/Filterball SC1&bull;Luft im W&auml;rmetr&auml;gersystem/ Heizsystem&bull;F&uuml;hler TC3 defekt&bull;I/O-Board defekt, 230 V fehlt&bull;I/O-Board defekt, kein PWM-Signal&bull;I/O-Board defekt, keine 230 V an Klemme 01 (51, N)&bull;I/O-Board defekt&bull;Umw&auml;lzpumpe defekt",                                                                                                     correction => "Einstellventile/Heizk&ouml;rperthermostate &uuml;berpr&uuml;fen&bull;Systemfilter/Filterball SC1 reinigen&bull;Das Heizsystem laut Anweisungen in der Gebrauchsanweisung entl&uuml;ften, Wasser im Heizsystem auff&uuml;llen&bull;F&uuml;hler TC3 im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen/F&uuml;hler TC3 wechseln&bull;&Uuml;berpr&uuml;fen, ob 230 V an Klemme 01 (51, N) am I/O-Board&bull;PWM-Signal an Klemme 36,37 am I/O-Board abschalten, Umw&auml;lzpumpe auf 100 % erh&ouml;hen&bull;&Uuml;berpr&uuml;fen, ob 230 V an Klemme 01 (51, N) am I/O-Board&bull;I/O-Board wechseln&bull;Umw&auml;lzpumpe wechseln"};
	$ErrorCodeList{"0430"} = {errorcode => "A01",                                    subcode => "5376",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z2 Frostschutz des Kondensators",                                                                                                                                                                                                                                                                                          checklist => "Schlechte/ausgebliebene Zirkulation im W&auml;rmetr&auml;gersystem/Heizsystem&bull;Schmutz im Systemfilter/Filterball SC1&bull;Luft im W&auml;rmetr&auml;gersystem/ Heizsystem&bull;F&uuml;hler TC3 defekt&bull;I/O-Board defekt, 230 V fehlt&bull;I/O-Board defekt, kein PWM-Signal&bull;I/O-Board defekt, keine 230 V an Klemme 01 (51, N)&bull;I/O-Board defekt&bull;Umw&auml;lzpumpe defekt",                                                                                                     correction => "Einstellventile/Heizk&ouml;rperthermostate &uuml;berpr&uuml;fen&bull;Systemfilter/Filterball SC1 reinigen&bull;Das Heizsystem laut Anweisungen in der Gebrauchsanweisung entl&uuml;ften, Wasser im Heizzystem auff&uuml;llen&bull;F&uuml;hler TC3 im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen/ F&uuml;hler TC3 wechseln&bull;&Uuml;berpr&uuml;fen, ob 230 V an Klemme 01 (51, N) am I/O-Board&bull;PWM-Signal an Klemme 36,37 am I/O-Board abschalten, Umw&auml;lzpumpe auf 100 % erh&ouml;hen&bull;&Uuml;berpr&uuml;fen, ob 230 V an Klemme 01 (51, N) am I/O-Board&bull;I/O-Board wechseln&bull;Umw&auml;lzpumpe wechseln"};
	$ErrorCodeList{"0431"} = {errorcode => "A01",                                    subcode => "5378",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z1 fehlgeschlagene; Abtauung",                                                                                                                                                                                                                                                                                             checklist => "Temperatur des Heizsystems zu niedrig&bull;F&uuml;hler TL2 defekt",                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Mehrere Thermostate des Heizzystems &ouml;ffnen&bull;F&uuml;hler TL2 im Vergleich zur aktuellen F&uuml;hlertabelle und tats&auml;chlichen Temperatur &uuml;berpr&uuml;fen. Bei Abweichungen F&uuml;hler TL2 wechseln."                                                                                                                                                                           };
	$ErrorCodeList{"0432"} = {errorcode => "A01",                                    subcode => "5380",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z2 fehlgeschlagene Abtauung",                                                                                                                                                                                                                                                                                              checklist => "Temperatur des Heizsystems zu niedrig&bull;F&uuml;hler TL2 defekt",                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Mehrere Thermostate des Heizzystems &ouml;ffnen&bull;F&uuml;hler TL2 im Vergleich zur aktuellen F&uuml;hlertabelle und tats&auml;chlichen Temperatur &uuml;berpr&uuml;fen. Bei Abweichungen F&uuml;hler TL2 wechseln."                                                                                                                                                                           };
	$ErrorCodeList{"0433"} = {errorcode => "A01",                                    subcode => "5394",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z1 Interner Kompressortreiberfehler 1",                                                                                                                                                                                                                                                                                    checklist => "Unterbrechung/Kurzschluss im Kabelbaum zwischen Inverter und Kompressor&bull;Stromfehler im Kompressor&bull;Inverter au&szlig;er Funktion",                                                                                                                                                                                                                                                                                                                                                           correction => "Kabelbaum zwischen Inverter und Kompressor &uuml;berpr&uuml;fen&bull;Widerstand zwischen Wicklungen und Erdung messen; bei Widerstand; < 10 kOhm, Kompressor wechseln&bull;Inverter wechseln"                                                                                                                                                                                                    };
	$ErrorCodeList{"0434"} = {errorcode => "A01",                                    subcode => "5396",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z2 Interner Kompressortreiberfehler 1",                                                                                                                                                                                                                                                                                    checklist => "Unterbrechung/Kurzschluss im Kabelbaum zwischen Inverter und Kompressor&bull;Stromfehler im Kompressor&bull;Inverter au&szlig;er Funktion",                                                                                                                                                                                                                                                                                                                                                           correction => "Kabelbaum zwischen Inverter und Kompressor &uuml;berpr&uuml;fen&bull;Widerstand zwischen Wicklungen und Erdung messen; bei Widerstand; < 10 kOhm, Kompressor wechseln&bull;Inverter wechseln"                                                                                                                                                                                                    };
	$ErrorCodeList{"0435"} = {errorcode => "A01",                                    subcode => "5408",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z1 Lufteintrittstemperaturf&uuml;hler TL2 Unterbrechung",                                                                                                                                                                                                                                                                  checklist => "Lufteintrittsf&uuml;hler TL2 T0/Signalkabel Unterbrechung&bull;Schraubklemme an I/O-Board zu Lufteintrittsf&uuml;hler TL2 nicht angezogen&bull;Lufteintrittsf&uuml;hler TL2 T0/Signalkabel Unterbrechung&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                        correction => "Ohmwert von F&uuml;hler TL2/Signalkabel; im Vergleich zur aktuellen F&uuml;hlertabelle und der tats&auml;chlichen Temperatur, getrennt von der I/OBoard-Klemme, &uuml;berpr&uuml;fen&bull;Schraubklemme &uuml;berpr&uuml;fen&bull;Lufteintrittsf&uuml;hler/Signalkabel F&uuml;hler TL2 wechseln&bull;I/O-Board wechseln"                                                                         };
	$ErrorCodeList{"0436"} = {errorcode => "A01",                                    subcode => "5409",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z1 Lufteintrittstemperaturf&uuml;hler TL2 kurzgeschlossen",                                                                                                                                                                                                                                                                checklist => "Au&szlig;entemperaturf&uuml;hler TL2/Signalkabel kurzgeschlossen&bull;Lufteintrittsf&uuml;hler TC3 kurzgeschlossen&bull;Signalkabel zu Lufteintrittsf&uuml;hler/F&uuml;hler TL2 kurzgeschlossen&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                 correction => "Ohmwert von F&uuml;hler TL2/Signalkabel; im Vergleich zur aktuellen F&uuml;hlertabelle und der tats&auml;chlichen Temperatur, getrennt von der I/OBoard-Klemme, &uuml;berpr&uuml;fen&bull;Ohmwert im Vergleich zur aktuellen F&uuml;hlertabelle und tats&auml;chlichen Temperatur &uuml;berpr&uuml;fen/Lufteintrittsf&uuml;hler TL2 wechseln&bull;Signalkabel/F&uuml;hler zu Kondensatorf&uuml;hler TL2 reparieren/wechseln&bull;I/O-Board wechseln"};
	$ErrorCodeList{"0437"} = {errorcode => "A01",                                    subcode => "5411",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z2 Lufteintrittstemperaturf&uuml;hler TL2 Unterbrechung",                                                                                                                                                                                                                                                                  checklist => "Lufteintrittsf&uuml;hler TL2 T0/Signalkabel Unterbrechung&bull;Schraubklemme an I/O-Board zu Lufteintrittsf&uuml;hler TL2 nicht angezogen&bull;Lufteintrittsf&uuml;hler TL2 T0/Signalkabel Unterbrechung&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                        correction => "Ohmwert von F&uuml;hler TL2/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle und der tats&auml;chlichen Temperatur, getrennt von der I/OBoard-Klemme, &uuml;berpr&uuml;fen&bull;Schraubklemme &uuml;berpr&uuml;fen&bull;Lufteintrittsf&uuml;hler/Signalkabel F&uuml;hler TL2 wechseln&bull;I/O-Board wechseln"                                                                          };
	$ErrorCodeList{"0438"} = {errorcode => "A01",                                    subcode => "5412",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z1 Lufteintrittstemperaturf&uuml;hler TL2 kurzgeschlossen",                                                                                                                                                                                                                                                                checklist => "Au&szlig;entemperaturf&uuml;hler TL2/Signalkabel kurzgeschlossen&bull;Lufteintrittsf&uuml;hler TL2 kurzgeschlossen&bull;Signalkabel zu Lufteintrittsf&uuml;hler/F&uuml;hler TL2 kurzgeschlossen&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                 correction => "Ohmwert von F&uuml;hler TL2/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle und der tats&auml;chlichen Temperatur, getrennt von der I/OBoard-Klemme, &uuml;berpr&uuml;fen&bull;Ohmwert im Vergleich zur aktuellen F&uuml;hler-tabelle und tats&auml;chlichen Temperatur &uuml;berpr&uuml;fen/Lufteintrittsf&uuml;hler TL2 wechseln&bull;Signalkabel/F&uuml;hler zu Kondensatorf&uuml;hler TL2 reparieren/wechseln&bull;I/O-Board wechseln"};
	$ErrorCodeList{"0439"} = {errorcode => "A01",                                    subcode => "5414",                                               class => "WPL AR",      reason => "Warnung; W&auml;rmepumpe Z1 Fl&uuml;ssigkeitsleitung in Heizbetrieb F&uuml;hler TR3 Unterbrechung",                                                                                                                                                                                                                                                checklist => "F&uuml;hler TR3/Signalkabel Unterbrechung&bull;Schraubklemme an I/O-Board zu F&uuml;hler TR3 nicht angezogen&bull;F&uuml;hler TR3/Signalkabel Unterbrechung&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                                                     correction => "Ohmwert von F&uuml;hler TR3/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen; das Innenteil ist dabei von; I/O-Board-Klemme getrennt&bull;Schraubklemme &uuml;berpr&uuml;fen&bull;F&uuml;hler TR3 wechseln&bull;I/O-Board wechseln"                                                                                                                                };
	$ErrorCodeList{"0440"} = {errorcode => "A01",                                    subcode => "5415",                                               class => "WPL AR",      reason => "Warnung; W&auml;rmepumpe Z1 Fl&uuml;ssigkeitsleitung in Heizbetrieb F&uuml;hler TR3 kurzgeschlossen",                                                                                                                                                                                                                                              checklist => "W&auml;rmetr&auml;ger in F&uuml;hler TR3/Signalkabel kurzgeschlossen&bull;F&uuml;hler TR3 kurzgeschlossen&bull;Signalkabel/F&uuml;hler TR3 kurzgeschlossen&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                                                      correction => "Ohmwert von F&uuml;hler TR3/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen; das Innenteil ist dabei von; I/O-Board-Klemme getrennt&bull;Ohmwert im Vergleich zur aktuellen F&uuml;hler-tabelle &uuml;berpr&uuml;fen/F&uuml;hler TR3 wechseln&bull;F&uuml;hler TR3 wechseln&bull;I/O-Board wechseln"                                                              };
	$ErrorCodeList{"0441"} = {errorcode => "A01",                                    subcode => "5417",                                               class => "WPL AR",      reason => "Warnung; W&auml;rmepumpe Z2 Fl&uuml;ssigkeitsleitung in Heizbetrieb F&uuml;hler TR3 Unterbrechung",                                                                                                                                                                                                                                                checklist => "F&uuml;hler TR3/Signalkabel Unterbrechung&bull;Schraubklemme an I/O-Board zu F&uuml;hler TR3 nicht angezogen&bull;F&uuml;hler TR3/Signalkabel Unterbrechung&bull;I/O-Board au&szlig;er Funktion&bull;-",                                                                                                                                                                                                                                                                                              correction => "Ohmwert von F&uuml;hler TR3/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen; das Innenteil ist dabei von; I/O-Board-Klemme getrennt&bull;Schraubklemme an I/O-Board &uuml;berpr&uuml;fen&bull;F&uuml;hler TR3 wechseln&bull;I/O-Board wechseln&bull;-"                                                                                                            };
	$ErrorCodeList{"0442"} = {errorcode => "A01",                                    subcode => "5418",                                               class => "WPL AR",      reason => "Warnung; W&auml;rmepumpe Z2 Fl&uuml;ssigkeitsleitung in Heizbetrieb F&uuml;hler TR3 kurzgeschlossen",                                                                                                                                                                                                                                              checklist => "W&auml;rmetr&auml;ger in F&uuml;hler TR3/Signalkabel kurzgeschlossen&bull;F&uuml;hler TR3 kurzgeschlossen&bull;Signalkabel/F&uuml;hler TR3 kurzgeschlossen&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                                                      correction => "Ohmwert von F&uuml;hler TR3/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen; das Innenteil ist dabei von; I/O-Board-Klemme getrennt&bull;Ohmwert im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen/F&uuml;hler TR3 wechseln&bull;F&uuml;hler TR3 wechseln&bull;I/O-Board wechseln"                                                               };
	$ErrorCodeList{"0443"} = {errorcode => "A01",                                    subcode => "5420",                                               class => "WPL AR",      reason => "Warnung; W&auml;rmepumpe Z1 Fl&uuml;ssigkeitsleitung im K&uuml;hlbetrieb F&uuml;hler TR4 Unterbrechung",                                                                                                                                                                                                                                           checklist => "F&uuml;hler TR4/Signalkabel Unterbrechung&bull;Schraubklemme an I/O-Board zu F&uuml;hler TR4 nicht angezogen&bull;Lufteintrittsf&uuml;hler TR4/Signalkabel Unterbrechung&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                                        correction => "Ohmwert von F&uuml;hler TR4/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle und der tats&auml;chlichen Temperatur, getrennt von der I/OBoard-Klemme, &uuml;berpr&uuml;fen&bull;Schraubklemme an I/O-Board &uuml;berpr&uuml;fen&bull;Byt F&uuml;hler/Signalkabel F&uuml;hler TR4&bull;I/O-Board wechseln"                                                                               };
	$ErrorCodeList{"0444"} = {errorcode => "A01",                                    subcode => "5421",                                               class => "WPL AR",      reason => "Warnung; W&auml;rmepumpe Z1 Fl&uuml;ssigkeitsleitung im K&uuml;hlbetrieb F&uuml;hler TR4 kurzgeschlossen",                                                                                                                                                                                                                                         checklist => "F&uuml;hler TR4/Signalkabel kurzgeschlossen&bull;F&uuml;hler TR4 kurzgeschlossen&bull;Signalkabel/F&uuml;hler TR4 kurzgeschlossen&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                                                                               correction => "Ohmwert von F&uuml;hler TR4/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle und der tats&auml;chlichen Temperatur, getrennt von der I/OBoard-Klemme, &uuml;berpr&uuml;fen&bull;Ohmwert im Vergleich zur aktuellen F&uuml;hlertabelle und tats&auml;chlichen Temperatur &uuml;ber-pr&uuml;fen/F&uuml;hler TR4 wechseln&bull;Signalkabel/F&uuml;hler zu F&uuml;hler TR4 wechseln&bull;I/O-Board wechseln"};
	$ErrorCodeList{"0445"} = {errorcode => "A01",                                    subcode => "5423",                                               class => "WPL AR",      reason => "Warnung; W&auml;rmepumpe Z2 Fl&uuml;ssigkeitsleitung im K&uuml;hlbetrieb F&uuml;hler TR4 Unterbrechung",                                                                                                                                                                                                                                           checklist => "F&uuml;hler TR4/Signalkabel Unterbrechung&bull;Schraubklemme an I/O-Board zu F&uuml;hler TR4 nicht angezogen&bull;Lufteintrittsf&uuml;hler TR4/Signalkabel Unterbrechung&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                                        correction => "Ohmwert von F&uuml;hler TR4/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle und der tats&auml;chlichen Temperatur, getrennt von der I/OBoard-Klemme, &uuml;berpr&uuml;fen&bull;Schraubklemme an I/O-Board &uuml;berpr&uuml;fen&bull;Byt F&uuml;hler/Signalkabel F&uuml;hler TR4&bull;I/O-Board wechseln"                                                                               };
	$ErrorCodeList{"0446"} = {errorcode => "A01",                                    subcode => "5424",                                               class => "WPL AR",      reason => "Warnung; W&auml;rmepumpe Z2 Fl&uuml;ssigkeitsleitung im K&uuml;hlbetrieb F&uuml;hler TR4 kurzgeschlossen",                                                                                                                                                                                                                                         checklist => "F&uuml;hler TR4/Signalkabel kurzgeschlossen&bull;F&uuml;hler TR4 kurzgeschlossen&bull;Signalkabel/F&uuml;hler TR4 kurzgeschlossen&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                                                                               correction => "Ohmwert von F&uuml;hler TR4/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle und der tats&auml;chlichen Temperatur, getrennt von der I/OBoard-Klemme, &uuml;berpr&uuml;fen&bull;Ohmwert im Vergleich zur aktuellen F&uuml;hler-tabelle und tats&auml;chlichen Temperatur &uuml;ber-pr&uuml;fen/F&uuml;hler TR4 wechseln&bull;Signalkabel/F&uuml;hler zu F&uuml;hler TR4 wechseln&bull;I/O-Board wechseln"};
	$ErrorCodeList{"0447"} = {errorcode => "A01",                                    subcode => "5426",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z1 Heizgasf&uuml;hler TR5 Unterbrechung",                                                                                                                                                                                                                                                                                  checklist => "F&uuml;hler TR5/Signalkabel Unterbrechung&bull;Schraubklemme an I/O-Board zu F&uuml;hler TR5 nicht angezogen&bull;F&uuml;hler TR5/Signalkabel Unterbrechung&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                                                     correction => "Ohmwert von F&uuml;hler TR5/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen, getrennt von I/O-BoardKlemme.&bull;Schraubklemme an I/O-Board &uuml;berpr&uuml;fen&bull;F&uuml;hler TR5 wechseln&bull;I/O-Board wechseln"                                                                                                                                            };
	$ErrorCodeList{"0448"} = {errorcode => "A01",                                    subcode => "5427",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z1 Sauggasf&uuml;hler TR5 kurzgeschlossen",                                                                                                                                                                                                                                                                                checklist => "F&uuml;hler TR5/Signalkabel kurzgeschlossen&bull;F&uuml;hler TR5 kurzgeschlossen&bull;Signalkabel/F&uuml;hler TR5 kurzgeschlossen&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                                                                               correction => "Ohmwert von F&uuml;hler TR5/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen, getrennt von I/O-Board-Klemme&bull;Ohmwert im Vergleich zur aktuellen F&uuml;hler-tabelle &uuml;berpr&uuml;fen/F&uuml;hler TR5 wechseln&bull;F&uuml;hler TR5 wechseln&bull;I/O-Board wechseln"                                                                                       };
	$ErrorCodeList{"0449"} = {errorcode => "A01",                                    subcode => "5429",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z2 Heizgasf&uuml;hler TR5 Unterbrechung",                                                                                                                                                                                                                                                                                  checklist => "F&uuml;hler TR5/Signalkabel Unterbrechung&bull;Schraubklemme an I/O-Board zu F&uuml;hler TR5 nicht angezogen&bull;F&uuml;hler TR5/Signalkabel Unterbrechung&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                                                     correction => "Ohmwert von F&uuml;hler TR5/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen, getrennt von I/O-BoardKlemme&bull;Schraubklemme an I/O-Board &uuml;berpr&uuml;fen&bull;F&uuml;hler TR5 wechseln&bull;I/O-Board wechseln"                                                                                                                                             };
	$ErrorCodeList{"0450"} = {errorcode => "A01",                                    subcode => "5430",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z2 Sauggasf&uuml;hler TR5 kurzgeschlossen",                                                                                                                                                                                                                                                                                checklist => "F&uuml;hler TR5/Signalkabel kurzgeschlossen&bull;F&uuml;hler TR5 kurzgeschlossen&bull;Signalkabel/F&uuml;hler TR5 kurzgeschlossen&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                                                                               correction => "Ohmwert von F&uuml;hler TR5/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen, getrennt von I/O-Board-Klemme&bull;Ohmwert im Vergleich zur aktuellen F&uuml;hler-tabelle &uuml;berpr&uuml;fen/F&uuml;hler TR5 wechseln&bull;F&uuml;hler TR5 wechseln&bull;I/O-Board wechseln"                                                                                       };
	$ErrorCodeList{"0451"} = {errorcode => "A01",                                    subcode => "5432",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z1 Druckgeber; niedriger Druck JR0 Unterbrechung",                                                                                                                                                                                                                                                                         checklist => "Spiel in Kabelbaum/Klemme 17, 19 an IO-Board&bull;I/O-Board defekt&bull;Druckgeber JR0 defekt",                                                                                                                                                                                                                                                                                                                                                                                                       correction => "Kabelbaum und Klemmenanschl&uuml;sse 17, 19 an IO-Board &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, dass Klemme 19 mit ausgeschaltetem Druckgeber 5 V DC an Klemme 17 abgibt, wenn Spannung; < 0,5 V, IO-Board wechseln&bull;&Uuml;berpr&uuml;fen, dass Klemme 19 mit ausgeschaltetem Druckgeber 5 V DC an Klemme 17 abgibt, wenn Spannung; = 5 V, IO-Board wechseln"                        };
	$ErrorCodeList{"0452"} = {errorcode => "A01",                                    subcode => "5435",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z2 Druckgeber niedriger Druck JR0 Unterbrechung",                                                                                                                                                                                                                                                                          checklist => "Spiel in Kabelbaum/Klemme 17, 19 an IO-Board&bull;I/O-Board defekt&bull;Druckgeber JR0 defekt",                                                                                                                                                                                                                                                                                                                                                                                                       correction => "Kabelbaum und Klemmenanschl&uuml;sse 17, 19 an IO-Board &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, dass Klemme 19 mit ausgeschaltetem Druckgeber 5V DC an Klemme 17 abgibt, wenn Spannung; < 0,5 V, IO-Board wechseln&bull;&Uuml;berpr&uuml;fen, dass Klemme 19 mit ausgeschaltetem Druckgeber 5 V DC an Klemme 17 abgibt, wenn Spannung; = 5 V, IO-Board wechseln"                         };
	$ErrorCodeList{"0453"} = {errorcode => "A01",                                    subcode => "5438",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z1 Druckgeber hoher Druck JR1 Unterbrechung",                                                                                                                                                                                                                                                                              checklist => "Spiel in Kabelbaum/Klemme 16, 18 an IO-Board&bull;I/O-Board defekt&bull;Druckgeber JR1 defekt",                                                                                                                                                                                                                                                                                                                                                                                                       correction => "Kabelbaum und Klemmenanschl&uuml;sse 16, 18 an IO-Board &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, dass Klemme 19 mit ausgeschaltetem Druckgeber 5V DC an Klemme 17 abgibt, wenn Spannung; < 0,5 V, IO-Board wechseln&bull;&Uuml;berpr&uuml;fen, dass Klemme 19 mit ausgeschaltetem Druckgeber 5V DC an Klemme 17 abgibt, wenn Spannung; = 5 V, IO-Board wechseln"                          };
	$ErrorCodeList{"0454"} = {errorcode => "A01",                                    subcode => "5441",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z2; Druckgeber hoher Druck JR1 Unterbrechung",                                                                                                                                                                                                                                                                             checklist => "Spiel in Kabelbaum/Klemme 16, 18 an IO-Board&bull;I/O-Board defekt&bull;Druckgeber JR1 defekt",                                                                                                                                                                                                                                                                                                                                                                                                       correction => "Kabelbaum und Klemmenanschl&uuml;sse 16, 18 an IO-Board &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, dass Klemme 19 mit ausgeschaltetem Druckgeber 5V DC an Klemme 17 abgibt, wenn Spannung; < 0,5 V, IO-Board wechseln&bull;&Uuml;berpr&uuml;fen, dass Klemme 19 mit ausgeschaltetem Druckgeber 5V DC an Klemme 17 abgibt, wenn Spannung = 5 V,; IO-Board wechseln"                          };
	$ErrorCodeList{"0455"} = {errorcode => "A01",                                    subcode => "5445",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z1 Sammelalarm Gebl&auml;se",                                                                                                                                                                                                                                                                                                checklist => "Gebl&auml;se blockiert&bull;Gebl&auml;se gem&auml;&szlig; Test Au&szlig;eneinheit &uuml;berpr&uuml;fen&bull;I/O-Board defekt&bull;I/O-Board defekt&bull;Gebl&auml;se hat nicht 230 V Spannung&bull;Gebl&auml;se ohne 0: 10 V-Signal&bull;Gebl&auml;se defekt",                                                                                                                                                                                                                                        correction => "&Uuml;berpr&uuml;fen, dass Gebl&auml;se nicht physisch blockiert wird&bull;Test Au&szlig;eneinheit aktivieren&bull;&Uuml;berpr&uuml;fen, dass 230 V Spannung an Ausgang PL3 SSM Klemme 34 (75, N)&bull;Fehlt Spannung wie oben, I/O-Board wechseln&bull;&Uuml;berpr&uuml;fen, dass 230 V Spannung an Ausgang PL3 Klemme 32 (78, N)&bull;0: 10 V Spannung an Ausgang PL3 PWM Klemme 20 (20, 26) laut Test Au&szlig;eneinheit &uuml;berpr&uuml;fen&bull;Ist Spannung wie oben vorhanden, Gebl&auml;se wechseln"};
	$ErrorCodeList{"0456"} = {errorcode => "A01",                                    subcode => "5450",                                               class => "WPL AR",      reason => "Kondenswasserablauf in W&auml;rmepumpe Z1 blockiert",                                                                                                                                                                                                                                                                                              checklist => "Kondenswasserablauf in W&auml;rmepumpe ist blockiert&bull;Heizkabel im Kondenswasserablauf defekt, Ablauf zugefroren&bull;Heizkabel hat keine Spannung",                                                                                                                                                                                                                                                                                                                                              correction => "Ablauf &uuml;berpr&uuml;fen/reinigen&bull;Heizkabel &uuml;berpr&uuml;fen/Heizkabel wechseln&bull;&Uuml;berpr&uuml;fen, ob Ausgang EA1-Klemme 26 (79, N) bei manuellem Betrieb aus Heizkabel 230 V hat"                                                                                                                                                                                           };
	$ErrorCodeList{"0457"} = {errorcode => "A01",                                    subcode => "5452",                                               class => "WPL AR",      reason => "Interner Fehler am Inverter zum Kompressor",                                                                                                                                                                                                                                                                                                       checklist => "Inverter defekt",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                     correction => "Inverter wechseln"                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0458"} = {errorcode => "A01",                                    subcode => "5454",                                               class => "WPL AR",      reason => "Kondenswasserablauf in W&auml;rmepumpe Z2 blockiert",                                                                                                                                                                                                                                                                                              checklist => "Kondenswasserablauf in W&auml;rmepumpe ist blockiert&bull;Heizkabel im Kondenswasserablauf defekt, Ablauf zugefroren&bull;Heizkabel hat keine Spannung",                                                                                                                                                                                                                                                                                                                                              correction => "Tropfschale/Ablauf &uuml;berpr&uuml;fen/reinigen&bull;Heizkabel &uuml;berpr&uuml;fen/Heizkabel wechseln&bull;&Uuml;berpr&uuml;fen, ob Ausgang EA1-Klemme 26 (79, N) bei manuellem Betrieb aus Heizkabel 230 V hat"                                                                                                                                                                               };
	$ErrorCodeList{"0459"} = {errorcode => "A01",                                    subcode => "5456",                                               class => "WPL AR",      reason => "Interner Fehler am Inverter zum Kompressor in W&auml;rmepumpe Z2",                                                                                                                                                                                                                                                                                 checklist => "Inverter defekt",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                     correction => "Inverter wechseln"                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0460"} = {errorcode => "A01",                                    subcode => "5458",                                               class => "WPL AR",      reason => "Warnung Z2 Warmwasserf&uuml;hler TW1 Unterbrechung",                                                                                                                                                                                                                                                                                               checklist => "Warmwasserf&uuml;hler TW1/Signalkabel Unterbrechung&bull;Schraubklemme an I/O-Board zu Warmwasserf&uuml;hler TW1 nicht angezogen&bull;Warmwasserf&uuml;hler TW1/Signalkabel Unterbrechung&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                       correction => "Ohmwert von F&uuml;hler TW1/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle; &uuml;berpr&uuml;fen; das Innenteil ist dabei von I/OBoard-Klemme getrennt&bull;Schraubklemme &uuml;berpr&uuml;fen&bull;Vorlauff&uuml;hler TW1 wechseln&bull;I/O-Board wechseln"                                                                                                                          };
	$ErrorCodeList{"0461"} = {errorcode => "A01",                                    subcode => "5459",                                               class => "WPL AR",      reason => "Warnung Z2 Warmwasserf&uuml;hler TW1 kurzgeschlossen",                                                                                                                                                                                                                                                                                             checklist => "Warmwasserf&uuml;hler TW1/Signalkabel kurzgeschlossen&bull;W&auml;rmetr&auml;ger in F&uuml;hler TW1 kurzgeschlossen&bull;Signalkabel zu W&auml;rmetr&auml;ger in F&uuml;hler TW1 kurzgeschlossen&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                correction => "Ohmwert von F&uuml;hler TW1/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen; das Innenteil ist dabei von; I/O-Board-Klemme getrennt&bull;Ohmwert im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen/W&auml;rmetr&auml;ger in F&uuml;hler TW1 wechseln&bull;Signalkabel zu W&auml;rmetr&auml;ger aus F&uuml;hler TW1 reparieren/wechseln&bull;I/O-Board wechseln"};
	$ErrorCodeList{"0462"} = {errorcode => "A01",                                    subcode => "5465",                                               class => "WPL AR",      reason => "Kein Spannung zu W&auml;rmepumpe Z2",                                                                                                                                                                                                                                                                                                              checklist => "Niedrige/Keine Spannung zur W&auml;rmepumpe&bull;Niedrige eingehende Netzspannung",                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Hauptsicherungen in Bezug auf schlechten Kontakt/Defekt &uuml;berpr&uuml;fen&bull;Bei wiederholten Warnungen Kontakt zum Stromanbieter aufnehmen"                                                                                                                                                                                                                                                };
	$ErrorCodeList{"0463"} = {errorcode => "A01",                                    subcode => "5500",                                               class => "WPL AR",      reason => "Fu&szlig;hodenheizungsthermostat am externen Eingang ausgel&ouml;st",                                                                                                                                                                                                                                                                              checklist => "Sicherheitsthermostat f&uuml;r Fu&szlig;bodenheizung wurde ausgel&ouml;st",                                                                                                                                                                                                                                                                                                                                                                                                                           correction => "Thermostat zur&uuml;ckstellen, eventuelle Heizkurve einstellen"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0464"} = {errorcode => "A01",                                    subcode => "5503",                                               class => "WPL AR",      reason => "Warnung Problem mit Anschluss zur Belastungsw&auml;chter",                                                                                                                                                                                                                                                                                         checklist => "Kabelbaum/Anschl&uuml;sse fehlerhaft&bull;Loser Kontakt in CANbus-Anschl&uuml;ssen an Installer-Board Tower und Belastungsw&auml;chter&bull;Unterbrechung an CANbus-Kabel zwischen Installer-Board Tower und Belastungsw&auml;chter&bull;Falscher Typ CANbus-Kabel&bull;CANbus-Kabel zusammen mit Stromversorgung der W&auml;rmepumpe verlegt",                                                                                                                                                       correction => "Kabelbaum/Anschl&uuml;sse &uuml;berpr&uuml;fen&bull;CANbus-Anschl&uuml;sse an Installer-Board Tower und Belastungsw&auml;chter &uuml;berpr&uuml;fen&bull;CANbus-Kabel zwischen Installer-Board Tower und Belastungsw&auml;chter wechseln&bull;Zum richtigen Typ von Kabel wechseln&bull;CANbusund Stromkabel mit einem Abstand von mindestens 100 mm verlegen"                                   };
	$ErrorCodeList{"0465"} = {errorcode => "A01",                                    subcode => "5504",                                               class => "WPL AR",      reason => "Alarm Unterbrechung Anschluss zum Belastungsw&auml;chter",                                                                                                                                                                                                                                                                                         checklist => "Siehe Pr&uuml;fvorgang/Ursachen f&uuml;r Ursachencode 5432",                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5432"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0466"} = {errorcode => "A01",                                    subcode => "5506",                                               class => "WPL AR",      reason => "Kompressor startet nicht",                                                                                                                                                                                                                                                                                                                         checklist => "Vor&uuml;bergehende Funktionst&ouml;rung Inverter&bull;Interner Fehler im Inverter",                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Spannung zur W&auml;rmepumpe ausund wieder einschalten&bull;Inverter wechseln"                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"0467"} = {errorcode => "A01",                                    subcode => "5507",                                               class => "WPL AR",      reason => "MR1 hoher Druck Alarm Alarm B in W&auml;rmepumpe Z1",                                                                                                                                                                                                                                                                                              checklist => "Pressostat au&szlig;er Funktion&bull;Unterbrechung/schlechte Verbindung im Kabelbaum zwischen MR1 und Inverter&bull;Unterbrechung/schlechte Verbindung im Kabelbaum zwischen MR1 und Inverter",                                                                                                                                                                                                                                                                                                       correction => "Aufgebrochener Kreis < 43 bar = Pressostat wechseln&bull;Kabelbaum/Anschl&uuml;sse zwischen MR1 und Inverter &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob Hochdruckw&auml;chter angeschlossen ist"                                                                                                                                                                                         };
	$ErrorCodeList{"0468"} = {errorcode => "A01",                                    subcode => "5509",                                               class => "WPL AR",      reason => "MR1 hoher Druck Alarm Alarm B in W&auml;rmepumpe Z2",                                                                                                                                                                                                                                                                                              checklist => "Unterbrechung in Kabel/Pressostat MR1&bull;Pressostat au&szlig;er Funktion&bull;Unterbrechung/schlechte Verbindung im Kabelbaum zwischen MR1 und Inverter&bull;Unterbrechung/schlechte Verbindung im Kabelbaum zwischen MR1 und Inverter",                                                                                                                                                                                                                                                            correction => "An ersten Molexanschluss Ohmwert Pressostat messen/summieren, geschlossen = OK&bull;Aufgebrochener Kreis < 43bar = Pressostat wechseln&bull;Kabelbaum/Anschl&uuml;sse zwischen MR1 und Inverter &uuml;berpr&uuml;fen&bull;MR1 im Inverter verbinden, Warnung sollte verschwinden, wenn nicht, Inverter wechseln"                                                                                 };
	$ErrorCodeList{"0469"} = {errorcode => "A01",                                    subcode => "5512",                                               class => "WPL AR",      reason => "Kondensationstemperatur au&szlig;erhalb der Grenze in W&auml;rmepumpe Z1",                                                                                                                                                                                                                                                                         checklist => "Zu niedrige Verdampftemperatur im Verh&auml;ltnis zur Kondensationstemperatur",                                                                                                                                                                                                                                                                                                                                                                                                                       correction => "Warscheinlich zu niedrige Au&szlig;entemperatur. Siehe Diagramm Kapitel 4 in Installationshandbuch"                                                                                                                                                                                                                                                                                              };
	$ErrorCodeList{"0470"} = {errorcode => "A01",                                    subcode => "5514",                                               class => "WPL AR",      reason => "Warnung niedriger Druck JR0 in W&auml;rmepumpe Z1",                                                                                                                                                                                                                                                                                                checklist => "Blockierter/niedriger Luftstrom zu Luftw&auml;rmetauscher an W&auml;rmepumpe&bull;Gebl&auml;se au&szlig;er Funktion&bull;Gebl&auml;se hat nicht 230 V Spannung&bull;Gebl&auml;se ohne 0: 10 V-Signal&bull;Gebl&auml;se defekt&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                   correction => "Ausreichenden Luftstrom &uuml;ber Luftw&auml;rmetauscher sicherstellen&bull;Test Au&szlig;eneinheit aktivieren&bull;&Uuml;berpr&uuml;fen, dass 230 V Spannung an Ausgang PL3 Klemme 32 (78, N)&bull;0: 10 V Spannung an Ausgang PL3 PWM Klemme 20 (20, 26) laut Test Au&szlig;eneinheit &uuml;berpr&uuml;fen&bull;Ist Spannung wie oben vorhanden, Gebl&auml;se wechseln&bull;Fehlt Spannung wie oben, I/O-Board wechseln"};
	$ErrorCodeList{"0471"} = {errorcode => "A01",                                    subcode => "5517",                                               class => "WPL AR",      reason => "Falscher Kompressortreiber in W&auml;rmepumpe Z2 installiert",                                                                                                                                                                                                                                                                                     checklist => "Falsche Kombination aus Inverter/I/O-Board in W&auml;rmepumpe 2 installiert&bull;Bei Wechsel des Inverters, lasches Invertermodell&bull;Beim Wechsel des I/O-Boards, falsche Einstellung des Dreh-Encoders auf neuem I/OBoard",                                                                                                                                                                                                                                                                       correction => "Einstellung im Vergleich zum aktuellen elektrischen Schaltplan &uuml;berpr&uuml;fen&bull;Invertermodell &uuml;berpr&uuml;fen&bull;Einstellung des Dreh-Encoders im Vergleich zum ausgetauschten I/O-Board &uuml;berpr&uuml;fen"                                                                                                                                                                  };
	$ErrorCodeList{"0472"} = {errorcode => "A01",                                    subcode => "5519",                                               class => "WPL AR",      reason => "Zu hohe Kondensation in W&auml;rmepumpe Z2",                                                                                                                                                                                                                                                                                                       checklist => "Zu niedrige Verdampftemperatur im Verh&auml;ltnis zur Kondensationstemperatur&bull;Zu hohe Stromrate/niedriges Delta im Heizzystem",                                                                                                                                                                                                                                                                                                                                                                  correction => "Warscheinlich zu niedrige Au&szlig;entemperatur. Siehe Diagramm Kapitel 4 in Installationshandbuch&bull;Temperatur/Stromrate einstellen"                                                                                                                                                                                                                                                         };
	$ErrorCodeList{"0473"} = {errorcode => "A01",                                    subcode => "5521",                                               class => "WPL AR",      reason => "Alarm geringe Verdampfung JR0 in W&auml;rmepumpe Z2",                                                                                                                                                                                                                                                                                              checklist => "Siehe Pr&uuml;fvorgang/Ursachen f&uuml;r Ursachencode 5520",                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5520"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0474"} = {errorcode => "A01",                                    subcode => "5522",                                               class => "WPL AR",      reason => "Falsche Kombination aus W&auml;rmepumpe und Innenteil",                                                                                                                                                                                                                                                                                            checklist => "Falsche Kombination aus W&auml;rmepumpe/ Innenteil&bull;Beim Wechsel des I/O-Boards,; falsche Einstellung des Dreh-Encoders auf neuem I/O-Board&bull;Beim Wechsel des Installer-Boards, falsche Einstellung des Dreh-Encoders auf neuem Installer-Board",                                                                                                                                                                                                                                             correction => "Kombination &uuml;berpr&uuml;fen&bull;Einstellung des Dreh-Encoders im Vergleich zum ausgetauschten I/O-Board &uuml;berpr&uuml;fen&bull;Einstellung des Dreh-Encoders im Vergleich zum ausgetauschten Installer-Board &uuml;berpr&uuml;fen"                                                                                                                                                      };
	$ErrorCodeList{"0475"} = {errorcode => "A01",                                    subcode => "5524",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z1 PFC &Uuml;berstrom des Kompressortreibers",                                                                                                                                                                                                                                                                               checklist => "Siehe Pr&uuml;fvorgang/Ursachen f&uuml;r Ursachencode 5523",                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5523"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0476"} = {errorcode => "A01",                                    subcode => "5526",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z2 defr Au&szlig;enteil der; W&auml;rmepumpe reinigen",                                                                                                                                                                                                                                                                      checklist => "Verdampfer vereist",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Eis vorsichtig mit Warmwasser schmelzen"                                                                                                                                                                                                                                                                                                                                                         };
	$ErrorCodeList{"0477"} = {errorcode => "A01",                                    subcode => "5528",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z2 Kompressorf&uuml;hler TR1 Unterbrechung",                                                                                                                                                                                                                                                                               checklist => "Schraubklemme an I/O-Board zu Kompressorf&uuml;hler TR1 nicht angezogen&bull;Kompressorf&uuml;hler TR3/Signalkabel Unterbrechung&bull;I/O-Board au&szlig;er Funktion&bull;Kompressorf&uuml;hler TR1/Signalkabel kurzgeschlossen",                                                                                                                                                                                                                                                                     correction => "Schrauklemme &uuml;berpr&uuml;fen&bull;Kompressorf&uuml;hler TR1 wechseln&bull;I/O-Board wechseln&bull;Ohmwert von F&uuml;hler TR1/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle und der tats&auml;chlichen Temperatur, getrennt von der I/OBoard-Klemme, &uuml;berpr&uuml;fen."                                                                                                     };
	$ErrorCodeList{"0478"} = {errorcode => "A01",                                    subcode => "5530",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z2 Kompressorf&uuml;hler TR1 defekt",                                                                                                                                                                                                                                                                                        checklist => "Kompressorf&uuml;hler TR1/Signalkabel Unterbrechung",                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "Ohmwert von F&uuml;hler TR1/Signalkabel; im Vergleich zur aktuellen F&uuml;hlertabelle und der tats&auml;chlichen Temperatur, getrennt von der I/OBoard-Klemme, &uuml;berpr&uuml;fen"                                                                                                                                                                                                            };
	$ErrorCodeList{"0479"} = {errorcode => "A01",                                    subcode => "5531",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z1 Kompressorf&uuml;hler TR1 Unterbrechung",                                                                                                                                                                                                                                                                               checklist => "Schraubklemme an I/O-Board zu Kompressorf&uuml;hler TR1 nicht angezogen&bull;Kompressorf&uuml;hler TR3/Signalkabel Unterbrechung&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                                                                                correction => "Schrauklemme &uuml;berpr&uuml;fen&bull;Kompressorf&uuml;hler TR1 wechseln&bull;I/O-Board wechseln"                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0480"} = {errorcode => "A01",                                    subcode => "5532",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z1 Kompressorf&uuml;hler TR1 Kurzschluss",                                                                                                                                                                                                                                                                                 checklist => "Kompressorf&uuml;hler TR1/Signalkabel kurzgeschlossen&bull;Kompressorf&uuml;hler TR1 kurzgeschlossen&bull;Signalkabel zu Kompressorf&uuml;hler TR1 kurzgeschlossen&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                                              correction => "Ohmwert von F&uuml;hler TR1/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle und der tats&auml;chlichen Temperatur, getrennt von der I/OBoard-Klemme, &uuml;berpr&uuml;fen&bull;Ohmwert im Vergleich zur aktuellen; F&uuml;hlertabelle und tats&auml;chlichen Temperatur &uuml;berpr&uuml;fen/F&uuml;hler TR1 wechseln&bull;Signalkabel zu Kompressorf&uuml;hler TR1 reparieren/wechseln&bull;I/O-Board wechseln"};
	$ErrorCodeList{"0481"} = {errorcode => "A01",                                    subcode => "5534",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z2 Temperaturf&uuml;hler TA4 Unterbrechung",                                                                                                                                                                                                                                                                               checklist => "Schrauklemme an I/O-Board zu F&uuml;hler TA4 nicht angezogen&bull;F&uuml;hler TA4/Signalkabel Unterbrechung&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                                                                                                     correction => "Schrauklemme an I/O-Board &uuml;berpr&uuml;fen&bull;Byt F&uuml;hler/Signalkabel F&uuml;hler TA4&bull;I/O-Board wechseln"                                                                                                                                                                                                                                                                         };
	$ErrorCodeList{"0482"} = {errorcode => "A01",                                    subcode => "5535",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z2 Temperaturf&uuml;hler TA4 kurzgeschlossen",                                                                                                                                                                                                                                                                             checklist => "F&uuml;hler TA4/Signalkabel kurzgeschlossen&bull;F&uuml;hler TA4 kurzgeschlossen&bull;Signalkabel/F&uuml;hler TA4 kurzgeschlossen&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                                                                               correction => "Ohmwert von F&uuml;hler TA4/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle und der tats&auml;chlichen Temperatur, getrennt von der I/OBoard-Klemme, &uuml;berpr&uuml;fen&bull;Ohmwert im Vergleich zur aktuellen; F&uuml;hler-tabelle und tats&auml;chlichen Temperatur &uuml;ber-pr&uuml;fen/F&uuml;hler TA4 wechseln&bull;Signalkabel/F&uuml;hler zu F&uuml;hler TA4 wechseln&bull;I/O-Board wechseln"};
	$ErrorCodeList{"0483"} = {errorcode => "A01",                                    subcode => "5538",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z1 Temperaturf&uuml;hler TA4 Unterbrechung",                                                                                                                                                                                                                                                                               checklist => "Schrauklemme an I/O-Board zu F&uuml;hler TA4 nicht angezogen&bull;F&uuml;hler TA4/Signalkabel Unterbrechung&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                                                                                                     correction => "Schrauklemme an I/O-Board &uuml;berpr&uuml;fen&bull;Byt F&uuml;hler/Signalkabel F&uuml;hler TA4&bull;I/O-Board wechseln"                                                                                                                                                                                                                                                                         };
	$ErrorCodeList{"0484"} = {errorcode => "A01",                                    subcode => "5539",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z1 Temperaturf&uuml;hler TA4 kurzgeschlossen",                                                                                                                                                                                                                                                                             checklist => "F&uuml;hler TA4/Signalkabel kurzgeschlossen&bull;F&uuml;hler TA4 kurzgeschlossen&bull;Signalkabel/F&uuml;hler TA4 kurzgeschlossen&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                                                                               correction => "Ohmwert von F&uuml;hler TA4/Signalkabel im Vergleich zur aktuellen F&uuml;hlertabelle und der tats&auml;chlichen Temperatur, getrennt von der I/OBoard-Klemme, &uuml;berpr&uuml;fen&bull;Ohmwert im Vergleich zur aktuellen; F&uuml;hlertabelle und tats&auml;chlichen Temperatur &uuml;berpr&uuml;fen/F&uuml;hler TA4 wechseln&bull;Signalkabel/F&uuml;hler zu F&uuml;hler TA4 wechseln&bull;I/O-Board wechseln"};
	$ErrorCodeList{"0485"} = {errorcode => "A01",                                    subcode => "5542",                                               class => "8",           reason => "Warnung Z2 Versorgungsspannung zu niedrig",                                                                                                                                                                                                                                                                                                        checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0486"} = {errorcode => "A01",                                    subcode => "5543",                                               class => "4",           reason => "Warnung Z1 &Uuml;berhitzungstemperatur ist &uuml;ber dem Optimum",                                                                                                                                                                                                                                                                                 checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0487"} = {errorcode => "A01",                                    subcode => "5544",                                               class => "4",           reason => "Warnung Z2 &Uuml;berhitzungstemperatur ist &uuml;ber dem Optimum",                                                                                                                                                                                                                                                                                 checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0488"} = {errorcode => "A01",                                    subcode => "5547",                                               class => "4",           reason => "Warnung Z1 Sole-Eintrittstemperatur an TB0 ist zu niedrig",                                                                                                                                                                                                                                                                                        checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0489"} = {errorcode => "A01",                                    subcode => "5548",                                               class => "4",           reason => "Warnung Z2 Sole-Eintrittstemperatur an TB0 ist zu niedrig",                                                                                                                                                                                                                                                                                        checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0490"} = {errorcode => "A01",                                    subcode => "5551",                                               class => "4",           reason => "Warnung Z1 Sole-Austrittstemperatur an TB1 ist zu niedrig",                                                                                                                                                                                                                                                                                        checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0491"} = {errorcode => "A01",                                    subcode => "5552",                                               class => "4",           reason => "Warnung Z2 Sole-Austrittstemperatur an TB1 ist zu niedrig",                                                                                                                                                                                                                                                                                        checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0492"} = {errorcode => "A01",                                    subcode => "5555",                                               class => "4",           reason => "Warnung Z1 Sole-Eintrittstemperatur an TB0 ist zu hoch",                                                                                                                                                                                                                                                                                           checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0493"} = {errorcode => "A01",                                    subcode => "5556",                                               class => "4",           reason => "Warnung Z2 Sole-Eintrittstemperatur an TB0 ist zu hoch",                                                                                                                                                                                                                                                                                           checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0494"} = {errorcode => "A01",                                    subcode => "5559",                                               class => "4",           reason => "Warnung Z1 Sole-Aus-trittstemperatur an TB1; ist zu hoch",                                                                                                                                                                                                                                                                                         checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0495"} = {errorcode => "A01",                                    subcode => "5560",                                               class => "4",           reason => "Warnung Z2 Sole-Aus-trittstemperatur an TB1; ist zu hoch",                                                                                                                                                                                                                                                                                         checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0496"} = {errorcode => "A01",                                    subcode => "5563",                                               class => "8",           reason => "Warnung Z1 Hohe Temperaturdiff. zwischen TB0 und TB1",                                                                                                                                                                                                                                                                                             checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0497"} = {errorcode => "A01",                                    subcode => "5564",                                               class => "8",           reason => "Warnung Z2 Hohe Temperaturdiff. zwischen TB0 und TB1",                                                                                                                                                                                                                                                                                             checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0498"} = {errorcode => "A01",                                    subcode => "5567",                                               class => "8",           reason => "Warnung Z1 St&ouml;rung Solepumpe",                                                                                                                                                                                                                                                                                                                checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0499"} = {errorcode => "A01",                                    subcode => "5568",                                               class => "8",           reason => "Warnung Z2 St&ouml;rung Solepumpe",                                                                                                                                                                                                                                                                                                                checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0500"} = {errorcode => "A01",                                    subcode => "5571",                                               class => "4",           reason => "Warnung Z1 Sole-Eintrittstemperatursensor TB0 Kurzschluss",                                                                                                                                                                                                                                                                                        checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0501"} = {errorcode => "A01",                                    subcode => "5572",                                               class => "4",           reason => "Warnung Z2 Sole-Eintrittstemperatursensor TB0 Kurzschluss",                                                                                                                                                                                                                                                                                        checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0502"} = {errorcode => "A01",                                    subcode => "5573",                                               class => "4",           reason => "Warnung Z1 Sole-Eintrittstemperatursensor TB0 Kabelbruch",                                                                                                                                                                                                                                                                                         checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0503"} = {errorcode => "A01",                                    subcode => "5574",                                               class => "4",           reason => "Warnung Z2 Sole-Eintrittstemperatursensor TB0 Kabelbruch",                                                                                                                                                                                                                                                                                         checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0504"} = {errorcode => "A01",                                    subcode => "5577",                                               class => "4",           reason => "Warnung Z1 Sole-Austrittstemperatursensor TB1 Kurzschluss",                                                                                                                                                                                                                                                                                        checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0505"} = {errorcode => "A01",                                    subcode => "5578",                                               class => "4",           reason => "Warnung Z2 Sole-Austrittstemperatursensor TB1 Kurzschluss",                                                                                                                                                                                                                                                                                        checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0506"} = {errorcode => "A01",                                    subcode => "5579",                                               class => "4",           reason => "Warnung Z1 Sole-Austrittstemperatursensor TB1 Kabelbruch",                                                                                                                                                                                                                                                                                         checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0507"} = {errorcode => "A01",                                    subcode => "5580",                                               class => "4",           reason => "Warnung Z2 Sole-Austrittstemperatursensor TB1 Kabelbruch",                                                                                                                                                                                                                                                                                         checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0508"} = {errorcode => "A01",                                    subcode => "5587",                                               class => "4",           reason => "Warnung Vorlauftemperaturf&uuml;hler/K&uuml;hlung TK1 Kabelbruch",                                                                                                                                                                                                                                                                                 checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0509"} = {errorcode => "A01",                                    subcode => "5588",                                               class => "4",           reason => "Warnung Vorlauftstemperaturf&uuml;hler/K&uuml;hlung TK1 Kurzschluss",                                                                                                                                                                                                                                                                              checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0510"} = {errorcode => "A01",                                    subcode => "5590",                                               class => "4",           reason => "Warnung Eingangstemperaturf&uuml;hler/K&uuml;hlung TK2 Kabelbruch",                                                                                                                                                                                                                                                                                checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0511"} = {errorcode => "A01",                                    subcode => "5591",                                               class => "4",           reason => "Warnung Eingangstemperaturf&uuml;hler/K&uuml;hlung TK2 Kurzschluss",                                                                                                                                                                                                                                                                               checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0512"} = {errorcode => "A01",                                    subcode => "5597",                                               class => "4",           reason => "Warnung Z1 Verdampfertemp.-sensor TR4 Kurzschluss",                                                                                                                                                                                                                                                                                                checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0513"} = {errorcode => "A01",                                    subcode => "5598",                                               class => "4",           reason => "Warnung Z1 Verdampfertemp.-sensor TR4 Kabelbruch",                                                                                                                                                                                                                                                                                                 checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0514"} = {errorcode => "A01",                                    subcode => "5600",                                               class => "4",           reason => "Warnung Kompressor blockiert (IPM Fehler)",                                                                                                                                                                                                                                                                                                        checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0515"} = {errorcode => "A01",                                    subcode => "5602",                                               class => "4",           reason => "Warnung Stromaufnahme Inverter zu gro&szlig;",                                                                                                                                                                                                                                                                                                     checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0516"} = {errorcode => "A01",                                    subcode => "5604",                                               class => "4",           reason => "Warnung Gleichspannung au&szlig;erhalb zul&auml;ssigen Bereichs",                                                                                                                                                                                                                                                                                  checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0517"} = {errorcode => "A01",                                    subcode => "5606",                                               class => "4",           reason => "Warnung Hochoder Niederdruckschalter hat ausgel&ouml;st",                                                                                                                                                                                                                                                                                          checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0518"} = {errorcode => "A01",                                    subcode => "5608",                                               class => "4",           reason => "Warnung Eingangssp. zu klein (<140VAC) oder zu hoch (>300VAC)",                                                                                                                                                                                                                                                                                    checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0519"} = {errorcode => "A01",                                    subcode => "5610",                                               class => "4",           reason => "Warnung Kompressor-phasenfolge falsch angeschlossen",                                                                                                                                                                                                                                                                                              checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0520"} = {errorcode => "A01",                                    subcode => "5612",                                               class => "4",           reason => "Warnung Zu hohe Stromaufnahme am IGBT, Spule pr&uuml;fen",                                                                                                                                                                                                                                                                                         checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0521"} = {errorcode => "A01",                                    subcode => "5614",                                               class => "4",           reason => "Warnung Gleichspannung gr&ouml;&szlig;er als 420 VDC.",                                                                                                                                                                                                                                                                                            checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0522"} = {errorcode => "A01",                                    subcode => "5616",                                               class => "4",           reason => "Warnung Stromaufnahme des Kompressors zu gro&szlig;",                                                                                                                                                                                                                                                                                              checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0523"} = {errorcode => "A01",                                    subcode => "5618",                                               class => "4",           reason => "Warnung Heissgastemperatur zu hoch",                                                                                                                                                                                                                                                                                                               checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0524"} = {errorcode => "A01",                                    subcode => "5620",                                               class => "4",           reason => "Warnung CT Sensor Fehler",                                                                                                                                                                                                                                                                                                                         checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0525"} = {errorcode => "A01",                                    subcode => "5622",                                               class => "4",           reason => "Warnung Heissgassensor TR6 Fehler",                                                                                                                                                                                                                                                                                                                checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0526"} = {errorcode => "A01",                                    subcode => "5624",                                               class => "4",           reason => "Warnung Lufttemperatursensor  Ausseneinheit TL2 Fehler",                                                                                                                                                                                                                                                                                           checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0527"} = {errorcode => "A01",                                    subcode => "5626",                                               class => "4",           reason => "Warnung VerdampferMitteltemp.-sensor TR8 Fehler",                                                                                                                                                                                                                                                                                                  checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0528"} = {errorcode => "A01",                                    subcode => "5628",                                               class => "4",           reason => "Warnung Sauggastemperatursensor TR5 Fehler",                                                                                                                                                                                                                                                                                                       checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0529"} = {errorcode => "A01",                                    subcode => "5630",                                               class => "4",           reason => "Warnung Nennsparnnung zu hoch Thermische Belastung zu hoch",                                                                                                                                                                                                                                                                                       checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0530"} = {errorcode => "A01",                                    subcode => "5632",                                               class => "4",           reason => "Warnung Interne InverterKommunikation gest&ouml;rt",                                                                                                                                                                                                                                                                                               checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0531"} = {errorcode => "A01",                                    subcode => "5634",                                               class => "4",           reason => "Warnung Interne CAN-Kommunikation der Au&szlig;eneinheit gest&ouml;rt",                                                                                                                                                                                                                                                                            checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0532"} = {errorcode => "A01",                                    subcode => "5636",                                               class => "4",           reason => "Warnung Phasenreihenfolge falsch oder Phase fehlt",                                                                                                                                                                                                                                                                                                checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0533"} = {errorcode => "A01",                                    subcode => "5638",                                               class => "4",           reason => "Warnung Defektes EEPROM auf der Platine in Au&szlig;eneinheit",                                                                                                                                                                                                                                                                                    checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0534"} = {errorcode => "A01",                                    subcode => "5640",                                               class => "4",           reason => "Warnung Kondensationstemp. zu hoch (gr&ouml;&szlig;er 65°C)",                                                                                                                                                                                                                                                                                     checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0535"} = {errorcode => "A01",                                    subcode => "5642",                                               class => "4",           reason => "Warnung Temperatur am Frequenzumrichter zu hoch (gr&ouml;&szlig;er 85°C)",                                                                                                                                                                                                                                                                        checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0536"} = {errorcode => "A01",                                    subcode => "5644",                                               class => "4",           reason => "Warnung Kondensationstemp. zu niedrig",                                                                                                                                                                                                                                                                                                            checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0537"} = {errorcode => "A01",                                    subcode => "5646",                                               class => "4",           reason => "Warnung Frequenzumrichter Temperaturf&uuml;hler Fehler",                                                                                                                                                                                                                                                                                           checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0538"} = {errorcode => "A01",                                    subcode => "5648",                                               class => "4",           reason => "Warnung Kompressor Anlaufstrom zu hoch",                                                                                                                                                                                                                                                                                                           checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0539"} = {errorcode => "A01",                                    subcode => "5650",                                               class => "4",           reason => "Warnung Z1 Kondensatoraustrittstemperaturf&uuml;hler TR7 Kabelbruch",                                                                                                                                                                                                                                                                              checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0540"} = {errorcode => "A01",                                    subcode => "5651",                                               class => "4",           reason => "Warnung Z1 Kondensatoraustrittstemperaturf&uuml;hler TR7 Kurzschluss",                                                                                                                                                                                                                                                                             checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0541"} = {errorcode => "A01",                                    subcode => "5653",                                               class => "4",           reason => "Warnung Z2 Kondensatoraustrittstemperaturf&uuml;hler TR7 Kabelbruch",                                                                                                                                                                                                                                                                              checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0542"} = {errorcode => "A01",                                    subcode => "5654",                                               class => "4",           reason => "Warnung Z2 Kondensatoraustrittstemperaturf&uuml;hler TR7 Kurzschluss",                                                                                                                                                                                                                                                                             checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0543"} = {errorcode => "A01",                                    subcode => "5656",                                               class => "4",           reason => "Warnung Z1 Kondensatoreintrittstemperaturf&uuml;hler TR3 Kabelbruch",                                                                                                                                                                                                                                                                              checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0544"} = {errorcode => "A01",                                    subcode => "5657",                                               class => "4",           reason => "Warnung Z1 Kondensatoreintrittstemperaturf&uuml;hler TR3 Kurzschluss",                                                                                                                                                                                                                                                                             checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0545"} = {errorcode => "A01",                                    subcode => "5659",                                               class => "4",           reason => "Warnung Z2 Kondensatoreintrittstemperaturf&uuml;hler TR3 Kabelbruch",                                                                                                                                                                                                                                                                              checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0546"} = {errorcode => "A01",                                    subcode => "5660",                                               class => "4",           reason => "Warnung Z2 Kondensatoreintrittstemperaturf&uuml;hler TR3 Kurzschluss",                                                                                                                                                                                                                                                                             checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0547"} = {errorcode => "A01",                                    subcode => "5662",                                               class => "4",           reason => "Warnung Z1 Drucksensor Kabelbruch oder Kurzschluss",                                                                                                                                                                                                                                                                                               checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0548"} = {errorcode => "A01",                                    subcode => "5664",                                               class => "4",           reason => "Warnung Z1 Eintritt Verdampfertemperaturf&uuml;hler TR4 Kabelbruch oder Kurzschluss",                                                                                                                                                                                                                                                              checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0549"} = {errorcode => "A01",                                    subcode => "5666",                                               class => "4",           reason => "Warnung Z1 Verdampfungsdruck zu niedrig",                                                                                                                                                                                                                                                                                                          checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0550"} = {errorcode => "A01",                                    subcode => "5668",                                               class => "4",           reason => "Warnung Z1 Kommunikation zwischen Innenund Au&szlig;eneinheit gest&ouml;rt",                                                                                                                                                                                                                                                                       checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0551"} = {errorcode => "A01",                                    subcode => "5670",                                               class => "4",           reason => "Warnung Z1 Ventilator Au&szlig;eneinheit ist blockiert",                                                                                                                                                                                                                                                                                           checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0552"} = {errorcode => "A01",                                    subcode => "5672",                                               class => "4",           reason => "Warnung Z2 Drucksensor Kabelbruch oder Kurzschluss",                                                                                                                                                                                                                                                                                               checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0553"} = {errorcode => "A01",                                    subcode => "5674",                                               class => "4",           reason => "Warnung Z2 Eintritt Verdampfertemperaturf&uuml;hler TR4 Kabelbruch oder Kurzschluss",                                                                                                                                                                                                                                                              checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0554"} = {errorcode => "A01",                                    subcode => "5676",                                               class => "4",           reason => "Warnung Z2 Druck zu niedrig",                                                                                                                                                                                                                                                                                                                      checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0555"} = {errorcode => "A01",                                    subcode => "5678",                                               class => "4",           reason => "Warnung Z2 Kommunikation zwischen Innenund Au&szlig;eneinheit gest&ouml;rt",                                                                                                                                                                                                                                                                       checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0556"} = {errorcode => "A01",                                    subcode => "5680",                                               class => "4",           reason => "Warnung Z2 Ventilator Au&szlig;eneinheit ist blockiert",                                                                                                                                                                                                                                                                                           checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0557"} = {errorcode => "A01",                                    subcode => "5700",                                               class => "4",           reason => "Warnung Z1 Vorlauftemperatursensor TC2 Fehler",                                                                                                                                                                                                                                                                                                    checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0558"} = {errorcode => "A01",                                    subcode => "5701",                                               class => "4",           reason => "Warnung Z1 Kurzschluss Vorlauftemperaturf&uuml;hler TC2",                                                                                                                                                                                                                                                                                          checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0559"} = {errorcode => "A01",                                    subcode => "5703",                                               class => "4",           reason => "Warnung Z1 R&uuml;cklauftemperatursensor TC4 Fehler",                                                                                                                                                                                                                                                                                              checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0560"} = {errorcode => "A01",                                    subcode => "5704",                                               class => "4",           reason => "Warnung Z1 Kurzschluss R&uuml;cklauftemperaturf&uuml;hler TC4",                                                                                                                                                                                                                                                                                    checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0561"} = {errorcode => "A01",                                    subcode => "5706",                                               class => "4",           reason => "Warnung Z1 Einspritztemperatursensor TR2  Fehler",                                                                                                                                                                                                                                                                                                 checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0562"} = {errorcode => "A01",                                    subcode => "5707",                                               class => "4",           reason => "Warnung Z1 Kurzschluss Einspritztemperaturf&uuml;hler TR2",                                                                                                                                                                                                                                                                                        checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0563"} = {errorcode => "A01",                                    subcode => "5708",                                               class => "4",           reason => "Warnung Z1 Einspritztemperatursensor TR2  Fehler",                                                                                                                                                                                                                                                                                                 checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0564"} = {errorcode => "A01",                                    subcode => "5709",                                               class => "4",           reason => "Warnung Z1 Unterbrechung Temperaturf&uuml;hler Luftaustritt TL3",                                                                                                                                                                                                                                                                                  checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0565"} = {errorcode => "A01",                                    subcode => "5710",                                               class => "4",           reason => "Warnung Z1 Kurzschluss Temperaturf&uuml;hler Luftaustritt TL3",                                                                                                                                                                                                                                                                                    checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0566"} = {errorcode => "A01",                                    subcode => "5713",                                               class => "4",           reason => "Warnung Z1 Mitteldrucksensor Fehler",                                                                                                                                                                                                                                                                                                              checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0567"} = {errorcode => "A01",                                    subcode => "5717",                                               class => "4",           reason => "Warnung Z1 K&auml;ltemittelleckage. Expansionsventil arbeitet nicht richtig",                                                                                                                                                                                                                                                                      checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0568"} = {errorcode => "A01",                                    subcode => "5718",                                               class => "4",           reason => "Warnung Z1 Inverterfehler Spannungsdiff. zw den drei Eingangs-Phasen",                                                                                                                                                                                                                                                                             checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0569"} = {errorcode => "A01",                                    subcode => "5719",                                               class => "4",           reason => "Warnung Z1 Inverterfehler Ents&auml;ttigung",                                                                                                                                                                                                                                                                                                      checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0570"} = {errorcode => "A01",                                    subcode => "5720",                                               class => "4",           reason => "Warnung Z1 Inverterfehler Interner Rechenfehler",                                                                                                                                                                                                                                                                                                  checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0571"} = {errorcode => "A01",                                    subcode => "5721",                                               class => "4",           reason => "Warnung Z1 Inverterfehler Eingangsrelais offen",                                                                                                                                                                                                                                                                                                   checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0572"} = {errorcode => "A01",                                    subcode => "5722",                                               class => "4",           reason => "Warnung Z1 Inverterfehler Stromunterschiede zw. den drei IGBT`s",                                                                                                                                                                                                                                                                                  checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0573"} = {errorcode => "A01",                                    subcode => "5723",                                               class => "4",           reason => "Warnung Z1 Inverterfehler Stromunterschiede zw. den drei PFC IGBT`s.",                                                                                                                                                                                                                                                                             checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0574"} = {errorcode => "A01",                                    subcode => "5724",                                               class => "4",           reason => "Warnung Z1 Inverterfehler Kleinspannung Bereichs&uuml;berschreitung",                                                                                                                                                                                                                                                                              checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0575"} = {errorcode => "A01",                                    subcode => "5725",                                               class => "4",           reason => "Warnung Z1 Inverterfehler Motordrehzahl zu hoch",                                                                                                                                                                                                                                                                                                  checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0576"} = {errorcode => "A01",                                    subcode => "5726",                                               class => "4",           reason => "Warnung Z1 Inverterfehler DC Spannung zu niedrig",                                                                                                                                                                                                                                                                                                 checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0577"} = {errorcode => "A01",                                    subcode => "5727",                                               class => "4",           reason => "Warnung Z1 Inverterfehler Drehmoment Grenze erreicht",                                                                                                                                                                                                                                                                                             checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0578"} = {errorcode => "A01",                                    subcode => "5728",                                               class => "4",           reason => "Warnung Z1 Inverterfehler Modbus Kommunikation gest&ouml;rt",                                                                                                                                                                                                                                                                                      checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0579"} = {errorcode => "A01",                                    subcode => "5729",                                               class => "4",           reason => "Warnung Z1 Inverterfehler Verdichter-Scroll &Uuml;bertemperatur",                                                                                                                                                                                                                                                                                  checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0580"} = {errorcode => "A01",                                    subcode => "5730",                                               class => "4",           reason => "Warnung Z1 Inverterfehler Verdichter-Motor &Uuml;bertemperatur",                                                                                                                                                                                                                                                                                   checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0581"} = {errorcode => "A01",                                    subcode => "5731",                                               class => "4",           reason => "Warnung Z1 Inverterfehler Schaltkreis &Uuml;bertemperatur",                                                                                                                                                                                                                                                                                        checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0582"} = {errorcode => "A01",                                    subcode => "5732",                                               class => "4",           reason => "Warnung Z1 Inverterfehler IGBT`s &Uuml;bertemperatur",                                                                                                                                                                                                                                                                                             checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0583"} = {errorcode => "A01",                                    subcode => "5733",                                               class => "4",           reason => "Warnung Z1 Inverterfehler PFC IGBT's &Uuml;bertemperatur",                                                                                                                                                                                                                                                                                         checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0584"} = {errorcode => "A01",                                    subcode => "5734",                                               class => "4",           reason => "Warnung Z1 Inverter Kommunikation. zw. Signalund Hauptprozessor gest&ouml;rt",                                                                                                                                                                                                                                                                     checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0585"} = {errorcode => "A01",                                    subcode => "5735",                                               class => "4",           reason => "Warnung Z1 Inverter Kommunikation. zw. Hauptund Signalprozessor gest&ouml;rt",                                                                                                                                                                                                                                                                     checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0586"} = {errorcode => "A01",                                    subcode => "5736",                                               class => "4",           reason => "Warnung Z1 Temperaturdifferenz zw. den drei Inverter IGBT`s",                                                                                                                                                                                                                                                                                      checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0587"} = {errorcode => "A01",                                    subcode => "5737",                                               class => "4",           reason => "Warnung Z1 Temperaturdifferenz zw. den drei PFC IGBT`s",                                                                                                                                                                                                                                                                                           checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0588"} = {errorcode => "A01",                                    subcode => "5738",                                               class => "4",           reason => "Warnung Z1 Inverterfehler Verdichter-Scroll Temp. zu niedrig",                                                                                                                                                                                                                                                                                     checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0589"} = {errorcode => "A01",                                    subcode => "5739",                                               class => "4",           reason => "Warnung Z1 Inverterfehler Verdichter-Motor Temp. zu niedrig",                                                                                                                                                                                                                                                                                      checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0590"} = {errorcode => "A01",                                    subcode => "5740",                                               class => "4",           reason => "Warnung Z1 Inverterfehler Schaltkreis Temp. zu niedrig",                                                                                                                                                                                                                                                                                           checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0591"} = {errorcode => "A01",                                    subcode => "5741",                                               class => "4",           reason => "Warnung Z1 Inverterfehler IGBT`s Temp. zu niedrig",                                                                                                                                                                                                                                                                                                checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0592"} = {errorcode => "A01",                                    subcode => "5742",                                               class => "4",           reason => "Warnung Z1 Inverterfehler PFC IGBT's Temp. zu niedrig",                                                                                                                                                                                                                                                                                            checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0593"} = {errorcode => "A01",                                    subcode => "5743",                                               class => "4",           reason => "Warnung Z1 Inverter Kommunikation zw. AD-Wandler und Nebenprozessor gest&ouml;rt",                                                                                                                                                                                                                                                                 checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0594"} = {errorcode => "A01",                                    subcode => "5745",                                               class => "4",           reason => "Warnung Z1 Niederdrucksensor JR0 Druck zu hoch",                                                                                                                                                                                                                                                                                                   checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0595"} = {errorcode => "A01",                                    subcode => "5746",                                               class => "4",           reason => "Warnung Z1 Niederdrucksensor JR0 Druck zu niedrig",                                                                                                                                                                                                                                                                                                checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0596"} = {errorcode => "A01",                                    subcode => "5747",                                               class => "4",           reason => "Warnung Z1 Hochdrucksensor JR1 Druck zu niedrig",                                                                                                                                                                                                                                                                                                  checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0597"} = {errorcode => "A01",                                    subcode => "5748",                                               class => "4",           reason => "Warnung Z2 Vorlauftemperatursensor TC2 Fehler",                                                                                                                                                                                                                                                                                                    checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0598"} = {errorcode => "A01",                                    subcode => "5749",                                               class => "4",           reason => "Warnung Z2 Kurzschluss Vorlauftemperaturf&uuml;hler TC2",                                                                                                                                                                                                                                                                                          checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0599"} = {errorcode => "A01",                                    subcode => "5751",                                               class => "4",           reason => "Warnung Z2 R&uuml;cklauftemperatursensor TC4 Fehler",                                                                                                                                                                                                                                                                                              checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0600"} = {errorcode => "A01",                                    subcode => "5752",                                               class => "4",           reason => "Warnung Z2 Kurzschluss R&uuml;cklauftemperaturf&uuml;hler TC4",                                                                                                                                                                                                                                                                                    checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0601"} = {errorcode => "A01",                                    subcode => "5754",                                               class => "4",           reason => "Warnung Z2 Einspritztemperatursensor TR2 Fehler",                                                                                                                                                                                                                                                                                                  checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0602"} = {errorcode => "A01",                                    subcode => "5755",                                               class => "4",           reason => "Warnung Z2 Kurzschluss Einspritztemperaturf&uuml;hler TR2",                                                                                                                                                                                                                                                                                        checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0603"} = {errorcode => "A01",                                    subcode => "5757",                                               class => "4",           reason => "Warnung Z2 Unterbrechung Temperaturf&uuml;hler Luftaustritt TL3",                                                                                                                                                                                                                                                                                  checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0604"} = {errorcode => "A01",                                    subcode => "5758",                                               class => "4",           reason => "Warnung Z2 Kurzschluss Temperaturf&uuml;hler Luftaustritt TL3",                                                                                                                                                                                                                                                                                    checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0605"} = {errorcode => "A01",                                    subcode => "5761",                                               class => "4",           reason => "Warnung Z2 Mitteldrucksensor Fehler",                                                                                                                                                                                                                                                                                                              checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0606"} = {errorcode => "A01",                                    subcode => "5765",                                               class => "4",           reason => "Warnung Z2 K&auml;ltemittelleckage. Expansionsventil arbeitet nicht richtig",                                                                                                                                                                                                                                                                      checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0607"} = {errorcode => "A01",                                    subcode => "5766",                                               class => "4",           reason => "Warnung Z2 Inverterfehler Spannungsdiff. zw den drei Eingangs-Phasen",                                                                                                                                                                                                                                                                             checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0608"} = {errorcode => "A01",                                    subcode => "5767",                                               class => "4",           reason => "Warnung Z2 Inverterfehler Ents&auml;ttigung",                                                                                                                                                                                                                                                                                                      checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0609"} = {errorcode => "A01",                                    subcode => "5768",                                               class => "4",           reason => "Warnung Z2 Inverterfehler Interner Rechenfehler",                                                                                                                                                                                                                                                                                                  checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0610"} = {errorcode => "A01",                                    subcode => "5769",                                               class => "4",           reason => "Warnung Z2 Inverterfehler Eingangsrelais offen",                                                                                                                                                                                                                                                                                                   checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0611"} = {errorcode => "A01",                                    subcode => "5770",                                               class => "4",           reason => "Warnung Z2 Inverterfehler Stromunterschiede zw. den drei IGBT`s",                                                                                                                                                                                                                                                                                  checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0612"} = {errorcode => "A01",                                    subcode => "5771",                                               class => "4",           reason => "Warnung Z2 Inverterfehler Stromunterschiede zw. den drei PFC IGBT`s.",                                                                                                                                                                                                                                                                             checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0613"} = {errorcode => "A01",                                    subcode => "5772",                                               class => "4",           reason => "Warnung Z2 Inverterfehler Kleinspannung Bereichs&uuml;berschreitung",                                                                                                                                                                                                                                                                              checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0614"} = {errorcode => "A01",                                    subcode => "5773",                                               class => "4",           reason => "Warnung Z2 Inverterfehler Motordrehzahl zu hoch",                                                                                                                                                                                                                                                                                                  checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0615"} = {errorcode => "A01",                                    subcode => "5774",                                               class => "4",           reason => "Warnung Z2 Inverterfehler DC Spannung zu niedrig",                                                                                                                                                                                                                                                                                                 checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0616"} = {errorcode => "A01",                                    subcode => "5775",                                               class => "4",           reason => "Warnung Z2 Inverterfehler Drehmoment Grenze erreicht",                                                                                                                                                                                                                                                                                             checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0617"} = {errorcode => "A01",                                    subcode => "5776",                                               class => "4",           reason => "Warnung Z2 Inverterfehler Modbus Kommunikation gest&ouml;rt",                                                                                                                                                                                                                                                                                      checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0618"} = {errorcode => "A01",                                    subcode => "5777",                                               class => "4",           reason => "Warnung Z2 Inverterfehler Verdichter-Scroll &Uuml;bertemperatur",                                                                                                                                                                                                                                                                                  checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0619"} = {errorcode => "A01",                                    subcode => "5778",                                               class => "4",           reason => "Warnung Z2 Inverterfehler Verdichter-Motor &Uuml;bertemperatur",                                                                                                                                                                                                                                                                                   checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0620"} = {errorcode => "A01",                                    subcode => "5779",                                               class => "4",           reason => "Warnung Z2 Inverterfehler Schaltkreis &Uuml;bertemperatur",                                                                                                                                                                                                                                                                                        checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0621"} = {errorcode => "A01",                                    subcode => "5780",                                               class => "4",           reason => "Warnung Z2 Inverterfehler IGBT`s &Uuml;bertemperatur",                                                                                                                                                                                                                                                                                             checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0622"} = {errorcode => "A01",                                    subcode => "5781",                                               class => "4",           reason => "Warnung Z2 Inverterfehler PFC IGBT's &Uuml;bertemperatur",                                                                                                                                                                                                                                                                                         checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0623"} = {errorcode => "A01",                                    subcode => "5782",                                               class => "4",           reason => "Warnung Z2 Inverter Kommunikation. zw. Signalund Hauptprozessor gest&ouml;rt",                                                                                                                                                                                                                                                                     checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0624"} = {errorcode => "A01",                                    subcode => "5783",                                               class => "4",           reason => "Warnung Z2 Inverter Kommunikation. zw. Hauptund Signalprozessor gest&ouml;rt",                                                                                                                                                                                                                                                                     checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0625"} = {errorcode => "A01",                                    subcode => "5785",                                               class => "4",           reason => "Warnung Z2 Temperaturdifferenz zw. den drei PFC IGBT`s",                                                                                                                                                                                                                                                                                           checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0626"} = {errorcode => "A01",                                    subcode => "5786",                                               class => "4",           reason => "Warnung Z2 Inverterfehler Verdichter-Scroll Temp. zu niedrig",                                                                                                                                                                                                                                                                                     checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0627"} = {errorcode => "A01",                                    subcode => "5787",                                               class => "4",           reason => "Warnung Z2 Inverterfehler Verdichter-Motor Temp. zu niedrig",                                                                                                                                                                                                                                                                                      checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0628"} = {errorcode => "A01",                                    subcode => "5788",                                               class => "4",           reason => "Warnung Z2 Inverterfehler Schaltkreis Temp. zu niedrig",                                                                                                                                                                                                                                                                                           checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0629"} = {errorcode => "A01",                                    subcode => "5789",                                               class => "4",           reason => "Warnung Z2 Inverterfehler IGBT`s Temp. zu niedrig",                                                                                                                                                                                                                                                                                                checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0630"} = {errorcode => "A01",                                    subcode => "5790",                                               class => "4",           reason => "Warnung Z2 Inverterfehler PFC IGBT's Temp. zu niedrig",                                                                                                                                                                                                                                                                                            checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0631"} = {errorcode => "A01",                                    subcode => "5791",                                               class => "4",           reason => "Warnung Z2 Inverter Kommunikation zw.; AD-Wandler und Nebenprozessor gest&ouml;rt",                                                                                                                                                                                                                                                                checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0632"} = {errorcode => "A01",                                    subcode => "5793",                                               class => "4",           reason => "Warnung Z2 Niederdrucksensor JR0 Druck zu hoch",                                                                                                                                                                                                                                                                                                   checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0633"} = {errorcode => "A01",                                    subcode => "5794",                                               class => "4",           reason => "Warnung Z2 Niederdrucksensor JR0 Druck zu niedrig",                                                                                                                                                                                                                                                                                                checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0634"} = {errorcode => "A01",                                    subcode => "5795",                                               class => "4",           reason => "Warnung Z2 Hochdrucksensor JR1 Druck zu niedrig",                                                                                                                                                                                                                                                                                                  checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0635"} = {errorcode => "A01",                                    subcode => "5796",                                               class => "4",           reason => "Warnung Z1; Keine Leistungsabgabe",                                                                                                                                                                                                                                                                                                                checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0636"} = {errorcode => "A01",                                    subcode => "5797",                                               class => "4",           reason => "Warnung Z2; Keine Leistungsabgabe",                                                                                                                                                                                                                                                                                                                checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0637"} = {errorcode => "A01",                                    subcode => "5864",                                               class => "4",           reason => "Warnung Z1 Kommunikationsst&ouml;rung zwischen IWS und Inverter",                                                                                                                                                                                                                                                                                  checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0638"} = {errorcode => "A01",                                    subcode => "5866",                                               class => "4",           reason => "Warnung Z1 Kommunikationsfehler zwischen IWS und Inverter",                                                                                                                                                                                                                                                                                        checklist => "&Uuml;berpr&uuml;fenb, ob der Anschlussstecker f&uuml;r die Drehzahlregelung am Gebl&auml;se in der Au&szlig;eneinheit komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob der Anschlussstecker f&uuml;r die Versorgungsspannung am Gebl&auml;se in der Au&szlig;eneinheit komplett aufgesteckt ist&bull;Funktion des Gebl&auml;ses in der Au&szlig;eneinheit &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Anschlussstecker f&uuml;r die Drehzahlregelung  zums Gebl&auml;se im Steuerger&auml;t der Au&szlig;eneinheit komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob der Anschlussstecker f&uuml;r die Versorgungsspannung zum Gebl&auml;se im Steuerger&auml;t der Au&szlig;eneinheit komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel  f&uuml;r die Drehzahlregelung zum Gebl&auml;se besch&auml;digt ist&bull;-",correction => "Anschlussstecker f&uuml;r die Drehzahlregelung am Gebl&auml;se in der Au&szlig;eneinheit komplett aufstecken&bull;Anschlussstecker f&uuml;r die Versorgungsspannung am Gebl&auml;se in der Au&szlig;eneinheit komplett aufstecken&bull;Gebl&auml;se bzw. Gebl&auml;semotor  in der Au&szlig;eneinheit austauschen&bull;Anschlussstecker f&uuml;r die Drehzahlregelung zum Gebl&auml;se im Steuerger&auml;t der Au&szlig;eneinheit komplett aufstecken&bull;Anschlussstecker f&uuml;r die Versorgungsspannung zum Gebl&auml;se im Steuerger&auml;t der Au&szlig;eneinheit komplett aufstecken&bull;Anschlusskabel f&uuml;r die Drehzahlregelung zum Gebl&auml;se  in der Au&szlig;eneinheit austauschen&bull;Anschlusskabel f&uuml;r die Spannungsversorgung zum Gebl&auml;se in der Au&szlig;eneineit austauschen"};
	$ErrorCodeList{"0639"} = {errorcode => "A01",                                    subcode => "5868",                                               class => "4",           reason => "Warnung Z2 Kommunikationsst&ouml;rung zwischen IWS und Inverter",                                                                                                                                                                                                                                                                                  checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0640"} = {errorcode => "A01",                                    subcode => "5870",                                               class => "4",           reason => "Warnung Z2 Kommunikationsfehler zwischen IWS und Inverter",                                                                                                                                                                                                                                                                                        checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0641"} = {errorcode => "A01",                                    subcode => "5874",                                               class => "4",           reason => "Enteisung der Au&szlig;eneinheit fehlgeschlagen",                                                                                                                                                                                                                                                                                                  checklist => "&Uuml;berpr&uuml;fen, ob die Lamellen am Verdampfer vereist sind&bull;&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Abtauventil im K&auml;ltemittelkreis aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Abtauventil im K&auml;ltemittelkreis besch&auml;digt ist&bull;Ohmschen Widerstand der Magnetspulte am Abtauventil mesen&bull;-",                                                                                                                                     correction => "Aufstellort f&uuml;r die Au&szlig;eneinheit ver&auml;ndern&bull;Anschlussstecker  am Abtauventil im K&auml;ltemittekreis aufstecken&bull;Anschlusskabel zum Abtauventil im K&auml;ltemittelkreis austauschen&bull;Magentspule am Abtauventil im K&auml;ltemittelkreis austauschen&bull;Korrekte Umlaufmenge im Solekreis zwischen Au&szlig;enund Inneneinheit einstellen"                        };
	$ErrorCodeList{"0642"} = {errorcode => "A01",                                    subcode => "5877",                                               class => "4",           reason => "Warnung Z2 Einspritztemperatursensor TR2 Fehler",                                                                                                                                                                                                                                                                                                  checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0643"} = {errorcode => "A01",                                    subcode => "5880",                                               class => "4",           reason => "Warnung Z2 Sauggastemperatursensor TR5 Fehler",                                                                                                                                                                                                                                                                                                    checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0644"} = {errorcode => "A01",                                    subcode => "5883",                                               class => "4",           reason => "Warnung Z1 Sauggastemperatursensor TR5 Fehler",                                                                                                                                                                                                                                                                                                    checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0645"} = {errorcode => "A01",                                    subcode => "5901; 5904",                                         class => "8",           reason => "Interner Fehler im Steuerger&auml;t (MCU)",                                                                                                                                                                                                                                                                                                        checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Versorgungsspannung kurz unterbrechen&bull;Steuerger&auml;t (MCU) austauschen"                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"0646"} = {errorcode => "A01",                                    subcode => "5905",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Interner Fehler im Steuerger&auml;t (CU I/O)",                                                                                                                                                                                                                                                                                checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Versorgungsspannung kurz unterbrechen&bull;Steuerger&auml;t (CU I/O) austauschen"                                                                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"0647"} = {errorcode => "A01",                                    subcode => "5906",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Interner Fehler im Steuerger&auml;t  der Au&szlig;eneinheit",                                                                                                                                                                                                                                                                 checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Versorgungsspannung kurz unterbrechen&bull;Steuerger&auml;t in der Au&szlig;eneinheit austauschen"                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0648"} = {errorcode => "A01",                                    subcode => "5910",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Kein CAN Bussignal am Steuerger&auml;t (CU I/O) vorhanden",                                                                                                                                                                                                                                                                   checklist => "&Uuml;berpr&uuml;fen, ob das CAN Buskabel  am Steuerger&auml;t (CU I/O) korrekt angeschlossen ist.&bull;&Uuml;berpr&uuml;fen, ob das CAN Buskabel zwischen den Steuerger&auml;ten MC und CU I/O besch&auml;digt ist.&bull;-",                                                                                                                                                                                                                                                                         correction => "CAN Buskabel im Steuerger&auml;t (CU I/O) korrekt anschlie&szlig;en&bull;CAN Buskabel zwischen den Steuerger&auml;ten MCU und CU I/O austauschen&bull;Steuerger&auml;t (CU I/O) austauschen"                                                                                                                                                                                                     };
	$ErrorCodeList{"0649"} = {errorcode => "A01",                                    subcode => "5911",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: St&ouml;rung in der CAN Buskommunikation zwischen Innenund Au&szlig;eneinheit",                                                                                                                                                                                                                                               checklist => "&Uuml;berpr&uuml;fen, ob das CAN -Buskabel am Steuerger&auml;t der Inneneinheit korrekt angeschlossen ist&bull;&Uuml;berpr&uuml;fen, ob das CAN Buskabel am Steuerger&auml;t der Au&szlig;eneinheit korrekt angeschlossen ist.&bull;Abstand zwischen dem CAN Buskabel und dem Anschlusskabel f&uuml;r die Spannungsversorgung der Au&szlig;eneinheit &uuml;berpr&uuml;fen.&bull;-",                                                                                                                   correction => "CAN Buskabel im Steuerger&auml;t der Inneneinheit korrekt anschlie&szlig;en&bull;CAN Buskabel im Steuerger&auml;t der Au&szlig;eneinheit korrekt anschlie&szlig;en&bull;Abstand von mindestens 100 mm zwischen CAN Buskabel und Anschlusskabel f&uuml;r die Spannungsversorgung sicherstellen&bull;CAN Buskabel zwischen Au&szlig;enund Inneneinheit instandsetzen oder austauschen"             };
	$ErrorCodeList{"0650"} = {errorcode => "A01",                                    subcode => "5912",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Kein CAN Bussignal am Steuerger&auml;t (GHP-SC) vorhanden",                                                                                                                                                                                                                                                                   checklist => "&Uuml;berpr&uuml;fen, ob das CAN  Buskabel  am Steuerger&auml;t (GHP-SC) korrekt angeschlossen ist.&bull;&Uuml;berpr&uuml;fen, ob das CAN  Buskabel  am Steuerger&auml;t (CU I/O) korrekt angeschlossen ist.&bull;&Uuml;berpr&uuml;fen, ob das CAN Buskabel zwischen den Steuerger&auml;ten GHP-SC und CU I/O besch&auml;digt ist.&bull;Spannungsversorgung am Steuerger&auml;t CU I/O pr&uuml;fen&bull;-",                                                                                           correction => "CAN Buskabel im Steuerger&auml;t (GHP-SC) korrekt anschlie&szlig;en&bull;CAN Buskabel im Steuerger&auml;t (CU I/O) korrekt anschlie&szlig;en&bull;CAN Buskabel zwischen den Steuerger&auml;ten GHP-SC und CU I/O austauschen&bull;Spannungsversorgung  am Steuerger&auml;t (CU I/O) korrigieren&bull;Steuerger&auml;t CU I/O tauschen&bull;Steuerger&auml;t GHP-SC tauschen"                     };
	$ErrorCodeList{"0651"} = {errorcode => "A01",                                    subcode => "5920",                                               class => "8",           reason => "Gas-W&auml;rmepumpe:  Keine R&uuml;ckmeldespannung an Anschlussklemme X39",                                                                                                                                                                                                                                                                        checklist => "&Uuml;berpf&uuml;en, ob die Br&uuml;cke an der Anschlussklemme X39 montiert ist&bull;-",                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Br&uuml;cke an de Anschlussklemme X39 montieren&bull;Anschlussstecker X39 auf der Leiterplatte im Steuerger&auml;t (CU I/O komplett aufstecken"                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0652"} = {errorcode => "A01",                                    subcode => "5921",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Hochdruckw&auml;chter (25 bar) im K&auml;ltemiitelkreis hat angesprochen",                                                                                                                                                                                                                                                    checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Hochdruckf&uuml;hler PTI11 im K&auml;ltemittelkreis komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das  Anschlusskabel vom Hochdruckf&uuml;hler PTI11 am Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Motor am 2 -Wege Ventil  besch&auml;digt ist&bull;&Uuml;berpr&uuml;fen, ob der Motor am 2 Wege Ventil korrekt montiert ist&bull;&Uuml;berpr&uuml;fen, ob der Motor am 2 Wege Ventil defekt ist&bull;Funktion des 2 Wege Ventiles vor dem Verfl&uuml;ssiger &uuml;berpr&uuml;fen&bull;-",correction => "Anschlussstecker am Hochdruckf&uuml;hler PTI11 im K&auml;ltemittelkreis komplett aufstecken&bull;Anschlusskabel vom Hochdruckf&uuml;hler PTI11 am Steuerger&auml;t (CU I/O) korrekt anschlie&szlig;en&bull;Anschlusskabel zum Motor am 2 Wege Ventil austauschen&bull;Motor am 2 Wege Ventil korrekt auf das Ventilunterteil montieren&bull;Motor am 2 Wege Ventil austauschen&bull;2 Wege Ventil vor dem Verfl&uuml;ssiger austauschen&bull;Werkskundendienst benachrichtigen"};
	$ErrorCodeList{"0653"} = {errorcode => "A01",                                    subcode => "5922",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Hochdruckbegrenzer (27bar) im K&auml;ltemittelkreis hat angesprochen",                                                                                                                                                                                                                                                        checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Hochdruckf&uuml;hler PTI11 im K&auml;ltemittelkreis komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das  Anschlusskabel vom Hochdruckf&uuml;hler PTI11 am Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Motor am 2 -Wege Ventil  besch&auml;digt ist&bull;&Uuml;berpr&uuml;fen, ob der Motor am 2 Wege Ventil korrekt montiert ist&bull;&Uuml;berpr&uuml;fen, ob der Motor am 2 Wege Ventil defekt ist&bull;Funktion des 2 Wege Ventiles vor dem Verfl&uuml;ssiger &uuml;berpr&uuml;fen&bull;-",correction => "Anschlussstecker am Hochdruckf&uuml;hler PTI11 im K&auml;ltemittelkreis komplett aufstecken und manuellen Reset am Druckbegrenzer durchf&uuml;hten&bull;Anschlusskabel vom Hochdruckf&uuml;hler PTI11 am Steuerger&auml;t (CU I/O) korrekt anschlie&szlig;en und manuellen Reset am Druckbegrenzer durchf&uuml;hren&bull;Anschlusskabel zum Motor am 2 Wege Ventil austauschen und manuellen Reset am Druckbegrenzer durchf&uuml;hren&bull;Motor am 2 Wege Ventil korrekt auf das Ventilunterteil montieren und manuellen Rest am Druckbegrenzer durchf&uuml;hren&bull;Motor am 2 Wege Ventil austauschen&bull;2 Wege Ventil vor dem Verfl&uuml;ssiger austauschen und manuellen Reset am Druckbegrenzer durchf&uuml;hren&bull;Werkskundendienst benachrichtigen"};
	$ErrorCodeList{"0654"} = {errorcode => "A01",                                    subcode => "5923",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Sicherheitstemperaturbegrenzer am Austreiberfu&szlig; hat angesprochen",                                                                                                                                                                                                                                                      checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Sicherheitstemperaturbegrenzer TSHI-9 komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Sicherheitstemperaturbegrenzer TSHI-9 besch&auml;digt ist&bull;Ohmschen Widerstand des Sicherheitstemperaturbegrenzers TSHI-9 messen&bull;Brennereinstellung bei Nennw&auml;rmebelastung &uuml;berpr&uuml;fen&bull;-",                                                                                                           correction => "Anschlussstecker am Sicherheitstemperaturbegrenzer TSHI-9 komplett aufstecken&bull;Anschlusskabel zum Sicherheitstemperaturbegrenzer TSHI-9 austauschen&bull;Sicherheitstemperaturbegrenzer TSHI-9 austauschen&bull;Brennereinstellwerte bei Nennw&auml;rmebelastung korrigieren&bull;Werkskundendienst benachrichtigen"                                                                         };
	$ErrorCodeList{"0655"} = {errorcode => "A01",                                    subcode => "5924",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Abgastemperaturbegrenzer am Austreiber hat angesprochen",                                                                                                                                                                                                                                                                     checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Abgassicherheitstemperaturbegrenzer TSHI-10 komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Abgassicherheitstemperaturbegrenzer TSHI-10 besch&auml;digt ist&bull;Ohmschen Widerstand des Abgassicherheitstemperaturbegrenzers TSHI-10 messen&bull;Brennereinstellung bei Nennw&auml;rmebelastung &uuml;berpr&uuml;fen&bull;-",                                                                                         correction => "Anschlussstecker am Abgassicherheitstemperaturbegrenzer TSHI-10 komplett aufstecken&bull;Anschlusskabel zum Abgassicherheitstemperaturbegrenzer TSHI-10 austauschen&bull;Abgassicherheitstemperaturbegrenzer TSHI-10 austauschen&bull;Brennereinstellwerte bei Nennw&auml;rmebelastung korrigieren&bull;Werkskundendienst benachrichtigen"                                                       };
	$ErrorCodeList{"0656"} = {errorcode => "A01",                                    subcode => "5925",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Abgastemperaturbegrenzer nach dem W&auml;rmetauscher hat angesprochen",                                                                                                                                                                                                                                                       checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Abgassicherheitstemperaturbegrenzer TSHI-11 komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Abgassicherheitstemperaturbegrenzer TSHI-11 besch&auml;digt sit&bull;Ohmschen Widerstand des  Abgassicherheitstemperaturbegrenzers TSHI-11 messen&bull;Brennereinstellwerte bei maximaler Nennw&auml;rmebelastung &uuml;berpr&uuml;fen&bull;Stellung der Wartungsh&auml;hne zwischen W&auml;rmepumpe und Pufferspeicher &uuml;berpr&uuml;fen.&bull;&Uuml;berpr&uuml;fen, ob sich Luft in den Verbindungsleitungen zwischen W&auml;rmepumpe und Pufferspeicher befindet.&bull;Funktion der Umw&auml;lzpumpe &uuml;berpr&uuml;fen&bull;Volumenstrom &uuml;berpr&uuml;fen",correction => "Anschlussstecker am Abgassicherheitstemperaturbegrenzer TSHI-11 komplett aufstecke&bull;Anschlusskabel zum Abgassicherheitstemperaturbegrenzer TSHI-11 austauschen&bull;Abgassicherheitstemperaturbegrenzer TSHI-11 austauschen&bull;Brennereinstellung bei maximaler Nennw&auml;rmelbelastung korrigieren&bull;Wartungsh&auml;hne zwischen W&auml;rmepumpe und Pufferspeicher &ouml;ffnen&bull;Verbindungsleitungen zwischen W&auml;rmepumpe und Pufferspeicher entl&uuml;ften&bull;Umw&auml;lzpumpe anwerfen ggf. austauschen&bull;Volumenstrom auf den zul&auml;ssigen Wert (7 100 l/ min) erh&ouml;hen"};
	$ErrorCodeList{"0657"} = {errorcode => "A01",                                    subcode => "5930",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Steuerger&auml;t erh&auml;lt unzul&auml;ssige Werte vom R&uuml;cklauftemperaturf&uuml;hler",                                                                                                                                                                                                                                  checklist => "&Uuml;berpr&uuml;fen ob der Anschlussstecker am Vorlauftemperaturf&uuml;hler TTI-6 komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel vom Vorlauftemperaturf&uuml;hler TTI-6 am Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;Ohmschen Widerstand des Temperaturf&uuml;hlers TTI-6 messen&bull;&Uuml;berpr&uuml;fen, ob der Vorlauftemperaturf&uuml;hler TTI-6 korrekten Kontakt an der Vorlaufleitung hat.&bull;Position des Vorlauftemperaturf&uuml;hlers TTI-6 an der Vorlaufleitung &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob die Isolierung am Vorlauftemperaturf&uuml;hler TTI-6 vorhanden bzw. besch&auml;digt ist&bull;-",correction => "Anschlussstecker am Vorlauftemperaturf&uuml;hler TTI-6 komplett aufstecken.&bull;Anschlusskabel vom Vorlauftemperaturf&uuml;hler TTI-6 am Steuerger&auml;t (CU i/O) korrekt anschlie&szlig;en&bull;Vorlauftemperaturf&uuml;hler TTI-6 austauschen&bull;Kontakt zwischen Vorlauftemperaturf&uuml;hler TTI-6 und Vorlaufleitung herstellen&bull;Vorlauftemperaturf&uuml;hler TTI-6 an der korrekten Position der Vorlaufleitung montieren&bull;Isolierung am Vorlauftemperaturf&uuml;hler TTI-6 austauschen bzw. montieren&bull;Kabelbaum austauschen"};
	$ErrorCodeList{"0658"} = {errorcode => "A01",                                    subcode => "5931",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Steuerger&auml;t erh&auml;lt unzul&auml;ssige Werte vom R&uuml;cklauftemperaturf&uuml;hler",                                                                                                                                                                                                                                  checklist => "&Uuml;berpr&uuml;fen ob der Anschlussstecker am  R&uuml;cklauftemperaturf&uuml;hler TTI-5 komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel vom R&uuml;cklauftemperaturf&uuml;hler TTI-5 am Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;Ohmschen Widerstand des R&uuml;cklauftemperaturf&uuml;hlers TTI-5 messen&bull;&Uuml;berpr&uuml;fen, ob der R&uuml;cklauftemperaturf&uuml;hler; TTI-5 korrekten Kontakt an der Vorlaufleitung hat.&bull;Position des R&uuml;cklauftemperaturf&uuml;hlers TTI-5 an der Vorlaufleitung &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob die Isolierung am R&uuml;cklauftemperaturf&uuml;hler TTI-5 vorhanden bzw. besch&auml;digt ist&bull;-",correction => "Anschlussstecker am R&uuml;cklauftemperaturf&uuml;hler TTI-5 komplett aufstecken.&bull;Anschlusskabel vom R&uuml;cklauftemperaturf&uuml;hler TTI-5 am Steuerger&auml;t (CU i/O) korrekt anschlie&szlig;en&bull;R&uuml;cklauftemperaturf&uuml;hler TTI-5 austauschen&bull;Kontakt zwischen R&uuml;cklauftemperaturf&uuml;hler TTI-5 und R&uuml;cklaufleitung herstellen&bull;R&uuml;cklauftemperaturf&uuml;hler TTI-5 an der korrekten Position der R&uuml;cklaufleitung montieren&bull;Isolierung am R&uuml;cklauftemperaturf&uuml;hler TTI-6 austauschen bzw. montieren&bull;Kabelbaum austauschen"};
	$ErrorCodeList{"0659"} = {errorcode => "A01",                                    subcode => "5932",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Steuerger&auml;t erh&auml;lt unzul&auml;ssige Werte vom Temperaturf&uuml;hler im Solevorlauf",                                                                                                                                                                                                                                checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Temperaturf&uuml;hler TTI-7 komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel vom Temperaturf&uuml;hler TTI-7 im Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Temperaturf&uuml;hler TTI-7 besch&auml;digt ist&bull;Ohmschen Widerstand des Temperaturf&uuml;hlers TTI-7 &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-7 korrekten Kontakt zur Solevorlaufleitung hat&bull;&Uuml;berpr&uuml;fen, ob die Isolierung am Temperaturf&uuml;hler TTI-7 vorhanden bzw besch&auml;digt ist&bull;Solekreis auf Entnahmeleistung &uuml;berpr&uuml;fen&bull;-",correction => "Anschlussstecker am Temperaturf&uuml;hler TTI-7 komplett aufstecken&bull;Anschlusskabel vom Temperaturf&uuml;hler TTI-7 korrekt im Steuerger&auml;t (CU I/O) anschlie&szlig;en&bull;Kabelbaum autauschen&bull;Temperaturf&uuml;hler TTI-7 austauschen&bull;Temperaturf&uuml;hler TTI-7 an der korrekten Position der Solevorlaufleitung montieren&bull;Isolierung am Temperaturf&uuml;hler TTI-7 austauschen bzw. montieren&bull;Korrekten Volumenstrom im Solekreis eiinstellen&bull;Umw&auml;lzpumpe im Solekreis austauschen"};
	$ErrorCodeList{"0660"} = {errorcode => "A01",                                    subcode => "5933",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Steuerger&auml;t erh&auml;lt unzul&auml;ssige Werte vom Soler&uuml;cklauftemperaturf&uuml;hler",                                                                                                                                                                                                                              checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Temperaturf&uuml;hler TTI-8 komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel vom Temperaturf&uuml;hler TTI-8 im Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Temperaturf&uuml;hler TTI-8 besch&auml;digt ist&bull;Ohmschen Widerstand des Temperaturf&uuml;hlers TTI-8 &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-8 korrekten Kontakt zur Soler&uuml;cklaufleitung hat&bull;&Uuml;berpr&uuml;fen, ob die Isolierung am Temperaturf&uuml;hler TTI-8 vorhanden bzw besch&auml;digt ist&bull;Solekreis auf Entnahmeleistung &uuml;berpr&uuml;fen&bull;-",correction => "Anschlussstecker am Temperaturf&uuml;hler TTI-8 komplett aufstecken&bull;Anschlusskabel vom Temperaturf&uuml;hler TTI-8 korrekt im Steuerger&auml;t (CU I/O) anschlie&szlig;en&bull;Kabelbaum autauschen&bull;Temperaturf&uuml;hler TTI-8 austauschen&bull;Temperaturf&uuml;hler TTI-8 an der korrekten Position der Soler&uuml;cklaufleitung montieren&bull;Isolierung am Temperaturf&uuml;hler TTI-8 austauschen bzw. montieren&bull;Korrekten Volumenstrom im Solekreis eiinstellen&bull;Umw&auml;lzpumpe im Solekreis austauschen"};
	$ErrorCodeList{"0661"} = {errorcode => "A01",                                    subcode => "5934",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Steuerger&auml;t erh&auml;lt unzul&auml;ssige Werte vom Temperaturf&uuml;hler am Verfl&uuml;ssigerausgang",                                                                                                                                                                                                                   checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Temperaturf&uuml;hler TTI-3 komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel vom Temperaturf&uuml;hler TTI-3 im Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Temperaturf&uuml;hler TTI-3 besch&auml;digt ist&bull;Ohmschen Widerstand des Temperaturf&uuml;hlers TTI-3 &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-3 korrekten Kontakt zur K&auml;ltemittelleitung am Verfl&uuml;ssigerausgang hat&bull;-",correction => "Anschlussstecker am Temperaturf&uuml;hler TTI-3 komplett aufstecken&bull;Anschlusskabel vom Temperaturf&uuml;hler TTI-3 korrekt im Steuerger&auml;t (CU I/O) anschlie&szlig;en&bull;Kabelbaum autauschen&bull;Temperaturf&uuml;hler TTI-3 austauschen&bull;Temperaturf&uuml;hler TTI-3 an der korrekten Position der K&auml;ltemittelleitung am Verfl&uuml;ssigerausgang montieren&bull;Isolierung am Temperaturf&uuml;hler TTI-3 austauschen bzw. montieren"};
	$ErrorCodeList{"0662"} = {errorcode => "A01",                                    subcode => "5935",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Steuerger&auml;t erh&auml;lt unzul&auml;ssige Werte vom Temperaturf&uuml;hler am Verfl&uuml;ssigerausgang (Heizwasser)",                                                                                                                                                                                                      checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Temperaturf&uuml;hler TTI-12 komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel vom Temperaturf&uuml;hler TTI-12 im Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Temperaturf&uuml;hler TTI-12 besch&auml;digt ist&bull;Ohmschen Widerstand des Temperaturf&uuml;hlers TTI-12 &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-12 korrekten Kontakt zur Heizwasserleitung am Verfl&uuml;ssigerausgang hat&bull;-",correction => "Anschlussstecker am Temperaturf&uuml;hler TTI-12 komplett aufstecken&bull;Anschlusskabel vom Temperaturf&uuml;hler TTI-12 korrekt im Steuerger&auml;t (CU I/O) anschlie&szlig;en&bull;Kabelbaum autauschen&bull;Temperaturf&uuml;hler TTI-12 austauschen&bull;Temperaturf&uuml;hler TTI-12 an der korrekten Position der Heizwasserleitung am Verfl&uuml;ssigerausgang montieren&bull;Isolierung am Temperaturf&uuml;hler TTI-12 austauschen bzw. montieren"};
	$ErrorCodeList{"0663"} = {errorcode => "A01",                                    subcode => "5936",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Steuerger&auml;t erh&auml;lt unzul&auml;ssige Werte vom Temperaturf&uuml;hler am Verdampferausgang",                                                                                                                                                                                                                          checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Temperaturf&uuml;hler TTI-4 komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel vom Temperaturf&uuml;hler TTI-4 im Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Temperaturf&uuml;hler TTI-4 besch&auml;digt ist&bull;Ohmschen Widerstand des Temperaturf&uuml;hlers TTI-4 &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-12 korrekten Kontakt zur K&auml;ltemittelleitung am Verdampferausgang hat&bull;-",correction => "Anschlussstecker am Temperaturf&uuml;hler TTI-4 komplett aufstecken&bull;Anschlusskabel vom Temperaturf&uuml;hler TTI-4 korrekt im Steuerger&auml;t (CU I/O) anschlie&szlig;en&bull;Kabelbaum autauschen&bull;Temperaturf&uuml;hler TTI-4 austauschen&bull;Temperaturf&uuml;hler TTI-4 an der korrekten Position der K&auml;ltemittelleitung am Verdampferausgang montieren&bull;Isolierung am Temperaturf&uuml;hler TTI-4 austauschen bzw. montieren"};
	$ErrorCodeList{"0664"} = {errorcode => "A01",                                    subcode => "5937",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Steuerger&auml;t erh&auml;lt unzul&auml;ssige Werte vom Temperaturf&uuml;hler am Austreiberfu&szlig;",                                                                                                                                                                                                                        checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Temperaturf&uuml;hler TTI-1 komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel vom Temperaturf&uuml;hler TTI-1 im Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Temperaturf&uuml;hler TTI-1 besch&auml;digt ist&bull;Ohmschen Widerstand des Temperaturf&uuml;hlers TTI-1 &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-1 korrekten Kontakt zum Generatorfu&szlig; hat&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-1 korrrekt isoliert ist&bull;Nennw&auml;rmebelastung am Brenner &uuml;berpr&uuml;fen&bull;Funktion der Umw&auml;lzpumpe im K&auml;ltemittelkreis (NH3) &uuml;berpr&uuml;fen&bull;-",correction => "Anschlussstecker am Temperaturf&uuml;hler TTI-1 komplett aufstecken&bull;Anschlusskabel vom Temperaturf&uuml;hler TTI-1 korrekt im Steuerger&auml;t (CU I/O) anschlie&szlig;en&bull;Kabelbaum autauschen&bull;Temperaturf&uuml;hler TTI-1 austauschen&bull;Temperaturf&uuml;hler TTI-1 an der korrekten Position am Austreiberfu&szlig; montieren&bull;Isolierung am Temperaturf&uuml;hler T1I-4 austauschen bzw. montieren&bull;Brennereinstellwert f&uuml;r die Nennw&auml;rmebelastung korrigieren&bull;Werkskundendienst benachrichtigen&bull;Werkskundendienst benachrichtigen"};
	$ErrorCodeList{"0665"} = {errorcode => "A01",                                    subcode => "5938",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Steuerger&auml;t erh&auml;lt unzul&auml;ssige Werte vom Temperaturf&uuml;hler f&uuml;r Hei&szlig;gas",                                                                                                                                                                                                                        checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Temperaturf&uuml;hler TTI-2 komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel vom Temperaturf&uuml;hler TTI-2 im Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Temperaturf&uuml;hler TTI-2 besch&auml;digt ist&bull;Ohmschen Widerstand des Temperaturf&uuml;hlers TTI-2 &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-1 korrekten Kontakt zur K&auml;ltemittelleitung am Verfl&uuml;ssigerausgang hat&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-2 korrrekt isoliert ist&bull;Nennw&auml;rmebelastung am Brenner &uuml;berpr&uuml;fen&bull;Funktion der Umw&auml;lzpumpe im K&auml;ltemittelkreis (NH3) &uuml;berpr&uuml;fen&bull;-",correction => "Anschlussstecker am Temperaturf&uuml;hler TTI-2 komplett aufstecken&bull;Anschlusskabel vom Temperaturf&uuml;hler TTI-2 korrekt im Steuerger&auml;t (CU I/O) anschlie&szlig;en&bull;Kabelbaum autauschen&bull;Temperaturf&uuml;hler TTI-2 austauschen&bull;Temperaturf&uuml;hler TTI-2 an der korrekten Position Der K&auml;ltemittelleitung am Verfl&uuml;ssigerausgang montieren&bull;Isolierung am Temperaturf&uuml;hler TT1-2 austauschen bzw. montieren&bull;Brennereinstellwert f&uuml;r die Nennw&auml;rmebelastung korrigieren&bull;Werkskundendienst benachrichtigen&bull;Werkskundendienst benachrichtigen"};
	$ErrorCodeList{"0666"} = {errorcode => "A01",                                    subcode => "5939",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Steuerger&auml;t erh&auml;lt unzul&auml;ssige Werte vom Niederdruckf&uuml;hler (K&auml;ltemittel)",                                                                                                                                                                                                                           checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Niederdruckf&uuml;hler PTI12 im K&auml;ltemittelkreis komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das  Anschlusskabel vom Niederdruckf&uuml;hler PTI12 am Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Niederdruckf&uuml;hler PTI12 im K&auml;ltemittelkreis besch&auml;digt ist&bull;&Uuml;berpr&uuml;fen, ob der Motor am 2 Wege Ventil korrekt montiert ist&bull;&Uuml;berpr&uuml;fen, ob der Motor am 2 Wege Ventil defekt ist&bull;Funktion des 2 Wege Ventiles vor dem Verfl&uuml;ssiger &uuml;berpr&uuml;fen&bull;-",correction => "Anschlussstecker am Niederdruckf&uuml;hler PTI12 im K&auml;ltemittelkreis komplett aufstecken&bull;Anschlusskabel vom Niederdruckf&uuml;hler PTI12 am Steuerger&auml;t (CU I/O) korrekt anschlie&szlig;en&bull;Kabelbaum austauschen&bull;Motor am 2 Wege Ventil korrekt auf das Ventilunterteil montieren&bull;Motor am 2 Wege Ventil austauschen&bull;2 Wege Ventil vor dem Verfl&uuml;ssiger austauschen&bull;Werkskundendienst benachrichtigen"};
	$ErrorCodeList{"0667"} = {errorcode => "A01",                                    subcode => "5940",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Steuerger&auml;t erh&auml;lt unzul&auml;ssige Werte vom Hochdruckf&uuml;hler (K&auml;ltemittel)",                                                                                                                                                                                                                             checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Hochdruckf&uuml;hler PTI11 im K&auml;ltemittelkreis komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das  Anschlusskabel vom Hochdruckf&uuml;hler PTI11 am Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Motor am 2 -Wege Ventil  besch&auml;digt ist&bull;&Uuml;berpr&uuml;fen, ob der Motor am 2 Wege Ventil korrekt montiert ist&bull;&Uuml;berpr&uuml;fen, ob der Motor am 2 Wege Ventil defekt ist&bull;Funktion des 2 Wege Ventiles vor dem Verfl&uuml;ssiger &uuml;berpr&uuml;fen&bull;-",correction => "Anschlussstecker am Hochdruckf&uuml;hler PTI11 im K&auml;ltemittelkreis komplett aufstecken&bull;Anschlusskabel vom Hochdruckf&uuml;hler PTI11 am Steuerger&auml;t (CU I/O) korrekt anschlie&szlig;en&bull;Anschlusskabel zum Motor am 2 Wege Ventil austauschen&bull;Motor am 2 Wege Ventil korrekt auf das Ventilunterteil montieren&bull;Motor am 2 Wege Ventil austauschen&bull;2 Wege Ventil vor dem Verfl&uuml;ssiger austauschen&bull;Werkskundendienst benachrichtigen"};
	$ErrorCodeList{"0668"} = {errorcode => "A01",                                    subcode => "5941",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Volumenstrom im Prim&auml;rkreis liegt au&szlig;erhalb des zul&auml;ssigen Bereichs",                                                                                                                                                                                                                                         checklist => "Stellung der Wartungsh&auml;hne zwischen W&auml;rmepumpe und Pufferspeicher &uuml;berpr&uuml;fen.&bull;&Uuml;berpr&uuml;fen, ob sich Luft in den Verbindungsleitungen zwischen W&auml;rmepumpe und Pufferspeicher befindet.&bull;Funktion der Umw&auml;lzpumpe &uuml;berpr&uuml;fen&bull;Volumenstrom &uuml;berpr&uuml;fen&bull;Volumenstrom &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Volumenstromsensor FTI18 aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel vom Volumenstromsensor FTI18 im Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;Volumenstrommesser FTI18 auf wasserseitige Verschmutzungen &uuml;berpr&uuml;fen&bull;-",correction => "Wartungsh&auml;hne zwischen W&auml;rmepumpe und Pufferspeicher &ouml;ffnen&bull;Verbindungsleitungen zwischen W&auml;rmepumpe und Pufferspeicher entl&uuml;ften&bull;Umw&auml;lzpumpe anwerfen ggf. austauschen&bull;Volumenstrom auf den zul&auml;ssigen Wert (7 100 l/ min) reduzieren&bull;Volumenstrom auf den zul&auml;ssigen Wert (7 100 l/ min) erh&ouml;hen&bull;Anschlussstecker am Volumenstromsensor FTI18 komplett aufstecken&bull;Anschlusskabel vom Volumenstromsensor FTI18 am Steuerger&auml;t (CU I/O) korrekt anschlie&szlig;en&bull;Ablagerungen/ Schmutz aus Volumenstromsensor FTI18 entfernen&bull;Volumenstromsensor FTI18 austauschen"};
	$ErrorCodeList{"0669"} = {errorcode => "A01",                                    subcode => "5942",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Volumenstrom im Solekreis liegt au&szlig;erhalb des zul&auml;ssigen Bereichs",                                                                                                                                                                                                                                                checklist => "Stellung der Wartungsh&auml;hne im Solekreis &uuml;berpr&uuml;fen.&bull;Filter im Solekreis auf Verschmutzung &uuml;berpr&uuml;fen.&bull;Funktion der Umw&auml;lzpumpe  im Solekreis &uuml;berpr&uuml;fen&bull;Volumenstrom  im Solekreis &uuml;berpr&uuml;fen&bull;Volumenstrom im Solekreis &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Volumenstromsensor FTI19 im Solekreis aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel vom Volumenstromsensor FTI19 im Steuerger&auml;t (CU I/O) korrekt angeschlossen ist&bull;Volumenstrommesser FTI19 im Solekreis auf wasserseitige Verschmutzungen &uuml;berpr&uuml;fen&bull;-",correction => "Wartungsh&auml;hne im Solekreis &ouml;ffnen&bull;Filter im Solekreis reinigen&bull;Umw&auml;lzpumpe anwerfen ggf. austauschen&bull;Volumenstrom auf den zul&auml;ssigen Wert im Solekreis  (12 100 l/min) reduzieren&bull;Volumenstrom auf den zul&auml;ssigen Wert im Solekreis  (12 100 l/min) erh&ouml;hen&bull;Anschlussstecker am Volumenstromsensor FTI19 im Solekreis komplett aufstecken&bull;Anschlusskabel vom Volumenstromsensor FTI19 am Steuerger&auml;t (CU I/O) korrekt anschlie&szlig;en&bull;Ablagerungen/ Schmutz aus Volumenstromsensor FTI19 im Solekreis entfernen&bull;Volumenstromsensor FTI19 im Solekreisaustauschen"};
	$ErrorCodeList{"0670"} = {errorcode => "A01",                                    subcode => "5943",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Unzul&auml;ssige Lufteintrittstemperatur an der Au&szlig;eneinheit",                                                                                                                                                                                                                                                          checklist => "&Uuml;berpr&uuml;fen, ob der Lufteintrittstemperaturf&uuml;hler an der Au&szlig;eneinheit verschmutzt ist&bull;&Uuml;berpr&uuml;fen, ob der Anschlusstecker am Lufteintrittstemperaturf&uuml;hler in der Au&szlig;eneinheit aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Lufteintrittstemperaturf&uuml;hler in der Au&szlig;eneinheit besch&auml;digt ist&bull;-",                                                                                                            correction => "Ablagerungen amLufteintrittstemperaturf&uuml;hler in der Au&szlig;eneinheit entfernen&bull;Anschlussstecker am Lufteintrittstemperaturf&uuml;hler in der Au&szlig;eneinheit aufstecken&bull;Anschlusskabel zum Lufteintrittstemperaturf&uuml;hler in der Au&szlig;eneinheit austauschen&bull;Lufteintrittstemperaturf&uuml;hler in der Au&szlig;eneinheit austauschen"                           };
	$ErrorCodeList{"0671"} = {errorcode => "A01",                                    subcode => "5944",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Unzul&auml;ssige Luftaustrittstemperatur an der Au&szlig;eneinheit",                                                                                                                                                                                                                                                          checklist => "&Uuml;berpr&uuml;fen, ob der Luftaustrittstemperaturf&uuml;hler an der Au&szlig;eneinheit verschmutzt ist&bull;&Uuml;berpr&uuml;fen, ob der Anschlusstecker am Luftaustrittstemperaturf&uuml;hler in der Au&szlig;eneinheit aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Luftaustrittstemperaturf&uuml;hler in der Au&szlig;eneinheit besch&auml;digt ist&bull;-",                                                                                                            correction => "Ablagerungen amLuftaustrittstemperaturf&uuml;hler in der Au&szlig;eneinheit entfernen&bull;Anschlussstecker am Luftaustrittstemperaturf&uuml;hler in der Au&szlig;eneinheit aufstecken&bull;Anschlusskabel zum Luftaustrittstemperaturf&uuml;hler in der Au&szlig;eneinheit austauschen&bull;Luftaustrittstemperaturf&uuml;hler in der Au&szlig;eneinheit austauschen"                           };
	$ErrorCodeList{"0672"} = {errorcode => "A01",                                    subcode => "5950",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Temperaturdifferenz im Solekreis liegt au&szlig;erhalb des zul&auml;ssigen Bereichs",                                                                                                                                                                                                                                         checklist => "Filter im Solekreis reinigen&bull;Volumenstrom im Solekreis &uuml;berpr&uuml;fen&bull;Volumenstrom im Solekreis &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-7 korrekten Kontakt zur Solevorlaufleitung hat&bull;&Uuml;berpr&uuml;fen, ob die Isolierung am Temperaturf&uuml;hler TTI-7 am Solevorlauf vorhanden bzw besch&auml;digt ist&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-8 korrekten Kontakt zur Soler&uuml;cklaufleitung hat&bull;&Uuml;berpr&uuml;fen, ob ob die Isolierung am Temperaturf&uuml;hler TTI-8 am Soler&uuml;cklauf vorhanden  bzw. besch&auml;digt ist&bull;Entnahmeleistung der Erdsonde &uuml;berpr&uuml;fen&bull;-",correction => "Filter im Solekreis reinigen&bull;Volumenstrom im Solekreis auf den zul&auml;ssigen Wert reduzieren (12 100 l/min)&bull;Volumenstrom im Solekreis auf den zul&auml;ssigen Wert erh&ouml;hen (12 100 l/min)&bull;Temperaturf&uuml;hler TTI-7 an der korrekten Position der Solevorlaufleitung montieren&bull;Isolierung am Temperaturf&uuml;hler TTI-7 austauschen bzw. montieren&bull;Temperaturf&uuml;hler TTI-8 an der korrekten Position der Soler&uuml;cklaufleitung montieren&bull;Isolierung am Temperaturf&uuml;hler TTI-8 austauschen bzw. montieren&bull;Erdsonde &uuml;berpr&uuml;fen&bull;Werkskundendienst beauftragen"};
	$ErrorCodeList{"0673"} = {errorcode => "A01",                                    subcode => "5951",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Temperaturdifferenz zwischen Sole und K&auml;ltemittelaustritt nach Verdampfer  ist au&szlig;erhalb des zul&auml;ssigen Bereiches",                                                                                                                                                                                           checklist => "&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-7 korrekten Kontakt zur Solevorlaufleitung hat&bull;&Uuml;berpr&uuml;fen, ob die Isolierung am Temperaturf&uuml;hler TTI-7 am Solevorlauf vorhanden bzw besch&auml;digt ist.&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-4 korrekten Kontakt zur K&auml;ltemittelleitung am Verdampferausgang hat.&bull;&Uuml;berpr&uuml;fen, ob die Isolierung am Temperaturf&uuml;hler TTI-4 vorhanden bzw. besch&auml;digt ist.&bull;Volumenstrom im Solekreis &uuml;berpr&uuml;fen",correction => "Temperaturf&uuml;hler TTI-7 an der korrekten Position der Solevorlaufleitung montieren&bull;Isolierung am Temperaturf&uuml;hler TTI-7 austauschen bzw. montieren&bull;Temperaturf&uuml;hler TTI-4 an der korrekten Position der K&auml;ltemittelleitung am Verdampferausgang montieren&bull;Isolierung am Temperaturf&uuml;hler TTI-4 austauschen bzw. montieren&bull;Volumenstrom im Solekreis auf die zul&auml;ssigen Werte (12 -100 l/min) erh&ouml;hen"};
	$ErrorCodeList{"0674"} = {errorcode => "A01",                                    subcode => "5952",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Temperaturdifferenz zwischen Verfl&uuml;ssigeraustritt K&auml;ltekreis und Heizungsr&uuml;cklauf ist au&szlig;erhalb des zul&auml;ssigen Bereiches",                                                                                                                                                                          checklist => "&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-3 korrekten Kontakt zur K&auml;ltemittelleitung am Verfl&uuml;ssigerausgang hat&bull;&Uuml;berpr&uuml;fen, ob die Isolierung am Temperaturf&uuml;hler TTI-3 vorhanden bzw. besch&auml;digt ist&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-5 korrekten Kontakt an der R&uuml;ckllaufleitung hat.&bull;Position des R&uuml;cklauftemperaturf&uuml;hlers TTI-5 an der Vorlaufleitung &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob die Isolierung am R&uuml;cklauftemperaturf&uuml;hler TTI-5 vorhanden bzw. besch&auml;digt ist",correction => "Temperaturf&uuml;hler TTI-3 an der korrekten Position der K&auml;ltemittelleitung am Verfl&uuml;ssigerausgang montieren&bull;Isolierung am Temperaturf&uuml;hler TTI-3 austauschen bzw. montieren&bull;Kontakt zwischen R&uuml;cklauftemperaturf&uuml;hler TTI-5 und R&uuml;cklaufleitung herstellen&bull;R&uuml;cklauftemperaturf&uuml;hler TTI-5 an der korrekten Position der R&uuml;cklaufleitung montieren&bull;Isolierung am R&uuml;cklauftemperaturf&uuml;hler TTI-5 austauschen bzw. montieren"};
	$ErrorCodeList{"0675"} = {errorcode => "A01",                                    subcode => "5953",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Druckdifferenz im K&auml;ltemittelkreis ist au&szlig;erhalb des zul&auml;ssigen Bereichs",                                                                                                                                                                                                                                    checklist => "Funktion der L&ouml;sungspumpe im K&auml;ltemittelkreis (NH3) &uuml;berpr&uuml;fen&bull;Funktion der Regelventile im K&auml;ltemittelkreis (NH3) &uuml;berpr&uuml;fen&bull;-",                                                                                                                                                                                                                                                                                                                        correction => "Werkskundendienst benachrichtigen&bull;Werkskundendienst benachrichtigen&bull;Werkskundendienst benachrichtigen"                                                                                                                                                                                                                                                                                 };
	$ErrorCodeList{"0676"} = {errorcode => "A01",                                    subcode => "5954",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Luftseitige Temperaturdifferenz am Verdampfer zu hoch",                                                                                                                                                                                                                                                                       checklist => "&Uuml;berpr&uuml;fen, ob die Lamellen am Verdampfer in der Au&szlig;eneinheit verschmutzt sind&bull;&Uuml;berpr&uuml;fen, ob die Wartungsh&auml;hne in den Verbindungsleitungen zwischen Innenund Au&szlig;eneinheit geschlossen sind&bull;&Uuml;berpr&uuml;fen, ob de Vorund R&uuml;cklauf an der Au&szlig;eneinheit korrekt angeschlossen ist&bull;&Uuml;berpr&uuml;fen, ob sich Schmutzablagerungen im Filter in den Verbindungsleitungen zwischen Innenund Au&szlig;eneinheit befinden&bull;&Uuml;berpr&uuml;fen, ob die Umw&auml;lzpumpe im Solekreis zwischen Innenund Au&szlig;eneinheit blockiert ist&bull;-",correction => "Luftseitige Lamellen am Verdampfer in der Au&szlig;eneinheit reinigen&bull;Wartungsh&auml;hne in den Verbindungsleitugnen zwischen Innenund Au&szlig;eneineit &ouml;ffnen&bull;Vorund R&uuml;cklauf an der Au&szlig;eneinheit korrekt anschlie&szlig;en&bull;Filter in den Verbindungsleitugnen zwischen Innenund Au&szlig;eneinheit reinigen&bull;Umw&auml;lzpumpe im Solekreis zwischen Innenund Au&szlig;eneinheit anwerfen bzw. austauschen&bull;Zul&auml;ssige Solefl&uuml;ssigkeit verwenden"};
	$ErrorCodeList{"0677"} = {errorcode => "A01",                                    subcode => "5960",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: St&ouml;rmeldung vom Feuerungsautomat",                                                                                                                                                                                                                                                                                       checklist => "Hauptabsperreinrichtung &uuml;berpr&uuml;fen&bull;Ger&auml;teabsperrhan &uuml;berpr&uuml;fen&bull;Gasanschlu&szlig;flie&szlig;druck bei Nennw&auml;rmebelastung messen&bull;&Uuml;berpr&uuml;fen, ob der Anschlusstecker an der Ionisationselektrode komplett aufgesteckt ist&bull;Ionisationsstrom messen, Ionisationselektrode &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob das Z&uuml;ndkabel am Z&uuml;ndtransformator komplett aufgesteckt ist&bull;Z&uuml;ndkabel auf Besch&auml;digungen &uuml;berpr&uuml;fen&bull;Anschlusskabel f&uuml;r die Ionisationselektrode auf Besch&auml;digungen &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob die Anschlusstecker an den Sicherheitsventilen in der Gasarmatur komplett aufgesteckt sind&bull;Ohmschen Widerstand der Sicherheitsventile an der Gasarmatur messen&bull;&Uuml;berpr&uuml;fen, ob sich Schmutzablagerungen im Filter der Gasarmatur befinden&bull;Brennereinstellwert bei Nennw&auml;rmebelastung &uuml;berpr&uuml;fen&bull;Brennereinstellwert bei minimaler Nennw&auml;rmebelastung &uuml;berpr&uuml;fen&bull;Abgasanlage auf freien Querschnitt, Dichtheit und maximal zul&auml;ssige L&auml;nge inkl. Umlenkungen &uuml;berpr&uuml;fen&bull;Verbrennungsluftmenge &uuml;berpr&uuml;fen",correction => "Hauptabsperreinrichtung &ouml;ffnen&bull;Ger&auml;teabsperrhahn &ouml;ffnen&bull;Ger&auml;t stilllegen und Gasleitung &uuml;berpr&uuml;fen&bull;Anschlusstecker an der Ionisationselektrode komplett aufstecken&bull;Ionsiatonslektrode austauschen&bull;Z&uuml;ndkabel am Z&uuml;ndtransformator komplett aufstecken&bull;Z&uuml;ndkabel austauschen&bull;Anschlusskabel zur Ionisationselektrode austauschen&bull;Anschlussstecker an den Sicherheitsventilen der Gasarmatur komplett aufstecken&bull;Gasarmatur austauschen&bull;Schmutzablagerungen aus dem Filter in der Gasarmatur entfernen&bull;Brenner bei Nennw&auml;rmebelastung korrekt einstellen&bull;Brenner bei minimaler Nennbelastung korrekt einstellen&bull;Abgasanlage instandsetzen bzw. umbauen&bull;Verbrennungsluftverbund sicherstellen bzw. zu geringe Gr&ouml;&szlig;e der L&uuml;ftungs&ouml;ffnung korrigieren"};
	$ErrorCodeList{"0678"} = {errorcode => "A01",                                    subcode => "5960",                                               class => "8",           reason => "",                                                                                                                                                                                                                                                                                                                                                 checklist => "&Uuml;berpr&uuml;fen, ob die Absperrmembran in der Mischeinrichtung besch&auml;digt oder verschmutz ist&bull;&Uuml;berpr&uuml;fen, ob der Anschlussstecker f&uuml;r die Spannungsversorgung (230 VAC) am Gebl&auml;se komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob der Anschlussstecker f&uuml;r die Versorgungsspannung zum Gebl&auml;se im Feuerungsautomat komlett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel f&uuml;r die Spannungsversorgung (230 VAC) zum Gebl&auml;se besch&auml;digt ist&bull;&Uuml;berpr&uuml;fen, ob der  Anschlussstecker f&uuml;r die Spannungsversorgung (230 VAC) am Z&uuml;ndtransformator aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob der Anschlussstecker f&uuml;r die Spannungsversorgung (230 VAC) zum Z&uuml;ndtransformator im Feuerungsautomat komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel f&uuml;r die Spannungsversorgung zum Z&uuml;ndtransformator besch&auml;digt ist&bull;Funktion des Z&uuml;ndtransformators &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob die korrekte Gasart im Masterregler eingestellt ist&bull;-",correction => "Absperrmembran inder Mischeinrichtung austauschen bzw. reinigen&bull;Anschlussstecker f&uuml;r die Spannungsversorgung (230 VAC) am Gebl&auml;se komplett aufstecken&bull;Anschlussstecker f&uuml;r die Spannungsversorgung (230 VAC) zum Gebl&auml;se im Feuerungsautomat komplett aufstecken&bull;Anschlusskabel f&uuml;r die Spannungsversorung (230 VAC) zum Gebl&auml;se austauschen&bull;Anschlussstecker f&uuml;r die Spannungsversorgung (230 VAC) an Z&uuml;ndtransformator komplett aufstecken&bull;Anschlussstecker f&uuml;r die Spannungsversorung (230 VAC) zum Z&uuml;ndtransformator im Feuerungsautomat komplett aufstecken&bull;Anschlusskabel f&uuml;r die Spannungsversorgung (230 VAC) zum Z&uuml;ndtransformator austauschen&bull;Z&uuml;ndtransformator austauschen&bull;Korrekte Gasart im Masterregler einstellen&bull;Feuerungsautomat austauschen"};
	$ErrorCodeList{"0679"} = {errorcode => "A01",                                    subcode => "5962",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Drehzahl vom  Abluftventilator liegt au&szlig;erhalb des zul&auml;ssigen Bereiches",                                                                                                                                                                                                                                          checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker f&uuml;r die Drehzahlregelung am Abluftventilator komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel f&uuml;r die Drehzahlregelung am Abluftventilator besch&auml;digt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel f&uuml;r die Spannungsversorgung des Abluftventilators besch&auml;digt ist&bull;Funktion des Abluftventilators &uuml;berpr&uuml;fen&bull;-",                                                         correction => "Anschlussstecker f&uuml;r die Drehzahlregelung am Abluftventilator komplett aufstecken&bull;Anschlusskabel f&uuml;r die Drehzahlregelung des Abluftventilators austauschen&bull;Anschlusskabel f&uuml;r die Spannungsversorgung des Abluftventilators austauschen&bull;Abluftventilator austauschen&bull;Vorwiderstand f&uuml;r die Drehzahlbegrenzung des Abluftventilators austauschen"        };
	$ErrorCodeList{"0680"} = {errorcode => "A01",                                    subcode => "5963",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Gebl&auml;sedrehzahl au&szlig;erhalb des zul&auml;ssigen Bereiches",                                                                                                                                                                                                                                                          checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker f&uuml;r die Drehzahlregelung am Gebl&auml;se  komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel f&uuml;r die Drehzahlregelung am Gebl&auml;se besch&auml;digt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel f&uuml;r die Spannungsversorgung des Gebl&auml;ses besch&auml;digt ist&bull;Funktion des Gebl&auml;sese &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Anschlussstecker f&uuml;r die Spannungsversorgung (230 VAC) am Gebl&auml;se komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob der Anschlussstecker f&uuml;r die Spannungsversorgung (230 VAC) des Gebl&auml;sese im Feuerungsautomat komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel f&uuml;r die Spannungsversorgung (230 VAC) zum Gebl&auml;se besch&auml;digt ist&bull;-",correction => "Anschlussstecker f&uuml;r die Drehzahlregelung am Gebl&auml;se komplett aufstecken&bull;Anschlusskabel f&uuml;r die Drehzahlregelung des Gebl&auml;ses austauschen&bull;Anschlusskabel f&uuml;r die Spannungsversorgung des Gebl&auml;ses austauschen&bull;Gebl&auml;se austauschen&bull;Anschlussstecker f&uuml;r die Spannungsversorgung (230 VAC) am Gebl&auml;se komplett aufstecken&bull;Anschlussstecker f&uuml;r die Spannungsversorgung (230 VAC) zum Gebl&auml;se im Feuerungsautomat komplett aufstecken&bull;Anschlusskabel f&uuml;r die Spannungsversorgung (230 VAC) zum Gebl&auml;se austauschen&bull;Feuerungsautomat austauschen"};
	$ErrorCodeList{"0681"} = {errorcode => "A01",                                    subcode => "5964",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Fehler 2-Wege-Ventil am Verfl&uuml;ssiger (Heizungswasser)",                                                                                                                                                                                                                                                                  checklist => "&Uuml;berpr&uuml;fen Sie die Position des Motors am 2-Wege-Ventil&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel f&uuml;r das R&uuml;ckmeldesignal am Motor f&uuml;r das 2-WegeVentil besch&auml;digt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel f&uuml;r die Spannungsversorgung des Motors am 2-WegeVentil  besch&auml;digt ist&bull;&Uuml;berpr&uuml;fen, ob der Motor vom 2-Wege-Ventil blockiert ist&bull;-",                                                                        correction => "Position des Motors am 2-Wege-Ventil korrigieren&bull;Anschlusskabel f&uuml;r das R&uuml;ckmeldesignal am Motor des 2-Wege-Ventiles austausche&bull;Anschlusskabel f&uuml;r die Spannungsversorgung des Motors am 2-Wege-Ventil austauschen&bull;Motor am 2-Wege-Ventil austauschen&bull;2-Wege-Ventil austauschen"                                                                              };
	$ErrorCodeList{"0682"} = {errorcode => "A01",                                    subcode => "5965",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: St&ouml;rung am Gebl&auml;se in der Au&szlig;eneinheit",                                                                                                                                                                                                                                                                      checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker f&uuml;r die Drehzahlregelung am Gebl&auml;se in der Au&szlig;eneinheit komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob der Anschlussstecker f&uuml;r die Versorgungsspannung am Gebl&auml;se in der Au&szlig;eneinheit komplett aufgesteckt ist&bull;Funktion des Gebl&auml;ses in der Au&szlig;eneinheit &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Anschlussstecker f&uuml;r die Drehzahlregelung  zums Gebl&auml;se im Steuerger&auml;t der Au&szlig;eneinheit komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob der Anschlussstecker f&uuml;r die Versorgungsspannung zum Gebl&auml;se im Steuerger&auml;t der Au&szlig;eneinheit komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel  f&uuml;r die Drehzahlregelung zum Gebl&auml;se besch&auml;digt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel f&uuml;r die Versorgungsspannung zum Gebl&auml;se in der Au&szlig;eneinheit besch&auml;digt ist&bull;-",correction => "Anschlussstecker f&uuml;r die Drehzahlregelung am Gebl&auml;se in der Au&szlig;eneinheit komplett aufstecken&bull;Anschlussstecker f&uuml;r die Versorgungsspannung am Gebl&auml;se in der Au&szlig;eneinheit komplett aufstecken&bull;Gebl&auml;se bzw. Gebl&auml;semotor  in der Au&szlig;eneinheit austauschen&bull;Anschlussstecker f&uuml;r die Drehzahlregelung zum Gebl&auml;se im Steuerger&auml;t der Au&szlig;eneinheit komplett aufstecken&bull;Anschlussstecker f&uuml;r die Versorgungsspannung zum Gebl&auml;se im Steuerger&auml;t der Au&szlig;eneinheit komplett aufstecken&bull;Anschlusskabel f&uuml;r die Drehzahlregelung zum Gebl&auml;se in der Au&szlig;eneinheit austauschen&bull;Anschlusskabel f&uuml;r die Spannungsversorgung zum Gebl&auml;se in der Au&szlig;eneineit austauschen&bull;Steuerger&auml;t f&uuml;r die Gebl&auml;segelung in der Au&szlig;eneinheit austauschen"};
	$ErrorCodeList{"0683"} = {errorcode => "A01",                                    subcode => "5966",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Unzul&auml;ssige Gebl&auml;sedrehzahl in der Au&szlig;eneinheit",                                                                                                                                                                                                                                                             checklist => "&Uuml;berpr&uuml;fenb, ob der Anschlussstecker f&uuml;r die Drehzahlregelung am Gebl&auml;se in der Au&szlig;eneinheit komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob der Anschlussstecker f&uuml;r die Versorgungsspannung am Gebl&auml;se in der Au&szlig;eneinheit komplett aufgesteckt ist&bull;Funktion des Gebl&auml;ses in der Au&szlig;eneinheit &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Anschlussstecker f&uuml;r die Drehzahlregelung  zums Gebl&auml;se im Steuerger&auml;t der Au&szlig;eneinheit komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob der Anschlussstecker f&uuml;r die Versorgungsspannung zum Gebl&auml;se im Steuerger&auml;t der Au&szlig;eneinheit komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel  f&uuml;r die Drehzahlregelung zum Gebl&auml;se besch&auml;digt ist&bull;-",correction => "Anschlussstecker f&uuml;r die Drehzahlregelung am Gebl&auml;se in der Au&szlig;eneinheit komplett aufstecken&bull;Anschlussstecker f&uuml;r die Versorgungsspannung am Gebl&auml;se in der Au&szlig;eneinheit komplett aufstecken&bull;Gebl&auml;se bzw. Gebl&auml;semotor in der Au&szlig;eneinheit austauschen&bull;Anschlussstecker f&uuml;r die Drehzahlregelung zum Gebl&auml;se im Steuerger&auml;t der Au&szlig;eneinheit komplett aufstecken&bull;Anschlussstecker f&uuml;r die Versorgungsspannung zum Gebl&auml;se im Steuerger&auml;t der Au&szlig;eneinheit komplett aufstecken&bull;Anschlusskabel f&uuml;r die Drehzahlregelung zum Gebl&auml;se  in der Au&szlig;eneinheit austauschen&bull;Anschlusskabel f&uuml;r die Spannungsversorgung zum Gebl&auml;se in der Au&szlig;eneineit austauschen"};
	$ErrorCodeList{"0684"} = {errorcode => "A01",                                    subcode => "5970",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: F&uuml;llstand des K&auml;ltemittels (NH3) im Austreiber zu niedrig",                                                                                                                                                                                                                                                         checklist => "Anschlusskabel zum F&uuml;llstandssensor &uuml;berpr&uuml;fen&bull;F&uuml;llstandssensor f&uuml;r das K&auml;ltemittel im Austreiber &uuml;berpr&uuml;fen&bull;Anlagenhydraulik &uuml;berpr&uuml;fen&bull;Fremdgas aus K&auml;ltemittelkreis (NH3) entfernen&bull;Funktion der Regelventile im K&auml;ltemittelkreis (NH3) &uuml;berpr&uuml;fen&bull;-",                                                                                                                                              correction => "Anschlusskabel zum F&uuml;llstandssensor austauschen&bull;Werkskundendienst benachrichtigen&bull;Anlagenhydraulik korrigieren&bull;Werkskundendienst benachrichtigen&bull;Werkskundendienst benachrichtigen&bull;Werkskundendienst benachrichtigen"                                                                                                                                              };
	$ErrorCodeList{"0685"} = {errorcode => "A01",                                    subcode => "5971",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: R&uuml;ckstau Abgaskondensat im Abgasw&auml;rmetauscher",                                                                                                                                                                                                                                                                     checklist => "&Uuml;berpr&uuml;fen, ob der Kondensatanschlu&szlig; am Abgasw&auml;rmetauscher durch Ablagerungen blockiert ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Sensor FLS besch&auml;digt ist&bull;Funktion des Sensors FLS &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob die Kondensatleitung mit Gef&auml;lle verlegt worden ist&bull;-",                                                                                                                                               correction => "R&uuml;ckst&auml;nde am Kondensatanschlu&szlig; im Abgasw&auml;rmetauscher entfernn&bull;Anschlusskabel zum Sensor FLS austauschen&bull;Sensor FLS am Abgasw&auml;rmetauscher austauschen&bull;Kondensatleitung mit Gef&auml;lle verlegen&bull;Externen Kondensatsiphon reinigen"                                                                                                                };
	$ErrorCodeList{"0686"} = {errorcode => "A01",                                    subcode => "5972",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Aufheizzeit imK&auml;ltekreis nach Start zu lang",                                                                                                                                                                                                                                                                            checklist => "Ohmschen Widerstand des R&uuml;cklauftemperaturf&uuml;hlers TTI-5 messen&bull;&Uuml;berpr&uuml;fen, ob der R&uuml;cklauftemperaturf&uuml;hler; TTI-5 korrekten Kontakt an der Vorlaufleitung hat.&bull;Position des R&uuml;cklauftemperaturf&uuml;hlers TTI-5 an der Vorlaufleitung &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob die Isolierung am R&uuml;cklauftemperaturf&uuml;hler TTI-5 vorhanden bzw. besch&auml;digt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum R&uuml;cklauftemperaturf&uuml;hler TTI-5 besch&auml;digt ist&bull;Ohmschen Widerstand des Temperaturf&uuml;hlers TTI-6 messen&bull;&Uuml;berpr&uuml;fen, ob der Vorlauftemperaturf&uuml;hler TTI-6 korrekten Kontakt an der Vorlaufleitung hat.&bull;Position des Vorlauftemperaturf&uuml;hlers TTI-6 an der Vorlaufleitung &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob die Isolierung am Vorlauftemperaturf&uuml;hler TTI-6 vorhanden bzw. besch&auml;digt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Temperaturf&uuml;hler TTI-6 besch&auml;digt ist",correction => "R&uuml;cklauftemperaturf&uuml;hler TTI-5 austauschen&bull;Kontakt zwischen R&uuml;cklauftemperaturf&uuml;hler TTI-5 und R&uuml;cklaufleitung herstellen&bull;R&uuml;cklauftemperaturf&uuml;hler TTI-5 an der korrekten Position der R&uuml;cklaufleitung montieren&bull;Isolierung am R&uuml;cklauftemperaturf&uuml;hler TTI-5 austauschen bzw. montieren&bull;Kabelbaum austauschen&bull;Vorlauftemperaturf&uuml;hler TTI-6 austauschen&bull;Kontakt zwischen Vorlauftemperaturf&uuml;hler TTI-6 und Vorlaufleitung herstellen&bull;Vorlauftemperaturf&uuml;hler TTI-6 an der korrekten Position der Vorlaufleitung montieren&bull;Isolierung am Vorlauftemperaturf&uuml;hler TTI-6 austauschen bzw. montieren&bull;Kabelbaum austauschen"};
	$ErrorCodeList{"0687"} = {errorcode => "A01",                                    subcode => "5972",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Aufheizzeit imK&auml;ltekreis nach Start zu lang",                                                                                                                                                                                                                                                                            checklist => "&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Temperaturf&uuml;hler TTI-12 besch&auml;digt ist&bull;Ohmschen Widerstand des Temperaturf&uuml;hlers TTI-12 &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Temperaturf&uuml;hler TTI-12 korrekten Kontakt zur Heizwasserleitung am Verfl&uuml;ssigerausgang hat&bull;&Uuml;berpr&uuml;fen, ob die Isolierung am Temperaturf&uuml;hler TTI-12 vorhanden bzw. besch&auml;digt ist&bull;Brennereinstellung bei minimaler Nennw&auml;rmebelastung &uuml;berpr&uuml;fen&bull;Brennereinstellung  bei maximaler Nennw&auml;rmebelastung &uuml;berpr&uuml;fen&bull;Funktion der Regelventile im K&auml;ltemittelkreis &uuml;berp&uuml;fen&bull;-",correction => "Kabelbaum autauschen&bull;Temperaturf&uuml;hler TTI-12 austauschen&bull;Temperaturf&uuml;hler TTI-12 an der korrekten Position der Heizwasserleitung am Verfl&uuml;ssigerausgang montieren&bull;Isolierung am Temperaturf&uuml;hler TTI-12 austauschen bzw. montieren&bull;Brennereinstellung bei minimaler Nennw&auml;rmebelastung korrigieren&bull;Brennereinstellung bei maximaler Nennw&auml;rmebelastung korrigieren&bull;Werkskundendienst benachrichtigen&bull;Werkskundendienst benachrichtigen&bull;Brennereinstellung bei maximaler Nennw&auml;rmebelastung korrigieren&bull;Werkskundendienst benachrichtigen"};
	$ErrorCodeList{"0688"} = {errorcode => "A01",                                    subcode => "5973",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: L&ouml;sungsmitteldrossel ist komplett ge&ouml;ffnet",                                                                                                                                                                                                                                                                        checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker f&uuml;r die L&ouml;sungsmitteldrossel im Steuerger&auml;t (CU I/O) komplett aufgesteckt ist&bull;Anschlusskabel zum Magnet an der L&ouml;sungsmitteldrossel auf Besch&auml;digungen &uuml;berpr&uuml;fen&bull;Ohmschen Widerstand am Magnet der L&ouml;sungemitteldrossel messe&bull;Motor an der L&ouml;sungsmittelpumpe &uuml;berpr&uuml;fen&bull;-",                                                                                                correction => "Anschlussstecker f&uuml;r die L&ouml;sungsmitteldrossel am Steuerger&auml;t (CU I/O) komplett aufstecken&bull;Anschlusskabel zum Magnet an der L&ouml;sungsmitteldrossel austauschen&bull;Magnet an der L&ouml;sungsmitteldrossel austauschen&bull;Motor an der L&ouml;sungspumpe austauschen&bull;Werkskundendienst benachrichtigen"                                                            };
	$ErrorCodeList{"0689"} = {errorcode => "A01",                                    subcode => "5974",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Enteisung der Au&szlig;eneinheit fehlgeschlagen",                                                                                                                                                                                                                                                                             checklist => "&Uuml;berpr&uuml;fen, ob die Lamellen am Verdampfer vereist sind&bull;&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Abtauventil im K&auml;ltemittelkreis aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel zum Abtauventil im K&auml;ltemittelkreis besch&auml;digt ist&bull;Ohmschen Widerstand der Magnetspulte am Abtauventil mesen&bull;-",                                                                                                                                     correction => "Aufstellort f&uuml;r die Au&szlig;eneinheit ver&auml;ndern&bull;Anschlussstecker am Abtauventil im K&auml;ltemittelkreis aufstecken&bull;Anschlusskabel zum Abtauventil im K&auml;ltemittelkreis austauschen&bull;Magentspule am Abtauventil im K&auml;ltemittelkreis austauschen&bull;Korrekte Umlaufmenge im Solekreis zwischen Au&szlig;enund Inneneinheit einstellen"                        };
	$ErrorCodeList{"0690"} = {errorcode => "A01",                                    subcode => "5992",                                               class => "8",           reason => "Gas-W&auml;rmepumpe: Sicherheitskette offen",                                                                                                                                                                                                                                                                                                      checklist => "&Uuml;berpr&uuml;fen, ob die Br&uuml;cke zwischen den Anschlussklemmen 62/64 bzw. 63/64 fehlt&bull;&Uuml;berpr&uuml;fen, ob der Anschlussstecker auf der Leiterplatte im Steuerger&auml;t GHP-SC komplett aufgesteckt ist&bull;-",                                                                                                                                                                                                                                                                    correction => "Br&uuml;cke zwischen den Anschlussklemmen 62/64 bzw. 63/64 im Steuerger&auml;t GHP-SC einbauen&bull;Anchlussstecker auf der Leiterplatte im Steuerger&auml;t GHP-SC komplett aufstecken&bull;Ursache f&uuml;r das Ansprechen der externen Sicherheitseinrichtung beseitigen"                                                                                                                     };
	$ErrorCodeList{"0691"} = {errorcode => "A01",                                    subcode => "6200",                                               class => "4",           reason => "Warnung Z2 Vorlauftemperatursensor T0 Kurzschluss",                                                                                                                                                                                                                                                                                                checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0692"} = {errorcode => "A01",                                    subcode => "6201",                                               class => "4",           reason => "Warnung Z2 Vorlauftemperatursensor T0 Kabelbruch",                                                                                                                                                                                                                                                                                                 checklist => "siehe Serviceunterlage der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0693"} = {errorcode => "A01",                                    subcode => "7400",                                               class => "8",           reason => "Gebl&auml;se Abluft steht",                                                                                                                                                                                                                                                                                                                        checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Gebl&auml;se Abluft pr&uuml;fen, Steckerverbindunge auf dem Steuerger&auml;t pr&uuml;fen."                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0694"} = {errorcode => "A01",                                    subcode => "7401",                                               class => "8",           reason => "Gebl&auml;se Zuluft steht",                                                                                                                                                                                                                                                                                                                        checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Gebl&auml;se Abluft pr&uuml;fen, Steckerverbindunge auf dem Steuerger&auml;t pr&uuml;fen."                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0695"} = {errorcode => "A01",                                    subcode => "7402",                                               class => "8",           reason => "Die Bypassklappe h&auml;ngt fest",                                                                                                                                                                                                                                                                                                                 checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Stellung der Bypassklappe pr&uuml;fen, ggf. l&ouml;sen und schmieren."                                                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"0696"} = {errorcode => "A01",                                    subcode => "7403",                                               class => "8",           reason => "Kurzschluss oder Unterbrechung im F&uuml;hler oder elektrischen Anschlussleitung vom Au&szlig;enlufttemperaturf&uuml;hler S1",                                                                                                                                                                                                                     checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "F&uuml;hler und Anschlussleitung pr&uuml;fen, ggf. ersetzen"                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0697"} = {errorcode => "A01",                                    subcode => "7404",                                               class => "8",           reason => "Kurzschluss oder Unterbrechung im F&uuml;hler oder elektrischen Anschlussleitung vom Zulufttemperaturf&uuml;hler S2",                                                                                                                                                                                                                              checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "F&uuml;hler und Anschlussleitung pr&uuml;fen, ggf. ersetzen"                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0698"} = {errorcode => "A01",                                    subcode => "7405",                                               class => "8",           reason => "Kurzschluss oder Unterbrechung im F&uuml;hler oder elektrischen Anschlussleitung vom Ablufttemperaturf&uuml;hler S3",                                                                                                                                                                                                                              checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "F&uuml;hler und Anschlussleitung pr&uuml;fen, ggf. ersetzen"                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0699"} = {errorcode => "A01",                                    subcode => "7406",                                               class => "8",           reason => "Kurzschluss oder Unterbrechung im F&uuml;hler oder elektrischen Anschlussleitung vom Fortlufttemperaturf&uuml;hler S4",                                                                                                                                                                                                                            checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "F&uuml;hler und Anschlussleitung pr&uuml;fen, ggf. ersetzen"                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0700"} = {errorcode => "A01",                                    subcode => "7407",                                               class => "8",           reason => "Kurzschluss oder Unterbrechung im F&uuml;hler oder elektrischen Anschlussleitung vom Raumlufttemperaturf&uuml;hler S5",                                                                                                                                                                                                                            checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "F&uuml;hler und Anschlussleitung pr&uuml;fen, ggf. ersetzen"                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0701"} = {errorcode => "A01",                                    subcode => "7408",                                               class => "8",           reason => "Luftfeuchtef&uuml;hler befindet sich im Abluftkanal des L&uuml;ftungsger&auml;tes",                                                                                                                                                                                                                                                                checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Luftfeuchtef&uuml;hler im Zuluftkanal positionieren."                                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"0702"} = {errorcode => "A01",                                    subcode => "7409",                                               class => "8",           reason => "Au&szlig;enlufttemperatur niedriger als -13°C und Vorheizer au&szlig;er Betrieb",                                                                                                                                                                                                                                                                 checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "&Uuml;ber Fernbedienung oder Konfigurationstool sicherstellen, dass der Vorheizer aktiv ist.; Pr&uuml;fen ob der eingebauter Vorheizer funktioniert. Kundendienst anrufen."                                                                                                                                                                                                                      };
	$ErrorCodeList{"0703"} = {errorcode => "A01",                                    subcode => "7410",                                               class => "8",           reason => "Zulufttemperatur niedriger als 15°C",                                                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Pr&uuml;fen, ob die Heizung des Geb&auml;udes funktioniert. Pr&uuml;fen, ob der eingebaute Vorheizer funktioniert. Pr&uuml;fen ob der Vorheizer aktiv ist. Stellung der Bypassklappe pr&uuml;fen, ggf. l&ouml;sen und schmieren. Kundendienst anrufen."                                                                                                                                          };
	$ErrorCodeList{"0704"} = {errorcode => "A01",                                    subcode => "7411",                                               class => "8",           reason => "Ein Temperaturf&uuml;hler misst eine Temperatur gr&ouml;&szlig;er 70°C",                                                                                                                                                                                                                                                                          checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Ger&auml;t vom Netz trennen. Kundendienst anrufen."                                                                                                                                                                                                                                                                                                                                              };
	$ErrorCodeList{"0705"} = {errorcode => "A02",                                    subcode => "816",                                                class => "A",           reason => "Keine Kommunikation mit Ger&auml;teelektronik (keine Auswirkung auf das Regelverhalten, aber keine Bedienm&ouml;glichkeit)",                                                                                                                                                                                                                       checklist => "Pr&uuml;fen, ob die Ger&auml;teelektronik BC10 richtig montiert ist bzw. richtig sitzt&bull;Pr&uuml;fen, ob alle Busleitungen zwischen Ger&auml;teelektronik BC10 und Ger&auml;teelektronik MC10 richtig montiert sind&bull;Ger&auml;teelektronik defekt",                                                                                                                                                                                                                                            correction => "Kaskadenmodul richtig montieren&bull;Busleitung richtig montieren&bull;Ger&auml;teelektronik BC10 austauschen"                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"0706"} = {errorcode => "A03",                                    subcode => "816",                                                class => "4",           reason => "Keine Buskommunikation zwischen Universalschaltmodul und Ger&auml;teelektronik (zwischen UBA-H3 und UIM/LM10)",                                                                                                                                                                                                                                    checklist => "Die Verdrahtung zwischen Universalschaltmodul und Ger&auml;teelektronik pr&uuml;fen&bull;Verbindungsleitung zwischen Universalschaltmodul und Ger&auml;teelektronik pr&uuml;fen&bull;Universalschaltmodul defekt",                                                                                                                                                                                                                                                                                    correction => "Bei Fehlverdrahtung den Verdrahtungsfehler korrigieren&bull;Bei Defekt Datenleitung ersetzen&bull;Modul austauschen"                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"0707"} = {errorcode => "A1",                                     subcode => "33",                                                 class => "B",           reason => "Heizung wegen Trockenlauf der Heizungspumpe ausgeschaltet",                                                                                                                                                                                                                                                                                        checklist => "Pumpenlager ist trockengelaufen",                                                                                                                                                                                                                                                                                                                                                                                                                                                                     correction => "Umw&auml;lzpumpe austauschen"                                                                                                                                                                                                                                                                                                                                                                    };
	$ErrorCodeList{"0708"} = {errorcode => "A9",                                     subcode => "11",                                                 class => "",            reason => "Warmwassertemperaturf&uuml;hler Ger&auml;t falsch montiert oder SchichtladespeicherNTC3 defekt",                                                                                                                                                                                                                                                   checklist => "Montageort des Warmwassertemperaturf&uuml;hler &uuml;berpr&uuml;fen&bull;Kontakt zwischen Warmwassertemperaturf&uuml;hler und Warmwasserrohr &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                   correction => "Warmwassertemperaturf&uuml;hler am Warmwasserrohr montieren&bull;Warmwassertemperaturf&uuml;hler korrekt auf die Einbauposition montieren"                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0709"} = {errorcode => "A11",                                    subcode => "801",                                                class => "14",          reason => "Systemst&ouml;rung Bedieneinheit",                                                                                                                                                                                                                                                                                                                 checklist => "Im St&ouml;rungsspeicher der Bedieneinheit pr&uuml;fen, ob diese St&ouml;rung einmalig oder h&auml;ufiger vorhanden ist",                                                                                                                                                                                                                                                                                                                                                                             correction => "Wenn die St&ouml;rung mehr als ein Mal vorhanden ist, dann ist die Bedieneinheit auszutauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0710"} = {errorcode => "A11",                                    subcode => "802",                                                class => "A",           reason => "Uhrzeit noch nicht eingestellt",                                                                                                                                                                                                                                                                                                                   checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Uhrzeit an Bedieneinheit einstellen"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"0711"} = {errorcode => "A11",                                    subcode => "803",                                                class => "A",           reason => "Datum noch nicht eingestellt",                                                                                                                                                                                                                                                                                                                     checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Datum am Bedieneinheit einstellen"                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0712"} = {errorcode => "A11",                                    subcode => "804",                                                class => "A",           reason => "Systemst&ouml;rung Bedieneinheit",                                                                                                                                                                                                                                                                                                                 checklist => "Im St&ouml;rungsspeicher der Bedieneinheit pr&uuml;fen, ob diese St&ouml;rung einmalig oder h&auml;ufiger vorhanden ist",                                                                                                                                                                                                                                                                                                                                                                             correction => "Datum am Bedieneinheit einstellen"                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0713"} = {errorcode => "A11",                                    subcode => "805; 806",                                           class => "A",           reason => "Der Raumtemperaturf&uuml;hler der Bedieneinheit ist defekt.",                                                                                                                                                                                                                                                                                      checklist => "Anzeige nur bis Bedieneinheit Version 1.03",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Bedieneinheit durch eine aktuelle Version ersetzen"                                                                                                                                                                                                                                                                                                                                              };
	$ErrorCodeList{"0714"} = {errorcode => "A11",                                    subcode => "821; 822",                                           class => "A",           reason => "Keine Heizkreis-Fernbedienung; (821 = Heizkreis 1;; 822 = Heizkreis 2)",                                                                                                                                                                                                                                                                           checklist => "In der Installationsebene die Parametrierung pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "Entweder eine andere Heizkreisart als; „Raumvorlauf“ oder „Raumleistung“ ausw&auml;hlen oder dem Heizkreis eine Fernbedienung zuordnen"                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0715"} = {errorcode => "A11",                                    subcode => "823; 824",                                           class => "A",           reason => "Keine Heizkreis-Fernbedienung; (823 = Heizkreis 1;; 824 = Heizkreis 2)",                                                                                                                                                                                                                                                                           checklist => "In der Installationsebene die Parametrierung pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "Entweder eine andere Frostschutzart ausw&auml;hlen oder dem Heizkreis eine Fernbedienung zuordnen"                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0716"} = {errorcode => "A11",                                    subcode => "825",                                                class => "14",          reason => "Zwei Master-Bedieneinheiten im System",                                                                                                                                                                                                                                                                                                            checklist => "In der Installationsebene die Parametrierung pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "Die Bedieneinheit f&uuml;r den Heizkreis 1 als Master anmelden"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0717"} = {errorcode => "A11",                                    subcode => "826; 827",                                           class => "A",           reason => "Heizkreis-2-Bedieneinheit-; Fernbedienung (826 = Heizkreis 1;; 827 = Heizkreis 2)",                                                                                                                                                                                                                                                                checklist => "Bei Anschluss eines externen Raumtemperaturf&uuml;hlers die Verbindungsleitung zu diesem F&uuml;hler auf Kurzschluss pr&uuml;fen&bull;Bei Verwendung des Raumtemperaturf&uuml;hlers in der Bedieneinheit im Men&uuml; „Monitor“ die Raumtemperatur abfragen",                                                                                                                                                                                                                                     correction => "Kurzschluss beheben&bull;Sollte kein Wert angezeigt werden, ist die Bedieneinheit defekt und muss ausgetauscht werden"                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"0718"} = {errorcode => "A11",                                    subcode => "1000",                                               class => "12",          reason => "Systemkonfiguration nicht best&auml;tigt",                                                                                                                                                                                                                                                                                                         checklist => "Systemkonfiguration nicht vollst&auml;ndig durchgef&uuml;hrt",                                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "System vollst&auml;ndig konfigurieren und best&auml;tigen"                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0719"} = {errorcode => "A11",                                    subcode => "1004",                                               class => "8",           reason => "Autark-Bedieneinheit: Keine Kommunikation mit Funktionsmodul",                                                                                                                                                                                                                                                                                     checklist => "Eine Autark-Bedieneinheit wird laut Konfiguration verlangt, wird aber nicht gefunden.&bull;-",                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Verdrahtung EMS-BUS pr&uuml;fen&bull;Typ und Funktion Autark-Bedieneinheit pr&uuml;fen"                                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"0720"} = {errorcode => "A11",                                    subcode => "1006",                                               class => "8",           reason => "Autark-Bedieneinheit: Keine Betriebsart gew&auml;hlt.",                                                                                                                                                                                                                                                                                            checklist => "Die Autark-Bedieneinheit muss auf die; zur Anlage passende Betriebsart eingestellt werden.&bull;-",                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Einstellung der Betriebsart der AutarkBedieneinheit pr&uuml;fen&bull;Ggfs. Konfigurationsassistent erneut starten"                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0721"} = {errorcode => "A11",                                    subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus",                                                                                                                                                                                                                                                                                            checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder einschalten Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung&bull;-",                                                                                                                                                                                                                                     correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t ausund wieder einschalten&bull;Busleitung reparieren bzw. austauschen&bull;Defekten EMS-BUS-Teilnehmer austauschen"                                                                                                                                                                                                                            };
	$ErrorCodeList{"0722"} = {errorcode => "A11",                                    subcode => "1030; 1033; 1034; 1035; 1036",                       class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen",                                                                                                                                                                                                                                                                                                checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0723"} = {errorcode => "A11",                                    subcode => "1037",                                               class => "14",          reason => "Au&szlig;entemperaturf&uuml;hler defekt; Ersatzbetrieb Heizung aktiv",                                                                                                                                                                                                                                                                             checklist => "Konfiguration pr&uuml;fen; Mit der gew&auml;hlten Einstellung ist ein Au&szlig;entemperaturf&uuml;hler erforderlich&bull;Verbindungsleitung zwischen Regelger&auml;t und Au&szlig;entemperaturf&uuml;hler auf Durchgang pr&uuml;fen&bull;Elektrischen Anschluss der Verbindungsleitung im Au&szlig;entemperaturf&uuml;hler bzw. am Stecker im Regelger&auml;t pr&uuml;fen&bull;Au&szlig;entemperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Au&szlig;entemperaturf&uuml;hlers im Regelger&auml;t laut Tabelle pr&uuml;fen",correction => "Konfiguration &auml;ndern&bull;Wenn kein Durchgang vorhanden ist, die St&ouml;rung beheben&bull;Falls Schrauben oder ein Stecker lose sind, das Kontaktproblem beheben&bull;Wenn Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Regelger&auml;t austauschen"       };
	$ErrorCodeList{"0724"} = {errorcode => "A11",                                    subcode => "1038",                                               class => "16",          reason => "Datum/Zeit ung&uuml;ltiger Wert",                                                                                                                                                                                                                                                                                                                  checklist => "Datum/Zeit noch nicht eingestellt&bull;Spannungsversorgung &uuml;ber l&auml;ngere Zeit ausgefallen",                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Datum/Zeit einstellen&bull;Spannungsausf&auml;lle vermeiden"                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0725"} = {errorcode => "A11",                                    subcode => "1039",                                               class => "16",          reason => "W&auml;rmeerzeuger nicht f&uuml;r Estrichtrocknung mit ungemischten Heizkreisen geeignet",                                                                                                                                                                                                                                                         checklist => "W&auml;rmeerzeuger ist nicht f&uuml;r Estrichtrocknung mit ungemischten Heizkreisen geeignet",                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Ungemischten Heizkreis in gemischten Heizkreis umbauen"                                                                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"0726"} = {errorcode => "A11",                                    subcode => "1040",                                               class => "16",          reason => "Estrichtrocknung mit ungemischten Heizkreisen nur mit Gesamtanlage m&ouml;glich",                                                                                                                                                                                                                                                                  checklist => "Konfiguration der Estrichtrocknung pr&uuml;fen;; die gew&auml;hlte Einstellung k&ouml;nnte zu einer &Uuml;berheizung f&uuml;hren; die Estrichtrocknung wird daher nicht ausgef&uuml;hrt",                                                                                                                                                                                                                                                                                                             correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0727"} = {errorcode => "A11",                                    subcode => "1041",                                               class => "16",          reason => "Spannungsausfall w&auml;hrend Estrichtrocknung",                                                                                                                                                                                                                                                                                                   checklist => "W&auml;hrend der Estrichtrocknung ist ein l&auml;ngerer Spannungsausfall aufgetreten Estrichtrocknung wird fortgesetzt",                                                                                                                                                                                                                                                                                                                                                                              correction => "Spannungsausf&auml;lle vermeiden"                                                                                                                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"0728"} = {errorcode => "A11",                                    subcode => "1042",                                               class => "14",          reason => "Interner Fehler",                                                                                                                                                                                                                                                                                                                                  checklist => "Interner Fehler:; Zugriff auf Uhrenbaustein blockiert",                                                                                                                                                                                                                                                                                                                                                                                                                                               correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0729"} = {errorcode => "A11",                                    subcode => "1051; 1052; 1053; 1054; 1055; 1056; 1057; 1058",     class => "14",          reason => "Keine Kommunikation Modul externer Raumtemperaturf&uuml;hler  (1051 = Heizkreis 1;; 1052 = Heizkreis 2;; 1053 = Heizkreis 3;; 1054 = Heizkreis 4;; 1055 = Heizkreis 5;; 1056 = Heizkreis 6;; 1057 = Heizkreis 7;; 1058 = Heizkreis 8)",                                                                                                            checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); Mit der gew&auml;hlten Einstellung ist ein Modul externer Raumtemperaturf&uuml;hler erforderlich&bull;Die Verbindungsleitung EMS zum Modul externer Raumtemperaturf&uuml;hler auf Besch&auml;digung pr&uuml;fen; die Busspannung am Modul externer Raumtemperaturf&uuml;hler muss; zwischen 12  und 15 V DC liegen&bull;Modul externer Raumtemperaturf&uuml;hler defekt",                                                                     correction => "Konfiguration &auml;ndern&bull;Besch&auml;digte Kabel austauschen&bull;Modul externer Raumtemperaturf&uuml;hler austauschen"                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0730"} = {errorcode => "A11",                                    subcode => "1091",                                               class => "8",           reason => "Keine Kommunikation mit Brennstoffzelle",                                                                                                                                                                                                                                                                                                          checklist => "Kommunikation unterbrochen",                                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "BUS-Verbindung &uuml;berpr&uuml;fen"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"0731"} = {errorcode => "A11",                                    subcode => "1092",                                               class => "8",           reason => "Konfigurationsfehler: Brennstoffzelle nicht verwendet",                                                                                                                                                                                                                                                                                            checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0732"} = {errorcode => "A11",                                    subcode => "1093",                                               class => "8",           reason => "Keine Kommunikation mit Funkmodul",                                                                                                                                                                                                                                                                                                                checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0733"} = {errorcode => "A11",                                    subcode => "3001; 3002; 3003; 3004; 3005; 3006; 3007; 3008",     class => "14",          reason => "Konfigurationsfehler: Modul externer Raumtemperaturf&uuml;hler nicht verwendet (3001 = Heizkreis 1;; 3002 = Heizkreis 2;; 3003 = Heizkreis 3;; 3004 = Heizkreis 4;; 3005 = Heizkreis 5;; 3006 = Heizkreis 6;; 3007 = Heizkreis 7;; 3008 = Heizkreis 8)",                                                                                           checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); Es ist ein Modul externer Raumtemperaturf&uuml;hler im System, das mit der gew&auml;hlten; Einstellung nicht verwendet wird",                                                                                                                                                                                                                                                                                                                 correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0734"} = {errorcode => "A11",                                    subcode => "3011; 3012",                                         class => "14",          reason => "Konfigurationsfehler: Mischermodul nicht verwen-",                                                                                                                                                                                                                                                                                                 checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul)&bull;Es ist ein Mischermodul im System, das mit der&bull;gew&auml;hlten Einstellung nicht verwendet wird&bull;-&bull;-",                                                                                                                                                                                                                                                                                                                       correction => "Konfiguration &auml;ndern&bull;-&bull;-"                                                                                                                                                                                                                                                                                                                                                         };
	$ErrorCodeList{"0735"} = {errorcode => "A11",                                    subcode => "3061; 3062",                                         class => "14",          reason => "Keine Kommunikation mit Mischermodul",                                                                                                                                                                                                                                                                                                             checklist => "Konfiguration pr&uuml;fen (Adresseinstellung); Mit der gew&auml;hlten Einstellung ist ein Mischermo-&bull;dul erforderlich&bull;Die Verbindungsleitung EMS zum Mischermodul auf Besch&auml;digung pr&uuml;fen; Die Busspannung am Mischermodul muss zwischen 12 und 15 V DC liegen&bull;Mischermodul defekt&bull;-",                                                                                                                                                                                  correction => "Konfiguration &auml;ndern&bull;-&bull;Besch&auml;digte Kabel austauschen&bull;Mischermodul austauschen&bull;-"                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"0736"} = {errorcode => "A11",                                    subcode => "3071; 3072",                                         class => "14",          reason => "Keine Kommunikation mit Fernbedienung",                                                                                                                                                                                                                                                                                                            checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); mit der gew&auml;hlten Einstellung ist eine&bull;Fernbedienung erforderlich&bull;Die Verbindungsleitung EMS zur Fernbedienung auf Besch&auml;digung pr&uuml;fen; Die Busspannung an der Fernbedienung muss zwischen 12 und 15 V DC liegen&bull;Fernbedienung defekt&bull;-",                                                                                                                                                                  correction => "Konfiguration &auml;ndern&bull;-&bull;Besch&auml;digte Kabel austauschen&bull;Fernbedienung austauschen&bull;-"                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0737"} = {errorcode => "A11",                                    subcode => "3081; 3082; 3083; 3084; 3085; 3086; 3087; 3088",     class => "16",          reason => "Konfigurationsfehler: Fernbedienung nicht verwendet (3081 = Heizkreis 1;; 3082 = Heizkreis 2;; 3083 = Heizkreis 3;; 3084 = Heizkreis 4;; 3085 = Heizkreis 5;; 3086 = Heizkreis 6;; 3087 = Heizkreis 7;; 3088 = Heizkreis 8)",                                                                                                                      checklist => "Konfiguration pr&uuml;fen; Es ist eine Fernbedienung im System, die mit der gew&auml;hlten Einstellung nicht verwendet wird",                                                                                                                                                                                                                                                                                                                                                                         correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0738"} = {errorcode => "A11",                                    subcode => "3091; 3092; 3093; 3094; 3095; 3096; 3097; 3098",     class => "14",          reason => "Raumtemperaturf&uuml;hler defekt; (3091 = Heizkreis 1;; 3092 = Heizkreis 2;; 3093 = Heizkreis 3;; 3094 = Heizkreis 4;; 3095 = Heizkreis 5;; 3096 = Heizkreis 6;; 3097 = Heizkreis 7;; 3098 = Heizkreis 8)",                                                                                                                                        checklist => "1. Systemregler Installationsort; 2. Systemregler Einstellung Regelungsart; 3. Systemregler Einstellung Frostschutz; 4. Systemregler Einstellung Raumeinfluss; 5. Systemregler oder Fernbedienung defekt",                                                                                                                                                                                                                                                                                            correction => "1. RC300 im Wohnraum installieren (nicht am Kessel) oder; 2. Regelungsart Heizkreis von Raumgef&uuml;hrt auf Au&szlig;entemperaturgef&uuml;hrt umstellen; 3. Frostschutz von Raum auf Au&szlig;en umstellen; 4. Einstellung Raumeinfluss auf 0 stellen; 5. oder Fernbedienung austauschen"                                                                                                       };
	$ErrorCodeList{"0739"} = {errorcode => "A11",                                    subcode => "3111; 3112; 3113; 3114; 3115; 3116; 3117; 3118",     class => "14",          reason => "Konfigurationsfehler: falsche Fernbedienung; (3111 = Heizkreis 1;; 3112 = Heizkreis 2;; 3113 = Heizkreis 3;; 3114 = Heizkreis 4;; 3115 = Heizkreis 5;; 3116 = Heizkreis 6;; 3117 = Heizkreis 7;; 3118 = Heizkreis 8)",                                                                                                                             checklist => "Konfiguration pr&uuml;fen Widerspr&uuml;chliche Konfigurationen der; Fernbedienung hinterlegt, z. B. eingestellt RC100, tats&auml;chlich RC200",                                                                                                                                                                                                                                                                                                                                                      correction => "Konfiguration &auml;ndern bzw. Fernbedienung austauschen"                                                                                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"0740"} = {errorcode => "A11",                                    subcode => "3151; 3152; 3153; 3154; 3155; 3156; 3157; 3158",     class => "8",           reason => "Heizkreismodul inkompatibel",                                                                                                                                                                                                                                                                                                                      checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0741"} = {errorcode => "A11",                                    subcode => "4001",                                               class => "12",          reason => "Konfigurationsfehler: Warmwasserger&auml;t und Regler stimmen nicht &uuml;berein",                                                                                                                                                                                                                                                                 checklist => "Konfiguration pr&uuml;fen; Widerspr&uuml;chliche Konfigurationen in Ger&auml;teelektronik und Systemregler hinterlegt",                                                                                                                                                                                                                                                                                                                                                                               correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0742"} = {errorcode => "A11",                                    subcode => "4011; 4012",                                         class => "14",          reason => "Keine Kommunikation mit Modul Warmwasser (4011 =; Warmwassersystem 1;; 4012 =; Warmwassersystem 2)",                                                                                                                                                                                                                                               checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); mit der gew&auml;hlten Einstellung ist ein Modul Warmwasser erforderlich&bull;Die Verbindungsleitung EMS zum Modul Warmwasser auf Besch&auml;digung pr&uuml;fen; Die Busspannung am Modul Warmwasser muss zwischen 12 und 15 V DC liegen&bull;Modul Warmwasser defekt",                                                                                                                                                                       correction => "Konfiguration &auml;ndern&bull;Besch&auml;digte Kabel austauschen&bull;Modul Warmwasser austauschen"                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"0743"} = {errorcode => "A11",                                    subcode => "4013",                                               class => "8",           reason => "Keine Kommunikation mit Frischwasser-Modul.",                                                                                                                                                                                                                                                                                                      checklist => "Falsche Warmwasserkonfiguration; &Uuml;berpr&uuml;fen, ob ein Frischwasser-Modul vorhanden ist&bull;-",                                                                                                                                                                                                                                                                                                                                                                                               correction => "Bei vorhandenem Frischwasser-Modul die Konfiguration im Men&uuml; Warmwasser auf FrischwasserModul &auml;ndern&bull;Bei fehlendem Frischwasser-Modul die Konfiguration im Men&uuml; Warmwasser &auml;ndern"                                                                                                                                                                                      };
	$ErrorCodeList{"0744"} = {errorcode => "A11",                                    subcode => "4021; 4022",                                         class => "16",          reason => "Konfigurationsfehler: Modul Warmwasser nicht verwendet; (4021 =; Warmwassersystem 1;; 4022 =; Warmwassersystem 2)",                                                                                                                                                                                                                                checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); Es ist ein Modul Warmwasser im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird",                                                                                                                                                                                                                                                                                                                                          correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0745"} = {errorcode => "A11",                                    subcode => "4023",                                               class => "8",           reason => "Konfigurationsfehler: Frischwasser-Modul; nicht verwendet",                                                                                                                                                                                                                                                                                        checklist => "Falsche Warmwasserkonfiguration",                                                                                                                                                                                                                                                                                                                                                                                                                                                                     correction => "Konfiguration im Men&uuml; Warmwasser auf Frischwasser-Modul &auml;ndern"                                                                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"0746"} = {errorcode => "A11",                                    subcode => "4033",                                               class => "8",           reason => "Fehler in Konfiguration Frischwassersystem",                                                                                                                                                                                                                                                                                                       checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Konfiguration Frischwassersystem pr&uuml;fen (Drehcodierschalter und Parametrierung Bedieneinheit)"                                                                                                                                                                                                                                                                                              };
	$ErrorCodeList{"0747"} = {errorcode => "A11",                                    subcode => "6001",                                               class => "16",          reason => "Konfigurationsfehler: Solarmodul nicht verwendet",                                                                                                                                                                                                                                                                                                 checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); Es ist ein Solarmodul im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird",                                                                                                                                                                                                                                                                                                                                                correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0748"} = {errorcode => "A11",                                    subcode => "6002",                                               class => "8",           reason => "Solarmodul inkompatibel",                                                                                                                                                                                                                                                                                                                          checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0749"} = {errorcode => "A11",                                    subcode => "6004",                                               class => "14",          reason => "Keine Kommunikation Solarmodul",                                                                                                                                                                                                                                                                                                                   checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); mit der gew&auml;hlten Einstellung ist ein Solarmodul erforderlich&bull;Die Verbindungsleitung EMS zum Solarmodul auf Besch&auml;digung pr&uuml;fen; die Busspannung am Solarmodul muss zwischen 12 und 15 V DC liegen&bull;Solarmodul defekt",                                                                                                                                                                                               correction => "Konfiguration &auml;ndern&bull;Besch&auml;digte Kabel austauschen&bull;Modul austauschen"                                                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"0750"} = {errorcode => "A12",                                    subcode => "815",                                                class => "16",          reason => "Temperaturf&uuml;hler hydraulische Weiche defekt",                                                                                                                                                                                                                                                                                                 checklist => "Verbindungsleitung zwischen Weichenmodul und Weichentemperaturf&uuml;hler pr&uuml;fen&bull;Elektrischen Anschluss der Verbindungsleitung am Weichenmodul pr&uuml;fen&bull;Weichentemperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Kollektortemperaturf&uuml;hlers am Weichenmodul laut Tabelle pr&uuml;fen",                                                                                                                                                  correction => "Wenn ein Defekt vorliegt, F&uuml;hler austauschen&bull;Falls Schrauben oder ein Stecker lose sind, das Kontaktproblem beheben&bull;Wenn Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Weichenmodul austauschen"                                                   };
	$ErrorCodeList{"0751"} = {errorcode => "A12",                                    subcode => "816",                                                class => "4",           reason => "Keine Kommunikation mit Weichenmodul (Auswirkung: wenn das WM10 in Ordnung ist, wird die HK1-Pumpe dauerhaft eingeschaltet)",                                                                                                                                                                                                                      checklist => "Pr&uuml;fen, ob die Verbindungsleitung EMS am Weichenmodul und am System angeschlossen/aufgesteckt ist&bull;Die Verbindungsleitung EMS auf Besch&auml;digungen pr&uuml;fen&bull;Polung der Verbindungsleitung zwischen Weichenmodul und EMS pr&uuml;fen&bull;Weichenmodul defekt",                                                                                                                                                                                                                    correction => "St&ouml;rung in der Steckverbindung beheben  (das Modul muss immer eine Verbindung zum; System haben; es k&ouml;nnen im Bedarfsfall auch die Anschlussklemmen der Bedieneinheit im Regelger&auml;t verwendet werden)&bull;Besch&auml;digte Kabel austauschen&bull;St&ouml;rung in der Polung beheben&bull;Weichenmodul austauschen"                                                              };
	$ErrorCodeList{"0752"} = {errorcode => "A12",                                    subcode => "1709",                                               class => "4",           reason => "Etagenstation: Modul erh&auml;lt unzul&auml;ssige Werte f&uuml;r den Temperaturf&uuml;hler in der hydraulischen Weiche",                                                                                                                                                                                                                           checklist => "Konfiguration pr&uuml;fen. Mit der gew&auml;hlten Einstellung wird ein Temperaturf&uuml;hler an der hydraulischen Weiche ben&ouml;tigt&bull;&Uuml;berpr&uuml;fen, ob ein Temperaturf&uuml;hler von der hydraulischen Weiche montiert ist&bull;Anschlussstecker am Temperaturf&uuml;hler f&uuml;r die hydraulische Weiche &uuml;berpr&uuml;fen&bull;Anschlusskabel zum Temperaturf&uuml;hler in der hydraulischen Weiche &uuml;berpr&uuml;fen&bull;-",                                                 correction => "Konfiguration &auml;ndern&bull;Temperaturf&uuml;hler von der hydraulischen Weiche montieren&bull;Anschlussstecker am Temperaturf&uuml;hler in der hydraulischen Weiche aufstecken&bull;Anschlusskabel zum Temperaturf&uuml;hler in der hydraulischen Weiche austauschen&bull;Temperaturf&uuml;hler in der hydraulischen Weiche austauschen"                                                      };
	$ErrorCodeList{"0753"} = {errorcode => "A12",                                    subcode => "1907",                                               class => "4",           reason => "Etagenstation: Modul erh&auml;lt unzul&auml;ssige Werte vom Temperaturf&uuml;hler in der hydraulischen Weiche",                                                                                                                                                                                                                                    checklist => "Konfiguration pr&uuml;fen. Mit der gew&auml;hlten Einstellung ist ein F&uuml;hler an der hydraulischen Weiche angebracht&bull;Verbindungsleitung zwischen CH Modul und Weichenf&uuml;hler pr&uuml;fen&bull;Weichenf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Weichenf&uuml;hlers laut Tabelle pr&uuml;fen",                                                                                                                                                       correction => "Konfiguration &auml;ndern&bull;Verbindung ordnungsgem&auml;&szlig; herstellen&bull;Wenn die Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Modul austauschen"                                                                                                      };
	$ErrorCodeList{"0754"} = {errorcode => "A16",                                    subcode => "815",                                                class => "14",          reason => "Temperaturf&uuml;hler hydraulische Weiche defekt (Pumpeneffizienzmodul) (Auswirkung: Pumpe geht in Betriebsart „Brennerleistung“)",                                                                                                                                                                                                            checklist => "Temperaturf&uuml;hler FK des Pumpeneffizienzmoduls defekt oder nicht korrekt angeschlossen&bull;-",                                                                                                                                                                                                                                                                                                                                                                                                   correction => "F&uuml;hleranschluss pr&uuml;fen&bull;Weichentemperaturf&uuml;hler auf falsche Einbauposition oder auf Bruchstelle pr&uuml;fen"                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0755"} = {errorcode => "A16",                                    subcode => "816",                                                class => "14",          reason => "Keine Kommunikation mit Pumpeneffizienzmodul (Auswirkung: Notbetrieb, Kesselpumpe l&auml;uft mit max. eingestellter Ausgangsspannung)",                                                                                                                                                                                                            checklist => "EMS-BUS defekt oder nicht korrekt angeschlossen&bull;Sicherung Pumpeneffizienzmodul pr&uuml;fen (wenn LED am Pumpeneffizienzmodul aus)",                                                                                                                                                                                                                                                                                                                                                              correction => "Busanschluss korrigieren&bull;Sicherung austauschen"                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"0756"} = {errorcode => "A18",                                    subcode => "825",                                                class => "14",          reason => "Zwei Master-Bedieneinheiten im System",                                                                                                                                                                                                                                                                                                            checklist => "In der Installationsebene die Parametrierung pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "Die Bedieneinheit f&uuml;r den Heizkreis 1 als Master anmelden"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0757"} = {errorcode => "A2",                                     subcode => "27",                                                 class => "",            reason => "Abgasaustritt aus der Brennkammer",                                                                                                                                                                                                                                                                                                                checklist => "Verteilerrohr vertauscht&bull;Zu gro&szlig;e Gasmenge&bull;W&auml;rmeblock abgasseitig zugesetzt&bull;Signal des F&uuml;hler Abgas&uuml;berwachung liegt au&szlig;erhalb der Kennlinie&bull;Signalauswertung auf Leiterplatte defekt",                                                                                                                                                                                                                                                                correction => "Brennereinstellwerte bei Nennbelastung einstellen&bull;W&auml;rmeblock abgasseitig reinigen&bull;F&uuml;hler (Abgas&uuml;berwachung) an der Brennkammer austauschen&bull;Leiterplatte austauschen&bull;Verteilerrohr tauschen"                                                                                                                                                                   };
	$ErrorCodeList{"0758"} = {errorcode => "A21 A22 A23 A24 A25",                    subcode => "806",                                                class => "A",           reason => "Raumtemperaturf&uuml;hler f&uuml;r Heizkreis defekt; (A21 = Heizkreis 1; A22 = Heizkreis 2; A23 = Heizkreis 3; A24 = Heizkreis 4; A25 = Heizkreis 5)",                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Fernbedienung austauschen"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0759"} = {errorcode => "A21 A22 A23 A24 A25",                    subcode => "816",                                                class => "A",           reason => "Keine Kommunikation mit Bedieneinheit Heizkreis (A21 = Heizkreis 1;; A22 = Heizkreis 2; A23 = Heizkreis 3; A24 = Heizkreis 4; A25 = Heizkreis 5); (Auswirkung: max. Vorlauftemperatur f&uuml;r Raumeinfluss, Schaltuhroptimierung und Raumregler)",                                                                                                checklist => "In der Serviceebene der Bedieneinheit unter Men&uuml;punkt „P1“ pr&uuml;fen, ob Bedieneinheit von der Adresseinstellung dem richtigen Heizkreis zugeordnet wurde; Ursache: falsche Adresse eingestellt, Bedieneinheit falsch verdrahtet&bull;Bei RC20 RF: Batterie &uuml;berpr&uuml;fen (Anzeige „bAt“ im RC20 RF oder Display aus)&bull;Bei RC20 RF: Heizungsanlage ausgeschaltet&bull;Bei RC20 RF: Nach Austausch des RFM-Moduls ist der RC20 RF nicht am neuen RFM eingelernt worden&bull;Bedieneinheit defekt&bull;Ger&auml;teelektronik MC10/MC40/UBA/HM10 defekt",correction => "Entweder die Verdrahtung oder die softwareseitige Zuordnung &auml;ndern&bull;Batterie austauschen&bull;Heizungsanlage einschalten&bull;RC20 RF neu einlernen&bull;Bedieneinheit austauschen&bull;Regelger&auml;t/Ger&auml;teelektronik austauschen"                                                                                                                                              };
	$ErrorCodeList{"0760"} = {errorcode => "A21 A22 A23 A24 A25",                    subcode => "829",                                                class => "A",           reason => "RC2x ohne Heizkreis (A21 = Heizkreis 1; A22 = Heizkreis 2; A23 = Heizkreis 3; A24 = Heizkreis 4; A25 = Heizkreis 5)",                                                                                                                                                                                                                              checklist => "Zuordnung von Fernbedienung Bedieneinheit und Heizkreis pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "Einstellfehler korrigieren"                                                                                                                                                                                                                                                                                                                                                                      };
	$ErrorCodeList{"0761"} = {errorcode => "A21 A22 A23 A24 A25",                    subcode => "830",                                                class => "A",           reason => "Schwache Batterie FunkBedieneinheit Heizkreis (A21 = Heizkreis 1;; A22 = Heizkreis 2; A23 = Heizkreis 3; A24 = Heizkreis 4; A25 = Heizkreis 5)",                                                                                                                                                                                                   checklist => "Batteriespannung pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Batterie austauschen"                                                                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"0762"} = {errorcode => "A21 A22 A23 A24 A25",                    subcode => "839",                                                class => "A",           reason => "Keine Funkkommunikation mit Bedieneinheit Heizkreis (A21 = Heizkreis 1;; A22 = Heizkreis 2; A23 = Heizkreis 3; A24 = Heizkreis 4; A25 = Heizkreis 5)",                                                                                                                                                                                             checklist => "RC20 RF ist au&szlig;erhalb des Empfangbereichs&bull;Heizungsanlage oder Regelger&auml;t ausgeschaltet&bull;Nach dem Austausch von RFM20 oder RC20 RF ist das neue Ger&auml;t nicht wieder eingelernt worden&bull;Bedieneinheit oder RFM20 defekt",                                                                                                                                                                                                                                                   correction => "Bedieneinheit oder RFM20 neu positionieren&bull;Heizungsanlage oder Regelger&auml;t einschalten&bull;Bedieneinheit bzw. RFM20 neu einlernen&bull;Ger&auml;te nacheinander austauschen und auf Funktion pr&uuml;fen"                                                                                                                                                                              };
	$ErrorCodeList{"0763"} = {errorcode => "A21 A22 A23 A24 A25",                    subcode => "842",                                                class => "A",           reason => "Frostschutz gew&auml;hlt,; aber keine Fernbedienung Heizkreis; (A21 = Heizkreis 1; A22 = Heizkreis 2; A23 = Heizkreis 3; A24 = Heizkreis 4; A25 = Heizkreis 5)",                                                                                                                                                                                   checklist => "Zuordnung von Fernbedienung Bedieneinheit und Heizkreis bzw. die Parametrierung pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                         correction => "Einstellung korrigieren"                                                                                                                                                                                                                                                                                                                                                                         };
	$ErrorCodeList{"0764"} = {errorcode => "A21 A22 A23 A24 A25",                    subcode => "843",                                                class => "A",           reason => "Raumregelung gew&auml;hlt, aber keine Fernbedienung f&uuml;r Heizkreis; (A21 = Heizkreis 1; A22 = Heizkreis 2; A23 = Heizkreis 3; A24 = Heizkreis 4; A25 = Heizkreis 5)",                                                                                                                                                                          checklist => "Zuordnung von Fernbedienung Bedieneinheit und Heizkreis bzw. die Parametrierung pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                         correction => "Einstellung korrigieren"                                                                                                                                                                                                                                                                                                                                                                         };
	$ErrorCodeList{"0765"} = {errorcode => "A21 A22 A23 A24 A25 A26 A27 A28",        subcode => "1001",                                               class => "12",          reason => "Keine Kommunikation zwischen Systemregler und Fernbedienung; (A21 = Heizkreis 1; A22 = Heizkreis 2; A23 = Heizkreis 3; A24 = Heizkreis 4; A25 = Heizkreis 5; A26 = Heizkreis 6; A27 = Heizkreis 7; A28 = Heizkreis 8)",                                                                                                                            checklist => "Konfiguration pr&uuml;fen (Adresseinstellung) Mit der gew&auml;hlten Einstellung ist ein Systemregler notwendig&bull;Die Verbindungsleitung EMS zum Systemregler auf Besch&auml;digung pr&uuml;fen; Die Busspannung am Systemregler muss zwischen 12 und 15 V DC liegen&bull;Fernbedienung oder Systemregler defekt",                                                                                                                                                                                 correction => "Konfiguration &auml;ndern&bull;Besch&auml;digte Kabel austauschen&bull;Fernbedienung oder Systemregler austauschen"                                                                                                                                                                                                                                                                              };
	$ErrorCodeList{"0766"} = {errorcode => "A21",                                    subcode => "1007",                                               class => "8",           reason => "Verbindung zur Basisstation unterbrochen : A21 = Heizkreis 1",                                                                                                                                                                                                                                                                                     checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0767"} = {errorcode => "A22",                                    subcode => "1007",                                               class => "8",           reason => "Verbindung zur Basisstation unterbrochen : A22 = Heizkreis 2",                                                                                                                                                                                                                                                                                     checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0768"} = {errorcode => "A23",                                    subcode => "1007",                                               class => "8",           reason => "Verbindung zur Basisstation unterbrochen : A23 = Heizkreis 3",                                                                                                                                                                                                                                                                                     checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0769"} = {errorcode => "A24",                                    subcode => "1007",                                               class => "8",           reason => "Verbindung zur Basisstation unterbrochen : A24 = Heizkreis 4",                                                                                                                                                                                                                                                                                     checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0770"} = {errorcode => "A25",                                    subcode => "1007",                                               class => "8",           reason => "Verbindung zur Basisstation unterbrochen : A25 = Heizkreis 5",                                                                                                                                                                                                                                                                                     checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0771"} = {errorcode => "A26",                                    subcode => "1007",                                               class => "8",           reason => "Verbindung zur Basisstation unterbrochen : A26 = Heizkreis 6",                                                                                                                                                                                                                                                                                     checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0772"} = {errorcode => "A27",                                    subcode => "1007",                                               class => "8",           reason => "Verbindung zur Basisstation unterbrochen : A27 = Heizkreis 7",                                                                                                                                                                                                                                                                                     checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0773"} = {errorcode => "A28",                                    subcode => "1007",                                               class => "8",           reason => "Verbindung zur Basisstation unterbrochen : A28 = Heizkreis 8",                                                                                                                                                                                                                                                                                     checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0774"} = {errorcode => "A21",                                    subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber",                                                                                                                                                                                                                                                                                                                    checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen",                                                                                                                                                                                                                                                                                                                                                                                                                                                     correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t"                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0775"} = {errorcode => "A22",                                    subcode => "1010",                                               class => "",            reason => "BUS-Verbindung EMS plus (A21 = Heizkreis 1;",                                                                                                                                                                                                                                                                                                      checklist => "wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder; einschalten; Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung",                                                                                                                                                                                                                                                                                          correction => "ausund wieder einschalten&bull;Busleitung reparieren bzw. austauschen"                                                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"0776"} = {errorcode => "A23",                                    subcode => "1010",                                               class => "",            reason => "",                                                                                                                                                                                                                                                                                                                                                 checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => ""                                                                                                                                                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"0777"} = {errorcode => "A24",                                    subcode => "1010",                                               class => "",            reason => "A22 = Heizkreis 2;",                                                                                                                                                                                                                                                                                                                               checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0778"} = {errorcode => "A25",                                    subcode => "1010",                                               class => "",            reason => "A23 = Heizkreis 3; A24 = Heizkreis 4;",                                                                                                                                                                                                                                                                                                            checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Defekten EMS-BUS-Teilnehmer austauschen"                                                                                                                                                                                                                                                                                                                                                         };
	$ErrorCodeList{"0779"} = {errorcode => "A26",                                    subcode => "1010",                                               class => "",            reason => "",                                                                                                                                                                                                                                                                                                                                                 checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => ""                                                                                                                                                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"0780"} = {errorcode => "A27",                                    subcode => "1010",                                               class => "",            reason => "A25 = Heizkreis 5;",                                                                                                                                                                                                                                                                                                                               checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0781"} = {errorcode => "A28",                                    subcode => "1010",                                               class => "",            reason => "A26 = Heizkreis 6; A27 = Heizkreis 7;",                                                                                                                                                                                                                                                                                                            checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0782"} = {errorcode => "A21 A22",                                subcode => "1030; 1033",                                         class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen",                                                                                                                                                                                                                                                                                                checklist => "Bedieneinheit austauschen",                                                                                                                                                                                                                                                                                                                                                                                                                                                                           correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0783"} = {errorcode => "A23",                                    subcode => "1034",                                               class => "",            reason => "(A21 = Heizkreis 1;",                                                                                                                                                                                                                                                                                                                              checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0784"} = {errorcode => "A24",                                    subcode => "1035",                                               class => "",            reason => "A22 = Heizkreis 2;",                                                                                                                                                                                                                                                                                                                               checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0785"} = {errorcode => "A25 A26",                                subcode => "1036",                                               class => "",            reason => "A23 = Heizkreis 3; A24 = Heizkreis 4;",                                                                                                                                                                                                                                                                                                            checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0786"} = {errorcode => "A27",                                    subcode => "",                                                   class => "",            reason => "A25 = Heizkreis 5;",                                                                                                                                                                                                                                                                                                                               checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0787"} = {errorcode => "A28",                                    subcode => "",                                                   class => "",            reason => "A26 = Heizkreis 6; A27 = Heizkreis 7;",                                                                                                                                                                                                                                                                                                            checklist => "&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0788"} = {errorcode => "A21 A22",                                subcode => "1031",                                               class => "8",           reason => "Verbindung mit Basisstation fehlgeschlagen.",                                                                                                                                                                                                                                                                                                      checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0789"} = {errorcode => "A23",                                    subcode => "",                                                   class => "",            reason => "(A21 = Heizkreis 1;",                                                                                                                                                                                                                                                                                                                              checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0790"} = {errorcode => "A24",                                    subcode => "",                                                   class => "",            reason => "A22 = Heizkreis 2;",                                                                                                                                                                                                                                                                                                                               checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0791"} = {errorcode => "A25",                                    subcode => "",                                                   class => "",            reason => "A23 = Heizkreis 3;",                                                                                                                                                                                                                                                                                                                               checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0792"} = {errorcode => "A26",                                    subcode => "",                                                   class => "",            reason => "A24 = Heizkreis 4;",                                                                                                                                                                                                                                                                                                                               checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0793"} = {errorcode => "A27",                                    subcode => "",                                                   class => "",            reason => "A25 = Heizkreis 5;",                                                                                                                                                                                                                                                                                                                               checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0794"} = {errorcode => "A28",                                    subcode => "",                                                   class => "",            reason => "A26 = Heizkreis 6; A27 = Heizkreis 7;",                                                                                                                                                                                                                                                                                                            checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0795"} = {errorcode => "A21 A22 A23 A24",                        subcode => "1038",                                               class => "16",          reason => "Datum/Zeit ung&uuml;ltiger Wert (A21 = Heizkreis 1;; A22 = Heizkreis 2; A23 = Heizkreis 3;",                                                                                                                                                                                                                                                       checklist => "Datum/Zeit noch nicht eingestellt&bull;Spannungsversorgung &uuml;ber l&auml;ngere Zeit ausgefallen",                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Datum/Zeit einstellen&bull;Spannungsausf&auml;lle vermeiden"                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0796"} = {errorcode => "A25",                                    subcode => "",                                                   class => "",            reason => "A24 = Heizkreis 4;",                                                                                                                                                                                                                                                                                                                               checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0797"} = {errorcode => "A26",                                    subcode => "",                                                   class => "",            reason => "A25 = Heizkreis 5;",                                                                                                                                                                                                                                                                                                                               checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0798"} = {errorcode => "A27",                                    subcode => "",                                                   class => "",            reason => "A26 = Heizkreis 6;",                                                                                                                                                                                                                                                                                                                               checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0799"} = {errorcode => "A28",                                    subcode => "",                                                   class => "",            reason => "A27 = Heizkreis 7; A28 = Heizkreis 8)",                                                                                                                                                                                                                                                                                                            checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0800"} = {errorcode => "A21 A22",                                subcode => "1042",                                               class => "14",          reason => "Interner Fehler:; Zugriff auf Uhrenbaustein",                                                                                                                                                                                                                                                                                                      checklist => "Interner Fehler",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                     correction => "Bedieneinheit austauschen"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0801"} = {errorcode => "A23 A24",                                subcode => "",                                                   class => "",            reason => "blockiert",                                                                                                                                                                                                                                                                                                                                        checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0802"} = {errorcode => "A25",                                    subcode => "",                                                   class => "",            reason => "",                                                                                                                                                                                                                                                                                                                                                 checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0803"} = {errorcode => "A26",                                    subcode => "",                                                   class => "",            reason => "",                                                                                                                                                                                                                                                                                                                                                 checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0804"} = {errorcode => "A27",                                    subcode => "",                                                   class => "",            reason => "",                                                                                                                                                                                                                                                                                                                                                 checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0805"} = {errorcode => "A28",                                    subcode => "",                                                   class => "",            reason => "",                                                                                                                                                                                                                                                                                                                                                 checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0806"} = {errorcode => "A21 A22 A23",                            subcode => "3141; 3142; 3143",                                   class => "14",          reason => "Feuchtigkeitsf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                                  checklist => "Feuchtigkeitsf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                                                                                                                                                                                     correction => "Bedieneinheit mit integriertem Feuchtigkeitsf&uuml;hler austauschen"                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"0807"} = {errorcode => "A24",                                    subcode => "3144",                                               class => "",            reason => "",                                                                                                                                                                                                                                                                                                                                                 checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0808"} = {errorcode => "A25",                                    subcode => "3145",                                               class => "",            reason => "",                                                                                                                                                                                                                                                                                                                                                 checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0809"} = {errorcode => "A26",                                    subcode => "3146",                                               class => "",            reason => "",                                                                                                                                                                                                                                                                                                                                                 checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0810"} = {errorcode => "A27",                                    subcode => "3147",                                               class => "",            reason => "",                                                                                                                                                                                                                                                                                                                                                 checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0811"} = {errorcode => "A28",                                    subcode => "3148",                                               class => "",            reason => "",                                                                                                                                                                                                                                                                                                                                                 checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0812"} = {errorcode => "A21",                                    subcode => "3161; 3162; 3163; 3164; 3165; 3166; 3167; 3168",     class => "8",           reason => "Batterie schwach : A21 = Heizkreis 1",                                                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0813"} = {errorcode => "A22",                                    subcode => "3161; 3162; 3163; 3164; 3165; 3166; 3167; 3168",     class => "8",           reason => "Batterie schwach : A22 = Heizkreis 2",                                                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0814"} = {errorcode => "A23",                                    subcode => "3161; 3162; 3163; 3164; 3165; 3166; 3167; 3168",     class => "8",           reason => "Batterie schwach : A23 = Heizkreis 3",                                                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0815"} = {errorcode => "A24",                                    subcode => "3161; 3162; 3163; 3164; 3165; 3166; 3167; 3168",     class => "8",           reason => "Batterie schwach : A24 = Heizkreis 4",                                                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0816"} = {errorcode => "A25",                                    subcode => "3161; 3162; 3163; 3164; 3165; 3166; 3167; 3168",     class => "8",           reason => "Batterie schwach : A25 = Heizkreis 5",                                                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0817"} = {errorcode => "A26",                                    subcode => "3161; 3162; 3163; 3164; 3165; 3166; 3167; 3168",     class => "8",           reason => "Batterie schwach : A26 = Heizkreis 6",                                                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0818"} = {errorcode => "A27",                                    subcode => "3161; 3162; 3163; 3164; 3165; 3166; 3167; 3168",     class => "8",           reason => "Batterie schwach : A27 = Heizkreis 7",                                                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0819"} = {errorcode => "A28",                                    subcode => "3161; 3162; 3163; 3164; 3165; 3166; 3167; 3168",     class => "8",           reason => "Batterie schwach : A28 = Heizkreis 8",                                                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0820"} = {errorcode => "A3",                                     subcode => "17",                                                 class => "",            reason => "Abgastemperaturf&uuml;hler an der Str&ouml;mungssicherung defekt",                                                                                                                                                                                                                                                                                 checklist => "Kein Signal von Abgas&uuml;berwachung an der Str&ouml;mungssicherung&bull;Signal des F&uuml;hlers f&uuml;r die Abgas&uuml;berwachung liegt au&szlig;erhalb der Kennlinie&bull;Signalunterbrechung im Anschlusskabel&bull;Signalauswertung auf Leiterplatte defekt",                                                                                                                                                                                                                                   correction => "F&uuml;hler (Abgas&uuml;berwachung) austauschen&bull;Kabelbaum auf Besch&auml;digungen &uuml;berpr&uuml;fen&bull;Kabelbaum austauschen&bull;Leiterplatte austauschen; Stecker an der Abgas&uuml;berwachung (Str&ouml;mungssicherung) aufstecken"                                                                                                                                                 };
	$ErrorCodeList{"0821"} = {errorcode => "A31",                                    subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus : A31 = Heizkreis 1",                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder; einschalten; Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung",                                                                                                                                                                                                                                          correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t aus und wieder einschalten&bull;Busleitung reparieren bzw. austauschen"                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"0822"} = {errorcode => "A32",                                    subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus : A32 = Heizkreis 2",                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder; einschalten; Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung",                                                                                                                                                                                                                                          correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t aus und wieder einschalten&bull;Busleitung reparieren bzw. austauschen"                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"0823"} = {errorcode => "A33",                                    subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus : A33 = Heizkreis 3",                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder; einschalten; Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung",                                                                                                                                                                                                                                          correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t aus und wieder einschalten&bull;Busleitung reparieren bzw. austauschen"                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"0824"} = {errorcode => "A34",                                    subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus : A34 = Heizkreis 4",                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder; einschalten; Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung",                                                                                                                                                                                                                                          correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t aus und wieder einschalten&bull;Busleitung reparieren bzw. austauschen"                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"0825"} = {errorcode => "A35",                                    subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus : A35 = Heizkreis 5",                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder; einschalten; Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung",                                                                                                                                                                                                                                          correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t aus und wieder einschalten&bull;Busleitung reparieren bzw. austauschen"                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"0826"} = {errorcode => "A36",                                    subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus : A36 = Heizkreis 6",                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder; einschalten; Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung",                                                                                                                                                                                                                                          correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t aus und wieder einschalten&bull;Busleitung reparieren bzw. austauschen"                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"0827"} = {errorcode => "A37",                                    subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus : A37 = Heizkreis 7",                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder; einschalten; Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung",                                                                                                                                                                                                                                          correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t aus und wieder einschalten&bull;Busleitung reparieren bzw. austauschen"                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"0828"} = {errorcode => "A38",                                    subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus : A38 = Heizkreis 8",                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder; einschalten; Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung",                                                                                                                                                                                                                                          correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t aus und wieder einschalten&bull;Busleitung reparieren bzw. austauschen"                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"0829"} = {errorcode => "A31 A32",                                subcode => "1030; 1035",                                         class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen",                                                                                                                                                                                                                                                                                                checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0830"} = {errorcode => "A33",                                    subcode => "1036",                                               class => "",            reason => "(A31 = Heizkreis 1;",                                                                                                                                                                                                                                                                                                                              checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0831"} = {errorcode => "A34 A35",                                subcode => "1038",                                               class => "",            reason => "A32 = Heizkreis 2; A33 = Heizkreis 3;",                                                                                                                                                                                                                                                                                                            checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0832"} = {errorcode => "A36",                                    subcode => "",                                                   class => "",            reason => "A34 = Heizkreis 4;",                                                                                                                                                                                                                                                                                                                               checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0833"} = {errorcode => "A37",                                    subcode => "",                                                   class => "",            reason => "A35 = Heizkreis 5;",                                                                                                                                                                                                                                                                                                                               checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0834"} = {errorcode => "A38",                                    subcode => "",                                                   class => "",            reason => "A36 = Heizkreis 6; A37 = Heizkreis 7;",                                                                                                                                                                                                                                                                                                            checklist => "&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0835"} = {errorcode => "A31",                                    subcode => "3021; 3022; 3023; 3024; 3025; 3026; 3027; 3028",     class => "14",          reason => "Heizkreis Vorlauftemperaturf&uuml;hler defekt : A31 = Heizkreis 1",                                                                                                                                                                                                                                                                                checklist => "Konfiguration pr&uuml;fen; Mit der gew&auml;hlten Einstellung ist ein Vorlauftemperaturf&uuml;hler erforderlich&bull;Verbindungsleitung zwischen Modul Mischer und Vorlauftemperaturf&uuml;hler pr&uuml;fen&bull;Vorlauftemperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Vorlauftemperaturf&uuml;hlers am Modul Mischer laut&bull;Tabelle pr&uuml;fen",                                                                                                       correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0836"} = {errorcode => "A32",                                    subcode => "3021; 3022; 3023; 3024; 3025; 3026; 3027; 3028",     class => "14",          reason => "Heizkreis Vorlauftemperaturf&uuml;hler defekt : A32 = Heizkreis 2",                                                                                                                                                                                                                                                                                checklist => "Konfiguration pr&uuml;fen; Mit der gew&auml;hlten Einstellung ist ein Vorlauftemperaturf&uuml;hler erforderlich&bull;Verbindungsleitung zwischen Modul Mischer und Vorlauftemperaturf&uuml;hler pr&uuml;fen&bull;Vorlauftemperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Vorlauftemperaturf&uuml;hlers am Modul Mischer laut&bull;Tabelle pr&uuml;fen",                                                                                                       correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0837"} = {errorcode => "A33",                                    subcode => "3021; 3022; 3023; 3024; 3025; 3026; 3027; 3028",     class => "14",          reason => "Heizkreis Vorlauftemperaturf&uuml;hler defekt : A33 = Heizkreis 3",                                                                                                                                                                                                                                                                                checklist => "Konfiguration pr&uuml;fen; Mit der gew&auml;hlten Einstellung ist ein Vorlauftemperaturf&uuml;hler erforderlich&bull;Verbindungsleitung zwischen Modul Mischer und Vorlauftemperaturf&uuml;hler pr&uuml;fen&bull;Vorlauftemperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Vorlauftemperaturf&uuml;hlers am Modul Mischer laut&bull;Tabelle pr&uuml;fen",                                                                                                       correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0838"} = {errorcode => "A34",                                    subcode => "3021; 3022; 3023; 3024; 3025; 3026; 3027; 3028",     class => "14",          reason => "Heizkreis Vorlauftemperaturf&uuml;hler defekt : A34 = Heizkreis 4",                                                                                                                                                                                                                                                                                checklist => "Konfiguration pr&uuml;fen; Mit der gew&auml;hlten Einstellung ist ein Vorlauftemperaturf&uuml;hler erforderlich&bull;Verbindungsleitung zwischen Modul Mischer und Vorlauftemperaturf&uuml;hler pr&uuml;fen&bull;Vorlauftemperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Vorlauftemperaturf&uuml;hlers am Modul Mischer laut&bull;Tabelle pr&uuml;fen",                                                                                                       correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0839"} = {errorcode => "A35",                                    subcode => "3021; 3022; 3023; 3024; 3025; 3026; 3027; 3028",     class => "14",          reason => "Heizkreis Vorlauftemperaturf&uuml;hler defekt : A35 = Heizkreis 5",                                                                                                                                                                                                                                                                                checklist => "Konfiguration pr&uuml;fen; Mit der gew&auml;hlten Einstellung ist ein Vorlauftemperaturf&uuml;hler erforderlich&bull;Verbindungsleitung zwischen Modul Mischer und Vorlauftemperaturf&uuml;hler pr&uuml;fen&bull;Vorlauftemperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Vorlauftemperaturf&uuml;hlers am Modul Mischer laut&bull;Tabelle pr&uuml;fen",                                                                                                       correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0840"} = {errorcode => "A36",                                    subcode => "3021; 3022; 3023; 3024; 3025; 3026; 3027; 3028",     class => "14",          reason => "Heizkreis Vorlauftemperaturf&uuml;hler defekt : A36 = Heizkreis 6",                                                                                                                                                                                                                                                                                checklist => "Konfiguration pr&uuml;fen; Mit der gew&auml;hlten Einstellung ist ein Vorlauftemperaturf&uuml;hler erforderlich&bull;Verbindungsleitung zwischen Modul Mischer und Vorlauftemperaturf&uuml;hler pr&uuml;fen&bull;Vorlauftemperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Vorlauftemperaturf&uuml;hlers am Modul Mischer laut&bull;Tabelle pr&uuml;fen",                                                                                                       correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0841"} = {errorcode => "A37",                                    subcode => "3021; 3022; 3023; 3024; 3025; 3026; 3027; 3028",     class => "14",          reason => "Heizkreis Vorlauftemperaturf&uuml;hler defekt : A37 = Heizkreis 7",                                                                                                                                                                                                                                                                                checklist => "Konfiguration pr&uuml;fen; Mit der gew&auml;hlten Einstellung ist ein Vorlauftemperaturf&uuml;hler erforderlich&bull;Verbindungsleitung zwischen Modul Mischer und Vorlauftemperaturf&uuml;hler pr&uuml;fen&bull;Vorlauftemperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Vorlauftemperaturf&uuml;hlers am Modul Mischer laut&bull;Tabelle pr&uuml;fen",                                                                                                       correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0842"} = {errorcode => "A38",                                    subcode => "3021; 3022; 3023; 3024; 3025; 3026; 3027; 3028",     class => "14",          reason => "Heizkreis Vorlauftemperaturf&uuml;hler defekt : A38 = Heizkreis 8",                                                                                                                                                                                                                                                                                checklist => "Konfiguration pr&uuml;fen; Mit der gew&auml;hlten Einstellung ist ein Vorlauftemperaturf&uuml;hler erforderlich&bull;Verbindungsleitung zwischen Modul Mischer und Vorlauftemperaturf&uuml;hler pr&uuml;fen&bull;Vorlauftemperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Vorlauftemperaturf&uuml;hlers am Modul Mischer laut&bull;Tabelle pr&uuml;fen",                                                                                                       correction => "-&bull;Verbindung ordnungsgem&auml;&szlig; herstellen&bull;Wenn die Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, die Spannungswerte jedoch nicht &uuml;bereinstimmen,&bull;dann das Mischermodul austauschen"                                                                                                                };
	$ErrorCodeList{"0843"} = {errorcode => "A31",                                    subcode => "3051; 3052; 3053; 3054; 3055; 3056; 3057; 3058",     class => "16",          reason => "Konfigurationsfehler: Vorlauftemperaturf&uuml;hler nicht verwendet : A31 = Heizkreis 1",                                                                                                                                                                                                                                                           checklist => "Konfigurationsfehler: Vorlauftemperaturf&uuml;hler nicht verwendetKonfiguration pr&uuml;fen; Es ist ein Vorlauftemperaturf&uuml;hler im System, der mit der gew&auml;hlten Einstellung nicht verwendet wird.",                                                                                                                                                                                                                                                                                        correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0844"} = {errorcode => "A32",                                    subcode => "3051; 3052; 3053; 3054; 3055; 3056; 3057; 3058",     class => "16",          reason => "Konfigurationsfehler: Vorlauftemperaturf&uuml;hler nicht verwendet : A32 = Heizkreis 2",                                                                                                                                                                                                                                                           checklist => "Konfigurationsfehler: Vorlauftemperaturf&uuml;hler nicht verwendetKonfiguration pr&uuml;fen; Es ist ein Vorlauftemperaturf&uuml;hler im System, der mit der gew&auml;hlten Einstellung nicht verwendet wird.",                                                                                                                                                                                                                                                                                        correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0845"} = {errorcode => "A33",                                    subcode => "3051; 3052; 3053; 3054; 3055; 3056; 3057; 3058",     class => "16",          reason => "Konfigurationsfehler: Vorlauftemperaturf&uuml;hler nicht verwendet : A33 = Heizkreis 3",                                                                                                                                                                                                                                                           checklist => "Konfigurationsfehler: Vorlauftemperaturf&uuml;hler nicht verwendetKonfiguration pr&uuml;fen; Es ist ein Vorlauftemperaturf&uuml;hler im System, der mit der gew&auml;hlten Einstellung nicht verwendet wird.",                                                                                                                                                                                                                                                                                        correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0846"} = {errorcode => "A34",                                    subcode => "3051; 3052; 3053; 3054; 3055; 3056; 3057; 3058",     class => "16",          reason => "Konfigurationsfehler: Vorlauftemperaturf&uuml;hler nicht verwendet : A34 = Heizkreis 4",                                                                                                                                                                                                                                                           checklist => "Konfigurationsfehler: Vorlauftemperaturf&uuml;hler nicht verwendetKonfiguration pr&uuml;fen; Es ist ein Vorlauftemperaturf&uuml;hler im System, der mit der gew&auml;hlten Einstellung nicht verwendet wird.",                                                                                                                                                                                                                                                                                        correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0847"} = {errorcode => "A35",                                    subcode => "3051; 3052; 3053; 3054; 3055; 3056; 3057; 3058",     class => "16",          reason => "Konfigurationsfehler: Vorlauftemperaturf&uuml;hler nicht verwendet : A35 = Heizkreis 5",                                                                                                                                                                                                                                                           checklist => "Konfigurationsfehler: Vorlauftemperaturf&uuml;hler nicht verwendetKonfiguration pr&uuml;fen; Es ist ein Vorlauftemperaturf&uuml;hler im System, der mit der gew&auml;hlten Einstellung nicht verwendet wird.",                                                                                                                                                                                                                                                                                        correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0848"} = {errorcode => "A36",                                    subcode => "3051; 3052; 3053; 3054; 3055; 3056; 3057; 3058",     class => "16",          reason => "Konfigurationsfehler: Vorlauftemperaturf&uuml;hler nicht verwendet : A36 = Heizkreis 6",                                                                                                                                                                                                                                                           checklist => "Konfigurationsfehler: Vorlauftemperaturf&uuml;hler nicht verwendetKonfiguration pr&uuml;fen; Es ist ein Vorlauftemperaturf&uuml;hler im System, der mit der gew&auml;hlten Einstellung nicht verwendet wird.",                                                                                                                                                                                                                                                                                        correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0849"} = {errorcode => "A37",                                    subcode => "3051; 3052; 3053; 3054; 3055; 3056; 3057; 3058",     class => "16",          reason => "Konfigurationsfehler: Vorlauftemperaturf&uuml;hler nicht verwendet : A37 = Heizkreis 7",                                                                                                                                                                                                                                                           checklist => "Konfigurationsfehler: Vorlauftemperaturf&uuml;hler nicht verwendetKonfiguration pr&uuml;fen; Es ist ein Vorlauftemperaturf&uuml;hler im System, der mit der gew&auml;hlten Einstellung nicht verwendet wird.",                                                                                                                                                                                                                                                                                        correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0850"} = {errorcode => "A38",                                    subcode => "3051; 3052; 3053; 3054; 3055; 3056; 3057; 3058",     class => "16",          reason => "Konfigurationsfehler: Vorlauftemperaturf&uuml;hler nicht verwendet : A38 = Heizkreis 8",                                                                                                                                                                                                                                                           checklist => "Konfigurationsfehler: Vorlauftemperaturf&uuml;hler nicht verwendetKonfiguration pr&uuml;fen; Es ist ein Vorlauftemperaturf&uuml;hler im System, der mit der gew&auml;hlten Einstellung nicht verwendet wird.",                                                                                                                                                                                                                                                                                        correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0851"} = {errorcode => "A31 A32 A33 A34 A35 A36 A37 A38",        subcode => "3101; 3102; 3103; 3104; 3105; 3106; 3107; 3108",     class => "16",          reason => "Keine Kommunikation zwischen Bedieneinheit; und Mischermodul (A31/3101 = Heizkreis 1; A32/3102 = Heizkreis 2; A33/3103 = Heizkreis 3; A34/3104 = Heizkreis 4; A35/3105 = Heizkreis 5; A36/3106 = Heizkreis 6; A37/3107 = Heizkreis 7; A38/3108 = Heizkreis 8)",                                                                                    checklist => "Pr&uuml;fen, ob die Adresseinstellung an Mischermodul und Bedieneinheit richtig gew&auml;hlt wurde&bull;Die EMS-Verbindungsleitung auf Besch&auml;digungen pr&uuml;fen&bull;Bedieneinheit defekt",                                                                                                                                                                                                                                                                                                    correction => "F&uuml;r Heizkreis 2 z. B. auf 2 stellen&bull;Besch&auml;digte Kabel austauschen&bull;Bedieneinheit austauschen"                                                                                                                                                                                                                                                                                 };
	$ErrorCodeList{"0852"} = {errorcode => "A31 A32 A33 A34 A35 A36 A37 A38",        subcode => "3121; 3122; 3123; 3124; 3125; 3126; 3127; 3128",     class => "10",          reason => "Temperaturbegrenzung ausgel&ouml;st; (A31/3121 = Heizkreis 1; A32/3122 = Heizkreis 2; A33/3123 = Heizkreis 3; A34/3124 = Heizkreis 4; A35/3125 = Heizkreis 5; A36/3126 = Heizkreis 6; A37/3127 = Heizkreis 7; A38/3128 = Heizkreis 8)",                                                                                                            checklist => "Anschlussklemme MC15 und MC16 nicht best&uuml;ckt mit Br&uuml;cke oder Temperaturbegrenzer&bull;Einstellung des Temperaturbegrenzers pr&uuml;fen&bull;Mischer nicht korrekt angeschlossen&bull;Mischer defekt",                                                                                                                                                                                                                                                                                       correction => "Br&uuml;cke oder Temperaturbegrenzer anschlie&szlig;en an Anschlussklemme MC15 und MC16&bull;Einstellung des Temperaturbegrenzers korrigieren&bull;Mischer korrekt anschlie&szlig;en&bull;Defekten Mischer austauschen"                                                                                                                                                                          };
	$ErrorCodeList{"0853"} = {errorcode => "A32 A33 A34 A35",                        subcode => "807",                                                class => "A",           reason => "Vorlauftemperaturf&uuml;hler Heizkreis defekt; (A32 = Heizkreis 1; A33 = Heizkreis 2; A34 = Heizkreis 3; A35 = Heizkreis 4); (Auswirkung: das Stellglied MM10 wird sofort stromlos geschaltet, bei MM50 bzw. MM100 wird vorher das Stellglied in Mittelstellung gefahren)",                                                                        checklist => "Verbindungsleitung zwischen Regelger&auml;t und Vorlauftemperaturf&uuml;hler pr&uuml;fen&bull;Elektrischen Anschluss der Verbindungsleitung im Regelger&auml;t pr&uuml;fen&bull;Vorlauftemperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Vorlauftemperaturf&uuml;hlers im Regelger&auml;t laut Tabelle pr&uuml;fen",                                                                                                                                           correction => "Wenn ein Defekt vorliegt, F&uuml;hler austauschen&bull;Falls Schrauben oder ein Stecker lose sind, das Kontaktproblem beheben&bull;Wenn Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Regelger&auml;t (Kaskadenmodul) austauschen"                                };
	$ErrorCodeList{"0854"} = {errorcode => "A32 A33 A34 A35",                        subcode => "816",                                                class => "A",           reason => "Keine Kommunikation mit Heizkreis-Mischermodul (A32 = Heizkreis 1;; A33 = Heizkreis 2; A34 = Heizkreis 3; A35 = Heizkreis 4)",                                                                                                                                                                                                                     checklist => "Pr&uuml;fen, ob die Adresseinstellung am Mischermodul richtig gew&auml;hlt wurde&bull;Die Verbindungsleitung EMS auf Besch&auml;digungen pr&uuml;fen&bull;Polung der Verbindungsleitung zwischen Mischermodul und EMS pr&uuml;fen&bull;Mischermodul defekt",                                                                                                                                                                                                                                          correction => "F&uuml;r Heizkreis 2 z. B. auf 2 stellen&bull;Besch&auml;digte Kabel austauschen&bull;St&ouml;rung in der Polung beheben&bull;Mischermodul austauschen"                                                                                                                                                                                                                                          };
	$ErrorCodeList{"0855"} = {errorcode => "A31/3131 A32/3132 A33/3133 A34/3134",    subcode => "14",                                                 class => "",            reason => "Feuchtigkeit auf Zuleitungsrohr detektiert (Feuchtew&auml;chter hat ausgel&ouml;st, Br&uuml;cke MD1 des Mischermoduls wurde geschlossen)",                                                                                                                                                                                                         checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0856"} = {errorcode => "A4",                                     subcode => "24; 57",                                             class => "",            reason => "Abgasaustritt aus Str&ouml;mungssicherung",                                                                                                                                                                                                                                                                                                        checklist => "Thermische Abgasklappe ist blockiert bzw. verschmutzt; Falscher Typ der thermischen Abgasklappe eingebaut; Minimaler Schornsteinzug wird unterschritten, Verbrennungsluftmenge zu gering, Abgasansaugung durch mechanische Entl&uuml;ftungseinrichtungen (z. B. Dunstabzugshaube, Abluftventilator)&bull;",                                                                                                                                                                                           correction => "Thermische Abgasklappe reinigen oder austauschen&bull;Thermische Abgasklappe austauschen"                                                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"0857"} = {errorcode => "A41 A42",                                subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus (A41 = Warmwassersystem 1;; A42 = Warmwassersystem 2)",                                                                                                                                                                                                                                      checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist; Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder einschalten; pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung&bull;",                                                                                                                                                                                                                                    correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t ausund wieder einschalten&bull;Busleitung reparieren bzw. austauschen&bull;Defekten EMS-BUS-Teilnehmer austauschen"                                                                                                                                                                                                                            };
	$ErrorCodeList{"0858"} = {errorcode => "A41 A42",                                subcode => "1030; 1036",                                         class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen; (A41 = Warmwassersystem 1;; A42 =Warmwassersystem 2)",                                                                                                                                                                                                                                          checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0859"} = {errorcode => "A41 A42",                                subcode => "1038",                                               class => "16",          reason => "Zeit/Datum ung&uuml;ltiger Wert; (A41 = Warmwassersystem 1;; A42 = Warmwassersystem 2)",                                                                                                                                                                                                                                                           checklist => "Datum/Zeit noch nicht eingestellt&bull;Spannungsversorgung &uuml;ber l&auml;ngere Zeit ausgefallen",                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Datum/Zeit einstellen&bull;Spannungsausf&auml;lle vermeiden"                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0860"} = {errorcode => "A41 A42",                                subcode => "4031; 4032",                                         class => "14",          reason => "Warmwasser-Temperaturf&uuml;hler defekt; (A41 = Warmwassersystem 1;; A42 = Warmwassersystem 2)",                                                                                                                                                                                                                                                   checklist => "Konfiguration pr&uuml;fen; Mit der gew&auml;hlten Einstellung ist ein Warmwasser-Temperaturf&uuml;hler erforderlich&bull;Verbindungsleitung zwischen Regelger&auml;t und Warmwasser-Temperaturf&uuml;hler pr&uuml;fen&bull;Elektrischen Anschluss der Verbindungsleitung im Regelger&auml;t pr&uuml;fen&bull;Warmwasser-Temperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Warmwasser-Temperaturf&uuml;hlers im Regelger&auml;t laut Tabelle pr&uuml;fen&bull;Beim GB142 sollte ein Reset durchgef&uuml;hrt werden",correction => "Konfiguration &auml;ndern&bull;Verbindung ordnungsgema&szlig; herstellen&bull;Falls Schrauben oder ein Stecker lose sind, das Kontaktproblem beheben&bull;Wenn Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Regelger&auml;t (Kaskadenmodul) austauschen&bull;Sollte nach dem Reset die St&ouml;rung noch bestehen, dann ist die Ger&auml;teelektronik UBA im W&auml;rmeerzeuger auszutauschen"};
	$ErrorCodeList{"0861"} = {errorcode => "A41",                                    subcode => "4041",                                               class => "8",           reason => "Warmwasser: Solltemperatur nicht erreicht",                                                                                                                                                                                                                                                                                                        checklist => "siehe A41/4051",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      correction => "siehe A41/4051"                                                                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0862"} = {errorcode => "A41 A42",                                subcode => "4051; 4052",                                         class => "14",          reason => "Thermische Desinfektion misslungen; (A41/4051 = Warmwassersystem 1;; A42/4052 = Warmwassersystem 2)",                                                                                                                                                                                                                                              checklist => "Pr&uuml;fen, ob evtl. st&auml;ndig Wasser aufgrund von Zapfungen oder einer Leckage aus dem Warmwasserspeicher entnommen wird&bull;Die Position des Warmwasser-Temperaturf&uuml;hlers pr&uuml;fen, evtl. ist dieser falsch angebracht&bull;Die Verbindungsrohre zwischen Kessel und Speicher kontrollieren und nach Montageanleitung pr&uuml;fen, ob diese richtig angeschlossen sind&bull;Pr&uuml;fen, ob die Heizschlange im Speicher vollst&auml;ndig entl&uuml;ftet ist&bull;Entsprechend den technischen Unterlagen pr&uuml;fen, ob die eingebaute Speicherladepumpe &uuml;ber die erforderliche Leistung verf&uuml;gt&bull;Wenn der Warmwasser-Vorrang abgew&auml;hlt wurde und Heizung und Warmwasser im Parallelbetrieb laufen, kann evtl. die Leistung des Kessels nicht ausreichen&bull;Zu gro&szlig;e Verluste Zirkulationsleitung&bull;Den Warmwasser-Temperaturf&uuml;hler laut Tabelle pr&uuml;fen",correction => "Ggf. st&auml;ndige Warmwasserentnahme unterbinden&bull;Warmwasser-Temperaturf&uuml;hler richtig positionieren&bull;Bei Fehlern in der Verrohrung diese beheben&bull;Ggf. entl&uuml;ften&bull;Wenn Abweichungen bestehen, die Pumpe austauschen&bull;Warmwasserbereitung auf „Vorrang“ einstellen&bull;Zirkulationsleitung pr&uuml;fen&bull;Ggf. Abweichungen zu den Tabellenwerten den F&uuml;hler austauschen"};
	$ErrorCodeList{"0863"} = {errorcode => "A41 A42",                                subcode => "4061; 4062",                                         class => "16",          reason => "Warmwasser: Solltemperatur nicht erreicht (A41/4061 =; Warmwassersystem 1; A42/4062 =; Warmwassersystem 2)",                                                                                                                                                                                                                                       checklist => "Pr&uuml;fen, ob evtl. st&auml;ndig Wasser aufgrund von Zapfungen oder einer Leckage aus dem Warmwasserspeicher entnommen wird&bull;Die Position des Warmwasser-Temperaturf&uuml;hlers pr&uuml;fen, evtl. ist dieser falsch angebracht&bull;&bull;&Uuml;ber Funktionstest Speicherladepumpe einschalten und pr&uuml;fen, ob 230 V AC an; den Anschlussklemmen anliegen (siehe Schaltplan Modul Heizkreis)&bull;Pumpenfunktion pr&uuml;fen&bull;&bull;Die Verbindungsrohre zwischen Kessel und Speicher kontrollieren und nach Installationsanleitung pr&uuml;fen, ob diese richtig angeschlossen sind&bull;Pr&uuml;fen, ob die Heizschlange im Speicher vollst&auml;ndig entl&uuml;ftet ist&bull;Nach den technischen Dokumenten pr&uuml;fen, ob die eingebaute Speicherladepumpe &uuml;ber die erforderliche Leistung verf&uuml;gt&bull;Wenn der Warmwasser-Vorrang abgew&auml;hlt wurde und Heizung und Warmwasser im Parallelbetrieb laufen, kann evtl. die Leistung des Kessels nicht ausreichen&bull;Zu gro&szlig;e Zirkulationsleitungserluste&bull;Den Warmwasser-Temperaturf&uuml;hler laut Tabelle pr&uuml;fen",correction => "Evtl. st&auml;ndige Warmwasserentnahme unterbinden&bull;Warmwasser-Temperaturf&uuml;hler richtig positionieren&bull;Unzureichende W&auml;rmeversorgung (Kesselleistung, Temperatur)&bull;Wenn keine Spannung am Ausgang f&uuml;r die Speicherladepumpe anliegt, dann ist das Modul Warmwasser defekt und muss ausgetauscht werden&bull;Wenn die Speicherladepumpe mit Spannung versorgt wird, trotzdem nicht l&auml;uft, ist die Speicherladepumpe defekt und muss ausgetauscht werden&bull;Wenn die Speicherladepumpe nicht mit Spannung versorgt wird, dann liegt ein Problem in der Verbindungsleitung zwischen Regelger&auml;t und Pumpe vor; Hierzu Schraubklemmen und Kabel pr&uuml;fen&bull;Bei St&ouml;rungen in der Verrohrung diese beheben&bull;Ggf. entl&uuml;ften&bull;Wenn Abweichungen bestehen, die Pumpe austauschen&bull;Warmwasserbereitung auf „Vorrang“ einstellen&bull;Zirkulationsleitung pr&uuml;fen&bull;Bei Abweichungen zu den Tabellenwerten den F&uuml;hler austauschen"};
	$ErrorCodeList{"0864"} = {errorcode => "A41",                                    subcode => "4071",                                               class => "8",           reason => "ung&uuml;ltige Position Warmwassersolltemperaturschalter",                                                                                                                                                                                                                                                                                         checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0865"} = {errorcode => "A41",                                    subcode => "4081; 4082",                                         class => "16",          reason => "Modul:; t&auml;gliche Aufheizung abgebrochen",                                                                                                                                                                                                                                                                                                     checklist => "Pumpe der t&auml;glichen Aufheizung blockiert&bull;Massenstrom der t&auml;glichen Aufheizung zu klein oder kein Massenstrom&bull;Leistung der Nachheizung zu klein",                                                                                                                                                                                                                                                                                                                                  correction => "Blockierung der Pumpe der t&auml;glichen Aufheizung beheben&bull;Ausreichenden Massenstrom f&uuml;r t&auml;gliche Aufheizung gew&auml;hrleisten&bull;Nachheizleistung pr&uuml;fen oder t&auml;gliche Aufheizung in zapfarme Zeiten legen, sodass die Leistung ausreicht"                                                                                                                         };
	$ErrorCodeList{"0866"} = {errorcode => "A41",                                    subcode => "4091 4092",                                          class => "16",          reason => "Warmwasser-Modul Br&uuml;cke MC1 fehlt; (4091 = Warmwassermodul; 1; 4092 = Warmwassermo-; dul 2)",                                                                                                                                                                                                                                                 checklist => "Br&uuml;cke MC1 pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                                                         correction => "Br&uuml;cke MC1 einlegen"                                                                                                                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"0867"} = {errorcode => "A41",                                    subcode => "4161",                                               class => "8",           reason => "Warmwasser-Ladesystem/ Schichtladespeicher: Temperaturf&uuml;hler R&uuml;cklauf Zirkulation defekt",                                                                                                                                                                                                                                               checklist => "F&uuml;hler &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "F&uuml;hlerwert &uuml;berpr&uuml;fen und ggfs. austauschen"                                                                                                                                                                                                                                                                                                                                      };
	$ErrorCodeList{"0868"} = {errorcode => "A41",                                    subcode => "4171",                                               class => "8",           reason => "Warmwasser-Ladesystem/ Schichtladespeicher: Temperaturf&uuml;hler W&auml;rmetauscher Ausgang defekt",                                                                                                                                                                                                                                              checklist => "F&uuml;hler &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "F&uuml;hlerwert &uuml;berpr&uuml;fen und ggfs. austauschen"                                                                                                                                                                                                                                                                                                                                      };
	$ErrorCodeList{"0869"} = {errorcode => "A41",                                    subcode => "4181",                                               class => "8",           reason => "Warmwasser-Ladesystem/ Schichtladespeicher: Temperaturf&uuml;hler Speicher Mitte defekt",                                                                                                                                                                                                                                                          checklist => "F&uuml;hler &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "F&uuml;hlerwert &uuml;berpr&uuml;fen und ggfs. austauschen"                                                                                                                                                                                                                                                                                                                                      };
	$ErrorCodeList{"0870"} = {errorcode => "A41",                                    subcode => "4191",                                               class => "8",           reason => "Warmwasser-Ladesystem/ Schichtladespeicher: Temperaturf&uuml;hler Speicher unten defekt",                                                                                                                                                                                                                                                          checklist => "F&uuml;hler &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "F&uuml;hlerwert &uuml;berpr&uuml;fen und ggfs. austauschen"                                                                                                                                                                                                                                                                                                                                      };
	$ErrorCodeList{"0871"} = {errorcode => "A41",                                    subcode => "4211",                                               class => "8",           reason => "Warmwasser-Ladesystem/ Schichtladespeicher: Solltemperatur W&auml;rme-tauscher &uuml;berschritten",                                                                                                                                                                                                                                                checklist => "F&uuml;hler &uuml;berpr&uuml;fen&bull;&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Die Position der Warmwasser-Temperaturf&uuml;hler pr&uuml;fen, evtl. ist dieser falsch angebracht&bull;Hydraulik auf Durchluss pr&uuml;fen (Absperrventil geschlossen?)&bull;Pumpenmodulation &uuml;berpr&uuml;fen anhand SystemBedieneinheit (Funktionstest Warmwasser)&bull;Dimensionierung Pumpe, Rohrnetz, W&auml;rmetauscher &uuml;berpr&uuml;fen"                                          };
	$ErrorCodeList{"0872"} = {errorcode => "A41",                                    subcode => "6002",                                               class => "16",          reason => "Solarmodul: t&auml;gliche Aufheizung abgebrochen",                                                                                                                                                                                                                                                                                                 checklist => "Pumpe der t&auml;glichen Aufheizung blockiert&bull;Massenstrom der t&auml;glichen Aufheizung zu klein oder kein Massenstrom&bull;Leistung der Nachheizung zu klein",                                                                                                                                                                                                                                                                                                                                  correction => "Blockierung der Pumpe der t&auml;glichen Aufheizung beheben&bull;Ausreichenden Massenstrom f&uuml;r t&auml;gliche Aufheizung gew&auml;hrleisten&bull;Nachheizleistung pr&uuml;fen oder t&auml;gliche Aufheizung in zapfarme Zeiten legen, sodass die Leistung ausreicht"                                                                                                                         };
	$ErrorCodeList{"0873"} = {errorcode => "A41 A42",                                subcode => "6008",                                               class => "16",          reason => "Solarmodul: t&auml;gliche Aufheizung abgebrochen; (A41 = Warmwassersystem 1;; A42 = Warmwassersystem 2)",                                                                                                                                                                                                                                          checklist => "Pumpe der t&auml;glichen Aufheizung blockiert&bull;Massenstrom der t&auml;glichen Aufheizung zu klein oder kein Massenstrom&bull;Leistung der Nachheizung zu klein",                                                                                                                                                                                                                                                                                                                                  correction => "Blockierung der Pumpe der t&auml;glichen Aufheizung beheben&bull;Ausreichenden Massenstrom f&uuml;r t&auml;gliche Aufheizung gew&auml;hrleisten&bull;Nachheizleistung pr&uuml;fen oder t&auml;gliche Aufheizung in zapfarme Zeiten legen, sodass die Leistung ausreicht"                                                                                                                         };
	$ErrorCodeList{"0874"} = {errorcode => "A43",                                    subcode => "4014",                                               class => "8",           reason => "Frischwasserstation: Erweiterungsmodul 1; nicht erkannt",                                                                                                                                                                                                                                                                                          checklist => "Modul &uuml;berpr&uuml;fen: Spannungsversorgung vorhanden, BUS installiert, Adresskodierschalter korrekt eingestellt",                                                                                                                                                                                                                                                                                                                                                                                correction => "Modul &uuml;berpr&uuml;fen: Spannungsversorgung herstellen, BUS installieren, Adresskodierschalter einstellen gem&auml;&szlig; technischen Unterlagen"                                                                                                                                                                                                                                           };
	$ErrorCodeList{"0875"} = {errorcode => "A43",                                    subcode => "4015",                                               class => "8",           reason => "Frischwasserstation: Erweiterungsmodul 2; nicht erkannt",                                                                                                                                                                                                                                                                                          checklist => "siehe A43/4014",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0876"} = {errorcode => "A43",                                    subcode => "4016",                                               class => "8",           reason => "Frischwasserstation: Erweiterungsmodul 3; nicht erkannt",                                                                                                                                                                                                                                                                                          checklist => "siehe A43/4014",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0877"} = {errorcode => "A43",                                    subcode => "4024",                                               class => "8",           reason => "Frischwasserstation: Falscher Typ Erweiterungsmodul 1 erkannt",                                                                                                                                                                                                                                                                                    checklist => "Modultyp &uuml;berpr&uuml;fen (Typenschild) Stellung Adresscodierschalter pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                               correction => "Modul tauschen"                                                                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"0878"} = {errorcode => "A43",                                    subcode => "4025",                                               class => "8",           reason => "Frischwasserstation: Falscher Typ Erweiterungsmodul 2 erkannt",                                                                                                                                                                                                                                                                                    checklist => "siehe A43/4024",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0879"} = {errorcode => "A43",                                    subcode => "4026",                                               class => "8",           reason => "Frischwasserstation: Falscher Typ Erweiterungsmodul 3 erkannt",                                                                                                                                                                                                                                                                                    checklist => "siehe A43/4024",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0880"} = {errorcode => "A43",                                    subcode => "4043",                                               class => "8",           reason => "Frischwasserstation: Warmwassersolltemperatur nicht erreicht",                                                                                                                                                                                                                                                                                     checklist => "Temperatur Pufferspeicher &uuml;berpr&uuml;fen;; Hydraulik &uuml;berpr&uuml;fen: Absperrh&auml;hne, Luft; im System; Temperaturf&uuml;hler &uuml;berpr&uuml;fen; Pumpe &uuml;berpr&uuml;fen; Volumenstrom &uuml;berpr&uuml;fen; Plattenw&auml;rmetauscher &uuml;berpr&uuml;fen; bei Friwa-Kaskade: Kaskadenventil &uuml;berpr&uuml;fen; Dimensionierung Rohrnetz &uuml;berpr&uuml;fen",                                                                                                               correction => "Temperatur Pufferspeicher erh&ouml;hen, Absperrh&auml;hne &ouml;ffnen, Entl&uuml;ften, Temperaturf&uuml;hler tauschen, Pumpe tauschen, Volumenstrom einstellen, Plattenw&auml;rmetauscher entkalken, Temperatur, Pufferspeicher bzw. F&uuml;hlerposition anpassen, Kaskadenventil tauschen, Rohrnetzberechnung durchf&uuml;hren"                                                                 };
	$ErrorCodeList{"0881"} = {errorcode => "A43",                                    subcode => "4053",                                               class => "8",           reason => "Frischwasserstation: Thermische Desinfektion misslungen",                                                                                                                                                                                                                                                                                          checklist => "siehe A43/4043",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0882"} = {errorcode => "A43",                                    subcode => "4083",                                               class => "8",           reason => "Frischwasserstation: T&auml;gliche Aufheizung abgebrochen",                                                                                                                                                                                                                                                                                        checklist => "siehe A43/4043",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0883"} = {errorcode => "A43",                                    subcode => "4103",                                               class => "8",           reason => "Frischwasserstation: Temperaturf&uuml;hler Warmwasserauslauf defekt",                                                                                                                                                                                                                                                                              checklist => "Anschlusskabel und Anschlussklemme &uuml;berpr&uuml;fen; F&uuml;hler pr&uuml;fen (Typ, F&uuml;hlerwert)",                                                                                                                                                                                                                                                                                                                                                                                             correction => "Installation korrigieren F&uuml;hler tauschen"                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"0884"} = {errorcode => "A43",                                    subcode => "4113",                                               class => "8",           reason => "Frischwasserstation: Temperaturf&uuml;hler Einlauf defekt",                                                                                                                                                                                                                                                                                        checklist => "Anschlusskabel und Anschlussklemme &uuml;berpr&uuml;fen; F&uuml;hler pr&uuml;fen (Typ, F&uuml;hlerwert)",                                                                                                                                                                                                                                                                                                                                                                                             correction => "Installation korrigieren F&uuml;hler tauschen"                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"0885"} = {errorcode => "A43",                                    subcode => "4123",                                               class => "8",           reason => "Frischwasserstation: Volumenstromf&uuml;hler defekt",                                                                                                                                                                                                                                                                                              checklist => "Anschlusskabel und Anschlussklemme &uuml;berpr&uuml;fen; F&uuml;hler pr&uuml;fen (Typ, F&uuml;hlerwert)",                                                                                                                                                                                                                                                                                                                                                                                             correction => "Installation korrigieren F&uuml;hler tauschen"                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"0886"} = {errorcode => "A43",                                    subcode => "4133",                                               class => "8",           reason => "Frischwasserstation: Warmwassersolltemperatur &uuml;berschritten",                                                                                                                                                                                                                                                                                 checklist => "Temperatur im Pufferspeicher &uuml;berpr&uuml;fen Volumenstr&ouml;me &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                           correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0887"} = {errorcode => "A43",                                    subcode => "4143",                                               class => "8",           reason => "Frischwasserstation noch nicht gestartet",                                                                                                                                                                                                                                                                                                         checklist => "Frischwassersystem &uuml;ber Bedieneinheit starten",                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0888"} = {errorcode => "A43",                                    subcode => "4153",                                               class => "8",           reason => "Frischwasserstation: Temperaturf&uuml;hler Warmwassereinlauf defekt",                                                                                                                                                                                                                                                                              checklist => "Anschlusskabel und Anschlussklemme &uuml;berpr&uuml;fen; F&uuml;hler pr&uuml;fen (Typ, F&uuml;hlerwert)",                                                                                                                                                                                                                                                                                                                                                                                             correction => "Installation korrigieren F&uuml;hler tauschen"                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"0889"} = {errorcode => "A43",                                    subcode => "4163",                                               class => "8",           reason => "Frischwasserstation: Temperaturf&uuml;hler R&uuml;cklauf Zirkulation defekt",                                                                                                                                                                                                                                                                      checklist => "Anschlusskabel und Anschlussklemme &uuml;berpr&uuml;fen; F&uuml;hler pr&uuml;fen (Typ, F&uuml;hlerwert)",                                                                                                                                                                                                                                                                                                                                                                                             correction => "Installation korrigieren F&uuml;hler tauschen"                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"0890"} = {errorcode => "A43",                                    subcode => "4233",                                               class => "8",           reason => "Frischwasserstation: Einlauftemperatur zu niedrig",                                                                                                                                                                                                                                                                                                checklist => "Puffertemperatur &uuml;berpr&uuml;fen Dimensionierung Rohrnetz &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "Puffertemperatur bzw. F&uuml;hlerposition anpassen Rohrnetzberechnung durchf&uuml;hren"                                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"0891"} = {errorcode => "A43",                                    subcode => "6062",                                               class => "8",           reason => "Solarmodul: 2 Bus-Master erkannt",                                                                                                                                                                                                                                                                                                                 checklist => "Kodierschalter am Solarmodul steht auf Stellung 10 f&uuml;r autarken Betrieb",                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Am Solarmodul den Kodierschalter auf 1 &auml;ndern, wenn ein EMS W&auml;rmeerzeuger; angeschlossen ist"                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"0892"} = {errorcode => "A44",                                    subcode => "1010",                                               class => "8",           reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS",                                                                                                                                                                                                                                                                                                 checklist => "-&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t Ausund wieder Einschalten&bull;Busleitung reparieren bzw. austauschen&bull;Defekten EMS-BUS-Teilnehmer austauschen"                                                                                                                                                                                                                            };
	$ErrorCodeList{"0893"} = {errorcode => "A45",                                    subcode => "1010",                                               class => "8",           reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS",                                                                                                                                                                                                                                                                                                 checklist => "-&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t Ausund wieder Einschalten&bull;Busleitung reparieren bzw. austauschen&bull;Defekten EMS-BUS-Teilnehmer austauschen"                                                                                                                                                                                                                            };
	$ErrorCodeList{"0894"} = {errorcode => "A46",                                    subcode => "1010",                                               class => "8",           reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS",                                                                                                                                                                                                                                                                                                 checklist => "-&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t Ausund wieder Einschalten&bull;Busleitung reparieren bzw. austauschen&bull;Defekten EMS-BUS-Teilnehmer austauschen"                                                                                                                                                                                                                            };
	$ErrorCodeList{"0895"} = {errorcode => "A48",                                    subcode => "4088",                                               class => "8",           reason => "Speicherumladesystem: T&auml;gliche Aufheizung fehlgeschlagen",                                                                                                                                                                                                                                                                                    checklist => "Temperatur Pufferspeicher &uuml;berpr&uuml;fen; Hydraulik &uuml;berpr&uuml;fen: Absperrh&auml;hne, Luft im System; Temperaturf&uuml;hler &uuml;berpr&uuml;fen;; Pumpen &uuml;berpr&uuml;fen; Volumenstrom &uuml;berpr&uuml;fen; Plattenw&auml;rmetauscher &uuml;berpr&uuml;fen; Dimensionierung Rohrnetz &uuml;berpr&uuml;fen; Dimensionierung W&auml;rmeerzeuger &uuml;berpr&uuml;fen",                                                                                                              correction => "Temperatur Pufferspeicher erh&ouml;hen; Absperrh&auml;hne &ouml;ffnen, Entl&uuml;ften; Temperaturf&uuml;hler tauschen; Pumpe tauschen; Volumenstrom einstellen; Plattenw&auml;rmetauscher entkalken; Temperatur Pufferspeicher bzw. F&uuml;hlerposition anpassen; Rohrnetzberechnung durchf&uuml;hren"                                                                                           };
	$ErrorCodeList{"0896"} = {errorcode => "A48",                                    subcode => "4178",                                               class => "8",           reason => "Speicherumladesystem: Temperaturf&uuml;hler W&auml;rmetauscher defekt",                                                                                                                                                                                                                                                                            checklist => "Anschlusskabel pr&uuml;fen Anschlussstecker pr&uuml;fen F&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                     correction => "Anschlusskabel tauschen Anschlussstecker korrekt installieren F&uuml;hler tauschen"                                                                                                                                                                                                                                                                                                              };
	$ErrorCodeList{"0897"} = {errorcode => "A48",                                    subcode => "4188",                                               class => "8",           reason => "Speicherumladesystem: Temperaturf&uuml;hler Speicher oben defekt",                                                                                                                                                                                                                                                                                 checklist => "Anschlusskabel pr&uuml;fen Anschlussstecker pr&uuml;fen F&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                     correction => "Anschlusskabel tauschen Anschlussstecker korrekt installieren F&uuml;hler tauschen"                                                                                                                                                                                                                                                                                                              };
	$ErrorCodeList{"0898"} = {errorcode => "A48",                                    subcode => "4198",                                               class => "8",           reason => "Speicherumladesystem: Temperaturf&uuml;hler Speicher unten defekt",                                                                                                                                                                                                                                                                                checklist => "Anschlusskabel pr&uuml;fen Anschlussstecker pr&uuml;fen F&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                     correction => "Anschlusskabel tauschen Anschlussstecker korrekt installieren F&uuml;hler tauschen"                                                                                                                                                                                                                                                                                                              };
	$ErrorCodeList{"0899"} = {errorcode => "A48",                                    subcode => "4218",                                               class => "8",           reason => "Speicherumladesystem: Solltemperatur W&auml;rmetauscher &uuml;berschritten",                                                                                                                                                                                                                                                                       checklist => "Puffertemperatur &uuml;berpr&uuml;fen Volumenstrom &uuml;berpr&uuml;fen Solltemperatur &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                         correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0900"} = {errorcode => "A48",                                    subcode => "4228",                                               class => "8",           reason => "Speicherumladesystem: Temperaturf&uuml;hler Pufferspeicher oben defekt",                                                                                                                                                                                                                                                                           checklist => "Anschlusskabel pr&uuml;fen Anschlussstecker pr&uuml;fen F&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                     correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0901"} = {errorcode => "A48",                                    subcode => "6062",                                               class => "8",           reason => "Solarmodul: 2 Bus-Master erkannt",                                                                                                                                                                                                                                                                                                                 checklist => "Kodierschalter am Solarmodul steht auf Stellung 10 f&uuml;r autarken Betrieb",                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Am Solarmodul den Kodierschalter auf 1 &auml;ndern, wenn ein EMS W&auml;rmeerzeuger angeschlossen ist"                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"0902"} = {errorcode => "A51",                                    subcode => "6052",                                               class => "8",           reason => "Pumpe 3. Solarspeicher blockiert",                                                                                                                                                                                                                                                                                                                 checklist => "Absperrh&auml;hne &uuml;berpr&uuml;fen, Luft im System, Pumpe &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Absperrh&auml;hne &ouml;ffnen, Entl&uuml;ften, Pumpe &uuml;berpr&uuml;fen"                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0903"} = {errorcode => "A51",                                    subcode => "6053",                                               class => "8",           reason => "F&uuml;hler 1. Kollektorfeld falsch positioniert",                                                                                                                                                                                                                                                                                                 checklist => "F&uuml;hlerposition &uuml;berpr&uuml;fen; F&uuml;hlerinstallation am Regelger&auml;t &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                           correction => "F&uuml;hlerposition korrigieren, F&uuml;hlerinstallation am Regelger&auml;t &uuml;berpr&uuml;fen (korrekte Anschlussklemme)"                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0904"} = {errorcode => "A51",                                    subcode => "6054",                                               class => "8",           reason => "F&uuml;hler 2. Kollektorfeld falsch positioniert",                                                                                                                                                                                                                                                                                                 checklist => "F&uuml;hlerposition &uuml;berpr&uuml;fen; F&uuml;hlerinstallation am Regelger&auml;t &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                           correction => "F&uuml;hlerposition korrigieren, F&uuml;hlerinstallation am Regelger&auml;t &uuml;berpr&uuml;fen (korrekte Anschlussklemme)"                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0905"} = {errorcode => "A51",                                    subcode => "6055",                                               class => "8",           reason => "Speicherf&uuml;hler als F&uuml;hler 1. Kollektorfeld verwendet",                                                                                                                                                                                                                                                                                   checklist => "Anschlusskabel &uuml;berpr&uuml;fen, Anschlussklemme &uuml;berpr&uuml;fen, F&uuml;hler &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                         correction => "F&uuml;hlerposition korrigieren, F&uuml;hlerinstallation am Regelger&auml;t &uuml;berpr&uuml;fen (korrekte Anschlussklemme)"                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0906"} = {errorcode => "A51",                                    subcode => "6056",                                               class => "8",           reason => "Speicherf&uuml;hler als F&uuml;hler 2. Kollektorfeld verwendet",                                                                                                                                                                                                                                                                                   checklist => "Anschlusskabel &uuml;berpr&uuml;fen, Anschlussklemme &uuml;berpr&uuml;fen, F&uuml;hler &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                         correction => "F&uuml;hlerposition korrigieren, F&uuml;hlerinstallation am Regelger&auml;t &uuml;berpr&uuml;fen (korrekte Anschlussklemme)"                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0907"} = {errorcode => "A51",                                    subcode => "6057",                                               class => "8",           reason => "Kollektorf&uuml;hler als F&uuml;hler Speicher unten verwendet",                                                                                                                                                                                                                                                                                    checklist => "Anschlusskabel &uuml;berpr&uuml;fen, Anschlussklemme &uuml;berpr&uuml;fen, F&uuml;hler &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                         correction => "F&uuml;hlerposition korrigieren, F&uuml;hlerinstallation am Regelger&auml;t &uuml;berpr&uuml;fen (korrekte Anschlussklemme)"                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0908"} = {errorcode => "A51",                                    subcode => "6058",                                               class => "8",           reason => "Kollektorf&uuml;hler als F&uuml;hler Umladespeicher verwendet",                                                                                                                                                                                                                                                                                    checklist => "Anschlusskabel &uuml;berpr&uuml;fen, Anschlussklemme &uuml;berpr&uuml;fen, F&uuml;hler &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                         correction => "F&uuml;hlerposition korrigieren, F&uuml;hlerinstallation am Regelger&auml;t &uuml;berpr&uuml;fen (korrekte Anschlussklemme)"                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0909"} = {errorcode => "A51",                                    subcode => "6059",                                               class => "8",           reason => "Kollektorf&uuml;hler als F&uuml;hler 2. Solarspeicher verwendet",                                                                                                                                                                                                                                                                                  checklist => "Anschlusskabel &uuml;berpr&uuml;fen, Anschlussklemme &uuml;berpr&uuml;fen, F&uuml;hler &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                         correction => "F&uuml;hlerposition korrigieren, F&uuml;hlerinstallation am Regelger&auml;t &uuml;berpr&uuml;fen (korrekte Anschlussklemme)"                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0910"} = {errorcode => "A51",                                    subcode => "6060",                                               class => "8",           reason => "Kollektorf&uuml;hler als F&uuml;hler 3. Solarspeicher verwendet",                                                                                                                                                                                                                                                                                  checklist => "Anschlusskabel &uuml;berpr&uuml;fen, Anschlussklemme &uuml;berpr&uuml;fen, F&uuml;hler &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                         correction => "F&uuml;hlerposition korrigieren, F&uuml;hlerinstallation am Regelger&auml;t &uuml;berpr&uuml;fen (korrekte Anschlussklemme)"                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0911"} = {errorcode => "A51",                                    subcode => "6061",                                               class => "8",           reason => "Solarmodul: t&auml;gliche Aufheizung abgebrochen",                                                                                                                                                                                                                                                                                                 checklist => "Hydraulik &uuml;berpr&uuml;fen: Absperrh&auml;hne, Luft; im System; Temperaturf&uuml;hler &uuml;berpr&uuml;fen; Volumenstrom &uuml;berpr&uuml;fen; Plattenw&auml;rmetauscher &uuml;berpr&uuml;fen; Dimensionierung Rohrnetz &uuml;berpr&uuml;fen; Dimensionierung W&auml;rmeerzeuger &uuml;berpr&uuml;fen; Pumpe &uuml;berpr&uuml;fen",                                                                                                                                                               correction => "Absperrh&auml;hne &ouml;ffnen, Luft im System Temperaturf&uuml;hler Position korrigieren bzw. austauschen, Volumenstrom einstellen, Plattenw&auml;rmetauscher reinigen, Dimensionierung Rohrnetz korrigieren, Dimensionierung W&auml;rmeerzeuger anpassen, Pumpe tauschen"                                                                                                                       };
	$ErrorCodeList{"0912"} = {errorcode => "A51",                                    subcode => "6062",                                               class => "8",           reason => "Solarmodul: 2 Bus-Master erkannt",                                                                                                                                                                                                                                                                                                                 checklist => "Kodierschalter am Solarmodul steht auf Stellung 10 f&uuml;r autarken Betrieb",                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Am Solarmodul den Kodierschalter auf 1 &auml;ndern, wenn ein EMS W&auml;rmeerzeuger angeschlossen ist"                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"0913"} = {errorcode => "A51",                                    subcode => "6063",                                               class => "8",           reason => "Solarmodul: Temperaturf&uuml;hler W&auml;rmetauscher Option S defekt",                                                                                                                                                                                                                                                                             checklist => "Konfiguration pr&uuml;fen. Mit der gew&auml;hlten Einstellung ist ein F&uuml;hler W&auml;rmetauscher Heizungsunterst&uuml;tzung am Solarmodul erforderlich&bull;Verbindungsleitung zwischen Modul Solar und F&uuml;hler W&auml;rmetauscher Heizungsunterst&uuml;tzung pr&uuml;fen&bull;F&uuml;hler W&auml;rmetauscher Heizungsunterst&uuml;tzung laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des F&uuml;hlers W&auml;rmetauscher Heizungsunterst&uuml;tzung am Solarmodul laut Tabelle pr&uuml;fen",correction => "Konfiguration &auml;ndern&bull;Verbindung ordnungsgem&auml;&szlig; herstellen&bull;Wenn die Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Modul austauschen"                                                                                                      };
	$ErrorCodeList{"0914"} = {errorcode => "A51",                                    subcode => "6070",                                               class => "8",           reason => "Solarsystem noch nicht gestartet",                                                                                                                                                                                                                                                                                                                 checklist => "Solarsystem &uuml;ber Bedieneinheit starten",                                                                                                                                                                                                                                                                                                                                                                                                                                                         correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0915"} = {errorcode => "A52",                                    subcode => "1010",                                               class => "8",           reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS",                                                                                                                                                                                                                                                                                                 checklist => "-&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                             correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t Ausund wieder Einschalten&bull;Busleitung reparieren bzw. austauschen Defekten EMS-BUS-Teilnehmer austauschen"                                                                                                                                                                                                                                 };
	$ErrorCodeList{"0916"} = {errorcode => "A5",                                     subcode => "9",                                                  class => "14",          reason => "Speichertemperaturf&uuml;hler 2 (SLS oberer Temperaturf&uuml;hler) defekt",                                                                                                                                                                                                                                                                        checklist => "Anschlu&szlig;stecker am Speichertemperaturf&uuml;hler 2 nicht aufgesteckt&bull;Anschlu&szlig;kabel zum Speichertemperaturf&uuml;hler 2 besch&auml;digt&bull;Speicherf&uuml;hler 2 (NTC2) defekt&bull;Signalauswertung auf Leiterplatte defekt",                                                                                                                                                                                                                                                      correction => "Kabelbaum austauschen&bull;Speicherf&uuml;hler 2 austauschen&bull;Leiterplatte austauschen&bull;Stecker am Speichertemperaturf&uuml;hler 2 aufstecken"                                                                                                                                                                                                                                           };
	$ErrorCodeList{"0917"} = {errorcode => "A51",                                    subcode => "812",                                                class => "A",           reason => "Einstellungsfehler Solarmodul",                                                                                                                                                                                                                                                                                                                    checklist => "Pr&uuml;fen, ob in der Parametrierung des Solarmoduls der Wert „Max Speicher“ niedriger eingestellt ist als der Wert „Min Speicher“",                                                                                                                                                                                                                                                                                                                                                         correction => "Parametrierungsfehler beheben"                                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"0918"} = {errorcode => "A51",                                    subcode => "813",                                                class => "14",          reason => "Kollektortemperaturf&uuml;hler defekt; (Auswirkung: die Solaranlage geht nicht in Betrieb)",                                                                                                                                                                                                                                                       checklist => "Konfiguration pr&uuml;fen; mit der gew&auml;hlten Einstellung ist ein Kollektortemperaturf&uuml;hler erforderlich&bull;Verbindungsleitung zwischen Solarmodul und Kollektortemperaturf&uuml;hler pr&uuml;fen&bull;Kollektortemperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Kollektortemperaturf&uuml;hlers am Solarmodul laut Tabelle pr&uuml;fen",                                                                                                          correction => "Konfiguration &auml;ndern&bull;Verbindung ordnungsgem&auml;&szlig; herstellen&bull;Sollten Werte nicht &uuml;bereinstimmen, den F&uuml;hler austauschen&bull;Sollten die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Modul austauschen"                                                                                                         };
	$ErrorCodeList{"0919"} = {errorcode => "A51",                                    subcode => "814",                                                class => "14",          reason => "Speichertemperaturf&uuml;hler unten defekt (Auswirkung: die Solaranlage; geht nicht in Betrieb)",                                                                                                                                                                                                                                                  checklist => "Konfiguration pr&uuml;fen; Mit der gew&auml;hlten Einstellung ist ein Speichertemperaturf&uuml;hler unten notwendig&bull;Verbindungsleitung zwischen Solarmodul und Speichertemperaturf&uuml;hler unten pr&uuml;fen&bull;Elektrischen Anschluss der Verbindungsleitung am Solarmodul pr&uuml;fen&bull;Speichertemperaturf&uuml;hler unten laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Speichertemperaturf&uuml;hlers unten am Solarmodul laut Tabelle pr&uuml;fen",            correction => "Konfiguration &auml;ndern&bull;Verbindung ordnungsgem&auml;&szlig; herstellen&bull;Falls Schrauben oder ein Stecker lose sind, das Kontaktproblem beheben&bull;Wenn Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Modul austauschen"                              };
	$ErrorCodeList{"0920"} = {errorcode => "A51",                                    subcode => "816",                                                class => "14",          reason => "Keine Kommunikation mit Solarmodul; (Auswirkung: keine solare Absenkung bei der Nachladung. Falls das Solarmodul in Ordnung ist, wird Solarbetrieb autark weitergeregelt)",                                                                                                                                                                        checklist => "Die Verbindungsleitung EMS auf Besch&auml;digungen pr&uuml;fen&bull;Polung der Verbindungsleitung zwischen Solarmodul und EMS pr&uuml;fen&bull;Solarmodul defekt",                                                                                                                                                                                                                                                                                                                                    correction => "Besch&auml;digte Kabel austauschen&bull;St&ouml;rung in der Polung beheben&bull;Solarmodul austauschen"                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"0921"} = {errorcode => "A51",                                    subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus",                                                                                                                                                                                                                                                                                            checklist => "Pr&uuml;fen, ob Busleitung zum/vom Solarmodul falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung zum/vom Solarmodul defekt ist; Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder einschalten; pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung&bull;",                                                                                                                                                                                              correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t ausund wieder einschalten&bull;Busleitung reparieren bzw. austauschen&bull;Defekten EMS-BUS-Teilnehmer austauschen"                                                                                                                                                                                                                            };
	$ErrorCodeList{"0922"} = {errorcode => "A51",                                    subcode => "1030; 1035; 1036",                                   class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen",                                                                                                                                                                                                                                                                                                checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0923"} = {errorcode => "A51",                                    subcode => "1038",                                               class => "16",          reason => "Datum/Zeit ung&uuml;ltiger Wert",                                                                                                                                                                                                                                                                                                                  checklist => "Datum/Zeit noch nicht eingestellt&bull;Spannungsversorgung &uuml;ber l&auml;ngere Zeit ausgefallen",                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Datum/Zeit einstellen&bull;Spannungsausf&auml;lle vermeiden"                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0924"} = {errorcode => "A51",                                    subcode => "6005",                                               class => "12",          reason => "Konfigurationsfehler: Solarmodul",                                                                                                                                                                                                                                                                                                                 checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); Es ist ein Erweiterungsmodul Solar im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird; Zul&auml;ssige Modulkombination mit Hilfe technischer Dokumentation pr&uuml;fen",                                                                                                                                                                                                                                                  correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0925"} = {errorcode => "A51",                                    subcode => "6006",                                               class => "12",          reason => "Keine Kommunikation mit Erweiterungsmodul Solar",                                                                                                                                                                                                                                                                                                  checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); mit der gew&auml;hlten Einstellung ist ein Erweiterungsmodul Solar erforderlich oder die Einstellung wird nicht erkannt&bull;Die Verbindungsleitung EMS zum Erweiterungsmodul Solar auf Besch&auml;digung pr&uuml;fen Die Busspannung am Erweiterungsmodul Solar muss zwischen 12 und 15 V DC liegen&bull;Erweiterungsmodul Solar defekt",                                                                                                    correction => "Konfiguration &auml;ndern&bull;Besch&auml;digte Kabel austauschen&bull;Erweiterungsmodul Solar austauschen"                                                                                                                                                                                                                                                                                      };
	$ErrorCodeList{"0926"} = {errorcode => "A51",                                    subcode => "6007",                                               class => "12",          reason => "Konfigurationsfehler: Erweiterungsmodul Solar",                                                                                                                                                                                                                                                                                                    checklist => "Konfiguration pr&uuml;fen; Widerspr&uuml;chliche Konfigurationen f&uuml;r Solarmodul und Erweiterungsmodul Solar hinterlegt",                                                                                                                                                                                                                                                                                                                                                                         correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0927"} = {errorcode => "A51",                                    subcode => "6011",                                               class => "14",          reason => "Solarmodul: Temperaturf&uuml;hler Speicher Mitte defekt",                                                                                                                                                                                                                                                                                          checklist => "Konfiguration pr&uuml;fen; Mit der gew&auml;hlten Einstellung ist ein F&uuml;hler Speicher Mitte am Solarmodul erforderlich&bull;Verbindungsleitung zwischen Solarmodul und F&uuml;hler Speicher Mitte pr&uuml;fen&bull;F&uuml;hler Speicher Mitte laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des F&uuml;hlers Speicher Mitte am Solarmodul laut Tabelle pr&uuml;fen",                                                                                                            correction => "Konfiguration &auml;ndern&bull;Verbindung ordnungsgem&auml;&szlig; herstellen&bull;Wenn die Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Modul austauschen"                                                                                                      };
	$ErrorCodeList{"0928"} = {errorcode => "A51",                                    subcode => "6012",                                               class => "14",          reason => "Solarmodul: Puffer Bypass Schaltung Temperaturf&uuml;hler Puffer oben defekt",                                                                                                                                                                                                                                                                     checklist => "Konfiguration pr&uuml;fen; mit der gew&auml;hlten Einstellung ist ein F&uuml;hler Puffer Bypass Schaltung Puffer am Solarmodul erforderlich&bull;Verbindungsleitung zwischen Solarmodul und F&uuml;hler Puffer Bypass Schaltung Puffer pr&uuml;fen&bull;F&uuml;hler Puffer Bypass Schaltung Puffer laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des F&uuml;hlers Puffer Bypass Schaltung Puffer am Solarmodul laut Tabelle pr&uuml;fen",                                            correction => "Konfiguration &auml;ndern&bull;Verbindung ordnungsgem&auml;&szlig; herstellen&bull;Wenn die Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Modul austauschen"                                                                                                      };
	$ErrorCodeList{"0929"} = {errorcode => "A51",                                    subcode => "6013",                                               class => "14",          reason => "Solarmodul: Puffer Bypass Schaltung Temperaturf&uuml;hler R&uuml;cklauf defekt",                                                                                                                                                                                                                                                                   checklist => "Konfiguration pr&uuml;fen; mit der gew&auml;hlten Einstellung ist ein Temperaturf&uuml;hler Heizungsr&uuml;cklauf in den Puffer am Solarmodul erforderlich&bull;Verbindungsleitung zwischen Solarmodul und Temperaturf&uuml;hler Heizungsr&uuml;cklauf in den Puffer pr&uuml;fen&bull;Temperaturf&uuml;hler Heizungsr&uuml;cklauf in den Puffer laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Temperaturf&uuml;hlers Heizungsr&uuml;cklauf in den Puffer am Solarmodul laut Tabelle pr&uuml;fen",correction => "Konfiguration &auml;ndern&bull;Verbindung ordnungsgem&auml;&szlig; herstellen&bull;Wenn die Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Modul austauschen"                                                                                                      };
	$ErrorCodeList{"0930"} = {errorcode => "A51",                                    subcode => "6014",                                               class => "14",          reason => "Solarmodul: Temperaturf&uuml;hler Kollektor 2 defekt",                                                                                                                                                                                                                                                                                             checklist => "Konfiguration pr&uuml;fen; Mit der gew&auml;hlten Einstellung ist ein F&uuml;hler Kollektor 2 am Solarmodul erforderlich&bull;Verbindungsleitung zwischen Solarmodul und F&uuml;hler Kollektor 2 pr&uuml;fen&bull;F&uuml;hler Kollektor 2 laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des F&uuml;hlers Kollektor 2 am Solarmodul laut Tabelle pr&uuml;fen",                                                                                                                        correction => "Konfiguration &auml;ndern&bull;Verbindung ordnungsgem&auml;&szlig; herstellen&bull;Wenn die Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Modul austauschen"                                                                                                      };
	$ErrorCodeList{"0931"} = {errorcode => "A51",                                    subcode => "6015",                                               class => "14",          reason => "Solarmodul: Temperaturf&uuml;hler Speicher unten defekt, Ersatzbetrieb aktiv bzw. F&uuml;hler Speicherreihenschaltung; Speicher 3 (Trinkwasser), nur Monitoring",                                                                                                                                                                                  checklist => "Konfiguration pr&uuml;fen; Mit der gew&auml;hlten Einstellung ist ein F&uuml;hler Speicher 3 unten am Solarmodul erforderlich&bull;Verbindungsleitung zwischen Solarmodul und F&uuml;hler Speicher 3 unten pr&uuml;fen&bull;F&uuml;hler Speicher 3 unten laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des F&uuml;hlers Speicher 3 unten am Solarmodul; laut Tabelle pr&uuml;fen",                                                                                                   correction => "Konfiguration &auml;ndern&bull;Verbindung ordnungsgem&auml;&szlig; herstellen&bull;Wenn die Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Modul austauschen"                                                                                                      };
	$ErrorCodeList{"0932"} = {errorcode => "A51",                                    subcode => "6016",                                               class => "14",          reason => "Solarmodul: Temperaturf&uuml;hler Speicher Mitte defekt, Ersatzbetrieb  aktiv bzw.; Referenzf&uuml;hler Umladung bei Speicherreihenschaltung Speicher 3; (Trinkwasser)",                                                                                                                                                                           checklist => "Konfiguration pr&uuml;fen; Mit der gew&auml;hlten Einstellung ist ein F&uuml;hler Speicher 3 Mitte am Solarmodul erforderlich&bull;Verbindungsleitung zwischen Solarmodul und F&uuml;hler Speicher 3 Mitte pr&uuml;fen&bull;F&uuml;hler Speicher 3 Mitte laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des F&uuml;hlers Speicher 3 Mitte am Solarmodul laut Tabelle pr&uuml;fen",                                                                                                    correction => "Konfiguration &auml;ndern&bull;Verbindung ordnungsgem&auml;&szlig; herstellen&bull;Wenn die Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Modul austauschen"                                                                                                      };
	$ErrorCodeList{"0933"} = {errorcode => "A51",                                    subcode => "6017",                                               class => "14",          reason => "Solarmodul: Temperaturf&uuml;hler Speicher unten defekt, Ersatzbetrieb  aktiv, Referenzf&uuml;hler; Speicher 2",                                                                                                                                                                                                                                   checklist => "Konfiguration pr&uuml;fen; Mit der gew&auml;hlten Einstellung ist ein F&uuml;hler Speicher 2 unten am Solarmodul erforderlich&bull;Verbindungsleitung zwischen Solarmodul und F&uuml;hler Speicher 2 unten pr&uuml;fen&bull;F&uuml;hler Speicher 2 unten laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des F&uuml;hlers Speicher 2 unten am Solarmodul laut Tabelle pr&uuml;fen",                                                                                                    correction => "Konfiguration &auml;ndern&bull;Verbindung ordnungsgem&auml;&szlig; herstellen&bull;Wenn die Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Modul austauschen"                                                                                                      };
	$ErrorCodeList{"0934"} = {errorcode => "A51",                                    subcode => "6018",                                               class => "14",          reason => "Solarmodul: Temperaturf&uuml;hler Speicher Mitte defekt, Ersatzbetrieb  aktiv, Referenzf&uuml;hler; Speicher 2",                                                                                                                                                                                                                                   checklist => "Konfiguration pr&uuml;fen; Mit der gew&auml;hlten Einstellung ist ein F&uuml;hler Speicher 2 Mitte am Solarmodul erforderlich&bull;Verbindungsleitung zwischen Solarmodul und F&uuml;hler Speicher 2 Mitte pr&uuml;fen&bull;F&uuml;hler Speicher 2 Mitte laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des F&uuml;hlers Speicher 2 Mitte am Solarmodul laut Tabelle pr&uuml;fen",                                                                                                    correction => "Konfiguration &auml;ndern&bull;Verbindung ordnungsgem&auml;&szlig; herstellen&bull;Wenn die Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Modul austauschen"                                                                                                      };
	$ErrorCodeList{"0935"} = {errorcode => "A51",                                    subcode => "6019",                                               class => "14",          reason => "Solarmodul: Temperaturf&uuml;hler externer W&auml;rmetauscher defekt, Ersatzbetrieb aktiv, Referenzf&uuml;hler Speicher 1",                                                                                                                                                                                                                        checklist => "Konfiguration pr&uuml;fen; Mit der gew&auml;hlten Einstellung ist ein F&uuml;hler Speicher Mitte am Solarmodul erforderlich&bull;Verbindungsleitung zwischen Solarmodul und F&uuml;hler Speicher Mitte pr&uuml;fen&bull;F&uuml;hler Speicher Mitte laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des F&uuml;hlers Speicher Mitte am Solarmodul laut Tabelle pr&uuml;fen",                                                                                                            correction => "Konfiguration &auml;ndern&bull;Verbindung ordnungsgem&auml;&szlig; herstellen&bull;Wenn die Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Modul austauschen"                                                                                                      };
	$ErrorCodeList{"0936"} = {errorcode => "A51",                                    subcode => "6020",                                               class => "14",          reason => "Solarmodul: Temperaturf&uuml;hler thermische Desinfektion/t&auml;gliche Aufheizung defekt",                                                                                                                                                                                                                                                        checklist => "Konfiguration pr&uuml;fen; Mit der gew&auml;hlten Einstellung ist ein F&uuml;hler Temperatur thermische Desinfektion/t&auml;gliche Aufheizung am Solarmodul erforderlich&bull;Verbindungsleitung zwischen Solarmodul und F&uuml;hler Temperatur thermische Desinfektion/t&auml;gliche Aufheizung pr&uuml;fen&bull;F&uuml;hler Temperatur thermische Desinfektion/t&auml;gliche Aufheizung laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des F&uuml;hlers Temperatur thermische Desinfektion/t&auml;gliche Aufheizung am Solarmodul laut Tabelle pr&uuml;fen",correction => "Konfiguration &auml;ndern&bull;Verbindung ordnungsgem&auml;&szlig; herstellen&bull;Wenn die Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Modul austauschen"                                                                                                      };
	$ErrorCodeList{"0937"} = {errorcode => "A51",                                    subcode => "6021",                                               class => "14",          reason => "Kollektortemperaturf&uuml;hler Kollektor 1 defekt",                                                                                                                                                                                                                                                                                                checklist => "Konfiguration pr&uuml;fen; Mit der gew&auml;hlten Einstellung ist ein Kollektortemperaturf&uuml;hler erforderlich&bull;Verbindungsleitung zwischen Solarmodul und Kollektortemperaturf&uuml;hler pr&uuml;fen&bull;Kollektortemperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Kollektortemperaturf&uuml;hlers am Solarmodul laut Tabelle pr&uuml;fen",                                                                                                          correction => "Konfiguration &auml;ndern&bull;Verbindung ordnungsgem&auml;&szlig; herstellen&bull;Wenn die Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Modul austauschen"                                                                                                      };
	$ErrorCodeList{"0938"} = {errorcode => "A51",                                    subcode => "6022",                                               class => "14",          reason => "Speichertemperaturf&uuml;hler unten defekt: Ersatzbetrieb aktiv",                                                                                                                                                                                                                                                                                  checklist => "Konfiguration pr&uuml;fen; Mit der gew&auml;hlten Einstellung ist ein Speichertemperaturf&uuml;hler unten notwendig&bull;Verbindungsleitung zwischen Solarmodul und Speichertemperaturf&uuml;hler unten pr&uuml;fen&bull;Elektrischen Anschluss der Verbindungsleitung am Solarmodul pr&uuml;fen&bull;Speichertemperaturf&uuml;hler unten laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Speichertemperaturf&uuml;hlers unten am Solarmodul laut Tabelle pr&uuml;fen",            correction => "Konfiguration &auml;ndern&bull;Verbindung ordnungsgem&auml;&szlig; herstellen&bull;Falls Schrauben oder ein Stecker lose sind, das Kontaktproblem beheben&bull;Wenn Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Modul austauschen"                              };
	$ErrorCodeList{"0939"} = {errorcode => "A51",                                    subcode => "6023",                                               class => "14",          reason => "Solarmodul: Solarpumpe blockiert oder Luft im System",                                                                                                                                                                                                                                                                                             checklist => "Pr&uuml;fen, ob Solarkreislauf vollst&auml;ndig entl&uuml;ftet ist&bull;Bei modulierenden Pumpen pr&uuml;fen, ob Steuerspannung 0: 10 V DC richtig gepolt ist&bull;Bei modulierenden Pumpen pr&uuml;fen, ob separate Spannungsversorgung mit 230 V AC sichergestellt ist&bull;&Uuml;ber Funktionstest Pumpe Kollektor 2 einschalten und pr&uuml;fen, ob 230 V AC an den Anschlussklemmen anliegen (siehe Schaltplan Solarmodul)&bull;Pr&uuml;fen, ob Speicher 1 beladen wird&bull;Im Funktionstest das 3-Wege-Ventil auf Beladung Speicher 1 schalten Kontrollieren, ob 230 V AC anliegen&bull;Pr&uuml;fen, ob Speicher 2 beladen wird&bull;Wenn 1 Verbraucher: Pumpenfunktion pr&uuml;fen Bei 2 Verbrauchern: im Verbraucherkreis 2 vorgehen wie bei einem Verbraucher&bull;Wenn 2 Verbraucher von einem Kollektor &uuml;ber ein 3-Wege-Ventil versorgt werden&bull;Im Funktionstest das 3-Wege-Ventil auf Beladung Speicher 2 schalten Kontrollieren, ob 230 V AC anliegen&bull;Die Position des F&uuml;hlers Speicher 1/2 unten pr&uuml;fen, evtl. ist dieser falsch angebracht&bull;Die Position des F&uuml;hlers Kollektor 2 pr&uuml;fen, evtl. ist dieser falsch angebracht&bull;Die Verbindungsrohre zwischen Kollektor und solaren Verbrauchern kontrollieren und nach Montageanleitung pr&uuml;fen, ob diese richtig angeschlossen sind&bull;Pr&uuml;fen, ob Pumpe Kollektor 2 richtig dimensioniert ist&bull;Die F&uuml;hler laut Tabellen pr&uuml;fen",correction => "Solaranlage entl&uuml;ften&bull;Wenn verpolt, Polung korrigieren Wenn keine Spannung ansteht, muss das Modul ausgetauscht werden&bull;Spannungsversorgung sicherstellen&bull;Wenn keine Spannung am Ausgang f&uuml;r die Pumpe Kollektor 2 anliegt, dann ist das Modul defekt und muss ausgetauscht werden&bull;Wenn statt Speicher 1 Speicher 2 beladen wird, muss die Polung, die Verbindungsleitung und das 3-Wege-Ventil gepr&uuml;ft werden&bull;Wenn keine Spannung am Ausgang Speicher 1 f&uuml;r das 3-Wege-Ventil anliegt, so ist das Modul defekt und muss ausgetauscht werden&bull;Wenn statt Speicher 2 Speicher 1 beladen wird, muss die Polung, die Verbindungsleitung und das 3-Wege-Ventil gepr&uuml;ft werden&bull;Wenn die Pumpe Kollektor 2 mit Spannung versorgt wird und trotzdem nicht l&auml;uft, so ist die Pumpe Kollektor 2 defekt und muss ausgetauscht werden&bull;Wenn die Pumpe Kollektor 2 nicht mit Spannung versorgt wird, dann liegt ein Problem in der Verbindungsleitung zwischen Modul und Pumpe Kollektor 2 vor; hierzu Schraubklemmen und Kabel pr&uuml;fen&bull;Wenn keine Spannung am Ausgang Speicher 2 f&uuml;r das 3-Wege-Ventil anliegt, so ist das Modul defekt und muss ausgetauscht werden&bull;F&uuml;hler Speicher 2 unten richtig positionieren&bull;F&uuml;hler Kollektor 2 richtig positionieren&bull;Bei Fehlern in der Verrohrung diese beheben&bull;Wenn Abweichungen bestehen, die Pumpe Kollektor 2 austauschen&bull;Bei Abweichungen zu den Tabellenwerten die F&uuml;hler austauschen"};
	$ErrorCodeList{"0940"} = {errorcode => "A51",                                    subcode => "6024",                                               class => "14",          reason => "Solarmodul: Pumpe Nachladung blockiert oder; Luft im System",                                                                                                                                                                                                                                                                                      checklist => "Die Position des F&uuml;hlers Speicher 1 unten pr&uuml;fen, evtl. ist dieser falsch angebracht&bull;Die Position F&uuml;hler Speicher Mitte am Solarmodul bzw. F&uuml;hler Warmwasser pr&uuml;fen, evtl. sind diese falsch angebracht&bull;&Uuml;ber Funktionstest Trinkwasserumladepumpe einschalten und pr&uuml;fen, ob 230 V AC an den Anschlussklemmen anliegen (siehe Schaltplan Solarmodul)&bull;Pumpenfunktion pr&uuml;fen&bull;&bull;Die Verbindungsrohre zwischen den; beiden Speichern kontrollieren und nach Installationsanleitung pr&uuml;fen&bull;Pr&uuml;fen, ob Trinkwasserumladekreislauf vollst&auml;ndig entl&uuml;ftet ist&bull;Pr&uuml;fen, ob Trinkwasserumladepumpe richtig dimensioniert ist&bull;Die F&uuml;hler laut Tabellen pr&uuml;fen",correction => "F&uuml;hler Speicher 1 unten richtig positionieren&bull;F&uuml;hler Speicher Mitte am Modul bzw. F&uuml;hler Warmwasser richtig positionieren&bull;Wenn keine Spannung am Ausgang f&uuml;r die Trinkwasserumladepumpe anliegt, so ist das Modul defekt und muss ausgetauscht werden&bull;Wenn die Trinkwasserumladepumpe mit Spannung versorgt wird und trotzdem nicht l&auml;uft, dann ist die Trinkwasserumladepumpe defekt und muss ausgetauscht werden&bull;Wenn die Trinkwasserumladepumpe nicht mit Spannung versorgt wird, so liegt ein Problem in der Verbindungsleitung zwischen Modul und Trinkwasserumladepumpe vor Hierzu Schraubklemmen und Kabel pr&uuml;fen&bull;Bei St&ouml;rungen in der Verrohrung diese beheben&bull;Solaranlage entl&uuml;ften&bull;Wenn Abweichungen bestehen, dann die Trinkwasserumladepumpe austauschen&bull;Bei Abweichungen zu den Tabellenwerten die F&uuml;hler austauschen"};
	$ErrorCodeList{"0941"} = {errorcode => "A51",                                    subcode => "6025",                                               class => "14",          reason => "Solarmodul: Pumpe Zusatzspeicher blockiert, Regelung erkennt, dass kein Fl&uuml;ssigkeitsumlauf vorhanden ist",                                                                                                                                                                                                                                    checklist => "Die Position des F&uuml;hlers Speicher 2 unten pr&uuml;fen, evtl. ist dieser falsch angebracht&bull;Die Position des F&uuml;hlers Kollektor 1 pr&uuml;fen, evtl. ist dieser falsch angebracht&bull;&Uuml;ber Funktionstest Pumpe Zusatzspeicher einschalten und pr&uuml;fen, ob 230 V AC an den Anschlussklemmen anliegen (siehe Schaltplan Solarmodul)&bull;Bei modulierenden Pumpen pr&uuml;fen,; ob Steuerspannung 0: 10 V DC richtig gepolt ist&bull;Bei modulierenden Pumpen pr&uuml;fen,; ob separate Spannungsversorgung mit 230 V AC sichergestellt ist&bull;Pumpenfunktion pr&uuml;fen&bull;&bull;Die Verbindungsrohre zwischen Kollektor 1 und solarem Verbraucher 2 kontrollieren und nach Installationsanleitung pr&uuml;fen, ob diese richtig angeschlossen sind&bull;Pr&uuml;fen, ob Solarkreislauf vollst&auml;ndig entl&uuml;ftet ist&bull;Pr&uuml;fen, ob Pumpe Zusatzspeicher richtig dimensioniert ist&bull;Die F&uuml;hler laut Tabellen pr&uuml;fen",correction => "F&uuml;hler Speicher 2 unten richtig positionieren&bull;F&uuml;hler Kollektor 1 richtig positionieren&bull;Wenn keine Spannung am Ausgang f&uuml;r die Pumpe Zusatzspeicher anliegt, so ist das Modul defekt und muss ausgetauscht werden&bull;Wenn verpolt, Polung korrigieren; Wenn keine Spannung ansteht, muss das Modul ausgetauscht werden&bull;Spannungsversorgung sicherstellen&bull;Wenn die Pumpe Zusatzspeicher mit; Spannung versorgt wird und trotzdem nicht l&auml;uft, dann ist die Pumpe Zusatzspeicher defekt und muss ausgetauscht werden&bull;Wenn die Pumpe Zusatzspeicher nicht mit Spannung versorgt wird, dann liegt ein Problem in der Verbindungsleitung zwischen Modul und Pumpe Zusatzspeicher vor; Hierzu Schraubklemmen und Kabel pr&uuml;fen&bull;Bei St&ouml;rungen in der Verrohrung diese beheben&bull;Solaranlage entl&uuml;ften&bull;Wenn Abweichungen bestehen, die Pumpe Zusatzspeicher austauschen&bull;Bei Abweichungen zu den Tabellenwerten die F&uuml;hler austauschen"};
	$ErrorCodeList{"0942"} = {errorcode => "A51",                                    subcode => "6026",                                               class => "14",          reason => "Pumpe externer W&auml;rmetauscher blockiert oder; Luft im System",                                                                                                                                                                                                                                                                                 checklist => "Die Position F&uuml;hler Speicher 1/2 unten pr&uuml;fen, evtl. ist dieser falsch angebracht&bull;Die Position des F&uuml;hlers Temperatur externer W&auml;rmetauscher pr&uuml;fen, evtl. ist dieser falsch angebracht&bull;&Uuml;ber Funktionstest Pumpe externer W&auml;rmetauscher einschalten und pr&uuml;fen, ob 230 V AC/0: 10 V DC an den Anschlussklemmen anliegen; (siehe Schaltplan Solarmodul)&bull;Pumpenfunktion pr&uuml;fen&bull;&bull;Die Verbindungsrohre zwischen W&auml;rmetauscher und solaren Verbrauchern kontrollieren und nach Installationsanleitung pr&uuml;fen, ob diese richtig angeschlossen sind&bull;Pr&uuml;fen, ob W&auml;rmetauscherkreislauf vollst&auml;ndig entl&uuml;ftet ist&bull;Pr&uuml;fen, ob Pumpe externer W&auml;rmetauscher richtig dimensioniert ist&bull;Die F&uuml;hler laut Tabellen pr&uuml;fen",correction => "F&uuml;hler Speicher 1/2 unten richtig positionieren&bull;F&uuml;hler Temperatur externer W&auml;rmetauscher richtig positionieren&bull;Wenn keine Spannung am Ausgang f&uuml;r die Pumpe externer W&auml;rmetauscher anliegt, dann ist das Modul defekt und muss ausgetauscht werden&bull;Wenn die Pumpe externer W&auml;rmetauscher; mit Spannung versorgt wird und trotzdem nicht l&auml;uft, dann ist die Pumpe defekt und muss ausgetauscht werden&bull;Wenn die Pumpe externer W&auml;rmetauscher nicht mit Spannung versorgt wird, dann liegt ein Problem in der Verbindungsleitung zwischen Modul und Pumpe externer W&auml;rmetauscher vor Hierzu Schraubklemmen und Kabel pr&uuml;fen&bull;Bei St&ouml;rungen in der Verrohrung diese beheben&bull;Anlage entl&uuml;ften&bull;Wenn Abweichungen bestehen, die Pumpe externer W&auml;rmetauscher austauschen&bull;Bei Abweichungen zu den Tabellenwerten die F&uuml;hler austauschen"};
	$ErrorCodeList{"0943"} = {errorcode => "A51",                                    subcode => "6027",                                               class => "14",          reason => "Solarmodul: Pumpe thermische Desinfektion blockiert",                                                                                                                                                                                                                                                                                              checklist => "Die Position F&uuml;hler Speicher 1/2 unten pr&uuml;fen, evtl. ist dieser falsch angebracht&bull;Die Position des F&uuml;hlers Temperatur externer W&auml;rmetauscher pr&uuml;fen, evtl. ist dieser falsch angebracht&bull;&Uuml;ber Funktionstest Pumpe externer W&auml;rmetauscher einschalten und pr&uuml;fen, ob; 230 V AC/0: 10 V DC an den Anschluss-klemmen anliegen (siehe Schaltplan Solarmodul)&bull;Pumpenfunktion pr&uuml;fen&bull;&bull;Die Verbindungsrohre zwischen W&auml;rmetauscher und solaren Verbrauchern kontrollieren und nach Installationsanleitung pr&uuml;fen, ob diese richtig angeschlossen sind&bull;Pr&uuml;fen, ob W&auml;rmetauscherkreislauf vollst&auml;ndig entl&uuml;ftet ist&bull;Pr&uuml;fen, ob Pumpe externer W&auml;rmetauscher richtig dimensioniert ist",correction => "F&uuml;hler Speicher 1/2 unten richtig positionieren&bull;F&uuml;hler Temperatur externer W&auml;rmetauscher richtig positionieren&bull;Wenn keine Spannung am Ausgang f&uuml;r die Pumpe externer W&auml;rmetauscher anliegt, dann ist das Modul defekt und muss ausgetauscht werden&bull;Wenn die Pumpe externer W&auml;rmetauscher mit Spannung versorgt wird und trotzdem nicht l&auml;uft, dann ist die Pumpe defekt und muss ausgetauscht werden&bull;Wenn die Pumpe externer W&auml;rmetauscher  nicht mit Spannung versorgt wird, dann liegt ein Problem in der Verbindungsleitung zwischen Modul und Pumpe externer W&auml;rmetauscher vor Hierzu Schraubklemmen und Kabel pr&uuml;fen&bull;Bei St&ouml;rungen in der Verrohrung diese beheben&bull;Anlage entl&uuml;ften&bull;Wenn Abweichungen bestehen, die Pumpe externer W&auml;rmetauscher austauschen"};
	$ErrorCodeList{"0944"} = {errorcode => "A51",                                    subcode => "6028",                                               class => "14",          reason => "Solarmodul: Solarpumpe blockiert oder Luft im System",                                                                                                                                                                                                                                                                                             checklist => "Die Position der F&uuml;hlers Speicher 1/2 unten pr&uuml;fen, evtl. ist dieser falsch angebracht&bull;Die Position des F&uuml;hlers Kollektor 1 pr&uuml;fen, evtl. ist dieser falsch angebracht&bull;&Uuml;ber Funktionstest Pumpe Kollektor 1 einschalten und pr&uuml;fen, ob 230 V AC an den Anschlussklemmen anliegen (siehe Schaltplan Solarmodul)&bull;Bei modulierenden Pumpen pr&uuml;fen,; ob Steuerspannung 0: 10 V DC richtig gepolt ist&bull;Bei modulierenden Pumpen pr&uuml;fen, ob separate Spannungsversorgung mit 230 V AC sichergestellt ist&bull;Wenn 1 Verbraucher: Pumpenfunktion pr&uuml;fen Bei 2 Verbrauchern: im Verbraucherkreis 2 vorgehen wie bei einem Verbraucher&bull;Wenn 2 Verbraucher von einem Kollektor &uuml;ber ein 3-Wege-Ventil versorgt werden&bull;Im Funktionstest das 3-Wege-Ventil auf Beladung Speicher 1 schalten Kontrollieren ob 230 V AC anliegen&bull;Pr&uuml;fen, ob Speicher 1 beladen wird&bull;Im Funktionstest das 3-Wege-Ventil auf Beladung Speicher 2 schalten Kontrollieren, ob 230 V AC anliegen&bull;Pr&uuml;fen, ob Speicher 2 beladen wird&bull;Die Verbindungsrohre zwischen Kollektor und solaren Verbrauchern kontrollieren und nach Montageanleitung pr&uuml;fen, ob diese richtig angeschlossen sind&bull;Pr&uuml;fen, ob Solarkreislauf vollst&auml;ndig entl&uuml;ftet ist&bull;Pr&uuml;fen, ob Pumpe Kollektor 1 richtig dimensioniert ist&bull;Die F&uuml;hler laut Tabellen pr&uuml;fen",correction => "F&uuml;hler Speicher 1/2 unten richtig positionieren&bull;Kollektortemperaturf&uuml;hler richtig positionieren&bull;Wenn keine Spannung am Ausgang f&uuml;r die Pumpe Kollektor 1anliegt, dann ist das Modul defekt und muss ausgetauscht werden&bull;Wenn verpolt, Polung korrigieren Wenn keine Spannung ansteht, muss das Modul ausgetauscht werden&bull;Spannungsversorgung sicherstellen&bull;Wenn die Pumpe Kollektor 1 mit Spannung versorgt wird und trotzdem nicht l&auml;uft, so ist die Pumpe Kollektor 1 defekt und muss ausgetauscht werden&bull;Wenn die Pumpe Kollektor 1 nicht mit Spannung versorgt wird, dann liegt ein Problem in der Verbindungsleitung zwischen Modul und Pumpe Kollektor 1 vor; hierzu Schraubklemmen und Kabel pr&uuml;fen&bull;Wenn keine Spannung am Ausgang Speicher 1 f&uuml;r das 3-Wege-Ventil anliegt, dann ist das Modul defekt und muss ausgetauscht werden&bull;Wenn statt Speicher 1 Speicher 2 beladen wird, m&uuml;ssen die Polung, die Verbindungsleitung und das 3-Wege-Ventil gepr&uuml;ft werden&bull;Wenn keine Spannung am Ausgang Speicher 2 f&uuml;r das 3-Wege-Ventil anliegt, dann ist das Modul defekt und muss getauscht werden&bull;Wenn statt Speicher 2 Speicher 1 beladen wird, m&uuml;ssen die Polung, die Verbindungsleitung und das 3-Wege-Ventil gepr&uuml;ft werden&bull;Bei Fehlern in der Verrohrung diese beheben&bull;Solaranlage entl&uuml;ften&bull;Wenn Abweichungen bestehen,; die Pumpe Kollektor 1 austauschen&bull;Bei Abweichungen zu den Tabellenwerten die F&uuml;hler austauschen"};
	$ErrorCodeList{"0945"} = {errorcode => "A51",                                    subcode => "6029",                                               class => "14",          reason => "Solarmodul: Mischertemperaturf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                  checklist => "Konfiguration pr&uuml;fen; Mit der gew&auml;hlten Einstellung ist ein Mischertemperaturf&uuml;hler am Solarmodul erforderlich&bull;Verbindungsleitung zwischen Solarmodul und Mischertemperaturf&uuml;hler pr&uuml;fen&bull;Mischertemperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Mischertemperaturf&uuml;hlers und am Solarmodul laut Tabelle pr&uuml;fen",                                                                                                correction => "Konfiguration &auml;ndern&bull;Verbindung ordnungsgem&auml;&szlig; herstellen&bull;Wenn die Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Modul austauschen"                                                                                                      };
	$ErrorCodeList{"0946"} = {errorcode => "A51",                                    subcode => "6032",                                               class => "14",          reason => "Zu hoher Massenstrom im Solarkreis des; 1. Kollektorfeldes",                                                                                                                                                                                                                                                                                       checklist => "Massenstrom im Solarkreis zu hoch&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                                       correction => "Massenstrom im Solarkreis korrekt einstellen; (z. B. Pumpenstufe reduzieren), ggf. an der Solarstation Drossel weiter schlie&szlig;en&bull;Richtwerte 20: 40 l/m2 Kollektorfl&auml;che und Stunde&bull;Einstellung f&uuml;r Kollektorfl&auml;che und -typ sowie Standortfaktor im Men&uuml; „Solaroptimierung“"                                                                              };
	$ErrorCodeList{"0947"} = {errorcode => "A51",                                    subcode => "6033",                                               class => "14",          reason => "Zu niedriger Massenstrom im Solarkreis des; 1. Kollektorfeldes",                                                                                                                                                                                                                                                                                   checklist => "Massenstrom im Solarkreis zu niedrig&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "Massenstrom im Solarkreis korrekt einstellen (z. B. Pumpenstufe erh&ouml;hen), ggf. an der Solarstation Drossel weiter &ouml;ffnen&bull;Richtwerte 20: 40 l/m2 Kollektorfl&auml;che und Stunde&bull;Einstellung f&uuml;r Kollektorfl&auml;che und -typ sowie Standortfaktor im Men&uuml; „Solaroptimierung“"                                                                                 };
	$ErrorCodeList{"0948"} = {errorcode => "A51",                                    subcode => "6034",                                               class => "14",          reason => "Zu hoher Massenstrom im Solarkreis des; 2. Kollektorfeldes",                                                                                                                                                                                                                                                                                       checklist => "Massenstrom im Solarkreis zu hoch&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                                       correction => "Massenstrom im Solarkreis korrekt einstellen; (z. B. Pumpenstufe reduzieren), ggf. an der Solarstation Drossel weiter schlie&szlig;en&bull;Richtwerte 20: 40 l/m2 Kollektorfl&auml;che und Stunde&bull;Einstellung f&uuml;r Kollektorfl&auml;che und -typ sowie Standortfaktor im Men&uuml; „Solaroptimierung“"                                                                              };
	$ErrorCodeList{"0949"} = {errorcode => "A51",                                    subcode => "6035",                                               class => "14",          reason => "Zu niedriger Massenstrom im Solarkreis des; 2. Kollektorfeldes",                                                                                                                                                                                                                                                                                   checklist => "Massenstrom im Solarkreis zu niedrig&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "Massenstrom im Solarkreis korrekt einstellen (z. B. Pumpenstufe erh&ouml;hen), ggf. an der Solarstation Drossel weiter &ouml;ffnen&bull;Richtwerte 20: 40 l/m2 Kollektorfl&auml;che und Stunde&bull;Einstellung f&uuml;r Kollektorfl&auml;che und -typ sowie Standortfaktor im Men&uuml; „Solaroptimierung“"                                                                                 };
	$ErrorCodeList{"0950"} = {errorcode => "A51",                                    subcode => "6036",                                               class => "14",          reason => "Temperaturf&uuml;hler am Solarspeicher 1 defekt",                                                                                                                                                                                                                                                                                                  checklist => "Temperaturf&uuml;hler am Solarspeicher 1 pr&uuml;fen und ggf. austauschen",                                                                                                                                                                                                                                                                                                                                                                                                                           correction => "Temperaturf&uuml;hler am Solarspeicher 1 pr&uuml;fen und ggf. austauschen"                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"0951"} = {errorcode => "A51",                                    subcode => "6042",                                               class => "16",          reason => "Solarmodul: thermische Desinfektion abgebrochen",                                                                                                                                                                                                                                                                                                  checklist => "Pumpe der thermischen Desinfektion blockiert&bull;Massenstrom der thermischen Desinfektion zu klein bzw. kein Massenstrom&bull;Leistung der Nachheizung zu klein",                                                                                                                                                                                                                                                                                                                                    correction => "Blockierung der Pumpe der thermischen Desinfektion beheben&bull;Ausreichenden Massenstrom f&uuml;r thermische Desinfektion gew&auml;hrleisten&bull;Nachheizleistung pr&uuml;fen bzw. thermische Desinfektion in zapfarme Zeiten legen,; sodass die Leistung ausreicht"                                                                                                                           };
	$ErrorCodeList{"0952"} = {errorcode => "A51",                                    subcode => "6043",                                               class => "16",          reason => "Kein Signal von der VolumenstromMesseinrichtung des W&auml;rmemengenz&auml;hlers",                                                                                                                                                                                                                                                                 checklist => "Verbindungsleitung zwischen Modul und Volumenstrom-Messeinrichtung des W&auml;rmemengenz&auml;hlers pr&uuml;fen&bull;Volumenstrom-Messeinrichtung des W&auml;rmemengenz&auml;hlers pr&uuml;fen&bull;Absperrventil geschlossen&bull;Pr&uuml;fen, ob Solarkreis vollst&auml;ndig entl&uuml;ftet ist&bull;Solarpumpe f&ouml;rdert nicht",                                                                                                                                                                correction => "M&ouml;gliches Kontaktproblem beheben&bull;Volumenstrom-Messeinrichtung des W&auml;rmemengenz&auml;hlers pr&uuml;fen und ggfs. austauschen&bull;Absperrventil &ouml;ffnen&bull;Anlage entl&uuml;ften&bull;Solarpumpe pr&uuml;fen"                                                                                                                                                                };
	$ErrorCodeList{"0953"} = {errorcode => "A51",                                    subcode => "6044",                                               class => "14",          reason => "Vorlauftemperaturf&uuml;hler des W&auml;rmemengenz&auml;hlers defekt",                                                                                                                                                                                                                                                                             checklist => "Verbindungsleitung zwischen Modul und Vorlauftemperaturf&uuml;hler des W&auml;rmemengenz&auml;hlers pr&uuml;fen&bull;Elektrischen Anschluss der Verbindungsleitung im Modul pr&uuml;fen&bull;Vorlauftemperaturf&uuml;hler des W&auml;rmemengenz&auml;hlers laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Temperaturf&uuml;hlers im Regelger&auml;t laut Tabelle pr&uuml;fen",                                                                                                    correction => "Wenn ein Defekt vorliegt, Temperaturf&uuml;hler austauschen&bull;Falls Schrauben oder ein Stecker lose sind, das Kontaktproblem beheben&bull;Wenn die Werte nicht &uuml;bereinstimmen, den Temperaturf&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Modul austauschen"                                       };
	$ErrorCodeList{"0954"} = {errorcode => "A51",                                    subcode => "6045",                                               class => "14",          reason => "R&uuml;cklauftemperaturf&uuml;hler des W&auml;rmemengenz&auml;hlers defekt",                                                                                                                                                                                                                                                                       checklist => "Verbindungsleitung zwischen Modul und R&uuml;cklauftemperaturf&uuml;hler des W&auml;rmemengenz&auml;hlers pr&uuml;fen&bull;Elektrischen Anschluss der Verbindungsleitung im Modul pr&uuml;fen&bull;Temperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Temperaturf&uuml;hlers im Regelger&auml;t laut Tabelle pr&uuml;fen",                                                                                                                                      correction => "Wenn ein Defekt vorliegt, Temperaturf&uuml;hler austauschen&bull;Wenn Schrauben oder ein Stecker lose sind, dann das Kontaktproblem beheben&bull;Wenn die Werte nicht &uuml;bereinstimmen, den Temperaturf&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Modul austauschen"                                   };
	$ErrorCodeList{"0955"} = {errorcode => "A51",                                    subcode => "6046; 6047",                                         class => "14",          reason => "W&auml;rmequellen-Temperaturf&uuml;hler f&uuml;r freien Temperaturdifferenzregler defekt",                                                                                                                                                                                                                                                         checklist => "Verbindungsleitung zwischen Modul und W&auml;rmequellen-Temperaturf&uuml;hler pr&uuml;fen&bull;Elektrischen Anschluss der Verbindungsleitung im Modul pr&uuml;fen&bull;Temperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Temperaturf&uuml;hlers im Regelger&auml;t laut Tabelle pr&uuml;fen",                                                                                                                                                                  correction => "Wenn ein Defekt vorliegt, Temperaturf&uuml;hler austauschen&bull;Wenn Schrauben oder ein Stecker lose sind, das Kontaktproblem beheben&bull;Wenn die Werte nicht &uuml;bereinstimmen, dann den Temperaturf&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Modul austauschen"                                   };
	$ErrorCodeList{"0956"} = {errorcode => "A51",                                    subcode => "6048",                                               class => "14",          reason => "Pumpe f&uuml;r Umladung mit W&auml;rmetauscher blockiert",                                                                                                                                                                                                                                                                                         checklist => "Die Position des Temperaturf&uuml;hlers Speicher 1 oben pr&uuml;fen, evtl. ist dieser falsch angebracht&bull;Die Position des Temperaturf&uuml;hlers Speicher 3 unten pr&uuml;fen, evtl. ist dieser falsch angebracht&bull;&Uuml;ber Funktionstest Umladepumpe einschalten und pr&uuml;fen, ob 230 V AC/0: 10 V DC an den Anschlussklemmen anliegen (siehe Schaltplan Solarmodul)&bull;Pumpenfunktion pr&uuml;fen&bull;&bull;Die Verbindungsrohre zwischen Speicher 1 und 3 kontrollieren und nach Installationsanleitung pr&uuml;fen, ob diese richtig angeschlossen sind&bull;Pr&uuml;fen, ob Solarkreis vollst&auml;ndig entl&uuml;ftet ist&bull;Pr&uuml;fen, ob Pumpe richtig dimensioniert ist&bull;Die F&uuml;hler laut Tabellen pr&uuml;fen",correction => "Temperaturf&uuml;hler richtig positionieren&bull;Temperaturf&uuml;hler richtig positionieren&bull;Wenn keine Spannung am Ausgang f&uuml;r die Umladepumpe anliegt, so ist das Modul defekt und muss ausgetauscht werden&bull;Wenn die Pumpe mit Spannung versorgt wird und trotzdem nicht l&auml;uft, so ist die Pumpe defekt und muss ausgetauscht werden&bull;Wenn die Pumpe nicht mit Spannung versorgt wird, so liegt ein Problem in der Verbindungsleitung zwischen Modul und Pumpe vor; Hierzu Schraubklemmen und Kabel pr&uuml;fen&bull;Bei St&ouml;rungen in der Verrohrung diese beheben&bull;Anlage entl&uuml;ften&bull;Wenn Abweichungen bestehen, die Pumpe austauschen&bull;Bei Abweichungen zu den Tabellenwerten die F&uuml;hler austauschen"};
	$ErrorCodeList{"0957"} = {errorcode => "A51",                                    subcode => "6049",                                               class => "14",          reason => "Temperaturf&uuml;hler Speicher oben defekt; Umladung mit W&auml;rmetauscher: Referenzf&uuml;hler im Speicher 3 unten",                                                                                                                                                                                                                             checklist => "Verbindungsleitung zwischen Modul und Temperaturf&uuml;hler Speicher oben pr&uuml;fen&bull;Elektrischen Anschluss der Verbindungsleitung im Modul pr&uuml;fen&bull;Temperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Temperaturf&uuml;hlers im Regelger&auml;t laut Tabelle pr&uuml;fen",                                                                                                                                                                      correction => "Wenn ein Defekt vorliegt, F&uuml;hler austauschen&bull;Wenn Schrauben oder ein Stecker lose sind, das Kontaktproblem beheben&bull;Wenn Werte nicht &uuml;bereinstimmen,  dann den Temperaturf&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, das Modul austauschen"                                                     };
	$ErrorCodeList{"0958"} = {errorcode => "A51",                                    subcode => "6050",                                               class => "14",          reason => "Temperaturf&uuml;hler Speicher 3 unten defekt, Ersatz-betrieb aktiv",                                                                                                                                                                                                                                                                              checklist => "Verbindungsleitung zwischen Modul und Temperaturf&uuml;hler Speicher 3 unten pr&uuml;fen&bull;Elektrischen Anschluss der Verbindungsleitung im Modul pr&uuml;fen&bull;Temperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Temperaturf&uuml;hlers im Regelger&auml;t laut Tabelle pr&uuml;fen",                                                                                                                                                                   correction => "Wenn ein Defekt vorliegt, F&uuml;hler austauschen&bull;Wenn Schrauben oder ein Stecker lose sind, dann das Kontaktproblem beheben&bull;Wenn die Werte nicht &uuml;bereinstimmen, den Temperaturf&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, aber die Spannungswerte nicht &uuml;bereinstimmen, dann das Modul austauschen"                                             };
	$ErrorCodeList{"0959"} = {errorcode => "A52",                                    subcode => "1036",                                               class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen : A61 = Heizkreis 1",                                                                                                                                                                                                                                                                            checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0960"} = {errorcode => "A52",                                    subcode => "1036",                                               class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen : A62 = Heizkreis 2",                                                                                                                                                                                                                                                                            checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0961"} = {errorcode => "A52",                                    subcode => "1036",                                               class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen : A63 = Heizkreis 3",                                                                                                                                                                                                                                                                            checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0962"} = {errorcode => "A52",                                    subcode => "1036",                                               class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen : A64 = Heizkreis 4",                                                                                                                                                                                                                                                                            checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0963"} = {errorcode => "A52",                                    subcode => "1036",                                               class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen : A65 = Heizkreis 5",                                                                                                                                                                                                                                                                            checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0964"} = {errorcode => "A52",                                    subcode => "1036",                                               class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen : A66 = Heizkreis 6",                                                                                                                                                                                                                                                                            checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0965"} = {errorcode => "A52",                                    subcode => "1036",                                               class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen : A67 = Heizkreis 7",                                                                                                                                                                                                                                                                            checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0966"} = {errorcode => "A52",                                    subcode => "6010",                                               class => "14",          reason => "Keine Kommunikation Hauptmodul Solar",                                                                                                                                                                                                                                                                                                             checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); mit der gew&auml;hlten Einstellung ist ein Hauptmodul Solar erforderlich&bull;Die Verbindungsleitung EMS zum Haupt-modul Solar auf Besch&auml;digung pr&uuml;fen; Die Busspannung am Hauptmodul Solar muss zwischen 12 und 15 V DC liegen&bull;Hauptmodul Solar defekt",                                                                                                                                                                      correction => "Konfiguration &auml;ndern&bull;Besch&auml;digte Leitung austauschen&bull;Hauptmodul Solar austauschen"                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"0967"} = {errorcode => "A6",                                     subcode => "26",                                                 class => "",            reason => "Abgastemperaturf&uuml;hler in Brennkammer defekt",                                                                                                                                                                                                                                                                                                 checklist => "Kein Signal von F&uuml;hler Abgas&uuml;berwachnung Brennkammer&bull;Signal des F&uuml;hler Abgas&uuml;berwachung liegt au&szlig;erhalb der Kennlinie&bull;Signalunterbrechung im Anschlusskabel&bull;Signalauswertung auf Leiterplatte defekt",                                                                                                                                                                                                                                                       correction => "F&uuml;hler (Abgas&uuml;berwachung) an der Brennkammer austauschen&bull;Kabelbaum austauschen&bull;Leiterplatte austauschen&bull;Steckkontakte am F&uuml;hler aufstecken"                                                                                                                                                                                                                        };
	$ErrorCodeList{"0968"} = {errorcode => "A61",                                    subcode => "1005",                                               class => "12",          reason => "Systemkonfiguration nicht best&auml;tigt: A61 = Heizkreis 1",                                                                                                                                                                                                                                                                                      checklist => "Systemkonfiguration nicht best&auml;tigt",                                                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "Inbetriebnahme &uuml;ber die Bedieneinheit vollst&auml;ndig durchf&uuml;hren und „Konfiguration best&auml;tigen“ aktivieren."                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"0969"} = {errorcode => "A62",                                    subcode => "1005",                                               class => "8",           reason => "Systemkonfiguration nicht best&auml;tigt: A62 = Heizkreis 2",                                                                                                                                                                                                                                                                                      checklist => "Systemkonfiguration nicht best&auml;tigt",                                                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "Inbetriebnahme &uuml;ber die Bedieneinheit vollst&auml;ndig durchf&uuml;hren und „Konfiguration best&auml;tigen“ aktivieren."                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"0970"} = {errorcode => "A63",                                    subcode => "1005",                                               class => "8",           reason => "Systemkonfiguration nicht best&auml;tigt: A63 = Heizkreis 3",                                                                                                                                                                                                                                                                                      checklist => "Systemkonfiguration nicht best&auml;tigt",                                                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "Inbetriebnahme &uuml;ber die Bedieneinheit vollst&auml;ndig durchf&uuml;hren und „Konfiguration best&auml;tigen“ aktivieren."                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"0971"} = {errorcode => "A64",                                    subcode => "1005",                                               class => "8",           reason => "Systemkonfiguration nicht best&auml;tigt: A64 = Heizkreis 4",                                                                                                                                                                                                                                                                                      checklist => "Systemkonfiguration nicht best&auml;tigt",                                                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "Inbetriebnahme &uuml;ber die Bedieneinheit vollst&auml;ndig durchf&uuml;hren und „Konfiguration best&auml;tigen“ aktivieren."                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"0972"} = {errorcode => "A65",                                    subcode => "1005",                                               class => "8",           reason => "Systemkonfiguration nicht best&auml;tigt: A65 = Heizkreis 5",                                                                                                                                                                                                                                                                                      checklist => "Systemkonfiguration nicht best&auml;tigt",                                                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "Inbetriebnahme &uuml;ber die Bedieneinheit vollst&auml;ndig durchf&uuml;hren und „Konfiguration best&auml;tigen“ aktivieren."                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"0973"} = {errorcode => "A66",                                    subcode => "1005",                                               class => "8",           reason => "Systemkonfiguration nicht best&auml;tigt: A66 = Heizkreis 6",                                                                                                                                                                                                                                                                                      checklist => "Systemkonfiguration nicht best&auml;tigt",                                                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "Inbetriebnahme &uuml;ber die Bedieneinheit vollst&auml;ndig durchf&uuml;hren und „Konfiguration best&auml;tigen“ aktivieren."                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"0974"} = {errorcode => "A67",                                    subcode => "1005",                                               class => "8",           reason => "Systemkonfiguration nicht best&auml;tigt: A67 = Heizkreis 7",                                                                                                                                                                                                                                                                                      checklist => "Systemkonfiguration nicht best&auml;tigt",                                                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "Inbetriebnahme &uuml;ber die Bedieneinheit vollst&auml;ndig durchf&uuml;hren und „Konfiguration best&auml;tigen“ aktivieren."                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"0975"} = {errorcode => "A68",                                    subcode => "1005",                                               class => "8",           reason => "Systemkonfiguration nicht best&auml;tigt: A67 = Heizkreis 8",                                                                                                                                                                                                                                                                                      checklist => "Systemkonfiguration nicht best&auml;tigt",                                                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "Inbetriebnahme &uuml;ber die Bedieneinheit vollst&auml;ndig durchf&uuml;hren und „Konfiguration best&auml;tigen“ aktivieren."                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"0976"} = {errorcode => "A61",                                    subcode => "1007",                                               class => "8",           reason => "Verbindung zur Basisstation unterbrochen: A61 = Heizkreis 1",                                                                                                                                                                                                                                                                                      checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0977"} = {errorcode => "A62",                                    subcode => "1007",                                               class => "8",           reason => "Verbindung zur Basisstation unterbrochen: A62 = Heizkreis 2",                                                                                                                                                                                                                                                                                      checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0978"} = {errorcode => "A63",                                    subcode => "1007",                                               class => "8",           reason => "Verbindung zur Basisstation unterbrochen: A63 = Heizkreis 3",                                                                                                                                                                                                                                                                                      checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0979"} = {errorcode => "A64",                                    subcode => "1007",                                               class => "8",           reason => "Verbindung zur Basisstation unterbrochen: A64 = Heizkreis 4",                                                                                                                                                                                                                                                                                      checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0980"} = {errorcode => "A65",                                    subcode => "1007",                                               class => "8",           reason => "Verbindung zur Basisstation unterbrochen: A65 = Heizkreis 5",                                                                                                                                                                                                                                                                                      checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0981"} = {errorcode => "A66",                                    subcode => "1007",                                               class => "8",           reason => "Verbindung zur Basisstation unterbrochen: A66 = Heizkreis 6",                                                                                                                                                                                                                                                                                      checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0982"} = {errorcode => "A67",                                    subcode => "1007",                                               class => "8",           reason => "Verbindung zur Basisstation unterbrochen: A67 = Heizkreis 7",                                                                                                                                                                                                                                                                                      checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0983"} = {errorcode => "A68",                                    subcode => "1007",                                               class => "8",           reason => "Verbindung zur Basisstation unterbrochen: A68 = Heizkreis 8",                                                                                                                                                                                                                                                                                      checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"0984"} = {errorcode => "A61",                                    subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus : A61 = Heizkreis 1",                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder einschalten Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung.&bull;",                                                                                                                                                                                                                                     correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t ausund wieder einschalten&bull;Busleitung reparieren bzw. austauschen&bull;Defekten EMS-BUS-Teilnehmer austauschen"                                                                                                                                                                                                                            };
	$ErrorCodeList{"0985"} = {errorcode => "A62",                                    subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus : A62 = Heizkreis 2",                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder einschalten Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung.&bull;",                                                                                                                                                                                                                                     correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t ausund wieder einschalten&bull;Busleitung reparieren bzw. austauschen&bull;Defekten EMS-BUS-Teilnehmer austauschen"                                                                                                                                                                                                                            };
	$ErrorCodeList{"0986"} = {errorcode => "A63",                                    subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus : A63 = Heizkreis 3",                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder einschalten Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung.&bull;",                                                                                                                                                                                                                                     correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t ausund wieder einschalten&bull;Busleitung reparieren bzw. austauschen&bull;Defekten EMS-BUS-Teilnehmer austauschen"                                                                                                                                                                                                                            };
	$ErrorCodeList{"0987"} = {errorcode => "A64",                                    subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus : A64 = Heizkreis 4",                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder einschalten Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung.&bull;",                                                                                                                                                                                                                                     correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t ausund wieder einschalten&bull;Busleitung reparieren bzw. austauschen&bull;Defekten EMS-BUS-Teilnehmer austauschen"                                                                                                                                                                                                                            };
	$ErrorCodeList{"0988"} = {errorcode => "A65",                                    subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus : A65 = Heizkreis 5",                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder einschalten Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung.&bull;",                                                                                                                                                                                                                                     correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t ausund wieder einschalten&bull;Busleitung reparieren bzw. austauschen&bull;Defekten EMS-BUS-Teilnehmer austauschen"                                                                                                                                                                                                                            };
	$ErrorCodeList{"0989"} = {errorcode => "A66",                                    subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus : A66 = Heizkreis 6",                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder einschalten Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung.&bull;",                                                                                                                                                                                                                                     correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t ausund wieder einschalten&bull;Busleitung reparieren bzw. austauschen&bull;Defekten EMS-BUS-Teilnehmer austauschen"                                                                                                                                                                                                                            };
	$ErrorCodeList{"0990"} = {errorcode => "A67",                                    subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus : A67 = Heizkreis 7",                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder einschalten Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung.&bull;",                                                                                                                                                                                                                                     correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t ausund wieder einschalten&bull;Busleitung reparieren bzw. austauschen&bull;Defekten EMS-BUS-Teilnehmer austauschen"                                                                                                                                                                                                                            };
	$ErrorCodeList{"0991"} = {errorcode => "A68",                                    subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus : A68 = Heizkreis 8",                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder einschalten Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung.&bull;",                                                                                                                                                                                                                                     correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t ausund wieder einschalten&bull;Busleitung reparieren bzw. austauschen&bull;Defekten EMS-BUS-Teilnehmer austauschen"                                                                                                                                                                                                                            };
	$ErrorCodeList{"0992"} = {errorcode => "A61",                                    subcode => "1030; 1033; 1034: 1035; 1036",                       class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen: A61 = Heizkreis 1",                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0993"} = {errorcode => "A62",                                    subcode => "1030; 1033; 1034: 1035; 1036",                       class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen: A62 = Heizkreis 2",                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0994"} = {errorcode => "A63",                                    subcode => "1030; 1033; 1034: 1035; 1036",                       class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen: A63 = Heizkreis 3",                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0995"} = {errorcode => "A64",                                    subcode => "1030; 1033; 1034: 1035; 1036",                       class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen: A64 = Heizkreis 4",                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0996"} = {errorcode => "A65",                                    subcode => "1030; 1033; 1034: 1035; 1036",                       class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen: A65 = Heizkreis 5",                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0997"} = {errorcode => "A66",                                    subcode => "1030; 1033; 1034: 1035; 1036",                       class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen: A66 = Heizkreis 6",                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0998"} = {errorcode => "A67",                                    subcode => "1030; 1033; 1034: 1035; 1036",                       class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen: A67 = Heizkreis 7",                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"0999"} = {errorcode => "A68",                                    subcode => "1030; 1033; 1034: 1035; 1036",                       class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen: A68 = Heizkreis 8",                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"1000"} = {errorcode => "A61 A62 A63 A64 A65 A66 A67 A68",        subcode => "1037",                                               class => "14",          reason => "Au&szlig;entemperaturf&uuml;hler defekt; (A61 = Heizkreis 1; A62 = Heizkreis 2; A63 = Heizkreis 3; A64 = Heizkreis 4; A65 = Heizkreis 5; A66 = Heizkreis 6; A67 = Heizkreis 7; A68 = Heizkreis 8)",                                                                                                                                                checklist => "Konfiguration pr&uuml;fen; mit der gew&auml;hlten Einstellung ist ein Au&szlig;entemperaturf&uuml;hler erforderlich&bull;Verbindungsleitung zwischen Regelger&auml;t und Au&szlig;entemperaturf&uuml;hler auf Durchgang pr&uuml;fen&bull;Elektrischen Anschluss der Verbindungsleitung im Au&szlig;entemperaturf&uuml;hler bzw. am Stecker im Regelger&auml;t pr&uuml;fen&bull;Au&szlig;entemperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Au&szlig;entemperaturf&uuml;hlers im Regelger&auml;t laut Tabelle pr&uuml;fen",correction => "Konfiguration &auml;ndern&bull;Wenn kein Durchgang vorhanden ist, die St&ouml;rung beheben&bull;Falls Schrauben oder ein Stecker lose sind, das Kontaktproblem beheben&bull;Wenn Werte nicht &uuml;bereinstimmen, dann den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, die Spannungswerte jedoch nicht &uuml;berein-; stimmen, dann das Regelger&auml;t austauschen"  };
	$ErrorCodeList{"1001"} = {errorcode => "A61 A62 A63 A64 A65 A66 A67 A68",        subcode => "1038",                                               class => "16",          reason => "Datum/Zeit ung&uuml;ltiger Wert (A61 = Heizkreis 1;; A62 = Heizkreis 2; A63 = Heizkreis 3; A64 = Heizkreis 4; A65 = Heizkreis 5; A66 = Heizkreis 6; A67 = Heizkreis 7; A68 = Heizkreis 8)",                                                                                                                                                        checklist => "Datum/Zeit noch nicht eingestellt&bull;Spannungsversorgung &uuml;ber l&auml;ngere Zeit ausgefallen",                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Datum/Zeit einstellen&bull;Spannungsausf&auml;lle vermeiden"                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"1002"} = {errorcode => "A61 A62 A63 A64 A65 A66 A67 A68",        subcode => "1042",                                               class => "14",          reason => "Interner Fehler:; Zugriff auf Uhrenbaustein blockiert",                                                                                                                                                                                                                                                                                            checklist => "Interner Fehler",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                     correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1003"} = {errorcode => "A61",                                    subcode => "1051; 1052; 1053; 1054; 1055; 1056; 1057; 1058",     class => "14",          reason => "Keine Kommunikation Modul externer Raumtemperaturf&uuml;hler : A61 = Heizkreis 1",                                                                                                                                                                                                                                                                 checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul) Mit der gew&auml;hlten Einstellung ist ein Modul externer F&uuml;hler Raumtemperaturf&uuml;hler erforderlich Die Verbindungsleitung EMS zum Modul externer Raumtemperaturf&uuml;hler auf Besch&auml;digung pr&uuml;fen; die Busspannung am Modul externer Raumtemperaturf&uuml;hler muss zwischen 12 und 15 V DC liegen&bull;Modul externer F&uuml;hler Raumtemperatur defekt",                                                                correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1004"} = {errorcode => "A62",                                    subcode => "1051; 1052; 1053; 1054; 1055; 1056; 1057; 1058",     class => "14",          reason => "Keine Kommunikation Modul externer Raumtemperaturf&uuml;hler : A62 = Heizkreis 2",                                                                                                                                                                                                                                                                 checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul) Mit der gew&auml;hlten Einstellung ist ein Modul externer F&uuml;hler Raumtemperaturf&uuml;hler erforderlich Die Verbindungsleitung EMS zum Modul externer Raumtemperaturf&uuml;hler auf Besch&auml;digung pr&uuml;fen; die Busspannung am Modul externer Raumtemperaturf&uuml;hler muss zwischen 12 und 15 V DC liegen&bull;Modul externer F&uuml;hler Raumtemperatur defekt",                                                                correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1005"} = {errorcode => "A63",                                    subcode => "1051; 1052; 1053; 1054; 1055; 1056; 1057; 1058",     class => "14",          reason => "Keine Kommunikation Modul externer Raumtemperaturf&uuml;hler : A63 = Heizkreis 3",                                                                                                                                                                                                                                                                 checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul) Mit der gew&auml;hlten Einstellung ist ein Modul externer F&uuml;hler Raumtemperaturf&uuml;hler erforderlich Die Verbindungsleitung EMS zum Modul externer Raumtemperaturf&uuml;hler auf Besch&auml;digung pr&uuml;fen; die Busspannung am Modul externer Raumtemperaturf&uuml;hler muss zwischen 12 und 15 V DC liegen&bull;Modul externer F&uuml;hler Raumtemperatur defekt",                                                                correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1006"} = {errorcode => "A64",                                    subcode => "1051; 1052; 1053; 1054; 1055; 1056; 1057; 1058",     class => "14",          reason => "Keine Kommunikation Modul externer Raumtemperaturf&uuml;hler : A64 = Heizkreis 4",                                                                                                                                                                                                                                                                 checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul) Mit der gew&auml;hlten Einstellung ist ein Modul externer F&uuml;hler Raumtemperaturf&uuml;hler erforderlich Die Verbindungsleitung EMS zum Modul externer Raumtemperaturf&uuml;hler auf Besch&auml;digung pr&uuml;fen; die Busspannung am Modul externer Raumtemperaturf&uuml;hler muss zwischen 12 und 15 V DC liegen&bull;Modul externer F&uuml;hler Raumtemperatur defekt",                                                                correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1007"} = {errorcode => "A65",                                    subcode => "1051; 1052; 1053; 1054; 1055; 1056; 1057; 1058",     class => "14",          reason => "Keine Kommunikation Modul externer Raumtemperaturf&uuml;hler : A65 = Heizkreis 5",                                                                                                                                                                                                                                                                 checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul) Mit der gew&auml;hlten Einstellung ist ein Modul externer F&uuml;hler Raumtemperaturf&uuml;hler erforderlich Die Verbindungsleitung EMS zum Modul externer Raumtemperaturf&uuml;hler auf Besch&auml;digung pr&uuml;fen; die Busspannung am Modul externer Raumtemperaturf&uuml;hler muss zwischen 12 und 15 V DC liegen&bull;Modul externer F&uuml;hler Raumtemperatur defekt",                                                                correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1008"} = {errorcode => "A66",                                    subcode => "1051; 1052; 1053; 1054; 1055; 1056; 1057; 1058",     class => "14",          reason => "Keine Kommunikation Modul externer Raumtemperaturf&uuml;hler : A66 = Heizkreis 6",                                                                                                                                                                                                                                                                 checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul) Mit der gew&auml;hlten Einstellung ist ein Modul externer F&uuml;hler Raumtemperaturf&uuml;hler erforderlich Die Verbindungsleitung EMS zum Modul externer Raumtemperaturf&uuml;hler auf Besch&auml;digung pr&uuml;fen; die Busspannung am Modul externer Raumtemperaturf&uuml;hler muss zwischen 12 und 15 V DC liegen&bull;Modul externer F&uuml;hler Raumtemperatur defekt",                                                                correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1009"} = {errorcode => "A67",                                    subcode => "1051; 1052; 1053; 1054; 1055; 1056; 1057; 1058",     class => "14",          reason => "Keine Kommunikation Modul externer Raumtemperaturf&uuml;hler : A67 = Heizkreis 7",                                                                                                                                                                                                                                                                 checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul) Mit der gew&auml;hlten Einstellung ist ein Modul externer F&uuml;hler Raumtemperaturf&uuml;hler erforderlich Die Verbindungsleitung EMS zum Modul externer Raumtemperaturf&uuml;hler auf Besch&auml;digung pr&uuml;fen; die Busspannung am Modul externer Raumtemperaturf&uuml;hler muss zwischen 12 und 15 V DC liegen&bull;Modul externer F&uuml;hler Raumtemperatur defekt",                                                                correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1010"} = {errorcode => "A68",                                    subcode => "1057; 1058",                                         class => "14",          reason => "Keine Kommunikation Modul externer Raumtemperaturf&uuml;hler : A68 = Heizkreis 8",                                                                                                                                                                                                                                                                 checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul) Mit der gew&auml;hlten Einstellung ist ein Modul externer F&uuml;hler Raumtemperaturf&uuml;hler erforderlich Die Verbindungsleitung EMS zum Modul externer Raumtemperaturf&uuml;hler auf Besch&auml;digung pr&uuml;fen; die Busspannung am Modul externer Raumtemperaturf&uuml;hler muss zwischen 12 und 15 V DC liegen&bull;Modul externer F&uuml;hler Raumtemperatur defekt",                                                                correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1011"} = {errorcode => "A61",                                    subcode => "1081; 1082; 1083;1084; 1085; 1086; 1087; 1088",      class => "14",          reason => "Zwei Master-Bedieneinheiten im System: : A61 = Heizkreis 1",                                                                                                                                                                                                                                                                                       checklist => "In der Installationsebene die Parametrierung pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "Die Bedieneinheit f&uuml;r den Heizkreis als Master anmelden"                                                                                                                                                                                                                                                                                                                                    };
	$ErrorCodeList{"1012"} = {errorcode => "A62",                                    subcode => "1081; 1082; 1083;1084; 1085; 1086; 1087; 1088",      class => "14",          reason => "Zwei Master-Bedieneinheiten im System: : A62 = Heizkreis 2",                                                                                                                                                                                                                                                                                       checklist => "In der Installationsebene die Parametrierung pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "Die Bedieneinheit f&uuml;r den Heizkreis als Master anmelden"                                                                                                                                                                                                                                                                                                                                    };
	$ErrorCodeList{"1013"} = {errorcode => "A63",                                    subcode => "1081; 1082; 1083;1084; 1085; 1086; 1087; 1088",      class => "14",          reason => "Zwei Master-Bedieneinheiten im System: : A63 = Heizkreis 3",                                                                                                                                                                                                                                                                                       checklist => "In der Installationsebene die Parametrierung pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "Die Bedieneinheit f&uuml;r den Heizkreis als Master anmelden"                                                                                                                                                                                                                                                                                                                                    };
	$ErrorCodeList{"1014"} = {errorcode => "A64",                                    subcode => "1081; 1082; 1083;1084; 1085; 1086; 1087; 1088",      class => "14",          reason => "Zwei Master-Bedieneinheiten im System: : A64 = Heizkreis 4",                                                                                                                                                                                                                                                                                       checklist => "In der Installationsebene die Parametrierung pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "Die Bedieneinheit f&uuml;r den Heizkreis als Master anmelden"                                                                                                                                                                                                                                                                                                                                    };
	$ErrorCodeList{"1015"} = {errorcode => "A65",                                    subcode => "1081; 1082; 1083;1084; 1085; 1086; 1087; 1088",      class => "14",          reason => "Zwei Master-Bedieneinheiten im System: : A65 = Heizkreis 5",                                                                                                                                                                                                                                                                                       checklist => "In der Installationsebene die Parametrierung pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "Die Bedieneinheit f&uuml;r den Heizkreis als Master anmelden"                                                                                                                                                                                                                                                                                                                                    };
	$ErrorCodeList{"1016"} = {errorcode => "A66",                                    subcode => "1081; 1082; 1083;1084; 1085; 1086; 1087; 1088",      class => "14",          reason => "Zwei Master-Bedieneinheiten im System: : A66 = Heizkreis 6",                                                                                                                                                                                                                                                                                       checklist => "In der Installationsebene die Parametrierung pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "Die Bedieneinheit f&uuml;r den Heizkreis als Master anmelden"                                                                                                                                                                                                                                                                                                                                    };
	$ErrorCodeList{"1017"} = {errorcode => "A67",                                    subcode => "1081; 1082; 1083;1084; 1085; 1086; 1087; 1088",      class => "14",          reason => "Zwei Master-Bedieneinheiten im System: : A67 = Heizkreis 7",                                                                                                                                                                                                                                                                                       checklist => "In der Installationsebene die Parametrierung pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "Die Bedieneinheit f&uuml;r den Heizkreis als Master anmelden"                                                                                                                                                                                                                                                                                                                                    };
	$ErrorCodeList{"1018"} = {errorcode => "A68",                                    subcode => "1081; 1082; 1083;1084; 1085; 1086; 1087; 1088",      class => "14",          reason => "Zwei Master-Bedieneinheiten im System: : A68 = Heizkreis 8",                                                                                                                                                                                                                                                                                       checklist => "In der Installationsebene die Parametrierung pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "Die Bedieneinheit f&uuml;r den Heizkreis als Master anmelden"                                                                                                                                                                                                                                                                                                                                    };
	$ErrorCodeList{"1019"} = {errorcode => "A61",                                    subcode => "3001; 3002; 3003; 3004; 3005; 3006; 3007; 3008",     class => "14",          reason => "Konfigurationsfehler: Modul externer Raumtemperaturf&uuml;hler nicht verwendet : A61 = Heizkreis 1",                                                                                                                                                                                                                                               checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul) Es ist ein Modul externer Raumtemperaturf&uuml;hler im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird ",                                                                                                                                                                                                                                                                                                                  correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1020"} = {errorcode => "A62",                                    subcode => "3001; 3002; 3003; 3004; 3005; 3006; 3007; 3008",     class => "14",          reason => "Konfigurationsfehler: Modul externer Raumtemperaturf&uuml;hler nicht verwendet : A62 = Heizkreis 2",                                                                                                                                                                                                                                               checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul) Es ist ein Modul externer Raumtemperaturf&uuml;hler im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird ",                                                                                                                                                                                                                                                                                                                  correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1021"} = {errorcode => "A63",                                    subcode => "3001; 3002; 3003; 3004; 3005; 3006; 3007; 3008",     class => "14",          reason => "Konfigurationsfehler: Modul externer Raumtemperaturf&uuml;hler nicht verwendet : A63 = Heizkreis 3",                                                                                                                                                                                                                                               checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul) Es ist ein Modul externer Raumtemperaturf&uuml;hler im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird ",                                                                                                                                                                                                                                                                                                                  correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1022"} = {errorcode => "A64",                                    subcode => "3001; 3002; 3003; 3004; 3005; 3006; 3007; 3008",     class => "14",          reason => "Konfigurationsfehler: Modul externer Raumtemperaturf&uuml;hler nicht verwendet : A64 = Heizkreis 4",                                                                                                                                                                                                                                               checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul) Es ist ein Modul externer Raumtemperaturf&uuml;hler im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird ",                                                                                                                                                                                                                                                                                                                  correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1023"} = {errorcode => "A65",                                    subcode => "3001; 3002; 3003; 3004; 3005; 3006; 3007; 3008",     class => "14",          reason => "Konfigurationsfehler: Modul externer Raumtemperaturf&uuml;hler nicht verwendet : A65 = Heizkreis 5",                                                                                                                                                                                                                                               checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul) Es ist ein Modul externer Raumtemperaturf&uuml;hler im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird ",                                                                                                                                                                                                                                                                                                                  correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1024"} = {errorcode => "A66",                                    subcode => "3001; 3002; 3003; 3004; 3005; 3006; 3007; 3008",     class => "14",          reason => "Konfigurationsfehler: Modul externer Raumtemperaturf&uuml;hler nicht verwendet : A66 = Heizkreis 6",                                                                                                                                                                                                                                               checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul) Es ist ein Modul externer Raumtemperaturf&uuml;hler im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird ",                                                                                                                                                                                                                                                                                                                  correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1025"} = {errorcode => "A67",                                    subcode => "3001; 3002; 3003; 3004; 3005; 3006; 3007; 3008",     class => "14",          reason => "Konfigurationsfehler: Modul externer Raumtemperaturf&uuml;hler nicht verwendet : A67 = Heizkreis 7",                                                                                                                                                                                                                                               checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul) Es ist ein Modul externer Raumtemperaturf&uuml;hler im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird ",                                                                                                                                                                                                                                                                                                                  correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1026"} = {errorcode => "A68",                                    subcode => "3001; 3002; 3003; 3004; 3005; 3006; 3007; 3008",     class => "14",          reason => "Konfigurationsfehler: Modul externer Raumtemperaturf&uuml;hler nicht verwendet : A68 = Heizkreis 8",                                                                                                                                                                                                                                               checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul) Es ist ein Modul externer Raumtemperaturf&uuml;hler im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird ",                                                                                                                                                                                                                                                                                                                  correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1027"} = {errorcode => "A61",                                    subcode => "3011; 3012; 3013; 3014; 3015; 3016; 3017; 3018",     class => "14",          reason => "Konfigurationsfehler: Mischermodul nicht verwendet : A61 = Heizkreis 1",                                                                                                                                                                                                                                                                           checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul) Es ist ein Mischermodul im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird.",                                                                                                                                                                                                                                                                                                                                              correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1028"} = {errorcode => "A62",                                    subcode => "3011; 3012; 3013; 3014; 3015; 3016; 3017; 3018",     class => "14",          reason => "Konfigurationsfehler: Mischermodul nicht verwendet : A62 = Heizkreis 2",                                                                                                                                                                                                                                                                           checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul) Es ist ein Mischermodul im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird.",                                                                                                                                                                                                                                                                                                                                              correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1029"} = {errorcode => "A63",                                    subcode => "3011; 3012; 3013; 3014; 3015; 3016; 3017; 3018",     class => "14",          reason => "Konfigurationsfehler: Mischermodul nicht verwendet : A63 = Heizkreis 3",                                                                                                                                                                                                                                                                           checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul) Es ist ein Mischermodul im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird.",                                                                                                                                                                                                                                                                                                                                              correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1030"} = {errorcode => "A64",                                    subcode => "3011; 3012; 3013; 3014; 3015; 3016; 3017; 3018",     class => "14",          reason => "Konfigurationsfehler: Mischermodul nicht verwendet : A64 = Heizkreis 4",                                                                                                                                                                                                                                                                           checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul) Es ist ein Mischermodul im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird.",                                                                                                                                                                                                                                                                                                                                              correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1031"} = {errorcode => "A65",                                    subcode => "3011; 3012; 3013; 3014; 3015; 3016; 3017; 3018",     class => "14",          reason => "Konfigurationsfehler: Mischermodul nicht verwendet : A65 = Heizkreis 5",                                                                                                                                                                                                                                                                           checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul) Es ist ein Mischermodul im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird.",                                                                                                                                                                                                                                                                                                                                              correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1032"} = {errorcode => "A66",                                    subcode => "3011; 3012; 3013; 3014; 3015; 3016; 3017; 3018",     class => "14",          reason => "Konfigurationsfehler: Mischermodul nicht verwendet : A66 = Heizkreis 6",                                                                                                                                                                                                                                                                           checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul) Es ist ein Mischermodul im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird.",                                                                                                                                                                                                                                                                                                                                              correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1033"} = {errorcode => "A67",                                    subcode => "3011; 3012; 3013; 3014; 3015; 3016; 3017; 3018",     class => "14",          reason => "Konfigurationsfehler: Mischermodul nicht verwendet : A67 = Heizkreis 7",                                                                                                                                                                                                                                                                           checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul) Es ist ein Mischermodul im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird.",                                                                                                                                                                                                                                                                                                                                              correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1034"} = {errorcode => "A68",                                    subcode => "3011; 3012; 3013; 3014; 3015; 3016; 3017; 3018",     class => "14",          reason => "Konfigurationsfehler: Mischermodul nicht verwendet : A68 = Heizkreis 8",                                                                                                                                                                                                                                                                           checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul) Es ist ein Mischermodul im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird.",                                                                                                                                                                                                                                                                                                                                              correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1035"} = {errorcode => "A61",                                    subcode => "3061; 3062; 3063; 3064; 3065; 3066; 3067; 3068",     class => "14",          reason => "Keine Kommunikation mit Mischermodul : A61 = Heizkreis 1",                                                                                                                                                                                                                                                                                         checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); mit der gew&auml;hlten Einstellung ist ein Mischermodul erforderlich&bull;Die Verbindungsleitung EMS zum Mischermodul auf Besch&auml;digung pr&uuml;fen; Die Busspannung am Mischermodul muss zwischen 12 und 15 V DC liegen&bull;Mischermodul defekt.",                                                                                                                                                                                      correction => "Konfiguration &auml;ndern -&bull;Besch&auml;digte Kabel austauschen&bull;Mischermodul austauschen"                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1036"} = {errorcode => "A62",                                    subcode => "3061; 3062; 3063; 3064; 3065; 3066; 3067; 3068",     class => "14",          reason => "Keine Kommunikation mit Mischermodul : A62 = Heizkreis 2",                                                                                                                                                                                                                                                                                         checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); mit der gew&auml;hlten Einstellung ist ein Mischermodul erforderlich&bull;Die Verbindungsleitung EMS zum Mischermodul auf Besch&auml;digung pr&uuml;fen; Die Busspannung am Mischermodul muss zwischen 12 und 15 V DC liegen&bull;Mischermodul defekt.",                                                                                                                                                                                      correction => "Konfiguration &auml;ndern -&bull;Besch&auml;digte Kabel austauschen&bull;Mischermodul austauschen"                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1037"} = {errorcode => "A63",                                    subcode => "3061; 3062; 3063; 3064; 3065; 3066; 3067; 3068",     class => "14",          reason => "Keine Kommunikation mit Mischermodul : A63 = Heizkreis 3",                                                                                                                                                                                                                                                                                         checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); mit der gew&auml;hlten Einstellung ist ein Mischermodul erforderlich&bull;Die Verbindungsleitung EMS zum Mischermodul auf Besch&auml;digung pr&uuml;fen; Die Busspannung am Mischermodul muss zwischen 12 und 15 V DC liegen&bull;Mischermodul defekt.",                                                                                                                                                                                      correction => "Konfiguration &auml;ndern -&bull;Besch&auml;digte Kabel austauschen&bull;Mischermodul austauschen"                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1038"} = {errorcode => "A64",                                    subcode => "3061; 3062; 3063; 3064; 3065; 3066; 3067; 3068",     class => "14",          reason => "Keine Kommunikation mit Mischermodul : A64 = Heizkreis 4",                                                                                                                                                                                                                                                                                         checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); mit der gew&auml;hlten Einstellung ist ein Mischermodul erforderlich&bull;Die Verbindungsleitung EMS zum Mischermodul auf Besch&auml;digung pr&uuml;fen; Die Busspannung am Mischermodul muss zwischen 12 und 15 V DC liegen&bull;Mischermodul defekt.",                                                                                                                                                                                      correction => "Konfiguration &auml;ndern -&bull;Besch&auml;digte Kabel austauschen&bull;Mischermodul austauschen"                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1039"} = {errorcode => "A65",                                    subcode => "3061; 3062; 3063; 3064; 3065; 3066; 3067; 3068",     class => "14",          reason => "Keine Kommunikation mit Mischermodul : A65 = Heizkreis 5",                                                                                                                                                                                                                                                                                         checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); mit der gew&auml;hlten Einstellung ist ein Mischermodul erforderlich&bull;Die Verbindungsleitung EMS zum Mischermodul auf Besch&auml;digung pr&uuml;fen; Die Busspannung am Mischermodul muss zwischen 12 und 15 V DC liegen&bull;Mischermodul defekt.",                                                                                                                                                                                      correction => "Konfiguration &auml;ndern -&bull;Besch&auml;digte Kabel austauschen&bull;Mischermodul austauschen"                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1040"} = {errorcode => "A66",                                    subcode => "3061; 3062; 3063; 3064; 3065; 3066; 3067; 3068",     class => "14",          reason => "Keine Kommunikation mit Mischermodul : A66 = Heizkreis 6",                                                                                                                                                                                                                                                                                         checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); mit der gew&auml;hlten Einstellung ist ein Mischermodul erforderlich&bull;Die Verbindungsleitung EMS zum Mischermodul auf Besch&auml;digung pr&uuml;fen; Die Busspannung am Mischermodul muss zwischen 12 und 15 V DC liegen&bull;Mischermodul defekt.",                                                                                                                                                                                      correction => "Konfiguration &auml;ndern -&bull;Besch&auml;digte Kabel austauschen&bull;Mischermodul austauschen"                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1041"} = {errorcode => "A67",                                    subcode => "3061; 3062; 3063; 3064; 3065; 3066; 3067; 3068",     class => "14",          reason => "Keine Kommunikation mit Mischermodul : A67 = Heizkreis 7",                                                                                                                                                                                                                                                                                         checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); mit der gew&auml;hlten Einstellung ist ein Mischermodul erforderlich&bull;Die Verbindungsleitung EMS zum Mischermodul auf Besch&auml;digung pr&uuml;fen; Die Busspannung am Mischermodul muss zwischen 12 und 15 V DC liegen&bull;Mischermodul defekt.",                                                                                                                                                                                      correction => "Konfiguration &auml;ndern -&bull;Besch&auml;digte Kabel austauschen&bull;Mischermodul austauschen"                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1042"} = {errorcode => "A68",                                    subcode => "3061; 3062; 3063; 3064; 3065; 3066; 3067; 3068",     class => "14",          reason => "Keine Kommunikation mit Mischermodul : A68 = Heizkreis 8",                                                                                                                                                                                                                                                                                         checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); mit der gew&auml;hlten Einstellung ist ein Mischermodul erforderlich&bull;Die Verbindungsleitung EMS zum Mischermodul auf Besch&auml;digung pr&uuml;fen; Die Busspannung am Mischermodul muss zwischen 12 und 15 V DC liegen&bull;Mischermodul defekt.",                                                                                                                                                                                      correction => "Konfiguration &auml;ndern -&bull;Besch&auml;digte Kabel austauschen&bull;Mischermodul austauschen"                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1043"} = {errorcode => "A61",                                    subcode => "3071; 3072; 3073; 3074; 3075; 3076; 3077; 3078",     class => "8",           reason => "Keine Kommunikation mit Fernbedienung : A61 = Heizkreis 1",                                                                                                                                                                                                                                                                                        checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); mit der gew&auml;hlten Einstellung ist ein Mischermodul erforderlich&bull;Die Verbindungsleitung EMS zum Mischermodul auf Besch&auml;digung pr&uuml;fen; Die Busspannung am Mischermodul muss zwischen 12 und 15 V DC liegen&bull;Mischermodul defekt.",                                                                                                                                                                                      correction => "Konfiguration &auml;ndern -&bull;Besch&auml;digte Kabel austauschen&bull;Mischermodul austauschen"                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1044"} = {errorcode => "A62",                                    subcode => "3071; 3072; 3073; 3074; 3075; 3076; 3077; 3078",     class => "8",           reason => "Keine Kommunikation mit Fernbedienung : A62 = Heizkreis 2",                                                                                                                                                                                                                                                                                        checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); mit der gew&auml;hlten Einstellung ist ein Mischermodul erforderlich&bull;Die Verbindungsleitung EMS zum Mischermodul auf Besch&auml;digung pr&uuml;fen; Die Busspannung am Mischermodul muss zwischen 12 und 15 V DC liegen&bull;Mischermodul defekt.",                                                                                                                                                                                      correction => "Konfiguration &auml;ndern -&bull;Besch&auml;digte Kabel austauschen&bull;Mischermodul austauschen"                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1045"} = {errorcode => "A63",                                    subcode => "3071; 3072; 3073; 3074; 3075; 3076; 3077; 3078",     class => "8",           reason => "Keine Kommunikation mit Fernbedienung : A63 = Heizkreis 3",                                                                                                                                                                                                                                                                                        checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); mit der gew&auml;hlten Einstellung ist ein Mischermodul erforderlich&bull;Die Verbindungsleitung EMS zum Mischermodul auf Besch&auml;digung pr&uuml;fen; Die Busspannung am Mischermodul muss zwischen 12 und 15 V DC liegen&bull;Mischermodul defekt.",                                                                                                                                                                                      correction => "Konfiguration &auml;ndern -&bull;Besch&auml;digte Kabel austauschen&bull;Mischermodul austauschen"                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1046"} = {errorcode => "A64",                                    subcode => "3071; 3072; 3073; 3074; 3075; 3076; 3077; 3078",     class => "8",           reason => "Keine Kommunikation mit Fernbedienung : A64 = Heizkreis 4",                                                                                                                                                                                                                                                                                        checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); mit der gew&auml;hlten Einstellung ist ein Mischermodul erforderlich&bull;Die Verbindungsleitung EMS zum Mischermodul auf Besch&auml;digung pr&uuml;fen; Die Busspannung am Mischermodul muss zwischen 12 und 15 V DC liegen&bull;Mischermodul defekt.",                                                                                                                                                                                      correction => "Konfiguration &auml;ndern -&bull;Besch&auml;digte Kabel austauschen&bull;Mischermodul austauschen"                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1047"} = {errorcode => "A65",                                    subcode => "3071; 3072; 3073; 3074; 3075; 3076; 3077; 3078",     class => "8",           reason => "Keine Kommunikation mit Fernbedienung : A65 = Heizkreis 5",                                                                                                                                                                                                                                                                                        checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); mit der gew&auml;hlten Einstellung ist ein Mischermodul erforderlich&bull;Die Verbindungsleitung EMS zum Mischermodul auf Besch&auml;digung pr&uuml;fen; Die Busspannung am Mischermodul muss zwischen 12 und 15 V DC liegen&bull;Mischermodul defekt.",                                                                                                                                                                                      correction => "Konfiguration &auml;ndern -&bull;Besch&auml;digte Kabel austauschen&bull;Mischermodul austauschen"                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1048"} = {errorcode => "A66",                                    subcode => "3071; 3072; 3073; 3074; 3075; 3076; 3077; 3078",     class => "8",           reason => "Keine Kommunikation mit Fernbedienung : A66 = Heizkreis 6",                                                                                                                                                                                                                                                                                        checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); mit der gew&auml;hlten Einstellung ist ein Mischermodul erforderlich&bull;Die Verbindungsleitung EMS zum Mischermodul auf Besch&auml;digung pr&uuml;fen; Die Busspannung am Mischermodul muss zwischen 12 und 15 V DC liegen&bull;Mischermodul defekt.",                                                                                                                                                                                      correction => "Konfiguration &auml;ndern -&bull;Besch&auml;digte Kabel austauschen&bull;Mischermodul austauschen"                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1049"} = {errorcode => "A67",                                    subcode => "3071; 3072; 3073; 3074; 3075; 3076; 3077; 3078",     class => "8",           reason => "Keine Kommunikation mit Fernbedienung : A67 = Heizkreis 7",                                                                                                                                                                                                                                                                                        checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); mit der gew&auml;hlten Einstellung ist ein Mischermodul erforderlich&bull;Die Verbindungsleitung EMS zum Mischermodul auf Besch&auml;digung pr&uuml;fen; Die Busspannung am Mischermodul muss zwischen 12 und 15 V DC liegen&bull;Mischermodul defekt.",                                                                                                                                                                                      correction => "Konfiguration &auml;ndern -&bull;Besch&auml;digte Kabel austauschen&bull;Mischermodul austauschen"                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1050"} = {errorcode => "A68",                                    subcode => "3071; 3072; 3073; 3074; 3075; 3076; 3077; 3078",     class => "8",           reason => "Keine Kommunikation mit Fernbedienung : A68 = Heizkreis 8",                                                                                                                                                                                                                                                                                        checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); mit der gew&auml;hlten Einstellung ist ein Mischermodul erforderlich&bull;Die Verbindungsleitung EMS zum Mischermodul auf Besch&auml;digung pr&uuml;fen; Die Busspannung am Mischermodul muss zwischen 12 und 15 V DC liegen&bull;Mischermodul defekt.",                                                                                                                                                                                      correction => "Konfiguration &auml;ndern -&bull;Besch&auml;digte Kabel austauschen&bull;Mischermodul austauschen"                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1051"} = {errorcode => "A61",                                    subcode => "3091; 3092; 3093; 3094; 3095; 3096; 3097; 3098",     class => "14",          reason => "Raumtemperaturf&uuml;hler defekt : A61 = Heizkreis 1",                                                                                                                                                                                                                                                                                             checklist => "Systemregler oder Fernbedienung defekt",                                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Autokonfiguration erneut starten; Alle Teilnehmer m&uuml;ssen am EMS BUS sein&bull;Systemregler oder Fernbedienung austauschen"                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1052"} = {errorcode => "A62",                                    subcode => "3091; 3092; 3093; 3094; 3095; 3096; 3097; 3098",     class => "14",          reason => "Raumtemperaturf&uuml;hler defekt : A62 = Heizkreis 2",                                                                                                                                                                                                                                                                                             checklist => "Systemregler oder Fernbedienung defekt",                                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Autokonfiguration erneut starten; Alle Teilnehmer m&uuml;ssen am EMS BUS sein&bull;Systemregler oder Fernbedienung austauschen"                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1053"} = {errorcode => "A63",                                    subcode => "3091; 3092; 3093; 3094; 3095; 3096; 3097; 3098",     class => "14",          reason => "Raumtemperaturf&uuml;hler defekt : A63 = Heizkreis 3",                                                                                                                                                                                                                                                                                             checklist => "Systemregler oder Fernbedienung defekt",                                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Autokonfiguration erneut starten; Alle Teilnehmer m&uuml;ssen am EMS BUS sein&bull;Systemregler oder Fernbedienung austauschen"                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1054"} = {errorcode => "A64",                                    subcode => "3091; 3092; 3093; 3094; 3095; 3096; 3097; 3098",     class => "14",          reason => "Raumtemperaturf&uuml;hler defekt : A64 = Heizkreis 4",                                                                                                                                                                                                                                                                                             checklist => "Systemregler oder Fernbedienung defekt",                                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Autokonfiguration erneut starten; Alle Teilnehmer m&uuml;ssen am EMS BUS sein&bull;Systemregler oder Fernbedienung austauschen"                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1055"} = {errorcode => "A65",                                    subcode => "3091; 3092; 3093; 3094; 3095; 3096; 3097; 3098",     class => "14",          reason => "Raumtemperaturf&uuml;hler defekt : A65 = Heizkreis 5",                                                                                                                                                                                                                                                                                             checklist => "Systemregler oder Fernbedienung defekt",                                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Autokonfiguration erneut starten; Alle Teilnehmer m&uuml;ssen am EMS BUS sein&bull;Systemregler oder Fernbedienung austauschen"                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1056"} = {errorcode => "A66",                                    subcode => "3091; 3092; 3093; 3094; 3095; 3096; 3097; 3098",     class => "14",          reason => "Raumtemperaturf&uuml;hler defekt : A66 = Heizkreis 6",                                                                                                                                                                                                                                                                                             checklist => "Systemregler oder Fernbedienung defekt",                                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Autokonfiguration erneut starten; Alle Teilnehmer m&uuml;ssen am EMS BUS sein&bull;Systemregler oder Fernbedienung austauschen"                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1057"} = {errorcode => "A67",                                    subcode => "3091; 3092; 3093; 3094; 3095; 3096; 3097; 3098",     class => "14",          reason => "Raumtemperaturf&uuml;hler defekt : A67 = Heizkreis 7",                                                                                                                                                                                                                                                                                             checklist => "Systemregler oder Fernbedienung defekt",                                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Autokonfiguration erneut starten; Alle Teilnehmer m&uuml;ssen am EMS BUS sein&bull;Systemregler oder Fernbedienung austauschen"                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1058"} = {errorcode => "A68",                                    subcode => "3091; 3092; 3093; 3094; 3095; 3096; 3097; 3098",     class => "14",          reason => "Raumtemperaturf&uuml;hler defekt : A68 = Heizkreis 8",                                                                                                                                                                                                                                                                                             checklist => "Systemregler oder Fernbedienung defekt",                                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Autokonfiguration erneut starten; Alle Teilnehmer m&uuml;ssen am EMS BUS sein&bull;Systemregler oder Fernbedienung austauschen"                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1059"} = {errorcode => "A61",                                    subcode => "3161; 3162; 3163; 3164; 3165; 3166; 3167; 3168",     class => "8",           reason => "Batterie Bedieneinheit schwach : A61 = Heizkreis 1",                                                                                                                                                                                                                                                                                               checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1060"} = {errorcode => "A62",                                    subcode => "3161; 3162; 3163; 3164; 3165; 3166; 3167; 3168",     class => "8",           reason => "Batterie Bedieneinheit schwach : A62 = Heizkreis 2",                                                                                                                                                                                                                                                                                               checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1061"} = {errorcode => "A63",                                    subcode => "3161; 3162; 3163; 3164; 3165; 3166; 3167; 3168",     class => "8",           reason => "Batterie Bedieneinheit schwach : A63 = Heizkreis 3",                                                                                                                                                                                                                                                                                               checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1062"} = {errorcode => "A64",                                    subcode => "3161; 3162; 3163; 3164; 3165; 3166; 3167; 3168",     class => "8",           reason => "Batterie Bedieneinheit schwach : A64 = Heizkreis 4",                                                                                                                                                                                                                                                                                               checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1063"} = {errorcode => "A65",                                    subcode => "3161; 3162; 3163; 3164; 3165; 3166; 3167; 3168",     class => "8",           reason => "Batterie Bedieneinheit schwach : A65 = Heizkreis 5",                                                                                                                                                                                                                                                                                               checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1064"} = {errorcode => "A66",                                    subcode => "3161; 3162; 3163; 3164; 3165; 3166; 3167; 3168",     class => "8",           reason => "Batterie Bedieneinheit schwach : A66 = Heizkreis 6",                                                                                                                                                                                                                                                                                               checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1065"} = {errorcode => "A67",                                    subcode => "3161; 3162; 3163; 3164; 3165; 3166; 3167; 3168",     class => "8",           reason => "Batterie Bedieneinheit schwach : A67 = Heizkreis 7",                                                                                                                                                                                                                                                                                               checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1066"} = {errorcode => "A68",                                    subcode => "3161; 3162; 3163; 3164; 3165; 3166; 3167; 3168",     class => "8",           reason => "Batterie Bedieneinheit schwach : A68 = Heizkreis 8",                                                                                                                                                                                                                                                                                               checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1067"} = {errorcode => "A61",                                    subcode => "6001",                                               class => "16",          reason => "Konfigurationsfehler: Solarmodul nicht verwendet : A61 = Heizkreis 1",                                                                                                                                                                                                                                                                             checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); Es ist ein Solarmodul im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird",                                                                                                                                                                                                                                                                                                                                                correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1068"} = {errorcode => "A62",                                    subcode => "6001",                                               class => "16",          reason => "Konfigurationsfehler: Solarmodul nicht verwendet : A62 = Heizkreis 2",                                                                                                                                                                                                                                                                             checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); Es ist ein Solarmodul im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird",                                                                                                                                                                                                                                                                                                                                                correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1069"} = {errorcode => "A63",                                    subcode => "6001",                                               class => "16",          reason => "Konfigurationsfehler: Solarmodul nicht verwendet : A63 = Heizkreis 3",                                                                                                                                                                                                                                                                             checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); Es ist ein Solarmodul im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird",                                                                                                                                                                                                                                                                                                                                                correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1070"} = {errorcode => "A64",                                    subcode => "6001",                                               class => "16",          reason => "Konfigurationsfehler: Solarmodul nicht verwendet : A64 = Heizkreis 4",                                                                                                                                                                                                                                                                             checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); Es ist ein Solarmodul im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird",                                                                                                                                                                                                                                                                                                                                                correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1071"} = {errorcode => "A65",                                    subcode => "6001",                                               class => "16",          reason => "Konfigurationsfehler: Solarmodul nicht verwendet : A65 = Heizkreis 5",                                                                                                                                                                                                                                                                             checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); Es ist ein Solarmodul im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird",                                                                                                                                                                                                                                                                                                                                                correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1072"} = {errorcode => "A66",                                    subcode => "6001",                                               class => "16",          reason => "Konfigurationsfehler: Solarmodul nicht verwendet : A66 = Heizkreis 6",                                                                                                                                                                                                                                                                             checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); Es ist ein Solarmodul im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird",                                                                                                                                                                                                                                                                                                                                                correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1073"} = {errorcode => "A67",                                    subcode => "6001",                                               class => "16",          reason => "Konfigurationsfehler: Solarmodul nicht verwendet : A67 = Heizkreis 7",                                                                                                                                                                                                                                                                             checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); Es ist ein Solarmodul im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird",                                                                                                                                                                                                                                                                                                                                                correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1074"} = {errorcode => "A68",                                    subcode => "6001",                                               class => "16",          reason => "Konfigurationsfehler: Solarmodul nicht verwendet : A68 = Heizkreis 8",                                                                                                                                                                                                                                                                             checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul); Es ist ein Solarmodul im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird",                                                                                                                                                                                                                                                                                                                                                correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1075"} = {errorcode => "A7",                                     subcode => "1",                                                  class => "",            reason => "St&ouml;rung: WarmwasserTemperaturf&uuml;hler",                                                                                                                                                                                                                                                                                                    checklist => "1. Kein Signal vom WarmwasserTemperaturf&uuml;hler vorhanden&bull;Verbindungskabel zum WarmwasserTemperaturf&uuml;hler ist besch&auml;digt&bull;Warmwasser-Temperaturf&uuml;hler defekt&bull;Signalauswertung auf Leiterplatte defekt",                                                                                                                                                                                                                                                               correction => "1. Stecker am WarmwasserTemperaturf&uuml;hler aufstecken&bull;Kabelbaum austauschen&bull;Warmwasser-Temperaturf&uuml;hler austauschen&bull;Leiterplatte austauschen"                                                                                                                                                                                                                             };
	$ErrorCodeList{"1076"} = {errorcode => "A7",                                     subcode => "9; 10",                                              class => "",            reason => "St&ouml;rung: SpeicherladeTemperaturf&uuml;hler",                                                                                                                                                                                                                                                                                                  checklist => "Stecker f&uuml;r Speichertemperaturf&uuml;hler 2 ist nicht aufgesteckt&bull;Anschlu&szlig;kabel zum Speicherladef&uuml;hler ist besch&auml;digt&bull;Speicherf&uuml;hler 2 defekt&bull;Signalauswertung auf Leiterplatte defekt",                                                                                                                                                                                                                                                                     correction => "Stecker f&uuml;r Speichertemperaturf&uuml;hler 2 ist nicht aufgesteckt&bull;Anschlu&szlig;kabel zum Speicherladef&uuml;hler ist besch&auml;digt&bull;Speicherf&uuml;hler 2 defekt&bull;Signalauswertung auf Leiterplatte defekt"                                                                                                                                                                 };
	$ErrorCodeList{"1077"} = {errorcode => "A7",                                     subcode => "28",                                                 class => "",            reason => "Temperaturf&uuml;hler am Brenner defekt",                                                                                                                                                                                                                                                                                                          checklist => "Kein Signal vom Vorlauftemperaturf&uuml;hler am Brenner&bull;Signal vom Brenner-Vorlauftemperaturf&uuml;hler au&szlig;erhalb Kennlinie&bull;Signalunterbrechung im Anschlusskabel&bull;Signalauswertung auf Leiterplatte defekt",                                                                                                                                                                                                                                                                     correction => "Vorlauftemperaturf&uuml;hler austauschen&bull;Kabelbaum austauschen&bull;Leiterplatte austauschen&bull;Stecker am Vorlauftemperaturf&uuml;hler komplett aufstecken"                                                                                                                                                                                                                              };
	$ErrorCodeList{"1078"} = {errorcode => "A71",                                    subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus : A61 = Heizkreis 1",                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder einschalten Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung&bull;",                                                                                                                                                                                                                                      correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t ausund wieder einschalten&bull;Busleitung reparieren bzw. austauschen&bull;Defekten EMS-BUS-Teilnehmer austauschen"                                                                                                                                                                                                                            };
	$ErrorCodeList{"1079"} = {errorcode => "A72",                                    subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus : A62 = Heizkreis 2",                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder einschalten Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung&bull;",                                                                                                                                                                                                                                      correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t ausund wieder einschalten&bull;Busleitung reparieren bzw. austauschen&bull;Defekten EMS-BUS-Teilnehmer austauschen"                                                                                                                                                                                                                            };
	$ErrorCodeList{"1080"} = {errorcode => "A73",                                    subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus : A63 = Heizkreis 3",                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder einschalten Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung&bull;",                                                                                                                                                                                                                                      correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t ausund wieder einschalten&bull;Busleitung reparieren bzw. austauschen&bull;Defekten EMS-BUS-Teilnehmer austauschen"                                                                                                                                                                                                                            };
	$ErrorCodeList{"1081"} = {errorcode => "A74",                                    subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus : A64 = Heizkreis 4",                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder einschalten Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung&bull;",                                                                                                                                                                                                                                      correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t ausund wieder einschalten&bull;Busleitung reparieren bzw. austauschen&bull;Defekten EMS-BUS-Teilnehmer austauschen"                                                                                                                                                                                                                            };
	$ErrorCodeList{"1082"} = {errorcode => "A75",                                    subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus : A65 = Heizkreis 5",                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder einschalten Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung&bull;",                                                                                                                                                                                                                                      correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t ausund wieder einschalten&bull;Busleitung reparieren bzw. austauschen&bull;Defekten EMS-BUS-Teilnehmer austauschen"                                                                                                                                                                                                                            };
	$ErrorCodeList{"1083"} = {errorcode => "A76",                                    subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus : A66 = Heizkreis 6",                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder einschalten Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung&bull;",                                                                                                                                                                                                                                      correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t ausund wieder einschalten&bull;Busleitung reparieren bzw. austauschen&bull;Defekten EMS-BUS-Teilnehmer austauschen"                                                                                                                                                                                                                            };
	$ErrorCodeList{"1084"} = {errorcode => "A77",                                    subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus : A67 = Heizkreis 7",                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder einschalten Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung&bull;",                                                                                                                                                                                                                                      correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t ausund wieder einschalten&bull;Busleitung reparieren bzw. austauschen&bull;Defekten EMS-BUS-Teilnehmer austauschen"                                                                                                                                                                                                                            };
	$ErrorCodeList{"1085"} = {errorcode => "A78",                                    subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus : A68 = Heizkreis 8",                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder einschalten Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung&bull;",                                                                                                                                                                                                                                      correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t ausund wieder einschalten&bull;Busleitung reparieren bzw. austauschen&bull;Defekten EMS-BUS-Teilnehmer austauschen"                                                                                                                                                                                                                            };
	$ErrorCodeList{"1086"} = {errorcode => "A71",                                    subcode => "1030; 1035; 1036",                                   class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen : A61 = Heizkreis 1",                                                                                                                                                                                                                                                                            checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"1087"} = {errorcode => "A72",                                    subcode => "1030; 1035; 1036",                                   class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen : A62 = Heizkreis 2",                                                                                                                                                                                                                                                                            checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"1088"} = {errorcode => "A73",                                    subcode => "1030; 1035; 1036",                                   class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen : A63 = Heizkreis 3",                                                                                                                                                                                                                                                                            checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"1089"} = {errorcode => "A74",                                    subcode => "1030; 1035; 1036",                                   class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen : A64 = Heizkreis 4",                                                                                                                                                                                                                                                                            checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"1090"} = {errorcode => "A75",                                    subcode => "1030; 1035; 1036",                                   class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen : A65 = Heizkreis 5",                                                                                                                                                                                                                                                                            checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"1091"} = {errorcode => "A76",                                    subcode => "1030; 1035; 1036",                                   class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen : A66 = Heizkreis 6",                                                                                                                                                                                                                                                                            checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"1092"} = {errorcode => "A77",                                    subcode => "1030; 1035; 1036",                                   class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen : A67 = Heizkreis 7",                                                                                                                                                                                                                                                                            checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"1093"} = {errorcode => "A78",                                    subcode => "1030; 1035; 1036",                                   class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen : A68 = Heizkreis 8",                                                                                                                                                                                                                                                                            checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"1094"} = {errorcode => "A71",                                    subcode => "1038",                                               class => "16",          reason => "Datum/Zeit ung&uuml;ltiger Wert (A71 = Heizkreis 1;; A72 = Heizkreis 2; A73 = Heizkreis 3; A74 = Heizkreis 4; A75 = Heizkreis 5; A76 = Heizkreis 6; A77 = Heizkreis 7; A78 = Heizkreis 8)",                                                                                                                                                        checklist => "Datum/Zeit noch nicht eingestellt&bull;Spannungsversorgung &uuml;ber l&auml;ngere Zeit ausgefallen",                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Datum/Zeit einstellen&bull;Spannungsausf&auml;lle vermeiden"                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"1095"} = {errorcode => "A72",                                    subcode => "1038",                                               class => "16",          reason => "Datum/Zeit ung&uuml;ltiger Wert (A71 = Heizkreis 1;; A72 = Heizkreis 2; A73 = Heizkreis 3; A74 = Heizkreis 4; A75 = Heizkreis 5; A76 = Heizkreis 6; A77 = Heizkreis 7; A78 = Heizkreis 8)",                                                                                                                                                        checklist => "Datum/Zeit noch nicht eingestellt&bull;Spannungsversorgung &uuml;ber l&auml;ngere Zeit ausgefallen",                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Datum/Zeit einstellen&bull;Spannungsausf&auml;lle vermeiden"                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"1096"} = {errorcode => "A73",                                    subcode => "1038",                                               class => "16",          reason => "Datum/Zeit ung&uuml;ltiger Wert (A71 = Heizkreis 1;; A72 = Heizkreis 2; A73 = Heizkreis 3; A74 = Heizkreis 4; A75 = Heizkreis 5; A76 = Heizkreis 6; A77 = Heizkreis 7; A78 = Heizkreis 8)",                                                                                                                                                        checklist => "Datum/Zeit noch nicht eingestellt&bull;Spannungsversorgung &uuml;ber l&auml;ngere Zeit ausgefallen",                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Datum/Zeit einstellen&bull;Spannungsausf&auml;lle vermeiden"                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"1097"} = {errorcode => "A74",                                    subcode => "1038",                                               class => "16",          reason => "Datum/Zeit ung&uuml;ltiger Wert (A71 = Heizkreis 1;; A72 = Heizkreis 2; A73 = Heizkreis 3; A74 = Heizkreis 4; A75 = Heizkreis 5; A76 = Heizkreis 6; A77 = Heizkreis 7; A78 = Heizkreis 8)",                                                                                                                                                        checklist => "Datum/Zeit noch nicht eingestellt&bull;Spannungsversorgung &uuml;ber l&auml;ngere Zeit ausgefallen",                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Datum/Zeit einstellen&bull;Spannungsausf&auml;lle vermeiden"                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"1098"} = {errorcode => "A75",                                    subcode => "1038",                                               class => "16",          reason => "Datum/Zeit ung&uuml;ltiger Wert (A71 = Heizkreis 1;; A72 = Heizkreis 2; A73 = Heizkreis 3; A74 = Heizkreis 4; A75 = Heizkreis 5; A76 = Heizkreis 6; A77 = Heizkreis 7; A78 = Heizkreis 8)",                                                                                                                                                        checklist => "Datum/Zeit noch nicht eingestellt&bull;Spannungsversorgung &uuml;ber l&auml;ngere Zeit ausgefallen",                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Datum/Zeit einstellen&bull;Spannungsausf&auml;lle vermeiden"                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"1099"} = {errorcode => "A76",                                    subcode => "1038",                                               class => "16",          reason => "Datum/Zeit ung&uuml;ltiger Wert (A71 = Heizkreis 1;; A72 = Heizkreis 2; A73 = Heizkreis 3; A74 = Heizkreis 4; A75 = Heizkreis 5; A76 = Heizkreis 6; A77 = Heizkreis 7; A78 = Heizkreis 8)",                                                                                                                                                        checklist => "Datum/Zeit noch nicht eingestellt&bull;Spannungsversorgung &uuml;ber l&auml;ngere Zeit ausgefallen",                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Datum/Zeit einstellen&bull;Spannungsausf&auml;lle vermeiden"                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"1100"} = {errorcode => "A77",                                    subcode => "1038",                                               class => "16",          reason => "Datum/Zeit ung&uuml;ltiger Wert (A71 = Heizkreis 1;; A72 = Heizkreis 2; A73 = Heizkreis 3; A74 = Heizkreis 4; A75 = Heizkreis 5; A76 = Heizkreis 6; A77 = Heizkreis 7; A78 = Heizkreis 8)",                                                                                                                                                        checklist => "Datum/Zeit noch nicht eingestellt&bull;Spannungsversorgung &uuml;ber l&auml;ngere Zeit ausgefallen",                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Datum/Zeit einstellen&bull;Spannungsausf&auml;lle vermeiden"                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"1101"} = {errorcode => "A78",                                    subcode => "1038",                                               class => "16",          reason => "Datum/Zeit ung&uuml;ltiger Wert (A71 = Heizkreis 1;; A72 = Heizkreis 2; A73 = Heizkreis 3; A74 = Heizkreis 4; A75 = Heizkreis 5; A76 = Heizkreis 6; A77 = Heizkreis 7; A78 = Heizkreis 8)",                                                                                                                                                        checklist => "Datum/Zeit noch nicht eingestellt&bull;Spannungsversorgung &uuml;ber l&auml;ngere Zeit ausgefallen",                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Datum/Zeit einstellen&bull;Spannungsausf&auml;lle vermeiden"                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"1102"} = {errorcode => "A8",                                     subcode => "46",                                                 class => "8",           reason => "Falsche BUS-Konfiguration",                                                                                                                                                                                                                                                                                                                        checklist => "BUS-Modul und 2-Draht-BUS Regler sind beide vorhanden&bull;BUS-Modul ausbauen",                                                                                                                                                                                                                                                                                                                                                                                                                       correction => "Im Regelger&auml;t Bus Regler abklemmen&bull;BUS-Modul ausbauen"                                                                                                                                                                                                                                                                                                                                 };
	$ErrorCodeList{"1103"} = {errorcode => "A8",                                     subcode => "61",                                                 class => "",            reason => "St&ouml;rung in Buskommunikation",                                                                                                                                                                                                                                                                                                                 checklist => "Verbindungskabel zwischen CANBusregler und BUS-Modul unterbrochen&bull;Stecker am BUS-Modul nicht korrekt eingerastet",                                                                                                                                                                                                                                                                                                                                                                               correction => "Verbindungskabel zwischen CAN-Busregler; und BUS-Modul instandsetzen bzw. austauschen&bull;Stecker vom BUS-Modul abziehen und neu einstecken"                                                                                                                                                                                                                                                    };
	$ErrorCodeList{"1104"} = {errorcode => "A8",                                     subcode => "310",                                                class => "B",           reason => "Keine Kommunikation mit dem EMS W&auml;rmeerzeuger",                                                                                                                                                                                                                                                                                               checklist => "Es kann keine Kommunikation zwischen Hybridmodul (Inneneinheit) und dem EMS W&auml;rmeerzeuger hergestellt werden&bull;Kontakte, Kabel und Verbindungen zwischen Hybridmodul und EMS W&auml;rmeerzeuger pr&uuml;fen",                                                                                                                                                                                                                                                                                 correction => "EMS-Verbindung zwischen Hybridmodul (Inneneinheit) und dem EMS W&auml;rmeerzeuger &uuml;berpr&uuml;fen&bull;Besch&auml;digte Leitungen und Kontakte austauschen, ggf. Ger&auml;teelektronik austauschen"                                                                                                                                                                                         };
	$ErrorCodeList{"1105"} = {errorcode => "A8",                                     subcode => "323",                                                class => "",            reason => "Falsche BUS-Konfiguration",                                                                                                                                                                                                                                                                                                                        checklist => "BUS-Modul und 2-Draht-BUS Regler sind beide vorhanden BUS-Modul ausbauen",                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "Heatronic 3® Bus Regler abklemmen BUS-Modul ausbauen"                                                                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"1106"} = {errorcode => "A8",                                     subcode => "323",                                                class => "",            reason => "St&ouml;rung in Buskommunikation",                                                                                                                                                                                                                                                                                                                 checklist => "Verbindungskabel zwischen CAN-Busregler und BUS-Modul unterbrochen Stecker am BUSModul nicht korrekt eingerastet",                                                                                                                                                                                                                                                                                                                                                                                    correction => "Verbindungskabel zwischen CAN-Busregler und BUS-Modul instandsetzen bzw. austauschen; Stecker vom BUS-Modul abziehen und neu einstecken"                                                                                                                                                                                                                                                         };
	$ErrorCodeList{"1107"} = {errorcode => "A8",                                     subcode => "470",                                                class => "10",          reason => "Keine Kommunikation mit dem Systemregler",                                                                                                                                                                                                                                                                                                         checklist => "Kontakte, Kabel und Verbindungen zwischen Hybridmodul und Systemregler pr&uuml;fen&bull;Systemregler pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                    correction => "Besch&auml;digte Leitungen und Kontakte austauschen&bull;Ggf. austauschen"                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1108"} = {errorcode => "A9",                                     subcode => "66",                                                 class => "",            reason => "SchichtladespeicherTemperaturf&uuml;hler 3 St&ouml;rung",                                                                                                                                                                                                                                                                                          checklist => "Luft im Warmwasserkreis&bull;Anschlu&szlig;stecker an der Schichtladepumpe nicht komplett aufgesteckt&bull;Schichtladepumpe um 180° verdreht eingebaut&bull;Schichtladepumpe blockiert&bull;Plattenw&auml;rmetauscher verschmutzt oder verkalkt",                                                                                                                                                                                                                                                    correction => "Warmwasserkreis entl&uuml;ften&bull;Stecker f&uuml;r die Schichtladepumpe komplett aufstecken und verrasten&bull;Pumpe in Flussrichtung einbauen&bull;Pumpe austauschen&bull;Plattenwarmetauscher austauschen"                                                                                                                                                                                   };
	$ErrorCodeList{"1109"} = {errorcode => "A91",                                    subcode => "1001",                                               class => "8",           reason => "Keine Kommunikation zwischen Systemregler und Fernbedienung",                                                                                                                                                                                                                                                                                      checklist => "Konfiguration pr&uuml;fen (Adresseinstellung).; Mit der gew&auml;hlten Einstellung ist ein Systemregler notwendig.&bull;Die Verbindungsleitung EMS zum Systemregler auf Besch&auml;digung pr&uuml;fen. Busspannung am Systemregler muss zwischen 12-15V DC liegen.&bull;Fernbedienung oder Systemregler defekt",                                                                                                                                                                                      correction => "Konfiguration &auml;ndern&bull;Besch&auml;digte Kabel austauschen&bull;Fernbedienung oder Systemregler austauschen"                                                                                                                                                                                                                                                                              };
	$ErrorCodeList{"1110"} = {errorcode => "A91",                                    subcode => "1009",                                               class => "8",           reason => "Keine Teilnehmer im Funknetz erkannt",                                                                                                                                                                                                                                                                                                             checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "Funk-Teilnehmer einlernen"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1111"} = {errorcode => "A91",                                    subcode => "1010",                                               class => "8",           reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS",                                                                                                                                                                                                                                                                                                 checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist. Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder einschalten. Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung.&bull;",                                                                                                                                                                                                                                   correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t Ausund wieder Einschalten&bull;Busleitung reparieren bzw. austauschen.&bull;Defekten EMS-BUS-Teilnehmer austauschen"                                                                                                                                                                                                                           };
	$ErrorCodeList{"1112"} = {errorcode => "A91",                                    subcode => "1030; 1033; 1034; 1035; 1036",                       class => "8",           reason => "Interner Fehler Funkzentrale, Ger&auml;t tauschen",                                                                                                                                                                                                                                                                                                checklist => "Nur f&uuml;r interne Qualit&auml;tskontrolle. Fertigungsfehler/Bauteilefehler",                                                                                                                                                                                                                                                                                                                                                                                                                       correction => "Nur f&uuml;r interne Qualit&auml;tskontrolle. Fertigungsfehler/Bauteilefehler."                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1113"} = {errorcode => "A91",                                    subcode => "1042",                                               class => "8",           reason => "Interner Fehler Funkzentrale: Zugriff auf Uhrenbaustein blockiert",                                                                                                                                                                                                                                                                                checklist => "",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1114"} = {errorcode => "A91",                                    subcode => "1080",                                               class => "8",           reason => "Zwei Master Bedieneinheiten im System",                                                                                                                                                                                                                                                                                                            checklist => "In der Installationsebene die Parametrierung pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "Konfiguration Bedieneinheiten pr&uuml;fen: nur eine  als Master anmelden, weitere Bedieneinheiten als Fernbedienung anmelden."                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"1115"} = {errorcode => "A91",                                    subcode => "1101; 1102; 1103; 1104; 1105; 1106; 1107; 1108",     class => "8",           reason => "Heizk&ouml;rperventil 1-8: Batterie schwach",                                                                                                                                                                                                                                                                                                      checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Batterie tauschen"                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1116"} = {errorcode => "A91",                                    subcode => "1110",                                               class => "8",           reason => "Kommunikation mit mindestens einem Heizk&ouml;rperventil unterbrochen.",                                                                                                                                                                                                                                                                           checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Batterie tauschen"                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1117"} = {errorcode => "A91",                                    subcode => "1111; 1112; 1113; 1114; 1115; 1116; 1117; 1118",     class => "8",           reason => "Heizk&ouml;rperventil 1-8: Batterie zu schwach",                                                                                                                                                                                                                                                                                                   checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Batterie tauschen"                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1118"} = {errorcode => "A91",                                    subcode => "1121; 1122; 1123; 1124; 1125; 1126; 1127; 1128",     class => "4",           reason => "Heizk&ouml;rperventil 1-8: Funksignal schwach",                                                                                                                                                                                                                                                                                                    checklist => "Keine/fehlerhafte Kommunikation mit RF-Zentrale&bull;&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Funkempfangsst&auml;rke an der Funkkomponente pr&uuml;fen&bull;Entfernung der Funkkomponenten zueinander reduzieren&bull;Abschirmung minimieren (Anzahl der W&auml;nde, Stahlbetondecken)&bull;Winkel der Funkkomponenten zueinander optimieren (optimal: „senkrecht“ zu W&auml;nden und Decken)"                                                                                            };
	$ErrorCodeList{"1119"} = {errorcode => "A91",                                    subcode => "1131; 1132; 1133; 1134; 1135; 1136; 1137; 1138",     class => "8",           reason => "Keine Adaption am FunkThermostat-Ventil 1-8 m&ouml;glich",                                                                                                                                                                                                                                                                                         checklist => "Oberteil am Ventil 1-8 ist nicht auf das Ventilunterteil montiert",                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Oberteil auf Ventil 1-8 montieren"                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1120"} = {errorcode => "A91",                                    subcode => "1141; 1142; 1143; 1144; 1145; 1146; 1147; 1148",     class => "8",           reason => "Zu geringer Ventilhub am Funk-Thermostat-Ventil 1-8",                                                                                                                                                                                                                                                                                              checklist => "Falschen Adapter am Unterteil vom Ventil 1-8 montiert.&bull;Batterie zu schwach am Ventil 1-8",                                                                                                                                                                                                                                                                                                                                                                                                       correction => "Korrekten Adapter am Ventil 1-8 montieren&bull;Batterien am Ventil 1-8 austauschen"                                                                                                                                                                                                                                                                                                              };
	$ErrorCodeList{"1121"} = {errorcode => "A91",                                    subcode => "1151; 1152; 1153; 1154; 1155; 1156; 1157; 1158",     class => "8",           reason => "Keine Ventilbewegung am Funk-Thermostat-Ventil 1-8 m&ouml;glich",                                                                                                                                                                                                                                                                                  checklist => "Ventilstift am Ventil 1-8 blockiert",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "Blockierung am Ventilstift beim Ventil 1-8 beseitigen bzw. austauschen"                                                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"1122"} = {errorcode => "A91",                                    subcode => "1160",                                               class => "8",           reason => "Drahtloser Au&szlig;entemperaturf&uuml;hler: Batterie schwach",                                                                                                                                                                                                                                                                                    checklist => "Solarzelle am Funk-Au&szlig;enf&uuml;hler &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                      correction => "Abdeckung oder Verschmutzung Solarzelle am Funk-Au&szlig;enf&uuml;hler beseitigen"                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1123"} = {errorcode => "A91",                                    subcode => "1161",                                               class => "8",           reason => "Drahtloser Au&szlig;entemperaturf&uuml;hler: Funksignal schwach",                                                                                                                                                                                                                                                                                  checklist => "Signalst&auml;rke pr&uuml;fen&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "Funk-Repeater erforderlich&bull;Falls vorhanden: Funktion Repeater &uuml;berpr&uuml;fen"                                                                                                                                                                                                                                                                                                         };
	$ErrorCodeList{"1124"} = {errorcode => "A91",                                    subcode => "3011",                                               class => "8",           reason => "Konfigurationsfehler: Mischermodul nicht verwendet",                                                                                                                                                                                                                                                                                               checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul). Es ist ein Mischermodul im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird.",                                                                                                                                                                                                                                                                                                                                             correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1125"} = {errorcode => "A91",                                    subcode => "6001",                                               class => "8",           reason => "Konfigurationsfehler: Solarmodul nicht; verwendet",                                                                                                                                                                                                                                                                                                checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul). Es ist ein Solarmodul im System, das mit der gew&auml;hlten Einstellung nicht verwendet wird.",                                                                                                                                                                                                                                                                                                                                               correction => "Konfiguration &auml;ndern"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1126"} = {errorcode => "A91",                                    subcode => "6004",                                               class => "8",           reason => "Keine Kommunikation Solarmodul",                                                                                                                                                                                                                                                                                                                   checklist => "Konfiguration pr&uuml;fen (Adresseinstellung am Modul). Mit der gew&auml;hlten Einstellung ist ein Solarmodul erforderlich.&bull;Die Verbindungsleitung EMS zum Solarmodul auf Besch&auml;digung pr&uuml;fen; Busspannung am Solarmodul muss zwischen 12-15V DC liegen&bull;Solarmodul defekt",                                                                                                                                                                                                       correction => "Konfiguration &auml;ndern&bull;Besch&auml;digte Kabel austauschen&bull;Modul austauschen"                                                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"1127"} = {errorcode => "AC",                                     subcode => "45",                                                 class => "",            reason => "Keine Kommunikation mit vorhandenem CAN-BUS-Modul",                                                                                                                                                                                                                                                                                                checklist => "Stecker vom Verbindungskabel nicht komplett aufgesteckt&bull;BUS-Modul defekt&bull;Signalauswertung auf Leiterplatte defekt",                                                                                                                                                                                                                                                                                                                                                                         correction => "BUS-Modul austauschen&bull;Leiterplatte austauschen&bull;Stecker vom Verbindungskabel aufstecken"                                                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"1128"} = {errorcode => "AD",                                     subcode => "2",                                                  class => "",            reason => "Speichertemperaturf&uuml;hler (SLS unterer Temperaturf&uuml;hler)",                                                                                                                                                                                                                                                                                checklist => "Temperaturf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Speichertemperaturf&uuml;hler und Anschlusskabel pr&uuml;fen ggf. austauschen"                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"1129"} = {errorcode => "AD",                                     subcode => "74",                                                 class => "",            reason => "Temperaturf&uuml;hler Pufferspeicher defekt",                                                                                                                                                                                                                                                                                                      checklist => "Temperaturf&uuml;hler f&uuml;r den Pufferspeicher falsch angeschlossen&bull;Temperaturf&uuml;hler f&uuml;r den Pufferspeicher defekt&bull;Signalauswertung auf Leiterplatte defekt",                                                                                                                                                                                                                                                                                                                  correction => "Speichertemperaturf&uuml;hler (NTC) austauschen&bull;Leiterplatte austauschen&bull;Temperaturf&uuml;hler f&uuml;r den Pufferspeicher anschlie&szlig;en"                                                                                                                                                                                                                                          };
	$ErrorCodeList{"1130"} = {errorcode => "AD1 AD2 AD3 AD4 AD5 AD6 AD7 AD8",        subcode => "815",                                                class => "14",          reason => "Temperaturf&uuml;hler hydraulische Weiche defekt (Pumpeneffizienzmodul)",                                                                                                                                                                                                                                                                          checklist => "Temperaturf&uuml;hler FK des Pumpeneffizienzmodul defekt oder nicht korrekt; angeschlossen&bull;",                                                                                                                                                                                                                                                                                                                                                                                                    correction => "F&uuml;hleranschluss pr&uuml;fen&bull;Weichenf&uuml;hler auf falsche Einbauposition oder auf Bruchstelle pr&uuml;fen"                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1131"} = {errorcode => "AD1",                                    subcode => "817",                                                class => "14",          reason => "Lufttemperaturf&uuml;hler defekt (Auswirkung: Gebl&auml;sedrehzahl kann nicht mehr optimal ausgeregelt werden)",                                                                                                                                                                                                                                   checklist => "Lufttemperaturf&uuml;hler inkl. Steckverbindung der Ger&auml;teelektronik SAFe pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Steckverbindung richtig aufstecken oder ggf. Temperaturf&uuml;hler austauschen"                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1132"} = {errorcode => "AD1",                                    subcode => "818",                                                class => "14",          reason => "W&auml;rmeerzeuger bleibt kalt (Der Kessel ist 30 min lang unter der Pumpenlogiktemperatur (47°C), obwohl der Brenner an ist.; Auswirkung: Die Anlage wird unterversorgt)",                                                                                                                                                                       checklist => "Die Version der Ger&auml;teelektronik SAFe pr&uuml;fen&bull;Evtl. liegen Probleme im Bereich des Kesself&uuml;hlers bzw. Luftprobleme im Kessel vor&bull;Es wurde ein Weichenmodul installiert, obwohl dieses nicht installiert werden darf, und damit ist die Pumpenlogik au&szlig;er Betrieb&bull;Die Auslegung der Kesselleistung bezogen auf die Anlage pr&uuml;fen&bull;Unter Umst&auml;nden ist die Kesselleistung nicht f&uuml;r einen Parallelbetrieb von Warmwasser und Heizung ausgelegt und es ist trotzdem Parallelbetrieb von Heizung und Warmwasser parametriert",correction => "Ist die Feuerungsautomat-Version nicht mindestens 2.14, sollte die Ger&auml;teelektronik SAFe ausgetauscht werden&bull;Anlage entl&uuml;ften bzw. Anschl&uuml;sse und Anschlusskabel des Kesself&uuml;hlers pr&uuml;fen oder den F&uuml;hler austauschen&bull;Weichenmodul deaktivieren&bull;Ggf. gr&ouml;&szlig;eren Kessel ausw&auml;hlen&bull;Heizungsanlage auf „Warmwasser-Vorrang“ einstellen"};
	$ErrorCodeList{"1133"} = {errorcode => "AD1",                                    subcode => "819",                                                class => "14",          reason => "&Ouml;lvorw&auml;rmer meldet Dauersignal; (Vom &Ouml;lvorw&auml;rmer wird ein Freigabesignal empfangen, obwohl er ausgeschaltet; ist. Auswirkung: Der Brenner versucht so zu starten)",                                                                                                                                                            checklist => "Verbindungsleitung zwischen Ger&auml;teelektronik SAFe und &Ouml;lvorw&auml;rmer optisch auf Besch&auml;digungen pr&uuml;fen&bull;Verbindungsleitung zwischen Ger&auml;teelektronik SAFe und &Ouml;lvorw&auml;rmer abziehen und auf Kurzschl&uuml;sse pr&uuml;fen&bull;&Ouml;lvorw&auml;rmer im kalten Zustand pr&uuml;fen",                                                                                                                                                                          correction => "Sollte das Kabel besch&auml;digt sein, dieses austauschen&bull;Sollten Kurzschl&uuml;sse vorhanden sein, Kabel austauschen&bull;Sollte eine Verbindung vorhanden sein, &Ouml;lvorw&auml;rmer austauschen"                                                                                                                                                                                        };
	$ErrorCodeList{"1134"} = {errorcode => "AD1",                                    subcode => "820",                                                class => "14",          reason => "&Ouml;l-Betriebstemperatur wird nicht erreicht; (Der &Ouml;lvorw&auml;rmer gibt kein Signal zur&uuml;ck, dass das &Ouml;l seine Betriebstemperatur erreicht hat. Nach ca.; 6 min versucht der Brenner trotzdem; zu starten)",                                                                                                                      checklist => "Verbindungsleitung zwischen Ger&auml;teelektronik SAFe und &Ouml;lvorw&auml;rmer optisch auf Besch&auml;digungen pr&uuml;fen&bull;Pr&uuml;fen, ob an der Ger&auml;teelektronik SAFe und am &Ouml;lvorw&auml;rmer die Stecker richtig aufgesteckt sind&bull;Verbindungsleitung zwischen Ger&auml;teelektronik SAFe und &Ouml;lvorw&auml;rmer auf Durchgang pr&uuml;fen&bull;Im Men&uuml; „Service“ der Bedieneinheit den Men&uuml;punkt „Relaistest“ aufrufen (RC30) bzw. „Funktionstest“ (RC35/RC300) und hier den &Ouml;lvorw&auml;rmer einschalten; Mit der Hand pr&uuml;fen, ob der &Ouml;lvorw&auml;rmer sich erw&auml;rmt&bull;",correction => "Falls das Kabel besch&auml;digt ist, dieses austauschen&bull;Stecker richtig aufstecken&bull;Erw&auml;rmt sich der &Ouml;lvorw&auml;rmer und die St&ouml;rungsmeldung bleibt erhalten, dann ist; der Schaltkontakt im &Ouml;lvorw&auml;rmer defekt; Der &Ouml;lvorw&auml;rmer muss ausgetauscht werden&bull;Erw&auml;rmt sich der &Ouml;lvorw&auml;rmer und die St&ouml;rungsmeldung bleibt erhalten, dann ist; der Schaltkontakt im &Ouml;lvorw&auml;rmer defekt; Der &Ouml;lvorw&auml;rmer muss ausgetauscht werden&bull;Erw&auml;rmt sich der &Ouml;lvorw&auml;rmer nicht, ist das Heizelement im &Ouml;lvorw&auml;rmer defekt; Der &Ouml;lvorw&auml;rmer muss ausgetauscht werden"};
	$ErrorCodeList{"1135"} = {errorcode => "AD1 AD2 AD3 AD4 AD5 AD6 AD7 AD8",        subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung; EMS plus; (AD1 = Heizkreis 1; AD2 = Heizkreis 2; AD3 = Heizkreis 3; AD4 = Heizkreis 4; AD5 = Heizkreis 5; AD6 = Heizkreis 6; AD7 = Heizkreis 7; AD8 = Heizkreis 8)",                                                                                                                                 checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist Erweiterungsmodule vom EMS-BUS; entfernen und Regelger&auml;t ausund wieder einschalten; Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung&bull;",                                                                                                                                                                                                                                    correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t Ausund wieder Einschalten&bull;Busleitung reparieren bzw. austauschen&bull;Defekten EMS-BUS-Teilnehmer austauschen"                                                                                                                                                                                                                            };
	$ErrorCodeList{"1136"} = {errorcode => "AD1 AD2 AD3 AD4 AD5 AD6 AD7 AD8",        subcode => "1030",                                               class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen (AD1 = Heizkreis 1;; AD2 = Heizkreis 2; AD3 = Heizkreis 3; AD4 = Heizkreis 4; AD5 = Heizkreis 5; AD6 = Heizkreis 6; AD7 = Heizkreis 7; AD8 = Heizkreis 8)",                                                                                                                                      checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"1137"} = {errorcode => "AE1 AE2 AE3 AE4 AE5 AE6 AE7 AE8",        subcode => "815",                                                class => "14",          reason => "Temperaturf&uuml;hler hydraulische Weiche defekt (Pumpen-; effizienzmodul)",                                                                                                                                                                                                                                                                       checklist => "Temperaturf&uuml;hler FK des Pumpeneffizienzmodul defekt oder nicht korrekt; angeschlossen&bull;",                                                                                                                                                                                                                                                                                                                                                                                                    correction => "F&uuml;hleranschluss pr&uuml;fen&bull;Weichenf&uuml;hler auf falsche Einbauposition oder auf Bruchstelle pr&uuml;fen"                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1138"} = {errorcode => "AE1 AE2 AE3 AE4 AE5 AE6 AE7 AE8",        subcode => "1010",                                               class => "14",          reason => "Keine Kommunikation &uuml;ber BUS-Verbindung EMS plus; (AE1 = Heizkreis 1; AE2 = Heizkreis 2; AE3 = Heizkreis 3; AE4 = Heizkreis 4; AE5 = Heizkreis 5; AE6 = Heizkreis 6; AE7 = Heizkreis 7; AE8 = Heizkreis 8)",                                                                                                                                  checklist => "Pr&uuml;fen, ob Busleitung falsch angeschlossen wurde&bull;Pr&uuml;fen, ob Busleitung defekt ist&bull;Erweiterungsmodule vom EMS-BUS entfernen und Regelger&auml;t ausund wieder einschalten&bull;Pr&uuml;fen, ob St&ouml;rungsursache Modul oder Modulverdrahtung",                                                                                                                                                                                                                                  correction => "Verdrahtungsfehler beseitigen und Regelger&auml;t Ausund wieder Einschalten&bull;Busleitung reparieren bzw. austauschen&bull;Defekten EMS-BUS-Teilnehmer austauschen&bull;-"                                                                                                                                                                                                                     };
	$ErrorCodeList{"1139"} = {errorcode => "AE1 AE2 AE3 AE4 AE5 AE6 AE7 AE8",        subcode => "1030",                                               class => "0",           reason => "Interner Datenfehler, Regelger&auml;t austauschen (AE1 = Heizkreis 1;; AE2 = Heizkreis 2; AE3 = Heizkreis 3; AE4 = Heizkreis 4; AE5 = Heizkreis 5; AE6 = Heizkreis 6; AE7 = Heizkreis 7; AE8 = Heizkreis 8)",                                                                                                                                      checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"1140"} = {errorcode => "AY",                                     subcode => "311",                                                class => "4",           reason => "Alle W&auml;rmeerzeuger verriegelt",                                                                                                                                                                                                                                                                                                               checklist => "W&auml;rmeerzeuger &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                                             correction => "St&ouml;rung am W&auml;rmeerzeuger beheben"                                                                                                                                                                                                                                                                                                                                                      };
	$ErrorCodeList{"1141"} = {errorcode => "AY",                                     subcode => "312",                                                class => "10",          reason => "Alle W&auml;rmeerzeuger blockiert",                                                                                                                                                                                                                                                                                                                checklist => "W&auml;rmeerzeuger &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                                             correction => "St&ouml;rung am W&auml;rmeerzeuger beheben"                                                                                                                                                                                                                                                                                                                                                      };
	$ErrorCodeList{"1142"} = {errorcode => "AY",                                     subcode => "313",                                                class => "10",          reason => "W&auml;rmeerzeuger gesperrt oder blockiert",                                                                                                                                                                                                                                                                                                       checklist => "EMS-Kessel pr&uuml;fen&bull;Hybridsystem: Au&szlig;entemperaturf&uuml;hler ist defekt (Stromkreis offen/Kurzschluss)&bull;Hybridsystem: St&ouml;rung des Str&ouml;mungsschalter-Selbsttests beim Einschalten&bull;Hybridsystem: Kurzschluss am Temperaturf&uuml;hler am Eingang des Verfl&uuml;ssigers&bull;Hybridsystem: Offener Stromkreis; am Temperaturf&uuml;hler am Eingang des Verfl&uuml;ssigers&bull;Hybridsystem: Kurzschluss am Temperaturf&uuml;hler am Ausgang des Verfl&uuml;ssigers&bull;Hybridsystem: Offener Stromkreis am Temperaturf&uuml;hler am Ausgang des Verfl&uuml;ssigers",correction => "St&ouml;rung am EMS-Kessel beheben&bull;Kontakte, Kabel und Verbindungen zwischen W&auml;rmeerzeuger und Au&szlig;entemperaturf&uuml;hler pr&uuml;fen und ggf. reparieren&bull;Geh&auml;use des Au&szlig;entemperaturf&uuml;hlers &ouml;ffnen, F&uuml;hler pr&uuml;fen und ggf. austauschen&bull;Str&ouml;mungsschalter pr&uuml;fen und ggf. austauschen&bull;Kontakte, Kabel, Verbindungen und Temperaturf&uuml;hler pr&uuml;fen und ggf. austauschen&bull;Kontakte, Kabel, Verbindungen und Temperaturf&uuml;hler pr&uuml;fen und ggf. austauschen&bull;Kontakte, Kabel, Verbindungen und Temperaturf&uuml;hler pr&uuml;fen und ggf. austauschen"};
	$ErrorCodeList{"1143"} = {errorcode => "B1",                                     subcode => "31",                                                 class => "",            reason => "Kodierstecker defekt oder ung&uuml;ltig",                                                                                                                                                                                                                                                                                                          checklist => "Falscher Kodierstecker eingebaut, Interne St&ouml;rung im Kodierstecker",                                                                                                                                                                                                                                                                                                                                                                                                                             correction => "Kodierstecker austauschen"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1144"} = {errorcode => "B1",                                     subcode => "32",                                                 class => "",            reason => "Kodierstecker defekt oder ung&uuml;ltig",                                                                                                                                                                                                                                                                                                          checklist => "falsche Leiterplatte vorhanden",                                                                                                                                                                                                                                                                                                                                                                                                                                                                      correction => "Leiterplatte austauschen"                                                                                                                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"1145"} = {errorcode => "B2",                                     subcode => "67",                                                 class => "",            reason => "Interner Fehler",                                                                                                                                                                                                                                                                                                                                  checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Regelger&auml;t auf Grundeinstellung zur&uuml;cksetzen (-> Servicefunktion 8.E)"                                                                                                                                                                                                                                                                                                                 };
	$ErrorCodeList{"1146"} = {errorcode => "C0",                                     subcode => "288",                                                class => "4",           reason => "Kurzschluss Wasserdruckf&uuml;hler",                                                                                                                                                                                                                                                                                                               checklist => "Siehe auch St&ouml;rung 2E/207 und 2L/266&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                                     correction => "Wasserdruck pr&uuml;fen&bull;Ggf. ist auch der Druckf&uuml;hler defekt und muss ausgetauscht werden"                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1147"} = {errorcode => "C0",                                     subcode => "289",                                                class => "4",           reason => "Kurzschluss Wasserdruckf&uuml;hler",                                                                                                                                                                                                                                                                                                               checklist => "Verbindungsleitung des Wasserdruckf&uuml;hlers auf Kurzschluss pr&uuml;fen&bull;Wasserdruckf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                                                                                                       correction => "Verbindungsleitung austauschen bzw. den Kurzschluss beseitigen&bull;Wasserdruckf&uuml;hler austauschen"                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"1148"} = {errorcode => "C0",                                     subcode => "568",                                                class => "4",           reason => "Unterbrechung Wasserdruckf&uuml;hler",                                                                                                                                                                                                                                                                                                             checklist => "Kabelverbindung zum Wasserdruckf&uuml;hler pr&uuml;fen&bull;Wasserdruckf&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                      correction => "Ggf. Unterbrechung beseitigen&bull;Wasserdruckf&uuml;hler austauschen"                                                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"1149"} = {errorcode => "C0",                                     subcode => "569",                                                class => "4",           reason => "Kurzschluss Wasserdruckf&uuml;hler",                                                                                                                                                                                                                                                                                                               checklist => "Kabelverbindung zum Wasserdruckf&uuml;hler pr&uuml;fen&bull;Wasserdruckf&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                      correction => "Ggf. Kurzschluss beseitigen&bull;Wasserdruckf&uuml;hler austauschen"                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1150"} = {errorcode => "C1",                                     subcode => "18",                                                 class => "",            reason => "Gebl&auml;se bleibt stehen oder Differenzdruckschalter &ouml;ffnet w&auml;hrend Brennerbetrieb",                                                                                                                                                                                                                                                   checklist => "Stecker am Differenzdruckschalter lose&bull;Gebl&auml;se ist blockiert&bull;Signalunterbrechung im Anschlusskabel zum Differenzdruckschalter&bull;Kein Signal vom Gebl&auml;se&bull;Differenzdruckschalter defekt&bull;Signalauswertung auf Leiterplatte defekt&bull;",                                                                                                                                                                                                                               correction => "Steckkontakte an Differenzdruckschalter aufstecken&bull;Gebl&auml;se austauschen&bull;Kabelbaum austauschen&bull;Kabelbaum austauschen&bull;Differenzdruckschalter austauschen&bull;Leiterplatte austauschen&bull;Anschlusskabel zum Gebl&auml;se austauschen"                                                                                                                                   };
	$ErrorCodeList{"1151"} = {errorcode => "C1",                                     subcode => "264",                                                class => "4",           reason => "Lufttransport w&auml;hrend Betriebsphase ausgefallen",                                                                                                                                                                                                                                                                                             checklist => "Anschlusstecker am Gebl&auml;se &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob das Gebl&auml;se blockiert ist&bull;",                                                                                                                                                                                                                                                                                                                                                                             correction => "Anschlussstecker am Gebl&auml;se aufstecken&bull;Gebl&auml;se austauschen&bull;Steuerger&auml;t austauschen"                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"1152"} = {errorcode => "C3",                                     subcode => "50",                                                 class => "",            reason => "Abgasaustritt aus Brennkammer, Heizger&auml;t verriegelt",                                                                                                                                                                                                                                                                                         checklist => "Verteilerrohr vertauscht&bull;Zu gro&szlig;e Gasmenge&bull;W&auml;rmeblock abgasseitig zugesetzt&bull;Signal des F&uuml;hler Abgas&uuml;berwachung liegt au&szlig;erhalb der Kennlinie&bull;Signalauswertung auf Leiterplatte defekt",                                                                                                                                                                                                                                                                correction => "Brennereinstellwerte bei Nennbelastung einstellen&bull;W&auml;rmeblock abgasseitig reinigen&bull;F&uuml;hler (Abgas&uuml;berwachung) an der Brennkammer austauschen&bull;Leiterplatte austauschen&bull;Verteilerrohr tauschen"                                                                                                                                                                   };
	$ErrorCodeList{"1153"} = {errorcode => "C4",                                     subcode => "23",                                                 class => "",            reason => "Differenzdruckschalter &ouml;ffnet nicht bei stehendem Gebl&auml;se",                                                                                                                                                                                                                                                                              checklist => "Schaltkontakt ist blockiert beim Differenzdruckschalter&bull;Signalauswertung auf Leiterplatte defekt",                                                                                                                                                                                                                                                                                                                                                                                               correction => "Differenzdruckschalter austauschen&bull;Leiterplatte austauschen"                                                                                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"1154"} = {errorcode => "C4",                                     subcode => "273",                                                class => "4",           reason => "Betriebsunterbrechung: Brenner und Gebl&auml;se",                                                                                                                                                                                                                                                                                                  checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "Gebl&auml;se und Brenner starten automatisch nach der Sicherheitskontrolle"                                                                                                                                                                                                                                                                                                                      };
	$ErrorCodeList{"1155"} = {errorcode => "C6",                                     subcode => "20",                                                 class => "",            reason => "Gebl&auml;se l&auml;uft nicht an oder Differenzdruck-; schalter schlie&szlig;t nicht bei Gebl&auml;seanlauf",                                                                                                                                                                                                                                      checklist => "Stecker am Gebl&auml;se lose&bull;Gebl&auml;se ist blockiert&bull;Signalunterbrechung im Anschlusskabel zum Differenzdruckschalter&bull;Kein Signal vom Gebl&auml;se&bull;Abgasrohrl&auml;nge zu lang&bull;Anzahl der Umlenkungen im Abgassystem zu hoch&bull;Verbrennungsluftmenge zu gering&bull;Differenzdruckschalter defekt&bull;&bull;",                                                                                                                                                        correction => "Stecker aufstecken&bull;Gebl&auml;se austauschen&bull;Kabelbaum austauschen&bull;Anschlusskabel zum Gebl&auml;se austauschen&bull;Abgasrohrl&auml;nge ermitteln und mit den freigegebenen Rohrl&auml;ngen abgleichen&bull;Abgassystem&bull;Verbrennungsluftverbund zu klein bzw. zu geringe Gr&ouml;&szlig;e der L&uuml;ftungs&ouml;ffnung&bull;Steckkontakte an Differenzdruckschalter aufstecken&bull;Differenzdruckschalter austauschen&bull;Leiterplatte austauschen"};
	$ErrorCodeList{"1156"} = {errorcode => "C6",                                     subcode => "71",                                                 class => "8",           reason => "Gebl&auml;se l&auml;uft nicht an oder Differenzdruckschalter schlie&szlig;t nicht bei Gebl&auml;seanlauf",                                                                                                                                                                                                                                         checklist => "Gebl&auml;se ist defekt&bull;Signalauswertung auf Leiterplatte defekt",                                                                                                                                                                                                                                                                                                                                                                                                                               correction => "Gebl&auml;se austauschen&bull;Leiterplatte austauschen"                                                                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"1157"} = {errorcode => "C6",                                     subcode => "215",                                                class => "8",           reason => "Gebl&auml;se zu schnell",                                                                                                                                                                                                                                                                                                                          checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Gebl&auml;se austauschen"                                                                                                                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"1158"} = {errorcode => "C6",                                     subcode => "216",                                                class => "8",           reason => "Gebl&auml;se zu langsam",                                                                                                                                                                                                                                                                                                                          checklist => "Internen Kondensatsiphon &uuml;berpr&uuml;fen&bull;W&auml;rmeblock abgasseitig auf Ablagerungen &uuml;berpr&uuml;fen&bull;",                                                                                                                                                                                                                                                                                                                                                                          correction => "Internen Kondensatsiphon reinigen&bull;W&auml;rmeblock abgasseitig reinigen&bull;Gebl&auml;se austauschen"                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1159"} = {errorcode => "C7",                                     subcode => "214",                                                class => "8",           reason => "Gebl&auml;se wird w&auml;hrend Sicherheitszeit abgeschaltet",                                                                                                                                                                                                                                                                                      checklist => "Anschlusstecker am Gebl&auml;se &uuml;berpr&uuml;fen&bull;Anschlusskabel zum Gebl&auml;se &uuml;berpr&uuml;fen&bull;",                                                                                                                                                                                                                                                                                                                                                                                correction => "Anschlussstecker am Gebl&auml;se aufstecken&bull;Anschlusskabel zum Gebl&auml;se austauschen&bull;Gebl&auml;se austauschen"                                                                                                                                                                                                                                                                      };
	$ErrorCodeList{"1160"} = {errorcode => "C7",                                     subcode => "217",                                                class => "8",           reason => "Kein Lufttransport nach mehreren Minuten",                                                                                                                                                                                                                                                                                                         checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Steuerger&auml;t/ Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1161"} = {errorcode => "C8",                                     subcode => "68",                                                 class => "",            reason => "Au&szlig;entemperaturf&uuml;hler an Ger&auml;teelektronik erkannt, aber nicht zul&auml;ssig Anschluss muss an; BUS-Modul erfolgen",                                                                                                                                                                                                                checklist => "Au&szlig;entemperaturf&uuml;hler ist nicht am BUS-Modul angeschlossen",                                                                                                                                                                                                                                                                                                                                                                                                                               correction => "Au&szlig;entemperaturf&uuml;hler am BUS-Modul anschlie&szlig;en"                                                                                                                                                                                                                                                                                                                                 };
	$ErrorCodeList{"1162"} = {errorcode => "CA",                                     subcode => "286",                                                class => "4",           reason => "Temperatur R&uuml;cklauftemperaturf&uuml;hler zu hoch",                                                                                                                                                                                                                                                                                            checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Ger&auml;t startet neu"                                                                                                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"1163"} = {errorcode => "CC",                                     subcode => "3",                                                  class => "",            reason => "St&ouml;rung: Au&szlig;entemperaturf&uuml;hler",                                                                                                                                                                                                                                                                                                   checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Au&szlig;entemperaturf&uuml;hler und Anschlusskabel auf Unterbrechung pr&uuml;fen, ggf. austauschen&bull;Au&szlig;entemperaturf&uuml;hler korrekt an den Anschlussklemmen A und F anschlie&szlig;en"                                                                                                                                                                                             };
	$ErrorCodeList{"1164"} = {errorcode => "CC",                                     subcode => "65",                                                 class => "",            reason => "Au&szlig;entemperaturf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                          checklist => "Au&szlig;entemperaturf&uuml;hler ist nicht an den Anschlussklemmen A und F angeschlossen&bull;Au&szlig;entemperaturf&uuml;hler hat Unterbrechung oder Kurzschluss&bull;Kabel zum Au&szlig;entemperaturf&uuml;hler hat Unterbrechung oder Kurzschluss&bull;BUS-Modul defekt",                                                                                                                                                                                                                          correction => "Au&szlig;entemperaturf&uuml;hler am BUS-Modul anschlie&szlig;en&bull;Kabel zum Au&szlig;entemperaturf&uuml;hler instandsetzen oder austauschen&bull;Au&szlig;entemperaturf&uuml;hler austauschen&bull;BUS-Modul austauschen"                                                                                                                                                                     };
	$ErrorCodeList{"1165"} = {errorcode => "CC",                                     subcode => "90",                                                 class => "10",          reason => "Au&szlig;entemperaturf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                          checklist => "F&uuml;hler anhand der Spannungsund Widerstandswerte kontrollieren&bull;F&uuml;hleranschluss kontrollieren",                                                                                                                                                                                                                                                                                                                                                                                          correction => "Ggf. F&uuml;hler austauschen&bull;F&uuml;hler wieder richtig anschlie&szlig;en"                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1166"} = {errorcode => "CU",                                     subcode => "240",                                                class => "4",           reason => "Kurzschluss R&uuml;cklauftemperaturf&uuml;hler",                                                                                                                                                                                                                                                                                                   checklist => "Verbindungsleitung des R&uuml;cklauftemperaturf&uuml;hlers auf Kurzschluss pr&uuml;fen&bull;R&uuml;cklauftemperaturf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                                                                               correction => "Verbindungsleitung austauschen bzw. den Kurzschluss beseitigen&bull;R&uuml;cklauftemperaturf&uuml;hler austauschen"                                                                                                                                                                                                                                                                              };
	$ErrorCodeList{"1167"} = {errorcode => "CY",                                     subcode => "241",                                                class => "4",           reason => "R&uuml;cklauftemperaturf&uuml;hler Kontakt lose oder defekt",                                                                                                                                                                                                                                                                                      checklist => "Verbindungsleitung des R&uuml;cklauftemperaturf&uuml;hlers auf Unterbrechung pr&uuml;fen&bull;R&uuml;cklauftemperaturf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                                                                             correction => "Verbindungsleitung austauschen bzw. die Unterbrechung beseitigen&bull;R&uuml;cklauftemperaturf&uuml;hler austauschen"                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1168"} = {errorcode => "CY",                                     subcode => "566",                                                class => "4",           reason => "Regelger&auml;t erh&auml;lt unzul&auml;ssige Werte vom R&uuml;cklauftemperaturf&uuml;hler",                                                                                                                                                                                                                                                        checklist => "Verbindungsleitung zwischen Regelger&auml;t und R&uuml;cklauftemperaturf&uuml;hler pr&uuml;fen&bull;Elektrischen Anschluss der Verbindungsleitung im Regelger&auml;t pr&uuml;fen&bull;R&uuml;cklauftemperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des R&uuml;cklauftemperaturf&uuml;hlers im Regelger&auml;t laut Tabelle pr&uuml;fen",                                                                                                                         correction => "Wenn ein Defekt vorliegt, F&uuml;hler austauschen&bull;Falls Schrauben oder ein Stecker lose sind, das Kontaktproblem beheben&bull;Wenn die Werte nicht mit den Tabellenwerten &uuml;bereinstimmen, den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, die Spannungswerte jedoch nicht &uuml;bereinstimmen, das Regelger&auml;t (Kaskadenmodul) austauschen"             };
	$ErrorCodeList{"1169"} = {errorcode => "CY",                                     subcode => "567",                                                class => "4",           reason => "Regelger&auml;t erh&auml;lt unzul&auml;ssige Werte vom R&uuml;cklauftemperaturf&uuml;hler",                                                                                                                                                                                                                                                        checklist => "Verbindungsleitung zwischen Regelger&auml;t und R&uuml;cklauftemperaturf&uuml;hler pr&uuml;fen&bull;Elektrischen Anschluss der Verbindungsleitung im Regelger&auml;t pr&uuml;fen&bull;R&uuml;cklauftemperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des R&uuml;cklauftemperaturf&uuml;hlers im Regelger&auml;t laut Tabelle pr&uuml;fen",                                                                                                                         correction => "Wenn ein Defekt vorliegt, F&uuml;hler austauschen&bull;Falls Schrauben oder ein Stecker lose sind, das Kontaktproblem beheben&bull;Wenn die Werte nicht mit den Tabellenwerten &uuml;bereinstimmen, den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, die Spannungswerte jedoch nicht &uuml;bereinstimmen, das Regelger&auml;t (Kaskadenmodul) austauschen"             };
	$ErrorCodeList{"1170"} = {errorcode => "CY",                                     subcode => "573",                                                class => "4",           reason => "Feuerungsautomat erh&auml;lt unzul&auml;ssige Werte vom Vorlauftemperaturf&uuml;hler",                                                                                                                                                                                                                                                             checklist => "Verbindungsleitung zwischen Regelger&auml;t und Vorlauftemperaturf&uuml;hler pr&uuml;fen&bull;Elektrischen Anschluss der Verbindungsleitung im Regelger&auml;t pr&uuml;fen&bull;Vorlauftemperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Vorlauftemperaturf&uuml;hlers im Regelger&auml;t laut Tabelle pr&uuml;fen",                                                                                                                                           correction => "Wenn ein Defekt vorliegt, F&uuml;hler austauschen&bull;Falls Schrauben oder ein Stecker lose sind, das Kontaktproblem beheben&bull;Wenn die Werte nicht &uuml;bereinstimmen, den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, die Spannungswerte jedoch nicht &uuml;bereinstimmen, das Regelger&auml;t MC10 austauschen"                                               };
	$ErrorCodeList{"1171"} = {errorcode => "CY",                                     subcode => "574",                                                class => "4",           reason => "Feuerungsautomat erh&auml;lt unzul&auml;ssige Werte vom Vorlauftemperaturf&uuml;hler (Vorlauftemperaturf&uuml;hler Kurzschluss, Temperatur; >150 °C, GB402 >130 °C)",                                                                                                                                                                            checklist => "Verbindungsleitung zwischen Regelger&auml;t und R&uuml;cklauftemperaturf&uuml;hler pr&uuml;fen&bull;Elektrischen Anschluss der Verbindungsleitung im Regelger&auml;t pr&uuml;fen&bull;Vorlauftemperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des R&uuml;cklauftemperaturf&uuml;hlers im Regelger&auml;t laut Tabelle pr&uuml;fen",                                                                                                                               correction => "Wenn ein Defekt vorliegt, F&uuml;hler austauschen&bull;Falls Schrauben oder ein Stecker lose sind, das Kontaktproblem beheben&bull;Wenn die Werte nicht &uuml;bereinstimmen, den F&uuml;hler austauschen&bull;Wenn die F&uuml;hlerwerte gestimmt haben, die Spannungswerte jedoch nicht &uuml;bereinstimmen, das Regelger&auml;t MC10 austauschen"                                               };
	$ErrorCodeList{"1172"} = {errorcode => "D1",                                     subcode => "37",                                                 class => "",            reason => "R&uuml;cklauftemperaturf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                        checklist => "Kein Signal vom R&uuml;cklauftemperaturf&uuml;hler&bull;Signal vom R&uuml;cklauff&uuml;hler au&szlig;erhalb Kennlinie&bull;Signalunterbrechung im Anschlusskabel&bull;Signalauswertung auf Leiterplatte defekt",                                                                                                                                                                                                                                                                                      correction => "R&uuml;cklauff&uuml;hler austauschen&bull;Kabelbaum austauschen&bull;Steckkontakte am F&uuml;hler aufstecken&bull;Leiterplatte austauschen"                                                                                                                                                                                                                                                      };
	$ErrorCodeList{"1173"} = {errorcode => "D1",                                     subcode => "69",                                                 class => "",            reason => "Zus&auml;tzlicher Gradientf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                     checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Temperaturf&uuml;hler und Anschlusskabel auf Unterbrechung oder Kurzschluss pr&uuml;fen, ggf. tauschen"                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"1174"} = {errorcode => "D1",                                     subcode => "84",                                                 class => "10",          reason => "R&uuml;cklauftemperaturf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                        checklist => "R&uuml;cklauftemperaturf&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                                      correction => "F&uuml;hler wieder richtig anschlie&szlig;en, ggf. austauschen"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1175"} = {errorcode => "D1",                                     subcode => "85",                                                 class => "10",          reason => "R&uuml;cklauftemperaturf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                        checklist => "R&uuml;cklauftemperaturf&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                                      correction => "F&uuml;hler wieder richtig anschlie&szlig;en, ggf. austauschen"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1176"} = {errorcode => "D1",                                     subcode => "89",                                                 class => "10",          reason => "Vorlaufoder R&uuml;cklauftemperatur au&szlig;erhalb des zul&auml;ssigen Bereichs",                                                                                                                                                                                                                                                                 checklist => "Vorlaufoder R&uuml;cklauftemperatur am HybridManager au&szlig;erhalb des zul&auml;ssigen Bereichs",                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Der W&auml;rmeerzeuger erw&auml;rmt zun&auml;chst das Wasser, bevor die Luft-Wasser-W&auml;rmepumpe gestartet wird"                                                                                                                                                                                                                                                                              };
	$ErrorCodeList{"1177"} = {errorcode => "D1",                                     subcode => "241",                                                class => "8",           reason => "R&uuml;cklauftemperaturf&uuml;hler Ger&auml;t Kontakt lose oder defekt",                                                                                                                                                                                                                                                                           checklist => "Verbindungsleitung des R&uuml;cklauf Temperaturf&uuml;hlers auf Unterbrechung pr&uuml;fen&bull;R&uuml;cklauftemperaturf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                                                                            correction => "Verbindungsleitung austauschen bzw. die Unterbrechung beseitigen&bull;R&uuml;cklaufsensor austauschen"                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"1178"} = {errorcode => "D3",                                     subcode => "22",                                                 class => "8",           reason => "Externer W&auml;chter ausgel&ouml;st, kein Heizbetrieb m&ouml;glich",                                                                                                                                                                                                                                                                              checklist => "Br&uuml;cke zwischen den Anschlussklemmen PR und PO fehlt&bull;R&uuml;ckmeldespannung an Anschlussklemme 9 fehlt&bull;Br&uuml;cke zwischen den Anschlussklemmen 8 und 9 fehlt",                                                                                                                                                                                                                                                                                                                       correction => "Br&uuml;cke zwischen den Anschlussklemmen PR und PO einbauen&bull;Begrenzer f&uuml;r die Fu&szlig;bodenheizung &uuml;berpr&uuml;fen&bull;Br&uuml;cke zwischen den Anschlussklemmen 8 und 9 einbauen"                                                                                                                                                                                             };
	$ErrorCodeList{"1179"} = {errorcode => "D3",                                     subcode => "232",                                                class => "8",           reason => "W&auml;rmeerzeuger durch externen Schaltkontakt verriegelt",                                                                                                                                                                                                                                                                                       checklist => "Anschlussstecker f&uuml;r den externen Schaltkontakt &uuml;berpr&uuml;fen&bull;Belegung des Anschlussstecker f&uuml;r den externen Schaltkontakt &uuml;berpr&uuml;fen/Ablaufleitung, Schwimmerschalter und Funktion der Kondensathebepumpe &uuml;berpr&uuml;fen&bull;Schaltpunkt des externen Temperaturw&auml;chters &uuml;berpr&uuml;fen&bull;Anschlusskabel zum externen Temperaturw&auml;chter &uuml;berpr&uuml;fen&bull;Funktion des externen Temrperaturw&auml;chters &uuml;berpr&uuml;fen",    correction => "Anschlusstecker f&uuml;r den externen Schaltkontakt aufstecken&bull;Br&uuml;cke einbauen/ Kondensathebepumpe nach Herstellerangaben &uuml;berpr&uuml;fen&bull;Schaltpunkt des externen Temperaturw&auml;chters an das System anpassen&bull;Anschlusskabel zum externen Temperaturw&auml;chter austauschen&bull;Externer Temperaturw&auml;chter austauschen"                                      };
	$ErrorCodeList{"1180"} = {errorcode => "D3",                                     subcode => "311",                                                class => "4",           reason => "Alle W&auml;rmeerzeuger verriegelt",                                                                                                                                                                                                                                                                                                               checklist => "W&auml;rmeerzeuger &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                                             correction => "St&ouml;rung am W&auml;rmeerzeuger beheben"                                                                                                                                                                                                                                                                                                                                                      };
	$ErrorCodeList{"1181"} = {errorcode => "D3",                                     subcode => "312",                                                class => "10",          reason => "Alle W&auml;rmeerzeuger blockiert",                                                                                                                                                                                                                                                                                                                checklist => "W&auml;rmeerzeuger &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                                             correction => "St&ouml;rung am W&auml;rmeerzeuger beheben"                                                                                                                                                                                                                                                                                                                                                      };
	$ErrorCodeList{"1182"} = {errorcode => "D4",                                     subcode => "19",                                                 class => "8",           reason => "Brenner vor&uuml;bergehend ausgeschaltet, aufgrund zu schnell ansteigender Vorlauftemperatur",                                                                                                                                                                                                                                                     checklist => "Kessel l&auml;uft selbst&auml;ndig nach kurzer Zeit wieder an",                                                                                                                                                                                                                                                                                                                                                                                                                                       correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1183"} = {errorcode => "D5",                                     subcode => "15",                                                 class => "",            reason => "Externer Vorlauftemperaturf&uuml;hler hydraulische Weiche defekt oder zwei Temperaturf&uuml;hler angeschlossen",                                                                                                                                                                                                                                   checklist => "Vorlauftemperaturf&uuml;hler f&uuml;r die hydraulische Weiche nicht angeschlossen&bull;Zwei externe Vorlauftemperaturf&uuml;hlerf&uuml;r die hydraulische Weiche angeschlossen&bull;Vorlauftemperaturf&uuml;hleran der hydraulischen Weiche defekt&bull;Signalauswertung auf Leiterplatte defekt",                                                                                                                                                                                                    correction => "Parameter O in der Servicefunktion 7.d einstellen&bull;Vorlauftemperaturf&uuml;hler an der hydraulischen Weiche ausbauen&bull;Externen Vorlauftemperaturf&uuml;hler austauschen&bull;Leiterplatte austauschen"                                                                                                                                                                                   };
	$ErrorCodeList{"1184"} = {errorcode => "D7",                                     subcode => "43",                                                 class => "",            reason => "Regelventil &ouml;ffnet nicht im Betrieb",                                                                                                                                                                                                                                                                                                         checklist => "Stecker f&uuml;r das Regelventil an der Gasarmatur ist nicht aufgesteckt&bull;Anschlusskabel zum Regelventil an der Gasarmatur ist besch&auml;digt&bull;Regelventil an Gasarmatur defekt&bull;Signalauswertung auf Leiterplatte defekt",                                                                                                                                                                                                                                                              correction => "Stecker am Regelventil der Gasarmatur aufstecken&bull;Kabelbaum austauschen&bull;Gasarmatur austauschen&bull;Leiterplatte austauschen"                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"1185"} = {errorcode => "E1",                                     subcode => "242; 243; 244; 245; 256; 247; 248; 249; 255; 257",   class => "4",           reason => "Systemst&ouml;rung Ger&auml;teelektronik UBA/Basiscontroller oder Kesselidentifikationsmodul KIM defekt",                                                                                                                                                                                                                                          checklist => "Kontakt zwischen Ger&auml;teelektronik UBA und Montagefu&szlig; sowie Basiscontroller und der Grundplatte und allen sonstigen Steckverbindungen pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                         correction => "Kontaktproblem beseitigen, ggf. Ger&auml;teelektronik UBA oder Kesselidentifikationsmodul KIM austauschen"                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1186"} = {errorcode => "E2",                                     subcode => "21",                                                 class => "",            reason => "Vorlauftemperaturf&uuml;hler im W&auml;rmeerzeuger defekt",                                                                                                                                                                                                                                                                                        checklist => "Kein Signal vom Vorlauftemperaturf&uuml;hler&bull;Signal vom Vorlauftemperaturf&uuml;hler au&szlig;erhalb Kennlinie&bull;Signalunterbrechung im Anschlusskabel&bull;Signalauswertung auf Leiterplatte defekt",                                                                                                                                                                                                                                                                                        correction => "Stecker am Vorlauftemperaturf&uuml;hler komplett naufstecken&bull;Vorlauftemperaturf&uuml;hler austauschen&bull;Kabelbaum austauschen&bull;Leiterplatte austauschen"                                                                                                                                                                                                                             };
	$ErrorCodeList{"1187"} = {errorcode => "E2",                                     subcode => "86",                                                 class => "10",          reason => "Vorlauftemperaturf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                              checklist => "Vorlauftemperaturf&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "F&uuml;hler wieder richtig anschlie&szlig;en, ggf. austauschen"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1188"} = {errorcode => "E2",                                     subcode => "87",                                                 class => "10",          reason => "Vorlauftemperaturf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                              checklist => "Vorlauftemperaturf&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "F&uuml;hler wieder richtig anschlie&szlig;en, ggf. austauschen"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1189"} = {errorcode => "E2",                                     subcode => "222",                                                class => "8",           reason => "Kurzschluss Vorlauftemperaturf&uuml;hler Ger&auml;t",                                                                                                                                                                                                                                                                                              checklist => "Kennlinie des Vorlauftemperaturf&uuml;hlers &uuml;berpr&uuml;fen&bull;Anschlusskabel zum Vorlauftemperaturf&uuml;hler &uuml;berpr&uuml;fen&bull;",                                                                                                                                                                                                                                                                                                                                                    correction => "Vorlauftemperaurf&uuml;hler austauschen&bull;Anschlusskabel zum Vorlauftemperaurf&uuml;hler austauschen&bull;Steuerger&auml;t/Feuerungsautomat austauschen"                                                                                                                                                                                                                                      };
	$ErrorCodeList{"1190"} = {errorcode => "E2",                                     subcode => "223",                                                class => "8",           reason => "Vorlauftemperaturf&uuml;hler Ger&auml;t Kontakt lose oder defekt",                                                                                                                                                                                                                                                                                 checklist => "Anschlussstecker am Vorlauftemperaturf&uuml;hler &uuml;berpr&uuml;fen&bull;Kennlinie des Vorlauftemperaturf&uuml;hlers &uuml;berpr&uuml;fen&bull;Anschlusskabel zum Vorlauftemperaturf&uuml;hler &uuml;berpr&uuml;fen&bull;",                                                                                                                                                                                                                                                                         correction => "Anschlussstecker am Vorlauftemperaturf&uuml;hler aufstecken&bull;Vorlauftemperaturf&uuml;hler austauschen&bull;Anschlusskabel zum Vorlauftemperaturf&uuml;hler austauschen&bull;Steuerger&auml;t/Feuerungsautomat austauschen"                                                                                                                                                                   };
	$ErrorCodeList{"1191"} = {errorcode => "E2",                                     subcode => "573",                                                class => "4",           reason => "Feuerungsautomat erh&auml;lt unzul&auml;ssige Werte vom Vorlauftemperaturf&uuml;hler",                                                                                                                                                                                                                                                             checklist => "Verbindungsleitung zwischen Regelger&auml;t und Vorlauftemperaturf&uuml;hler pr&uuml;fen&bull;Elektrischer Anschluss der Verbindungsleitung im Regelger&auml;t pr&uuml;fen&bull;Vorlauftemperaturf&uuml;hler laut Tabelle pr&uuml;fen&bull;Spannung an den Anschlussklemmen des Vorlauftemperaturf&uuml;hlers im Regelger&auml;t laut Tabelle pr&uuml;fen",                                                                                                                                           correction => "Wenn ein Defekt vorliegt, F&uuml;hler austauschen&bull;Falls Schrauben oder ein Stecker lose sind, das Kontaktproblem beheben&bull;Sollten Werte nicht &uuml;bereinstimmen, den F&uuml;hler austauschen&bull;Sollten die F&uuml;hlerwerte gestimmt haben, die Spannungswerte jedoch nicht &uuml;bereinstimmen, das Regelger&auml;t (Ger&auml;teelektronik) austauschen"                          };
	$ErrorCodeList{"1192"} = {errorcode => "E4",                                     subcode => "16",                                                 class => "",            reason => "Kaltwassereinlauftemperaturf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                    checklist => "Stecker am Kaltwassertemperaturf&uuml;hler nicht aufgesteckt&bull;Anschlusskabel f&uuml;r Kaltwassertemperaturf&uuml;hler (NTC) besch&auml;digt&bull;Temperaturf&uuml;hler f&uuml;r Kaltwasser defekt&bull;Signalauswertung auf Leiterplatte defekt",                                                                                                                                                                                                                                                 correction => "Stecker aufstecken&bull;Kabelbaum austauschen&bull;F&uuml;hler austauschen&bull;Leiterplatte austauschen"                                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"1193"} = {errorcode => "E5",                                     subcode => "29",                                                 class => "",            reason => "Brenner vor&uuml;bergehend ausgeschaltet, aufgrund zu hoher Temperatur am Brenner",                                                                                                                                                                                                                                                                checklist => "Vorlauftemperatur am Brenner zu hoch&bull;Signal vom Brenner-Vorlauftemperaturf&uuml;hler au&szlig;erhalb Kennlinie&bull;Brennerdeck wasserseitig zugesetzt",                                                                                                                                                                                                                                                                                                                                         correction => "Ger&auml;t entl&uuml;ften&bull;Vorlauftemperaturf&uuml;hler am Brenner austauschen&bull;Brennerdeck austauschen"                                                                                                                                                                                                                                                                                 };
	$ErrorCodeList{"1194"} = {errorcode => "E7",                                     subcode => "28",                                                 class => "",            reason => "Temperaturf&uuml;hler am Brenner defekt",                                                                                                                                                                                                                                                                                                          checklist => "Kein Signal vom Vorlauftemperaturf&uuml;hler am Brenner&bull;Signal vom Brenner-Vorlauftemperaturf&uuml;hler au&szlig;erhalb Kennlinie&bull;Signalunterbrechung im Anschlusskabel&bull;Signalauswertung auf Leiterplatte defekt",                                                                                                                                                                                                                                                                     correction => "Stecker am Vorlauftemperaturf&uuml;hler komplett aufstecken&bull;F&uuml;hler austauschen&bull;Kabelbaum austauschen&bull;Leiterplatte austauschen"                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1195"} = {errorcode => "E8",                                     subcode => "207",                                                class => "4",           reason => "Betriebsdruck zu niedrig",                                                                                                                                                                                                                                                                                                                         checklist => "Kein Vordruck mehr im Membranausdehnungsgef&auml;&szlig; vorhanden&bull;Heizungssicherheitsventil undicht&bull;Schmutzablagerungen an der Messstelle im Drucksensor&bull;Signalauswerung f&uuml;r den Drucksensor im Feuerungsautomat (SAFe) defekt",                                                                                                                                                                                                                                                 correction => "Stickstoff nachf&uuml;llen bzw. Membranausdehnungsgef&auml;&szlig; austauschen&bull;Heizungssicherheitsventil austauschen&bull;Drucksensor reinigen bzw. austauschen&bull;Feuerungsautomat austauschen"                                                                                                                                                                                          };
	$ErrorCodeList{"1196"} = {errorcode => "E9",                                     subcode => "55",                                                 class => "",            reason => "Sicherheitsabschaltung Brenner, weil Sicherheitstemperaturbegrenzer ausgel&ouml;st hat",                                                                                                                                                                                                                                                           checklist => "Keine Umlaufwassermenge vorhanden&bull;Betriebsdruck der Heizungsanlage zu niedrig&bull;Vorlauftemperaturf&uuml;hlernicht in Tauchh&uuml;lse gesteckt&bull;Umw&auml;lzpumpe blockiert&bull;Der Stecker am Abgas-Sicherheitstemperaturbegrenzer hat keinen Kontakt&bull;Stecker am Sicherheitstemperaturbegrenzer hat keinen Kontakt&bull;Abgas-Sicherheitstemperaturbegrenzer hat Unterbrechung&bull;Sicherheitstemperaturbegrenzer hat Unterbrechung&bull;Verdr&auml;ngungsk&ouml;rper nicht montiert&bull;W&auml;rmeblock wasserseitig verschmutzt",correction => "Absperrh&auml;hne &ouml;ffnen&bull;Wasser nachf&uuml;llen, bis der gew&uuml;nschte Betriebsdruck erreicht ist&bull;Vorlauftemperaturf&uuml;hler in Tauchh&uuml;lse stecken&bull;Umw&auml;lzpumpe anwerfen, Pumpenmotor austauschen&bull;Stecker am Abgas-Sicherheitstemperaturbegrenzer vollst&auml;ndig aufstecken&bull;Stecker am Sicherheitstemperaturbegrenzer vollst&auml;ndig aufstecken&bull;Sicherheitstemperaturbegrenzer austauschen&bull;Sicherheitstemperaturbegrenzer austauschen&bull;Verdr&auml;ngungsk&ouml;rper einbauen&bull;W&auml;rmeblock austauschen"};
	$ErrorCodeList{"1197"} = {errorcode => "E9",                                     subcode => "88",                                                 class => "10",          reason => "H&ouml;here R&uuml;cklauftemperatur als Vorlauftemperatur",                                                                                                                                                                                                                                                                                        checklist => "Pr&uuml;fen, ob Vorund R&uuml;cklauf vertauscht sind&bull;DIP-Schalter Stellung auf der Leiterplatte auf dem W&auml;rmepumpenmodul pr&uuml;fen (Normalbetrieb)&bull;Die F&uuml;hler und die Position der F&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                                                    correction => "Vorund R&uuml;cklauf richtig anschlie&szlig;en&bull;DIP-Schalter auf Normalbetreib stellen&bull;Die F&uuml;hler richtig positionieren"                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"1198"} = {errorcode => "E9",                                     subcode => "224",                                                class => "8",           reason => "Sicherheitstemperaturbegrenzer hat ausgel&ouml;st",                                                                                                                                                                                                                                                                                                checklist => "Absperreinrichtungen und Hydrauliksystem &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob die Umw&auml;lzpumpe blockiert ist&bull;Einstellungen f&uuml;r die Umw&auml;lzpumpe &uuml;berpr&uuml;fen&bull;Anschlussstecker am Sicherheitstemperaturbegrenzer &uuml;berpr&uuml;fen&bull;Sicherheitstemperaturbegrenzer &uuml;berpr&uuml;fen&bull;Anschlussstecker an der Umw&auml;lzpumpe &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob sich noch Luft im Heizsystem befindet",                   correction => "Absperreinrichtungen &ouml;ffnen bzw. Anlagenhydraulik ab&auml;ndern&bull;Umw&auml;lzpumpe anwerfen bzw. austauschen&bull;Einstellungen f&uuml;r die Umw&auml;lzpumpe an die Anlagenhydraulik anpasssen&bull;Anschlussstecker am Sicherheitstemperaturbegrenzer aufstecken&bull;Sicherheitstemperaturbegrenzer austauchen&bull;Anschlussstecker an der Umw&auml;lzpumpe korrekt aufstecken und verrasten&bull;Heizsystem entl&uuml;ften"};
	$ErrorCodeList{"1199"} = {errorcode => "E9",                                     subcode => "276",                                                class => "4",           reason => "Temperatur am Vorlauftemperaturf&uuml;hler Ger&auml;tzu hoch",                                                                                                                                                                                                                                                                                     checklist => "Absperreinrichtungen und Hydrauliksystem &uuml;berpr&uuml;fen&bull;Anschlussstecker an der Umw&auml;lzpumpe &uuml;berpr&uuml;fen&bull;Umw&auml;lzpumpenlauf &uuml;berpr&uuml;fen&bull;Einstellungen f&uuml;r die Umw&auml;lzpumpe &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                              correction => "Absperreinrichtungen &ouml;ffnen bzw. Anlagenhydraulik ab&auml;ndern&bull;Anschlusstecker an der Umw&auml;lzpumpe aufstecken&bull;Umw&auml;lzpumpe austauschen&bull;Einstellungen f&uuml;r die Umw&auml;lzpumpe an die Anlagenhydraulik anpasssen"                                                                                                                                               };
	$ErrorCodeList{"1200"} = {errorcode => "EA",                                     subcode => "49",                                                 class => "n.a.",        reason => "Keine Flammbildung innerhalb Sicherheitszeit",                                                                                                                                                                                                                                                                                                     checklist => "Hauptabsperreinrichtung &uuml;berpr&uuml;fen&bull;Ger&auml;teabsperrhahn &uuml;berpr&uuml;fen&bull;Gasanschlu&szlig;flie&szlig;druck bei Nennw&auml;rmebelastung messen&bull;Ionisationsstrom messen&bull;Ionisationselektrode &uuml;berpr&uuml;fen&bull;Schutzleiteranschlu&szlig; im Steuerger&auml;t &uuml;berpr&uuml;fen&bull;Z&uuml;ndkabel auf Besch&auml;digungen &uuml;berpr&uuml;fen&bull;Anschlusskabel f&uuml;r die Ionisationselektrode auf Besch&auml;digungen &uuml;berpr&uuml;fen&bull;Ohmschen Widerstand der Sicherheitsventile an der Gasarmatur messen&bull;Brennereinstellwert bei Nennw&auml;rmebelastung bzw. eingebaute Brennerd&uuml;sen &uuml;berpr&uuml;fen&bull;Brennereinstellwert bei kleinster Leistung &uuml;berpr&uuml;fen&bull;Abgasanlage &uuml;berpr&uuml;fen&bull;Verbrennungsluftversorgung &uuml;berpr&uuml;fen&bull;W&auml;rmeblock auf abgasseitige Ablagerungen &uuml;berpr&uuml;f&uuml;en&bull;",correction => "Hauptabsperreinrichtung &ouml;ffnen&bull;Ger&auml;teabsperrhahn &ouml;ffnen&bull;Ger&auml;t stilllegen und Gasleitung &uuml;berpr&uuml;fen&bull;Signalauswertung auf Leiterplatte defekt&bull;Ionisationselektrode austauschen&bull;Schutzleiteranschluss (PE) im Schaltkasten herstellen&bull;Z&uuml;ndkabel austauschen&bull;Anschlusskabel zur Ionisationselektrode austauschen&bull;Gasarmatur austauschen&bull;Brenner korrekt einstellen bzw. Brennerd&uuml;sen austauschen&bull;Brenner bei minimaler Nennbelastung einstellen&bull;Abgasanlage umbauen&bull;Verbrennungsluftverbund zu klein bzw. zu geringe Gr&ouml;&szlig;e der L&uuml;ftungs&ouml;ffnung&bull;W&auml;rmeblock abgasseitig reinigen&bull;Leiterplatte austauschen"};
	$ErrorCodeList{"1201"} = {errorcode => "EA",                                     subcode => "227",                                                class => "8",           reason => "Kein Flammensignal nach Z&uuml;ndung",                                                                                                                                                                                                                                                                                                             checklist => "Hauptabsperreinrichtung &uuml;berpr&uuml;fen&bull;Ger&auml;teabsperrhan &uuml;berpr&uuml;fen&bull;Gasanschlu&szlig;flie&szlig;druck bei Nennw&auml;rmebelastung messen&bull;Ionisationsstrom messen&bull;Ionisationselektrode &uuml;berpr&uuml;fen&bull;Schutzleiteranschlu&szlig; im Steuerger&auml;t &uuml;berpr&uuml;fen&bull;Z&uuml;ndkabel auf Besch&auml;digungen &uuml;berpr&uuml;fen&bull;Anschlusskabel f&uuml;r die Ionisationselektrode auf Besch&auml;digungen &uuml;berpr&uuml;fen&bull;Ohmschen Widerstand der Sicherheitsventile an der Gasarmatur messen&bull;Brennereinstellwert bei Nennw&auml;rmebelastung bzw. eingebaute Brennerd&uuml;sen &uuml;berpr&uuml;fen&bull;Brennereinstellwert bei kleinster Leistung &uuml;berpr&uuml;fen&bull;Abgasanlage &uuml;berpr&uuml;fen&bull;Verbrennungsluftversorgung &uuml;berpr&uuml;fen&bull;W&auml;rmeblock auf abgasseitige Ablagerungen &uuml;berpr&uuml;f&uuml;en&bull;",correction => "Hauptabsperreinrichtung &ouml;ffnen&bull;Ger&auml;teabsperrhahn &ouml;ffnen&bull;Ger&auml;t stilllegen und Gasleitung &uuml;berpr&uuml;fen&bull;Signalauswertung auf Leiterplatte defekt&bull;Ionisationselektrode austauschen&bull;Schutzleiteranschluss (PE) im Stchaltkasten herstellen&bull;Z&uuml;ndkabel austauschen&bull;Anschlusskabel zur Ionisationselektrode austauschen&bull;Gasarmatur austauschen&bull;Brenner korrekt einstellen bzw. Brennerd&uuml;sen austauschen&bull;Brenner bei minimaler Nennbelastung einstellen&bull;Abgasanlage umbauen&bull;Verbrennungsluftverbund zu klein bzw. zu geringe Gr&ouml;&szlig;e der L&uuml;ftungs&ouml;ffnung&bull;W&auml;rmeblock abgasseitig reinigen&bull;Steuerger&auml;t / Feuerungsautomat austauschen"};
	$ErrorCodeList{"1202"} = {errorcode => "EA",                                     subcode => "229",                                                class => "4",           reason => "Flamme w&auml;hrend  Brennerbetrieb ausgefallen",                                                                                                                                                                                                                                                                                                  checklist => "Hauptabsperreinrichtung &uuml;berpr&uuml;fen&bull;Ger&auml;teabsperrhan &uuml;berpr&uuml;fen&bull;Gasanschlu&szlig;flie&szlig;druck bei Nennw&auml;rmebelastung messen&bull;Ionisationsstrom messen&bull;Ionisationselektrode &uuml;berpr&uuml;fen&bull;Schutzleiteranschlu&szlig; im Steuerger&auml;t &uuml;berpr&uuml;fen&bull;Z&uuml;ndkabel auf Besch&auml;digungen &uuml;berpr&uuml;fen&bull;Anschlusskabel f&uuml;r die Ionisationselektrode auf Besch&auml;digungen &uuml;berpr&uuml;fen&bull;Ohmschen Widerstand der Sicherheitsventile an der Gasarmatur messen&bull;Brennereinstellwert bei Nennw&auml;rmebelastung bzw. eingebaute Brennerd&uuml;sen &uuml;berpr&uuml;fen&bull;Brennereinstellwert bei kleinster Leistung &uuml;berpr&uuml;fen&bull;Abgasanlage &uuml;berpr&uuml;fen&bull;Verbrennungsluftversorgung &uuml;berpr&uuml;fen&bull;W&auml;rmeblock auf abgasseitige Ablagerungen &uuml;berpr&uuml;f&uuml;en&bull;",correction => "Hauptabsperreinrichtung &ouml;ffnen&bull;Ger&auml;teabsperrhahn &ouml;ffnen&bull;Ger&auml;t stilllegen und Gasleitung &uuml;berpr&uuml;fen&bull;Signalauswertung auf Leiterplatte defekt&bull;Ionisationselektrode austauschen&bull;Schutzleiteranschluss (PE) im Schaltkasten herstellen&bull;Z&uuml;ndkabel austauschen&bull;Anschlusskabel zur Ionisationselektrode austauschen&bull;Gasarmatur austauschen&bull;Brenner korrekt einstellen bzw. Brennerd&uuml;sen austauschen&bull;Brenner bei minimaler Nennbelastung einstellen&bull;Abgasanlage umbauen&bull;Verbrennungsluftverbund zu klein bzw. zu geringe Gr&ouml;&szlig;e der L&uuml;ftungs&ouml;ffnung&bull;W&auml;rmeblock abgasseitig reinigen&bull;Steuerger&auml;t/ Feuerungsautomat austauschen"};
	$ErrorCodeList{"1203"} = {errorcode => "EA",                                     subcode => "230",                                                class => "8",           reason => "Interne St&ouml;rung Ger&auml;teelektronik, Modulationsspule St&ouml;rung",                                                                                                                                                                                                                                                                        checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Ionisation pr&uuml;fen (au&szlig;erhalb Grenzwert), Steuerger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"1204"} = {errorcode => "EA",                                     subcode => "234",                                                class => "8",           reason => "Elektrische St&ouml;rung Gasarmatur",                                                                                                                                                                                                                                                                                                              checklist => "Anschlusskabel der Gasarmatur pr&uuml;fen&bull;Gasarmatur &uuml;berpr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                      correction => "Anschlusskabel austauschen&bull;Gasarmatur austauschen"                                                                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"1205"} = {errorcode => "EA",                                     subcode => "252; 253",                                           class => "4",           reason => "Systemst&ouml;rung Ger&auml;teelektronik UBA/Basiscontroller",                                                                                                                                                                                                                                                                                     checklist => "Kontakt zwischen Ger&auml;teelektronik UBA; und Montagefu&szlig; sowie Basiscontroller und der Grundplatte und allen sonstigen Steckverbindungen pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                        correction => "Kontaktproblem beseitigen, ggf. Ger&auml;teelektronik UBA oder Kesselidentifikationsmodul austauschen"                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"1206"} = {errorcode => "EA",                                     subcode => "261",                                                class => "8",           reason => "Zeitfehler bei erster Sicherheitszeit",                                                                                                                                                                                                                                                                                                            checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Steuerger&auml;t/ Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1207"} = {errorcode => "EC",                                     subcode => "251; 256",                                           class => "4",           reason => "Systemst&ouml;rung Ger&auml;teelektronik UBA/Basiscontroller",                                                                                                                                                                                                                                                                                     checklist => "Kontakt zwischen Ger&auml;teelektronik UBA; und Montagefu&szlig; sowie Basiscontroller und der Grundplatte und allen sonstigen Steckverbindungen pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                        correction => "Kontaktproblem beseitigen, ggf. Ger&auml;teelektronik UBA oder Kesselidentifikationsmodul austauschen"                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"1208"} = {errorcode => "EE",                                     subcode => "547",                                                class => "4",           reason => "Systemst&ouml;rung Kodierstecker BIM",                                                                                                                                                                                                                                                                                                             checklist => "Pr&uuml;fen, ob die Ger&auml;teelektronik SAFe defekt ist&bull;Sollte die St&ouml;rung nach Austausch der Ger&auml;teelektronik SAFe weiterhin vorliegen, ist der Kodierstecker defekt",                                                                                                                                                                                                                                                                                                              correction => "Ger&auml;teelektronik SAFe austauschen&bull;Kodierstecker BIM durch Buderus Kundendienst austauschen lassen"                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"1209"} = {errorcode => "EE",                                     subcode => "554",                                                class => "10",          reason => "Systemst&ouml;rung Ger&auml;teelektronik",                                                                                                                                                                                                                                                                                                         checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Ger&auml;teelektronik SAFe austauschen"                                                                                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"1210"} = {errorcode => "EE",                                     subcode => "601",                                                class => "4",           reason => "Systemst&ouml;rung Ger&auml;teelektronik SAFe",                                                                                                                                                                                                                                                                                                    checklist => "F&uuml;hlerleitung zum Kesself&uuml;hler pr&uuml;fen&bull;Steckverbindung pr&uuml;fen&bull;&bull;&bull;F&uuml;hlerwerte nach Tabelle pr&uuml;fen&bull;Spannungswerte am F&uuml;hler laut Tabelle pr&uuml;fen",                                                                                                                                                                                                                                                                                        correction => "Bei Besch&auml;digung austauschen&bull;Bei Verschmutzung reinigen oder ggf. austauschen&bull;Bei Besch&auml;digung austauschen&bull;Bei losem Stecker den Stecker wieder aufstecken&bull;Bei Abweichungen den F&uuml;hler austauschen&bull;Bei Abweichungen die Ger&auml;teelektronik SAFe austauschen"                                                                                          };
	$ErrorCodeList{"1211"} = {errorcode => "EE",                                     subcode => "602",                                                class => "4",           reason => "Messung Abgastemperaturf&uuml;hler SAFe gest&ouml;rt",                                                                                                                                                                                                                                                                                             checklist => "F&uuml;hlerleitung pr&uuml;fen&bull;Steckverbindung pr&uuml;fen&bull;&bull;&bull;F&uuml;hlerwerte nach Tabelle pr&uuml;fen&bull;Spannungswerte am F&uuml;hler laut Tabelle pr&uuml;fen",                                                                                                                                                                                                                                                                                                              correction => "Bei Besch&auml;digung austauschen&bull;Bei Verschmutzung reinigen oder ggf. austauschen&bull;Bei Besch&auml;digung austauschen&bull;Bei losem Stecker den Stecker wieder aufstecken&bull;Bei Abweichungen den F&uuml;hler austauschen&bull;Bei Abweichungen die Ger&auml;teelektronik SAFe austauschen"                                                                                          };
	$ErrorCodeList{"1212"} = {errorcode => "EE",                                     subcode => "603; 604; 605; 606; 607; 608; 609; 610; 611",        class => "4",           reason => "Systemst&ouml;rung Feuerungsautomat SAFe",                                                                                                                                                                                                                                                                                                         checklist => "&Uuml;ber die Bedieneinheit im Servicemen&uuml; unter dem Men&uuml;punkt „Diagnose/Fehler“ (RC35) bzw.; „Diagnose/St&ouml;rungsanzeigen/St&ouml;rungshistorie“ (RC300) pr&uuml;fen, ob diese St&ouml;rung h&auml;ufiger vorhanden ist",                                                                                                                                                                                                                                                       correction => "Wenn diese St&ouml;rung h&auml;ufiger auftritt oder nicht behoben werden kann, ist der Feuerungsautomat SAFe auszutauschen"                                                                                                                                                                                                                                                                      };
	$ErrorCodeList{"1213"} = {errorcode => "EE",                                     subcode => "612",                                                class => "4",           reason => "Systemst&ouml;rung Ger&auml;teelektronik SAFe",                                                                                                                                                                                                                                                                                                    checklist => "Steckverbindung pr&uuml;fen&bull;Kabel zum R&uuml;cklauftemperaturf&uuml;hler pr&uuml;fen&bull;Widerstandswerte des R&uuml;cklauftemperaturf&uuml;hlers pr&uuml;fen&bull;",                                                                                                                                                                                                                                                                                                                           correction => "Bei losem Stecker diesen wieder richtig aufstecken&bull;Bei Besch&auml;digung austauschen; bei Verschmutzung reinigen oder ggf. austauschen&bull;Bei Abweichungen den R&uuml;cklauftemperaturf&uuml;hler austauschen&bull;Regelger&auml;t ausund wieder einschalten"                                                                                                                             };
	$ErrorCodeList{"1214"} = {errorcode => "EE",                                     subcode => "613",                                                class => "4",           reason => "Systemst&ouml;rung Feuerungsautomat (Aufeinanderfolgende Messungen der Vorlauftemperatur weichen zu; stark voneinander ab)",                                                                                                                                                                                                                       checklist => "Widerstandswerte des Vorlauftemperaturf&uuml;hlers pr&uuml;fen&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Bei Abweichungen den Vorlauftemperaturf&uuml;hler austauschen&bull;Regelger&auml;t ausund wieder einschalten"                                                                                                                                                                                                                                                                                    };
	$ErrorCodeList{"1215"} = {errorcode => "EE",                                     subcode => "620; 621; 622; 623",                                 class => "4",           reason => "Systemst&ouml;rung Feuerungsautomat",                                                                                                                                                                                                                                                                                                              checklist => "&Uuml;ber Bedieneinheit im Servicemen&uuml; unter Men&uuml;punkt „Diagnose/Fehler“ (RC35) bzw. „Diagnose/ St&ouml;rungsanzeigen/St&ouml;rungshistorie“ (RC300) pr&uuml;fen, ob diese St&ouml;rung h&auml;ufiger vorhanden ist",                                                                                                                                                                                                                                                               correction => "Wenn diese St&ouml;rung h&auml;ufiger auftritt oder nicht behoben werden kann, ist der Feuerungsautomat auszutauschen"                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"1216"} = {errorcode => "EE",                                     subcode => "625",                                                class => "4",           reason => "Systemst&ouml;rung Feuerungsautomat",                                                                                                                                                                                                                                                                                                              checklist => "Pr&uuml;fen, ob der Winkelhalter evtl. falsch montiert oder verschmutzt ist&bull;&Uuml;ber Bedieneinheit im Servicemen&uuml; unter Men&uuml;punkt „Diagnose/Fehler“ (RC35) bzw. „Diagnose/ St&ouml;rungsanzeigen/St&ouml;rungshistorie“ (RC300) pr&uuml;fen, ob diese St&ouml;rung h&auml;ufiger vorhanden ist",                                                                                                                                                                              correction => "Winkelhalter richtig montieren, reinigen bzw. austauschen&bull;Wenn diese St&ouml;rung h&auml;ufiger auftritt oder nicht behoben werden kann, ist der Feuerungsautomat auszutauschen"                                                                                                                                                                                                            };
	$ErrorCodeList{"1217"} = {errorcode => "EE",                                     subcode => "626",                                                class => "4",           reason => "Systemst&ouml;rung Feuerungsautomat",                                                                                                                                                                                                                                                                                                              checklist => "Kabel zwischen Ger&auml;teelektronik und Z&uuml;ndelektrode auf Unterbrechungen oder Besch&auml;digungen pr&uuml;fen&bull;Z&uuml;ndelektrode auf Besch&auml;digungen pr&uuml;fen&bull;Sollte an den Elektroden oder den Kabeln keine St&ouml;rung vorgelegen haben, ist der Feuerungsautomat defekt",                                                                                                                                                                                                 correction => "Defekte Kabel austauschen bzw. Unterbrechungen beheben&bull;Defekte Elektrode austauschen&bull;Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"1218"} = {errorcode => "EE",                                     subcode => "627",                                                class => "4",           reason => "Systemst&ouml;rung Feuerungsautomat",                                                                                                                                                                                                                                                                                                              checklist => "Diese St&ouml;rung entsteht durch einen defekten Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                                                                                    };
	$ErrorCodeList{"1219"} = {errorcode => "EE",                                     subcode => "630; 631; 640",                                      class => "4",           reason => "Systemst&ouml;rung Feuerungsautomat",                                                                                                                                                                                                                                                                                                              checklist => "&Uuml;ber die Bedieneinheit im Servicemen&uuml;; unter dem Men&uuml;punkt „Diagnose/Fehler“ (RC35) bzw. „Diagnose/St&ouml;rungsanzeigen/St&ouml;rungshistorie“ (RC300) pr&uuml;fen, ob diese St&ouml;rung h&auml;ufiger vorhanden ist",                                                                                                                                                                                                                                                       correction => "Wenn diese St&ouml;rung h&auml;ufiger auftritt oder nicht behoben werden kann, ist der Feuerungsautomat auszutauschen"                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"1220"} = {errorcode => "EE",                                     subcode => "641",                                                class => "4",           reason => "Systemst&ouml;rung Feuerungsautomat",                                                                                                                                                                                                                                                                                                              checklist => "&Uuml;ber die Bedieneinheit im Servicemen&uuml;; unter dem Men&uuml;punkt „Diagnose/Fehler“ (RC35) bzw. „Diagnose/St&ouml;rungsanzeigen/St&ouml;rungshistorie“ (RC300) pr&uuml;fen, ob diese St&ouml;rung h&auml;ufiger vorhanden ist",                                                                                                                                                                                                                                                       correction => "Wenn diese St&ouml;rung h&auml;ufiger auftritt oder nicht behoben werden kann, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"1221"} = {errorcode => "EE",                                     subcode => "650",                                                class => "4",           reason => "Kodierstecker BIM defekt",                                                                                                                                                                                                                                                                                                                         checklist => "&Uuml;ber die Bedieneinheit im Servicemen&uuml;; unter dem Men&uuml;punkt „Diagnose/Fehler“ (RC35) bzw. „Diagnose/St&ouml;rungsanzeigen/ St&ouml;rungshistorie“ (RC300) pr&uuml;fen, ob; diese St&ouml;rung h&auml;ufiger vorhanden ist",                                                                                                                                                                                                                                                     correction => "Kodierstecker BIM austauschen"                                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"1222"} = {errorcode => "EE",                                     subcode => "651",                                                class => "4",           reason => "Systemst&ouml;rung Feuerungsautomat SAFe und Kodierstecker defekt",                                                                                                                                                                                                                                                                                checklist => "ACHTUNG:; Diese St&ouml;rung kann nicht entriegelt werden!",                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Feuerungsautomat und Kodierstecker BIM austauschen"                                                                                                                                                                                                                                                                                                                                              };
	$ErrorCodeList{"1223"} = {errorcode => "EE",                                     subcode => "652",                                                class => "4",           reason => "Kodierstecker defekt",                                                                                                                                                                                                                                                                                                                             checklist => "&Uuml;ber die Bedieneinheit im Servicemen&uuml;; unter dem Men&uuml;punkt „Diagnose/Fehler“ (RC35) bzw. „Diagnose/St&ouml;rungsanzeigen/St&ouml;rungshistorie“ (RC300) pr&uuml;fen, ob; diese St&ouml;rung h&auml;ufiger vorhanden ist",                                                                                                                                                                                                                                                      correction => "Wenn die St&ouml;rung h&auml;ufiger auftritt oder nicht behoben werden kann, ist der Kodierstecker BIM auszutauschen"                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1224"} = {errorcode => "EE",                                     subcode => "653",                                                class => "4",           reason => "Systemst&ouml;rung Feuerungsautomat",                                                                                                                                                                                                                                                                                                              checklist => "&Uuml;ber die Bedieneinheit im Servicemen&uuml;; unter dem Men&uuml;punkt „Diagnose/Fehler“ (RC35) bzw. „Diagnose/St&ouml;rungsanzeigen/St&ouml;rungshistorie“ (RC300) pr&uuml;fen, ob; diese St&ouml;rung h&auml;ufiger vorhanden ist",                                                                                                                                                                                                                                                      correction => "Feuerungsautomat SAFe austauschen"                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1225"} = {errorcode => "EE",                                     subcode => "654; 655; 656; 657; 658; 659",                       class => "4",           reason => "Systemst&ouml;rung Feuerungsautomat",                                                                                                                                                                                                                                                                                                              checklist => "&Uuml;ber die Bedieneinheit im Servicemen&uuml;; unter dem Men&uuml;punkt „Diagnose/Fehler“ (RC35) bzw. „Diagnose/St&ouml;rungsanzeigen/St&ouml;rungshistorie“ (RC300) pr&uuml;fen, ob diese; St&ouml;rung h&auml;ufiger vorhanden ist",                                                                                                                                                                                                                                                      correction => "Wenn diese St&ouml;rung h&auml;ufiger auftritt oder nicht behoben werden kann, ist der Feuerungsautomat auszutauschen"                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"1226"} = {errorcode => "EE",                                     subcode => "660",                                                class => "4",           reason => "Kodierstecker BIM Kommunikation gest&ouml;rt",                                                                                                                                                                                                                                                                                                     checklist => "Pr&uuml;fen, ob die Ger&auml;teelektronik richtig montiert ist",                                                                                                                                                                                                                                                                                                                                                                                                                                      correction => "Sollten Feuerungsautomat und Kodierstecker offensichtlich richtig montiert sein, zun&auml;chst Ger&auml;teelektronik austauschen; Wird die St&ouml;rung dadurch nicht behoben, dann Kodierstecker austauschen"                                                                                                                                                                                   };
	$ErrorCodeList{"1227"} = {errorcode => "EE",                                     subcode => "661",                                                class => "4",           reason => "Systemst&ouml;rung BIM Kodierstecker",                                                                                                                                                                                                                                                                                                             checklist => "Zun&auml;chst Feuerungsautomat austauschen und pr&uuml;fen, ob die St&ouml;rung dadurch behoben wurde",                                                                                                                                                                                                                                                                                                                                                                                               correction => "Sollte die St&ouml;rung nicht behoben sein, den Buderus Kundendiensts informieren"                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1228"} = {errorcode => "EE",                                     subcode => "662",                                                class => "4",           reason => "Kodierstecker BIM kann nicht gelesen werden",                                                                                                                                                                                                                                                                                                      checklist => "Pr&uuml;fen, ob Ger&auml;teelektronik und Kodierstecker richtig montiert sind",                                                                                                                                                                                                                                                                                                                                                                                                                       correction => "Sollten Ger&auml;teelektronik und Kodierstecker offensichtlich richtig montiert sein, zun&auml;chst Ger&auml;teelektronik austauschen; Wird die St&ouml;rung dadurch nicht behoben, dann Kodierstecker austauschen"                                                                                                                                                                              };
	$ErrorCodeList{"1229"} = {errorcode => "EF",                                     subcode => "254",                                                class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik/Basiscontroller (UBA3 oder KIM)",                                                                                                                                                                                                                                                                         checklist => "Kontakt zwischen Ger&auml;teelektronik und Montagefu&szlig; sowie Basiscontroler und der Grundplatte und allen sonstigen Steckverbindungen pr&uuml;fen&bull;Hybridsystem: Bauteilfehler im Hybridmodul",                                                                                                                                                                                                                                                                                              correction => "Kontaktproblem beseitigen, ggf. Ger&auml;teelektronik oder Kesselidentifikationsmodul austauschen&bull;Hybridmodul pr&uuml;fen und ggf. ersetzen"                                                                                                                                                                                                                                                };
	$ErrorCodeList{"1230"} = {errorcode => "EF",                                     subcode => "345",                                                class => "8",           reason => "Kein Temperaturanstieg nach Brennerstart",                                                                                                                                                                                                                                                                                                         checklist => "Sind alle Absperrh&auml;hne ge&ouml;ffnet?&bull;Ist Luft im System?&bull;Ist mindestens ein Thermostatventil ge&ouml;ffnet?&bull;Umw&auml;lzpumpe pr&uuml;fen&bull;Sicherheitsund Vorlauftemperaturf&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                                                          correction => "Absperrh&auml;hne &ouml;ffnen&bull;Anlage entl&uuml;ften&bull;Ein Thermostatventil &ouml;ffnen&bull;Ggf. Umw&auml;lzpumpe tauschen&bull;Ggf. Sicherheitsoder Vorlauftemperaturf&uuml;hler austauschen"                                                                                                                                                                                           };
	$ErrorCodeList{"1231"} = {errorcode => "EF",                                     subcode => "349",                                                class => "4",           reason => "Differenz zwischen Vorlaufund R&uuml;cklauftemperatur zu gro&szlig;",                                                                                                                                                                                                                                                                              checklist => "Sind alle Absperrh&auml;hne ge&ouml;ffnet?&bull;Betr&auml;gt der Wasserdruck in der Anlage mindestens 1 bar?&bull;Ist mindestens ein Thermostatventil ge&ouml;ffnet?&bull;Vorlaufund R&uuml;cklaufsensor pr&uuml;fen&bull;Umw&auml;lzpumpe pr&uuml;fen",                                                                                                                                                                                                                                              correction => "Absperrh&auml;hne &ouml;ffnen&bull;Bei zu geringem Wasserdruck, Wasser nachf&uuml;llen und Anlage entl&uuml;ften.&bull;Ein Thermostatventil &ouml;ffnen&bull;Ggf. Vorlauf-oder R&uuml;cklaufsensor austauschen&bull;Ggf. Umw&auml;lzpumpe austauschen"                                                                                                                                           };
	$ErrorCodeList{"1232"} = {errorcode => "EF",                                     subcode => "553",                                                class => "8",           reason => "Zu viele Flammenabrisse",                                                                                                                                                                                                                                                                                                                          checklist => "O2-Wert in der Verbrennungsluft zu niedrig&bull;Flammenf&uuml;hler verschmutzt&bull;Spiegel vom Winkelhalter verschmutzt&bull;Flammenf&uuml;hler defekt&bull;Abweichende Brennereinstellwerte&bull;Signalauswertung im Feuerungsautomat (SAFe) defekt",                                                                                                                                                                                                                                               correction => "Undichtheit im Abgassystem beseitigen&bull;Flammenf&uuml;hler reinigen&bull;Spiegel vom Winkelhalter reinigen&bull;Flammenf&uuml;hler austauschen&bull;Brenner korrekt einstellen&bull;Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                      };
	$ErrorCodeList{"1233"} = {errorcode => "EH",                                     subcode => "250; 258; 262",                                      class => "4",           reason => "Systemst&ouml;rung Ger&auml;teelektronik UBA/Basiscontroller",                                                                                                                                                                                                                                                                                     checklist => "Kontakt zwischen Ger&auml;teelektronik UBA; und Montagefu&szlig; sowie Basiscontroller BC und der Grundplatte und allen sonstigen Steckverbindungen pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                     correction => "Kontaktproblem beseitigen, ggf. Ger&auml;teelektronik UBA oder Kesselidentifikationsmodul KIM austauschen"                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1234"} = {errorcode => "EL",                                     subcode => "259",                                                class => "4",           reason => "Systemst&ouml;rung Ger&auml;teelektronik UBA/Basiscontroller",                                                                                                                                                                                                                                                                                     checklist => "Kontakt zwischen Ger&auml;teelektronik UBA und Montagefu&szlig; sowie Basiscontroller; und der Grundplatte und allen sonstigen Steckverbindungen pr&uuml;fen&bull;Kesselidentifikationsmodul nicht richtig aufgesteckt oder defekt",                                                                                                                                                                                                                                                                  correction => "Kontaktproblem beseitigen, ggf. Ger&auml;teelektronik UBA oder Ger&auml;teelektronik BC25 austauschen&bull;Kesselidentifikationsmodul richtig aufstecken oder durch Buderus Service austauschen lassen"                                                                                                                                                                                          };
	$ErrorCodeList{"1235"} = {errorcode => "EL",                                     subcode => "279; 290",                                           class => "4",           reason => "Systemst&ouml;rung Ger&auml;teelektronik UBA/Basiscontroller",                                                                                                                                                                                                                                                                                     checklist => "Kontakt zwischen Ger&auml;teelektronik UBA und Montagefu&szlig; sowie Basiscontroller; und der Grundplatte und allen sonstigen Steckverbindungen pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                        correction => "Kontaktproblem beseitigen, ggf. Ger&auml;teelektronik UBA oder Kesselidentifikationsmodul austauschen"                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"1236"} = {errorcode => "EP",                                     subcode => "287",                                                class => "4",           reason => "Systemst&ouml;rung Ger&auml;teelektronik UBA/Basiscontroller",                                                                                                                                                                                                                                                                                     checklist => "Kontakt zwischen Ger&auml;teelektronik UBA und Montagefu&szlig; sowie Basiscontroller; und der Grundplatte und allen sonstigen Steckverbindungen pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                        correction => "Kontaktproblem beseitigen, ggf. Ger&auml;teelektronik UBA oder Kesselidentifikationsmodul austauschen"                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"1237"} = {errorcode => "EU",                                     subcode => "690",                                                class => "4",           reason => "Relais im Umschaltmodul schaltet nicht nach Vorgabe",                                                                                                                                                                                                                                                                                              checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Umschaltmodul austauschen"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1238"} = {errorcode => "EU",                                     subcode => "691",                                                class => "4",           reason => "R&uuml;ckmeldung vom Umschaltmodul, obwohl Relais im Umschaltmodul nicht angesteuert wird",                                                                                                                                                                                                                                                        checklist => "Anschl&uuml;sse des Moduls auf Kurzschluss pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Kurzschluss beheben, ggf. Umschaltmodul austauschen"                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1239"} = {errorcode => "EU",                                     subcode => "692; 693; 694; 695; 696; 697; 698; 699",             class => "4",           reason => "Systemst&ouml;rung",                                                                                                                                                                                                                                                                                                                               checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Umschaltmodul austauschen"                                                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1240"} = {errorcode => "EY",                                     subcode => "263",                                                class => "4",           reason => "Systemst&ouml;rung Ger&auml;teelektronik UBA/Basiscontroller",                                                                                                                                                                                                                                                                                     checklist => "Kontakt zwischen Ger&auml;teelektronik UBA; und Montagefu&szlig; sowie Basiscontroller BC10 und der Grundplatte und allen sonstigen Steckverbindungen pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                   correction => "Kontaktproblem beseitigen, ggf. Ger&auml;teelektronik oder Kesselidentifikationsmodul austauschen"                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1241"} = {errorcode => "EY",                                     subcode => "390",                                                class => "8",           reason => "Fehler bei der Initialisierung der 3WV",                                                                                                                                                                                                                                                                                                           checklist => "Falsche Werte in der KIM definiert",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Kessel-Idendifikations-Modul/Kodierstecker austauschen"                                                                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"1242"} = {errorcode => "F0",                                     subcode => "25",                                                 class => "n.a.",        reason => "Interne St&ouml;rung, W&auml;rmeerzeuger",                                                                                                                                                                                                                                                                                                         checklist => "Interner Fehler beim &Uuml;berpr&uuml;fen des Abgassensors an der Str&ouml;mungssicherung",                                                                                                                                                                                                                                                                                                                                                                                                           correction => "Leiterplatte austauschen"                                                                                                                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"1243"} = {errorcode => "F0",                                     subcode => "35",                                                 class => "n.a.",        reason => "Interne St&ouml;rung, W&auml;rmeerzeuger",                                                                                                                                                                                                                                                                                                         checklist => "Interner Fehler auf de Leiterplatte",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "Leiterplatte austauschen"                                                                                                                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"1244"} = {errorcode => "F0",                                     subcode => "51",                                                 class => "n.a.",        reason => "Interne St&ouml;rung, W&auml;rmeerzeuger",                                                                                                                                                                                                                                                                                                         checklist => "Interner Fehler im Programmablauf&bull;Z&uuml;ndkabel ist besch&auml;digt&bull;Anschlusskabel zu den Sicherheitsventilen in der Gasarmatur besch&auml;digt&bull;Interner Fehler auf der Leiterplatte",                                                                                                                                                                                                                                                                                                correction => "Reset &uuml;ber Servicefunktion 8.E durchf&uuml;hren&bull;Z&uuml;ndkabel austauschen&bull;Anchlusskabel zu den Sicherheitsventilen in der Gasarmatur austauschen&bull;Leiterplatte austauschen"                                                                                                                                                                                                  };
	$ErrorCodeList{"1245"} = {errorcode => "F0",                                     subcode => "237",                                                class => "8",           reason => "Systemst&ouml;rung",                                                                                                                                                                                                                                                                                                                               checklist => "Interner Fehler im Steuerger&auml;t&bull;Anschlussstecker am Kodierstecker ist nicht komplett aufgesteckt",                                                                                                                                                                                                                                                                                                                                                                                           correction => "Steuerger&auml;t austauschen&bull;Anschlussstecker am Kodierstecker komplett aufstecken"                                                                                                                                                                                                                                                                                                         };
	$ErrorCodeList{"1246"} = {errorcode => "F0",                                     subcode => "238",                                                class => "8",           reason => "Ger&auml;teelektronik ist defekt",                                                                                                                                                                                                                                                                                                                 checklist => "Anschlussstecker an der Gasarmatur ist nicht vollst&auml;ndig aufgesteckt&bull;Anschlusskabel zu den Sicherheitsventilen in der Gasarmatur sind besch&auml;digt&bull;Signauswertung im Steuerger&auml;t/ Feuerungsautomat defekt",                                                                                                                                                                                                                                                                    correction => "Anschlussstecker an der Gasarmatur komplett aufstecken&bull;Anschlusskabel zu den Sicherheitsventilen in der Gasarmatur austauschen&bull;Steuerger&auml;t/Feerungsautomat austauschen"                                                                                                                                                                                                           };
	$ErrorCodeList{"1247"} = {errorcode => "F0",                                     subcode => "239",                                                class => "8",           reason => "Systemst&ouml;rung",                                                                                                                                                                                                                                                                                                                               checklist => "Anschlussstecker an der Gasarmatur ist nicht vollst&auml;ndig aufgesteckt&bull;Anschlusskabel zu den Sicherheitsventilen in der Gasarmatur sind besch&auml;digt&bull;Signauswertung im Steuerger&auml;t/ Feuerungsautomat defekt",                                                                                                                                                                                                                                                                    correction => "Anschlussstecker an der Gasarmatur komplett aufstecken&bull;Anschlusskabel zu den Sicherheitsventilen in der Gasarmatur austauschen&bull;Steuerger&auml;t/Feuerungsautomat austauschen"                                                                                                                                                                                                          };
	$ErrorCodeList{"1248"} = {errorcode => "F0",                                     subcode => "242",                                                class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik/Basiscontroller",                                                                                                                                                                                                                                                                                         checklist => "Interner Fehler im Steuerger&auml;t/ Feuerungautomat",                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Steuerger&auml;t/Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"1249"} = {errorcode => "F0",                                     subcode => "243",                                                class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik/Basiscontroller",                                                                                                                                                                                                                                                                                         checklist => "Interner Fehler im Steuerger&auml;t/ Feuerungautomat",                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Steuerger&auml;t/Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"1250"} = {errorcode => "F0",                                     subcode => "244",                                                class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik/Basiscontroller",                                                                                                                                                                                                                                                                                         checklist => "Interner Fehler im Steuerger&auml;t/ Feuerungautomat",                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Steuerger&auml;t/Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"1251"} = {errorcode => "F0",                                     subcode => "245",                                                class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik/Basiscontroller",                                                                                                                                                                                                                                                                                         checklist => "Interner Fehler im Steuerger&auml;t/ Feuerungautomat",                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Steuerger&auml;t/Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"1252"} = {errorcode => "F0",                                     subcode => "246",                                                class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik/Basiscontroller",                                                                                                                                                                                                                                                                                         checklist => "Interner Fehler im Steuerger&auml;t/ Feuerungautomat",                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Steuerger&auml;t/Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"1253"} = {errorcode => "F0",                                     subcode => "247",                                                class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik/Basiscontroller",                                                                                                                                                                                                                                                                                         checklist => "Interner Fehler im Steuerger&auml;t/ Feuerungautomat",                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Steuerger&auml;t/Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"1254"} = {errorcode => "F0",                                     subcode => "248",                                                class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik/Basiscontroller",                                                                                                                                                                                                                                                                                         checklist => "Interner Fehler im Steuerger&auml;t/ Feuerungautomat",                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Steuerger&auml;t/Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"1255"} = {errorcode => "F0",                                     subcode => "249",                                                class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik/Basiscontroller",                                                                                                                                                                                                                                                                                         checklist => "Interner Fehler im Steuerger&auml;t/ Feuerungautomat",                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Steuerger&auml;t/Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"1256"} = {errorcode => "F0",                                     subcode => "250",                                                class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik/Basiscontroller",                                                                                                                                                                                                                                                                                         checklist => "Interner Fehler im Steuerger&auml;t/ Feuerungautomat",                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Steuerger&auml;t/Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"1257"} = {errorcode => "F0",                                     subcode => "251",                                                class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik/Basiscontroller",                                                                                                                                                                                                                                                                                         checklist => "Interner Fehler im Steuerger&auml;t/ Feuerungautomat",                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Steuerger&auml;t/Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"1258"} = {errorcode => "F0",                                     subcode => "252",                                                class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik/Basiscontroller",                                                                                                                                                                                                                                                                                         checklist => "Interner Fehler im Steuerger&auml;t/ Feuerungautomat",                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Steuerger&auml;t/Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"1259"} = {errorcode => "F0",                                     subcode => "253",                                                class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik/Basiscontroller",                                                                                                                                                                                                                                                                                         checklist => "Interner Fehler im Steuerger&auml;t/ Feuerungautomat",                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Steuerger&auml;t/Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"1260"} = {errorcode => "F0",                                     subcode => "254",                                                class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik/Basiscontroller",                                                                                                                                                                                                                                                                                         checklist => "Interner Fehler im Steuerger&auml;t/ Feuerungautomat",                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Steuerger&auml;t/Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"1261"} = {errorcode => "F0",                                     subcode => "255",                                                class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik/Basiscontroller",                                                                                                                                                                                                                                                                                         checklist => "Interner Fehler im Steuerger&auml;t/ Feuerungautomat",                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Steuerger&auml;t/Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"1262"} = {errorcode => "F0",                                     subcode => "256",                                                class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik/Basiscontroller",                                                                                                                                                                                                                                                                                         checklist => "Interner Fehler im Steuerger&auml;t/Feuerungsautomat&bull;Anschlussstecker am Kessel-IdendifikationsModul/Kodierstecker ist nicht komplett aufgesteckt&bull;Anschlussstecker f&uuml;r das Kessel-IdendifikationsModul/Kodierstecker ist nicht am Steuer-ger&auml;t/ Feuerungsautomat aufgesteckt",                                                                                                                                                                                                    correction => "Steuerger&auml;t/Feuerungsautomat austauschen&bull;Anschlussstecker am Kessel-IdendifikationsModul/Kodierstecker komplett aufstecken&bull;Anschlussstecker f&uuml;r das Kessel-IdendifikationsModul/Kodierstecker am Steuerger&auml;t/Feuerungsautomat komplett aufstecken"                                                                                                                      };
	$ErrorCodeList{"1263"} = {errorcode => "F0",                                     subcode => "258; 259; 262",                                      class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik/Basiscontroller",                                                                                                                                                                                                                                                                                         checklist => "Interner Fehler im Steuerger&auml;t/ Feuerungautomat",                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Steuerger&auml;t/Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"1264"} = {errorcode => "F0",                                     subcode => "263",                                                class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik/Basiscontroller",                                                                                                                                                                                                                                                                                         checklist => "Interner Fehler im Steuerger&auml;t/Feuerungs-automat&bull;Anschlussstecker am Kessel-IdendifikationsModul/ Kodierstecker ist nicht komplett aufgesteckt&bull;Anschlussstecker f&uuml;r das Kessel-IdendifikationsModul/Kodierstecker ist nicht am Steuerger&auml;t/Feuerungsautomat aufgesteckt",                                                                                                                                                                                                    correction => "Steuerger&auml;t/Feuerungsautomat austauschen&bull;Anschlussstecker am Kessel-IdendifikationsModul/Kodierstecker komplett aufstecken&bull;Anschlussstecker f&uuml;r das Kessel-IdendifikationsModul/Kodierstecker am Steuerger&auml;t/Feuerungsautomat komplett aufstecken"                                                                                                                      };
	$ErrorCodeList{"1265"} = {errorcode => "F0",                                     subcode => "272",                                                class => "8",           reason => "Systemst&ouml;rung",                                                                                                                                                                                                                                                                                                                               checklist => "Interner Fehler im Steuerger&auml;t/Feuerung-automat",                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Steuerger&auml;t/Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"1266"} = {errorcode => "F0",                                     subcode => "278",                                                class => "8",           reason => "F&uuml;hlertest fehlgeschlagen",                                                                                                                                                                                                                                                                                                                   checklist => "Kabelbaum besch&auml;digt",                                                                                                                                                                                                                                                                                                                                                                                                                                                                           correction => "Kabelbaum oder entsprechenden Teil davon ersetzen"                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1267"} = {errorcode => "F0",                                     subcode => "279",                                                class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik/Basiscontroller",                                                                                                                                                                                                                                                                                         checklist => "Interner Fehler im Steuerger&auml;t/ Feuerungautomat",                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Steuerger&auml;t austauschen"                                                                                                                                                                                                                                                                                                                                                                    };
	$ErrorCodeList{"1268"} = {errorcode => "F0",                                     subcode => "280",                                                class => "8",           reason => "Zeitfehler bei Wieder-anlaufversuch",                                                                                                                                                                                                                                                                                                              checklist => "Fehler im Programmablauf des Steuerger&auml;tes/ Feuerungsautomat",                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Steuerger&auml;t/Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"1269"} = {errorcode => "F0",                                     subcode => "287",                                                class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik/Basiscontroller",                                                                                                                                                                                                                                                                                         checklist => "Kurzschluss an der Ionisationselektrode&bull;Anschlusskabel zur Ionisationselektrode besch&auml;digt&bull;Interner Fehler im Steuerger&auml;t",                                                                                                                                                                                                                                                                                                                                                       correction => "Ionisationselektrode austauschen&bull;Anschlusskabel zur Ionisationselektrode austauschen&bull;Steuerger&auml;t austauschen"                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"1270"} = {errorcode => "F0",                                     subcode => "290",                                                class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik/Basiscontroller",                                                                                                                                                                                                                                                                                         checklist => "Interner Fehler im Steuerger&auml;t/ Feuerungautomat",                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Steuerger&auml;t/Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"1271"} = {errorcode => "F0",                                     subcode => "338",                                                class => "8",           reason => "Zu viele erfolglose Brennerstartversuche",                                                                                                                                                                                                                                                                                                         checklist => "Sind alle Absperrh&auml;hne ge&ouml;ffnet?&bull;Ist Luft im System?&bull;Ist mindestens ein Thermostatventil ge&ouml;ffnet?&bull;Umw&auml;lzpumpe pr&uuml;fen&bull;Sicherheitsund Vorlauftemperaturf&uuml;hler pr&uuml;fen",                                                                                                                                                                                                                                                                          correction => "Absperrh&auml;hne &ouml;ffnen&bull;Anlage entl&uuml;ften&bull;Ein Thermostatventil &ouml;ffnen&bull;Ggf. Umw&auml;lzpumpe tauschen&bull;Ggf. Sicherheitsoder Vorlauftemperaturf&uuml;hler tauschen"                                                                                                                                                                                              };
	$ErrorCodeList{"1272"} = {errorcode => "F0",                                     subcode => "356",                                                class => "4",           reason => "Versorgungsspannung; f&uuml;r den W&auml;rmeerzeuger ist zu niedrig",                                                                                                                                                                                                                                                                              checklist => "Netzspannung liegt unter 196 VAC",                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "Versorgungsspannung von mindestens 196 VAC herstellen"                                                                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"1273"} = {errorcode => "F0",                                     subcode => "363",                                                class => "-",           reason => "Systemst&ouml;rung Ger&auml;teelektronik/Basiscontroller",                                                                                                                                                                                                                                                                                         checklist => "Fehler beim Test des Ionisationssignales",                                                                                                                                                                                                                                                                                                                                                                                                                                                            correction => "Steuerger&auml;t/Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"1274"} = {errorcode => "F0",                                     subcode => "381",                                                class => "4",           reason => "W&auml;rmeanforderung blockiert weil 24 h vorbei",                                                                                                                                                                                                                                                                                                 checklist => "Kurze Unterbrechung der W&auml;rmeversorgung nach 24 Stunden (keine St&ouml;rung)",                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1275"} = {errorcode => "F0",                                     subcode => "500",                                                class => "8",           reason => "Interne St&ouml;rung Fremdbrennermodul, keine Spannung Sicherheitsrelais",                                                                                                                                                                                                                                                                         checklist => "1. Feuerungsautomat entriegeln&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Fehler nicht reproduzierbar&bull;Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1276"} = {errorcode => "F0",                                     subcode => "501",                                                class => "8",           reason => "Interne St&ouml;rung Fremdbrennermodul, Sicherheitsrelais h&auml;ngt",                                                                                                                                                                                                                                                                             checklist => "1. Feuerungsautomat entriegeln&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Fehler nicht reproduzierbar&bull;Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1277"} = {errorcode => "F0",                                     subcode => "502",                                                class => "8",           reason => "Interne St&ouml;rung Fremdbrennermodul, keine; Spannung Brennstoffrelais 1",                                                                                                                                                                                                                                                                       checklist => "Relais f&uuml;r das Magnetventil 1 im Feuerungsautomat (SAFe) schaltet nicht durch&bull;Feuerungsautomat (SAFe) defekt",                                                                                                                                                                                                                                                                                                                                                                              correction => "2. Feuerungsautomat (SAFe) austauschen&bull;Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                 };
	$ErrorCodeList{"1278"} = {errorcode => "F0",                                     subcode => "503",                                                class => "8",           reason => "Interne St&ouml;rung Fremdbrennermodul, Brennstoffrelais 1 h&auml;ngt",                                                                                                                                                                                                                                                                            checklist => "1. Feuerungsautomat entriegeln&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Fehler nicht reproduzierbar&bull;Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1279"} = {errorcode => "F0",                                     subcode => "508",                                                class => "8",           reason => "Flammensignal zu hoch",                                                                                                                                                                                                                                                                                                                            checklist => "Flammenf&uuml;hlerstrom au&szlig;erhalb des Mess-bereiches im Feuerngsautomat/1. Feuerungsautomat entriegeln&bull;",                                                                                                                                                                                                                                                                                                                                                                                  correction => "2. Feuerungsautomat (SAFe) austauschen&bull;Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                 };
	$ErrorCodeList{"1280"} = {errorcode => "F0",                                     subcode => "509",                                                class => "8",           reason => "Eingang Flammenf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                                checklist => "Fremdlicht vorhanden&bull;Flammenf&uuml;hler defekt&bull;Fehler bei der Auswertung des Flammensignals im Feuerungsautomat (SAFe)",                                                                                                                                                                                                                                                                                                                                                                    correction => "Ursache f&uuml;r Fremdlicht beseitigen&bull;Flammenf&uuml;hler austauschen&bull;Feuerungsautomat (Safe) austauschen"                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1281"} = {errorcode => "F0",                                     subcode => "547",                                                class => "8",           reason => "Systemst&ouml;rung Kodierstecker",                                                                                                                                                                                                                                                                                                                 checklist => "Falsche Brenner-Idendifikationsmodul montiert&bull;Interner Fehler im Feuerungsautomat (SAFe)",                                                                                                                                                                                                                                                                                                                                                                                                       correction => "Brenner-Idendifikationsmodul austauschen&bull;Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                                      };
	$ErrorCodeList{"1282"} = {errorcode => "F0",                                     subcode => "554",                                                class => "4",           reason => "Systemst&ouml;rung Ger&auml;teelektronik",                                                                                                                                                                                                                                                                                                         checklist => "Interner Fehler im Feuerungsautomat (SAFe)",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1283"} = {errorcode => "F0",                                     subcode => "586",                                                class => "8",           reason => "Ger&auml;teelektronik alter Softwarestand",                                                                                                                                                                                                                                                                                                        checklist => "Parameter vom Schaltmodul UX 15 k&ouml;nnen mit der vorhandenen Software im Feuerungsautomat (SAFe) nicht verarbeitet werden",                                                                                                                                                                                                                                                                                                                                                                        correction => "Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1284"} = {errorcode => "F0",                                     subcode => "601",                                                class => "8",           reason => "Systemst&ouml;rung Ger&auml;teelektronik",                                                                                                                                                                                                                                                                                                         checklist => "Signalumwandlung im Feuerungsautomat (SAFe) defekt",                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1285"} = {errorcode => "F0",                                     subcode => "602",                                                class => "8",           reason => "Messung Abgastemperaturf&uuml;hler gest&ouml;rt",                                                                                                                                                                                                                                                                                                  checklist => "Signalumwandlung im Feuerungsautomat (SAFe) defekt",                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1286"} = {errorcode => "F0",                                     subcode => "603",                                                class => "8",           reason => "Systemst&ouml;rung Feuerungsautomat",                                                                                                                                                                                                                                                                                                              checklist => "Signalumwandlung im Feuerungsautomat (SAFe) defekt",                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1287"} = {errorcode => "F0",                                     subcode => "604; 605",                                           class => "8",           reason => "Systemst&ouml;rung Feuerungsautomat",                                                                                                                                                                                                                                                                                                              checklist => "Unzul&auml;ssige Referenzspannung f&uuml;r den ersten/zweiten Mikroprozessor im Feuerungsautomat (SAFe)",                                                                                                                                                                                                                                                                                                                                                                                             correction => "Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1288"} = {errorcode => "F0",                                     subcode => "606; 607",                                           class => "8",           reason => "Systemst&ouml;rung Feuerungsautomat",                                                                                                                                                                                                                                                                                                              checklist => "Kein F&uuml;hlertest durch den ersten/zweiten Mikroprozessor im Feuerungsautomat (SAFe)",                                                                                                                                                                                                                                                                                                                                                                                                             correction => "Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1289"} = {errorcode => "F0",                                     subcode => "608",                                                class => "8",           reason => "Systemst&ouml;rung Feuerungsautomat",                                                                                                                                                                                                                                                                                                              checklist => "Interner Fehler im Feuerungsautomat (SAFe)",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1290"} = {errorcode => "F0",                                     subcode => "610",                                                class => "8",           reason => "Systemst&ouml;rung Feuerungsautomat",                                                                                                                                                                                                                                                                                                              checklist => "Interner Fehler im Feuerungsautomat (SAFe)",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1291"} = {errorcode => "F0",                                     subcode => "611",                                                class => "8",           reason => "Systemst&ouml;rung Feuerungsautomat",                                                                                                                                                                                                                                                                                                              checklist => "Interner Fehler im Feuerungsautomat (SAFe)",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1292"} = {errorcode => "F0",                                     subcode => "612",                                                class => "8",           reason => "Systemst&ouml;rung Feuerungsautomat",                                                                                                                                                                                                                                                                                                              checklist => "Interner Fehler im Feuerungsautomat (SAFe)",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1293"} = {errorcode => "F0",                                     subcode => "613",                                                class => "8",           reason => "Systemst&ouml;rung Feuerungsautomat",                                                                                                                                                                                                                                                                                                              checklist => "Interner Fehler im Feuerungsautomat (SAFe)",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1294"} = {errorcode => "F0",                                     subcode => "620",                                                class => "8",           reason => "Systemst&ouml;rung Feuerungsautomat",                                                                                                                                                                                                                                                                                                              checklist => "Interner Fehler im Feuerungsautomat (SAFe)",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1295"} = {errorcode => "F0",                                     subcode => "621",                                                class => "8",           reason => "Systemst&ouml;rung Feuerungsautomat",                                                                                                                                                                                                                                                                                                              checklist => "Interner Fehler im Feuerungsautomat (SAFe)",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1296"} = {errorcode => "F0",                                     subcode => "622",                                                class => "8",           reason => "Systemst&ouml;rung Feuerungsautomat",                                                                                                                                                                                                                                                                                                              checklist => "Interner Fehler im Feuerungsautomat (SAFe)",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1297"} = {errorcode => "F0",                                     subcode => "623",                                                class => "8",           reason => "Systemst&ouml;rung Feuerungsautomat",                                                                                                                                                                                                                                                                                                              checklist => "Interner Fehler im Feuerungsautomat (SAFe)",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1298"} = {errorcode => "F0",                                     subcode => "625",                                                class => "8",           reason => "Systemst&ouml;rung Feuerungsautomat",                                                                                                                                                                                                                                                                                                              checklist => "Interner Fehler im Feuerungsautomat (SAFe)",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1299"} = {errorcode => "F0",                                     subcode => "627",                                                class => "8",           reason => "Systemst&ouml;rung Feuerungsautomat",                                                                                                                                                                                                                                                                                                              checklist => "Interner Fehler im Feuerungsautomat (SAFe)",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1300"} = {errorcode => "F0",                                     subcode => "631",                                                class => "8",           reason => "Systemst&ouml;rung Feuerungsautomat",                                                                                                                                                                                                                                                                                                              checklist => "Interner Fehler im Feuerungsautomat (SAFe)",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1301"} = {errorcode => "F0",                                     subcode => "632",                                                class => "8",           reason => "Systemst&ouml;rung Feuerungsautomat",                                                                                                                                                                                                                                                                                                              checklist => "Interner Fehler im Feuerungsautomat (SAFe)",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1302"} = {errorcode => "F0",                                     subcode => "640",                                                class => "8",           reason => "Systemst&ouml;rung Feuerungsautomat",                                                                                                                                                                                                                                                                                                              checklist => "Interner Fehler im Feuerungsautomat (SAFe)",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1303"} = {errorcode => "F0",                                     subcode => "641",                                                class => "8",           reason => "Systemst&ouml;rung Feuerungsautomat",                                                                                                                                                                                                                                                                                                              checklist => "Interner Fehler im Feuerungsautomat (SAFe)",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1304"} = {errorcode => "F0",                                     subcode => "650",                                                class => "8",           reason => "Kodierstecker defekt",                                                                                                                                                                                                                                                                                                                             checklist => "Brenner-Idendifikations-Modul (BIM) falsch programmiert",                                                                                                                                                                                                                                                                                                                                                                                                                                             correction => "Brenner-Idendifikations-Modul (BIM) austauschen"                                                                                                                                                                                                                                                                                                                                                 };
	$ErrorCodeList{"1305"} = {errorcode => "F0",                                     subcode => "651",                                                class => "8",           reason => "Systemst&ouml;rung Feuerungsautomat und Kodierstecker defekt",                                                                                                                                                                                                                                                                                     checklist => "Feuerungsautomat (SAFe) ist nicht vollst&auml;ndig mit dem Brenner-Idendifikations-Modul (BIM) verbunden&bull;Falsches Brenner-Idendifikations-Modul (BIM) eingebaut&bull;Feuerungsautomat (SAFe) defekt",                                                                                                                                                                                                                                                                                            correction => "Feuerungsautomat (SAFe) korrekt mit dem Brenner-Idendifikations-Modul (BIM) verbinden&bull;Brenner-Idendifikations-Modul (BIM) austauschen&bull;Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                             };
	$ErrorCodeList{"1306"} = {errorcode => "F0",                                     subcode => "652",                                                class => "8",           reason => "Kodierstecker defekt",                                                                                                                                                                                                                                                                                                                             checklist => "Interner Fehler im Brenner-Idendifikations-Modul (BIM)",                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Brenner-Idendifikations-Modul (BIM) austauschen"                                                                                                                                                                                                                                                                                                                                                 };
	$ErrorCodeList{"1307"} = {errorcode => "F0",                                     subcode => "653; 654; 655; 656; 657; 658; 659",                  class => "8",           reason => "Systemst&ouml;rung Feuerungsautomat",                                                                                                                                                                                                                                                                                                              checklist => "Interner Fehler im Feuerungsautomat (SAFe)",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Feuerungsautomat (SAFe) austauschen"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1308"} = {errorcode => "F0",                                     subcode => "660",                                                class => "8",           reason => "Kodierstecker Kommunikation gest&ouml;rt",                                                                                                                                                                                                                                                                                                         checklist => "Kommunikationsst&ouml;rung im BrennerIdendifikations-Modul (BIM)",                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "Brenner-Idendifikations-Modul (BIM) austauschen"                                                                                                                                                                                                                                                                                                                                                 };
	$ErrorCodeList{"1309"} = {errorcode => "F0",                                     subcode => "661",                                                class => "8",           reason => "Systemst&ouml;rung Kodierstecker",                                                                                                                                                                                                                                                                                                                 checklist => "Interner Fehler im Brenner-IdendifikationsModul (BIM)",                                                                                                                                                                                                                                                                                                                                                                                                                                               correction => "Brenner-Idendifikations-Modul (BIM) austauschen"                                                                                                                                                                                                                                                                                                                                                 };
	$ErrorCodeList{"1310"} = {errorcode => "F0",                                     subcode => "662",                                                class => "8",           reason => "Kodierstecker kann nicht gelesen werden",                                                                                                                                                                                                                                                                                                          checklist => "Interner Fehler im Brenner-IdendifikationsModul (BIM)",                                                                                                                                                                                                                                                                                                                                                                                                                                               correction => "Brenner-Idendifikations-Modul (BIM) austauschen"                                                                                                                                                                                                                                                                                                                                                 };
	$ErrorCodeList{"1311"} = {errorcode => "F0",                                     subcode => "690",                                                class => "8",           reason => "Relais im Umschaltmodul schaltet nicht nach Vorgabe",                                                                                                                                                                                                                                                                                              checklist => "Relais im Schaltmodul UX15 defekt",                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Schaltmodul UX15 austauschen"                                                                                                                                                                                                                                                                                                                                                                    };
	$ErrorCodeList{"1312"} = {errorcode => "F0",                                     subcode => "691",                                                class => "8",           reason => "R&uuml;ckmeldung Umschaltmodul, obwohl Relais dort nicht angesteuert",                                                                                                                                                                                                                                                                             checklist => "Interner Fehler im Schaltmodul UX15",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "Schaltmodul UX15 austauschen"                                                                                                                                                                                                                                                                                                                                                                    };
	$ErrorCodeList{"1313"} = {errorcode => "F0",                                     subcode => "692; 693; 694; 695; 696; 697; 698; 699",             class => "8",           reason => "Systemst&ouml;rung",                                                                                                                                                                                                                                                                                                                               checklist => "Interner Fehler im Schaltmodul UX15",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "Schaltmodul UX15 austauschen"                                                                                                                                                                                                                                                                                                                                                                    };
	$ErrorCodeList{"1314"} = {errorcode => "F7",                                     subcode => "52",                                                 class => "",            reason => "Flammensignal erkannt, obwohl keine Gasfreigabe erfolgt ist, W&auml;rmeerzeuger verriegelt",                                                                                                                                                                                                                                                       checklist => "-&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       correction => "Elektroden pr&uuml;fen, ggf. tauschen&bull;Abgassystem pr&uuml;fen, ggf. reinigen oder instandsetzen&bull;Leiterplatte auf Feuchtigkeit pr&uuml;fen, ggf. trocknen"                                                                                                                                                                                                                              };
	$ErrorCodeList{"1315"} = {errorcode => "F7",                                     subcode => "228",                                                class => "8",           reason => "Flammensignal trotz nicht vorhandener Flamme",                                                                                                                                                                                                                                                                                                     checklist => "&Uuml;berpr&uuml;fen, ob eine Brennerflamme vorhanden ist&bull;Ionisationskabel auf Besch&auml;digung &uuml;berpr&uuml;fen&bull;",                                                                                                                                                                                                                                                                                                                                                                    correction => "Gasaramatur austauschen&bull;Ionisationskabel austauschen&bull;Steuerger&auml;t/ Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"1316"} = {errorcode => "F7",                                     subcode => "328",                                                class => "8",           reason => "Netzspannung kurzzeitig unterbrochen",                                                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Elektrische Hausinstallation hinsichtlich Spannungsunterbrechung zum W&auml;rmerezeuger &uuml;berpr&uuml;fen."                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"1317"} = {errorcode => "FA",                                     subcode => "53",                                                 class => "",            reason => "Undichtigkeit an Sicherheitsventilen erkannt, W&auml;rmeerzeuger verriegelt",                                                                                                                                                                                                                                                                      checklist => "-&bull;&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "Gasarmatur pr&uuml;fen, ggf. tauschen&bull;Kondensatsiphon reinigen&bull;Elektroden und Anschlusskabel pr&uuml;fen, ggf. tauschen&bull;Abgassystem pr&uuml;fen, ggf. reinigen oder instandsetzen"                                                                                                                                                                                                };
	$ErrorCodeList{"1318"} = {errorcode => "FA",                                     subcode => "306",                                                class => "8",           reason => "Flammensignal nach Schlie&szlig;en der Brennstoffversorgung",                                                                                                                                                                                                                                                                                      checklist => "&Uuml;berpr&uuml;fen, ob eine Brennerflamme vorhanden ist&bull;Ionisationskabel auf Besch&auml;digung &uuml;berpr&uuml;fen&bull;",                                                                                                                                                                                                                                                                                                                                                                    correction => "Gasaramatur austauschen&bull;Ionisationskabel austauschen&bull;Steuerger&auml;t/ Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"1319"} = {errorcode => "FD",                                     subcode => "54",                                                 class => "",            reason => "Entst&ouml;rtaste gedr&uuml;ckt oder interne St&ouml;rung, W&auml;rmeerzeuger verriegelt",                                                                                                                                                                                                                                                         checklist => "-&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                             correction => "Reset-Taste dr&uuml;cken und halten, bis das Display 00 zeigt&bull;Kabelbaum zum Sicherheitstemperaturbegrenzer und Gasarmatur auf Masseschluss pr&uuml;fen"                                                                                                                                                                                                                                     };
	$ErrorCodeList{"1320"} = {errorcode => "-H",                                     subcode => "200",                                                class => "8",           reason => "W&auml;rmeerzeuger im Heizbetrieb",                                                                                                                                                                                                                                                                                                                checklist => "W&auml;rmeerzeuger ist im Heizbetrieb",                                                                                                                                                                                                                                                                                                                                                                                                                                                               correction => "Betriebsmeldung, keine St&ouml;rung"                                                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1321"} = {errorcode => "-H",                                     subcode => "2501",                                               class => "BC",          reason => "W&auml;rmeanforderung wegen Frostschutz",                                                                                                                                                                                                                                                                                                          checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1322"} = {errorcode => "-H",                                     subcode => "2502",                                               class => "BC",          reason => "W&auml;rmeanforderung wegen Notbetrieb",                                                                                                                                                                                                                                                                                                           checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1323"} = {errorcode => "-H",                                     subcode => "2506",                                               class => "BC",          reason => "W&auml;rmeanforderung wegen Heizbetrieb",                                                                                                                                                                                                                                                                                                          checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1324"} = {errorcode => "-H",                                     subcode => "2509",                                               class => "BC",          reason => "Interner Staus",                                                                                                                                                                                                                                                                                                                                   checklist => "Interner Status",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                     correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1325"} = {errorcode => "-H",                                     subcode => "201",                                                class => "8",           reason => "W&auml;rmeerzeuger im WW-Betrieb",                                                                                                                                                                                                                                                                                                                 checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1326"} = {errorcode => "-H",                                     subcode => "2507",                                               class => "8",           reason => "W&auml;rmeanforderung wegen WW-Betrieb",                                                                                                                                                                                                                                                                                                           checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1327"} = {errorcode => "-H",                                     subcode => "309",                                                class => "8",           reason => "Heizund W&auml;rmebetrieb parallel",                                                                                                                                                                                                                                                                                                               checklist => "Betriebsmeldung, keine St&ouml;rung, Hybrid system l&auml;uft gleichzeitig im Heizund Warmwasserbetrieb (parallel)",                                                                                                                                                                                                                                                                                                                                                                                  correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1328"} = {errorcode => "-H",                                     subcode => "2508",                                               class => "8",           reason => "W&auml;rmeanforderung wegen WW/Heizbetrieb",                                                                                                                                                                                                                                                                                                       checklist => "Betriebsmeldung, keine St&ouml;rung",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1329"} = {errorcode => "HH",                                     subcode => "311",                                                class => "10",          reason => "Keine Kommunikation mit dem Systemregler",                                                                                                                                                                                                                                                                                                         checklist => "Kontakte, Kabel und Verbindungen zwischen Hybridmodul und Systemregler pr&uuml;fen&bull;Systemregler pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                    correction => "Besch&auml;digte Leitungen und Kontakte austauschen&bull;Ggf. ersetzen"                                                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"1330"} = {errorcode => "H01",                                    subcode => "1011",                                               class => "18",          reason => "St&ouml;rung: Abgastemperatur zu hoch; (Sobald die Abgastemperatur einen Sollwert ubersteigt, wird diese Wartungsmeldung erzeugt und kann erst wieder geloscht werden, wenn am RC3x der Befehl; „Wartungsmeldung zurucksetzen“ gegeben wird. Auswirkung: Die Leistung des Kessels; wird begrenzt)",                                            checklist => "Kessel auf Verschmutzung pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Kessel reinigen"                                                                                                                                                                                                                                                                                                                                                                                 };
	$ErrorCodeList{"1331"} = {errorcode => "H01",                                    subcode => "5203",                                               class => "8",           reason => "Alarm Aussentemperaturf&uuml;hler T1 Fehler",                                                                                                                                                                                                                                                                                                      checklist => "Anschluss Au&szlig;entemperaturf&uuml;hler pr&uuml;fen Au&szlig;entemperatuf&uuml;hler liegt au&szlig;erhalb des Messbereichs (Ohm-Werte pr&uuml;fen, s.u. Werte in Installationsanleitung)",                                                                                                                                                                                                                                                                                                         correction => "Au&szlig;entemperaturf&uuml;hler montieren Au&szlig;entemperaturf&uuml;hler tauschen Installationsmodul tauschen"                                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"1332"} = {errorcode => "H01",                                    subcode => "5206",                                               class => "WPL AR",      reason => "Alarm Vorlauff&uuml;hler T0 defekt",                                                                                                                                                                                                                                                                                                               checklist => "Siehe Pr&uuml;fvorgang/Ursachen f&uuml;r Ursachencode 5204, 5205",                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5204, 5205"                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1333"} = {errorcode => "H01",                                    subcode => "5209",                                               class => "8",           reason => "Alarm Temperaturf&uuml;hler TC1 Fehler",                                                                                                                                                                                                                                                                                                           checklist => "&Uuml;berpr&uuml;fen, ob der Anschlussstecker am Temperaturf&uuml;hler TC1 komplett aufgesteckt ist&bull;Ohmschen Widerstand des Temperaturf&uuml;hler TC1 messen&bull;-",                                                                                                                                                                                                                                                                                                                            correction => "Anschlussstecker am Temperaturf&uuml;hler TC1 komplett aufstecken.&bull;Temperaturf&uuml;hler TC 1 austauschen&bull;Anschlusskabel zum Temperatuf&uuml;hler TC1 austauschen"                                                                                                                                                                                                                     };
	$ErrorCodeList{"1334"} = {errorcode => "H01",                                    subcode => "5212",                                               class => "WPL AR",      reason => "Alarm Z2 W&auml;rmetr&auml;ger in F&uuml;hler TC1 defekt",                                                                                                                                                                                                                                                                                         checklist => "Siehe Pr&uuml;fvorgang/Ursachen f&uuml;r Ursachencode 5210, 5211",                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5210, 5211"                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1335"} = {errorcode => "H01",                                    subcode => "5215",                                               class => "WPL AR",      reason => "Alarm Z1 W&auml;rme-; tr&auml;ger in F&uuml;hler TC0 defekt",                                                                                                                                                                                                                                                                                      checklist => "Siehe Pr&uuml;fvorgang/Ursachen f&uuml;r Ursachencode 5213, 5214",                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5213, 5214"                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1336"} = {errorcode => "H01",                                    subcode => "5218",                                               class => "WPL AR",      reason => "Alarm Z2 W&auml;rmetr&auml;ger in F&uuml;hler TC0 defekt",                                                                                                                                                                                                                                                                                         checklist => "Siehe Pr&uuml;fvorgang/Ursachen f&uuml;r Ursachencode 5217, 5216",                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5216, 5217"                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1337"} = {errorcode => "H01",                                    subcode => "5221",                                               class => "4",           reason => "Alarm Z1 Sole Eintrittstemperaturf&uuml;hler TB0 Fehler",                                                                                                                                                                                                                                                                                          checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Temperaturf&uuml;hler TC3 austauschen"                                                                                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"1338"} = {errorcode => "H01",                                    subcode => "5236",                                               class => "WPL AR",      reason => "Alarm Schwimmbadtemperaturf&uuml;hler TP1 defekt",                                                                                                                                                                                                                                                                                                 checklist => "Siehe Pr&uuml;fvorgang/Ursachen f&uuml;r Ursachencode 5234, 5235",                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5234, 5235"                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1339"} = {errorcode => "H01",                                    subcode => "5239",                                               class => "8",           reason => "Alarm Warmwasserf&uuml;hler TW1 Fehler",                                                                                                                                                                                                                                                                                                           checklist => "&Uuml;berpr&uuml;fen, ob das Anschlusskabel vom Warmwassertemperaturf&uuml;hler TW1 im Steuerger&auml;t korrekt angeschlossen ist&bull;Ohmschen Widerstand des Warmwassertemperaturf&uuml;hlers TW1 messen&bull;-",                                                                                                                                                                                                                                                                                   correction => "Anschlusskabel vom Warmwassertemperaturf&uuml;hler TW1 im Steuerger&auml;t korrekt anschlie&szlig;en&bull;Warmwassertemperaturf&uuml;hler TW1 austauschen&bull;Anschlusskabel zum Warmwassertemperaturf&uuml;hler TW1 austauschen"                                                                                                                                                               };
	$ErrorCodeList{"1340"} = {errorcode => "H01",                                    subcode => "5249",                                               class => "WPL AR",      reason => "Alarm Z1 W&auml;rmepumpe gestoppt wegen mehrfach zu hoher Vorlauftemperatur",                                                                                                                                                                                                                                                                      checklist => "Heizkurve/Warmwassertemperatur zu hoch eingestellt&bull;Schmutz im Systemfilter/Filterball SC1",                                                                                                                                                                                                                                                                                                                                                                                                      correction => "Temperaturen &uuml;berpr&uuml;fen/einstellen&bull;Systemfilter/Filterball SC1 reinigen"                                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"1341"} = {errorcode => "H01",                                    subcode => "5251",                                               class => "WPL AR",      reason => "Alarm Z2 W&auml;rmepumpe gestoppt wegen mehrfach zu hoher Vorlauftemperatur",                                                                                                                                                                                                                                                                      checklist => "Heizkurve/Warmwassertemperatur zu hoch eingestellt&bull;Schmutz im Systemfilter/Filterball SC1",                                                                                                                                                                                                                                                                                                                                                                                                      correction => "Temperaturen &uuml;berpr&uuml;fen/einstellen&bull;Systemfilter/Filterball SC1 reinigen"                                                                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"1342"} = {errorcode => "H01",                                    subcode => "5252",                                               class => "WPL AR",      reason => "Warnung Z1 Volumenstrom zw. Au&szlig;enund Inneneinheit eingeschr&auml;nkt (Filter pr&uuml;fen)",                                                                                                                                                                                                                                                  checklist => "Schlechte Zirkulation im W&auml;rmetr&auml;gersystem/Heizsystem",                                                                                                                                                                                                                                                                                                                                                                                                                                     correction => "Einstellventile/Heizk&ouml;rperthermostate &uuml;berpr&uuml;fen"                                                                                                                                                                                                                                                                                                                                 };
	$ErrorCodeList{"1343"} = {errorcode => "H01",                                    subcode => "5266",                                               class => "WPL AR",      reason => "Alarm Heizpumpenkarte Z1 getrennt",                                                                                                                                                                                                                                                                                                                checklist => "Siehe Ursachen f&uuml;r Warnung 5265 oben",                                                                                                                                                                                                                                                                                                                                                                                                                                                           correction => "Siehe Ma&szlig;nahmen f&uuml;r Warnung 5265 oben"                                                                                                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"1344"} = {errorcode => "H01",                                    subcode => "5268",                                               class => "WPL AR",      reason => "Alarm Heizpumpenkarte Z2 getrennt",                                                                                                                                                                                                                                                                                                                checklist => "Siehe Ursachen f&uuml;r Warnung 5267 oben",                                                                                                                                                                                                                                                                                                                                                                                                                                                           correction => "Siehe Ma&szlig;nahmen f&uuml;r Warnung 5267 oben"                                                                                                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"1345"} = {errorcode => "H01",                                    subcode => "5275",                                               class => "WPL AR",      reason => "Alarm Fremdstromanonde funktioniert nicht",                                                                                                                                                                                                                                                                                                        checklist => "Diode an Fremdstromanodenkarte leuchtet rot&bull;Betr&auml;gt die Spannung an Klemme 45, 46 am Installer-Board > 1 V DC, ist das Installer-Board au&szlig;er Funktion&bull;Gr&uuml;ne Leuchtdiode an Fremdstromanodenkarte leuchtet&bull;F&uuml;hrt Klemme X1 auf der Fremdstromanodenkarte 230 V und leuchtet die Leuchtdiode nicht gr&uuml;n, ist die Fremdstromanodenkarte au&szlig;er Funktion&bull;Luftw&auml;rmetauscher an W&auml;rmepumpe verschmutzt/verstopft",                             correction => "Anschluss/Kabelbaum an Klemme X2 und Fremdstromanodenklemme im Tank &uuml;berpr&uuml;fen&bull;Installer-Board wechseln&bull;&Uuml;berpr&uuml;fen, ob Klemme X1 an Fremdstromkarte 230 V f&uuml;hrt&bull;Fremdstromanodenkarte wechseln&bull;Luftw&auml;rmetauscher an W&auml;rmepumpe reinigen"                                                                                                  };
	$ErrorCodeList{"1346"} = {errorcode => "H01",                                    subcode => "5283",                                               class => "WPL AR",      reason => "JR1 Alarm W&auml;rmepumpe Z2 reinigen",                                                                                                                                                                                                                                                                                                            checklist => "Blockierter/niedriger Luftstrom zu Luftw&auml;rmetauscher an W&auml;rmepumpe&bull;Gebl&auml;se au&szlig;er Funktion&bull;Gebl&auml;se hat nicht 230 V Spannung&bull;Gebl&auml;se ohne 0: 10 V-Signal&bull;Gebl&auml;se defekt&bull;I/O-Board au&szlig;er Funktion&bull;Schlechte/ausgebliebene Zirkulation im W&auml;rmetr&auml;gersystem/Heizsystem",                                                                                                                                                correction => "Ausreichenden Luftstrom &uuml;ber Luftw&auml;rmetauscher sicherstellen&bull;Test Au&szlig;eneinheit aktivieren&bull;&Uuml;berpr&uuml;fen, dass 230V Spannung an Ausgang PL3 Klemme 32 (78, N)&bull;0: 10 V Spannung an Ausgang PL3 PWM Klemme 20 (20, 26) laut Test Au&szlig;eneinheit &uuml;berpr&uuml;fen&bull;Ist Spannung wie oben vorhanden, Gebl&auml;se wechseln&bull;Fehlt Spannung wie oben, I/O-Board wechseln&bull;Einstellventile/Heizk&ouml;rperthermostate &uuml;berpr&uuml;fen"};
	$ErrorCodeList{"1347"} = {errorcode => "H01",                                    subcode => "5285",                                               class => "WPL AR",      reason => "Warnung Frostschutz ist aktiv",                                                                                                                                                                                                                                                                                                                    checklist => "F&uuml;hler TC3 defekt&bull;I/O-Board defekt, kein PWM-Signal&bull;I/O-Board defekt, kein PWM-Signal&bull;I/O-Board defekt, keine 230 V an Klemme 01 (51, N)&bull;I/O-Board defekt&bull;Umw&auml;lzpumpe defekt&bull;Luftw&auml;rmetauscher an W&auml;rmepumpe verschmutzt/verstopft",                                                                                                                                                                                                                correction => "F&uuml;hler TC3 im Vergleich zur aktuellen F&uuml;hlertabelle &uuml;berpr&uuml;fen/F&uuml;hler TC3 wechseln&bull;&Uuml;berpr&uuml;fen, ob 230 V an Klemme 01 (51, N) am I/O-Board&bull;PWM-Signal an Klemme 36,37 am I/O-Board abschalten, Umw&auml;lzpumpe auf 100 % erh&ouml;hen&bull;&Uuml;berpr&uuml;fen, ob 230 V an Klemme 01 (51, N) am I/O-Board&bull;I/O-Board wechseln&bull;Umw&auml;lzpumpe wechseln&bull;Luftw&auml;rmetauscher an W&auml;rmepumpe reinigen"};
	$ErrorCodeList{"1348"} = {errorcode => "H01",                                    subcode => "5293",                                               class => "WPL AR",      reason => "JR0 Alarm W&auml;rmepumpe Z1 reinigen",                                                                                                                                                                                                                                                                                                            checklist => "Blockierter/niedriger Luftstrom zu Luftw&auml;rmetauscher an W&auml;rmepumpe&bull;Gebl&auml;se au&szlig;er Funktion&bull;Gebl&auml;se hat nicht 230 V Spannung&bull;Gebl&auml;se ohne 0: 10 V-Signal&bull;Gebl&auml;se defekt&bull;I/O-Board au&szlig;er Funktion&bull;Raumtemperatur zu niedrig eingestellt, Vorlauftemperatur T0 < Taupunkt",                                                                                                                                                       correction => "Ausreichenden Luftstrom &uuml;ber Luftw&auml;rmetauscher sicherstellen&bull;Test Au&szlig;eneinheit aktivieren&bull;&Uuml;berpr&uuml;fen, dass 230 V Spannung an Ausgang PL3 Klemme 32 (78, N)&bull;0: 10 V Spannung an Ausgang PL3 PWM Klemme 20 (20, 26) laut Test Au&szlig;eneinheit &uuml;berpr&uuml;fen&bull;Ist Spannung wie oben vorhanden, Gebl&auml;se wechseln&bull;Fehlt Spannung wie oben, I/O-Board wechseln&bull;Raumtemperatur &uuml;ber dem Taupunkt einstellen"};
	$ErrorCodeList{"1349"} = {errorcode => "H01",                                    subcode => "5294",                                               class => "WPL AR",      reason => "Warnung Kondensationsw&auml;chter wurde ausgel&ouml;st",                                                                                                                                                                                                                                                                                           checklist => "Kurzschluss im Feuchtigkeitsf&uuml;hler&bull;Installer-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Ohmwert&bull;Spannung an Klemme MK2 (34, 35) am InstallerBoard mit herausgezogenem Kabel zum Feuchtigkeitsf&uuml;hler messen, bei Spannung unter 2,5 V DC, Installer-Board wechseln"                                                                                                                                                                                                             };
	$ErrorCodeList{"1350"} = {errorcode => "H01",                                    subcode => "5295",                                               class => "WPL AR",      reason => "Alarm Kondensations-w&auml;chter wurde ausgel&ouml;st",                                                                                                                                                                                                                                                                                            checklist => "Siehe Ursachen f&uuml;r Warnung 5294",                                                                                                                                                                                                                                                                                                                                                                                                                                                                correction => "Siehe Ma&szlig;nahmen f&uuml;r Warnung 5294"                                                                                                                                                                                                                                                                                                                                                     };
	$ErrorCodeList{"1351"} = {errorcode => "H01",                                    subcode => "5296",                                               class => "WPL AR",      reason => "K&uuml;hlbetrieb Alarm W&auml;rmepumpe Z1; Fehler an 4-Wege-Ventil kann nicht zu K&uuml;hlung wechseln",                                                                                                                                                                                                                                           checklist => "Spule zum 4-Wege-Ventil au&szlig;er Funktion&bull;Funktion des 4-Wege-Ventils &uuml;berpr&uuml;fen&bull;Keine 230 V auf Klemme VR4 (81, N)&bull;Fehler an F&uuml;hler (TR3, TR4, TC3, TC0)&bull;4-Wege-Ventil st&ouml;rt/au&szlig;er Funktion",                                                                                                                                                                                                                                                       correction => "Widerstand in Magnetspule an 4-Wege-Ventil &uuml;berpr&uuml;fen (1,3-1,5 kOhm), wenn nicht Spule wechseln&bull;Aktivieren Test-Au&szlig;eneinheit, siehe separates Dokument&bull;I/O-Board wechseln&bull;&Uuml;berpr&uuml;fen, ob der F&uuml;hler sich in der richtigen Position befindet, F&uuml;hlerwert im Vergleich zur aktuellen F&uuml;hlertabelle und Temperatur &uuml;berpr&uuml;fen&bull;4-Wege-Ventil wechseln"};
	$ErrorCodeList{"1352"} = {errorcode => "H01",                                    subcode => "5297",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z1 Fehler an 4-Wege-Ventil, kann nicht zu Heizen wechseln",                                                                                                                                                                                                                                                                  checklist => "Funktion des 4-Wege-Ventils &uuml;berpr&uuml;fen&bull;Fehler an F&uuml;hler (TR4, TR3, TC0, TC3)&bull;230 V an Klemme VR4 im K&uuml;hlbetrieb/ Abtauungsbetrieb&bull;4-Wege-Ventil st&ouml;rt/au&szlig;er Funktion",                                                                                                                                                                                                                                                                                  correction => "Aktivieren Test-Au&szlig;eneinheit, siehe separates Dokument&bull;&Uuml;berpr&uuml;fen, ob der F&uuml;hler sich in der richtigen Position befindet, F&uuml;hlerwert im Vergleich zur aktuellen F&uuml;hlertabelle und Temperatur &uuml;berpr&uuml;fen&bull;I/O-Board wechseln&bull;4-Wege-Ventil wechseln"                                                                                       };
	$ErrorCodeList{"1353"} = {errorcode => "H01",                                    subcode => "5299",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z1 zu hoher Druck an JR1",                                                                                                                                                                                                                                                                                                   checklist => "Siehe Warnung 5298 oben",                                                                                                                                                                                                                                                                                                                                                                                                                                                                             correction => "Siehe Ma&szlig;nahmen f&uuml;r Warnung 5298 oben"                                                                                                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"1354"} = {errorcode => "H01",                                    subcode => "5301",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z2 ausgel&ouml;ster Hochdruckpressostat",                                                                                                                                                                                                                                                                                    checklist => "Siehe Warnung 5298, Alarm 5299 oben",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "Siehe Ma&szlig;nahmen f&uuml;r Warnung 5265, Alarm 5299 oben"                                                                                                                                                                                                                                                                                                                                    };
	$ErrorCodeList{"1355"} = {errorcode => "H01",                                    subcode => "5303",                                               class => "WPL AR",      reason => "Alarm Z1 hohe Temperatur am Kompressortreiber",                                                                                                                                                                                                                                                                                                    checklist => "Siehe Warnung 5302 oben&bull;Inverter defekt",                                                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Siehe Ma&szlig;nahmen f&uuml;r Warnung 5302 oben&bull;Inverter wechseln"                                                                                                                                                                                                                                                                                                                         };
	$ErrorCodeList{"1356"} = {errorcode => "H01",                                    subcode => "5305",                                               class => "WPL AR",      reason => "Alarm Z2 hohe Temperatur am Kompressortreiber W&auml;rmepumpe 2",                                                                                                                                                                                                                                                                                  checklist => "Siehe Alarm 5303 oben",                                                                                                                                                                                                                                                                                                                                                                                                                                                                               correction => "Siehe Ma&szlig;nahmen f&uuml;r Alarm 5303 oben"                                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1357"} = {errorcode => "H01",                                    subcode => "5311",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z1 Hei&szlig;gastemperatur zu hoch",                                                                                                                                                                                                                                                                                         checklist => "Siehe Warnung 5310 oben",                                                                                                                                                                                                                                                                                                                                                                                                                                                                             correction => "Siehe Ma&szlig;nahmen f&uuml;r Warnung 5310 oben"                                                                                                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"1358"} = {errorcode => "H01",                                    subcode => "5313",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z2 Hei&szlig;gastemperatur zu hoch",                                                                                                                                                                                                                                                                                         checklist => "Sauggas-&Uuml;berhitzung zu hoch",                                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "&Uuml;berpr&uuml;fen, dass Sauggas-&Uuml;berhitzung 10C nicht kontinuierlich &uuml;bersteigt"                                                                                                                                                                                                                                                                                                    };
	$ErrorCodeList{"1359"} = {errorcode => "H01",                                    subcode => "5316",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z1 Heizgasf&uuml;hler TR6 defekt",                                                                                                                                                                                                                                                                                           checklist => "Siehe Pr&uuml;fvorgang/Ursachen f&uuml;r Ursachencode 5314, 5315",                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5314, 5315"                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1360"} = {errorcode => "H01",                                    subcode => "5319",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z2 Heizgasf&uuml;hler TR6 W&auml;rmepumpe 2 defekt",                                                                                                                                                                                                                                                                         checklist => "Siehe Pr&uuml;fvorgang/Ursachen f&uuml;r Ursachencode 5317, 5318",                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5317, 5318"                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1361"} = {errorcode => "H01",                                    subcode => "5322",                                               class => "8",           reason => "Alarm Temperaturf&uuml;hler TC3 Fehler",                                                                                                                                                                                                                                                                                                           checklist => "&Uuml;berpr&uuml;fen, ob das Anschlusskabel vom Temperaturf&uuml;hler TC3 im Steuerger&auml;t korrekt angeschlossen ist&bull;-",                                                                                                                                                                                                                                                                                                                                                                      correction => "Anschlusskabel vom Temperaturf&uuml;hler TC3 im Steuerger&auml;t korrekt anschlie&szlig;en&bull;Temperaturf&uuml;hler TC3 austauschen"                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"1362"} = {errorcode => "H01",                                    subcode => "5325",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z2 Kondensatorf&uuml;hler TC3 defekt",                                                                                                                                                                                                                                                                                       checklist => "Siehe Pr&uuml;fvorgang/Ursachen f&uuml;r Ursachencode 5323, 5324",                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5323, 5324"                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1363"} = {errorcode => "H01",                                    subcode => "5331",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z1 Kommunikationsfehler zum Kompressortreiber",                                                                                                                                                                                                                                                                              checklist => "Keine Spannung zum Au&szlig;enteil&bull;St&ouml;rung am MODbus&bull;CANbus-Kabel zwischen IDU, ODU fehlerhaft verlegt&bull;I/O-Board au&szlig;er Funktion&bull;Inverter au&szlig;er Funktion",                                                                                                                                                                                                                                                                                                        correction => "230/400 V-Anschl&uuml;sse in Tower und ODU &uuml;berpr&uuml;fen&bull;MODbus-Kabel/Anschlussklemmen zwischen I/O-Board und Inverter &uuml;berpr&uuml;fen&bull;Kabelbaum und Verlegung (mindestens 100 mm zwischen Stromkabel und Canbus-Kabel) &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob 12 V DC an MODbus-Klemme (31, 34) am I/O-Board vorhanden sind; sind keine 12 V DC vorhanden, I/O-Board wechseln&bull;Inverter wechseln"};
	$ErrorCodeList{"1364"} = {errorcode => "H01",                                    subcode => "5333",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z2 Kommunikationsfehler zum Kompressortreiber",                                                                                                                                                                                                                                                                              checklist => "Keine Spannung zum Au&szlig;enteil&bull;St&ouml;rung am MODbus&bull;CANbus-Kabel zwischen IDU, ODU fehlerhaft verlegt&bull;I/O-Board au&szlig;er Funktion&bull;Inverter au&szlig;er Funktion",                                                                                                                                                                                                                                                                                                        correction => "230/400 V-Anschl&uuml;sse in Tower und ODU &uuml;berpr&uuml;fen&bull;MODbus-Kabel/Anschlussklemmen zwischen I/O-Board und Inverter &uuml;berpr&uuml;fen&bull;Kabelbaum und Verlegung (mindestens 100 mm zwischen Stromkabel und Canbus-Kabel) &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob 12 V DC an MODbus-; Klemme (31, 34) am I/O-Board vorhanden sind; sind keine 12 V DC vorhanden, I/O-Board wechseln&bull;Inverter wechseln"};
	$ErrorCodeList{"1365"} = {errorcode => "H01",                                    subcode => "5347",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z1 Eingangsspannung; zu niedrig",                                                                                                                                                                                                                                                                                          checklist => "Kontakt in Spannungsversorgung zu IDU/ODU schlecht&bull;Niedrige eingehende Netzspannung",                                                                                                                                                                                                                                                                                                                                                                                                            correction => "Spannungsversorgung &uuml;berpr&uuml;fen&bull;Bei wiederholten Warnungen Kontakt zum Stromanbieter aufnehmen"                                                                                                                                                                                                                                                                                    };
	$ErrorCodeList{"1366"} = {errorcode => "H01",                                    subcode => "5352",                                               class => "WPL AR",      reason => "Warnung; W&auml;rmepumpe Z2 asynchroner Betrieb am Kompressor W&auml;rmepumpe 2",                                                                                                                                                                                                                                                                  checklist => "Spiel im Kabelbaum zwischen Kompressor&bull;Spiel im Kabelbaum zwischen Kompressor und Inverter&bull;&Ouml;l/Fl&uuml;ssigkeit im Kompressor bei Startversuch&bull;Kompressorf&uuml;hler TR1 zeigt falschen Wert&bull;Wenn I/O-Board ausgetauscht wird, fehlerhafte Einstellung des Drehgebers am I/O-Board",                                                                                                                                                                                          correction => "Kabelbaum/Anschl&uuml;sse zwischen&bull;Kabelbaum/Anschl&uuml;sse zwischen Kompressor und Inverter &uuml;berpr&uuml;fen&bull;Kompressorf&uuml;hler TR1 zeigt falschen Wert&bull;Kompressorf&uuml;hler TR1 im Vergleich zur aktuellen F&uuml;hlertabelle und tats&auml;chlichen Temperatur &uuml;berpr&uuml;fen&bull;Einstellung im Vergleich zum aktuellen elektrischen Schaltplan &uuml;berpr&uuml;fen"};
	$ErrorCodeList{"1367"} = {errorcode => "H01",                                    subcode => "5355",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z1 &Uuml;berstrom Kompressor",                                                                                                                                                                                                                                                                                               checklist => "Siehe Pr&uuml;fvorgang/; Ursachen f&uuml;r Ursachencode 5354",                                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5354"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1368"} = {errorcode => "H01",                                    subcode => "5357",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z2 &Uuml;berstrom am Kompressor",                                                                                                                                                                                                                                                                                            checklist => "Siehe Pr&uuml;fvorgang/; Ursachen f&uuml;r Ursachencode 5356",                                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5356"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1369"} = {errorcode => "H01",                                    subcode => "5361",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z2 PFC &Uuml;berstrom des Kompressortreibers",                                                                                                                                                                                                                                                                               checklist => "Siehe Pr&uuml;fvorgang/; Ursachen f&uuml;r Ursachencode 5360",                                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5360"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1370"} = {errorcode => "H01",                                    subcode => "5367",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z1 niedrige &Uuml;berhitzung",                                                                                                                                                                                                                                                                                               checklist => "Siehe Pr&uuml;fvorgang/; Ursachen f&uuml;r Ursachencode 5366",                                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5366"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1371"} = {errorcode => "H01",                                    subcode => "5369",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z2 niedrige &Uuml;berhitzung",                                                                                                                                                                                                                                                                                               checklist => "Siehe Pr&uuml;fvorgang/; Ursachen f&uuml;r Ursachencode 5368",                                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5368"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1372"} = {errorcode => "H01",                                    subcode => "5375",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z1 Frostschutz des Kondensators",                                                                                                                                                                                                                                                                                            checklist => "Siehe Pr&uuml;fvorgang/; Ursachen f&uuml;r Ursachencode 5374",                                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5374"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1373"} = {errorcode => "H01",                                    subcode => "5377",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z2 Frostschutz des Kondensators",                                                                                                                                                                                                                                                                                            checklist => "Siehe Pr&uuml;fvorgang/; Ursachen f&uuml;r Ursachencode 5376",                                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5376"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1374"} = {errorcode => "H01",                                    subcode => "5387",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z1 des Kompressortreibers ist zu warm",                                                                                                                                                                                                                                                                                      checklist => "Schlechte W&auml;rme&uuml;bertragung zur K&uuml;hlschlange&bull;Bei wiederholtem Alarm, Inverter defekt",                                                                                                                                                                                                                                                                                                                                                                                             correction => "Verschraubung zur K&uuml;hlschlange &uuml;berpr&uuml;fen&bull;Inverter wechseln"                                                                                                                                                                                                                                                                                                                 };
	$ErrorCodeList{"1375"} = {errorcode => "H01",                                    subcode => "5389",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z2 des Kompressortreibers ist zu warm",                                                                                                                                                                                                                                                                                      checklist => "Schlechte W&auml;rme&uuml;bertragung zur K&uuml;hlschlange&bull;Bei wiederholtem Alarm, Inverter defekt",                                                                                                                                                                                                                                                                                                                                                                                             correction => "Verschraubung zur K&uuml;hlschlange &uuml;berpr&uuml;fen&bull;Inverter wechseln"                                                                                                                                                                                                                                                                                                                 };
	$ErrorCodeList{"1376"} = {errorcode => "H01",                                    subcode => "5395",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z1 Interner Kompressortreiberfehler 1",                                                                                                                                                                                                                                                                                      checklist => "Siehe Pr&uuml;fvorgang/; Ursachen f&uuml;r Ursachencode 5394",                                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5394"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1377"} = {errorcode => "H01",                                    subcode => "5397",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z2 Interner Kompressortreiberfehler 1",                                                                                                                                                                                                                                                                                      checklist => "Siehe Pr&uuml;fvorgang/; Ursachen f&uuml;r Ursachencode 5396",                                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5396"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1378"} = {errorcode => "H01",                                    subcode => "5410",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z1 Lufteintrittstemperatur-f&uuml;hler TL2 defekt",                                                                                                                                                                                                                                                                          checklist => "Siehe Pr&uuml;fvorgang/; Ursachen f&uuml;r Ursachencode 5408, 5409",                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5408, 5409"                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1379"} = {errorcode => "H01",                                    subcode => "5413",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z1 Lufteintrittstemperatur-f&uuml;hler TL2 defekt",                                                                                                                                                                                                                                                                          checklist => "Siehe Pr&uuml;fvorgang/; Ursachen f&uuml;r Ursachencode 5411, 5412",                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5411, 5412"                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1380"} = {errorcode => "H01",                                    subcode => "5416",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z1 Fl&uuml;ssigkeitsleitung in Heizbetrieb F&uuml;hler TR3 defekt",                                                                                                                                                                                                                                                          checklist => "Siehe Pr&uuml;fvorgang/; Ursachen f&uuml;r Ursachencode 5414, 5415",                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5414, 5415"                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1381"} = {errorcode => "H01",                                    subcode => "5419",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z2 Fl&uuml;ssigkeitsleitung in Heizbetrieb F&uuml;hler TR3 defekt",                                                                                                                                                                                                                                                          checklist => "Siehe Pr&uuml;fvorgang/; Ursachen f&uuml;r Ursachencode 5417, 5418",                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5417, 5418"                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1382"} = {errorcode => "H01",                                    subcode => "5422",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z1 Fl&uuml;ssigkeitsleitung im K&uuml;hlbetrieb F&uuml;hler TR4 defekt",                                                                                                                                                                                                                                                     checklist => "Siehe Pr&uuml;fvorgang/; Ursachen f&uuml;r Ursachencode 5421, 5421",                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5420, 5421"                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1383"} = {errorcode => "H01",                                    subcode => "5425",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z2 Fl&uuml;ssigkeitsleitung im K&uuml;hlbetrieb F&uuml;hler TR4 defekt",                                                                                                                                                                                                                                                     checklist => "Siehe Pr&uuml;fvorgang/; Ursachen f&uuml;r Ursachencode 5421, 5421",                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5420, 5421"                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1384"} = {errorcode => "H01",                                    subcode => "5428",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z1 Sauggasf&uuml;hler TR5 defekt",                                                                                                                                                                                                                                                                                           checklist => "Siehe Pr&uuml;fvorgang/; Ursachen f&uuml;r Ursachencode 5426, 5427",                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5426, 5427"                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1385"} = {errorcode => "H01",                                    subcode => "5431",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z2 Sauggasf&uuml;hler TR5 defekt",                                                                                                                                                                                                                                                                                           checklist => "Siehe Pr&uuml;fvorgang/; Ursachen f&uuml;r Ursachencode 5429, 5430",                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5429, 5430"                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1386"} = {errorcode => "H01",                                    subcode => "5434",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z1 Druckgeber; niedriger Druck JR0 defekt",                                                                                                                                                                                                                                                                                  checklist => "Siehe Pr&uuml;fvorgang/; Ursachen f&uuml;r Ursachencode 5432",                                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5432"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1387"} = {errorcode => "H01",                                    subcode => "5437",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z2 Druckgeber niedriger Druck JR0 defekt",                                                                                                                                                                                                                                                                                   checklist => "Siehe Pr&uuml;fvorgang/; Ursachen f&uuml;r Ursachencode 5435",                                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5435"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1388"} = {errorcode => "H01",                                    subcode => "5440",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z1 Druckgeber hoher Druck JR1 defekt",                                                                                                                                                                                                                                                                                       checklist => "Siehe Pr&uuml;fvorgang/; Ursachen f&uuml;r Ursachencode 5438",                                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5438"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1389"} = {errorcode => "H01",                                    subcode => "5443",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z2 Druckgeber hoher Druck JR1 defekt",                                                                                                                                                                                                                                                                                       checklist => "Siehe Pr&uuml;fvorgang/; Ursachen f&uuml;r Ursachencode 5441",                                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5441"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1390"} = {errorcode => "H01",                                    subcode => "5446",                                               class => "WPL AR",      reason => "Schl&auml;uche zwischen Innenteil und W&auml;rmepumpe Z1 wurden gewechselt",                                                                                                                                                                                                                                                                       checklist => "Schl&auml;uche (Steig/R&uuml;cklauf) zwischen W&auml;rmepumpe und Innenteil verwechselt&bull;Defekt F&uuml;hler TC3, TC0",                                                                                                                                                                                                                                                                                                                                                                            correction => "Schl&auml;uche an den richtigen Anschluss anschlie&szlig;en&bull;Platzierung des F&uuml;hlers &uuml;berpr&uuml;fen, F&uuml;hler; im Vergleich zur aktuellen F&uuml;hlertabelle und tats&auml;chlichen Temperatur &uuml;berpr&uuml;fen/bei Abweichung F&uuml;hler wechseln"                                                                                                                       };
	$ErrorCodeList{"1391"} = {errorcode => "H01",                                    subcode => "5448",                                               class => "WPL AR",      reason => "Zu wenig K&uuml;hlmittel W&auml;rmepumpe Z1",                                                                                                                                                                                                                                                                                                      checklist => "Zu wenig K&uuml;hlmittel in W&auml;rmepumpe&bull;Undichtigkeit im K&uuml;hlkreis",                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "K&uuml;hlmittelmenge &uuml;berpr&uuml;fen. Daran denken, beim Entleeren/Auff&uuml;llen des K&uuml;hlmittels die; „Entleeren/Auff&uuml;llen“-Funktion zu&bull;Undichtigkeit &uuml;berpr&uuml;fen/beheben"                                                                                                                                                                                     };
	$ErrorCodeList{"1392"} = {errorcode => "H01",                                    subcode => "5451",                                               class => "WPL AR",      reason => "Zu niedrige Stromrate durch W&auml;rmepumpe Z1 bei Abtauung",                                                                                                                                                                                                                                                                                      checklist => "Temperatur des Heizsystems zu niedrig&bull;Schmutz im Systemfilter/Filterball SC1&bull;Luft im Heizsystem&bull;kein Durchfluss durch Heizk&ouml;rper&bull;I/O-Board defekt, kein PWM-Signal zur Umw&auml;lzpumpe&bull;I/O-Board defekt, keine 230 V an Klemme 01 (51, N)&bull;I/O-Board defekt&bull;Umw&auml;lzpumpe defekt",                                                                                                                                                                         correction => "Mehrere Thermostate des Heizzystems &ouml;ffnen&bull;Systemfilter/Filterball SC1 reinigen&bull;Das Heizsystem laut Anweisungen in der Gebrauchsanweisung entl&uuml;ften, Wasser im Heizzystem&bull;Einstellventile/Heizk&ouml;rperthermostate &uuml;berpr&uuml;fen&bull;PWM-Signal an Klemme 36,37 am; I/O-Board abschalten, Umw&auml;lzpumpe auf 100 % erh&ouml;hen&bull;&Uuml;berpr&uuml;fen, ob 230 V an Klemme 01 (51, N) am I/O-Board&bull;I/O-Board wechseln&bull;Umw&auml;lzpumpe wechseln"};
	$ErrorCodeList{"1393"} = {errorcode => "H01",                                    subcode => "5460",                                               class => "WPL AR",      reason => "Alarm Z2 Warmwasserf&uuml;hler TW1 defekt",                                                                                                                                                                                                                                                                                                        checklist => "Siehe Pr&uuml;fvorgang/; Ursachen f&uuml;r Ursachencode 5458, 5459",                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5458, 5459"                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1394"} = {errorcode => "H01",                                    subcode => "5461",                                               class => "WPL AR",      reason => "Alarm; W&auml;rmepumpe Z2 Fehler an 4-Wege-Ventil, kann nicht zu K&uuml;hlung wechseln",                                                                                                                                                                                                                                                           checklist => "Spule zum 4-Wege-Ventil au&szlig;er Funktion&bull;Funktion des 4-Wege-Ventils &uuml;berpr&uuml;fen&bull;Keine 230 V auf Klemme VR4 (81, N)&bull;Fehler an F&uuml;hler (TR3, TR4, TC3, TC0)&bull;4-Wege-Ventil st&ouml;rt/au&szlig;er Funktion",                                                                                                                                                                                                                                                       correction => "Widerstand in Magnetspule an 4-Wege-Ventil &uuml;berpr&uuml;fen (1,3-1,5 kOhm), wenn nicht Spule wechseln&bull;Aktivieren Test-Au&szlig;eneinheit, siehe separates Dokument&bull;I/O-Board wechseln&bull;&Uuml;berpr&uuml;fen, ob der F&uuml;hler sich in der richtigen Position befindet, F&uuml;hlerwert im Vergleich zur aktuellen F&uuml;hlertabelle und Temperatur &uuml;berpr&uuml;fen&bull;4-Wege-Ventil wechseln"};
	$ErrorCodeList{"1395"} = {errorcode => "H01",                                    subcode => "5462",                                               class => "WPL AR",      reason => "Alarm; W&auml;rmepumpe Z2 Fehler an 4-Wege-Ventil, kann nicht zu Heizen wechseln",                                                                                                                                                                                                                                                                 checklist => "Funktion des 4-Wege-Ventils &uuml;berpr&uuml;fen&bull;Fehler an F&uuml;hler (TR4, TR3, TC0, TC3)&bull;230 V an Klemme VR4 im K&uuml;hlbetrieb/ Abtauungsbetrieb&bull;4-Wege-Ventil st&ouml;rt/au&szlig;er Funktion",                                                                                                                                                                                                                                                                                  correction => "Aktivieren Test-Au&szlig;eneinheit, siehe separates Dokument&bull;&Uuml;berpr&uuml;fen, ob der F&uuml;hler sich in der richtigen Position befindet, F&uuml;hlerwert im Vergleich zur aktuellen F&uuml;hlertabelle und Temperatur &uuml;berpr&uuml;fen&bull;I/O-Board wechseln&bull;4-Wege-Ventil wechseln"                                                                                       };
	$ErrorCodeList{"1396"} = {errorcode => "H01",                                    subcode => "5463",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z1 defr Au&szlig;enteil der W&auml;rmepumpe reinigen",                                                                                                                                                                                                                                                                       checklist => "Verdampfer vereist&bull;Temperatur des Heizsystems zu niedrig&bull;Schmutz im Systemfilter/Filterball SC1",                                                                                                                                                                                                                                                                                                                                                                                           correction => "Eis vorsichtig mit Warmwasser schmelzen&bull;Mehrere Thermostate des Heizzystems &ouml;ffnen&bull;Systemfilter/Filterball SC1 reinigen"                                                                                                                                                                                                                                                          };
	$ErrorCodeList{"1397"} = {errorcode => "H01",                                    subcode => "5501",                                               class => "WPL AR",      reason => "JR0 Alarm W&auml;rmepumpe Z2 reinigen",                                                                                                                                                                                                                                                                                                            checklist => "Blockierter/niedriger Luftstrom zu Luftw&auml;rmetauscher an W&auml;rmepumpe&bull;Gebl&auml;se gem&auml;&szlig; Test Au&szlig;eneinheit &uuml;berpr&uuml;fen&bull;Gebl&auml;se hat nicht 230 V Spannung&bull;Gebl&auml;se ohne 0: 10 V-Signal&bull;Gebl&auml;se defekt&bull;I/O-Board au&szlig;er Funktion&bull;Elektroheizer in Tropfschale au&szlig;er Funktion",                                                                                                                                   correction => "Ausreichenden Luftstrom &uuml;ber Luftw&auml;rmetauscher sicherstellen&bull;Test Au&szlig;eneinheit aktivieren&bull;&Uuml;berpr&uuml;fen, dass 230 V Spannung an Ausgang PL3 Klemme 32 (78, N)&bull;0: 10 V Spannung an Ausgang PL3 PWM Klemme 20 (20, 26) laut Test Au&szlig;eneinheit &uuml;berpr&uuml;fen&bull;Ist Spannung wie oben vorhanden, Gebl&auml;se wechseln&bull;Fehlt Spannung wie oben, I/O-Board wechseln&bull;Widerstand des Elektroheizers messen,; 75 W ~720 Ohm, 50 W ~1070 Ohm, bei Unterbrechung Elektroheizer wechseln"};
	$ErrorCodeList{"1398"} = {errorcode => "H01",                                    subcode => "5502",                                               class => "WPL AR",      reason => "Elektroheizer in Tropfschale in W&auml;rmepumpe Z2 au&szlig;er Funktion",                                                                                                                                                                                                                                                                          checklist => "F&uuml;hler TA4 au&szlig;er Funktion&bull;I/O-Board au&szlig;er Funktion&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                                                                                                                                        correction => "F&uuml;hlerplatzierung sowie F&uuml;hlerwert im Vergleich zur aktuellen F&uuml;hlertabelle und tats&auml;chlichen Temperatur &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob Ausgang EA0-Klemme 27 (80, N) bei manuellem Betrieb aus Tropfschalenw&auml;rmer 230V hat&bull;Fehlt Spannung wie oben, I/O-Board wechseln"                                                                       };
	$ErrorCodeList{"1399"} = {errorcode => "H01",                                    subcode => "5505",                                               class => "WPL AR",      reason => "Elektroheizer in Tropfschale in W&auml;rmepumpe Z1 au&szlig;er Funktion",                                                                                                                                                                                                                                                                          checklist => "Elektroheizer in Tropfschale au&szlig;er Funktion&bull;F&uuml;hler TA4 au&szlig;er Funktion&bull;I/O-Board au&szlig;er Funktion&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                                                                                                                                                                 correction => "Widerstand des Elektroheizers messen, 75 W; ~720 Ohm, 50 W ~1070 Ohm, bei Unterbrechung Elektroheizer wechseln&bull;F&uuml;hlerwert im Vergleich zur aktuellen F&uuml;hlertabelle und tats&auml;chlichen Temperatur &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob Ausgang EA0-Klemme 27; (80, N) bei manuellem Betrieb aus Tropfschalenw&auml;rmer 230 V hat&bull;Fehlt Spannung wie oben, I/O-Board wechseln"};
	$ErrorCodeList{"1400"} = {errorcode => "H01",                                    subcode => "5508",                                               class => "WPL AR",      reason => "MR1 hoher Druck Alarm A in W&auml;rmepumpe Z1",                                                                                                                                                                                                                                                                                                    checklist => "Siehe Warnung 5507 oben",                                                                                                                                                                                                                                                                                                                                                                                                                                                                             correction => "Siehe Ma&szlig;nahmen f&uuml;r Warnung 5507 oben"                                                                                                                                                                                                                                                                                                                                                };
	$ErrorCodeList{"1401"} = {errorcode => "H01",                                    subcode => "5510",                                               class => "WPL AR",      reason => "MR1 hoher Druck Alarm Alarm A in W&auml;rmepumpe Z1",                                                                                                                                                                                                                                                                                              checklist => "Pressostat au&szlig;er Funktion&bull;Unterbrechung/schlechte Verbindung im Kabelbaum zwischen MR1 und Inverter&bull;Unterbrechung/schlechte Verbindung im Kabelbaum zwischen MR1 und Inverter",                                                                                                                                                                                                                                                                                                       correction => "Aufgebrochener Kreis < 43 bar = Pressostat wechseln&bull;Kabelbaum/Anschl&uuml;sse zwischen MR1 und Inverter &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob Hochdruckw&auml;chter angeschlossen ist"                                                                                                                                                                                         };
	$ErrorCodeList{"1402"} = {errorcode => "H01",                                    subcode => "5513",                                               class => "WPL AR",      reason => "Zu hohe Kondensation in W&auml;rmepumpe Z1",                                                                                                                                                                                                                                                                                                       checklist => "Zu niedrige Verdampftemperatur im Verh&auml;ltnis zur Kondensationstemperatur&bull;Zu hohe Stromrate/; niedriges Delta im Heizzystem",                                                                                                                                                                                                                                                                                                                                                                correction => "Wahrscheinlich zu niedrige Au&szlig;entemperatur. Siehe Diagramm Kapitel 4 in Installationshandbuch&bull;Temperatur/Stromrate einstellen"                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"1403"} = {errorcode => "H01",                                    subcode => "5515",                                               class => "WPL AR",      reason => "Alarm geringe Verdampfung JR0 in W&auml;rmepumpe Z1",                                                                                                                                                                                                                                                                                              checklist => "Siehe Pr&uuml;fvorgang/; Ursachen f&uuml;r Ursachencode 5514",                                                                                                                                                                                                                                                                                                                                                                                                                                        correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5514"                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1404"} = {errorcode => "H01",                                    subcode => "5518",                                               class => "WPL AR",      reason => "Kondensationstemperatur au&szlig;erhalb der Grenze in W&auml;rmepumpe Z2",                                                                                                                                                                                                                                                                         checklist => "Zu niedrige Verdampftemperatur im Verh&auml;ltnis zur Kondensationstemperatur",                                                                                                                                                                                                                                                                                                                                                                                                                       correction => "Wahrscheinlich zu niedrige Au&szlig;entemperatur. Siehe Diagramm Kapitel 4 in Installationshandbuch"                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1405"} = {errorcode => "H01",                                    subcode => "5520",                                               class => "WPL AR",      reason => "Warnung niedriger Druck JR0 in W&auml;rmepumpe Z2",                                                                                                                                                                                                                                                                                                checklist => "Blockierter/niedriger Luftstrom zu Luftw&auml;rmetauscher an W&auml;rmepumpe&bull;Gebl&auml;se gem&auml;&szlig; Test Au&szlig;eneinheit &uuml;berpr&uuml;fen&bull;Gebl&auml;se hat nicht 230 V Spannung&bull;Gebl&auml;se ohne 0: 10 V-Signal&bull;Gebl&auml;se defekt&bull;I/O-Board au&szlig;er Funktion",                                                                                                                                                                                          correction => "Ausreichenden Luftstrom &uuml;ber Luftw&auml;rmetauscher sicherstellen.&bull;Test Au&szlig;eneinheit aktivieren&bull;&Uuml;berpr&uuml;fen, dass 230 V Spannung an Ausgang PL3 Klemme 32 (78, N)&bull;0: 10 V Spannung an Ausgang PL3 PWM Klemme 20 (20, 26) laut Test Au&szlig;eneinheit &uuml;berpr&uuml;fen&bull;Ist Spannung wie oben vorhanden, Gebl&auml;se wechseln&bull;Fehlt Spannung wie oben, I/O-Board wechseln"};
	$ErrorCodeList{"1406"} = {errorcode => "H01",                                    subcode => "5523",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z1 PFC &Uuml;berstrom des Kompressortreibers",                                                                                                                                                                                                                                                                             checklist => "Interner Fehler im Inverter",                                                                                                                                                                                                                                                                                                                                                                                                                                                                         correction => "Inverter wechseln"                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1407"} = {errorcode => "H01",                                    subcode => "5527",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z1 zu viel K&uuml;hlmittel in W&auml;rmepumpe",                                                                                                                                                                                                                                                                              checklist => "W&auml;rmepumpe &uuml;berf&uuml;llt",                                                                                                                                                                                                                                                                                                                                                                                                                                                                 correction => "W&auml;rmepumpe entleeren und Menge laut Typenschild auff&uuml;llen"                                                                                                                                                                                                                                                                                                                             };
	$ErrorCodeList{"1408"} = {errorcode => "H01",                                    subcode => "5529",                                               class => "WPL AR",      reason => "Warnung W&auml;rmepumpe Z2 Kompressorf&uuml;hler TR1 Kurzschluss",                                                                                                                                                                                                                                                                                 checklist => "Kompressorf&uuml;hler TR1 kurzgeschlossen&bull;Signalkabel zu Kompressorf&uuml;hler TR1 kurzgeschlossen&bull;I/O-Board au&szlig;er Funktion&bull;Siehe Pr&uuml;fvorgang/; Ursachen f&uuml;r Ursachencode 5314, 5315",                                                                                                                                                                                                                                                                                 correction => "Ohmwert im Vergleich zur aktuellen; F&uuml;hlertabelle und tats&auml;chlichen Temperatur &uuml;berpr&uuml;fen/F&uuml;hler TR1 wechseln&bull;Signalkabel zu Kompressorf&uuml;hler TR1 reparieren/wechseln&bull;I/O-Board wechseln&bull;Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5314, 5315"                                                                                      };
	$ErrorCodeList{"1409"} = {errorcode => "H01",                                    subcode => "5533",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z1 Kompressorf&uuml;hler TR1 defekt",                                                                                                                                                                                                                                                                                        checklist => "Siehe Pr&uuml;fvorgang/; Ursachen f&uuml;r Ursachencode 5531, 5532",                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5531, 5532"                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1410"} = {errorcode => "H01",                                    subcode => "5536",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z2 Temperaturf&uuml;hler TA4 defekt",                                                                                                                                                                                                                                                                                        checklist => "Siehe Pr&uuml;fvorgang/Ursachen f&uuml;r Ursachencode 5534, 5535",                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5534, 5535"                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1411"} = {errorcode => "H01",                                    subcode => "5537",                                               class => "WPL AR",      reason => "Schl&auml;uche zwischen Innenteil und W&auml;rmepumpe Z2 wurden gewechselt",                                                                                                                                                                                                                                                                       checklist => "Schl&auml;uche (Steig/R&uuml;cklauf) zwischen W&auml;rmepumpe und Innenteil verwechselt&bull;Defekt F&uuml;hler TC3, TC0",                                                                                                                                                                                                                                                                                                                                                                            correction => "Anschl&uuml;sse der Schl&auml;uche &uuml;berpr&uuml;fen&bull;F&uuml;hler im Vergleich zur aktuellen F&uuml;hlertabelle und tats&auml;chlichen Temperatur &uuml;berpr&uuml;fen/  bei Abweichung F&uuml;hler wechseln"                                                                                                                                                                             };
	$ErrorCodeList{"1412"} = {errorcode => "H01",                                    subcode => "5540",                                               class => "WPL AR",      reason => "Alarm W&auml;rmepumpe Z1 Temperaturf&uuml;hler TA4 defekt",                                                                                                                                                                                                                                                                                        checklist => "Siehe Pr&uuml;fvorgang/Ursachen f&uuml;r Ursachencode 5538, 5539",                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "Siehe m&ouml;gliche Ma&szlig;nahmen f&uuml;r Ursachencode 5538, 5539"                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1413"} = {errorcode => "H01",                                    subcode => "5541",                                               class => "WPL AR",      reason => "Kommunikation mit Schwimmbadkarte fehlgeschlagen",                                                                                                                                                                                                                                                                                                 checklist => "Loser Kontakt in CANbus-Anschl&uuml;ssen an Installer-Board Tower und Schwimmbadkarte&bull;Unterbrechung an CANbus-Kabel zwischen Installer-Board Tower und Schwimmbadkarte&bull;Falscher Typ CANbus-Kabel&bull;CANbus-Kabel zusammen mit Stromversorgung der W&auml;rmepumpe verlegt&bull;Fehlerhafte Erdung des CANbus-Kabels",                                                                                                                                                                     correction => "Kontakt in CANbus-Anschl&uuml;ssen an; Installer-Board Tower und Schwimmbadkarte &uuml;berpr&uuml;fen&bull;CANbus-Kabel zwischen Installer-Board Tower und Schwimmbadkarte wechseln&bull;Zum richtigen Typ von Kabel wechseln&bull;CANbusund Stromkabel mit einem Abstand von mindestens 100 mm verlegen&bull;Kabelschutz zur/von der Erdung entfernen/abschlie&szlig;en"                        };
	$ErrorCodeList{"1414"} = {errorcode => "H01",                                    subcode => "5545",                                               class => "8",           reason => "Alarm Z1 Beide SoleTemperatursensoren TB0 und TB1 sind defekt",                                                                                                                                                                                                                                                                                    checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1415"} = {errorcode => "H01",                                    subcode => "5546",                                               class => "8",           reason => "Alarm Z2 Beide SoleTemperatursensoren TB0 und TB1 sind defekt",                                                                                                                                                                                                                                                                                    checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1416"} = {errorcode => "H01",                                    subcode => "5549",                                               class => "8",           reason => "Alarm Z1 Sole-Eintrittstemperatur an TB0 ist zu niedrig",                                                                                                                                                                                                                                                                                          checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1417"} = {errorcode => "H01",                                    subcode => "5550",                                               class => "8",           reason => "Alarm Z2 Sole-Eintrittstemperatur an TB0 ist zu niedrig",                                                                                                                                                                                                                                                                                          checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1418"} = {errorcode => "H01",                                    subcode => "5553",                                               class => "8",           reason => "Alarm Z1 Sole-Austrittstemperatur an TB1 ist zu niedrig",                                                                                                                                                                                                                                                                                          checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1419"} = {errorcode => "H01",                                    subcode => "5554",                                               class => "8",           reason => "Alarm Z2 Sole-Austrittstemperatur an TB1 ist zu niedrig",                                                                                                                                                                                                                                                                                          checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1420"} = {errorcode => "H01",                                    subcode => "5557",                                               class => "8",           reason => "Alarm Z1 Sole-Eintrittstemperatur an TB0 ist zu hoch",                                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1421"} = {errorcode => "H01",                                    subcode => "5558",                                               class => "8",           reason => "Alarm Z2 Sole-Eintrittstemperatur an TB0 ist zu hoch",                                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1422"} = {errorcode => "H01",                                    subcode => "5561",                                               class => "8",           reason => "Alarm Z1 Sole-Austrittstemperatur an TB1 ist zu hoch",                                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1423"} = {errorcode => "H01",                                    subcode => "5562",                                               class => "8",           reason => "Alarm Z2 Sole-Austrittstemperatur an TB1 ist zu hoch",                                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1424"} = {errorcode => "H01",                                    subcode => "5565",                                               class => "8",           reason => "Alarm Z1 Hohe Temperaturdiff.  zwischen TB0 und TB1",                                                                                                                                                                                                                                                                                              checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1425"} = {errorcode => "H01",                                    subcode => "5566",                                               class => "8",           reason => "Alarm Z2 Hohe Temperaturdiff. zwischen TB0 und TB1",                                                                                                                                                                                                                                                                                               checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1426"} = {errorcode => "H01",                                    subcode => "5569",                                               class => "8",           reason => "Alarm Z1 Fehler Solepumpe",                                                                                                                                                                                                                                                                                                                        checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1427"} = {errorcode => "H01",                                    subcode => "5570",                                               class => "8",           reason => "Alarm Z2 Fehler Solepumpe",                                                                                                                                                                                                                                                                                                                        checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1428"} = {errorcode => "H01",                                    subcode => "5575",                                               class => "8",           reason => "Alarm Z1 Sole-Eintrittstemperatursensor TB0 Fehler",                                                                                                                                                                                                                                                                                               checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1429"} = {errorcode => "H01",                                    subcode => "5576",                                               class => "8",           reason => "Alarm Z2 Sole-Eintrittstemperatursensor TB0 Fehler",                                                                                                                                                                                                                                                                                               checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1430"} = {errorcode => "H01",                                    subcode => "5581",                                               class => "8",           reason => "Alarm Z1 Sole-Austrittstemperatursensor TB1 Fehler",                                                                                                                                                                                                                                                                                               checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1431"} = {errorcode => "H01",                                    subcode => "5582",                                               class => "8",           reason => "Alarm Z2 Sole-Austrittstemperatursensor TB1 Fehler",                                                                                                                                                                                                                                                                                               checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1432"} = {errorcode => "H01",                                    subcode => "5583",                                               class => "8",           reason => "Alarm Z1 Hochdrucksensor detektiert Probleme im Solekreis",                                                                                                                                                                                                                                                                                        checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1433"} = {errorcode => "H01",                                    subcode => "5584",                                               class => "8",           reason => "Alarm Z2 Hochdrucksensor detektiert Probleme im Solekreis",                                                                                                                                                                                                                                                                                        checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1434"} = {errorcode => "H01",                                    subcode => "5585",                                               class => "8",           reason => "Alarm Z1 Niederdruck-sensor detektiert Probleme im Solekreis",                                                                                                                                                                                                                                                                                     checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1435"} = {errorcode => "H01",                                    subcode => "5586",                                               class => "8",           reason => "Alarm Z2 Niederdruck-sensor detektiert Probleme im Solekreis",                                                                                                                                                                                                                                                                                     checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1436"} = {errorcode => "H01",                                    subcode => "5589",                                               class => "8",           reason => "Alarm Vorlauftemperaturf&uuml;hler/K&uuml;hlung TK1 Fehler",                                                                                                                                                                                                                                                                                       checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1437"} = {errorcode => "H01",                                    subcode => "5592",                                               class => "8",           reason => "Alarm Eingangstemperaturf&uuml;hler/K&uuml;hlung TK2 Fehler",                                                                                                                                                                                                                                                                                      checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1438"} = {errorcode => "H01",                                    subcode => "5593",                                               class => "4",           reason => "Warnung Keine Kommunikation mit Poolmodul",                                                                                                                                                                                                                                                                                                        checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1439"} = {errorcode => "H01",                                    subcode => "5594",                                               class => "8",           reason => "Alarm Z1 Luft im System",                                                                                                                                                                                                                                                                                                                          checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1440"} = {errorcode => "H01",                                    subcode => "5595",                                               class => "8",           reason => "Alarm Z2 Luft im System",                                                                                                                                                                                                                                                                                                                          checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1441"} = {errorcode => "H01",                                    subcode => "5596",                                               class => "8",           reason => "Alarm Z1 Bitte Luftfilter reinigen",                                                                                                                                                                                                                                                                                                               checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1442"} = {errorcode => "H01",                                    subcode => "5599",                                               class => "8",           reason => "Alarm Z1 Verdampfer-temp.sensor TR4 Fehler",                                                                                                                                                                                                                                                                                                       checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1443"} = {errorcode => "H01",                                    subcode => "5601",                                               class => "8",           reason => "Alarm Kompressor blockiert (IPM Fehler)",                                                                                                                                                                                                                                                                                                          checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1444"} = {errorcode => "H01",                                    subcode => "5603",                                               class => "8",           reason => "Alarm Stromaufnahme Inverter zu gro&szlig;",                                                                                                                                                                                                                                                                                                       checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1445"} = {errorcode => "H01",                                    subcode => "5605",                                               class => "8",           reason => "Alarm Gleichspannung au&szlig;erhalb zul&auml;ssigen Bereichs",                                                                                                                                                                                                                                                                                    checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1446"} = {errorcode => "H01",                                    subcode => "5607",                                               class => "8",           reason => "Alarm Hochoder Niederdruckschalter hat ausgel&ouml;st",                                                                                                                                                                                                                                                                                            checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1447"} = {errorcode => "H01",                                    subcode => "5609",                                               class => "8",           reason => "Alarm Eingangssp. zu klein (< 140VAC) oder zu hoch (> 300VAC)",                                                                                                                                                                                                                                                                                    checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1448"} = {errorcode => "H01",                                    subcode => "5611",                                               class => "8",           reason => "Alarm Kompressorphasenfolge falsch angeschlossen",                                                                                                                                                                                                                                                                                                 checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1449"} = {errorcode => "H01",                                    subcode => "5613",                                               class => "8",           reason => "Alarm Zu hohe Stromaufnahme am IGBT, Spule pr&uuml;fen",                                                                                                                                                                                                                                                                                           checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1450"} = {errorcode => "H01",                                    subcode => "5615",                                               class => "8",           reason => "Alarm Gleichspannung gr&ouml;&szlig;er als 420 VDC.",                                                                                                                                                                                                                                                                                              checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1451"} = {errorcode => "H01",                                    subcode => "5617",                                               class => "8",           reason => "Alarm Stromaufnahme des Kompressors zu gro&szlig;",                                                                                                                                                                                                                                                                                                checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1452"} = {errorcode => "H01",                                    subcode => "5619",                                               class => "8",           reason => "Alarm Heissgastemperatur zu hoch",                                                                                                                                                                                                                                                                                                                 checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1453"} = {errorcode => "H01",                                    subcode => "5621",                                               class => "8",           reason => "Alarm CT Sensor Fehler",                                                                                                                                                                                                                                                                                                                           checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1454"} = {errorcode => "H01",                                    subcode => "5623",                                               class => "8",           reason => "Alarm Heissgassensor TR6 Fehler",                                                                                                                                                                                                                                                                                                                  checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1455"} = {errorcode => "H01",                                    subcode => "5625",                                               class => "8",           reason => "Alarm Lufttemperatursensor Ausseneinheit TL2 Fehler",                                                                                                                                                                                                                                                                                              checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1456"} = {errorcode => "H01",                                    subcode => "5627",                                               class => "8",           reason => "Alarm VerdampferMitteltemp.-sensor TR8 Fehler",                                                                                                                                                                                                                                                                                                    checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1457"} = {errorcode => "H01",                                    subcode => "5629",                                               class => "8",           reason => "Alarm Sauggastemperatursensor TR5 Fehler",                                                                                                                                                                                                                                                                                                         checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1458"} = {errorcode => "H01",                                    subcode => "5631",                                               class => "8",           reason => "Alarm Nennsparnnung zu hoch Thermische Belastung zu hoch",                                                                                                                                                                                                                                                                                         checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1459"} = {errorcode => "H01",                                    subcode => "5633",                                               class => "8",           reason => "Alarm Interne Inverter-Kommunikation gest&ouml;rt",                                                                                                                                                                                                                                                                                                checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1460"} = {errorcode => "H01",                                    subcode => "5635",                                               class => "8",           reason => "Alarm Interne CAN-Kommunikation der Au&szlig;eneinheit gest&ouml;rt",                                                                                                                                                                                                                                                                              checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1461"} = {errorcode => "H01",                                    subcode => "5637",                                               class => "8",           reason => "Alarm Phasenreihenfolge falsch oder Phase fehlt",                                                                                                                                                                                                                                                                                                  checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1462"} = {errorcode => "H01",                                    subcode => "5639",                                               class => "8",           reason => "Alarm Defektes EEPROM auf der Platine in Au&szlig;eneinheit",                                                                                                                                                                                                                                                                                      checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1463"} = {errorcode => "H01",                                    subcode => "5641",                                               class => "8",           reason => "Alarm Kondensations-temp. zu hoch (gr&ouml;&szlig;er 65°C)",                                                                                                                                                                                                                                                                                      checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1464"} = {errorcode => "H01",                                    subcode => "5643",                                               class => "8",           reason => "Alarm Temperatur am Frequenzumrichter zu hoch (gr&ouml;&szlig;er 85°C)",                                                                                                                                                                                                                                                                          checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1465"} = {errorcode => "H01",                                    subcode => "5645",                                               class => "8",           reason => "Alarm Kondensationstemperatur zu niedrig",                                                                                                                                                                                                                                                                                                         checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1466"} = {errorcode => "H01",                                    subcode => "5647",                                               class => "8",           reason => "Alarm Frequenzumrichter Temperaturf&uuml;hler Fehler",                                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1467"} = {errorcode => "H01",                                    subcode => "5649",                                               class => "8",           reason => "Alarm Kompressor Anlaufstrom zu hoch",                                                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1468"} = {errorcode => "H01",                                    subcode => "5652",                                               class => "8",           reason => "Alarm Z1 Kondensatoraustrittstemperaturf&uuml;hler TR7 Fehler",                                                                                                                                                                                                                                                                                    checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1469"} = {errorcode => "H01",                                    subcode => "5655",                                               class => "8",           reason => "Alarm Z2 Kondensatoraustrittstemperaturf&uuml;hler TR7 Fehler",                                                                                                                                                                                                                                                                                    checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1470"} = {errorcode => "H01",                                    subcode => "5658",                                               class => "8",           reason => "Alarm Z1 Kondensatoreintrittstemperaturf&uuml;hler TR3 Fehler",                                                                                                                                                                                                                                                                                    checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1471"} = {errorcode => "H01",                                    subcode => "5661",                                               class => "8",           reason => "Alarm Z2 Kondensatoreintrittstemperaturf&uuml;hler TR3 Fehler",                                                                                                                                                                                                                                                                                    checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1472"} = {errorcode => "H01",                                    subcode => "5663",                                               class => "8",           reason => "Alarm Z1 Drucksensor Fehler",                                                                                                                                                                                                                                                                                                                      checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1473"} = {errorcode => "H01",                                    subcode => "5665",                                               class => "8",           reason => "Warnung Z1 Eintritt Verdampfertemperatur-f&uuml;hler TR4 Fehler",                                                                                                                                                                                                                                                                                  checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1474"} = {errorcode => "H01",                                    subcode => "5667",                                               class => "8",           reason => "Alarm Z1 Verdampfungsdruck zu niedrig",                                                                                                                                                                                                                                                                                                            checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1475"} = {errorcode => "H01",                                    subcode => "5669",                                               class => "8",           reason => "Alarm Z1 Kommunikation zwischen Innenund Au&szlig;eneinheit gest&ouml;rt",                                                                                                                                                                                                                                                                         checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1476"} = {errorcode => "H01",                                    subcode => "5671",                                               class => "8",           reason => "Alarm Z1 Ventilator Au&szlig;eneinheit ist blockiert",                                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1477"} = {errorcode => "H01",                                    subcode => "5673",                                               class => "8",           reason => "Alarm Z2 Drucksensor Fehler",                                                                                                                                                                                                                                                                                                                      checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1478"} = {errorcode => "H01",                                    subcode => "5675",                                               class => "8",           reason => "Warnung Z2 Eintritt Verdampfertemperatur-f&uuml;hler TR4 Fehler",                                                                                                                                                                                                                                                                                  checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1479"} = {errorcode => "H01",                                    subcode => "5677",                                               class => "8",           reason => "Alarm Z2 Druck zu niedrig",                                                                                                                                                                                                                                                                                                                        checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1480"} = {errorcode => "H01",                                    subcode => "5679",                                               class => "8",           reason => "Alarm Z2 Kommunikation zwischen Innenund Au&szlig;eneinheit gest&ouml;rt",                                                                                                                                                                                                                                                                         checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1481"} = {errorcode => "H01",                                    subcode => "5681",                                               class => "8",           reason => "Alarm Z2 Ventilator Au&szlig;eneinheit ist blockiert",                                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1482"} = {errorcode => "H01",                                    subcode => "5702",                                               class => "8",           reason => "Alarm Z1 Vorlauftemperatursensor TC2 Fehler",                                                                                                                                                                                                                                                                                                      checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1483"} = {errorcode => "H01",                                    subcode => "5705",                                               class => "4",           reason => "Alarm Z1 R&uuml;cklauftemperatursensor TC4 Fehler",                                                                                                                                                                                                                                                                                                checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1484"} = {errorcode => "H01",                                    subcode => "5711",                                               class => "8",           reason => "Alarm Z1 Temperaturf&uuml;hler Luftaustritt TL3 defekt",                                                                                                                                                                                                                                                                                           checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1485"} = {errorcode => "H01",                                    subcode => "5712",                                               class => "8",           reason => "Alarm Z1 WP-Typ DIP Schalter ist nicht korrekt eingestellt",                                                                                                                                                                                                                                                                                       checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1486"} = {errorcode => "H01",                                    subcode => "5714",                                               class => "8",           reason => "Alarm Z1 Einspritzdrucksensor defekt",                                                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1487"} = {errorcode => "H01",                                    subcode => "5715",                                               class => "8",           reason => "Alarm Z1 Keine Leistungsabgabe",                                                                                                                                                                                                                                                                                                                   checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1488"} = {errorcode => "H01",                                    subcode => "5716",                                               class => "8",           reason => "Alarm Z1 K&auml;ltemittelleckage. Expansionsventil arbeitet nicht richtig",                                                                                                                                                                                                                                                                        checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1489"} = {errorcode => "H01",                                    subcode => "5744",                                               class => "8",           reason => "Alarm Z1 Inverterfehler Fehlergrenze erreicht",                                                                                                                                                                                                                                                                                                    checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1490"} = {errorcode => "H01",                                    subcode => "5750",                                               class => "8",           reason => "Alarm Z2 Vorlauftemperatursensor TC2 Fehler",                                                                                                                                                                                                                                                                                                      checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1491"} = {errorcode => "H01",                                    subcode => "5753",                                               class => "8",           reason => "Alarm Z2 R&uuml;cklauftemperatursensor TC4 Fehler",                                                                                                                                                                                                                                                                                                checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1492"} = {errorcode => "H01",                                    subcode => "5756",                                               class => "8",           reason => "Alarm Z2 Einspritztemperatursensor TR2  Fehler",                                                                                                                                                                                                                                                                                                   checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1493"} = {errorcode => "H01",                                    subcode => "5759",                                               class => "8",           reason => "Alarm Z2 Temperaturf&uuml;hler Luftaustritt TL3 defekt",                                                                                                                                                                                                                                                                                           checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1494"} = {errorcode => "H01",                                    subcode => "5760",                                               class => "8",           reason => "Alarm Z2 WP-Typ DIP Schalter ist nicht korrekt eingestellt",                                                                                                                                                                                                                                                                                       checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1495"} = {errorcode => "H01",                                    subcode => "5762",                                               class => "8",           reason => "Alarm Z2 Einspritzdrucksensor defekt",                                                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1496"} = {errorcode => "H01",                                    subcode => "5763",                                               class => "8",           reason => "Alarm Z2 Keine Leistungsabgabe",                                                                                                                                                                                                                                                                                                                   checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1497"} = {errorcode => "H01",                                    subcode => "5764",                                               class => "8",           reason => "Alarm Z2 K&auml;ltemittelleckage. Expansionsventil arbeitet nicht richtig",                                                                                                                                                                                                                                                                        checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1498"} = {errorcode => "H01",                                    subcode => "5784",                                               class => "4",           reason => "Alarm Z2 Invertertemperatur Balance",                                                                                                                                                                                                                                                                                                              checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1499"} = {errorcode => "H01",                                    subcode => "5792",                                               class => "8",           reason => "Alarm Z2 Inverterfehler Fehlergrenze erreicht",                                                                                                                                                                                                                                                                                                    checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1500"} = {errorcode => "H01",                                    subcode => "5838",                                               class => "8",           reason => "Warnung Z2 Temperaturdifferenz zw. den drei Inverter IGBT`s",                                                                                                                                                                                                                                                                                      checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1501"} = {errorcode => "H01",                                    subcode => "5841",                                               class => "8",           reason => "Alarm Z1 Inverter Kommunikation. zw. Signalund Hauptprozessor gest&ouml;rt",                                                                                                                                                                                                                                                                       checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1502"} = {errorcode => "H01",                                    subcode => "5842",                                               class => "8",           reason => "Alarm Z1 Einspritztemperatursensor TR2  Fehler",                                                                                                                                                                                                                                                                                                   checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1503"} = {errorcode => "H01",                                    subcode => "5844",                                               class => "8",           reason => "Alarm Z1 Inverter Kommunikation. zw.Hauptund Signalprozessor gest&ouml;rt",                                                                                                                                                                                                                                                                        checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1504"} = {errorcode => "H01",                                    subcode => "5845",                                               class => "8",           reason => "Alarm Z1 Inverterfehler Verdichter-Scroll Temp.; zu niedrig",                                                                                                                                                                                                                                                                                      checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1505"} = {errorcode => "H01",                                    subcode => "5846",                                               class => "8",           reason => "Alarm Z1 Inverterfehler Verdichter-Motor Temp.; zu niedrig",                                                                                                                                                                                                                                                                                       checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1506"} = {errorcode => "H01",                                    subcode => "5847",                                               class => "8",           reason => "Alarm Z1 Inverterfehler Schaltkreis Temp. zu niedrig",                                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1507"} = {errorcode => "H01",                                    subcode => "5849",                                               class => "8",           reason => "Alarm Z1 Inverterfehler IGBT`s Temp. zu niedrig",                                                                                                                                                                                                                                                                                                  checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1508"} = {errorcode => "H01",                                    subcode => "5854",                                               class => "4",           reason => "Alarm Z2 Inverter Kommunikation. zw. Signalund Hauptprozessor gest&ouml;rt",                                                                                                                                                                                                                                                                       checklist => "&Uuml;berpr&uuml;fen, ob die Lamellen am Verdampfer in der Au&szlig;eneinheit verschmutzt sind&bull;&Uuml;berpr&uuml;fen, ob die Wartungsh&auml;hne in den Verbindungsleitungen zwischen Innenund Au&szlig;eneinheit geschlossen sind&bull;&Uuml;berpr&uuml;fen, ob de Vorund R&uuml;cklauf an der Au&szlig;eneinheit korrekt angeschlossen ist&bull;&Uuml;berpr&uuml;fen, ob sich Schmutzablagerungen im Filter in den Verbindungsleitungen zwischen Innenund Au&szlig;eneinheit befinden&bull;&Uuml;berpr&uuml;fen, ob die Umw&auml;lzpumpe im Solekreis zwischen Innenund Au&szlig;eneinheit blockiert ist&bull;",correction => "Luftseitige Lamellen am Verdampfer in der Au&szlig;eneinheit reinigen&bull;Wartungsh&auml;hne in den Verbindungsleitugnen zwischen Innenund Au&szlig;eneineit &ouml;ffnen&bull;Vorund R&uuml;cklauf an der Au&szlig;eneinheit korrekt anschlie&szlig;en&bull;Filter in den Verbindungsleitugnen zwischen Innenund Au&szlig;eneinheit reinigen&bull;Umw&auml;lzpumpe im Solekreis zwischen Innenund Au&szlig;eneinheit anwerfen bzw. austauschen&bull;Zul&auml;ssige Solefl&uuml;ssigkeit verwenden"};
	$ErrorCodeList{"1509"} = {errorcode => "H01",                                    subcode => "5855",                                               class => "8",           reason => "Alarm Z2 Inverter Kommunikation. zw. Hauptund Signalprozessor gest&ouml;rt",                                                                                                                                                                                                                                                                       checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1510"} = {errorcode => "H01",                                    subcode => "5856",                                               class => "8",           reason => "Alarm Z2 Inverterfehler Verdichter-Scroll Temp.; zu niedrig",                                                                                                                                                                                                                                                                                      checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1511"} = {errorcode => "H01",                                    subcode => "5857",                                               class => "8",           reason => "Alarm Z2 Inverterfehler Verdichter-Motor Temp.; zu niedrig",                                                                                                                                                                                                                                                                                       checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1512"} = {errorcode => "H01",                                    subcode => "5858",                                               class => "8",           reason => "Alarm Z2 Inverterfehler Schaltkreis Temp. zu niedrig",                                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1513"} = {errorcode => "H01",                                    subcode => "5859",                                               class => "8",           reason => "Alarm Z2 Inverterfehler IGBT`s Temp. zu niedrig",                                                                                                                                                                                                                                                                                                  checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1514"} = {errorcode => "H01",                                    subcode => "5860",                                               class => "8",           reason => "Alarm Z2 Inverterfehler PFC IGBT's Temp. zu niedrig",                                                                                                                                                                                                                                                                                              checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1515"} = {errorcode => "H01",                                    subcode => "5865",                                               class => "4",           reason => "Alarm Z1 Kommunikationsst&ouml;rung zwischen IWS und Inverter",                                                                                                                                                                                                                                                                                    checklist => "&Uuml;berpr&uuml;fenb, ob der Anschlussstecker f&uuml;r die Drehzahlregelung am Gebl&auml;se in der Au&szlig;eneinheit komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob der Anschlussstecker f&uuml;r die Versorgungsspannung am Gebl&auml;se in der Au&szlig;eneinheit komplett aufgesteckt ist&bull;Funktion des Gebl&auml;ses in der Au&szlig;eneinheit &uuml;berpr&uuml;fen&bull;&Uuml;berpr&uuml;fen, ob der Anschlussstecker f&uuml;r die Drehzahlregelung  zums Gebl&auml;se im Steuerger&auml;t der Au&szlig;eneinheit komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob der Anschlussstecker f&uuml;r die Versorgungsspannung zum Gebl&auml;se im Steuerger&auml;t der Au&szlig;eneinheit komplett aufgesteckt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel  f&uuml;r die Drehzahlregelung zum Gebl&auml;se besch&auml;digt ist&bull;&Uuml;berpr&uuml;fen, ob das Anschlusskabel f&uuml;r die Versorgungsspannung zum Gebl&auml;se in der Au&szlig;eneinheit besch&auml;digt ist&bull;",correction => "Anschlussstecker f&uuml;r die Drehzahlregelung am Gebl&auml;se in der Au&szlig;eneinheit komplett aufstecken&bull;Anschlussstecker f&uuml;r die Versorgungsspannung am Gebl&auml;se in der Au&szlig;eneinheit komplett aufstecken&bull;Gebl&auml;se bzw. Gebl&auml;semotor  in der Au&szlig;eneinheit austauschen&bull;Anschlussstecker f&uuml;r die Drehzahlregelung zum Gebl&auml;se im Steuerger&auml;t der Au&szlig;eneinheit komplett aufstecken&bull;Anschlussstecker f&uuml;r die Versorgungsspannung zum Gebl&auml;se im Steuerger&auml;t der Au&szlig;eneinheit komplett aufstecken&bull;Anschlusskabel f&uuml;r die Drehzahlregelung zum Gebl&auml;se in der Au&szlig;eneinheit austauschen&bull;Anschlusskabel f&uuml;r die Spannungsversorgung zum Gebl&auml;se in der Au&szlig;eneineit austauschen&bull;Steuerger&auml;t f&uuml;r die Gebl&auml;segelung in der Au&szlig;eneinheit austauschen"};
	$ErrorCodeList{"1516"} = {errorcode => "H01",                                    subcode => "5867",                                               class => "8",           reason => "Alarm Z1 Kommunikationsfehler zwischen IWS und Inverter",                                                                                                                                                                                                                                                                                          checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1517"} = {errorcode => "H01",                                    subcode => "5869",                                               class => "8",           reason => "Alarm Z2 Kommunikationsst&ouml;rung zwischen IWS und Inverter",                                                                                                                                                                                                                                                                                    checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1518"} = {errorcode => "H01",                                    subcode => "5871",                                               class => "8",           reason => "Alarm Z2 Kommunikationsfehler zwischen IWS und Inverter",                                                                                                                                                                                                                                                                                          checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1519"} = {errorcode => "H01",                                    subcode => "5878",                                               class => "8",           reason => "Alarm Z2 Sauggastemperatursensor TR5 Fehler",                                                                                                                                                                                                                                                                                                      checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1520"} = {errorcode => "H01",                                    subcode => "5881",                                               class => "8",           reason => "Alarm Z1 Sauggastemperatursensor TR5 Fehler",                                                                                                                                                                                                                                                                                                      checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1521"} = {errorcode => "H01",                                    subcode => "5884",                                               class => "8",           reason => "Alarm Z1 Inverterfehler",                                                                                                                                                                                                                                                                                                                          checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1522"} = {errorcode => "H01",                                    subcode => "5886",                                               class => "8",           reason => "Alarm Z2 Inverterfehler",                                                                                                                                                                                                                                                                                                                          checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1523"} = {errorcode => "H01",                                    subcode => "5888",                                               class => "8",           reason => "Alarm Z1 Mitteldrucksensor Fehler",                                                                                                                                                                                                                                                                                                                checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1524"} = {errorcode => "H01",                                    subcode => "5889",                                               class => "8",           reason => "Alarm Z1 Expansionsventil schliesst nicht",                                                                                                                                                                                                                                                                                                        checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1525"} = {errorcode => "H01",                                    subcode => "5890",                                               class => "8",           reason => "Alarm Z2 Mitteldrucksensor Fehler",                                                                                                                                                                                                                                                                                                                checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1526"} = {errorcode => "H01",                                    subcode => "5891",                                               class => "8",           reason => "Alarm Z2 Expansionsventil schliesst nicht",                                                                                                                                                                                                                                                                                                        checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1527"} = {errorcode => "H01",                                    subcode => "6202",                                               class => "8",           reason => "Alarm Z2 Vorlauftemperatursensor T0 Fehler",                                                                                                                                                                                                                                                                                                       checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1528"} = {errorcode => "H01",                                    subcode => "",                                                   class => "8",           reason => "Unterspannung am stromversorgung W&auml;rmepumpe Z2",                                                                                                                                                                                                                                                                                              checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1529"} = {errorcode => "H01",                                    subcode => "",                                                   class => "8",           reason => "Unterspannung am Stromversorgung W&auml;rmepumpe Z2",                                                                                                                                                                                                                                                                                              checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1530"} = {errorcode => "H02",                                    subcode => "1012",                                               class => "18",          reason => "Gebl&auml;se l&auml;uft nicht korrekt (Wenn die Luftersteuerung bei einer konstanten Vorgabe nach einer bestimmten Zeit mit dem PWM-; Signal bestimmte Grenzen unterschreitet, wird diese Wartungsmeldung erzeugt)",                                                                                                                               checklist => "Gebl&auml;se auf Verschmutzung pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                                          correction => "Ggf. reinigen oder austauschen"                                                                                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1531"} = {errorcode => "H03",                                    subcode => "1013",                                               class => "18",          reason => "Maximale Betriebsstunden erreicht; (Wenn eine bestimmte, vorher im RC3x eingestellte Anzahl an Betriebsstunden abgelaufen ist, wird diese Wartungsmeldung erzeugt)",                                                                                                                                                                               checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Wartung durchf&uuml;hren"                                                                                                                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"1532"} = {errorcode => "H04",                                    subcode => "1014",                                               class => "18",          reason => "Aktuelle Ionisation ist zu niedrig; (Wenn innerhalb der letzten Brennerstarts 5 niedrige Flammensignale am Ende der Sicherheitszeit festgestellt wurden, wird diese Wartungsmeldung erzeugt)",                                                                                                                                                     checklist => "Flammenf&uuml;hler und Winkelhalter (Spiegel) auf Verschmutzung pr&uuml;fen&bull;Mischsystem auf Verschmutzung pr&uuml;fen&bull;Brennereinstellung nach Vorgaben (Serviceanleitung) pr&uuml;fen&bull;Im Men&uuml; Monitor der Bedieneinheit in der; 1. und 2. Brennerstufe den Flammenstrom pr&uuml;fen, ob dieser ca. 50 μA betr&auml;gt",                                                                                                                                                        correction => "Wenn erforderlich reinigen&bull;Wenn erforderlich reinigen&bull;Wenn erforderlich korrigieren&bull;Sollte der Flammenstrom geringer sein, dann Flammenf&uuml;hler austauschen"                                                                                                                                                                                                                   };
	$ErrorCodeList{"1533"} = {errorcode => "H05",                                    subcode => "1015",                                               class => "18",          reason => "Z&uuml;ndungsdauer zu hoch (Wenn innerhalb der letzten Brennerstarts 5 hohe Zundverzuge festgestellt wurden, wird diese Wartungsmeldung erzeugt)",                                                                                                                                                                                                 checklist => "Pr&uuml;fen, ob eine korrekte Brennstoffversorgung sichergestellt ist; Kapitel „&Ouml;lversorgungsEinrichtung“ in Serviceanleitung beachten!&bull;Z&uuml;ndung mittels Relaistest (Bedieneinheit) pr&uuml;fen&bull;Mischsystem pr&uuml;fen&bull;&bull;&bull;Brennereinstellung nach Vorgaben pr&uuml;fen",                                                                                                                                                                                            correction => "Bei St&ouml;rung der Brennstoffversorgung diese beheben&bull;Z&uuml;ndelektrode auf Verschmutzung oder Besch&auml;digung (Elektrodenabstand) pr&uuml;fen, ggf. austauschen&bull;Brennstoffd&uuml;se austauschen&bull;&Ouml;labschlussventil des &Ouml;lvorw&auml;rmers austauschen&bull;Bei Bedarf reinigen&bull;Brennereinstellung wenn erforderlich korrigieren"                               };
	$ErrorCodeList{"1534"} = {errorcode => "H06",                                    subcode => "1016",                                               class => "18",          reason => "Zu viele Flammenunterbrechungen; (Bei den letzten Brennerstarts; kam es haufig zu Flammenabrissen)",                                                                                                                                                                                                                                               checklist => "Entst&ouml;rtaste am Feuerungsautomaten dr&uuml;cken und Ursache f&uuml;r Flammenabriss entsprechend folgenden Schritten suchen:&bull;Z&uuml;ndung mittels Relaistest der Bedieneinheit pr&uuml;fen&bull;Pr&uuml;fen, ob eine korrekte Brennstoffversorgung sichergestellt ist (siehe Kapitel „&Ouml;lversorgungsEinrichtung“ in Serviceanleitung)&bull;In der Serviceebene der Bedieneinheit den Men&uuml;punkt „Monitor“ aufrufen und hier den Flammenstrom pr&uuml;fen; Sollte dieser schwanken oder st&auml;ndig zu klein sein, folgendes Vorgehen w&auml;hlen:&bull;Flammenf&uuml;hler auf Verschmutzung pr&uuml;fen&bull;Position des Flammenf&uuml;hlers pr&uuml;fen (bei &Ouml;lkesseln Winkelhalter pr&uuml;fen)&bull;Kabel zwischen Feuerungsautomat und Flammenf&uuml;hler pr&uuml;fen&bull;Steckverbindung des Flammenf&uuml;hlers am Feuerungsautomaten pr&uuml;fen&bull;Brennstoffd&uuml;se pr&uuml;fen&bull;&Ouml;labschlussventil bei &Ouml;lkesseln optisch pr&uuml;fen&bull;Mischsystem bei &Ouml;lkesseln pr&uuml;fen&bull;Brennereinstellung pr&uuml;fen („Einstellwerte“)&bull;In der Serviceebene der Bedieneinheit den Men&uuml;punkt „St&ouml;rungsspeicher“ und danach das Untermen&uuml; „blockierende St&ouml;rungen“ aufrufen Sollte St&ouml;rungsmeldung 6L/516 vorliegen, pr&uuml;fen, ob Magnetventil 1 und 2 von der Zuordnung richtig am Feuerungsautomaten aufgesteckt sind&bull;Im Men&uuml; „Monitor“ der Bedieneinheit in der; 1. und 2. Brennerstufe den Flammenstrom pr&uuml;fen, ob dieser zu gering ist",correction => "-&bull;Z&uuml;ndelektrode auf Verschmutzung oder Besch&auml;digung (Elektrodenabstand) pr&uuml;fen, ggf. austauschen&bull;Bei St&ouml;rung der Brennstoffversorgung diese beheben&bull;-&bull;Brennereinstellung, &Ouml;lversorgung pr&uuml;fen, wenn erforderlich korrigieren&bull;-&bull;Brennstoffd&uuml;se ggf. austauschen&bull;&Ouml;labschlussventil des &Ouml;lvorw&auml;rmers ggf. austauschen&bull;Mischsystem reinigen oder austauschen&bull;Bei Abweichungen korrigieren&bull;Bei Vertauschungen diese beheben&bull;Sollte der Flammenstrom zu gering sein, dann Flammenf&uuml;hler reinigen bzw. austauschen"};
	$ErrorCodeList{"1535"} = {errorcode => "H07",                                    subcode => "1017",                                               class => "18",          reason => "Wasserdruck zu niedrig (Der Heizungsanlagendruck ist zu niedrig (< 0,2 bar bei Wandkesseln bzw. 0,8 bar bei GB312))",                                                                                                                                                                                                                              checklist => "Wasserdruck pr&uuml;fen&bull;Eventuell Druckf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                                                                                                                                                      correction => "Wasser nachf&uuml;llen und Anlage entl&uuml;ften&bull;Druckf&uuml;hler austauschen"                                                                                                                                                                                                                                                                                                              };
	$ErrorCodeList{"1536"} = {errorcode => "H08",                                    subcode => "1018",                                               class => "18",          reason => "Servicezeit abgelaufen (Wenn ein bestimmtes, vorher im RC3x eingestelltes Datum &uuml;berschritten wurde, wird diese Wartungsmeldung erzeugt)",                                                                                                                                                                                                    checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Wartung durchf&uuml;hren"                                                                                                                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"1537"} = {errorcode => "H09",                                    subcode => "1019",                                               class => "14",          reason => "Falscher Pumpentyp erkannt",                                                                                                                                                                                                                                                                                                                       checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Richtige Pumpe einbauen"                                                                                                                                                                                                                                                                                                                                                                         };
	$ErrorCodeList{"1538"} = {errorcode => "H10",                                    subcode => "1020",                                               class => "14",          reason => "Aktuelle Ionisation ist zu hoch; (Im Betrieb sind uber eine langere Zeit haufiger hohe Flammensignale gemessen worden)",                                                                                                                                                                                                                           checklist => "Das Flammensignal liegt oberhalb der Wartungsgrenze&bull;Elektrische Verbindung zwischen Flammenf&uuml;hler und Feuerungsautomat ist fehlerhaft&bull;Flammenf&uuml;hler ist defekt&bull;Feuerungsautomat ist defekt",                                                                                                                                                                                                                                                                                 correction => "Wartung durchf&uuml;hren&bull;Verbindungsleitungen zwischen Feuerungsautomat u. Flammenf&uuml;hler pr&uuml;fen u. St&ouml;rung beheben&bull;Flammenf&uuml;hler austauschen&bull;Feuerungsautomat austauschen"                                                                                                                                                                                    };
	$ErrorCodeList{"1539"} = {errorcode => "H1",                                     subcode => "70",                                                 class => "",            reason => "Vorlauftemperaturf&uuml;hler gemischter Heizkreis defekt",                                                                                                                                                                                                                                                                                         checklist => "Kein Signal vom Mischertemperaturf&uuml;hler&bull;Signal vom Mischerf&uuml;hler au&szlig;erhalb der Kennlinie&bull;Signalunterbrechung im Anschlusskabel&bull;Signalauswertung auf Leiterplatte defekt",                                                                                                                                                                                                                                                                                              correction => "Stecker am Mischertemperaturf&uuml;hler aufstecken&bull;F&uuml;hler austauschen&bull;Kabelbaum austauschen&bull;Leiterplatte austauschen"                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"1540"} = {errorcode => "H11",                                    subcode => "1021",                                               class => "16",          reason => "Temperaturf&uuml;hler des Schichtladespeichers defekt (Der Speichersensor ist defekt oder es besteht ein Kontaktproblem zum Sensor. Wartungsmeldung)",                                                                                                                                                                                             checklist => "Der F&uuml;hler SLS (Schicht-Lade-Sensor) oder Warmwasser ist falsch, d. h. nicht am Vorlaufrohr positioniert&bull;Der F&uuml;hler SLS (Schicht-Lade-Sensor) hat keinen Kontakt zum Vorlaufrohr&bull;Kabelverbindung zum F&uuml;hler SLS unterbrochen oder besch&auml;digt&bull;F&uuml;hler SLS defekt",                                                                                                                                                                                              correction => "Den F&uuml;hler SLS (Schicht-Lade-Sensor) bzw. Warmwasser richtig positionieren&bull;Den F&uuml;hler SLS (Schicht-Lade-Sensor) richtig am Vorlaufrohr befestigen&bull;Verbindungsleitung zum F&uuml;hler SLS reparieren bzw. austauschen&bull;F&uuml;hler austauschen"                                                                                                                           };
	$ErrorCodeList{"1541"} = {errorcode => "H12",                                    subcode => "1022",                                               class => "16",          reason => "Speichertemperaturf&uuml;hler defekt oder Kontaktprobleme",                                                                                                                                                                                                                                                                                        checklist => "Angezeigte Speichertemperatur auf Plausibilit&auml;t pr&uuml;fen. Steckverbindungen und Kabelbaum auf Kontakt pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                           correction => "Speichertemperaturf&uuml;hler ersetzen, Kontaktproblem beseitigen"                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1542"} = {errorcode => "H13",                                    subcode => "1023",                                               class => "18",          reason => "Maximale Betriebsdauer einschlie&szlig;lich Stand-by-Zeit ist erreicht",                                                                                                                                                                                                                                                                           checklist => "Wartung nach eingestellter Betriebszeit notwendig",                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "Wartung durchf&uuml;hren"                                                                                                                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"1543"} = {errorcode => "H14",                                    subcode => "1024",                                               class => "16",          reason => "Kommunikationsfehler: zwischen W&auml;rmeerzeuger und Bedieneinheit",                                                                                                                                                                                                                                                                              checklist => "Pr&uuml;fen, ob die Ger&auml;teelektronik richtig montiert ist bzw. richtig sitzt&bull;Pr&uuml;fen, ob die Busleitung zwischen Bedieneinheit und Ger&auml;teelektronik richtig montiert ist&bull;Ger&auml;teelektronik defekt",                                                                                                                                                                                                                                                                       correction => "Bedieneinheit richtig montieren&bull;Busleitung an der Ger&auml;teelektonik richtig montieren&bull;Ger&auml;teelektronik austauschen"                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1544"} = {errorcode => "H15",                                    subcode => "1025",                                               class => "16",          reason => "R&uuml;cklauftemperaturf&uuml;hler ist defekt",                                                                                                                                                                                                                                                                                                    checklist => "Kabelverbindung zum R&uuml;cklauftemperaturf&uuml;hler unterbrochen oder besch&auml;digt&bull;R&uuml;cklauftemperaturf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                                                                             correction => "Verbindungsleitung zum R&uuml;cklauftemperaturf&uuml;hler reparieren bzw. austauschen&bull;F&uuml;hler austauschen"                                                                                                                                                                                                                                                                              };
	$ErrorCodeList{"1545"} = {errorcode => "H16",                                    subcode => "1026",                                               class => "16",          reason => "Speichertemperaturf&uuml;hler Korrektur zu hoch",                                                                                                                                                                                                                                                                                                  checklist => "Temperaturunterschied zwischen Warmwasserr&uuml;cklauftemperaturf&uuml;hler und Warmwassertemperaturf&uuml;hler zu gro&szlig;&bull;F&uuml;hlerwerte pr&uuml;fen&bull;W&auml;rmetauscher auf Verkalkung pr&uuml;fen",                                                                                                                                                                                                                                                                                  correction => "-&bull;Ggf. F&uuml;hler austauschen&bull;W&auml;rmetauscher entkalken"                                                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"1546"} = {errorcode => "H17",                                    subcode => "1027",                                               class => "16",          reason => "Solarmodul: Speichertemperaturf&uuml;hler defekt (Das Regelger&auml;t erh&auml;lt unrealistische Werte vom Pufferf&uuml;hler (GBH172 [TS3]))",                                                                                                                                                                                                     checklist => "Kabelverbindung zum F&uuml;hler Puffer unterbrochen oder besch&auml;digt&bull;F&uuml;hler Puffer defekt",                                                                                                                                                                                                                                                                                                                                                                                             correction => "Verbindungsleitung zum F&uuml;hler Puffer reparieren bzw. austauschen&bull;F&uuml;hler austauschen"                                                                                                                                                                                                                                                                                              };
	$ErrorCodeList{"1547"} = {errorcode => "H18",                                    subcode => "1028",                                               class => "16",          reason => "Temperaturf&uuml;hler; 3-Wege-Mischer defekt",                                                                                                                                                                                                                                                                                                     checklist => "Kabelverbindung zum F&uuml;hler interner 3-WegeMischer unterbrochen oder besch&auml;digt&bull;F&uuml;hler interner 3-Wege-Mischer defekt",                                                                                                                                                                                                                                                                                                                                                            correction => "Verbindungsleitung zum F&uuml;hler interner 3-WegeMischer reparieren bzw. austauschen&bull;F&uuml;hler austauschen"                                                                                                                                                                                                                                                                              };
	$ErrorCodeList{"1548"} = {errorcode => "H19",                                    subcode => "1029",                                               class => "16",          reason => "3-Wege-Ventil Mischer defekt",                                                                                                                                                                                                                                                                                                                     checklist => "Montage des Mischertemperaturf&uuml;hlers pr&uuml;fen&bull;Motor des 3-Wege-Mischers pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                    correction => "Ggf. Mischertemperaturf&uuml;hler austauschen&bull;Ggf. Motor des 3-Wege-Mischers austauschen"                                                                                                                                                                                                                                                                                                   };
	$ErrorCodeList{"1549"} = {errorcode => "H20",                                    subcode => "1060",                                               class => "18",          reason => "Zu viele Startversuche des Kompressors oder klemmendes Kompressorrelais, W&auml;rmepumpe deaktiviert",                                                                                                                                                                                                                                             checklist => "St&ouml;rung des Kompressors der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "Leiterplattenschnittstelle der W&auml;rmepumpe austauschen"                                                                                                                                                                                                                                                                                                                                      };
	$ErrorCodeList{"1550"} = {errorcode => "H21",                                    subcode => "1061",                                               class => "18",          reason => "Zu viele Betriebsabbr&uuml;che ausgel&ouml;st durch K&auml;ltemitteldruck, W&auml;rmepumpe deaktiviert",                                                                                                                                                                                                                                           checklist => "St&ouml;rung aufgrund des K&auml;ltemitteldrucks",                                                                                                                                                                                                                                                                                                                                                                                                                                                    correction => "K&auml;ltemittelsystem pr&uuml;fen"                                                                                                                                                                                                                                                                                                                                                              };
	$ErrorCodeList{"1551"} = {errorcode => "H22",                                    subcode => "1062",                                               class => "18",          reason => "Zu viele Startversuche des Gebl&auml;ses, W&auml;rmepumpe deaktiviert",                                                                                                                                                                                                                                                                            checklist => "St&ouml;rung des Gebl&auml;ses der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Gebl&auml;se und Leiterplattenschnittstelle der W&auml;rmepumpe reinigen oder austauschen"                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1552"} = {errorcode => "H23",                                    subcode => "1063",                                               class => "18",          reason => "Zu viele Betriebsabbr&uuml;che ausgel&ouml;st durch Probleme in der Luftzufuhr",                                                                                                                                                                                                                                                                   checklist => "St&ouml;rung der Luftzufuhr der W&auml;rmepumpe",                                                                                                                                                                                                                                                                                                                                                                                                                                                     correction => "Gebl&auml;se und/oder Luftzuf&uuml;hrung reinigen"                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1553"} = {errorcode => "H24",                                    subcode => "1064",                                               class => "14",          reason => "W&auml;rmepumpen-Temperaturf&uuml;hler defekt",                                                                                                                                                                                                                                                                                                    checklist => "Temperaturf&uuml;hler der W&auml;rmepumpe defekt oder nicht angeschlossen",                                                                                                                                                                                                                                                                                                                                                                                                                           correction => "Temperaturf&uuml;hler anschlie&szlig;en oder austauschen"                                                                                                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"1554"} = {errorcode => "H25",                                    subcode => "1065",                                               class => "14",          reason => "Wasserdruckf&uuml;hler defekt oder nicht angeschlossen",                                                                                                                                                                                                                                                                                           checklist => "Wasserdruckf&uuml;hler defekt oder nicht angeschlossen",                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Wasserdruckf&uuml;hler anschlie&szlig;en oder austauschen"                                                                                                                                                                                                                                                                                                                                       };
	$ErrorCodeList{"1555"} = {errorcode => "H26",                                    subcode => "1066",                                               class => "20",          reason => "Verbrennung (&Ouml;l-LuftVerh&auml;ltnis) nicht optimal",                                                                                                                                                                                                                                                                                          checklist => "Pr&uuml;fen, ob korrekte Brennstoffversorgung sichergestellt ist (&Ouml;ltank, &Ouml;lfilter, &Ouml;lpumpe, Einspritzventil)&bull;Die Lambdasonde misst Werte abweichend vom optimalen Betriebspunkt&bull;Luftmenge nicht optimal auf &Ouml;lmenge abgestimmt",                                                                                                                                                                                                                                       correction => "Bei St&ouml;rung in der Brennstoffversorgung diese beheben, z. B. durch Reinigung, ggf. Austausch &Ouml;lfilter, &Ouml;lpumpe, Einspritzventil&bull;Pr&uuml;fung Position Lambdasonde, ggf. reinigen oder austauschen&bull;Gebl&auml;se auf Verschmutzung pr&uuml;fen, ggf. reinigen oder austauschen"                                                                                           };
	$ErrorCodeList{"1556"} = {errorcode => "H27",                                    subcode => "1032",                                               class => "8",           reason => "Warmwassereinlauff&uuml;hler defekt.",                                                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1557"} = {errorcode => "H28",                                    subcode => "1067",                                               class => "8",           reason => "W&auml;rmeblock verschmutzt oder Lambdasonde gedriftet",                                                                                                                                                                                                                                                                                           checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1558"} = {errorcode => "H29",                                    subcode => "1068",                                               class => "8",           reason => "Aussentemperaturf&uuml;hler oder Lambdasonde defekt.",                                                                                                                                                                                                                                                                                             checklist => "-",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   correction => "-"                                                                                                                                                                                                                                                                                                                                                                                               };
	$ErrorCodeList{"1559"} = {errorcode => "H3",                                     subcode => "73",                                                 class => "",            reason => "Mischer defekt",                                                                                                                                                                                                                                                                                                                                   checklist => "Mischertemperaturf&uuml;hler hat keinen Kontakt zur Rohrleitung&bull;Anschlusssstecker am Motor des Mischventiles ist nicht aufgesteckt&bull;Anschlusskabel zum Motor des Mischventiles ist besch&auml;digt&bull;Mischerventil mechanisch blockiert&bull;Motor des Mischventiles ist defekt&bull;interner Fehler auf der Leiterplatte f&uuml;r die Steuerung des Mischventiles",                                                                                                                      correction => "Mischertemperaturf&uuml;hler auf die Rohrleitung aufklipsen&bull;Stecker aufstecken&bull;Kabelbaum austauschen&bull;Mischventil austauschen&bull;Motor des Mischventiles austauschen&bull;Leiterplatte tauschen"                                                                                                                                                                                 };
	$ErrorCodeList{"1560"} = {errorcode => "H30",                                    subcode => "1067",                                               class => "8",           reason => "W&auml;rmeblock verschmutzt oder Lamdasonde gedriftet",                                                                                                                                                                                                                                                                                            checklist => "Lamdasondenersatzbetrieb &uuml;berpr&uuml;fen&bull;Silikondichtung auf Besch&auml;digung &uuml;berpr&uuml;fen&bull;Dichtung am PT 100 &uuml;berpr&uuml;fen&bull;Dichtung an der Luftklappe &uuml;berpr&uuml;fent",                                                                                                                                                                                                                                                                                    correction => "Lamdasonde austauschen&bull;Silikondichtung austauschen&bull;Temperaturf&uuml;hler PT 1000 austauschen&bull;Dichtung an der Luftklappe austauschen"                                                                                                                                                                                                                                              };
	$ErrorCodeList{"1561"} = {errorcode => "H30",                                    subcode => "1068",                                               class => "8",           reason => "Au&szlig;entemperaturf&uuml;hler oder Lambdasonde defekt",                                                                                                                                                                                                                                                                                         checklist => "Anschlusskontakte an der Lambdasonde &uuml;berpr&uuml;fen&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                     correction => "Kontaktproblem beseitigen&bull;Lamdasonde austauschen"                                                                                                                                                                                                                                                                                                                                           };
	$ErrorCodeList{"1562"} = {errorcode => "H30",                                    subcode => "1069",                                               class => "8",           reason => "Flammenabriss im Betrieb",                                                                                                                                                                                                                                                                                                                         checklist => "&Ouml;ltank leer&bull;Luft in der &Ouml;lleitung&bull;Ablagerungen auf der Ionisatonselektrode&bull;Ionisationselektrode verformt&bull;&Ouml;lpumpendruck zu niedrig&bull;Abgasseitige Ablagerungen am Kesselblock&bull;Gebl&auml;se verschmutzt",                                                                                                                                                                                                                                                    correction => "&Ouml;ltank bef&uuml;llen&bull;&Ouml;lleitung entl&uuml;ften&bull;Ionisationselektrode reinigen bzw. austauschen&bull;Ionisationselektrode austauschen&bull;&Ouml;lpumpendruck korrekt einstellen&bull;Abgaswege am Kesselblock reinigen&bull;Gebl&auml;se reinigen"                                                                                                                             };
	$ErrorCodeList{"1563"} = {errorcode => "H5",                                     subcode => "76",                                                 class => "20",          reason => "Geregelte Pumpe wurde nicht erkannt",                                                                                                                                                                                                                                                                                                              checklist => "Stecker f&uuml;r die geregelte Heizungsumw&auml;lzpume nicht aufgesteckt&bull;Signalunterbrechung im Kabel f&uuml;r das Steuersignal der drehzahlgeregelten Umw&auml;lzpumpe&bull;Drehzahlregelung in der Umw&auml;lzpumpe defekt",                                                                                                                                                                                                                                                                   correction => "Stecker aufstecken&bull;Kabelbaum austauschen&bull;Pumpe austauschen"                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1564"} = {errorcode => "HP",                                     subcode => "91",                                                 class => "4",           reason => "St&ouml;rung Au&szlig;eneinheit/ Wasserdurchfluss gest&ouml;rt",                                                                                                                                                                                                                                                                                   checklist => "St&ouml;rung an der W&auml;rmepumpe/ Wasserdurchfluss gest&ouml;rt&bull;LED-Signal am Hybridmodul pr&uuml;fen&bull;Durchfluss der Luft-Wasser-W&auml;rmepumpe pr&uuml;fen; Filter pr&uuml;fen&bull;Str&ouml;mungsschalter pr&uuml;fen&bull;System pr&uuml;fen",                                                                                                                                                                                                                                       correction => "-&bull;Filter reinigen ggf. austauschen&bull;Ggf. austauschen&bull;-"                                                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1565"} = {errorcode => "HP",                                     subcode => "479",                                                class => "4",           reason => "St&ouml;rung Str&ouml;mungsschalter W&auml;rmepumpe w&auml;hrend Selbsttest",                                                                                                                                                                                                                                                                      checklist => "Str&ouml;mungsschalter pr&uuml;fen",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Ggf. Str&ouml;mungsschalter austauschen"                                                                                                                                                                                                                                                                                                                                                         };
	$ErrorCodeList{"1566"} = {errorcode => "L1",                                     subcode => "2004; 2013",                                         class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1567"} = {errorcode => "L1",                                     subcode => "",                                                   class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1568"} = {errorcode => "L1",                                     subcode => "2017",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1569"} = {errorcode => "L1",                                     subcode => "2018; 2019",                                         class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1570"} = {errorcode => "L1",                                     subcode => "",                                                   class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1571"} = {errorcode => "L1",                                     subcode => "2020",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1572"} = {errorcode => "L1",                                     subcode => "2021",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1573"} = {errorcode => "L1",                                     subcode => "2022",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1574"} = {errorcode => "L1",                                     subcode => "2024",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1575"} = {errorcode => "L1",                                     subcode => "2025",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1576"} = {errorcode => "L1",                                     subcode => "2026",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1577"} = {errorcode => "L1",                                     subcode => "2027",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1578"} = {errorcode => "L1",                                     subcode => "2028",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1579"} = {errorcode => "L1",                                     subcode => "2029",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1580"} = {errorcode => "L1",                                     subcode => "2030",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1581"} = {errorcode => "L1",                                     subcode => "2031",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1582"} = {errorcode => "L1",                                     subcode => "2032",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1583"} = {errorcode => "L1",                                     subcode => "2033",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1584"} = {errorcode => "L1",                                     subcode => "2334",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1585"} = {errorcode => "L1",                                     subcode => "2040",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1586"} = {errorcode => "L1",                                     subcode => "2044",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1587"} = {errorcode => "L1",                                     subcode => "2048",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1588"} = {errorcode => "L1",                                     subcode => "2049",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1589"} = {errorcode => "L1",                                     subcode => "2054",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1590"} = {errorcode => "L1",                                     subcode => "2058",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1591"} = {errorcode => "L1",                                     subcode => "2059",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1592"} = {errorcode => "L1",                                     subcode => "2060",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1593"} = {errorcode => "L1",                                     subcode => "2061",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1594"} = {errorcode => "L1",                                     subcode => "2062",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1595"} = {errorcode => "L1",                                     subcode => "2063",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1596"} = {errorcode => "L1",                                     subcode => "2064",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1597"} = {errorcode => "L1",                                     subcode => "2065",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1598"} = {errorcode => "L1",                                     subcode => "2066",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1599"} = {errorcode => "L1",                                     subcode => "2068",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1600"} = {errorcode => "L1",                                     subcode => "2069",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1601"} = {errorcode => "L1",                                     subcode => "2070",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1602"} = {errorcode => "L1",                                     subcode => "2071",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1603"} = {errorcode => "L1",                                     subcode => "2072",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1604"} = {errorcode => "L1",                                     subcode => "2073",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1605"} = {errorcode => "L1",                                     subcode => "2074",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1606"} = {errorcode => "L1",                                     subcode => "2075",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1607"} = {errorcode => "L1",                                     subcode => "2076",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1608"} = {errorcode => "L1",                                     subcode => "2077",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1609"} = {errorcode => "L1",                                     subcode => "2078",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1610"} = {errorcode => "L1",                                     subcode => "2079",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1611"} = {errorcode => "L1",                                     subcode => "2080",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1612"} = {errorcode => "L1",                                     subcode => "2081",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1613"} = {errorcode => "L1",                                     subcode => "2082",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1614"} = {errorcode => "L1",                                     subcode => "2084",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1615"} = {errorcode => "L1",                                     subcode => "2085",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1616"} = {errorcode => "L1",                                     subcode => "2087",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1617"} = {errorcode => "L1",                                     subcode => "2088",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1618"} = {errorcode => "L1",                                     subcode => "2092",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1619"} = {errorcode => "L1",                                     subcode => "2096",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1620"} = {errorcode => "L1",                                     subcode => "2097",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1621"} = {errorcode => "L1",                                     subcode => "2098",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1622"} = {errorcode => "L1",                                     subcode => "2099",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1623"} = {errorcode => "L1",                                     subcode => "2101",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1624"} = {errorcode => "L1",                                     subcode => "2104",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1625"} = {errorcode => "L1",                                     subcode => "2108",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1626"} = {errorcode => "L1",                                     subcode => "2111",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1627"} = {errorcode => "L1",                                     subcode => "2116",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten",                                                                                                                                                                                                                                                                                                                                                                                                                                                                  correction => "Entriegeln, Anlage f&uuml;r 30 Sekunden spannungsfrei,  schalten, Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                                  };
	$ErrorCodeList{"1628"} = {errorcode => "LL",                                     subcode => "571",                                                class => "4",           reason => "Zu viele Wiederanl&auml;ufe trotz Entriegelung; (Es traten direkt hintereinander 15 Wiederanl&auml;ufe auf.; D. h.: Nach dem Entriegeln war immer noch das gleiche Problem am Brenner. ACHTUNG: Dieser Fehler l&auml;sst sich nur &uuml;ber die Taste am SAFe entriegeln oder beim einem BRM10 kann dies Problem nur durch AUSwieder EINschalten behoben werden)", checklist => "Es wurden laufend vorhandene St&ouml;rungen nur entriegelt und nicht behoben&bull;Ger&auml;teelektronik SAFe defekt",                                                                                                                                                                                                                                                                                                                                                                                 correction => "St&ouml;rungen, die zu den Entriegelungen gef&uuml;hrt haben, suchen und beheben&bull;Ger&auml;teelektronik SAFe austauschen"                                                                                                                                                                                                                                                                    };
	$ErrorCodeList{"1629"} = {errorcode => "LP",                                     subcode => "570",                                                class => "4",           reason => "Werden innerhalb einer bestimmten Zeit zu viele Entriegelungen uber die Schnittstelle empfangen, wird diese Fehlermeldung erzeugt. Achtung: Diese St&ouml;rung l&auml;sst sich nur &uuml;ber die Taste am SAFe entriegeln. Beim BRM10 kann dieses Problem nur durch Ausund wieder Einschalten behoben werden!",                                    checklist => "Es wurden laufend vorhandene St&ouml;rungen nur entriegelt und nicht behoben&bull;Es liegt eine Fehlfunktion des Basiscontrollers vor, wodurch st&auml;ndig entriegelt wird&bull;Es liegt eine Fehlfunktion in der Ger&auml;teelektronik SAFe vor",                                                                                                                                                                                                                                                   correction => "St&ouml;rungen, die zu den Entriegelungen gef&uuml;hrt haben, suchen und beheben&bull;Basiscontroller BC10 austauschen&bull;Ger&auml;teelektronik SAFe austauschen"                                                                                                                                                                                                                              };
	$ErrorCodeList{"1630"} = {errorcode => "L2",                                     subcode => "2067",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Identifikationsmodul (BCI)&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1631"} = {errorcode => "L2",                                     subcode => "2086",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Identifikationsmodul (BCI)&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1632"} = {errorcode => "L2",                                     subcode => "2089",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Identifikationsmodul (BCI)&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1633"} = {errorcode => "L2",                                     subcode => "2093",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Identifikationsmodul (BCI)&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1634"} = {errorcode => "L2",                                     subcode => "2094",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Identifikationsmodul (BCI)&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1635"} = {errorcode => "L2",                                     subcode => "2095",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Identifikationsmodul (BCI)&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1636"} = {errorcode => "L2",                                     subcode => "2102",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Identifikationsmodul (BCI)&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1637"} = {errorcode => "L2",                                     subcode => "2103",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Identifikationsmodul (BCI)&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1638"} = {errorcode => "L2",                                     subcode => "2105",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Identifikationsmodul (BCI)&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1639"} = {errorcode => "L2",                                     subcode => "2106",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Identifikationsmodul (BCI)&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1640"} = {errorcode => "L2",                                     subcode => "2107",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Identifikationsmodul (BCI)&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1641"} = {errorcode => "L2",                                     subcode => "2109",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Identifikationsmodul (BCI)&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1642"} = {errorcode => "L2",                                     subcode => "2110",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Identifikationsmodul (BCI)&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1643"} = {errorcode => "L2",                                     subcode => "2115",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Identifikationsmodul (BCI)&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1644"} = {errorcode => "L2",                                     subcode => "2117",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Identifikationsmodul (BCI)&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1645"} = {errorcode => "L2",                                     subcode => "2118",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Identifikationsmodul (BCI)&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1646"} = {errorcode => "L2",                                     subcode => "2119",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Identifikationsmodul (BCI)&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1647"} = {errorcode => "L2",                                     subcode => "2121",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Identifikationsmodul (BCI)&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1648"} = {errorcode => "L2",                                     subcode => "2122",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Identifikationsmodul (BCI)&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1649"} = {errorcode => "L2",                                     subcode => "2123",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Identifikationsmodul (BCI)&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1650"} = {errorcode => "L2",                                     subcode => "2124",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Identifikationsmodul (BCI)&bull;&bull;",                                                                                                                                                                                                                                                                                                                                                                                                                                              correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen"                                                                                                                                                                                                                                                                                            };
	$ErrorCodeList{"1651"} = {errorcode => "L3",                                     subcode => "2007",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten&bull;&bull;&bull;St&ouml;rung im Identifikationsmodul (BCI)",                                                                                                                                                                                                                                                                                                                                                                                                      correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen&bull;Kundendienst verst&auml;ndigen"                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"1652"} = {errorcode => "L3",                                     subcode => "2008",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten&bull;&bull;&bull;St&ouml;rung im Identifikationsmodul (BCI)",                                                                                                                                                                                                                                                                                                                                                                                                      correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen&bull;Kundendienst verst&auml;ndigen"                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"1653"} = {errorcode => "L3",                                     subcode => "2010",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten&bull;&bull;&bull;St&ouml;rung im Identifikationsmodul (BCI)",                                                                                                                                                                                                                                                                                                                                                                                                      correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen&bull;Kundendienst verst&auml;ndigen"                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"1654"} = {errorcode => "L3",                                     subcode => "2011",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten&bull;&bull;&bull;St&ouml;rung im Identifikationsmodul (BCI)",                                                                                                                                                                                                                                                                                                                                                                                                      correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen&bull;Kundendienst verst&auml;ndigen"                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"1655"} = {errorcode => "L3",                                     subcode => "2012",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten&bull;&bull;&bull;St&ouml;rung im Identifikationsmodul (BCI)",                                                                                                                                                                                                                                                                                                                                                                                                      correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen&bull;Kundendienst verst&auml;ndigen"                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"1656"} = {errorcode => "L3",                                     subcode => "2014",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten&bull;&bull;&bull;St&ouml;rung im Identifikationsmodul (BCI)",                                                                                                                                                                                                                                                                                                                                                                                                      correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen&bull;Kundendienst verst&auml;ndigen"                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"1657"} = {errorcode => "L3",                                     subcode => "2015",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten&bull;&bull;&bull;St&ouml;rung im Identifikationsmodul (BCI)",                                                                                                                                                                                                                                                                                                                                                                                                      correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen&bull;Kundendienst verst&auml;ndigen"                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"1658"} = {errorcode => "L3",                                     subcode => "2016",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten&bull;&bull;&bull;St&ouml;rung im Identifikationsmodul (BCI)",                                                                                                                                                                                                                                                                                                                                                                                                      correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen&bull;Kundendienst verst&auml;ndigen"                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"1659"} = {errorcode => "L3",                                     subcode => "2045",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten&bull;&bull;&bull;St&ouml;rung im Identifikationsmodul (BCI)",                                                                                                                                                                                                                                                                                                                                                                                                      correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen&bull;Kundendienst verst&auml;ndigen"                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"1660"} = {errorcode => "L3",                                     subcode => "2047",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten&bull;&bull;&bull;St&ouml;rung im Identifikationsmodul (BCI)",                                                                                                                                                                                                                                                                                                                                                                                                      correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen&bull;Kundendienst verst&auml;ndigen"                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"1661"} = {errorcode => "L3",                                     subcode => "2053",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten&bull;&bull;&bull;St&ouml;rung im Identifikationsmodul (BCI)",                                                                                                                                                                                                                                                                                                                                                                                                      correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen&bull;Kundendienst verst&auml;ndigen"                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"1662"} = {errorcode => "L3",                                     subcode => "2055",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten&bull;&bull;&bull;St&ouml;rung im Identifikationsmodul (BCI)",                                                                                                                                                                                                                                                                                                                                                                                                      correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen&bull;Kundendienst verst&auml;ndigen"                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"1663"} = {errorcode => "L3",                                     subcode => "2056",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten&bull;&bull;&bull;St&ouml;rung im Identifikationsmodul (BCI)",                                                                                                                                                                                                                                                                                                                                                                                                      correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen&bull;Kundendienst verst&auml;ndigen"                                                                                                                                                                                                                                                        };
	$ErrorCodeList{"1664"} = {errorcode => "L3",                                     subcode => "2057",                                               class => "4",           reason => "Interne St&ouml;rung",                                                                                                                                                                                                                                                                                                                             checklist => "St&ouml;rung im Feuerungsautomaten&bull;&bull;&bull;St&ouml;rung im Identifikationsmodul (BCI)",                                                                                                                                                                                                                                                                                                                                                                                                      correction => "Entriegeln&bull;Anlage f&uuml;r 30 Sekunden spannungsfrei schalten&bull;Feuerungsautomat austauschen&bull;Kundendienst verst&auml;ndigen"                                                                                                                                                                                                                                                        };

	### Define database for error class as hash
	$ClassCodeList{"1"}    = {description => "Schwerwiegende Störung – verriegelnd (interner Defekt)",                                                                    workaround => "&bull; Betroffenes Produkt außer Betrieb (verriegelnde Störung)\n&bull;Produkt muss ausgetauscht werden\n&bull;Beispiel: EEPROM oder interner ROM defekt"                                                                                                                                  };
	$ClassCodeList{"2"}    = {description => "Schwerwiegende Störung – verriegelnd (Sicherheitsabschaltung, Reset durch Installateur)",                                   workaround => "&bull; Betroffenes Produkt außer Betrieb (verriegelnde Störung)\n&bull; Reset durch Kunden nicht möglich\n&bull; Spezieller Reset erforderlich"                                                                                                                                            };
	$ClassCodeList{"4"}    = {description => "Schwerwiegende Störung – verriegelnd (Sicherheitsabschaltung, manueller Reset durch Bediener)",                             workaround => "&bull; Betroffenes Produkt außer Betrieb (verriegelnde Störung)\n&bull; Manueller Reset erforderlich\n&bull; Beispiel: Sicherheitstemperaturfühler defekt"                                                                                                                                 };
	$ClassCodeList{"6"}    = {description => "Schwerwiegende Störung – verriegelnd (Sicherheitsabschaltung, manueller Reset oder Neustart durch Spannungsunterbrechung)", workaround => "&bull; Betroffenes Produkt außer Betrieb (verriegelnde Störung)\n&bull; Manueller Reset oder Neustart erforderlich"                                                                                                                                                                        };
	$ClassCodeList{"8"}    = {description => "Schwerwiegende Störung – blockierend (Sicherheitsabschaltung mit automatischem Neustart)",                                  workaround => "&bull; Betroffenes Produkt zeitweise außer Betrieb (blockierende Störung)\n&bull; Manueller Reset nicht möglich\n&bull; Störungsanzeige wird nach einiger Zeit aufgehoben, tritt jedoch häufig erneut auf\n&bull; Beispiel: Flamme erlischt"                                               };
	$ClassCodeList{"10"}   = {description => "Schwerwiegende Störung – blockierend (Sicherheitsabschaltung)",                                                             workaround => "&bull; Betroffenes Produkt außer Betrieb (blockierende Störung)\n&bull; Manueller Reset nicht möglich\n&bull; Störungsanzeige wird durch Beheben der Ursache aufgehoben\n&bull; Beispiel: Überhitzung"                                                                                     };
	$ClassCodeList{"12"}   = {description => "Schwerwiegende Störung – Anlagenstörung (kontrollierte Abschaltung)",                                                       workaround => "&bull; Betroffenes Produkt außer Betrieb (blockierende Störung)\n&bull; Manueller Reset nicht möglich\n&bull; Störungsanzeige wird durch Beheben der Ursache aufgehoben\n&bull; Beispiel: blockierte Pumpe, primärer Warmwasser-Temperaturfühler defekt usw."                              };
	$ClassCodeList{"14"}   = {description => "Störung – Anlagenstörung",                                                                                                  workaround => "&bull; Betroffenes Produkt bleibt mit eingeschränkter Funktion in Betrieb\n&bull;  Manueller Reset nicht möglich\n&bull;  Störungsanzeige wird durch Beheben der Ursache aufgehoben\n&bull;  Beispiel: Außentemperaturfühler defekt"                                                       };
	$ClassCodeList{"16"}   = {description => "Geringfügige Störung – Anlagenstörung",                                                                                     workaround => "&bull; Betroffenes Produkt bleibt mit geringfügig eingeschränkter Funktion in Betrieb\n&bull;  Manueller Reset nicht möglich\n&bull;  Störungsanzeige wird durch Beheben der Ursache aufgehoben\n&bull;  Beispiel: sekundärer Warmwasser-Temperaturfühler defekt"                          };
	$ClassCodeList{"18"}   = {description => "Serviceanzeige – Wartung (Reset möglich)",                                                                                  workaround => "&bull; Betroffenes Produkt uneingeschränkt in Betrieb\n&bull;  Service erforderlich, um einer Störung vorzubeugen\n&bull;  Manueller Reset nach Service erforderlich\n&bull;  Beispiel: Betriebsstunden des Brenners bis zum nächsten Service überschritten"                               };
	$ClassCodeList{"20"}   = {description => "Serviceanzeige – Wartung (kein Reset möglich)",                                                                             workaround => "&bull; Betroffenes Produkt uneingeschränkt in Betrieb\n&bull;  Service erforderlich, um einer Störung vorzubeugen\n&bull;  Manueller Reset nicht möglich\n&bull;  Störungsanzeige wird durch Beheben der Ursache aufgehoben\n&bull;  Beispiel: Betriebsdruck der Heizungsanlage zu niedrig"};
	$ClassCodeList{"BC"}   = {description => "Betriebsmeldung, keine Störung",                                                                                            workaround => "&bull; Betriebsmeldung, keine Störung\n&bull;  Wird nicht an der Bedieneinheit angezeigt\n&bull;  Beispiel: nur EMS-Wärmeerzeuger in Betrieb"                                                                                                                                              };
	$ClassCodeList{"A"}    = {description => "-",                                                                                                                         workaround => "&bull; Systemfehler"                                                                                                                                                                                                                                                                       };
	$ClassCodeList{"B"}    = {description => "-",                                                                                                                         workaround => "&bull; Blockierend"                                                                                                                                                                                                                                                                        };



	### For debugging purpose only
	Log3 $name, 5, $name. " : km200_GettErrorMessage  - ErrorCode                : " . $ErrorCode;
	Log3 $name, 5, $name. " : km200_GettErrorMessage  - SubCode                  : " . $SubCode;
	Log3 $name, 5, $name. " : km200_GettErrorMessage  - Class                    : " . $Class;
	
	my @ErrorCodeMatches;
	my $SubCodeMatches;
	
	### Find ErrorCodes
	@ErrorCodeMatches = grep { $ErrorCodeList{$_}{errorcode} =~ /$ErrorCode/ } keys %ErrorCodeList;

	### For debugging purpose only
	Log3 $name, 5, $name. " : km200_GettErrorMessage  - ErrorCodeMatches in      : " . @ErrorCodeMatches . " entries.";
	
	### Search within the matching errorcodes for matching subcode
	foreach my $ErrorCodeId ( @ErrorCodeMatches ) {
		### If SubCode has been found, save ID
		if ($ErrorCodeList{$ErrorCodeId}{subcode} =~ /$SubCode/) {
			$SubCodeMatches = $ErrorCodeId;
		}
	}

	### For debugging purpose only
	Log3 $name, 5, $name. " : km200_GettErrorMessage  - Found Error- and Subcode : " . $SubCodeMatches;	

	### Formating the output
	my $ClassDescription;
	my $ClassWorkaround;
	my $ReasonText;
	my $ChecklistText;
	my $CorrectionText;

	### If the maximum length of Reading has been defined
	if ($hash->{READTEXTLEN} > 0) {
		### Format text
		$Text::Wrap::columns = $hash->{READTEXTLEN};
		$Text::Wrap::break   = '[\s+]';
		$Text::Wrap::huge    = 'wrap';
		$ClassDescription    = wrap("","", $ClassCodeList{$Class}{description}        ); 
		$ClassWorkaround     = wrap("","", $ClassCodeList{$Class}{workaround}         ); 
		$ReasonText          = wrap("","", $ErrorCodeList{$SubCodeMatches}{reason}    ); 
		$ChecklistText       = wrap("","", $ErrorCodeList{$SubCodeMatches}{checklist} ); 
		$CorrectionText      = wrap("","", $ErrorCodeList{$SubCodeMatches}{correction}); 
	}
	### If maximum of reading is 0 and therefore disabled
	else {
		$ClassDescription    = $ClassCodeList{$Class}{description}        ; 
		$ClassWorkaround     = $ClassCodeList{$Class}{workaround}         ; 
		$ReasonText          = $ErrorCodeList{$SubCodeMatches}{reason}    ; 
		$ChecklistText       = $ErrorCodeList{$SubCodeMatches}{checklist} ; 
		$CorrectionText      = $ErrorCodeList{$SubCodeMatches}{correction}; 
	}
	
	return($ClassDescription, $ClassWorkaround, $ReasonText, $ChecklistText, $CorrectionText);
}

####END####### Load database for errorcodes in hash ############################################################END#####
1;

###START###### Description for fhem commandref ################################################################START####
=pod
=item device
=item summary    Connects fhem to Buderus KM300, KM200, KM100, KM50
=item summary_DE Verbindet fhem mit Buderus KM300, KM200, KM100, KM50

=begin html

<a name="km200"></a>
<h3>KM200</h3>
<ul>
	<table>
		<tr>
			<td>
				The Buderus <a href="https://www.buderus.de/de/produkte/catalogue/alle-produkte/7719_Gateway-Logamatic-web-KM200-KM100-KM50">KM200, KM100  or KM50 (hereafter described as KMxxx)</a> is a communication device to establish a connection between the Buderus central heating control unit and the internet.<BR>
				It has been designed in order to allow the inhabitants accessing their heating system via his Buderus App <a href="http://www.buderus.de/Online_Anwendungen/Apps/fuer_den_Endverbrauch
				er/EasyControl-4848514.html"> EasyControl</a>.<BR>
				Furthermore it allows the maintenance companies to access the central heating control system to read and change settings.<BR>
				The km200 fhem-module enables read/write access to these parameters.<BR>
				<BR>
				In order to use the KMxxx with fhem, you must define the private password with the Buderus App <a href="http://www.buderus.de/Online_Anwendungen/Apps/fuer_den_Endverbraucher/EasyControl-4848514.html"> EasyControl</a> first.<BR>
				<BR>
				<b><u>Remark:</u></b><BR>
				Despite the instruction of the Buderus KMxxx Installation guide, the ports 5222 and 5223 should not be opened and allow access to the KMxxx module from outside.<BR>
				You should configure (or leave) your internet router with the respective settings.<BR>
				If you want to read or change settings on the heating system, you should access the central heating control system via your fhem system only.<BR>
				<BR>
				As soon the module has been defined within the fhem.cfg, the module is trying to obtain all known/possible services. <BR>
				After this initial contact, the module differs between a set of continuous (dynamically) changing values (e.g.: temperatures) and not changing static values (e.g.: Firmware version).<BR>
				This two different set of values can be bound to an individual polling interval. Refer to <a href="#KM200Attr">Attributes</a><BR> 
				<BR>
			</td>
		</tr>
	</table>
	<BR>  
	<table>
		<tr><td><a name="KM200define"></a><b>Define</b></td></tr>
		<tr><td><ul><code>define &lt;name&gt; km200 &lt;IPv4-address&gt; &lt;GatewayPassword&gt; &lt;PrivatePassword&gt;</code></ul></td></tr>
		<tr><td><ul><ul><code>&lt;name&gt;</code> : </td><td>The name of the device. Recommendation: "myKm200".</ul></ul></td></tr>
		<tr><td><ul><ul><code>&lt;IPv4-address&gt;</code> : </td><td>A valid IPv4 address of the KMxxx. You might look into your router which DHCP address has been given to the KMxxx.</ul></ul></td></tr>
		<tr><td><ul><ul><code>&lt;GatewayPassword&gt;</code> : </td><td>The gateway password which is provided on the type sign of the KMxxx.</ul></ul></td></tr>
		<tr><td><ul><ul><code>&lt;PrivatePassword&gt;</code> : </td><td>The private password which has been defined by the user via <a href="http://www.buderus.de/Online_Anwendungen/Apps/fuer_den_Endverbraucher/EasyControl-4848514.html"> EasyControl</a>.</ul></ul></td></tr>
	</table>
	<BR>

	<table>
		<tr><td><a name="KM200Set"></a><b>Set</b></td></tr>
		<tr><td><ul>The set function is able to change a value of a service which has the "writeable" - tag within the KMxxx service structure.<BR>Most of those values have an additional list of allowed values which are the only ones to be set.<BR>Other floatable type values can be changed only within their range of minimum and maximum value.<BR></ul></td></tr>
		<tr><td><ul><code>set &lt;service&gt; &lt;value&gt;</code></ul></td></tr>
		<tr><td><ul><ul><code>&lt;service&gt;</code> : </td><td>The name of the service which value shall be set. E.g.: "<code>/heatingCircuits/hc1/operationMode</code>"<BR></ul></ul></td></tr>
		<tr><td><ul><ul><code>&lt;value&gt;</code> : </td><td>A valid value for this service.<BR></ul></ul></td></tr>
	</table>
	<BR>
	<table>
		<tr><td><a name="KM200Get"></a><b>Get</b></td></tr>
		<tr><td><ul>          The get function is able to obtain a value of a service within the KMxxx service structure.<BR>The additional list of allowed values or their range of minimum and maximum value will not be handed back.<BR></ul></td></tr>
		<tr><td><ul>    <code>get &lt;service&gt; &lt;option&gt;</code></ul></td></tr>
		<tr><td><ul><ul><code>&lt;service&gt;</code> : </td><td>The name of the service which value shall be obtained. E.g.: "<code>/heatingCircuits/hc1/operationMode</code>"<BR>&nbsp;&nbsp;It returns only the value but not the unit or the range or list of allowed values possible.<BR></ul></ul></td></tr>
		<tr><td><ul><ul><code>&lt;option&gt;</code> : </td><td>The optional Argument for the result of the get-command e.g.:  "<code>json</code>"<BR>&nbsp;&nbsp;The following options are available:<BR>&nbsp;&nbsp;json - Returns the raw json-answer from the KMxxx as string.<BR></ul></ul></td></tr>
	</table>
	<BR>

	<table>
		<tr><td><a name="KM200Attr"></a><b>Attributes</b></td></tr>
		<tr><td><ul>The following user attributes can be used with the km200 module in addition to the global ones e.g. <a href="#room">room</a>.<BR></ul></td></tr>
	</table>
	<table>
		<tr><td><ul><ul><a name="IntervalDynVal" ></a><li><b><u><code>IntervalDynVal </code></u></b> : A valid polling interval for the dynamically changing values of the KMxxx. The value must be >=20s to allow the km200 module to perform a full polling procedure. <BR>The default value is 300s.                                                                                                                             <BR></li></ul></ul></td></tr>
		<tr><td><ul><ul><a name="PollingTimeout" ></a><li><b><u><code>PollingTimeout </code></u></b> : A valid time in order to allow the module to wait for a response of the KMxxx. Usually this value does not need to be changed but might in case of slow network or slow response.<BR>The default and minimum value is 5s.                                                                                                    <BR></li></ul></ul></td></tr>
		<tr><td><ul><ul><a name="DoNotPoll"      ></a><li><b><u><code>DoNotPoll      </code></u></b> : A list of services separated by blanks which shall not be downloaded due to repeatable crashes or irrelevant values.<BR>The list can be filled with the name of the top - hierarchy service, which means everything below that service will also be ignored.<BR>The default value (empty) therefore nothing will be ignored. <BR></li></ul></ul></td></tr>
		<tr><td><ul><ul><a name="ReadBackDelay"  ></a><li><b><u><code>ReadBackDelay  </code></u></b> : A valid time in milliseconds [ms] for the delay between writing and re-reading of values after using the "set" - command. The value must be >=0ms.<BR>The default value is 100 = 100ms = 0,1s.                                                                                                                               <BR></li></ul></ul></td></tr>
		<tr><td><ul><ul><a name="ReadingTextLen  ></a><li><b><u><code>ReadingTextLen </code></u></b> : A valid positive number of maximum text length which the error code readings shall be displayed<BR>The default value is 0 = disabled = unlimited text length.                                                                                                                                                                <BR></li></ul></ul></td></tr>
		<tr><td><ul><ul><a name="disable"        ></a><li><b><u><code>disable        </code></u></b> : Stops the device from further pollings and deletes the existing readings.<BR>The default value is 0 = activated<BR>                                                                                                                                                                                                          <BR></li></ul></ul></td></tr>
	</table>
</ul>
=end html

=begin html_DE

<a name="km200"></a>
<h3>KM200</h3>
<ul>
	<table>
		<tr>
			<td>
				Das Buderus <a href="https://www.buderus.de/de/produkte/catalogue/alle-produkte/7719_Gateway-Logamatic-web-KM200-KM100-KM50">KM200, KM100  or KM50 (ab hier als KMxxx beschrieben)</a> ist eine Schnittstelle zwischen der Buderus Zentralheizungssteuerung un dem Internet.<BR>
				Es wurde entwickelt um den Bewohnern den Zugang zu Ihrem Heizungssystem durch die Buderus App <a href="http://www.buderus.de/Online_Anwendungen/Apps/fuer_den_Endverbraucher/EasyControl-4848514.html"> EasyControl zu erlauben.</a>.<BR>
				Dar&uuml;ber hinaus erlaubt es nach vorheriger Freigabe dem Heizungs- bzw. Wartungsbetrieb die Heizungsanlage von aussen zu warten und Werte zu ver&auml;ndern.<BR>
				Das km200 fhem-Modul erlaubt den Lese-/Schreibzugriff dieser Parameter durch fhem.<BR>
				<BR>
				Um das KMxxx Ger&auml;t mit fhem nutzen zu k&ouml;nnen, mu&szlig; zun&auml;chst ein privates Passwort mit der Buderus Buderus App <a href="http://www.buderus.de/Online_Anwendungen/Apps/fuer_den_Endverbraucher/EasyControl-4848514.html"> EasyControl</a> - App gesetzt werden.<BR>
				<BR>
				<b><u>Anmerkung:</u></b><BR>
				Unabh&auml;ngig der Installationsanleitung des Buderus KMxxx Ger&auml;ts, sollten die Ports 5222 und 5223 am Router geschlossen bleiben um keinen Zugriff von au&szlig;en auf das Ger&auml;t zu erlauben.<BR>
				Der Router sollte entsprechend Konfiguriert bzw. so belassen werden.<BR>
				Wenn der Lese-/Schreibzugriff von aussen gew&uuml;nscht ist, so sollte man ausschlie&szlig;lich &uuml;ber das fhem-System auf die Zentralheizung zugreifen.<BR>
				<BR>
				Sobald das Modul in der fhem.cfg definiert ist, wird das Modul versuchen alle bekannten Services abzuklopfen ob diese in der angeschlossenen Konstellation &uuml;berhaupt vorhanden sind.<BR>
				Nach diesem Initial-Kontakt unterscheidet das Modul zwisachen einem Satz an Services die sich st&auml;ndig (dynamisch) &auml;ndern (z.B.: Vorlauftemperatur) sowie sich nicht st&auml;ndig (statisch) &auml;ndernden Werten (z.B.: Firmware Version).<BR>
				Diese beiden S&auml;tze an Services k&ouml;nnen mir einem individuellen Abfrageintervall versehen werden. Siehe <a href="#KM200Attr">Attributes</a><BR> 
				<BR>
			</td>
		</tr>
	</table>
	<BR>  
	<table>
		<tr><td><a name="KM200define"></a><b>Define</b></td></tr>
		<tr><td><ul><code>define &lt;name&gt; km200 &lt;IPv4-address&gt; &lt;GatewayPassword&gt; &lt;PrivatePassword&gt;</code></ul></td></tr>
		<tr><td><ul><ul><code>&lt;name&gt;</code> : </td><td>Der Name des Ger&auml;tes. Empfehlung: "myKm200".</ul></ul></td></tr>
		<tr><td><ul><ul><code>&lt;IPv4-address&gt;</code> : </td><td>Eine g&uuml;ltige IPv4 Adresse des KM200. Eventuell im Router nachschauen welche DHCP - Addresse dem KM200/KM50 vergeben wurde.</ul></ul></td></tr>
		<tr><td><ul><ul><code>&lt;GatewayPassword&gt;</code> : </td><td>Das gateway Passwort, welches auf dem Typenschild des KM200/KM50 zu finden ist.</ul></ul></td></tr>
		<tr><td><ul><ul><code>&lt;PrivatePassword&gt;</code> : </td><td>Das private Passwort, welches durch den User mit Hilfe der <a href="http://www.buderus.de/Online_Anwendungen/Apps/fuer_den_Endverbraucher/EasyControl-4848514.html"> EasyControl</a> - App vergeben wurde.</ul></ul></td></tr>
	</table>
	<BR>
	<table>
		<tr><td><a name="KM200Set"></a><b>Set</b></td></tr>
		<tr><td><ul>Die set Funktion &auml;ndert die Werte der Services welche das Flag "schreibbar" innerhalb der KMxxx Service Struktur besitzen.<BR>Die meisten dieser beschreibbaren Werte haben eine exklusive Liste von m&ouml;glichen Werten innerhalb dessen sich der neue Wert bewegen muss.<BR>Andere Flie&szlig;komma Werte haben einen maximum und minumum Wert, in dessen sich der neue Wert bewegen mu&szlig;.<BR></ul></td></tr>
		<tr><td><ul>    <code>set &lt;service&gt; &lt;value&gt;</code></ul></td></tr>
		<tr><td><ul><ul><code>&lt;service&gt;</code> : </td><td>Der Name des Service welcher gesetzt werden soll. Z.B.: "<code>/heatingCircuits/hc1/operationMode</code>"<BR></ul></ul></td></tr>
		<tr><td><ul><ul><code>&lt;value&gt;</code> : </td><td>Ein g&uuml;ltiger Wert f&uuml;r diesen Service.<BR></ul></ul></td></tr>
	</table>
	<BR>
	<table>
		<tr><td><a name="KM200Get"></a><b>Get</b></td></tr>
		<tr><td><ul>Die get-Funktion ist in der Lage einen Wert eines Service innerhalb der KMxxx Service Struktur auszulesen.<BR>Die zus&auml;tzliche Liste von erlaubten Werten oder der Wertebereich zwischen Minimum und Maximum wird nicht zur&uuml;ck gegeben.<BR></ul></td></tr>
		<tr><td><ul><code>get &lt;service&gt; &lt;option&gt;</code></ul></td></tr>
		<tr><td><ul><ul><code>&lt;service&gt;</code> : </td><td>Der Name des Service welcher ausgelesen werden soll. Z.B.:  "<code>/heatingCircuits/hc1/operationMode</code>"<BR>&nbsp;&nbsp;Es gibt nur den Wert, aber nicht die Werteliste oder den m&ouml;glichen Wertebereich zur&uuml;ck.<BR></ul></ul></td></tr>
		<tr><td><ul><ul><code>&lt;option&gt;</code> : </td><td>Das optionelle Argument f&uuml;r Ausgabe des get-Befehls Z.B.:  "<code>json</code>"<BR>&nbsp;&nbsp;Folgende Optionen sind verf&uuml;gbar:<BR>&nbsp;&nbsp;json - Gibt anstelle des Wertes, die gesamte Json Antwort des KMxxx als String zur&uuml;ck<BR></ul></ul></td></tr>
	</table>
	<BR>
	<table>
		<tr><td><a name="KM200Attr"></a><b>Attributes</b></td></tr>
		<tr><td><ul>Die folgenden Modul-spezifischen Attribute k&ouml;nnen neben den bekannten globalen Attributen gesetzt werden wie z.B.: <a href="#room">room</a>.<BR></ul></td></tr>
	</table>
	<table>
		<tr><td><ul><ul><a name="IntervalDynVal" ></a><li><b><u><code>IntervalDynVal </code></u></b> : Ein g&uuml;ltiges Abfrageintervall f&uuml;r die sich st&auml;ndig ver&auml;ndernden - dynamischen Werte der KMxxx Services. Der Wert muss gr&ouml;&szlig;er gleich >=20s sein um dem Modul gen&uuml;gend Zeit einzur&auml;umen eine volle Abfrage auszuf&uuml;hren bevor die n&auml;chste Abfrage startet.<BR>Der Default-Wert ist 300s.                                                                   <BR></li></ul></ul></td></tr>
		<tr><td><ul><ul><a name="PollingTimeout" ></a><li><b><u><code>PollingTimeout </code></u></b> : Ein g&uuml;ltiger Zeitwert um dem KMxxx gen&uuml;gend Zeit zur Antwort einzelner Werte einzur&auml;umen. Normalerweise braucht dieser Wert nicht ver&auml;ndert werden, muss jedoch im Falle eines langsamen Netzwerks erh&ouml;ht werden<BR>Der Default-Wert ist 5s.	                                                                                                                                  <BR></li></ul></ul></td></tr>
		<tr><td><ul><ul><a name="DoNotPoll"      ></a><li><b><u><code>DoNotPoll      </code></u></b> : Eine durch Leerzeichen (Blank) getrennte Liste von Services welche von der Abfrage aufgrund irrelevanter Werte oder fhem - Abst&uuml;rzen ausgenommen werden sollen.<BR>Die Liste kann auch Hierarchien von services enthalten. Dies bedeutet, das alle Services unterhalb dieses Services ebenfalls gel&ouml;scht werden.<BR>Der Default Wert ist (empty) somit werden alle bekannten Services abgefragt. <BR></li></ul></ul></td></tr>
		<tr><td><ul><ul><a name="ReadBackDelay"  ></a><li><b><u><code>ReadBackDelay  </code></u></b> : Ein g&uuml;ltiger Zeitwert in Mllisekunden [ms] f&uuml;r die Pause zwischen schreiben und zur&uuml;cklesen des Wertes durch den "set" - Befehl. Der Wert muss >=0ms sein.<BR>Der  Default-Wert ist 100 = 100ms = 0,1s.                                                                                                                                                                                     <BR></li></ul></ul></td></tr>
		<tr><td><ul><ul><a name="ReadingTextLen  ></a><li><b><u><code>ReadingTextLen </code></u></b> : Eine g&uuml;ltige positive Ganzzahl f&uuml;r die maximale L&auml;nge der Readings der Fehlermerldung.<BR>Der Default-Wert ist 0 = deaktiviert = maximale L&auml;nge.                                                                                                                                                                                                                                       <BR></li></ul></ul></td></tr>
		<tr><td><ul><ul><a name="disable"        ></a><li><b><u><code>disable        </code></u></b> : Deaktiviert das Device und l&ouml;scht alle bestehenden Readings.<BR>Der Default-Wert ist 0 = aktiviert.                                                                                                                                                                                                                                                                                                   <BR></li></ul></ul></td></tr>
	</table>
</ul>
=end html_DE

=for :application/json;q=META.json 73_km200.pm
{
  "abstract": "Connects fhem to Buderus KM300, KM200, KM100, KM50<BR>",
  "description": "The Buderus KM200, KM100  or KM50 (hereafter described as KMxxx) is a communication device to establish a connection between the Buderus central heating control unit and the internet.<BR>It has been designed in order to allow the inhabitants accessing their heating system via his Buderus App EasyControl.<BR>Furthermore it allows the maintenance companies to access the central heating control system to read and change settings.<BR>The km200 fhem-module enables read/write access to these parameters.<BR>",
  "x_lang": {
    "de": {
      "abstract": "Verbindet fhem mit Buderus KM300, KM200, KM100, KM50<BR>",
      "description": "Das Buderus KM200, KM100  or KM50 (ab hier als KMxxx beschrieben) ist eine Schnittstelle zwischen der Buderus Zentralheizungssteuerung un dem Internet.<BR>Es wurde entwickelt um den Bewohnern den Zugang zu Ihrem Heizungssystem durch die Buderus App EasyControl zu erlauben.<BR>Dar&uuml;ber hinaus erlaubt es nach vorheriger Freigabe dem Heizungs- bzw. Wartungsbetrieb die Heizungsanlage von aussen zu warten und Werte zu ver&auml;ndern.<BR>Das km200 fhem-Modul erlaubt den Lese-/Schreibzugriff dieser Parameter durch fhem.<BR>"
    }
  },
  "author": [
    "I am the maintainer matthias.deeke@deeke.eu"
  ],
  "x_fhem_maintainer": [
    "Sailor"
  ],
  "keywords": [
    "Buderus",
    "Bosch",
	"KM50",
    "KM100",
    "KM200",
    "KM300",
    "communication",
    "network"
  ],
  "prereqs": {
    "runtime": {
      "requires": {
        "FHEM": 5.00918623,
        "FHEM::Meta": 0.001006,
        "HttpUtils": 0,
        "JSON": 0,
        "perl": 5.014
      },
      "recommends": {
      },
      "suggests": {
      }
    }
  },
  "resources": {
    "x_support_community": {
		"rss": "https://forum.fhem.de/index.php/topic,25540.msg",
		"web": "https://forum.fhem.de/index.php/topic,25540.msg",
		"subCommunity" : {
			"rss" : "https://forum.fhem.de/index.php/topic,25540.msg",
			"title" : "This sub-board will be first contact point",
			"web" : "https://forum.fhem.de/index.php/topic,25540.msg"
		}
    },
	"x_wiki" : {
		"title" : "FHEM Wiki: Buderus Web Gateway",
		"web" : "https://wiki.fhem.de/wiki/Buderus_Web_Gateway"
	}
  },
  "x_support_status": "supported"
}
=end :application/json;q=META.json

=cut